const int MAX_CHANNELS = 16;

/*
	Class structure for in memory data file
	CData		- memory representation of data file w/ Channels[] Segments[] SelLists[]
				  m_bDataUpdated, m_bListsUpdated, m_bCursorsUpdated, m_bFitCurvesUpdated, m_bIdlUpdated, m_bConfigUpdated
				  m_cFileName, m_iStatus, m_dSampling, 
				  m_aSegments, m_aChannels, m_pDataFile, m_aLists
	 CChannel   - m_cUnits, m_cName, m_bVisible, m_bBaseline, m_bFilter, m_dFilterKhz, 
				  m_bAutoScale, m_cr, m_dScaling, m_dLowScale, m_dHighScale, m_dMin, 
				  m_dMax : channel configuration parameters 				
	 CSegment	- raw data for 1 segment, all channels and array of CChanSegDisp's
	  CChanSegDisp- decimated data for 1 segment for 1 channel 
	  CIdlList	- m_iOffset, m_iLength, m_iChannel, m_crColor, m_aIdlClasses[CIdlClass],
					m_aIdlDwells[CIdlDwell] : Idealization array of classes and dwells
	   CIdlClass- { m_dAmp, m_dStdDev } : one amplitude for idealization.
	   CIdlDwell- { m_iClass, m_iDuration } : One class,duration pair for IDL 
	 CSelList	- { m_cName, m_aItems[CSelListItem] } : Selection list 
	  CSelListItem - { m_cLabel, m_cNote, m_iStartSegment, m_iStartPoint, m_iEndSegment,
					m_iEndPoint}
	  
	CSelectionSet-Array of data sections ( for algorithms to process ) 
*/

const int DIGOUTLINES = 16;

//-- structure for parameters passed to data view when constructing a new data file 
class COptionsNewData { 
	public : 
		CString m_cFileName;
		double m_dScaling;
		double m_dSampling;
		int m_iPointsPerTrace;
		int m_iTracesPerPage;
		int m_iChannelCount;
	};

//--------------------------------------------- List of events and stimulii
const int EVTLST_ALLOCSIZE=1000;
class CEventList { 
	public :
		int m_iEvents;
		int m_iEventsAlloc;		// size of events arrays m_aiClass, m_aiDuration, m_aiStimIndex
		short * m_aiClass;		// class for each event
		int * m_aiDuration;		// duration for each event
		unsigned short * m_aiStimIndex;	// m_afStimulii index for each event 
		
		int m_iStimChannels;	// number of stimulus channels
		int * m_aiStimChans;	// list of stimulus channels
		
		float * m_afStimulii;	// array of sets of {m_iStimChan} stimulii levels.
		float * m_afBuffer;		// one stimulii level set for temporary comparing
		int m_iStimAlloc;		// size of m_afStimulii in floats, not sets.
		int m_iStimCount;		// Number of used m_afStimulii sets.
		
		CEventList( );
		~CEventList();
		
		// not in constructor - filled by called function
		bool Allocate( int iEvents, bool bStim );	
		bool AllocateStimChan( int iStimChans );	
		bool AddStimChan( int iCh );
		int FindOrAddStim( );	// TODO - sorted list ? 
		bool AddEvent( int iDuration, short iClass );
		bool CombineEvent( int iDuration, short iClass );
	};


//--------------------------------------------- Selection Lists
class CSelListItem : public CItem { 
	public : 
		CString m_cLabel;
		CString m_cNote;
		int m_iStartSegment;
		int m_iStartPoint;
		int m_iEndSegment;
		int m_iEndPoint;
		void Set( int iStartSegment, int iStartPoint, int iEndSegment, int iEndPoint );
		void Text( char * pch );
	};

class CSelList : public CItem { 
	public : 
		CSelList( const char * pchName=NULL );
		CString m_cName;
		CItemArray m_aItems;
		CSelListItem * Item( int iItem );		
		// add to list in correct position.  result = position 
		int AddListItem( int iStartSegment, int iStartPoint, int iEndSegment, int iEndPoint, const char * pchLabel=NULL, const char * pchNote=NULL );
	};

//--------------------------------------------- 
class CIdlClass : public CItem { 
	public : 
		double m_dAmp;				// idealization amplitude ( from model ) 
		double m_dStdDev;			// idealization standard deviation 
	};

class CIdlDwell : public CItem { 
	public : 
		short m_iClass;				// Class of dwell, used as index to amplitude 
		int m_iDuration;			// duration of dwell in points.
		short m_iState;				// reference from simulated data 
	};

class CIdlList : public CItem {
	public : 
		int m_iOffset;				// point within segment where this IDL starts 
		int m_iLength;				// calculated length of segment 
		int m_iChannel;				// Channel number that was idealized 
		COLORREF m_crColor;			// Display color for idealization

		CIdlList();
		
		CItemArray m_aIdlClasses;	// array of idealized classes 
		CIdlClass * IdlClass( int iClass );	
		void AllocateClasses( int iClasses );
				
		CItemArray m_aIdlDwells;	// array of dwells 
		CIdlDwell * IdlDwell( int iDwell );
		void AllocateDwells( int iDwells );
		void AddDwell( int iClass, int iDuration, int iState=-1 );
	};

//--------------------------------------------- Digital output events 
class CDigoutEvent : public CItem { 
	public : 
		int m_iSampleStart;
		int m_iSampleStop;
		int m_iLine;
		CDigoutEvent( int iLine=0, int iStart=0 );
	};

class CDigoutEventList { 
	protected : 
		CItemArray m_aEvents;
		int m_aiOn[32];						// last added event to turn on each line
	public : 
		unsigned int m_iActiveMask;		// bits flag for which digital lines are present
		unsigned int m_iAutoResetMask;	// bits flag for which channels are Auto Reset
		int m_iActiveLines;				// number of bits set in m_iActiveMask
		void SetMasks( int iActive, int iAutoReset );
			
		CDigoutEventList();
		CDigoutEvent * Event( int iEvent );
		int EventCount();
		void AddEvent( int iLine, int iStart );
		void StopEvent( int iLine, int iStop );
		
		void LoadArray( int iLen, unsigned int * aiData );
	};

//--------------------------------------------- Attributes of acquisition Channel 
class CChannel : public CItem { 
	public : 
		CChannel( );
		const CChannel & operator=( const CChannel & oSource );
	
		//-- attributes from the DataView config dialogue 
		CString m_cUnits;
		CString m_cName;
		bool m_bVisible;
		bool m_bBaseline;
		bool m_bFilter;
		double m_dFilterKhz;
		bool m_bAutoScale;
		COLORREF m_cr;				// display color of channel 

		double m_dChScaling;		// scaling for int's read from .qdf files
									// This is the value read from an existing .qdf file, 
									// or calculated at the creation pt. for new files.
		
		//-- calculated attributes 
		double m_dLowScale;			// visible portion of channel - as set by manual scaling
		double m_dHighScale;		
		double m_dMin, m_dMax;		// max /min of loaded portion of data (used for autoscale)
		double VisibleMax();
		double VisibleMin();

		CString ToString();		// combined Visible,baseline,filter&khz,autoscale,cr settings
		void FromString( const char * pchBuf );
	};

	
//-- count number of visible channels in array 
int VisibleChannels( CItemArray * paChannels );

//------------------------ ChannelSegmentDisplay - 1 channel, 1 segment, decimated for display
// contains array of low and high pixel height within bitmap based on current scaling
class CChanSegDisp : public CItem { 
	private : 
		short * m_piDispLow;		// Array of display data ( decimated m_pdData ) - low and ...
		short * m_piDispHigh;		// ... high pixel in bitmap to set.
		short m_iDispBase;			// location of baseline
		int m_iDispLength;			// size of arrays, i.e. pixel width of segment 

		double m_dMin,m_dMax;		// min , max of analog values
		double m_dRange;			// max-min - retained for efficiency
					
	public : 
		short YPixel( double dVal ) ;
		
		double AbsMax();			// max(abs(data))  Caution : ensure dMin,dMax are previously calculated.
	
		int m_iDispHeight;			// height of trace decimated to.

		void DeleteDisp( bool bInit=false );

		bool AllocateDisp( int iLen, bool bClear=false );
			
		//-- constructor of empty data block
		CChanSegDisp( );
			
		//-- destructor 
		~CChanSegDisp();

		//-- Convert loaded data to display block.  
		// pdData = array of doubles		iSamples = count of samples 
		// iSkip = skip count between samples (1 for non-interleaved, iChannels otherwise )
		// dPixelsPerSample - display factor for decimation 
		bool CreateDisp( double * pdData, int iDispLength, int iSamples, int iSkip, double dPixelsPerSample, int iHeight, CChannel * pCh );

		bool Render( HDC dc, int iBtm, int iOffset, int iWid, COLORREF cr, bool bBaseline );
		bool RenderDispIdl( HDC dc, CIdlList * pIdl, int iBtm, int iOffset, double dPixelsPerSample );
	};

//--------------------------------------------- Attributes of segment of data 
class CSegment : public CItem {
	public : 
		//-- 'Header' info 
		double m_dtStartTime;			// Start date and time of segment
		int m_iNodePos;					// Data Node position in file, for .qdf type files.  
		int m_iPos;						// Data position in file, for .qdf type files.  0 if data has not been written to file yet.
		int m_iChanged;					// 0-unchanged since loaded. 1-data has changed. 
										// TODO : m_iChanged=2 -data has changed length. (grown or shrunk).
		
		//-- data which is filled in by LoadData()
		int m_iChannels;				// 'width' of segment 
		int m_iSamples;					// length of segment;
		double * m_pdData;				// Array of data;  NULL if data has not yet been loaded from disk.

		//-- Load data digout list 
		bool m_bDigout;					// flag whether digital output data is present.
		CDigoutEventList m_oDigoutList;	// List of digital output events 
		
		//-- decimated digout list 
		unsigned int * m_piDigoutDisp;	// array of decimated digital output for display
		int m_iDigoutHeight;			// height of display allocated to digital data per active line 
		int m_iDigoutDispLength;		// size of pixel array decimated to.

		//--		
		bool m_bInterleaved;			// organization of data : true - sample major, false - channel major 

		CItemArray m_aIdl;				// array of idealization results
		CIdlList * IdlList(int iIdlList);
		void DeleteOverlappingIdls( int iStart, int iStop );

		//-- Get event list for specified channel and range.
		//-- call with arrays set to NULL to return event count.  Provide arrays on 2nd call.
		int IDLEventList( int iChannel, int iChannels, 
						int iRanges, int * aiRangeStart, int * aiRangeStop, double dSampling, 
						double dDeadTime, CEventList * pEventList );

		CSegment();
		~CSegment();					// delete m_pData array

		//-- typed access to m_aChanSegDisps
		CItemArray m_aChanSegDisps;	//-- array of display data 
		CChanSegDisp * ChanSegDisp( int iCh );		

		//-- Create buffer for all data.  Delete existing, allocate Samples*Channels doubles.
		bool AllocateData( bool bClear=false );		
		
		//-- Clear member contents.  Delete m_pData and set it to null.
		void DeleteData( );				

		//-- Copy block of data from memory.  
		bool CopyData( double * pData, int iSamples, int iChannels, bool bInterleaved );
		
		//-- Accept block of doubles.  Block will be delete[]'d by this class.
		bool SetData( double * pData, int iSamples, int iChannels, bool bInterleaved );

		//-- Read in samples from file (previously seek'd) to all channels 
		// If arrays are provided, the min and max values are also determined.
		//		todo - try using adMax and adMin options to speed up load process
		bool LoadDataShort( int iSamples, int iChannels, FILE * h, double * adScaling, 
						double dFileScaling, bool bInterleaved=true, double * adMin=NULL, double * adMax=NULL );
		bool LoadDataDouble( int iSamples, int iChannels, FILE * h, 
						bool bInterleaved=true, double * adMin=NULL, double * adMax=NULL );

		//-- Create disp for all visible channels 		
		bool CreateDisp( int iTraceHeight, double dPixPerSamp, bool bOverlap, bool bDigout, CItemArray * paChannels);
		
		//-- fill in m_dMax and m_dMin for all channels
		void SetRanges( CItemArray * paChannels );
		
		//-- Draw disp to device context (bitmap)
		bool Render( HDC dc, bool bOverlap, bool bDispIdl, double dPixPerSamp, int iTop, int iOffset, int iWid, CItemArray * paChannels, COLORREF * aDigoutColors );
		
		//-- Calculate number of traces required to draw whole segment
		int TraceCount( int iSamplesPerTrace );
		
		//-- Calculate average and stddev of a section of data
		bool CalculateSum( int iChStart, int iChStop, int iPtStart, int iPtStop, double * adSum, bool bInitToZero );
		bool CalculateVariance( int iChStart, int iChStop, int iPtStart, int iPtStop, double * adAvg, double * adVar, bool bInitToZero );
	};

//--------------------------------------------- Data view confiuration parameters
class CDataConfigure { 
	public : 
		CItemArray m_aChannels;		// Channel configurations
		bool AddChannel( bool bDefault=true );
		CChannel * Channel( int iChannel );
		int ChannelCount();

		int m_iTracesPerPage;	//-- traces visible
		int m_iPointsPerTrace;	//-- points visible per trace
		bool m_bOverlap;		//-- Whether channels within trace overlap or have divided separate draw areas

		bool m_bDispLists;		//-- Toggle display of lists
		bool m_bDispListLabels;	//-- Toggle display of list labels
		bool m_bDispCursors;	//-- Toggle display of cursors
		bool m_bDispFitCurves;  //-- Toggle display of fit curves
		bool m_bDispIdealizations;	//-- Toggle display of Idealizations
		bool m_bDispDigout;		//-- Toggle display of digital output if segment contains it.
		COLORREF m_acrDigout[DIGOUTLINES];	//-- digital output lines display colors 

		CString ToString( );
		bool FromStringDataConfig( CString * pc );
		CDataConfigure();

		//-- default configuration - static array, so 1 copy for all instances
		void LoadDefaultDataConfig();
		void CopyDefaultDataConfig();	
		void SetDefaultDataConfig();
	};

//--------------------------------------------- Data - collection of channels, data, lists, settings.
class CData { 
	//-- The following flags determine which portions of the file may have changed and need to be saved		
	public : 
		bool m_bDataUpdated;
		bool m_bListsUpdated;
		bool m_bCursorsUpdated;
		bool m_bFitCurvesUpdated;
		bool m_bIdlUpdated;
		bool m_bConfigUpdated;

		void SetDataUpdated(){ m_bDataUpdated=true; }
		void SetListsUpdated(){ m_bListsUpdated=true; }
		void SetCursorsUpdated(){ m_bCursorsUpdated=true; }
		void SetFitCurvesUpdated(){ m_bFitCurvesUpdated=true; }
		void SetIdlUpdated(){ m_bIdlUpdated=true; }
		void SetConfigUpdated(){ m_bConfigUpdated=true; }
		
		void ClearUpdated();

	//----- Interface for algorithms.		
	//-- status is used to pseudo lock the file when in use.  
	// 0=free 1=locked reading 2=locked&changing
	// if status is 1, processes should be able to short term lock the mutex 
	// and grab some data as long as no changes are made.  If status is 2 the 
	// data is undergoing changes and process should wait for access.
	//
	// TODO - check status before various operations in VwData.
	private :
		int m_iStatus;				
	public : 
		bool SetStatus( int iStatus );
		int GetStatus();		

	public : 
		CData();
		~CData();

		CDataConfigure m_oConfig;

		bool OpenQDF( const char * pchFileName );
		bool WriteQDF( const char * pchFileName );

		double m_dSampling;
		double m_dScaling;			// For .qdf files, A/D scaling factor

		CSegment * Segment( int iSeg );
		CItemArray m_aSegments;		// segments of data ( count is Channels * Segments ) 

		CDataFileInterface * m_pDataFile;	// ptr to file handler (parent class type)
		
		CSelList * List( int iList );		
		CSelList * List( const char * pchListName );		
		CItemArray m_aLists;		// array of lists 

		bool PointToSegOffset( int iPoint, int * piSegment, int * piOffset );

		bool CalculateStatistics( int iChStart, int iChStop, int iSegStart, int iPtStart, 
				int iSegStop, int iPtStop, int * piPoints, double * adAvg, double * adStd );
		
		bool GetEventList( int iChannel, int iDataSelect, int * aiCurrent, int iList, 
				bool bJoined, double dDeadTime, 
				CEventList * pEventList, double * pdSample );
	};

//--SelectionSet is a set of data ranges to operate on for algorithms.
// Setup() and GetNext() are used to iterate through it. 
// Data is copied to a local buffer for each segment being analyzed.
class CSelectionSetRange : public CItem { 
	public : 
		int m_iSeg;				// data segment extracting from 
		int m_iMin;				// index to first point
		int m_iMax;				// index to last point in set
		double m_dtStart;		// reference time of start 
		double m_dtStart0;		// reference time of start 
		int Size();
	};
	
class CSelectionSet  { 
	private :
		int m_iChannel;
		int m_iCurrent;
		int m_iMaxSize;
		int m_iCurrentSize;
		void AllocateBuffer( int iChannels );
		void SetupHelper( int iSeg0, int iOffset0, int iSeg1, int iOffset1 );
		
	public :
		CItemArray m_aRanges;
		CSelectionSetRange * Range( int iRange );

		double * m_adData;			// buffer for data loading
		double Datum( int iSample );
		double Datum( int iChannel, int iSample );
		
		short * m_aiClass;			// buffer for idealization class / -1 if none 
		short * m_aiState;			// buffer for idealization state / -1 if none
		int m_iAllocatedIDL;
		
		CData * m_pData;			// ptr to CData object
		CSelectionSet( );
		~CSelectionSet();
		
		bool Setup( CData * pData, int iChannel );		// Whole File 
		bool Setup( CData * pData, int iChannel, int iList );
		bool Setup( CData * pData, int iChannel, const char * pchList );
		bool Setup( CData * pData, int iChannel, int iSeg0, int iOffset0, int iSeg1, int iOffset1 );
		bool Setup( HWND hdlgData, int iMode, CData * pData, int iChannel );

		//--- calculate min / max value in range / whole set 
		bool RangeMinMax( int iRange, int iChannel, bool bInitialize, double * pdMin, double * pdMax );
		bool CalculateMinMax( int iChannel, double * pdMin, double * pdMax );
		
		bool GetRange( int iRange );
		bool GetFirst();
		bool GetNext();
		int RangeIndex();			// Index of current range within list, i.e. 0..n-1

		int CurrentSize();			// Size of current SelectionSetRange
		bool GetIdealized();
		int MaxSize();				// maximum of all range sizes 
		int TotalSize();			// Total of all range sizes
		CSelectionSetRange * m_pRange;
	};


