/*

CCombState
CStimulus{CStimuli}
CRateList{CRateListItem}
CModel{CModelClass,CModelState,CModelRate,CModelConstraint}
CMetaModel{CMetaModelState,CMetaModelPath}

*/

using namespace TNT;

const int CONSTRAINT_ERROR = -1;
const int CONSTRAINT_FIXRATE = 1;
const int CONSTRAINT_FIXEXP = 2;
const int CONSTRAINT_SCALERATE = 3;
const int CONSTRAINT_SCALEEXP = 4;
const int CONSTRAINT_LOOPBALANCE = 5;
const int CONSTRAINT_LOOPIMBALANCE = 6;

const int MAXCLASSES=60;

COLORREF DefaultClassColor( int iClass );

void CalcArrowPoint( POINT pt0, POINT pt1, POINT * pA, bool bNeg=false );

//-----------------------------------------------------------------------------
// MetaModel helper class used to iterate through all combined states for a
// specified number of channels and states. The results (MetaStates) must be 
// copied from m_ai between iterations.
class CCombState {	
	protected : 
		int t,h;
	public : 
		int m_iChannels, m_iStates;			
		int * m_ai;							// MetaState - number of channels in each state 
		CCombState( int iChannels, int iStates );	// Constructor - start all channels in state 0.
		bool NextCombination( );			// move 1 channel to a new state. Returns true until all combinations are exhausted.
		~CCombState();
	};
	
//-----------------------------------------------------------------------------
// CStimulus - a single name/value or name/expression which represents a pressure, 
//   voltage, agonist, or other variable.  Pre-exponential and exponential.
// CStimuli - a collection of CStimulus.  
class CStimulus : public CItem { 
	public : 
		CStimulus( const char * pchName, double dValue );
		CString m_cName;				// Stimulus Name for lookup 
		double m_dValue;				// Current Value ( or fixed value if expression is empty ) 
		int m_iChannel;					// when used as algorithm setup - data channel index in file.		
		// oExpression					// todo - time dependent variation 
		double Value(double t) { return m_dValue; }		
	};
class CStimuli { 
	public : 
		CItemArray m_aStim;				// Array of CStimulus
		void AddStimulus( const char * pchName, double dValue );	// add a fixed stimulus
		double StimulusValue( const char * pchName, double t, double dNotFoundResult=1.0 );
														// get value of stimulis at offset time t.
		int StimulusIndex( const char * pchName );	// get index of stimulus in list. -1 if d.n.e.
		CStimulus * Stimulus( const char * pchName );	// get stimulus reference.  NULL if d.n.e.
	};
	
/*	
//-----------------------------------------------------------------------------
// List of rate endpoints.  CRateList is a collection of CRateListItem.
// Allows search for rate or reverse of rate.
// Used by ApplyConstraints, but may useful elsewhere.
class CRateListItem : public CItem {		
	public : 
		int m_aiState[2];
		CRateListItem( int iState0, int iState1 );
	};	
class CRateList {
	public : 
		CItemArray m_aRows;
		bool HasRate( int iState0, int iState1 );	// check for rate 0->1 or 1->0 
		void AddRate( int iState0, int iState1 );	// add to list 
	};
*/	

//-----------------------------------------------------------------------------
// Model Class : {color, amp, stddev, conditioning}.
class CModelClass : public CItem { 
	public : 
		COLORREF m_cr;				// Display color ( in specific model, not global)
		double m_dAmp;				// current amplitude
		double m_dStdDev;			// standard deviation of current 
		double m_dConditioning;		// TODO - currently unused
		CModelClass();
		CModelClass( double dAmp, double dStd, double dCond=1.0, COLORREF cr=RGB(0,0,0) );
	};

//-----------------------------------------------------------------------------
// Model State : {class, start probability, x, y, EnergyLandscape values}. 
class CModelState : public CItem { 
	public : 
		//CModelState();
		CModelState( int iClass=0, double dX=0.0, double dY=0.0, double dStartProb=0.0 );

		double m_dStartProb;		// likelihood of starting in this state (scaled)
		int m_iClass;				// class this state belongs to
		int m_iGroup;				// TODO - currently unused
		double m_dXPos, m_dYPos;	// position within [0..100] 

		//-- Energy landscape	
		double m_dEnergyMin;		// energy level for landscape (w.r.t. stimulus) minimum...
		double m_dEnergyMax;		// ... and maximum 
		double Y( double dFrac );
	};

//-----------------------------------------------------------------------------
// Model Rate - One combined fwd/bwd rate in model 
// { states(endpoints), k0, k1, P, Q, EnergyLandscape values, Rate Limits }
class CModelRate : public CItem {
	public : 
		int m_aiStates[2];			// to and from state indices
		double m_adK0[2];			// k0  in { K = k0 * P * e^(k1 * q) }
		double m_adK1[2];			// k1
		bool m_abPDep[2];			// true if P (ligand) dependency
		CString m_acPName[2];		// P (ligand) dependent name
		bool m_abQDep[2];			// true if Q (voltage/pressure) dependency
		CString m_acQName[2];		// Q (voltage/pressure) dependent name
		bool m_abConstrainedK0[2];	// true if a constraint controls this k0 for Mil - numeric
		bool m_abConstrainedK1[2];	// true if a constraint controls this k1 for Mil - numeric

		double RateValueK0( bool bFwd, CStimuli * poStim=NULL, double t=0.0 );
									// Calculate P * k0 
		double RateValueK1( bool bFwd, CStimuli * poStim=NULL, double t=0.0 );
									// Calculate Q * k1 
		double RateValue( bool bFwd, CStimuli * poStim=NULL, double t=0.0 ); 
									// Calculate = k0 * P * exp( Q * k1 )

		CModelRate( CModelRate * pRef );
		CModelRate( int iSt0=0, int iSt1=1, double dK0Fd=100.0, double dK0Bk=100.0 );
		int IsHit( int iX, int iY, POINT pt0, POINT pt1, int iTextHgt, int iStateRadius );
		//double Rate( bool bFwd, CStimuli * poStim, double t );

		//-- Energy landscape	
		double m_dEnergyMin, m_dEnergyMax;	// energy level for landscape (w.r.t. stimulus)
		double Y( double dFrac );

		//-- MIL	
		double m_adK0Limit[2][2];	// Search limits for Mil DFP, (to be enforced in ztox?)
		double m_adK1Limit[2][2];	// Original k0 and k1 mult and div by dFactor. (could just store original instead).
		void SetLimits( double dFactor );
	};

//-----------------------------------------------------------------------------
// Model Constraint - constraint of various types (fixrate, loopbalance, etc. )
class CModelConstraint : public CItem { 
	public : 
		CModelConstraint( int iType, int iStates, int * aiStates=NULL, double dX=50.0, double dY=50.0 );
		CModelConstraint( const CModelConstraint & refConstraint );
		~CModelConstraint();
		void SetStates( int iStates, int i0, int i1, int i2=-1, int i3=-1 );	// shortcut for filling in 2 or 4 states 
		int m_iStates;				// # of states in list 
		int * m_aiStates;			// list of states 
		int m_iType;				// constraint type.  (Consider using an enum)
		double m_dXPos, m_dYPos;	// position within [0..100] 
		void Text( char * pchBuf );
		
		double m_adValue[2];			// working value for scale or loop 
		CModelRate * m_apAdjust[2];	// ptr to rate to adjust for loop balancing
		bool m_abAdjustBkwd[2];			// adjust rate is forward or backward component
		bool m_bError;
		
	};
		
//-----------------------------------------------------------------------------
//CModel - Class structure for in memory data file
class CModel { 
	public : 
		CModel( );
		CModel( const CModel & refModel );
		CModel( CModel * pModel1, CModel * pModel2, int iChannels1, int iChannels2 );
		const CModel & operator=( const CModel & refModel );
		void Copy( const CModel & refModel );

	//-- status is used to pseudo lock the file when in use.  0=free 1=locked reading 2=locked&changing
	// todo - mutex protect a section to check if status can be changed safely
	protected :
		int m_iStatus;				
	
	public : 
		bool SetStatus( int iStatus );
		int GetStatus();		
		
	public :
		//-- collections of classes, states, rates, constraints
		CItemArray m_aClasses;
		CModelClass * Class( int i ) const;
		double StateAmp( int iState );
		double StateStd( int iState );
		void AddClass();
		CModelClass * AddClass( double dAmp, double dStd, double dCond=0.0 );
		void DeleteClass(int iClass);

		CItemArray m_aStates;
		CModelState * State( int i ) const;		
		bool AddState( int iClass, double dXPos, double dZPos, double dProb=0.0 );
		bool DeleteState( int iState );
		double NormalizeStartProbabilities(double * ad=NULL);	
			// set all mi.pr - scale or set to 1/n if sum(fabs())!=1.0
			// if ad!=NULL, copy probabilities to this array instead of updating internal probs.
			
		CItemArray m_aRates;
		CModelRate * Rate( int i ) const;
		bool AddRate( int iState0, int iState1 );
		bool DeleteRate( int iRate );
		CModelRate * FindRate( int iState0, int iState1, bool * pbBwd=NULL );	// NULL if DNE
		int FindRateIndex( int iState0, int iState1, bool * pbBwd=NULL );	// -1
		bool HasPreExpDepend();				// Model has any pre-exponential dependency ? (P)
		bool HasExpDepend();				// Model has any exponential dependency ? (Q)
		void GetStimuliiNames( CStringArray & aNames ); // Fill array with all P and Q names used in model
		void SetRateLimits( double dLimitFactor );	// Call SetLimit(dLimitFactor) for every rate.

		CItemArray m_aConstraints;
		CModelConstraint * Constraint( int i ) const;
		bool AddConstraint( int iType, int iStates, int * aiStates, double dX=50.0, double dY=50.0 );
		bool DeleteConstraint( int iConstraint );
		int ConstraintStatus( int iConstraint );	//-2:non-loop constraint -1:loop not valid 0:loop,no exp rates int 1:loop containing exp rates
		void BuildConstraintsMtxVct( Array2D<double> & Mtx, Array1D<double> & Vct, int * piConstraintRows );	
		int ConstrainedRate( int iState0, int iState1 );	//0=no 1=fwd:fixed 2=fwd:scaled -1=bwd:fixed -2=bwd:scaled
		bool ConstrainedLoopRate( bool bImbal, int iState0, int iState1 );	// true if part of loop, fwd or bwd
		bool SetConstraintStatus();
		bool ApplyConstraints();
		
		bool UsesClass( int iClass );	// true if model has a state of class iClass
		int UsedClassCount(Array1D<int> * paiClasses=NULL);	 // Number of classes for which UsesClass is true.  Store used classes in vector if provided.
		
		Array2D<double> Q( CStimuli * poStim=NULL, double t=0.0 );			// create the Q matrix 

		//-- Free parameters from constraints
		int FreePar( int iConstraintRows, Array1D<double> & aX, Array2D<double> & aMtx, Array1D<double> & aVct, Array1D<double> & aZ );

		//-- Fill in provided arrays with amp,std for classes used in model. Return count of classes filled in.
		int GetAmpStdArray( int iSize, int * aiClass, double * adAmp, double *adStd, bool bSort );

		//-- New (simple closed-open) model 
		void ClearModel();
		void NewModel( );
		
		//-- disk file interface
		bool OpenQmf( const char * pchFileName );
		bool WriteQmf( const char * pchFileName );
		
	};

//-----------------------------------------------------------------------------
class CMetaModelState : public CModelState { 
	public : 
		int * m_aiStates;				// array of channel counts per state
		CMetaModelState( int iStates, int * aiStates, double dStartProb );	// construct with state channel counts
		~CMetaModelState();
	};
	
class CMetaModelPath : public CItem /*CModelRate*/ { 
	public :
		int m_aiStates[2];					// to and from metastate indices
		int m_iOriginalIndex;				// index of rate in original model, for derivative calcs
		bool m_bOriginalBwd;				// true = original is back rate 
		CModelRate * m_pOriginalRate;		// ptr to rate in original model 
		CMetaModelPath( int i, int j, int iRateIndex, CModelRate * pRate, bool bBwd );

		double RateValueK0( bool bFwd, CStimuli * poStim=NULL, double t=0.0 );
		double RateValueK1( bool bFwd, CStimuli * poStim=NULL, double t=0.0 );
		double RateValue( bool bFwd, CStimuli * poStim=NULL, double t=0.0 ); 
	};		

class CMetaModel : public CModel {
	//----- todo : MetaMarkov creates drug[model.iRates*2] and volt[] arrays - rqd ? 
	protected : 
		//-- find the extreme (min or max depending on open direction) conductance 
		//-- of state where class=-1(unassigned). Used by constructor to combine classes.
		int FindExtremeAmp( bool bPositive, double * adMetaAmp, int iAmps ); 
		//-- set all amps within .00001 of specified amp to specified class
		void SetExtremeAmp( double dAmp, double dStd, int iMetaStates, double * adMetaAmp, double * adMetaStd, double dTol=0.00001 );
		//-- calculate sum(abs(dif(channels))) for 2 metastates.  If it is 1+1, they are connected
		int MetaDistance( int i, int j, int iStates );
	public : 
		CMetaModel( );
		CMetaModel( CModel * pModel, int iChannels, bool bStd );
		void Setup( CModel * pModel, int iChannels, bool bStd );
		void SetupClassStateIndex( int iUsedClasses, int iStates ) ;
		
		CItemArray m_aPaths;
		CMetaModelPath * MetaPath( int i );
		void FillPathArray( Array2D<int> & ai1Paths, CStimuli * aoStim=NULL, double t=0.0  );

		CMetaModelState * MetaState( int i );		

		//----- Index[iMetaClasses][iMetaStates*2] is array where first nGroup[i] elements of 
		//----- index[i] are metastates of metaclass i and (iMetaState-nGroup[i]) elements which
		//----- start at index[i][iMetaState] are metastates not of meta class i.
		//----- Ex : (2 of 7 states have class 4) Index[4] = { 2,4,0,0,0,0,0,1,3,5,6,7,0,0 }
		Array1D<int> m_aiClassStateCount;
		Array2D<int> m_aiClassStateIndex;

		//-- if metastates i and j are connected, return true and set piFrom, piTo to original model state indices.
		bool MetaConnect( int i, int j, int iStates, int * piFrom, int * piTo );

		//-- 
		Array2D<double> QQ( CStimuli * poStim=NULL, double t=0.0, Array2D<double> * padQQ_q=NULL );

		double RateValueK0( CMetaModelPath * pMetaPath, bool bFwd, CStimuli * poStim=NULL, double t=0.0, int * piChannels=NULL );
		double RateValueK1( CMetaModelPath * pMetaPath, bool bFwd, CStimuli * poStim=NULL, double t=0.0 );
		double RateValue( CMetaModelPath * pMetaPath, bool bFwd, CStimuli * poStim=NULL, double t=0.0, int * piChannels=NULL ); 
		bool MakeRate( int iPath, CModelRate * pRate );

	};

