unit RotLabel;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses

RotRect, Math, Windows, Controls, StdCtrls, Graphics, Classes;


type

TRotLabel = class
protected

  prAutoFlip            :Boolean;
  prAutoSize            :Boolean;
  prCaption             :String;
  prColor               :TColor;
  prFontHeight          :Integer;
  prFontWeight          :Integer;
  prMinWidth            :Integer;
  prMinHeight           :Integer;
  prMaxWidth            :Integer;
  prMaxHeight           :Integer;
  prVisible             :Boolean;
  prRect                :TGrowRotRect;

  procedure prDoAutoSize;

  function prGetAnchors: TAnchors;
  procedure prSetAnchors(x :TAnchors);
  procedure prSetAutoSize(x :Boolean);
  procedure prSetCaption(x :String);
  procedure prSetMinWidth(x :Integer);
  procedure prSetMinHeight(x :Integer);
  procedure prSetMaxWidth(x :Integer);
  procedure prSetMaxHeight(x :Integer);

public

  OnMouseDown   :procedure (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer) of Object;
  OnMouseMove   :procedure (Sender: TObject; Shift: TShiftState; X, Y: Integer) of Object;
  OnMouseUp     :procedure (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer) of Object;
  OnMouseWheel  :procedure (Sender :TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled :Boolean) of Object;
  OnDblClick    :TNotifyEvent;

  constructor Create; virtual;
  destructor Destroy; override;

  procedure SetBounds(Left, Top, Width, Height :Integer);
  procedure Rotate(X0, Y0 :Integer; Rads :Double); // r-h rule

  procedure PaintTo(NewCanvas :TCanvas);
  procedure DoMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
  procedure DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
  procedure DoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
  procedure DoMouseWheel(Sender :TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled :Boolean);
  procedure DoDblClick(Sender: TObject; X, Y: Integer; var Handled :Boolean);

  property Anchors    :TAnchors      read prGetAnchors    write prSetAnchors;
  property AutoFlip   :Boolean       read prAutoFlip      write prAutoFlip;
  property AutoSize   :Boolean       read prAutoSize      write prSetAutoSize;
  property Caption    :String        read prCaption       write prSetCaption;
  property Color      :TColor        read prColor         write prColor;
  property FontHeight :Integer       read prFontHeight    write prFontHeight;
  property FontWeight :Integer       read prFontWeight    write prFontWeight;
  property MinWidth   :Integer       read prMinWidth      write prSetMinWidth;
  property MinHeight  :Integer       read prMinHeight     write prSetMinHeight;
  property MaxWidth   :Integer       read prMaxWidth      write prSetMaxWidth;
  property MaxHeight  :Integer       read prMaxHeight     write prSetMaxHeight;
  property Rect       :TGrowRotRect  read prRect;
  property Visible    :Boolean       read prVisible       write prVisible;

end;


implementation


constructor TRotLabel.Create;
begin
  prAutoFlip    := True;
  prAutoSize    := True;
  prCaption     := '';
  prColor       := clBlack;
  prFontHeight  := 0;
  prFontWeight  := 400;
  prMinWidth    := 1;
  prMinHeight   := 1;
  prMaxWidth    := 10000;
  prMaxHeight   := 10000;
  prVisible     := True;
  prRect        := TGrowRotRect.Create(0, 0, 100, 20);
end;

destructor TRotLabel.Destroy;
begin
  prRect.Free;
  inherited;
end;

procedure TRotLabel.SetBounds(Left, Top, Width, Height :Integer);
begin
  prRect.SetBounds(Left, Top, Width, Height);
end;

procedure TRotLabel.Rotate(X0, Y0 :Integer; Rads :Double);
begin
  prRect.Rotate(X0, Y0, Rads);
end;

procedure TRotLabel.PaintTo(NewCanvas :TCanvas);
var
  Extents          :TSize;
  StupidFontMgmt   :TFont;
  RestoreFont      :TFont;
  RestoreTextAlign :Integer;
  RestoreBkMode    :Integer;
  VisCaption       :String;
  Escapement       :Double;
  Esc10Deg         :Integer;
  Flipped          :Boolean;
  TL               :TPoint;
begin
  if not Visible then Exit;

  // get escapement
  Escapement       := prRect.BaseAngle;
  // depending on escapement, flip 180, get tl
  if Abs(Escapement) >= (PI/2 - 1e-12) then begin
    Escapement     := Escapement + PI;
    Flipped        := True;
  end else begin
    Flipped        := False;
  end;
  Esc10Deg         := Round(Escapement * (180.0/PI) * 10);

  // setup font
  RestoreFont      := NewCanvas.Font;
  StupidFontMgmt   := TFont.Create;
  NewCanvas.Font   := StupidFontMgmt;
  NewCanvas.Font.Handle := CreateFont(FontHeight, 0, Esc10Deg, Esc10Deg, FontWeight, 0, 0, 0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, DEFAULT_PITCH and FF_DONTCARE, nil);
  NewCanvas.Font.Color  := Color;
  RestoreTextAlign := GetTextAlign(NewCanvas.Handle);
  SetTextAlign(NewCanvas.Handle, TA_TOP or TA_LEFT);
  RestoreBkMode    := GetBkMode(NewCanvas.Handle);
  SetBkMode(NewCanvas.Handle, TRANSPARENT);

  // measure caption
  GetTextExtentPoint32(NewCanvas.Handle, PChar(Caption), Length(Caption), Extents);
  // set width, height if autosize
  if AutoSize then begin
    prRect.Width   := Extents.cx + 4;
    prRect.Height  := Extents.cy + 2;
    prDoAutoSize;
  end;

  // shorten caption
  VisCaption       := Caption;
  while (Length(VisCaption) > 3) and (Extents.cx > prRect.Width) do begin
    if VisCaption = Caption then begin
      VisCaption   := Copy(Caption, 0, Length(Caption)-1) + '...';
    end else begin
      VisCaption   := Copy(VisCaption, 0, Length(VisCaption)-4) + '...';
    end;
    GetTextExtentPoint32(NewCanvas.Handle, PChar(VisCaption), Length(VisCaption), Extents);
  end;

  // draw
  // SetBkMode(TRANSPARENT);
  // SetTextAlign(TA_TOP or TA_LEFT);
  if Flipped then begin
    TL             := prRect.BR;
  end else begin
    TL             := prRect.TL;
  end;
  TextOut(NewCanvas.Handle, TL.x, TL.y, PChar(VisCaption), Length(VisCaption));

  // restore font
  NewCanvas.Font   := RestoreFont;
  StupidFontMgmt.Free;
  SetTextAlign(NewCanvas.Handle, RestoreTextAlign);
  SetBkMode(NewCanvas.Handle, RestoreBkMode);
end;

procedure TRotLabel.DoMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
begin
  if Visible and Assigned(OnMouseDown) and prRect.PtInRect(X, Y) then begin
    OnMouseDown(Self, Button, Shift, X, Y);
    Handled := True;
  end;
end;

procedure TRotLabel.DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
begin
  if Visible and Assigned(OnMouseMove) and prRect.PtInRect(X, Y) then begin
    OnMouseMove(Self, Shift, X, Y);
    Handled := True;
  end;
end;

procedure TRotLabel.DoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
begin
  if Visible and Assigned(OnMouseUp) and prRect.PtInRect(X, Y) then begin
    OnMouseUp(Self, Button, Shift, X, Y);
    Handled := True;
  end;
end;

procedure TRotLabel.DoMouseWheel(Sender :TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled :Boolean);
begin
  if Visible and Assigned(OnMouseWheel) and prRect.PtInRect(MousePos.X, MousePos.Y) then begin
    OnMouseWheel(Self, Shift, WheelDelta, MousePos, Handled);
  end;
end;

procedure TRotLabel.DoDblClick(Sender :TObject; X, Y :Integer; var Handled :Boolean);
begin
  if Visible and Assigned(OnDblClick) and prRect.PtInRect(X, Y) then begin
    OnDblClick(Self);
    Handled := True;
  end;
end;

procedure TRotLabel.prDoAutoSize;
var
  w, h     :Integer;
begin
  if prAutoSize then begin
    w := prRect.Width;
    h := prRect.Height;

    if w < prMinWidth then begin
      prRect.Width := prMinWidth;
    end else if w > prMaxWidth then begin
      prRect.Width := prMaxWidth;
    end;
    if h < prMinHeight then begin
      prRect.Height := prMinHeight;
    end else if h > prMaxHeight then begin
      prRect.Height := prMaxHeight;
    end;
  end;
end;

function TRotLabel.prGetAnchors: TAnchors;
begin
  Result := prRect.Anchors;
end;

procedure TRotLabel.prSetAnchors(x :TAnchors);
begin
  prRect.Anchors  := x;
end;

procedure TRotLabel.prSetAutoSize(x :Boolean);
begin
  prAutoSize := x;
  prDoAutoSize;
end;

procedure TRotLabel.prSetCaption(x :String);
begin
  prCaption := x;
end;

procedure TRotLabel.prSetMinWidth(x :Integer);
begin
  prMinWidth := x;
  prDoAutoSize;
end;

procedure TRotLabel.prSetMinHeight(x :Integer);
begin
  prMinHeight := x;
  prDoAutoSize;
end;

procedure TRotLabel.prSetMaxWidth(x :Integer);
begin
  prMaxWidth := x;
  prDoAutoSize;
end;

procedure TRotLabel.prSetMaxHeight(x :Integer);
begin
  prMaxHeight := x;
  prDoAutoSize;
end;


end.
 