unit RotRect;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses

Math, Windows, Controls;


type


TRotRect = class
public

  TL        :TPoint;
  TR        :TPoint;
  BR        :TPoint;
  BL        :TPoint;

  constructor Create(Left, Top, Width, Height: Integer); virtual; //overload;
  //constructor Create(TopLeft, TopRight, BotLeft, BotRight :TPoint); overload;

  procedure CopyFrom(x :TRotRect); virtual;

  function PtInRect(Pt :TPoint): Boolean; overload;
  function PtInRect(X, Y: Integer): Boolean; overload;

  function BaseAngle: Double; // from bl to br, in rads from horizontal

  procedure SetBounds(Left, Top, Width, Height :Integer);
  procedure Rotate(X0, Y0 :Double; Rads :Double);

  procedure GetBoundingRect(Bounds :TRect);

end;

TGrowRotRect = class(TRotRect)
protected

  function prGetWidth: Integer;
  function prGetHeight: Integer;
  procedure prSetWidth(x :Integer);
  procedure prSetHeight(x :Integer);

public

  Anchors         :TAnchors;

  constructor Create(Top, Left, Width, Height :Integer); override;

  // procedure CopyFrom(x :TRotRect); override; // to copy Anchors as well

  property Width  :Integer      read prGetWidth      write prSetWidth;
  property Height :Integer      read prGetHeight     write prSetHeight;

end;


implementation


function Distance2D(A, B :TPoint): Double;
var
  dx, dy   :Integer;
begin
  dx       := B.x - A.x;
  dy       := B.y - A.y;
  Result   := Sqrt(dx*dx + dy*dy);
end;

function DotProdAngle(Ax, Ay, Bx, By, Cx, Cy :Integer): Integer;
var
  BAx, BAy, BCx, BCy :Integer;
begin
  BAx    := Ax - Bx;
  BAy    := Ay - By;
  BCx    := Cx - Bx;
  BCy    := Cy - By;
  Result := BAx * BCx + BAy * BCy;
end;

function CrossProdLenAngle(Ax, Ay, Bx, By, Cx, Cy :Integer): Integer;
var
  BAx, BAy, BCx, BCy :Integer;
begin
  BAx    := Ax - Bx;
  BAy    := Ay - By;
  BCx    := Cx - Bx;
  BCy    := Cy - By;
  Result := BAx * BCy - BAy * BCx;
end;

function RadiansInAngle(const A, B, C :TPoint): Double;
begin
  Result := ArcTan2( CrossProdLenAngle(A.x, A.y, B.x, B.y, C.x, C.y),
                     DotProdAngle(A.x, A.y, B.x, B.y, C.x, C.y) );
end;

function MidPoint(const A, B :TPoint): TPoint;
begin
  Result.x := (a.x + b.x) div 2;
  Result.y := (a.y + b.y) div 2;
end;

constructor TRotRect.Create(Left, Top, Width, Height: Integer);
begin
  SetBounds(Left, Top, Width, Height);
end;

procedure TRotRect.CopyFrom(x :TRotRect);
begin
  TL  := x.TL;
  TR  := x.TR;
  BR  := x.BR;
  BL  := x.BL;
end;

(*
constructor TRotRect.Create(TopLeft, TopRight, BotLeft, BotRight :TPoint);
begin
  TL  := TopLeft;
  TR  := TopRight;
  BL  := BotLeft;
  BR  := BotRight;
end;
*)
function TRotRect.PtInRect(Pt :TPoint): Boolean;
begin
  Result := PtInRect(Pt.X, Pt.Y);
end;

function TRotRect.PtInRect(X, Y: Integer): Boolean;
// the left-hand-rule goes clockwise
var
  Rads   :Double;
  Pt     :TPoint;
begin
  Result := False;
  Pt.X   := X;
  Pt.Y   := Y;
  Rads   := RadiansInAngle(TR, TL, Pt);
  if (Rads < 0.0) or (Rads > (PI/2)) then Exit;
  Rads   := RadiansInAngle(BL, BR, Pt);
  if (Rads < 0.0) or (Rads > (PI/2)) then Exit;
  Result := True;
end;

function TRotRect.BaseAngle: Double;
begin
  Result := - ArcTan2(BR.y - BL.y, BR.x - BL.x);
end;

procedure TRotRect.SetBounds(Left, Top, Width, Height :Integer);
begin
  TL.X  := Left;             TL.Y  := Top;
  TR.X  := Left + Width;     TR.Y  := Top;
  BL.X  := Left;             BL.Y  := Top + Height;
  BR.X  := TR.X;             BR.Y  := BL.Y;
end;

procedure TRotRect.Rotate(X0, Y0 :Double; Rads :Double);
var
  cosa, sina  :Double;

  function DoRot(P :TPoint): TPoint;
  begin
    Result.X := Round(X0 + ((P.x - X0)*cosa - (P.y - Y0)*sina));
    Result.Y := Round(Y0 + ((P.x - X0)*sina + (P.y - Y0)*cosa));
  end;
begin
  cosa := Cos(- Rads);      // (-) to convert to r-h-rule angles
  sina := Sin(- Rads);

  TL := DoRot(TL);
  TR := DoRot(TR);
  BL := DoRot(BL);
  BR := DoRot(BR);
end;

procedure TRotRect.GetBoundingRect(Bounds :TRect);
  procedure PushBounds(Pt :TPoint);
  begin
    if Pt.x < Bounds.Left     then Bounds.Left     := Pt.x;
    if Pt.x > Bounds.Right    then Bounds.Right    := Pt.x;
    if Pt.y < Bounds.Top      then Bounds.Top      := Pt.y;
    if Pt.y > Bounds.Bottom   then Bounds.Bottom   := Pt.y;
  end;
begin
  Bounds.Left    := TL.x;
  Bounds.Right   := TL.x;
  Bounds.Top     := TL.y;
  Bounds.Bottom  := TL.y;
  PushBounds(TR);
  PushBounds(BR);
  PushBounds(BL);
end;


constructor TGrowRotRect.Create(Top, Left, Width, Height :Integer);
begin
  inherited;
  Anchors := [];
end;

function TGrowRotRect.prGetWidth: Integer;
begin
  Result := Round( Distance2D(BL, BR) );
end;

function TGrowRotRect.prGetHeight: Integer;
begin
  Result := Round( Distance2D(BL, TL) );
end;

procedure TGrowRotRect.prSetWidth(x :Integer);
var
  Scale   :Double;
  Mid     :TPoint;
  dx, dy  :Integer;

  procedure Grow(var A, B :TPoint);
  begin
    dx  := b.x - a.x;
    dy  := b.y - a.y;
    b.x := Round(a.x + dx * scale);
    b.y := Round(a.y + dy * scale);
  end;

  procedure SetSide(var L, R :TPoint);
  begin
    if (not (akLeft in Anchors)) and (not (akRight in Anchors)) then begin
      Mid   := MidPoint(L, R);
      Grow(Mid, L);
      Grow(Mid, R);
    end else if (not (akLeft in Anchors)) then begin
      Grow(R, L);
    end else if (not (akRight in Anchors)) then begin
      Grow(L, R);
    end;
  end;
begin
  Scale := x / Distance2D(TL, TR);
  SetSide(TL, TR);
  SetSide(BL, BR);
end;

procedure TGrowRotRect.prSetHeight(x :Integer);
var
  Scale   :Double;
  Mid     :TPoint;
  dx, dy  :Integer;

  procedure Grow(var A, B :TPoint);
  begin
    dx  := b.x - a.x;
    dy  := b.y - a.y;
    b.x := Round(a.x + dx * scale);
    b.y := Round(a.y + dy * scale);
  end;

  procedure SetSide(var T, B :TPoint);
  begin
    if (not (akTop in Anchors)) and (not (akBottom in Anchors)) then begin
      Mid   := MidPoint(T, B);
      Grow(Mid, T);
      Grow(Mid, B);
    end else if (not (akTop in Anchors)) then begin
      Grow(B, T);
    end else if (not (akBottom in Anchors)) then begin
      Grow(T, B);
    end;
  end;
begin
  Scale := x / Distance2D(TL, BL);
  SetSide(TL, BL);
  SetSide(TR, BR);
end;


end.
 
