{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
  {
  by. John Bannen, 7-2004
  Model energy plot in OpenGL.
      Creates a new form and draws a series of 'ribbons' as parabolic sections
  connecting model states together with barriers for rates.  Image can be
  rotated / manipulated with keystrokes.
      Primary info is the m_aRates array which is displayed by the GLDrawRibbon
  function.  Rest is OpenGL setup, teardown.
      All drawing is done on the [-100,100] x,y,z axes.  X,Y are determined from
  placement of the states on the model window by the user.  Z axis is an energy
  level determined by starting at first state=0, progressing in breadth first
  search using a 'coloring' algorithm and the rates array and setting the levels
  for each state and rate (barrier).
      Wish list (TODO):
       (Fred)
        user configurable ribbon thickness
        Use of shading.
        Perspective view for better 3-d visual interpolation.
       (Gareth/Sean)
        superimpose of 2 models
  }

unit ModelEnergyPlot;

interface

uses
  OpenGL, Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  MATH, StdCtrls, ExtCtrls;

const
  GL_TEXTBASE : Integer = 1000;     // Lowest call list number
  GL_TEXTMIN : Integer = 32;        // Lowest drawable character
  GL_TEXTMAX : Integer = 127;       // Highest drawable character

  TIMERID : Integer = 83;           // Whatever .. ID number for windows timer.

  ROTATEANGLE : Double = 2.0;       // Degrees per rotation

type

  //----------------------------------------------------------------------------
  //----- Connection beween states
  TRibbon = class
  public
    m_nX0, m_nY0 : Double;           // starting x,y
    m_nZ0 : Array[0..1] of Double;   // starting z [With, without stimulus]
    m_nX1, m_nY1 : Double;           // ending x,y
    m_nZ1 : Array[0..1] of Double;   // ending z [With, without stimulus]
    m_nZb : Array[0..1] of Double;   // barrier height
    procedure GLDrawRibbon( nMinZ, nMaxZ, nPct : double; lColor : boolean );
  end;

  TRibbons = class
  private
    m_aRates : array of TRibbon;        // Array of rate 'ribbons'
    m_iCount : Integer;                 // Rates filled in
    m_iAlloc : Integer;              // Allocated size of array
  public
    Function Get(i :Integer) : TRibbon;
    Function GetCount() : Integer;
    Constructor Create();
    procedure AddRibbon(x0,y0,z0,zz0, x1,y1,z1,zz1, zb,zzb : Double);
    procedure Free;
  end;

  //----------------------------------------------------------------------------
  //----- Cube drawn at state positions
  TBlock = class
  public
    m_nX, m_nY : Double;   // center
    m_nZ : Array[0..1] of Double;
    m_nSz : Double;
    m_cName : String;
    m_rgbFaceColor : TColor;
    procedure GLDrawBlock( nPct : Double; lColor : boolean );
  end;

  TBlocks = class
  private
    m_aBlocks : array of TBlock;        // Array of state blocks
    m_iCount : Integer;                // States filled in
    m_iAlloc : Integer;             // Allocated size of array
  public
    Function Get(i :Integer) : TBlock;
    Function GetCount() : Integer;
    Constructor Create();
    procedure AddBlock( x,y,z,zz,nSz:Double; c:String; rgb:TColor );
    procedure Free;
  end;

  //----------------------------------------------------------------------------
  // Bar for 3-d Histogram  // Color scheme for bars needs to be addressed
  TBar = class
  public
    m_nX0, m_nX1, m_nY0, m_nY1, m_nZ0, m_nZ1 : Double;
    procedure GLDrawBar();
  end;

  TBars = class
  private
    m_aBars : array of TBar;            // Array of 3-d blocks
    m_iCount : Integer;
    m_iAlloc : Integer;
  public
    Function Get(i :Integer) : TBar;
    Function GetCount() : Integer;
    Constructor Create();
    procedure AddBar(x0,x1,y0,y1,z0,z1 :Double);
    procedure Free;
  end;

  //----------------------------------------------------------------------------
  //----- Collection of Ribbons and cubes forming an energy plot
  TEnergyPlot = class(TForm)
    sbStimulus: TScrollBar;
    lblTitle: TLabel;
    lblStimulus: TLabel;
    lblCaption1: TLabel;
    lblCaption2: TLabel;
    chkGrid: TCheckBox;
    chkStates: TCheckBox;
    btnReset: TButton;
    btnMovie: TButton;
    btnCopy: TButton;
    chkColor: TCheckBox;
    Button1: TButton;
    chkLegend: TCheckBox;
    btnExport: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormPaint(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure sbStimulusChange(Sender: TObject);
    procedure chkGridClick(Sender: TObject);
    procedure chkStatesClick(Sender: TObject);
    procedure btnResetClick(Sender: TObject);
    procedure btnMovieClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure btnCopyClick(Sender: TObject);
    procedure chkColorClick(Sender: TObject);
    procedure btnRotateClick(Sender: TObject);
    procedure chkLegendClick(Sender: TObject);
    procedure btnExportClick(Sender: TObject);

  private
		m_angleX,m_angleY,m_angleZ:single;  // rotation to viewpoint
    m_nMaxZ, m_nMinZ : double;          // Computed Z Range for scaling purposes.
    m_lTrack : Boolean;                 // Mouse tracking active ?
    m_iTrackX, m_iTrackY : Integer;     // Position of mouse click
    m_lCaption : Boolean;               // Whether to display x,y,z rotation angles
    m_iRotate : Integer;                // Whether rotate is on.  Continuously rotates.
    m_lDrawing : Boolean;
    procedure Rotate();
		procedure Draw;
		procedure CauseDraw(l : Boolean=true);


  //-- Ribbons, blocks, bars - drawing elements
  private
  	m_RC : HGLRC;
    m_aRates : TRibbons;
    m_aStates : TBlocks;
    m_aBars : TBars;

  public
    procedure AddRate(x0,y0,z0,zz0, x1,y1,z1,zz1, zb,zzb : Double );
    procedure AddState(x,y,z,zz, nSz : Double; c :String; rgb:TColor);
    procedure AddBar(x0,x1,y0,y1,z0,z1 :Double);
    procedure Free();
    procedure SetCaption(l : boolean);
    procedure Movie();
    procedure SetStimulusDependency(lDependant : boolean);

  end;

//------------------------------------------------------------------------------
// Single Global Energy Plot form (OpenGL does not support multiple windows well)
procedure CauseUpdate();
var
    g_frmEnergyPlot : TEnergyPlot ;


//------------------------------------------------------------------------------
implementation

{$R *.DFM}

//-- global timer function which triggers drawing in the global window
procedure CauseUpdate();
Begin
  g_frmEnergyPlot.Rotate();
  g_frmEnergyPlot.CauseDraw(false);
end;

procedure TEnergyPlot.Rotate();
Begin
  m_angleX := m_angleX + ROTATEANGLE;
  if m_angleX > 360 Then m_angleX := m_angleX - 360;
end;

//-- click on rotate to start a timer to CauseUpdate
procedure TEnergyPlot.btnRotateClick(Sender: TObject);
Begin
  m_iRotate := 1-m_iRotate;
  if m_iRotate = 0 Then
    KillTimer( WindowHandle, TIMERID )
  else Begin
    SetTimer( WindowHandle, TIMERID, 50, Addr(CauseUpdate) );
  end;
  ActiveControl:=nil;
end;

//-- Cause the Paint message to occur
procedure TEnergyPlot.CauseDraw(l:Boolean=true);
Begin
  if l then
    ActiveControl:=nil;
  if not g_frmEnergyPlot.m_lDrawing Then
    InvalidateRect( WindowHandle, nil, false);
end;

//-- main guts of the paint process
procedure TEnergyPlot.Draw;
var
  i, j : integer;
  n :double;
  c : String;
begin

  m_lDrawing := true;

  //----- Setup
	glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
	glEnable(GL_NORMALIZE);
	glShadeModel(GL_SMOOTH);  //glShadeModel(GL_FLAT);

	//glCullFace(GL_BACK);     // glDisable(GL_CULL_FACE);
	glLoadIdentity;
  //glTranslatef(0.5, 0.0, -2.0);
	glRotatef(m_angleX, 0.0, 1.0, 0.0);
	glRotatef(m_angleY, 1.0, 0.0, 0.0);
	glRotatef(m_angleZ, 0.0, 0.0, 1.0);

  //----- z=0 plane grid
  if chkGrid.State=cbChecked then begin
  	glColor3f(0.75, 0.75, 0.75 );
    for i := -10 to 10 do Begin
      glBegin(GL_LINE_STRIP);
      for j := -10 to 10 do
        glVertex3f( 10.0 * i, 10.0 * j, 0.0);
      glEnd();
    end;
    for j := -10 to 10 do Begin
      glBegin(GL_LINE_STRIP);
      for i := -10 to 10 do
        glVertex3f( 10.0 * i, 10.0 * j, 0.0);
      glEnd();
    end;
  end;

  //----- ribbons connecting states
  for i:=0 to m_aRates.GetCount()-1 do
    m_aRates.Get(i).GlDrawRibbon(m_nMinz,m_nMaxZ, 0.01 * sbStimulus.Position, cbChecked=chkColor.State );

  //----- Cubes as state markers
  if chkStates.State=cbChecked then
    for i:=0 to m_aStates.GetCount()-1 do
      m_aStates.Get(i).GlDrawBlock(0.01 * sbStimulus.Position, cbChecked=chkColor.State );

  //----- 3-d Bars
  for i:=0 to m_aBars.GetCount()-1 do
    m_aBars.Get(i).GlDrawBar();

  //----- Legend
  if (cbChecked=chkLegend.State) and ( cbChecked=chkColor.State) Then Begin
  	glLoadIdentity;           // Un-rotate
    glListBase(GL_TEXTBASE-GL_TEXTMIN);
    glColor3f(0.0,0.0,0.0);

    c := FloatToStr(m_nMinZ); // Minimum (text)
    glPushMatrix();
    glTranslate(110.0, -105.0, 0.0 );
    glScalef(6.0, 6.0, 6.0 );
    glCallLists(Length(c), GL_UNSIGNED_BYTE, PChar(c));
    glPopMatrix();

    c := FloatToStr(m_nMaxZ); // Maximum (text)
    glPushMatrix();
    glTranslate(110.0, 100.0, 0.0 );
    glScalef(6.0, 6.0, 6.0 );
    glCallLists(3, GL_UNSIGNED_BYTE, PChar(c));
    glPopMatrix();

	  glLoadIdentity;           // Color bar at right side
  	glBegin(GL_TRIANGLE_STRIP);
    for i := 0 to 100 do begin
      n := (1.0*i) / 100.0;
      if n < 0.5 Then
      	glColor3f( 0.0, 0.0, 1.0-n*2.0 )
      else
      	glColor3f( n*2.0-1.0, 0.0, 0.0 );
      //glColor3f( min(1.0,8.0*n*n*n), 0.0, min(1.0,8.0*(1.0-n)*(1.0-n)*(1.0-n)) );
    	//glColor3f(max(0.0, 1.0-2.0*n), 1.0-abs(1.0-2.0*n), max(0.0,2.0*n-1.0) );
      glVertex3f(120.0, 200.0*n-100.0, 0.0 );
      glVertex3f(125.0, 200.0*n-100.0, 0.0 );
    end;
  end;

  glEnd();
  glFlush();

  //----- Double buffering flush
	SwapBuffers(wglGetCurrentDC);

  //----- Text on top of image
  canvas.Font.Name := 'Arial';
  canvas.Font.Size := -16;
  lblTitle.Repaint();
  lblStimulus.Repaint();
  chkGrid.Repaint();
  chkStates.Repaint();
  chkColor.Repaint();
  chkLegend.Repaint();
  if m_lCaption Then Begin
    lblCaption2.Caption := Format('x:%6.1f y:%6.1f z:%6.1f',[m_angleX,m_angleY,m_angleZ]);
    lblCaption1.Repaint();
    lblCaption2.Repaint();
  end;

  //ValidateRect( WindowHandle, nil );

  m_lDrawing := false;

end;

{ ------------------------------------------------------------------------
  Draw curve from pt1 to pt2, passing through ptb - barrier.  Assembled
as a series of 4 parabolic sections.  For each section, determine 2 points
RIB_MARGIN units from x,y orthogonal to line (pt1->pt2) in same z plane.
Proceed from those two points along curves parallel to the pt1->ptb->pt2 curve.
------------------------------------------------------------------------ }
procedure TRibbon.GLDrawRibbon( nMinZ, nMaxZ, nPct : double; lColor:boolean );
const
  SEGMENTS_PER : Integer = 32;   // per quarter ribbon
  RIB_MARGIN : double = 0.5;
  RIB_HGT : double = 0.5;

var
  i :Integer;
  d, dx, dy, dxx, dyy, di : double;
  x0a, x0b, y0a, y0b : double ;
  x1a, y1a  : double ;
  xa, ya, xb, yb : double;
  z : double;
  pz : double;
  iPrefix : Integer;    // set to 1 or 1-.5*seg if leading ribbon is needed
  iSuffix : Integer;    // set to 4*seg or 4.5 *seg if trailing ribbon is needed

  Procedure ZColor( z : double );
  var
    pct : double;
  begin
    if lColor then Begin
      pct := ( z-nMinZ ) / (nMaxZ - nMinZ);
      if pct < 0.5 Then
      	glColor3f( 0.0, 0.0, 1.0-pct*2.0 )
      else
      	glColor3f( pct*2.0-1.0, 0.0, 0.0 );
    	//glColor3f( min(1.0,8.0*pct*pct*pct), 0.0, min(1.0,8.0*(1.0-pct)*(1.0-pct)*(1.0-pct)) );
    end else
    	glColor3f( 0.0, 0.0, 0.0 );
  end;

  Procedure PointAB();
  var
     pxa, pxb, pya, pyb : double;
  begin

    //----- connect with rectangular tube
    pxa:=xa;
    pxb:=xb;
    pya:=ya;
    pyb:=yb;

    xa:=xa+dxx;
    ya:=ya+dyy;
    xb:=xb+dxx;
    yb:=yb+dyy;
    glVertex3f( xb,  yb,  z );
    glVertex3f( pxb, pyb, pz-RIB_HGT );
    glVertex3f( xb,  yb,  z -RIB_HGT );
    glVertex3f( pxa, pya, pz-RIB_HGT );
    glVertex3f( xa,  ya,  z -RIB_HGT );
    glVertex3f( pxa, pya, pz);
    glVertex3f( xa,  ya,  z );
    glVertex3f( xb,  yb,  z );

    //----- Flat ribbon
    //glVertex3f( xa, ya, z );
    //glVertex3f( xb, yb, z );
  end;

  Function Z0() : double;
  Begin
    result := m_nZ0[0]*(1-nPct)+m_nZ0[1]*nPct;
  end;
  Function Z1() : double;
  Begin
    result := m_nZ1[0]*(1-nPct)+m_nZ1[1]*nPct;
  end;
  Function Zb() : double;
  Begin
    result := m_nZb[0]*(1-nPct)+m_nZb[1]*nPct;
  end;

begin
  //-- dist from pt0->pt1
  d := max(sqrt((m_nY1-m_nY0)*(m_nY1-m_nY0)+(m_nX1-m_nX0)*(m_nX1-m_nX0)),0.1);

  //-- unit of dist * margin - sep'd into x,y components
  dx := (m_nX1-m_nX0)*RIB_MARGIN/d;
  dy := (m_nY1-m_nY0)*RIB_MARGIN/d;

  //-- pt RIB_MARGIN units 90 deg from pt x0,y0 on pt0->pt1
  x0a := m_nX0 - dy;
  y0a := m_nY0 + dx;

  //-- pt RIB_MARGIN units -90 deg from pt x0,y0 on pt0->pt1
  x0b := m_nX0 + dy;
  y0b := m_nY0 - dx;

  //-- pt RIB_MARGIN units 90 deg from pt x1,y1 on pt0->pt1
  x1a := m_nX1 - dy;
  y1a := m_nY1 + dx;

  //-- pt RIB_MARGIN units -90 deg from pt x1,y1 on pt0->pt1
  //x1b := x1 + dy;
  //y1b := y1 - dx;

  //-- distance to move each segment
  dxx := (x1a-x0a)/4.0/SEGMENTS_PER;
  dyy := (y1a-y0a)/4.0/SEGMENTS_PER;

  //-- Start of curve - 2 points
	glBegin(GL_TRIANGLE_STRIP);
  xa := x0a;
  ya := y0a;
  xb := x0b;
  yb := y0b;
  z := Z0();

  // iPrefix and iSuffix are used for drawing a partial riibon before state0
  // or after state1 to complete the 'well' appearance.
  // could check : if( only rate leaving 0 state ... )
  iPrefix := 1 ;
  iSuffix := SEGMENTS_PER*4;
  if( not lColor ) Then Begin
    // adjust initial position for leading ribbon
    xa := xa - dxx * (SEGMENTS_PER Div 2);
    ya := ya - dyy * (SEGMENTS_PER Div 2);
    xb := xb - dxx * (SEGMENTS_PER Div 2);
    yb := yb - dyy * (SEGMENTS_PER Div 2);
    z := Z0() + (Zb()-Z0()) / 2.0 * 0.25 ;
    // leading ribbon ...
    iPrefix := 1 - SEGMENTS_PER Div 2;
    // trailing ribbon ...
    iSuffix := iSuffix + SEGMENTS_PER Div 2;
  end;

  pz := z;
	ZColor(z);  //glColor3f(1.0, 0.0, 0.0 );
  PointAB();

  //----- Parabola from 0% to 25% toward pt1, 0% - 50% toward barrier
  for i := iPrefix to SEGMENTS_PER do begin
    di := 1.0 * i / SEGMENTS_PER;     // approx. range (0 to 1]
    pz := z;
    z := Z0() + (Zb()-Z0())/2.0 * di * di ;
  	ZColor(z);  //glColor3f( 1.0-di, 0.0+di, 0.0 );
    PointAB();
  end;

  //----- Parabola from 25% to 50% toward pt1, 50%-100% toward barrier
  for i := SEGMENTS_PER+1 to SEGMENTS_PER*2 do begin
    di := 1.0 * (SEGMENTS_PER*2-i) / SEGMENTS_PER ; // approx from (1 to 0]
    pz := z;
    z := Zb() - (Zb()-Z0())/2.0 * di * di;
 	  ZColor(z);  //glColor3f( 0.0, 0.0+di, 1.0-di );
    PointAB();
  end;

  //----- Parabola from 50% to 75% toward pt1, 50% - 25% from barrier
  for i := 2*SEGMENTS_PER+1 to SEGMENTS_PER*3 do begin
    di := 1.0 * (i-2*SEGMENTS_PER) / SEGMENTS_PER;     // approx. range (0 to 1]
    pz := z;
    z := Zb() - (Zb()-Z1())/2.0 * di * di ;
  	ZColor(z);  //glColor3f( 0.0, 0.0+di, 1.0-di );
    PointAB();
  end;

  //----- Parabola from 75% to 100% toward pt1, 25% - 0% from barrier
  for i := 3*SEGMENTS_PER+1 to iSuffix do begin
    di := 1.0 * (SEGMENTS_PER*4-i) / SEGMENTS_PER ; // approx from (1 to 0]
    pz := z;
    z := Z1() + (Zb()-Z1())/2.0 * di * di;
  	ZColor(z);  //glColor3f( 1.0-di, 0.0+di, 0.0 );
    PointAB();
  end;
  glEnd();

end;

//----- One time Setup of OpenGL rendering environment.
procedure GLInit( iWidth, iHeight : Integer );
	{
const
  light0_position:TGLArrayf4=( -8.0, 8.0, -300.0, 1.0);
	ambient:  TGLArrayf4=( 0.8, 0.8, 0.8, 0.8 );
  }
begin
	// set viewing projection.  Draw within x,y,z [-100,100] then translate
  // by 200 on z, then rotate.

  //  glViewport(0,0,width,height);
	glMatrixMode(GL_PROJECTION);

  glOrtho(-141.0, 141.0, -141.0, 141.0, -141.0, 141.0);

	// positioning
	glMatrixMode(GL_MODELVIEW);
  {
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_POSITION, @light0_position);
	glLightfv(GL_LIGHT0, GL_AMBIENT, @ambient);
	glEnable(GL_LIGHT0);
  }
	glClearColor( 0.5, 0.5, 0.5, 1.0);		   // Background Color

  //----- Depth buffer checking
	glClearDepth(1.0);		    // Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);	// Enables Depth Testing
	glDepthFunc(GL_LESS);			// The Type Of Depth Test To Do

end;

//----- GL init for pixel format.
// iMode=0 : On screen double buffered.  iMode=1 : memory bitmap for clipboard.
procedure setupPixelFormat( DC:HDC; iMode:Integer );
var
	pfd:TPIXELFORMATDESCRIPTOR;

	pixelFormat:integer;
begin
		pfd.nSize := sizeof(TPIXELFORMATDESCRIPTOR);	// size
		pfd.nVersion := 1;						        // version
		pfd.dwFlags := PFD_SUPPORT_OPENGL or PFD_DRAW_TO_WINDOW
          or PFD_DOUBLEBUFFER;		// support double-buffering
		pfd.iPixelType := PFD_TYPE_RGBA;		  // color type
		pfd.cColorBits := 24;					      // prefered color depth
		pfd.cRedBits := 0; pfd.cRedShift := 0;		  // color bits (ignored)
		pfd.cGreenBits := 0; pfd.cGreenShift := 0;
		pfd.cBlueBits := 0; pfd.cBlueShift := 0;
		pfd.cAlphaBits := 0; pfd.cAlphaShift := 0;  // no alpha buffer
		pfd.cAccumBits :=  0;
		pfd.cAccumRedBits :=  0;  				    // no accumulation buffer,
		pfd.cAccumGreenBits :=  0;           // accum bits (ignored)
		pfd.cAccumBlueBits :=  0;
		pfd.cAccumAlphaBits :=  0;
		pfd.cDepthBits := 16;					      // depth buffer
		pfd.cStencilBits := 0;					      // no stencil buffer
		pfd.cAuxBuffers := 0;					      // no auxiliary buffers
		pfd.iLayerType := PFD_MAIN_PLANE;    // main layer
		pfd.bReserved :=  0;
		pfd.dwLayerMask :=  0;
		pfd.dwVisibleMask :=  0;
		pfd.dwDamageMask :=  0;              // no layer, visible, damage masks */

  if iMode=0 Then begin
  	pixelFormat := ChoosePixelFormat(DC, @pfd);
  	if (pixelFormat = 0) then begin
	  	MessageBox(WindowFromDC(DC), 'ChoosePixelFormat failed.', 'Error',MB_ICONERROR or MB_OK);
		  exit;
  	end;
  	if (SetPixelFormat(DC, pixelFormat, @pfd) <> TRUE) then begin
	  	MessageBox(WindowFromDC(DC), 'SetPixelFormat failed.', 'Error', MB_ICONERROR or MB_OK);
		  exit;
  	end;
  end else begin
    pfd.dwFlags := PFD_SUPPORT_OPENGL or PFD_DRAW_TO_BITMAP;
		pfd.cRedBits := 8;
		pfd.cGreenBits := 8;
		pfd.cBlueBits := 8;
		pfd.cDepthBits := 16;
  	pixelFormat := ChoosePixelFormat(DC, @pfd);
  	SetPixelFormat(DC, pixelFormat, @pfd);
    //TODO - error trapping
  end;

end;


procedure TEnergyPlot.FormCreate(Sender: TObject);
var
	DC:HDC;
  m_agmf : array[0..95] of GLYPHMETRICSFLOAT ;  // Text Call List (not req'd to persist ?)

begin
  m_aRates := TRibbons.Create();
  m_aStates := TBlocks.Create();
  m_aBars := TBars.Create();
  m_iRotate := 0;
  m_lDrawing := false;


  //----- OpenGL prep
	DC:=GetDC(Handle);
	SetupPixelFormat(DC, 0);
	m_RC:=wglCreateContext(DC);
	wglMakeCurrent(DC, m_RC);
	m_angleX := 0.0;
  m_angleY := -90.0;
  m_angleZ := 0.0;
  m_lTrack := false;
  SetCaption( True );
  m_nMaxZ := 0.0;
  m_nMinZ := 0.0;
	GLInit( ClientWidth, ClientHeight );

  //----- Build call lists for text
  canvas.Font.Name := 'Arial';
  //canvas.Font.Size := -16;    // unused ?
  wglUseFontOutlines(canvas.Handle,GL_TEXTMIN,GL_TEXTMAX-GL_TEXTMIN+1,GL_TEXTBASE, 0.0, 0.1, WGL_FONT_POLYGONS, @(m_agmf[0]) );

end;

//=========================== Form event handlers

procedure TEnergyPlot.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Free();
end;

procedure TEnergyPlot.FormPaint(Sender: TObject);
begin
	Draw;
end;

procedure TEnergyPlot.FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if m_lTrack then begin
    m_angleX := m_angleX + (x-m_iTrackX)/5;
    m_angleY := m_angleY + (y-m_iTrackY)/3;
    m_iTrackX := x;
    m_iTrackY := y;
    CauseDraw();
  end;
end;

procedure TEnergyPlot.FormMouseDown(Sender: TObject; Button: TMouseButton;  Shift: TShiftState; X, Y: Integer);
begin
  m_iTrackX := x;
  m_iTrackY := y;
  m_lTrack:=True;
end;

procedure TEnergyPlot.FormMouseUp(Sender: TObject; Button: TMouseButton;  Shift: TShiftState; X, Y: Integer);
begin
  m_lTrack:=False;
end;

procedure TEnergyPlot.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
	if (Key = 37) then m_angleX := m_angleX-4.5;      // Left
	if (Key = 39) then m_angleX := m_angleX+4.5;      // Right
	if (Key = 36) then m_angleZ := m_angleZ-4.5;      // CW
	if (Key = 35) then m_angleZ := m_angleZ+4.5;      // CCW
	if (Key = 38) then m_angleY := m_angleY-4.5;      // Up    and (m_angleY>-90.0)
	if (Key = 40) then m_angleY := m_angleY+4.5;      // Down and (m_angleY<0.0)
  if (Key = 88) Then SetCaption(not m_lCaption);    // Text for rotation
  if (Key = 45) and (sbStimulus.Visible)            // '-' = decrease stimulus
      and (sbStimulus.Position>sbStimulus.Min) Then
    sbStimulus.Position := max(sbStimulus.Position-1,sbStimulus.Min);
  if (Key = 43) and (sbStimulus.Visible)            // '+' = increase stimulus
      and (sbStimulus.Position<sbStimulus.Max) Then
    sbStimulus.Position := min(sbStimulus.Position+1,sbStimulus.Max);
	CauseDraw();
end;

procedure TEnergyPlot.Free();
begin
  m_aRates.Free();
  m_aStates.Free();
  m_aBars.Free();
end;

procedure TEnergyPlot.sbStimulusChange(Sender: TObject);
begin
//  ActiveControl:=nil;
  CauseDraw();
end;

procedure TEnergyPlot.chkGridClick(Sender: TObject);
begin
//  ActiveControl:=nil;
  CauseDraw();
end;

procedure TEnergyPlot.chkStatesClick(Sender: TObject);
begin
//  ActiveControl:=nil;
  CauseDraw();
end;

procedure TEnergyPlot.chkColorClick(Sender: TObject);
begin
//  ActiveControl:=nil;
  CauseDraw();
end;

procedure TEnergyPlot.chkLegendClick(Sender: TObject);
begin
//  ActiveControl:=nil;
  CauseDraw();
end;

procedure TEnergyPlot.btnResetClick(Sender: TObject);
begin
  m_angleX := 0;
  m_angleY := -90;
  m_angleZ := 0;
//  ActiveControl:=nil;
  CauseDraw();
end;

procedure TEnergyPlot.btnMovieClick(Sender: TObject);
begin
  Movie();
//  ActiveControl:=nil;
  CauseDraw();
end;

procedure TEnergyPlot.FormActivate(Sender: TObject);
begin
  // if multiple open GL windows will be available, this is required.
	wglMakeCurrent(GetDC(Handle), m_RC);
end;

procedure TEnergyPlot.FormResize(Sender: TObject);
begin
  glViewport(0,0,ClientWidth,ClientHeight);
  CauseDraw();
end;

//----- Add a Rate ribbon
procedure TEnergyPlot.AddRate(x0,y0,z0,zz0, x1,y1,z1,zz1, zb,zzb : Double);
begin
  m_aRates.AddRibbon(x0,y0,z0,zz0, x1,y1,z1,zz1, zb, zzb);
  m_nMinZ := min(min(m_nMinZ, z0 ), zz0);
  m_nMinZ := min(min(m_nMinZ, z1 ), zz1);
  m_nMaxZ := max(max(m_nMaxZ, zb ), zzb);
end;

//----- Add a State block to the array
procedure TEnergyPlot.AddState(x,y,z,zz, nSz : Double; c:String; rgb:TColor);
begin
  m_aStates.AddBlock(x,y,z,zz,nSz,c,rgb);
end;

//----- Add a Bar to the array
procedure TEnergyPlot.AddBar(x0,x1,y0,y1,z0,z1 : Double);
begin
  m_aBars.AddBar(x0,x1,y0,y1,z0,z1);
  m_nMinZ := min(m_nMinZ, z0 );
  m_nMaxZ := max(m_nMaxZ, z1 );
end;

//----- iterate through complete range of stimulus in a movie mode.
procedure TEnergyPlot.Movie();
var
  i : Integer;
  //iPrev : Integer;
Begin
  //iPrev := sbStimulus.Position;
  i := sbStimulus.Min;
  while i <= sbStimulus.Max do Begin
    sbStimulus.Position := i;
    inc(i,5);
  end;
  i := sbStimulus.Max;
  while i >= sbStimulus.Min do Begin
    sbStimulus.Position := i;
    dec(i,5);
  end;
  sbStimulus.Position := sbStimulus.Min;
  //sbStimulus.Position:=iPrev;
end;

//----- Indicate that the model has a stimulus and enable appropriate controls
procedure TEnergyPlot.SetStimulusDependency(lDependant : boolean);
begin
//  ActiveControl:=nil;
  btnMovie.Visible := lDependant;
  lblStimulus.Visible := lDependant;
  sbStimulus.Visible := lDependant;
end;

//----- Turn captions on / off.  Set x,y,z caption to correct values.
// TODO - no good reason for this - option could be removed.
procedure TEnergyPlot.SetCaption(l : boolean);
begin
  m_lCaption := l;
  if m_lCaption Then
    lblCaption2.Caption := Format('x:%6.1f y:%6.1f z:%6.1f',[m_angleX,m_angleY,m_angleZ]);
  lblCaption1.Visible := l;
  lblCaption2.Visible := l;
end;

//==============================================================================
//-----
Constructor TRibbons.Create();
begin
  //----- Initialize rates array to 0 ribbons
  m_iCount := 0;
  m_iAlloc := 10;
  SetLength( m_aRates, m_iAlloc );
end;

//----- Cleanup member array
Procedure TRibbons.Free();
var
  i : Integer;
begin
  for i := 0 to m_iCount-1 do
    m_aRates[i].Free;
  m_iCount := 0;
  m_iAlloc := 10;
  SetLength( m_aRates, m_iAlloc );
end;

//----- Add a Rate 'Ribbon' to the array
procedure TRibbons.AddRibbon(x0,y0,z0,zz0, x1,y1,z1,zz1, zb,zzb : Double);
begin
  if m_iCount >= m_iAlloc Then begin
    inc(m_iAlloc,10);
    SetLength( m_aRates, m_iAlloc );
  end;
  m_aRates[m_iCount] := TRibbon.Create;
  m_aRates[m_iCount].m_nX0 := x0;
  m_aRates[m_iCount].m_nY0 := y0;
  m_aRates[m_iCount].m_nZ0[0] := z0;
  m_aRates[m_iCount].m_nZ0[1] := zz0;
  m_aRates[m_iCount].m_nX1 := x1;
  m_aRates[m_iCount].m_nY1 := y1;
  m_aRates[m_iCount].m_nZ1[0] := z1;
  m_aRates[m_iCount].m_nZ1[1] := zz1;
  m_aRates[m_iCount].m_nZb[0] := zb;
  m_aRates[m_iCount].m_nZb[1] := zzb;

  inc(m_iCount);
end;

//-----
Function TRibbons.GetCount() : Integer;
begin
  result := m_iCount;
end;

//-----
Function TRibbons.Get(i:Integer):TRibbon;
begin
  result := m_aRates[max(0,min(m_iCount-1,i))];
end;

//==============================================================================
//-----
Constructor TBlocks.Create();
begin
  //----- Initialize States array to 0 states
  m_iCount := 0;
  m_iAlloc := 10;
  SetLength( m_aBlocks, m_iAlloc );
end;

//----- Cleanup member array
Procedure TBlocks.Free();
var
  i : Integer;
begin
  for i := 0 to m_iCount-1 do
    m_aBlocks[i].Free;
  m_iCount := 0;
  m_iAlloc := 10;
  SetLength( m_aBlocks, m_iAlloc );
end;

//-----
procedure TBlocks.AddBlock(x,y,z,zz, nSz : Double; c:String; rgb:TColor);
begin
  if m_iCount >= m_iAlloc Then begin
    inc(m_iAlloc,10);
    SetLength( m_aBlocks, m_iAlloc );
  end;
  m_aBlocks[m_iCount] := TBlock.Create;
  m_aBlocks[m_iCount].m_nX := x;
  m_aBlocks[m_iCount].m_nY := y;
  m_aBlocks[m_iCount].m_nZ[0] := z;
  m_aBlocks[m_iCount].m_nZ[1] := zz;
  m_aBlocks[m_iCount].m_nSz := nSz;
  m_aBlocks[m_iCount].m_cName := Copy(c,0,4);
  m_aBlocks[m_iCount].m_rgbFaceColor := rgb;

  inc(m_iCount);
end;

//-----
Function TBlocks.GetCount() : Integer;
begin
  result := m_iCount;
end;

//-----
Function TBlocks.Get(i:Integer):TBlock;
begin
  result := m_aBlocks[max(0,min(m_iCount-1,i))];
end;

//-----
procedure TBlock.GLDrawBlock(nPct : Double; lColor : boolean);
  Procedure Face(nRotX,nRotY,nRotz:double);
  begin
    glPushMatrix();
    if nRotX <>0.0 Then glRotatef(nRotX, 1.0, 0.0, 0.0);
    if nRotY <>0.0 Then glRotatef(nRotY, 0.0, 1.0, 0.0);
    if nRotZ <>0.0 Then glRotatef(nRotZ, 0.0, 0.0, 1.0);

    // Solid face ( current color )
    glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(-1.0, -1.0, 1.0);
    glVertex3f(-1.0,  1.0, 1.0);
    glVertex3f( 1.0, -1.0, 1.0);
    glVertex3f( 1.0,  1.0, 1.0);
    glEnd();
    // wire frame
	  glColor3f(0.0, 0.0, 0.0 );
    glBegin(GL_LINE_STRIP);
    glVertex3f(-1.02, -1.02, 1.02);
    glVertex3f( 1.02, -1.02, 1.02);
    glVertex3f( 1.02,  1.02, 1.02);
    glVertex3f(-1.02,  1.02, 1.02);
    glVertex3f(-1.02, -1.02, 1.02);
    glEnd();
    // Text Label
	  glColor3f(1.0, 1.0, 1.0 );
    glTranslate(m_nSz/-30.0 * Length(m_cName), m_nSz/-18.0, 1.1);
    glScalef(6.0/m_nSz, 6.0/m_nSz, 6.0/m_nSz);
    glListBase(GL_TEXTBASE-GL_TEXTMIN);
    glCallLists(Length(m_cName), GL_UNSIGNED_BYTE, PCHAR(m_cName));
    glPopMatrix();
  end;

  Function Z() : double;
  Begin
    result := m_nZ[0]*(1-nPct)+m_nZ[1]*nPct;
  end;

var
    pRGB : PByteArray;

begin
    glPushMatrix();
    glTranslate(m_nX, m_nY, Z() );
    glScale(m_nSz,m_nSz,m_nSz);

    pRGB := Addr(m_rgbFaceColor); // use with glColor3b which requires signed byte arguments.

    glColor3f(0.2,0.2,0.2);    // Front
    Face(  0.0,  0.0,  0.0);
    glColor3f(0.2,0.2,0.2);    // Back
    Face(180.0,  0.0,180.0);
    // Top
    if lColor Then
      glColor3b(pRGB[0] div 2, pRGB[1] div 2, pRGB[2] div 2)
    else
      glColor3f(0.7,0.7,0.7);
    Face( 90.0,  0.0,  0.0);

    // Bottom
    if lColor Then
      glColor3b(pRGB[0] div 2, pRGB[1] div 2, pRGB[2] div 2)
    else
      glColor3f(0.7,0.7,0.7);
    Face(-90.0,  0.0,180.0);

    // Right
    if lColor Then
      glColor3b(pRGB[0] div 4, pRGB[1] div 4, pRGB[2] div 4)
    else
      glColor3f(0.5,0.5,0.5);
    Face(  0.0, 90.0, 90.0);

    // Left
    if lColor Then
      glColor3b(pRGB[0] div 4, pRGB[1] div 4, pRGB[2] div 4)
    else
      glColor3f(0.5,0.5,0.5);
    Face(  0.0,-90.0,-90.0);

    glPopMatrix();
end;

//==============================================================================
//-----
procedure TBars.AddBar(x0,x1,y0,y1,z0,z1 :Double);
begin
  if m_iCount >= m_iAlloc Then begin
    inc(m_iAlloc,40);
    SetLength( m_aBars, m_iAlloc );
  end;
  m_aBars[m_iCount] := TBar.Create;
  m_aBars[m_iCount].m_nX0 := x0;
  m_aBars[m_iCount].m_nX1 := x1;
  m_aBars[m_iCount].m_nY0 := y0;
  m_aBars[m_iCount].m_nY1 := y1;
  m_aBars[m_iCount].m_nZ0 := z0;
  m_aBars[m_iCount].m_nZ1 := z1;
  inc(m_iCount);
end;

//-----
procedure TBar.GLDrawBar();
  const
    EXTR : double = 0.1;

  Procedure Face( x0,x1, y0,y1, z0,z1 : double );
  begin
    // Solid Face
    glBegin(GL_TRIANGLE_STRIP);
    glVertex3f(x0, y0, z0);
    if x1<>x0 then glVertex3f(x1, y0, z0) else glVertex3f(x0, y1, z0);
    if z1<>z0 then glVertex3f(x0, y0, z1) else glVertex3f(x0, y1, z0);
    glVertex3f(x1, y1, z1);
    glEnd();
    // wire frame
	  glColor3f(0.0, 0.0, 0.0 );
    glBegin(GL_LINE_STRIP);
    glVertex3f(x0-EXTR,y0-EXTR,z0-EXTR);
    if x1<>x0 then glVertex3f(x1+EXTR, y0-EXTR, z0-EXTR) else glVertex3f(x0-EXTR, y1+EXTR, z0-EXTR);
    glVertex3f(x1+EXTR,y1+EXTR,z1+EXTR);
    if z1<>z0 then glVertex3f(x0-EXTR, y0-EXTR, z1+EXTR) else glVertex3f(x0-EXTR, y1+EXTR, z0-EXTR);
    glVertex3f(x0-EXTR,y0-EXTR,z0-EXTR);
    glEnd();
  end;

begin
    //Solid faces
    glColor3f(1.0,0.0,0.0); Face( m_nX0, m_nX1, m_nY0, m_nY1, m_nZ0, m_nZ0 );
    glColor3f(1.0,1.0,0.0); Face( m_nX0, m_nX1, m_nY0, m_nY1, m_nZ1, m_nZ1 );
    glColor3f(1.0,0.0,1.0); Face( m_nX0, m_nX1, m_nY0, m_nY0, m_nZ0, m_nZ1 );
    glColor3f(0.0,1.0,0.0); Face( m_nX0, m_nX1, m_nY1, m_nY1, m_nZ0, m_nZ1 );
    glColor3f(0.0,1.0,1.0); Face( m_nX0, m_nX0, m_nY0, m_nY1, m_nZ0, m_nZ1 );
    glColor3f(0.0,0.0,1.0); Face( m_nX1, m_nX1, m_nY0, m_nY1, m_nZ0, m_nZ1 );
end;

//-----
Constructor TBars.Create();
begin
  //----- Initialize Bars Array to 0 Bars
  m_iCount := 0;
  m_iAlloc := 10;
  SetLength( m_aBars, m_iAlloc );
end;

//----- Cleanup member array
Procedure TBars.Free();
var
  i : Integer;
begin
  for i := 0 to m_iCount-1 do
    m_aBars[i].Free;
  m_iCount := 0;
  m_iAlloc := 10;
  SetLength( m_aBars, m_iAlloc );
end;

//-----
Function TBars.GetCount() : Integer;
begin
  result := m_iCount;
end;

//-----
Function TBars.Get(i:Integer):TBar;
begin
  result := m_aBars[max(0,min(m_iCount-1,i))];
end;

//----- Create a DI Bitmap and draw the GL image to it, then save on clipboard.
procedure TEnergyPlot.btnCopyClick(Sender: TObject);
var
  info : BITMAPINFO;
  bitmap : HBITMAP;
  dcBmp, dcWin : HDC;
  bits : Pointer;//GLubyte;
 	rc : HGLRC;
  i,j,k : Integer;
  m_agmf : array[0..95] of GLYPHMETRICSFLOAT ;  // Text Call List (not req'd to persist ?)
  iWid, iHgt : Integer;
  hMem : Pointer;
  Mem : Integer;

begin
  dcWin := GetDC(Handle);
  dcBmp := CreateCompatibleDC(dcWin);
  iWid := 4*(ClientWidth div 4);
  iHgt := ClientHeight;

  //----- Prepare bitmap for Open GL rendering
  info.bmiHeader.biSize := sizeof(BITMAPINFOHEADER);
  info.bmiHeader.biWidth := iWid;
  info.bmiHeader.biHeight := iHgt;
  info.bmiHeader.biPlanes := 1;
  info.bmiHeader.biBitCount := 24;
  info.bmiHeader.biCompression := BI_RGB;
  info.bmiHeader.biClrImportant:=0;
  info.bmiHeader.biClrUsed:=0;
  info.bmiHeader.biSizeImage:=0;
  info.bmiHeader.biXPelsPerMeter:=0;
  info.bmiHeader.biYPelsPerMeter:=0;

  bitmap := CreateDIBSection( dcBmp, info, DIB_RGB_COLORS, bits, 0, 0);
  SelectObject( dcBmp, bitmap );

	SetupPixelFormat( dcBmp, 1);
  rc := wglCreateContext(dcBmp);
	wglMakeCurrent(dcBmp, rc);
  GLInit( iWid, iHgt );

  SelectObject( dcBmp, GetStockObject(DEFAULT_GUI_FONT) );
  wglUseFontOutlines( dcBmp,GL_TEXTMIN,GL_TEXTMAX-GL_TEXTMIN+1,GL_TEXTBASE, 0.0, 0.1, WGL_FONT_POLYGONS, @(m_agmf[0]) );

  //----- Render to bitmap
  Draw;

  //----- copy bits to clipboard
  Mem := GlobalAlloc( GHND, sizeof(info)+iWid*iHgt*3+1 );
  hMem := GlobalLock(Mem);
  for i := 0 to sizeof(info)-1 do
    PByteArray(hMem)[i] := PByteArray( @info )[i];
  for i := 0 to iHgt-1 do
    for j := 0 to iWid-1 do Begin
      k := sizeof(info)+(i*iWid+j)*3;        // Rearrange bytes - why ?
      PByteArray(hMem)[k+1] := PByteArray( bits )[k];
      PByteArray(hMem)[k+2] := PByteArray( bits )[k+1];
      PByteArray(hMem)[k] := PByteArray( bits )[k+2];
    end;

  GlobalUnlock(Mem);

  if OpenClipboard(Application.Handle) Then Begin
    EmptyClipboard( );
    if SetClipboardData( CF_DIB, Mem ) = 0{NULL} Then
      GlobalFree(Mem);
    CloseClipboard();
  end else
    GlobalFree(Mem);

  //----- Clean up
	wglMakeCurrent(dcWin, m_RC);
  wglDeleteContext(rc);
  DeleteObject(bitmap);
  DeleteDC(dcBmp);
//  ActiveControl:=nil;
end;

procedure TEnergyPlot.btnExportClick(Sender: TObject);
begin
  Application.MessageBox( 'Under Construction', 'Movie Export', MB_OK);
  //ExportPrepare()
  // { Export, Adjust Stim }
  //ExportCleanup()
  //  ActiveControl:=nil;
end;


end.
