{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// by Chris c.2002
// interactive python text area
// fantastic complexity arises from keeping track of the most recent '>>>' in the presence of arbitrary editing
// slows down to a crawl before even 100 lines
// still gets confused if you delete the space to the right of '>>>'.
// (would be a failed experiment if there were anything better to take its place)

// UPDATE: PythonForDelphi now includes a full-featured editor; maybe part of it would fit here instead.


unit PythonReadEvalPrint;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  PythonEngine, QUB_Python_Thread,
  StdCtrls, ComCtrls;

type
  TPythonReadEvalPrint = class(TRichEdit)
  private
    { Private declarations }
  protected
    // properties
    prEngine                      :TPythonEngine;
    prEngineIO                    :TPythonInputOutput;
    prInited                      :Boolean;
    prOnSendData                  :TSendDataEvent;

    // track the previous insertion point, selection, and Length(Text)
    // so when something changes we can figure out what it was
    prLastSelStart                :Integer;
    prLastSelLength               :Integer;
    prLastSize                    :Integer;

    // when you press return, it triggers KeyPressed, then inserts CRLF,
    // then triggers SelectionChanged.
    // prReturnPressed is set in KeyPressed, then cleared when we handle it
    // in SelectionChanged.
    prRtnPressed                  :Boolean;

    // index in Text of the beginning of the current expression
    // used to get the current expression on RtnPressed,
    // and to handle properly edits before it (move prExprStart)
    // and after it (modify current expression)
    prExprStart                   :Integer;

    // whether the current expression is multi-line
    // (whether it takes an extra blank line to trigger evaluation)
    prExprMulti                   :Boolean;

    // set by properties
    // when both are set, it initializes (link Engine, IO, Self; print prompt)
    procedure prSetEngine(Engine :TPythonEngine);
    procedure prSetEngineIO(EngineIO :TPythonInputOutput);
    procedure prInitialize;

    procedure prSaveContext;   //  get prLastWhatever to prepare for next time

    procedure prDeleteExtras; // blow away trailing w.s. and prompts
    function  prReadExpr: String;
    function  prReadExprLastLine: String;
    procedure prInterpret(Expr :String);
    procedure prPromptStart;
    procedure prPromptCont(PrevLine :String); // indent like prevline, +tab if PrevLine ends in :

  public
    constructor Create(AOwner :TComponent); override;
    //destructor Destroy; override;

    // events for text
    procedure PythonKeyPress(Sender: TObject; var Key: Char);
    procedure PythonSelectionChange(Sender: TObject);
    // output from IO.OnSendData
    procedure PythonPrint(Sender: TObject; const Data : String );

  published
    property Engine               :TPythonEngine          read prEngine     write prSetEngine;
    property IO                   :TPythonInputOutput     read prEngineIO   write prSetEngineIO;

    // if this is nil, send expressions directly to the Engine
    property OnSendData           :TSendDataEvent         read prOnSendData write prOnSendData;
  end;

procedure Register;

implementation

constructor TPythonReadEvalPrint.Create(AOwner :TComponent);
begin
  inherited;

  // set own event handlers
  OnKeyPress            := PythonKeyPress;
  OnSelectionChange     := PythonSelectionChange;

  prEngine              := nil; // TPythonEngine.Create(Self);
  prEngineIO            := nil; // TPythonInputOutput.Create(Self);
  prInited              := False;
end;

procedure TPythonReadEvalPrint.PythonKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    prRtnPressed := True;
  end;
end;

procedure TPythonReadEvalPrint.PythonSelectionChange(Sender: TObject);
// event for pretty much anything that happens in the text, after it happens
var
  Growth            :Integer;
  here              :Integer;
  Expr, Line        :String;
begin
  Growth      := Length(Text) - prLastSize;
  if Growth <> 0 then begin
    if prLastSelStart < prExprStart then begin
      if prLastSelLength > 0 then begin
        if (prLastSelStart + prLastSelLength) <= prExprStart then begin
          prExprStart     := prExprStart + Growth;
        end else begin // sel spanned ExprStart
          prExprStart := prLastSelStart;
        end;
      end else begin // nothing selected
        prExprStart := prExprStart + Growth;
      end;
    end else begin
      if (SelStart < prLastSelStart) // backspace
      and (prLastSelStart = prExprStart) then begin
        prExprStart := prExprStart + Growth; // -= 1;
      end;
    end;
  end;

  if prRtnPressed then begin
    prRtnPressed := False; // some actions herein cause SelChanged to be called recursively
    prSaveContext; // maybe necessary for PythonIO output to display in the right place?
    here := SelStart;
    prDeleteExtras; // if text > here is just ws and prompts, destroy it
    if here = Length(Text) then begin
      Expr := prReadExpr; // from ExprStart to here, excluding any prompt chars at beginning of lines
      Line := prReadExprLastLine; // excluding trailing space,tab,rtn,nl
      if prExprMulti then begin
        if Length(Line) = 0 then begin
          prExprMulti := False;
          prExprStart := Length(Text);  // so PythonIO prints after expr being evaluated
          prInterpret(Expr);
          prPromptStart;
        end else begin
          prPromptCont(Line);
        end;
      end else begin
        if (Length(Line) = 0) or (Length(Expr) = 0) then begin
          prPromptStart;
        end else if Line[ Length(Line) ] = ':' then begin
          prExprMulti := True;
          prPromptCont(Line);
        end else begin
          prExprStart := Length(Text);
          prInterpret(Expr);
          prPromptStart;
        end;
      end;
    end;
  end;
  
  prSaveContext;
end;

procedure TPythonReadEvalPrint.PythonPrint(Sender: TObject; const Data : String );
var
  InsertBefore        :Integer;
  CharCount           :Integer;
  c                   :Char;
  OldLength           :Integer;
  OldPos              :Integer;
begin
  InsertBefore := 0;

  CharCount := Length(Lines[0]);
  while CharCount < prExprStart do begin
    Inc(InsertBefore);

    c := Text[CharCount+1];
    while (c = #13) or (c = #10) do begin
      Inc(CharCount);
      c := Text[CharCount+1];
    end;

    CharCount := CharCount + Length(Lines[InsertBefore]);
  end;

  OldLength := Length(Text);
  OldPos    := prLastSelStart;
  Lines.Insert(InsertBefore, Data);
  SelStart  := OldPos + (Length(Text) - OldLength);
end;

procedure TPythonReadEvalPrint.prSetEngine(Engine :TPythonEngine);
begin
  prEngine := Engine;
  prEngine.RedirectIO   := True;
  prInitialize;
end;

procedure TPythonReadEvalPrint.prSetEngineIO(EngineIO: TPythonInputOutput);
begin
  prEngineIO := EngineIO;
  prEngineIO.OnSendData := PythonPrint;
  prInitialize;
end;

procedure TPythonReadEvalPrint.prInitialize;
begin
  if (IO <> nil) and (Engine <> nil) then begin
    prEngine.IO           := prEngineIO;
  end;

  if (not prInited) and (IO <> nil) then begin
    prInited              := True;
    Clear;
    prExprStart := 0;
    prPromptStart;
    SelStart := Length(Text) + 1;
    prSaveContext;
  end;
end;

procedure TPythonReadEvalPrint.prSaveContext;
begin
  prLastSize       := Length(Text);
  prLastSelStart   := SelStart;
  prLastSelLength  := SelLength;
  prRtnPressed     := False;
end;

procedure TPythonReadEvalPrint.prDeleteExtras; // blow away trailing w.s. and prompts
var
  AllJunk         :Boolean;
  i               :Integer;
  c               :Char;
begin
  AllJunk := True;
  for i := SelStart to Length(Text) do begin
    c := Text[i];
    if (c <> ' ') and (c <> #9) and (c <> #13) and (c <> #10) and (c <> '>') and (c <> '.')
    then AllJunk := False;
  end;

  if AllJunk then begin
    SelLength := Length(Text) - SelStart;
    ClearSelection;
  end;
end;

function TPythonReadEvalPrint.prReadExpr: String;
var
  i              :Integer;
  c              :Char;
  FreshNL        :Boolean;
begin
  FreshNL := False;
  Result := '';
  for i := prExprStart + 1 to Length(Text) - 2 do begin    // skip final crlf
    c := Text[i];
    if (c = #13) or (c = #10) then begin
      Result := Result + c;
      FreshNL := True;
    end else if FreshNL and ((c = '>') or (c = '.')) then begin
      // don't include prompt
    end else if FreshNL and (c = ' ') then begin
      FreshNL := False; // allow one space after prompt
    end else begin
      Result := Result + c;
      FreshNL := False;
    end;
  end;
end;

function TPythonReadEvalPrint.prReadExprLastLine: String;
var
  i              :Integer;
  c              :Char;
  LineStart     :Integer;
  LineLen       :Integer;
begin
  Result := Lines[ Lines.Count - 1 ];
  if Length(Result) = 0 then begin
    Result := Lines[ Lines.Count - 2 ];
  end;

  LineStart := Length(Result); // assume it's all prompt
  for i := 1 to Length(Result) do begin
    c := Result[i];
    if (c <> '>') and (c <> '.') then begin
      if (c = ' ') then begin
        LineStart := i + 1;
      end else begin
        LineStart := i;
      end;
      Break;
    end;
  end;

  LineLen := 0; // assume it's all white space
  for i := Length(Result) downto LineStart do begin
    c := Result[i];
    if (c <> ' ') and (c <> #9) then begin
      LineLen := i - LineStart + 1;
      Break;
    end;
  end;

  Result := Copy(Result, LineStart, LineLen);
end;

procedure TPythonReadEvalPrint.prInterpret(Expr :String);
begin
  try
    prOnSendData(Self, Expr);
  except
    try
      prEngine.Run_CommandAsString(Expr, single_input);
    except
    end;
  end;
end;

procedure TPythonReadEvalPrint.prPromptStart;
begin
  Lines[ Lines.Count ] := '>>> ';
  prExprStart := Length( Text );
  prExprMulti := False;
end;

procedure TPythonReadEvalPrint.prPromptCont(PrevLine :String);
var
  i           :Integer;
  c           :Char;
  IndentStr   :String;
begin
  IndentStr := '';
  for i := 1 to Length(PrevLine) do begin
    c := PrevLine[i];
    if (c = ' ') or (c = #9) then IndentStr := IndentStr + c else Break;
  end;
  if PrevLine[ Length(PrevLine) ] = ':' then IndentStr := IndentStr + '    ';

  Lines[ Lines.Count ] := '... ' + IndentStr;
end;

procedure Register;
begin
  RegisterComponents('Python', [TPythonReadEvalPrint]);
end;

end.
