{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// by Chris
// defines an Action by its name, l-click, r-click, properties-tree, and layouts
// allows actions to be written in Python, or invoked therefrom
//
// 7-11-2004 : added Hint,ShortName,Color items with defaults to Create
//             Constructor for efficiency in calling code. {JB}
//
unit QUB_Action;

interface

uses

Classes, Graphics, SyncObjs,
QUB_Node,
QUB_Node_Python,
PythonEngine;

type TProcOfObj = procedure of object;
type TProcWithCompletion = procedure(OnActionDone :TNotifyEvent) of object;
// the completion's primary use is so Python can invoke an Action then wait for it to finish before returning.

type TQUB_Action = class
protected

  prName                :String;
  prShortName           :String;
  prColor               :TColor;
  prHint                :String;
  prDescription         :String;
  prVisLayouts          :TStringList; // in which layouts is this action visible

  prOnLClick            :TProcWithCompletion;
  prOnRClick            :TProcWithCompletion;
  prProperties          :TNodeMethPtr;

  prPyAction            :PPyObject;

  procedure prDefLClick(OnActionDone :TNotifyEvent);
  procedure prDefRClick(OnActionDone :TNotifyEvent);
  function prDefProperties( arg :INode ): INode;

public

  constructor Create( Name :String;
                      OnLClick, OnRClick :TProcWithCompletion;
                      Properties :TNodeMethPtr;
                      cHint : String = '';
                      cShortName : String = '';
                      clColor : TColor = clBlack );

  destructor Destroy; override;

  property Name         :String              read prName            write prName;
  property ShortName    :String              read prShortName       write prShortName;
  property Color        :TColor              read prColor           write prColor;
  property Hint         :String              read prHint            write prHint;
  property Description  :String              read prDescription     write prDescription;
  property VisLayouts   :TStringList         read prVisLayouts;

  property OnLClick     :TProcWithCompletion read prOnLClick        write prOnLClick;
  property OnRClick     :TProcWithCompletion read prOnRClick        write prOnRClick;
  property Properties   :TNodeMethPtr        read prProperties      write prProperties;

  property PyAction     :PPyObject           read prPyAction;

end;

// an action as visible from python

TQUB_Action_PyWrap = class(TPyObject)
protected
  prAction              :TQUB_Action;

  prPropertiesNode      :INode;

  procedure prProperties;
  
public

  function  GetAttr(key : PChar) : PPyObject; override;
  // function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  function Properties( args : PPyObject ) : PPyObject; cdecl;
  function RClick( args : PPyObject ) : PPyObject; cdecl;

  function  Call( args, kw : PPyObject) : PPyObject; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;

  property Action       :TQUB_Action        read prAction        write prAction;
end;

// invoking an action from python

TQUB_Action_Invocation = class
protected
  prAction              :TQUB_Action;
  prActionDone          :TEvent;

  procedure prSyncLClick;
  procedure prSyncRClick;

  procedure prOnActionDone(Sender :TObject);

public
  constructor Create(Action :TQUB_Action);
  destructor Destroy; override;

  procedure DoLClick; // from python thread; pref. after releasing GIL -- waits until ActionDone
  procedure DoRClick;
end;

// defining an action in python

TQUB_Action_Python = class(TQUB_Action)
protected

  prLClick        :PPyObject;
  prRClick        :PPyObject;
  prProps         :PPyObject;

  procedure prCleanup;
  procedure prCallNoArgs(PyFunc :PPyObject);

  procedure prDoLClick(OnActionDone :TNotifyEvent);
  procedure prDoRClick(OnActionDone :TNotifyEvent);
  function prDoProperties(arg :INode): INode;

  procedure prDoLClickPy;
  procedure prDoRClickPy;

public

  constructor CreatePy( Name :String; OnLClick, OnRClick, Properties :PPyObject );  // caller must hold python GIL
  destructor Destroy; override;

end;


procedure QUB_Action_InitPython;

var
  QUB_Action_PyType     :TPythonType;
  

implementation


uses

QUB_Viewport_Python,
QUB_Main;


constructor TQUB_Action.Create( Name :String;
                                OnLClick,
                                OnRClick :TProcWithCompletion;
                                Properties :TNodeMethPtr;
                                cHint : String;
                                cShortName : String;
                                clColor : TColor);
var
  ActionWrap              :TQUB_Action_PyWrap;
begin
  prName                := Name;
  if cShortName = '' Then prShortName := Name else prShortName := cShortName;
  prOnLClick            := OnLClick;
  prOnRClick            := OnRClick;
  prProperties          := Properties;

  if @prOnLClick   = nil then prOnLClick   := prDefLClick;
  if @prOnRClick   = nil then prOnRClick   := prDefRClick;
  if @prProperties = nil then prProperties := prDefProperties;

  prVisLayouts          := TStringList.Create;
  prVisLayouts.Sorted   := True; // speed lookup?

  PythonThread.BeginPython;
  prPyAction            := QUB_Action_PyType.CreateInstance;
  ActionWrap            := TQUB_Action_PyWrap( PythonToDelphi(prPyAction) );
  ActionWrap.Action     := Self;

  Color        := clColor;
  Hint         := cHint;
  Description  := '';

  PythonThread.EndPython;
end;

destructor TQUB_Action.Destroy;
begin
  prVisLayouts.Free;
  PythonThread.BeginPython;
  GetPythonEngine.Py_DECREF( prPyAction );
  PythonThread.EndPython;
end;

procedure TQUB_Action.prDefLClick(OnActionDone :TNotifyEvent);
begin
  if @OnActionDone <> nil then OnActionDone(Self);
end;

procedure TQUB_Action.prDefRClick(OnActionDone :TNotifyEvent);
begin
  OnLClick(OnActionDone);
end;

function TQUB_Action.prDefProperties( arg :INode ): INode;
begin
  Result := TNode.CreateNull;
end;


// *****************************************************************************

function TQUB_Action_PyWrap.GetAttr(key : PChar) : PPyObject;
begin
  if Action = nil then begin
    Result := inherited GetAttr(key);
    Exit;
  end;

  with GetPythonEngine do begin
    if key = 'name' then
      Result := PyString_FromString( PChar(Action.Name) )
    else if key = 'hint' then
      Result := PyString_FromString( PChar(Action.Hint ) )
    else if key = 'description' then
      Result := PyString_FromString( PChar(Action.Description ) )
    else
      Result := inherited GetAttr(key);
  end;
end;

function TQUB_Action_PyWrap.Properties( args : PPyObject ) : PPyObject; cdecl;
var
  arg                    :PPyObject;
  argNode                :INode;
  ResultNode             :INode;
begin
  Adjust(@Self);

  if Action = nil then begin
    ResultNode := TNode.CreateNull;
  end else begin
    with GetPythonEngine do begin
      arg      := nil;
      argNode  := nil;
      if ( PyArg_ParseTuple( args, '|O:Properties(qtr.Node)', [@arg] ) = 0 ) then begin
        ResultNode := nil;
      end else begin
        try
          if arg = nil then begin
            argNode := TNode.Create('Properties');
          end else if PyNode_IsNode( arg ) then begin
            argNode := PyNode_ToNode( arg );
          end;

          prPropertiesNode := argNode;
          PythonThread.SyncGUIProc( prProperties );
          ResultNode       := prPropertiesNode;
        except
          ResultNode := TNode.CreateNull;
        end;
      end;
    end;
  end;

  if ResultNode <> nil then begin
    Result  := PyNode_FromNode( ResultNode );
  end else begin
    Result  := nil;
  end;
end;

procedure TQUB_Action_PyWrap.prProperties;
begin
  prPropertiesNode := Action.Properties( prPropertiesNode );
end;

function TQUB_Action_PyWrap.RClick( args : PPyObject ) : PPyObject; cdecl;
var
  Invocation             :TQUB_Action_Invocation;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    Result := ReturnNone;

    if Action = nil then Exit;

    // release GIL
    PythonThread.EndPython;
    Invocation           := nil;

    try
      Invocation           := TQUB_Action_Invocation.Create(Action);
      Invocation.DoRClick;
    finally
      Invocation.Free;

      // reclaim GIL
      PythonThread.BeginPython;
    end;
  end;
end;

function TQUB_Action_PyWrap.Call( args, kw : PPyObject) : PPyObject;
var
  Invocation             :TQUB_Action_Invocation;
begin
  with GetPythonEngine do begin
    Result := ReturnNone;

    if Action = nil then Exit;

    // release GIL
    PythonThread.EndPython;
    Invocation           := nil;

    try
      Invocation           := TQUB_Action_Invocation.Create(Action);
      Invocation.DoLClick;
    finally
      Invocation.Free;

      // reclaim GIL
      PythonThread.BeginPython;
    end;
  end;
end;

class procedure TQUB_Action_PyWrap.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do
    begin
      AddMethod( 'Properties', @TQUB_Action_PyWrap.Properties, 'Action.Properties( qtr.Node )' );
      AddMethod( 'RClick', @TQUB_Action_PyWrap.RClick, 'Action.RClick()' );
    end;
end;

procedure QUB_Action_InitPython;
begin
  if QUB_Action_PyType <> nil then Exit;

  QUB_Action_PyType := TPythonType.Create(nil);
  QUB_Action_PyType.Engine := GetPythonEngine;
  QUB_Action_PyType.Module := PythonViewport.pmQUBTypes;
  QUB_Action_PyType.TypeName := 'ButtonActionType';
  QUB_Action_PyType.Services.Basic := [bsGetAttr, bsCall];
  QUB_Action_PyType.PyObjectClass := TQUB_Action_PyWrap;
  QUB_Action_PyType.Initialize;
end;


// *****************************************************************************
constructor TQUB_Action_Invocation.Create(Action :TQUB_Action);
begin
  prAction              := Action;
  prActionDone          := TEvent.Create(nil{security}, False{auto-reset}, False{init:off}, ''{unique});
end;

destructor TQUB_Action_Invocation.Destroy;
begin
  try
    prActionDone.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Action_Invocation.prOnActionDone(Sender :TObject);
begin
  prActionDone.SetEvent;
end;

procedure TQUB_Action_Invocation.DoLClick; // from python thread -- waits until ActionDone
var
  WaitResult       :TWaitResult;
begin
  SyncGUIProc( prSyncLClick );
  repeat
    WaitResult     := prActionDone.WaitFor( 1000 )
  until WaitResult = wrSignaled;

  prActionDone.ResetEvent;
end;

procedure TQUB_Action_Invocation.DoRClick;
var
  WaitResult       :TWaitResult;
begin
  SyncGUIProc( prSyncRClick );
  repeat
    WaitResult     := prActionDone.WaitFor( 1000 )
  until WaitResult = wrSignaled;
  
  prActionDone.ResetEvent;
end;

procedure TQUB_Action_Invocation.prSyncLClick;
begin
  prAction.OnLClick( prOnActionDone );
end;

procedure TQUB_Action_Invocation.prSyncRClick;
begin
  prAction.OnRClick( prOnActionDone );
end;

//****************************************************************************//

constructor TQUB_Action_Python.CreatePy( Name :String; OnLClick, OnRClick, Properties :PPyObject );
// caller must hold the python GIL
begin
  PythonThread.EndPython;
  inherited Create(Name, prDoLClick, prDoRClick, prDoProperties);
  PythonThread.BeginPython;
  
  prLClick        := OnLClick;
  prRClick        := OnRClick;
  prProps         := Properties;

  with GetPythonEngine do begin
    Py_XINCREF( prLClick );
    Py_XINCREF( prRClick );
    Py_XINCREF( prProps );
  end;
end;

destructor TQUB_Action_Python.Destroy;
begin
  try
    PythonThread.SneakPyProc( prCleanup );
  finally
    inherited;
  end;
end;

procedure TQUB_Action_Python.prCleanup;
begin
  with GetPythonEngine do begin
    Py_XDECREF( prLClick );
    Py_XDECREF( prRClick );
    Py_XDECREF( prProps );
  end;
end;

procedure TQUB_Action_Python.prCallNoArgs(PyFunc :PPyObject);
var
  arglist                      :PPyObject;
  resObj                       :PPyObject;
begin
  with GetPythonEngine do begin
    try
      arglist := PyTuple_New(0);
      resObj := PyEval_CallObject(PyFunc, arglist);
      Py_XDECREF(arglist);
      if resObj <> nil then begin
        Py_XDECREF(resObj);
      end else begin
        PyErr_Print;
      end;
    except
    end;
  end;
end;


procedure TQUB_Action_Python.prDoLClick(OnActionDone :TNotifyEvent);
begin
  if prLClick <> nil then PythonThread.SyncPyProc(prDoLClickPy, True, False, -1);
  if Assigned(OnActionDone) then OnActionDone(nil); // so far sender has been a task
end;

procedure TQUB_Action_Python.prDoLClickPy;
begin
  prCallNoArgs( prLClick );
end;

procedure TQUB_Action_Python.prDoRClick(OnActionDone :TNotifyEvent);
begin
  if prRClick <> nil then PythonThread.SyncPyProc(prDoRClickPy, True, False, -1);
  if Assigned(OnActionDone) then OnActionDone(nil); // so far sender has been a task
end;

procedure TQUB_Action_Python.prDoRClickPy;
begin
  prCallNoArgs( prRClick );
end;

function TQUB_Action_Python.prDoProperties(arg :INode): INode;
var
  arglist                      :PPyObject;
  resObj                       :PPyObject;
begin
  if prProps = nil then begin
    Result := TNode.CreateNull;
    Exit;
  end;

  with GetPythonEngine do begin
    PythonThread.BeginPython;

    try
      arglist := Py_BuildValue('(O)', [PyNode_FromNode( arg )]);
      resObj := PyEval_CallObject(prProps, arglist);
      Py_DECREF(arglist);
      if resObj = nil then begin
        PyErr_Print;
        Result := TNode.CreateNull;
      end else if PyNode_IsNode( resObj ) then begin
        Result := PyNode_ToNode( resObj );
        Py_DECREF(resObj);
      end else begin
        Result := TNode.CreateNull;
        Py_DECREF(resObj);
      end;
    except
      Result := TNode.CreateNull;
    end;

    PythonThread.EndPython;
  end;
end;

// *****************************************************************************
initialization
  QUB_Action_PyType        := nil;

// *****************************************************************************
end.
