unit QUB_ActionSelector;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// home to a category of actions
// presents a button for each, with visibility customizable by named layout
// manages the layout prefs in the registry
// exports actions to python as action = QUB.category['actionname']; action(); action.onRClick(); action.properties(node); action.description; etc

interface

uses

Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls, Buttons,
Registry,
QUB_Action,
QUB_Node,
PythonEngine, Menus;


const

  BUTTON_FONT_SIZE = 8;

type

TQUB_Python_ActionSelector = class;

TQUB_ActionSelector = class(TFrame)

  Caption               :TLabel;
    CaptionPopup: TPopupMenu;
    procedure CaptionPopupPopup(Sender: TObject);
    procedure CaptionMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);

protected

  prPyHandle            :TQUB_Python_ActionSelector; // python surrogate of Self

  prName                :String;
  prActions             :TStringList; // of TQUB_Action
  prButtons             :TList;      // of TSpeedButton; parallel to prActions
  prSmallButtons        :Boolean;

  prActiveLayout        :String;

  function prGetAction(i :Integer): TQUB_Action;
  function prGetActionByName(n :String): TQUB_Action;
  function prGetActionCount: Integer;
  function prGetActionVisible(i :Integer; Layout :String): Boolean;
  procedure prSetActionVisible(i :Integer; Layout :String; Visible :Boolean);
  procedure prSetActiveLayout(Layout :String);
  procedure prSetSmallButtons(x :Boolean);

  procedure prOnBtnClick(Sender :TObject);
  procedure prOnBtnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

  procedure prOnPopupItemClick(Sender :TObject);

  procedure prAddAction(Action :TQUB_Action);    // it will not appear until you next set activelayout

  procedure prRestoreDefaults; virtual;

public

  constructor CreateNamed(AOwner :TComponent; Name :String);
  destructor Destroy; override;

  procedure RestoreDefaults;

  procedure PropertiesRegistry(Read :Boolean); virtual;

  function IndexOf(ActionName :String): Integer;

  procedure AddPyAction( Action :TQUB_Action_Python );

  property Name                      :String            read prName;
  property Actions[i :Integer]       :TQUB_Action       read prGetAction; default;
  property ActionByName[n :String]   :TQUB_Action       read prGetActionByName;
  property ActionCount               :Integer           read prGetActionCount;
  property ActionVisible[i :Integer; Layout :String] :Boolean read prGetActionVisible   write prSetActionVisible;

  property ActiveLayout              :String            read prActiveLayout             write prSetActiveLayout;
  property SmallButtons              :Boolean           read prSmallButtons             write prSetSmallButtons;

  property PyHandle                  :TQUB_Python_ActionSelector  read prPyHandle;

end;


// is this list in use, or just for thematic consistency with the ViewportList in QUB_GUIInterface? -- Chris
TQUB_ActionSelector_List = class(TStringList)
protected


  function prGet(Index :Integer) :TQUB_ActionSelector;
  function prGetByName(Index :String) :TQUB_ActionSelector;


public


  destructor Destroy; override;


  procedure AddActionSelector(ActionSelector :TQUB_ActionSelector; par : TPanel );


  property ActionSelector[Index :Integer]  :TQUB_ActionSelector read prGet; default;
  property ByName[Index :String]           :TQUB_ActionSelector read prGetByName;


end;


TQUB_Python_ActionSelector = class(TPyObject)
protected
  prActionSelector    :TQUB_ActionSelector;

  prActionToAdd       :TQUB_Action_Python;
  procedure prAddButton;
  
public
  function  GetAttr(key : PChar) : PPyObject; override;

  // Mapping services
  function  MpLength : Integer; override;
  function  MpSubscript( obj : PPyObject) : PPyObject; override;
  function Keys( args :PPyObject ): PPyObject; cdecl;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;
  function AddButton( args :PPyObject ): PPyObject; cdecl;

  property ActionSelector  :TQUB_ActionSelector  read prActionSelector   write prActionSelector;
end;


procedure QUB_ActionSelector_InitPython;


implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Viewport_Python,
QUB_Node_Python,
QUB_Main,
QUB_Python_Thread;


const


ACTIONSEL_BUTTON_VPAD = 0;


var

  QUB_ActionSelectorPyType       :TPythonType;




constructor TQUB_ActionSelector.CreateNamed(AOwner :TComponent; Name :String);
var
  PyInst                  :PPyObject;
begin
  inherited Create(AOwner);

  prName            := Name;
  prActions         := TStringList.Create;
  prButtons         := TList.Create;

  Caption.Caption   := Name;
  Caption.Width     := Width;

  PythonThread.BeginPython;
  QUB_ActionSelector_InitPython;
  PyInst                    := QUB_ActionSelectorPyType.CreateInstance;
  prPyHandle                := TQUB_Python_ActionSelector( PythonToDelphi(PyInst) );
  prPyHandle.ActionSelector := Self;
  PyNode_PublishPyObj(PyInst, Name);
  PythonThread.EndPython;
end;

destructor TQUB_ActionSelector.Destroy;
var
  i                 :Integer;
begin
  try
    PropertiesRegistry(False); // Write

    for i := 0 to ActionCount - 1 do begin
      Actions[i].Free;
    end;
    prActions.Free;
    prButtons.Free;

    PythonThread.BeginPython;
    prPyHandle.ActionSelector := nil;
    GetPythonEngine.Py_DECREF( prPyHandle.GetSelf );
    PythonThread.EndPython;
  finally
    inherited;
  end;
end;

procedure TQUB_ActionSelector.RestoreDefaults;
var
  i, j                          :Integer;
begin
  for i := 0 to ActionCount - 1 do begin
    Actions[i].VisLayouts.Clear;

    for j := 0 to InterfaceList.Count - 1 do begin
      Actions[i].VisLayouts.Add( InterfaceList[j].Name );
    end;
  end;

  prRestoreDefaults;

  ActiveLayout := ActiveLayout;
end;

procedure TQUB_ActionSelector.prRestoreDefaults;
begin
end;

procedure TQUB_ActionSelector.prAddAction(Action :TQUB_Action);
var
  OldIx              :Integer;
  Button             :TSpeedButton;
  PopupItem          :TMenuItem;
begin

  QUB_MainForm.AddMenuTask(Name,Action);

  OldIx              := prActions.IndexOf(Action.Name);
  if OldIx >= 0 then begin
    Actions[OldIx].Free;
    prActions.Delete(OldIx);
    TSpeedButton(prButtons[OldIx]).Free;
    prButtons.Delete(OldIx);
    CaptionPopup.Items.Delete(OldIx);
  end;

  prActions.AddObject(Action.Name, Action);

  Button             := TSpeedButton.Create(Self);
  Button.Parent      := Self;
  Button.Visible     := False;
  Button.Flat        := True;
  Button.Anchors     := [akTop, akLeft, akRight];
  Button.Font.Style  := [fsBold];
  Button.Font.Size   := BUTTON_FONT_SIZE;
  if prSmallButtons then begin
    Button.Caption   := Action.ShortName;
  end else begin
    Button.Caption   := Action.Name;
  end;
  Button.Font.Color  := Action.Color;
  Button.Width       := Width;
  Button.Height      := Button.Height + ACTIONSEL_BUTTON_VPAD;
  Button.Hint        := Action.Hint;
  Button.ShowHint    := Action.Hint <> '';
  Button.OnClick     := prOnBtnClick;
  Button.OnMouseUp   := prOnBtnMouseUp;
  prButtons.Add(Button);

  PopupItem          := TMenuItem.Create(Self);
  PopupItem.Caption  := Action.Name;
  PopupItem.OnClick  := prOnPopupItemClick;
  CaptionPopup.Items.Add(PopupItem);
end;

function TQUB_ActionSelector.prGetAction(i :Integer): TQUB_Action;
begin
  Result := TQUB_Action(prActions.Objects[i]);
end;

function TQUB_ActionSelector.IndexOf(ActionName :String): Integer;
begin
  Result           := prActions.IndexOf(ActionName);
end;

procedure TQUB_ActionSelector.AddPyAction( Action :TQUB_Action_Python );
var
  i                :Integer;
begin
  for i := 0 to InterfaceList.Count - 1 do begin
    Action.VisLayouts.Add( InterfaceList[i].Name );
  end;

  prAddAction(Action);
  ActiveLayout := ActiveLayout;
end;

function TQUB_ActionSelector.prGetActionByName(n :String): TQUB_Action;
var
  Ix               :Integer;
begin
  Result           := nil;
  Ix               := prActions.IndexOf(n);
  if Ix >= 0 then begin
    Result := TQUB_Action(prActions.Objects[Ix]);
  end else begin
    for Ix := 0 to ActionCount - 1 do begin
      if Actions[Ix].ShortName = n then begin
        Result := Actions[Ix];
        Break;
      end;
    end;
  end;
end;

function TQUB_ActionSelector.prGetActionCount: Integer;
begin
  Result           := prActions.Count;
end;

function TQUB_ActionSelector.prGetActionVisible(i :Integer; Layout :String): Boolean;
begin
  Result           := Actions[i].VisLayouts.IndexOf(Layout) >= 0;
end;

procedure TQUB_ActionSelector.prSetActionVisible(i :Integer; Layout :String; Visible :Boolean);
var
  VisLayouts       :TStringList;
  Ix               :Integer;
begin
  VisLayouts       := Actions[i].VisLayouts;
  Ix               := VisLayouts.IndexOf(Layout);

  if (Ix >= 0) = Visible then Exit;  // unchanging

  if Visible then begin
    VisLayouts.Add(Layout);
  end else begin
    VisLayouts.Delete( Ix );
  end;

  ActiveLayout     := ActiveLayout; // force repositioning
end;

procedure TQUB_ActionSelector.prSetActiveLayout(Layout :String);
var
  i                  :Integer;
  iVis               :Integer;
  NextTop            :Integer;
  Button             :TSpeedButton;
begin
  prActiveLayout     := Layout;

  NextTop            := Caption.Height;

  Button             := nil;
  
  if SmallButtons then begin


    iVis             := 0;
    for i := 0 to ActionCount - 1 do begin
      Button           := TSpeedButton(prButtons[i]);
      Button.Visible   := ActionVisible[i, Layout];
      if Button.Visible then begin
        Button.Top     := NextTop;
        Button.Width   := Width div 2;
        Button.Left    := (iVis mod 2) * Button.Width;
        if ((iVis mod 2) = 1) then begin
          NextTop        := NextTop + Button.Height;
        end;
        Inc( iVis );
      end;
    end;

    if (iVis mod 2) = 1 then begin   // note that iVis is incremented after Button is assigned
      NextTop          := NextTop + Button.Height;   // (that warning is bogus)
    end;

  end else begin

    for i := 0 to ActionCount - 1 do begin
      Button           := TSpeedButton(prButtons[i]);
      Button.Visible   := ActionVisible[i, Layout];
      if Button.Visible then begin
        Button.Top     := NextTop;
        Button.Left    := 0;
        Button.Width   := Width;
        NextTop        := NextTop + Button.Height;
      end;
    end;

  end;
  
  Height           := NextTop;
 // if NextTop > Caption.Height then begin
    Caption.Visible  := True;
    Height           := Height + Caption.Height; // a little pad for looks
  //end else begin
 //   Caption.Visible  := False;
 // end;
end;

procedure TQUB_ActionSelector.prSetSmallButtons(x :Boolean);
var
  i                  :Integer;
begin
  if x then begin
    for i := 0 to ActionCount - 1 do begin
      TSpeedButton(prButtons[i]).Caption := Actions[i].ShortName;
    end;
  end else begin
    for i := 0 to ActionCount - 1 do begin
      TSpeedButton(prButtons[i]).Caption := Actions[i].Name;
    end;
  end;
  
  prSmallButtons     := x;
  ActiveLayout       := ActiveLayout;
end;

procedure TQUB_ActionSelector.CaptionMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Point   :TPoint;
begin
  inherited;
  Point.X := X;
  Point.Y := Y;
  Point   := Caption.ClientToScreen( Point );
  CaptionPopup.Popup(Point.X,Point.Y);
end;

procedure TQUB_ActionSelector.prOnBtnClick(Sender :TObject);
begin
  Actions[ prButtons.IndexOf(Sender) ].OnLClick(nil);
end;

procedure TQUB_ActionSelector.prOnBtnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbRight then Actions[ prButtons.IndexOf(Sender) ].OnRClick(nil);
end;

procedure TQUB_ActionSelector.CaptionPopupPopup(Sender: TObject);
var
  i                      :Integer;
begin
  for i := 0 to ActionCount - 1 do begin
    CaptionPopup.Items[i].Checked := ActionVisible[i, ActiveLayout];
  end;
end;

procedure TQUB_ActionSelector.prOnPopupItemClick(Sender :TObject);
var
  Item                   :TMenuItem;
  i                      :Integer;
begin
  Item                   := TMenuItem(Sender);
  for i := 0 to ActionCount - 1 do begin
    if Item = CaptionPopup.Items[i] then begin
      ActionVisible[i, ActiveLayout] := not ActionVisible[i, ActiveLayout];
      ActiveLayout := ActiveLayout; // refresh
      Break;
    end;
  end;
end;

procedure TQUB_ActionSelector.PropertiesRegistry(Read :Boolean);
var
  Reg      :TRegistry;
  SubReg   :TRegistry;
  i, j, ix :Integer;
  Action   :TQUB_Action;
  Defined  :TStringList;
begin

  Reg      := TRegistry.Create;
  Defined  := TStringList.Create;

  try

    with Reg do begin
  // foreach action access named subtree -- skip if reading and does not exist
  // action is visible in layout iff subtree has child named layout

  // read: Action.VisLayouts.Clear; .GetValueNames(VisLayouts); remove names with value of False
  // write: .GetValueNames; for x in VisLayouts do x=True, ValueNames.remove(x); for x in ValueNames do if x==True then x=False

      if OpenKey('Software\QUB\ActionSelector ' + Name, True) then begin

        SubReg         := TRegistry.Create;
        SubReg.RootKey := CurrentKey;
        with SubReg do begin

          if Read then begin

            for i := 0 to ActionCount - 1 do begin
              Action := Actions[i];

              if OpenKey(Action.Name, False) then begin
                Action.VisLayouts.Clear;
                GetValueNames(Action.VisLayouts);

                for j := Action.VisLayouts.Count - 1 downto 0 do begin
                  if not ReadBool(Action.VisLayouts[j]) then begin
                    Action.VisLayouts.Delete(j);
                  end;
                end;

                CloseKey;
              end;
            end;

          end else begin // Write

            for i := 0 to ActionCount - 1 do begin
              Action := Actions[i];

              if OpenKey(Action.Name, True) then begin
                Defined.Clear;
                GetValueNames(Defined);

                for j := 0 to Action.VisLayouts.Count - 1 do begin
                  WriteBool(Action.VisLayouts[j], True);
                  ix := Defined.IndexOf(Action.VisLayouts[j]);
                  if ix >= 0 then Defined.Delete(ix); // so Defined will contain only non-visible layouts
                end;
                for j := 0 to Defined.Count - 1 do begin
                  WriteBool(Defined[j], False);
                end;

                CloseKey;
              end;
            end;

          end; // read/write
        end; // subreg

        SubReg.Free;

      end;//Software
      CloseKey;

    end;//with

  finally

    Reg.Free;
    Defined.Free;

  end;

end;


//ActionSelector List **********************************************************
destructor TQUB_ActionSelector_List.Destroy;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      try
        Self[i].Free;
      except
        beep;
      end;
      Delete(i);
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_ActionSelector_List.AddActionSelector(ActionSelector :TQUB_ActionSelector; par : TPanel);
begin
  AddObject(ActionSelector.Name, ActionSelector);
  ActionSelector.Parent := par;
end;

function TQUB_ActionSelector_List.prGet(Index :Integer) :TQUB_ActionSelector;
begin
  if (Index >= 0) and (Index < Count) then Result := TQUB_ActionSelector(Objects[Index]) else Result := nil;
end;

function TQUB_ActionSelector_List.prGetByName(Index :String) :TQUB_ActionSelector;
var
  Idx :Integer;
begin
  Idx := IndexOf(Index);
  if Idx = -1 then Result := nil else Result := Self[Idx];
end;


//******************************************************************************
function TQUB_Python_ActionSelector.GetAttr(key : PChar) : PPyObject;
begin
  if ActionSelector = nil then begin
    Result := inherited GetAttr(key);
    Exit;
  end;

  with GetPythonEngine do begin
    Result := inherited GetAttr(key);
  end;
end;

function TQUB_Python_ActionSelector.MpLength : Integer;
begin
  if ActionSelector = nil then Result := 0 else Result := ActionSelector.ActionCount;
end;

function TQUB_Python_ActionSelector.MpSubscript( obj : PPyObject) : PPyObject;
var
  Action       :TQUB_Action;
begin
  with GetPythonEngine do begin
    if PyString_Check( obj ) then begin
      if ActionSelector = nil then begin
        Result := ReturnNone;
      end else begin
        Action := ActionSelector.ActionByName[ PyString_AsString( obj ) ];
        if Action = nil then begin
          Result := ReturnNone;
        end else begin
          Result := Action.PyAction;
          Py_INCREF( Result );
        end;
      end;
    end else begin
      PyErr_SetString( PyExc_TypeError^, PChar('get Action by name (string)') );
      Result := nil;
    end;
  end;
end;

class procedure TQUB_Python_ActionSelector.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do begin
    AddMethod( 'addButton',    @TQUB_Python_ActionSelector.AddButton,     'ActionSelector.addButton(name, onLclick, onRclick, propertiesFn, hint, description, shortName)'  );
    AddMethod( 'keys',         @TQUB_Python_ActionSelector.Keys,          'ActionSelector.keys()' );
  end;
end;

function TQUB_Python_ActionSelector.Keys( args :PPyObject ): PPyObject;
var
  i      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if ActionSelector = nil then begin
      Result := PyList_New( 0 );
      Exit;
    end;

    Result := PyList_New( 2*ActionSelector.ActionCount );
    for i := 0 to ActionSelector.ActionCount - 1 do begin
      PyList_SetItem( Result, 2*i, PyString_FromString( PChar(ActionSelector[i].Name) ) );
      PyList_SetItem( Result, 2*i+1, PyString_FromString( PChar(ActionSelector[i].ShortName) ) );
    end;
  end;
end;

function TQUB_Python_ActionSelector.AddButton( args :PPyObject ): PPyObject;
// ok ignore kw args because PythonForDelphi doesn't have pyarg_parseTupleAndKeywords
var
  FName, Hint, Description       :PChar;
  ShortName                      :PChar;
  OnLClick, OnRClick, Properties :PPyObject;
begin
  Adjust(@Self);
  if ActionSelector = nil then begin
    Result := GetPythonEngine.ReturnNone;
    Exit;
  end;

  with GetPythonEngine do begin
    FName                     := PChar('');
    OnRClick                  := nil;
    Properties                := nil;
    Hint                      := PChar('');
    Description               := PChar('');
    ShortName                 := PChar('');

    if PyArg_ParseTuple(args, 'sO|OOsss', [@FName, @OnLClick, @OnRClick, @Properties, @Hint, @Description, @ShortName]) <> 0 then begin

       if OnRClick = nil then OnRClick := OnLClick;

      if PyCallable_Check(OnLClick) = 0 then begin
        PyErr_SetString( PyExc_TypeError^, PChar('OnLClick must be callable') );
        Result                := nil;
      end else if PyCallable_Check(OnRClick) = 0 then begin
        PyErr_SetString( PyExc_TypeError^, PChar('OnRClick must be callable') );
        Result                := nil;
      end else if (Properties <> nil) and (PyCallable_Check(Properties) = 0) then begin
        PyErr_SetString( PyExc_TypeError^, PChar('Properties must be callable') );
        Result                := nil;
      end else begin

        prActionToAdd             := TQUB_Action_Python.CreatePy(FName, OnLClick, OnRClick, Properties);
        prActionToAdd.Hint        := Hint;
        prActionToAdd.Description := Description;
        if ShortName <> '' then prActionToAdd.ShortName := ShortName;
        
        PythonThread.SyncGUIProc( prAddButton );

        Result                := ReturnNone;

      end;

    end else begin
      Result                  := nil;
    end;
  end;
end;

//----- From AddButton() - Separate method to be executed in Python GUI thread.
procedure TQUB_Python_ActionSelector.prAddButton;
begin
  ActionSelector.AddPyAction( prActionToAdd );
end;


// *****************************************************************************
procedure QUB_ActionSelector_InitPython;
begin
  if QUB_ActionSelectorPyType <> nil then Exit;

  QUB_ActionSelectorPyType := TPythonType.Create(nil);
  QUB_ActionSelectorPyType.Engine := GetPythonEngine;
  QUB_ActionSelectorPyType.Module := PythonViewport.pmQUBTypes;
  QUB_ActionSelectorPyType.TypeName := 'ActionSelectorType';
  QUB_ActionSelectorPyType.Services.Basic   := [bsGetAttr];
  QUB_ActionSelectorPyType.Services.Mapping := [msLength, msSubscript];
  QUB_ActionSelectorPyType.PyObjectClass := TQUB_Python_ActionSelector;
  QUB_ActionSelectorPyType.Initialize;

  QUB_Action_InitPython;
end;

initialization
  QUB_ActionSelectorPyType        := nil;

// *****************************************************************************
end.
