{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// generates random numbers with various distributions


unit QUB_Deviates;


interface


{
 Returns a uniform random deviate between 0 and 1
 (exclusive of the endpoint values).

 Call with idum a negative integer to initialize;
 thereafter, do not alter idum between successive
 deviates in a sequence.

 RNMX should approximate the largest floating value
 that is less than 1.
}
//function Rand1(var idum :Integer) :Single; forward;


{
 Returns a normally distributed deviate with zero
 mean and unit variance, using rand1(idum) as the
 source of uniform deviates.

 Call with idum a negative integer to initialize;
 thereafter, do not alter idum between successive
 deviates in a sequence.

 Probability transformation law: If y is a function
 of x, then f(y)=f(x)|dx/dy|. If x is uniformly
 distributed, then f(y)=|dx/dy|.
}
//function GaussianDeviate(var idum :Integer) :Single; forward;


{
 Returns an exponentially distributed, positive, random
 deviate of unit mean, i.e. f(x)=exp(-x), using
 rand1(idum) as the source of uniform deviates.

 Dividing the returned value by a constant yields a
 variable y=x/w that has the probability distribution
 f(y)=w*exp(-wx).
}

//function ExponentialDeviate(var idum :Integer) :Single; forward;


{
 Returns an integer value that is drawn from a discrete random
 variable, using rand1(idum) as a source of uniform random
 deviate.

 The continuous interval (0,1) is devided into n
 parts with the length of the ith part equal to the
 probability that the discret variable assumes its ith value.

 By considering each part as a choice, the continuous-valued
 variable becomes a discrete-valued variable.
}

//function DiscreteDeviate(n :Integer; p :TPASingle; var idum :Integer) :Integer; forward;


uses


Math, AllocTypes;


function ExpDev :Double; forward;
function DisDev(n :Integer; p :TPADouble) :Integer; forward;
function DisDevFactor(n :Integer; p :TPADouble; f :Double) :Integer; forward;


implementation


function DisDev(n :Integer; p :TPADouble) :Integer;
var
  u, s  :Double;
  i, n1 :Integer;
begin

  u     := Random;
  i     := 0;
  s     := p[0];
  n1    := n - 1;

  //TO DO: LM. Check the implications of this change
  //while (s < u) and (i < n) do begin
  while (s < u) and (i < n1) do begin
    Inc(i);
    s := s + p[i];
  end;

  Result := i;

end;

function DisDevFactor(n :Integer; p :TPADouble; f :Double) :Integer;
var
  u, s  :Double;
  i, n1 :Integer;
begin

  u     := f * Random;
  i     := 0;
  s     := p[0];
  n1    := n - 1;

  //TO DO: LM. Check the implications of this change
  //while (s < u) and (i < n) do begin
  while (s < u) and (i < n1) do begin
    Inc(i);
    s := s + p[i];
  end;

  Result := i;

end;

function ExpDev :Double;
var
  u :Double;
begin
  repeat
    u := Random;
  until((u > 0) and (u < 1));
  Result := -ln(u);
end;




const


PI = 3.141592654;
Ran1_NTAB = 32;


{
const
Ran1_IA	  = 16807;
Ran1_IM	  = 2147483647;
Ran1_AM	  = (1.0 / Ran1_IM);
Ran1_IQ	  = 127773;
Ran1_IR	  = 2836;
Ran1_NDIV = (1 + (Ran1_IM - 1) / Ran1_NTAB);
Ran1_EPS  = 1.2e-7;
Ran1_RNMX = (1.0 - Ran1_EPS);
}

var


Ran1_iy        :Integer = 0;
//Ran1_iv        :Array[0..Ran1_NTAB - 1] of Integer;



{
function Ran1(var idum :Integer) :Single;
var
  j, k :Integer;
  temp :Single;
begin

  if (idum <= 0) or (Ran1_iy = 0) then begin

    if (-idum < 1) then idum := 1 else idum := -idum;

    for j := Ran1_NTAB + 7 downto 0 do begin

      k := Round(idum / Ran1_IQ);
      idum := Ran1_IA * (idum - k * Ran1_IQ) - Ran1_IR * k;
      if idum < 0 then idum := idum + Ran1_IM;
      if j < Ran1_NTAB then Ran1_iv[j] := idum;

    end;

    Ran1_iy := Ran1_iv[0];

  end;

  k := Round(idum / Ran1_IQ);
  idum := Ran1_IA * (idum - k * Ran1_IQ) - Ran1_IR * k;
  if idum < 0 then idum := idum + Ran1_IM;
  j := Round(Ran1_iy / Ran1_NDIV);
  Ran1_iy := Ran1_iv[j];
  Ran1_iv[j] := idum;

  temp := Ran1_AM * Ran1_iy;

  if (temp > Ran1_RNMX) then Result := Ran1_RNMX else Result := temp;

end;
}


(*
var


GaussianDeviate_iset  :Boolean = False;
GaussianDeviate_gset  :Single;



function GaussianDeviate(var idum :Integer) :Single;
var
  fac, rsq, v1, v2 :Single;
begin

  if not GaussianDeviate_iset then begin

    repeat

      v1  := 2.0 * Rand1(idum) - 1.0;
      v2  := 2.0 * Rand1(idum) - 1.0;
      rsq := v1 * v1 + v2 * v2;

    until ((rsq < 1.0) and (rsq > 0.0));

    fac := sqrt(-2.0 * ln{???}(rsq) / rsq);

    GaussianDeviate_gset := v1*fac;
    GaussianDeviate_iset := True;

    Result := v2*fac;

  end else begin

    GaussianDeviate_iset := False;
    Result := GaussianDeviate_gset;

  end;

end;


function ExponentialDeviate(var idum :Integer) :Single;
var
  u :Single;
begin
  repeat
    u := rand1(idum);
  until(u <> 0);
  Result := -ln(u);
end;


function DiscreteDeviate(n :Integer; p :TPASingle; var idum :Integer) :Integer;
var
  u, s :Single;
  i    :Integer;
begin
  u := Rand1(idum);
  i := 0;
  s := p[0];

  while (s < u) do begin
    Inc(i);
    s := s + p[i];
  end;

  Result := i;

end;
*)



{
function gammln(xx: Double): Double;
const
   stp            = 2.50662827465;
   half           = 0.5;
   one            = 1.0;
   fpf            = 5.5;
   cof            :Array[1..6] of Double = (76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5);
var
   x, y, tmp, ser :Double;
   j              :integer;
begin
   x   := xx;
   y   := xx;
   tmp := x + 5.5;
   tmp := tmp - (x + 0.5) * ln(tmp);
   ser := 1.000000000190015;;
   for j := 1 to 6 do begin
     y := y + 1;
     ser := ser + cof[j] / y;
   end;
   Result := -tmp + ln(2.5066282746310005 * ser / x);
end;
}


{
var


BnlDev_nold  :Integer = -1;
BnlDev_pold  :Single  = -1;
BnlDev_pclog :Single;
BnlDev_plog  :Single;
BnlDev_pc    :Single;
BnlDev_en    :Single;
BnlDev_oldg  :Single;


function BnlDev(pp :Double; n :Integer; var idum :Integer) :Double;
LABEL 1;
var
  am, em, g, angle, p, bnl, sq, t, y :Single;
  j                                  :Integer;
begin

  em := 0;

  if pp <= 0.5 then p := pp else p := 1.0 - pp;

  try

    //The binomial distribution is invariant under changing pp to 1-pp,
    //if we also change the answer to n minus itself;
    //we'll remember to do this below.

    am := n * p;
    //This is the mean of the deviate to be produced.

    if n < 25 then begin
    //Use the direct method while n is not too large. This can require up to 25 calls to ran1.

      bnl := 0.0;
      for j := 1 to n do begin
        t := ran1(idum);
        if t < p then begin
          bnl := bnl + 1.0;
        end;
      end;

    end else if am < 1.0 then begin
    //If fewer than one event is expected out of 25 or more trials,
    //then the distribution is quite accurately Poisson.
    //Use direct Poisson method.

      g := exp(-am);
      t := 1.0;

      for j := 0 to n do begin
        t := t*ran1(idum);
        if t < g then break;
      end;
      if j <= n then bnl := j else bnl := n;

     end else begin
     //Use the rejection method.

       if n <> BnlDev_nold then begin
       //If n has changed, then compute useful quantities
         BnlDev_en    := n;
         BnlDev_oldg  := gammln(BnlDev_en + 1.0);
         BnlDev_nold  := n;
       end;

       if p <> BnlDev_pold then begin
       //If p has changed, then compute useful quantities
         BnlDev_pc    := 1.0 - p;
         BnlDev_plog  := ln(p);
         BnlDev_pclog := ln(BnlDev_pc);
         BnlDev_pold  := p;
       end;

       sq := sqrt(2.0 * am * BnlDev_pc);
       //The following code should by now seem familiar:
       //rejection method with a Lorentzian comparison function

       repeat

         y := 0;
         repeat
           try
             angle := pi * ran1(idum);
             y     := sin(angle) / cos(angle);
             em    := sq * y + am;
           except
             p := p;
           end;
         until((em >= 0.0) and (em < (BnlDev_en + 1.0)));//Reject

         em := trunc(em);//Trick for integer-valued distribution
         t := BnlDev_oldg;
         t := t - gammln(em + 1.0);
         t := t - gammln(BnlDev_en - em + 1.0);
         t := t + em * BnlDev_plog;

         t := t + (BnlDev_en - em) * ln(BnlDev_pc); //BnlDev_pclog

         try
           t := exp(t);
         except
           p := p;
         end;
         t := t * 1.2 * sq * (1.0 + sqr(y));

       until(ran1(idum) <= t);//Reject. This happens about 1.5 times per deviate, on average.

       bnl := em;

    end;

  except

     Result := Round(p * n);
     Exit;

  end;

  if (p <> pp) then bnl := n - bnl;//Remember to undo the symmetry transformation

  Result := bnl;

end;
}


//******************************************************************************
end.
