{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// probably should be called QUB_Staircase_Model
// provides HMM specialization for staircase processes
// used by QUB_Task_Drift


unit QUB_Drift_Model;


interface


uses


Classes,

AllocTypes, Alloc,
qubopt_iface,

QUB_Matrices,

QUB_Q_Model, QUB_Q_Global,
QUB_MIP_Global,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix, QUB_QFS_Channels,
QUB_ExpCond_Parameters;


type


TQUB_Drift_Model               = class(TQUB_Q_Model)
protected


  prOptParams                :TQUB_MIP_Opt_PParams;


  procedure prAllocModelStuff; override;
  procedure prInitModelStuff; override;
  procedure prFreeModelStuff; override;

  function prGet_ChannelCount :Integer; override;


public


  //LRM - Specific:


  //Position units
  UnitCount                  :Integer;
  UnitStateCount             :Integer;
  UsedStateCount             :Integer;
  BlackUnit                  :Integer;
  BlackUnitOffset            :Integer;


  //Summing matrix
  SM                         :TPMDouble;

  //Conductance matrices .......................................................
  B                          :TPAPointer; //The conductance matrices

  //P Vector
  sP0                        :TPADouble;
  dsP0dq                     :TPMDouble;

  //Small A matrix
  sA                         :TPMDouble;

  //Small dA matrices
  sdAdq                      :TPAPointer;


  //A.B.SM powers and their derivatives
  ABSM                       :TPMPointer;
  dABSMdq                    :TPAPointer;

  //Temporary quantities .......................................................
  Vec1E                      :TPAExtended;


  TotalDwellCount            :Integer;
  TotalDataPoints            :Integer;
  Multiplications            :Integer;
  TruncatedJumpCount         :Integer;


  constructor Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_MIP_Opt_PParams);


  function IsPZero(Rate, Index :Integer) :Boolean; override;
  function IsQZero(Rate, Index :Integer) :Boolean; override;

  function Do_fParsToModel :Integer; override;

  function Do_A(t :Double) :Integer; override;
  function Do_dA_dq(t :Double; ExpCond :Integer) :Integer; override;

  function Do_APowers :Integer;
  function Do_dAPowers_dq :Integer;

  function Do_P0 :Integer; override;
  function Do_dP0_dq :Integer; override;


  property OptParams         :TQUB_MIP_Opt_PParams    read prOptParams;


end;


implementation


//******************************************************************************
constructor TQUB_Drift_Model.Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_MIP_Opt_PParams);
begin

  prOptParams := OptParams;

  QMinRate    := prOptParams.ModQMinRate;
  
  inherited Create(Model, DoInitialize, ExpCondParams);

end;



//****************************************************************************//
//                                                                            //
//                             C O N S T R A I N T S                          //
//                                                                            //
//****************************************************************************//
function TQUB_Drift_Model.IsPZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := False;
  if not prModel.Rates[Rate].P[Index] then Exit;
  //Not P-dependent --> not zero


  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].PName[Index]];
  if (Par <> nil) and (Par.Value = 0) and (not Par.IsVar) then begin
    Result := True;
  end;

end;

function TQUB_Drift_Model.IsQZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := True;
  if not prModel.Rates[Rate].Q[Index] then Exit;
  //Not Q-dependent --> zero

  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].QName[Index]];
  if (Par <> nil) then Exit;

  if (Par.Value <> 0) or (Par.IsVar) then begin
    Result := False;
  end;

end;



//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_Drift_Model.Do_fParsToModel :Integer;
begin

  Result := inherited Do_fParsToModel;

end;



//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Drift_Model.prAllocModelStuff;
var
  i, j, k, c, p :Integer;
begin


  inherited;


  //Drift - specific:


  //The class 0 (Black) states represent the unit

  //Count the unit states, that is, the black ones .............................
  UnitStateCount             := 0;
  for k := 0 to prModel.States.StateCount - 1 do begin
    if prModel.States[k].Cls = 0 then Inc(UnitStateCount);
  end;

  //Unit count must be equal to or higher than the max jump order + 1
  if prOptParams.ModRightOnly then begin
    UnitCount                := (prModel.States.StateCount - 1) div UnitStateCount;
    UsedStateCount           := UnitCount * UnitStateCount;
    BlackUnit                := 0;
    BlackUnitOffset          := 0;
  end else begin
    UnitCount                := ((((prModel.States.StateCount - UnitStateCount) div 2) - 1) div UnitStateCount) * 2 + 1;
    UsedStateCount           := UnitCount * UnitStateCount;
    BlackUnit                := UnitCount div 2;
    BlackUnitOffset          := 1;
  end;


  //small P0 ...................................................................
  sP0                        := DAlloc1(UsedStateCount);
  dsP0dq                     := DAlloc2(RateCount, UsedStateCount);


  //Small A ....................................................................
  sA                         := DAlloc2(UsedStateCount, UsedStateCount);


  //Small dAdq .................................................................
  sdAdq                      := PAlloc1(RateCount);
  for k := 0 to RateCount - 1 do begin
    sdAdq[k]                 := DAlloc2(UsedStateCount, UsedStateCount);
  end;


  //Summing matrix .............................................................
  SM                         := DAlloc2(StateCount, StateCount);//Just in case
  for i := 0 to UsedStateCount - 1 do begin
    for j := 0 to UsedStateCount - 1 do begin
      SM[i, j] := 0;
    end;
  end;
  for k := 0 to UnitCount - 1 do begin
    for i := 0 to UnitStateCount - 1 do begin
      SM[i + k*UnitStateCount, i + BlackUnit * UnitStateCount] := 1;
    end;
  end;


  //Conductance matrices .......................................................
  B                          := PAlloc1(UnitCount);
  for c := 0 to UnitCount - 1 do begin
    B[c]                     := DAlloc2(StateCount, StateCount);//Just in case
    //All zero
    for i := 0 to UsedStateCount - 1 do begin
      for j := 0 to UsedStateCount - 1 do begin
        TPMDouble(B[c])[i, j] := 0;
      end;
    end;
    //Diagonal very small
    for j := 0 to UsedStateCount - 1 do begin
      TPMDouble(B[c])[j, j] := 0;//1e-8;//To prevent zero state vector
    end;
    //Class dependent
    for j := 0 to UnitStateCount - 1 do begin
      TPMDouble(B[c])[j + c*UnitStateCount, j + c*UnitStateCount] := 1;
    end;
  end;


  //A.B.SM powers and their derivatives, for each jump .........................
  ABSM                         := PAlloc2(UnitCount, prOptParams.OptMaxPower + 1);
  for c := 0 to UnitCount - 1 do begin
    for p := 0 to prOptParams.OptMaxPower do begin
      ABSM[c, p] := DAlloc2(StateCount, StateCount);//Just in case
    end;
  end;

  dABSMdq                    := PAlloc1(RateCount);
  for k := 0 to RateCount - 1 do begin
    dABSMdq[k]               := PAlloc2(UnitCount, prOptParams.OptMaxPower + 1);
    for c := 0 to UnitCount - 1 do begin
      for p := 0 to prOptParams.OptMaxPower do begin
        TPMPointer(dABSMdq[k])[c, p] := DAlloc2(StateCount, StateCount);//Just in case
      end;
    end;
  end;


  //Temporary quantities .......................................................
  Vec1E                      := EAlloc1(StateCount);

end;

procedure TQUB_Drift_Model.prInitModelStuff;
begin

  inherited;

  MakeInitProbs(prModel, P0, StateCount);

end;


procedure TQUB_Drift_Model.prFreeModelStuff;
var
  k, c, p :Integer;
begin

  inherited;

  //LRM - specific:


  //Small P0 ...................................................................
  FreeMem(sP0);
  sP0 := nil;
  Free2(TPMChar(dsP0dq));
  dsP0dq := nil;


  //Small A ....................................................................
  Free2(TPMChar(sA));
  sA := nil;


  //Small dA ...................................................................
  if sdAdq <> nil then begin
    for k := 0 to RateCount - 1 do begin
      Free2(TPMChar(sdAdq[k]));
    end;
  end;
  FreeMem(sdAdq);
  sdAdq := nil;


  //Summing matrix
  Free2(TPMChar(SM));
  SM := nil;


  //Conductance matrices .......................................................
  if B <> nil then begin
    for c := 0 to UnitCount - 1 do begin
      Free2(TPMChar(B[c]));
    end;
  end;
  FreeMem(B);
  B := nil;


  //A.B.SM powers and their derivatives, for each jump .........................
  if ABSM <> nil then begin
    for c := 0 to UnitCount - 1 do begin
      for p := 0 to prOptParams.OptMaxPower do begin
        Free2(TPMChar(ABSM[c, p]));
      end;
    end;
  end;
  Free2(TPMChar(ABSM));
  ABSM := nil;

  if dABSMdq <> nil then begin
    for k := 0 to RateCount - 1 do begin
      for c := 0 to UnitCount - 1 do begin
        for p := 0 to prOptParams.OptMaxPower do begin
          Free2(TPMChar(TPMPointer(dABSMdq[k])[c, p]));
        end;
      end;
      Free2(TPMChar(dABSMdq[k]));
    end;
  end;
  FreeMem(dABSMdq);
  dABSMdq := nil;


  //Temporary quantities .......................................................
  FreeMem(Vec1E);
  Vec1E := nil;

end;



//****************************************************************************//
//                                                                            //
//                              P    V e c t o r                              //
//                                                                            //
//****************************************************************************//
function TQUB_Drift_Model.Do_P0 :Integer;
var
  i, d :Integer;
begin

  Result := inherited Do_P0;

  if prOptParams.ModRightOnly then begin
    d := 0;
  end else begin
    d := (StateCount - UsedStateCount) div 2;
  end;

  for i := 0 to UsedStateCount - 1 do begin
    sP0[i] := P0[i + d];
  end;

end;

function TQUB_Drift_Model.Do_dP0_dq :Integer;
var
  i, r, d, ec :Integer;
begin

  Result := inherited Do_dP0_dq;

  if prOptParams.ModRightOnly then begin
    d := 0;
  end else begin
    d := (StateCount - UsedStateCount) div 2;
  end;

  for ec := 0 to ExpCondCount - 1 do begin
    for r := 0 to RateCount - 1 do begin
      for i := 0 to UsedStateCount - 1 do begin
        TPMDouble(dP0dq[ec])[r, i] := TPMDouble(dP0dq[ec])[r, i + d];
      end;
    end;
  end;

end;



//****************************************************************************//
//                                                                            //
//                              A    M a t r i x                              //
//                                                                            //
//****************************************************************************//
function TQUB_Drift_Model.Do_A(t :Double) :Integer;
var
  f, i, j :Integer;
begin

  Result := inherited Do_A(t);

  //Copy into small A
  if prOptParams.ModRightOnly then begin
    f := 0;
  end else begin
    f := (StateCount - UsedStateCount) div 2;
  end;
  for i := 0 to UsedStateCount - 1 do begin
    for j := 0 to UsedStateCount - 1 do begin
      sA[i, j] := TPMDouble(A[0])[i + f, j + f];
    end;
  end;

  if Result = 0 then if Do_APowers <> 0 then Result := 1;

end;

function TQUB_Drift_Model.Do_dA_dq(t :Double; ExpCond :Integer) :Integer;
var
  f, k, i, j :Integer;
begin

  Result := inherited Do_dA_dq(t, ExpCond);

  //Use ExpCond !!!

  //Copy into small dA
  if prOptParams.ModRightOnly then begin
    f := 0;
  end else begin
    f := (StateCount - UsedStateCount) div 2;
  end;

  for k := 0 to RateCount - 1 do begin
    for i := 0 to UsedStateCount - 1 do begin
      for j := 0 to UsedStateCount - 1 do begin
        TPMDouble(sdAdq[k])[i, j] := TPMDouble(TPAPointer(dAdq[0])[k])[i + f, j + f];
      end;
    end;
  end;

  if Result = 0 then if Do_dAPowers_dq <> 0 then Result := 1;

end;

//A.B Powers
function TQUB_Drift_Model.Do_APowers :Integer;
var
  c, p :Integer;
begin

  Result := 0;

  //sABSM[c, p] (StateCount, StateCount);
  for c := 0 to UnitCount - 1 do begin
    p := 0;
    MulMatMat(sA, B[c], Mat1, UsedStateCount);
    MulMatMat(Mat1, SM, ABSM[c, p], UsedStateCount);
    for p := 1 to prOptParams.OptMaxPower do begin
      MulMatMat(ABSM[c, p - 1], ABSM[c, p - 1], ABSM[c, p], UsedStateCount);
    end;//MaxPower
  end;//UnitCount

end;

function TQUB_Drift_Model.Do_dAPowers_dq :Integer;
var
  k, c, p :Integer;
begin

  Result := 0;

  //dABdq(RateCount)
  for k := 0 to RateCount - 1 do begin
    //dABdq[k](UnitCount, prOptParams.OptMaxPower + 1);
    for c := 0 to UnitCount - 1 do begin
      p := 0;
      MulMatMat(sdAdq[k], B[c], Mat1, UsedStateCount);
      MulMatMat(Mat1, SM, TPMPointer(dABSMdq[k])[c, p], UsedStateCount);
      for p := 1 to prOptParams.OptMaxPower do begin
        MulMatMat(TPMPointer(dABSMdq[k])[c, p - 1], ABSM[c, p - 1], Mat1, UsedStateCount);
        MulMatMat(ABSM[c, p - 1], TPMPointer(dABSMdq[k])[c, p - 1], Mat2, UsedStateCount);
        SumMatMat(Mat1, Mat2, TPMPointer(dABSMdq[k])[c, p], UsedStateCount);
      end;//MaxPower + 1
    end;//UnitCount
  end;//RateCount

end;

function TQUB_Drift_Model.prGet_ChannelCount :Integer;
var
  Par    :TQUB_ExpCond_PParameter;
begin

  Result := inherited prGet_ChannelCount;

  Par    := ExpCondParams[0].ParameterByName['ChannelCount'];
  if Par <> nil then begin
    if Par.Value > 0 then begin
      try
        Result := Round(Par.Value);
      except
      end;
    end;
  end;

end;


//******************************************************************************
end.
