{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// stores experimental conditions e.g. Voltage and DeadTime
// converts to and from qubtree (mainly for Session files)
// there is a persistent confusion between 'Index' and 'Channel'.

unit QUB_ExpCond_Parameters;


interface


uses


SysUtils, Classes,
QUB_Node,
QUB_Events;


const


QUB_EXPCOND_MAX_PARAMETER                      = 100;


type


TQUB_ExpCond_Parameter                         = record
  Name                                         :String;
  Value                                        :Double;
  Cond_Value                                   :Double;
  Index                                        :Integer;
  IsVar                                        :Boolean;
  Channel                                      :Integer;
end;
TQUB_ExpCond_PParameter                        = ^TQUB_ExpCond_Parameter;


TQUB_ExpCond_Parameters                        = class(TStringList)
protected


  prEvents                                     :TQUB_Events;


  function prGetParameter(Index :Integer) :TQUB_ExpCond_PParameter;
  function prGetParameterByName(Index :String) :TQUB_ExpCond_PParameter;


public


  constructor Create;
  constructor CreateCopy(Params :TQUB_ExpCond_Parameters);
  destructor Destroy; override;


  procedure Clear; override;

  procedure ReadTree(Tree :INode);
  function ToTree: INode;

  function AddPar(SourcePar :TQUB_ExpCond_PParameter) :Boolean; overload;
  function AddPar(Name :String; Value :Double; Index :Integer; IsVar :Boolean) :Boolean; overload;
  function AddPar(Name :String) :Boolean; overload;
  function AddPar(Name :String; Value :Double; Index :Integer; IsVar :Boolean; Channel :Integer) :Boolean; overload;
  function DelPar(Index :Integer) :Boolean; overload;
  function DelPar(Index :String) :Boolean; overload;
  function ChangePar(Index :Integer; Value :Double) :Boolean; overload;
  function ChangePar(Index :String; Value :Double) :Boolean; overload;
  function ChangeParCond(Index :Integer; Cond_Value :Double) :Boolean; overload;
  function ChangeParCond(Index :String; Cond_Value :Double) :Boolean; overload;
  function ChangeParChannel(Index :Integer; Channel :Integer) :Boolean; overload;
  function ChangeParChannel(Index :String; Channel :Integer) :Boolean; overload;
  function ChangeParName(Index :Integer; const Name :String) :Boolean; overload;
  function ChangeParName(Index :String; const Name :String) :Boolean; overload;


  property Parameter      [Index :Integer]     :TQUB_ExpCond_PParameter     read prGetParameter;        default;
  property ParameterByName[Index :String]      :TQUB_ExpCond_PParameter     read prGetParameterByName;

  property Events                              :TQUB_Events                 read prEvents;


end;

//This a list of TQUB_ExpCond_Parameters, one for each experimental condition
TQUB_ExpCond_ParametersList                    = class(TList)
protected


  function prGetParameters(Index :Integer) :TQUB_ExpCond_Parameters;


public


  destructor Destroy; override;


  procedure Clear; override;


  property Parameters      [Index :Integer]    :TQUB_ExpCond_Parameters     read prGetParameters;        default;


end;


//This a list of TQUB_ExpCond_ParametersList, e.g., one for each file
TQUB_ExpCond_ParametersListList                    = class(TList)
protected


  function prGetParametersList(Index :Integer) :TQUB_ExpCond_ParametersList;


public


  destructor Destroy; override;


  procedure Clear; override;


  property ParametersList  [Index :Integer]    :TQUB_ExpCond_ParametersList read prGetParametersList;    default;


end;


implementation


//****************************************************************************//
//                                                                            //
//                            P A R A M E T E R S                             //
//                                                                            //
//****************************************************************************//
constructor TQUB_ExpCond_Parameters.Create;
begin
  prEvents := TQUB_Events.Create;
end;

constructor TQUB_ExpCond_Parameters.CreateCopy(Params :TQUB_ExpCond_Parameters);
var
  p          :Integer;
begin
  prEvents := TQUB_Events.Create;

  for p := 0 to Params.Count - 1 do begin
    AddPar(Params[p]);
  end;
end;

destructor TQUB_ExpCond_Parameters.Destroy;
begin
  try
    Clear;
    prEvents.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_ExpCond_Parameters.Clear;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      Dispose(Parameter[i]);
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_ExpCond_Parameters.ReadTree(Tree :INode);
var
  Par        :TQUB_ExpCond_PParameter;
  ParNode    :INode;
  XNode      :INode;
begin
  ParNode    := Tree.Child;
  while not ParNode.IsNull do begin
    if ParNode.DataCount > 0 then begin
      try
        Par  := ParameterByName[ParNode.Name];
        if Par <> nil then begin
          Par.Value := Double(ParNode.Data[0]);
        end else begin
          AddPar(ParNode.Name, Double(ParNode.Data[0]), -1, False);
          Par := ParameterByName[ParNode.Name{Count - 1}];//Big fucking problem!!! Don't know why though. Lorin.
        end;

        XNode := ParNode.Find('Cond_Value');
        if XNode.DataCount > 0 then begin
          Par.Cond_Value := Double(XNode.Data[0]);
        end;

        XNode := ParNode.Find('ChannelIndex');
        if XNode.DataCount > 0 then begin
          Par.Channel := Integer(XNode.Data[0]);
        end else begin
          Par.Channel := -1;
        end;
      except
      end;
    end;

    ParNode  := ParNode.Sibling;
  end;
  prEvents.RaiseNotifyEvent('OnChange', Self);
end;

function TQUB_ExpCond_Parameters.ToTree: INode;
var
  Par        :TQUB_ExpCond_PParameter;
  ParNode    :INode;
  i          :Integer;
begin
  Result     := TNode.Create('ExpCond');
  ParNode    := TNode.CreateNull;

  for i := 0 to Count - 1 do begin
    Par      := Parameter[i];
    ParNode  := Result.InsertChild(ParNode, Par.Name);
    ParNode.SetData(Par.Value);
    if Par.Value <> Par.Cond_Value then begin
      ParNode['Cond_Value'].SetData(Par.Cond_Value);
    end else begin
      ParNode['Cond_Value'].SetData(Par.Value);
    end;
    if Par.Channel >= 0 then ParNode['ChannelIndex'].SetData(Par.Channel);
  end;
end;

function TQUB_ExpCond_Parameters.AddPar(SourcePar :TQUB_ExpCond_PParameter) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Result := Count >= QUB_EXPCOND_MAX_PARAMETER;
  if Result then Exit;

  New(Par);
  Par.Name       := SourcePar.Name;
  Par.Value      := SourcePar.Value;
  Par.Cond_Value := SourcePar.Value;
  Par.Index      := SourcePar.Index;
  Par.IsVar      := SourcePar.IsVar;
  Par.Channel    := SourcePar.Channel;
  AddObject(Par.Name, TObject(Par));

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.AddPar(Name :String; Value :Double; Index :Integer; IsVar :Boolean) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Result := Count >= QUB_EXPCOND_MAX_PARAMETER;
  if Result then Exit;

  New(Par);
  Par.Name       := Name;
  Par.Value      := Value;
  Par.Cond_Value := Value;
  Par.Index      := Index;
  Par.IsVar      := IsVar;
  Par.Channel    := -1;
  AddObject(Name, TObject(Par));

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.AddPar(Name :String) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Result := Count >= QUB_EXPCOND_MAX_PARAMETER;
  if Result then Exit;

  New(Par);
  Par.Name       := Name;
  Par.Value      := 0;
  Par.Cond_Value := 0;
  Par.Index      := 0;
  Par.IsVar      := False;
  Par.Channel    := -1;
  AddObject(Name, TObject(Par));

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.AddPar(Name :String; Value :Double; Index :Integer; IsVar :Boolean; Channel :Integer) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Result := Count >= QUB_EXPCOND_MAX_PARAMETER;
  if Result then Exit;

  New(Par);
  Par.Name       := Name;
  Par.Value      := Value;
  Par.Cond_Value := Value;
  Par.Index      := Index;
  Par.IsVar      := IsVar;
  Par.Channel    := Channel;
  AddObject(Name, TObject(Par));

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.DelPar(Index :Integer) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin
  if Index < 0 then begin
    Result := True;
    Exit;
  end;

  Par := Parameter[Index];
  Result := Par = nil;
  if Result then Exit;

  Dispose(Par);
  Delete(Index);

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.DelPar(Index :String) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin
  Par := ParameterByName[Index];
  Result := Par = nil;
  if Result then Exit;

  Dispose(Par);
  Delete(IndexOf(Index));

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangePar(Index :Integer; Value :Double) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Par := Parameter[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Value := Value;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangePar(Index :String; Value :Double) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Par := ParameterByName[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Value := Value;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangeParCond(Index :Integer; Cond_Value :Double) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Par := Parameter[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Cond_Value := Cond_Value;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangeParCond(Index :String; Cond_Value :Double) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Par := ParameterByName[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Cond_Value := Cond_Value;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangeParChannel(Index :Integer; Channel :Integer) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Par := Parameter[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Channel := Channel;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangeParChannel(Index :String; Channel :Integer) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Par := ParameterByName[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Channel := Channel;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangeParName(Index :Integer; const Name :String) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin
  Par := Parameter[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Name       := Name;
  Strings[Index] := Name;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.ChangeParName(Index :String; const Name :String) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin
  Par := ParameterByName[Index];
  Result := Par = nil;
  if Result then Exit;

  Par.Name       := Name;
  Strings[IndexOfObject(TObject(Par))] := Name;

  prEvents.RaiseNotifyEvent('OnChange', Self);

end;

function TQUB_ExpCond_Parameters.prGetParameter(Index :Integer) :TQUB_ExpCond_PParameter;
begin

  if (Index < 0) or (Index >= Count) then begin
    Result := nil;
  end else begin
    Result := TQUB_ExpCond_PParameter(Objects[Index]);
  end;

end;

function TQUB_ExpCond_Parameters.prGetParameterByName(Index :String) :TQUB_ExpCond_PParameter;
begin
  Result := Parameter[IndexOf(Index)];
end;

//******************************************************************************
destructor TQUB_ExpCond_ParametersList.Destroy;
begin
  try
    Clear;
  finally
    inherited;
  end;
end;

procedure TQUB_ExpCond_ParametersList.Clear;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      Self[i].Free;
    end;
  finally
    inherited;
  end;
end;

function TQUB_ExpCond_ParametersList.prGetParameters(Index :Integer) :TQUB_ExpCond_Parameters;
begin

  if (Index < 0) or (Index >= Count) then begin
    Result := nil;
  end else begin
    Result := TQUB_ExpCond_Parameters(List[Index]);
  end;

end;

//******************************************************************************
destructor TQUB_ExpCond_ParametersListList.Destroy;
begin
  try
    Clear;
  finally
    inherited;
  end;
end;

procedure TQUB_ExpCond_ParametersListList.Clear;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      Self[i].Free;
    end;
  finally
    inherited;
  end;
end;

function TQUB_ExpCond_ParametersListList.prGetParametersList(Index :Integer) :TQUB_ExpCond_ParametersList;
begin

  if (Index < 0) or (Index >= Count) then begin
    Result := nil;
  end else begin
    Result := TQUB_ExpCond_ParametersList(List[Index]);
  end;

end;


//******************************************************************************
end.
