unit QUB_Exp_Model;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Classes,
AllocTypes, Alloc,
qubopt_iface,
QUB_Matrices,

QUB_Fit_Model,
QUB_Q_Global, QUB_Exp_Global;


type


TQUB_Exp_Model                 = class(TQUB_Fit_Model)
protected


  procedure prAllocModelStuff; override;
  procedure prInitModelStuff; override;
  procedure prFreeModelStuff; override;

  function prGet_cParCount :Integer; override;


public


  StateCount                 :Integer;    //State count --> StateCount-1 exponentials


  //Equilibrium current ........................................................
  Ieq                        :Double;
  dSSdIeq                    :Double;

  //Tau's ......................................................................
  Tau                        :TPADouble;  //Time constants
  dSSdTau                    :TPADouble;

  //Amp's ......................................................................
  Amp                        :TPADouble;  //PreExponentials
  dSSdAmp                    :TPADouble;


  constructor Create(var OptParams :TQUB_Exp_Opt_Params); override;


  function ResetSSandGR :Boolean; override;

  function Do_Mean_t(var mt :Double; t :Double) :Integer; override;
  function Do_dMean_t_dIeq(var dmt :Double; t :Double) :Integer;

  function Do_dMean_t_dTau(var dmt :Double; t :Double; k :Integer) :Integer;
  function Do_dMean_t_dAmp(var dmt :Double; t :Double; k :Integer) :Integer;

  function Do_fParsToModel :Integer; override;

  function Do_InitcPars :Integer; override;

  function Do_dSSdfPar :Integer; override;
  function Do_Varf2Varc :Integer; override;


end;


implementation


uses


QUB_Node;


//******************************************************************************
constructor TQUB_Exp_Model.Create(var OptParams :TQUB_Exp_Opt_Params);
begin

  Self.StateCount            := OptParams.StateCount;

  inherited Create(OptParams);

end;

function TQUB_Exp_Model.ResetSSandGR :Boolean;
var
  k :Integer;
begin

  Result := inherited ResetSSandGR;

  //dSSdIeq
  dSSdIeq  := 0;
  //dSSdTau's, dSSdAmp's
  for k := 0 to StateCount - 2 do begin
    dSSdTau[k] := 0;
    dSSdAmp[k] := 0;
  end;

end;


//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_Exp_Model.Do_fParsToModel :Integer;
var
  i :Integer;
begin

  Result := inherited Do_fParsToModel;

  //cParsToModel
  Ieq := cPars[0];
  for i := 0 to prOptParams.StateCount - 2 do begin
    Tau[i] := Exp(cPars[1 + 2*i + 0]);
    Amp[i] :=     cPars[1 + 2*i + 1];
  end;

end;


//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Exp_Model.prAllocModelStuff;
begin

  inherited;

  Tau                        := DAlloc1(StateCount);
  Amp                        := DAlloc1(StateCount);

  dSSdTau                    := DAlloc1(StateCount);
  dSSdAmp                    := DAlloc1(StateCount);

  //Temporary quantities
  Vec1                       := DAlloc1(StateCount);
  Vec2                       := DAlloc1(StateCount);
  Mat1                       := DAlloc2(StateCount, StateCount);
  Mat2                       := DAlloc2(StateCount, StateCount);

end;

procedure TQUB_Exp_Model.prInitModelStuff;
var
  i :Integer;
begin

  inherited;

  Ieq := prOptParams.Ieq;
  for i := 0 to StateCount - 2 do begin
    Tau[i] := prOptParams.Tau[i];
    Amp[i] := prOptParams.Amp[i];
  end;
end;


procedure TQUB_Exp_Model.prFreeModelStuff;
begin

  FreeMem(Tau);
  Tau := nil;
  FreeMem(Amp);
  Amp := nil;

  FreeMem(dSSdTau);
  dSSdTau := nil;
  FreeMem(dSSdAmp);
  dSSdAmp := nil;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
function TQUB_Exp_Model.Do_InitcPars :Integer;
var
  i :Integer;
begin

  Result := inherited Do_InitcPars;

  cPars[0] := Ieq;
  for i := 0 to prOptParams.StateCount - 2 do begin
    cPars[1 + 2*i + 0] := Ln(Tau[i]);
    cPars[1 + 2*i + 1] :=    Amp[i];
  end;

end;


//****************************************************************************//
//                                                                            //
//                        L O G  -  L I K E L I H O O D                       //
//                                                                            //
//****************************************************************************//

//Current mean
function TQUB_Exp_Model.Do_Mean_t(var mt :Double; t :Double) :Integer;
var
  i :Integer;
begin

  Result := inherited Do_Mean_t(mt, t);

  //mt = Ieq + Sum(i)[Ampi.exp(-t/Taui)]

  mt := Ieq;
  for i := 0 to StateCount - 2 do begin
    mt := mt + Amp[i]*exp(-t/Tau[i]);
  end;

end;

function TQUB_Exp_Model.Do_dMean_t_dIeq(var dmt :Double; t :Double) :Integer;
begin

  Result := 0;

  //mt = Ieq + Sum(i)[Ampi.exp(-t/Taui)]
  //dmt = 1

  dmt := 1;

end;

function TQUB_Exp_Model.Do_dMean_t_dTau(var dmt :Double; t :Double; k :Integer) :Integer;
begin
  Result := 0;

  //mt = Ieq + Sum(i)[Ampi.exp(-t/Taui)]
  //dmt = Ampi*(-t)*ln(Taui)*exp(-t/Taui)
  //dmt = Ampi*t/(Tau^2)*exp(-t/Taui)

  //dmt := -t*Amp[k]*ln(Tau[k])*exp(-t/Tau[k]);
  dmt := t*Amp[k]/(Tau[k]*Tau[k])*exp(-t/Tau[k]);

end;

function TQUB_Exp_Model.Do_dMean_t_dAmp(var dmt :Double; t :Double; k :Integer) :Integer;
begin
  Result := 0;

  //mt = Ieq + Sum(i)[Ampi.exp(-t/Taui)]
  //dmt = exp(-t/Taui)

  dmt := exp(-t/Tau[k]);

end;

function TQUB_Exp_Model.Do_dSSdfPar :Integer;
var
  k :Integer;
begin

  Result := inherited Do_dSSdfPar;

  //dSSdfPar
  dSSdfPar[0] := dSSdIeq;
  for k := 0 to StateCount - 2 do begin
    dSSdfPar[1 + 2*k + 0] := dSSdTau[k]*exp(fPars[1 + 2*k + 0]);//Tau=exp(fPar)
    dSSdfPar[1 + 2*k + 1] := dSSdAmp[k];
  end;

end;

function TQUB_Exp_Model.Do_Varf2Varc :Integer;
var
  k :Integer;
begin

  Result := inherited Do_Varf2Varc;

  //var(Ieq|Tau|Amp) ~= Sqr(d{Ieq|Tau|Amp}/dz) * var(z)
  //dIeq/dz =   1     * 1
  //dTau/dz = dTau/dc * 1 = Exp(c) * 1
  //dAmp/dz =   1     * 1

  //Ieq
  cParsVar[0] := Abs(fParsVar[0, 0]);
  //Tau, Amp
  for k := 1 to fParCount - 1 do begin
    if Odd(k) then begin
      //Tau
      cParsVar[k] := Sqr(Exp(cPars[k])) * Abs(fParsVar[k, k]);
    end else begin
      //Amp
      cParsVar[k] :=                      Abs(fParsVar[k, k]);
    end;
  end;
end;

function TQUB_Exp_Model.prGet_cParCount :Integer;
begin
  Result := inherited prGet_cParCount + 2*StateCount - 1;
end;


//******************************************************************************
end.
