unit QUB_Fit_Model;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Classes,
AllocTypes, Alloc,
qubopt_iface,
QUB_Matrices,

QUB_Q_Global, QUB_Exp_Global;


type


TQUB_Fit_Model                 = class
protected


  prOptParams                  :TQUB_Exp_Opt_PParams;


  procedure prAllocModelStuff; virtual;
  procedure prInitModelStuff; virtual;
  procedure prFreeModelStuff; virtual;

  procedure prAllocOptStuff; virtual;
  procedure prInitOptStuff; virtual;
  procedure prFreeOptStuff; virtual;

  function prGet_cParCount :Integer; virtual;


public


  //Parameters .................................................................

  //Free parameters ............................................................
  fPars                      :TPADouble;  //Free parameters, passed to the optimizer
  fParsVar                   :TPMDouble;  //Free parameters covariance
  fParCount                  :Integer;    //Free parameter count

  //Constrained parameters .....................................................
  cPars                      :TPADouble;  //Constrained parameters
  cParsVar                   :TPADouble;  //Constrained parameters variance
  cParCount                  :Integer;    //Constrained parameter count

  //Optimizer ..................................................................
  Error                      :Integer;
  Iterations                 :Integer;
  FuncCalls                  :Integer;
  GradCalls                  :Integer;
  SS                         :Double;


  //Sum-Of-Squares and its Derivatives
  dSSdfPar                   :TPADouble;  //Derivative of SS w.r.t each free parameter
  dSSdcPar                   :TPADouble;  //Derivative of SS w.r.t. each constrained parameter


  //Temporary quantities .......................................................
  Vec1                       :TPADouble;  //used for calculations
  Vec2                       :TPADouble;
  Mat1                       :TPMDouble;
  Mat2                       :TPMDouble;


  constructor Create(var OptParams :TQUB_Exp_Opt_Params); virtual;
  destructor Destroy; override;


  function ResetSSandGR :Boolean; virtual;

  function Do_Mean_t(var mt :Double; t :Double) :Integer; virtual;

  function Do_fParsToModel :Integer; virtual;

  function Do_InitfPars :Integer; virtual;
  function Do_InitcPars :Integer; virtual;

  function Do_dSSdfPar :Integer; virtual;
  function Do_Varf2Varc :Integer; virtual;


  property OptParams         :TQUB_Exp_Opt_PParams    read prOptParams;


end;


implementation


uses


QUB_Node;


//******************************************************************************
constructor TQUB_Fit_Model.Create(var OptParams :TQUB_Exp_Opt_Params);
begin

  inherited Create;


  prOptParams                := @OptParams;

  cParCount                  := prGet_cParCount;

  //First part, model-related quantities
  prAllocModelStuff;

  //Second part, optimization-related quantities
  prAllocOptStuff;

  //Initialize the model-related quantities
  prInitModelStuff;

  //Initialize optimization-related quantities
  prInitOptStuff;

end;

destructor TQUB_Fit_Model.Destroy;
begin
  try

    prFreeModelStuff;
    prFreeOptStuff;

  finally
    inherited;
  end;

end;

function TQUB_Fit_Model.ResetSSandGR :Boolean;
begin

  Result := False;

  SS := 0;
  
end;


//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_Fit_Model.Do_fParsToModel :Integer;
var
  i :Integer;
begin
  Result := 0;

  //fParsTocPars
  for i := 0 to cParCount - 1 do begin
    cPars[i] := fPars[i];
  end;

  //cParsToModel

end;


//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Fit_Model.prAllocModelStuff;
begin
end;

procedure TQUB_Fit_Model.prInitModelStuff;
begin
end;


procedure TQUB_Fit_Model.prFreeModelStuff;
begin

  //Temporary quantities .......................................................
  FreeMem(Vec1);
  Vec1 := nil;
  FreeMem(Vec2);
  Vec2 := nil;
  Free2(TPMChar(Mat1));
  Mat1 := nil;
  Free2(TPMChar(Mat2));
  Mat2 := nil;

end;


//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Fit_Model.prAllocOptStuff;
var
  i, j :Integer;
begin

  //Constrained parameters .....................................................
  cParCount                  := prGet_cParCount;
  cPars                      := DAlloc1(cParCount);
  cParsVar                   := DAlloc1(cParCount);

  //Free parameters ............................................................
  fPars                      := DAlloc1(cParCount);
  fParsVar                   := DAlloc2(cParCount, cParCount);
  for i := 0 to cParCount - 1 do begin
    fPars[i] := 0;//new. OK?
    for j := 0 to cParCount - 1 do begin
      fParsVar[i, j] := 0;
    end;
  end;
  for j := 0 to cParCount - 1 do begin
    fParsVar[j, j] := 1;
  end;

  //Derivatives
  dSSdfPar                   := DAlloc1(cParCount);
  dSSdcPar                   := DAlloc1(cParCount);

end;

procedure TQUB_Fit_Model.prInitOptStuff;
begin

  //Make initial cPars .........................................................
  Do_InitcPars;

  //Free Parameters ............................................................
  fParCount := cParCount;

  //Init fPars .................................................................
  Do_InitfPars;

end;

procedure TQUB_Fit_Model.prFreeOptStuff;
begin

  //Free parameters ............................................................
  FreeMem(fPars);
  fPars := nil;
  Free2(TPMChar(fParsVar));
  fParsVar := nil;

  //Constrained parameters .....................................................
  FreeMem(cPars);
  cPars := nil;
  FreeMem(cParsVar);
  cParsVar := nil;

  //Derivatives ................................................................
  FreeMem(dSSdfPar);
  dSSdfPar := nil;
  FreeMem(dSSdcPar);
  dSSdcPar := nil;

end;


//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
function TQUB_Fit_Model.Do_InitcPars :Integer;
begin

  Result := 0;

end;

function TQUB_Fit_Model.Do_InitfPars :Integer;
var
  i :Integer;
begin

  Result := 0;

  for i := 0 to cParCount - 1 do begin
    fPars[i] := cPars[i];
  end;

end;


//****************************************************************************//
//                                                                            //
//                        L O G  -  L I K E L I H O O D                       //
//                                                                            //
//****************************************************************************//

//Current mean
function TQUB_Fit_Model.Do_Mean_t(var mt :Double; t :Double) :Integer;
begin
  Result := 0;
end;

function TQUB_Fit_Model.Do_dSSdfPar :Integer;
begin
  Result := 0;
end;

function TQUB_Fit_Model.Do_Varf2Varc :Integer;
begin
  Result := 0;
end;

function TQUB_Fit_Model.prGet_cParCount :Integer;
begin
  Result := 0
end;


//******************************************************************************
end.
