{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Interface;
{
    Interfaces at one point had specific functionality.  Now the only difference
  is the default window layouts.  7 descendant classes have been removed and
  the methods DockViewportsDefault,DoPropertiesDisplayWriteDefaultViewport have
  been changed to include specific to the interface functionality.  This is prep
  work for allowing user configurable layouts and 'tutorial' layouts.
    Generic quality of layout should be extended to these two methods with hard
  coded defaults as data parameters.  Interface references would also need to be
  array indices with possibility of user interfaces add / remove.
} {4-20-2004 JB}

interface


uses


Windows, SysUtils, Classes, Messages, Buttons, Controls, ExtCtrls, Forms, Registry,

RegistryInterface,
QUB_Global,
QUB_Events,
QUB_Viewport;


const


INTERFACE_NONE                      = -1;
INTERFACE_ACQUISITION               = 0;
INTERFACE_SIMULATION                = 1;
INTERFACE_PREPROCESSING             = 2;
INTERFACE_MODELING                  = 3;
INTERFACE_SCRIPTS                   = 4;
INTERFACE_RESULTS                   = 5;
INTERFACE_MACROSCOPIC               = 6;

INTERFACE_USER                      = 7;  // start with # 7
INTERFACE_COUNTUSER                 = 8;
// Allow up to 8 user interfaces.  Increase this ? -> also add controls to
// QUB_UserInterfaceProperties form, and arrays in its FormCreate

INTERFACE_DEFAULT                   = INTERFACE_SIMULATION;


type


TQUB_Interface = class(TObject)
protected


  prActive               :Boolean;
  prFloating             :Boolean;
  prCanDragDock          :Boolean;

  prName                 :String;
  prID                   :Integer;

  prViewportList         :TQUB_Viewport_List;
  //prServers              :TQUB_ServerViewForm;

  prEvents               :TQUB_Events;

  prInterfaceButton      :TSpeedButton;

  prWorkSpace            :TPanel;

  prDontRegistry         :Boolean;


  procedure prSetFloating(NewFloating :Boolean);

  function prPercentsToBounds(l, t, w, h :Single) :TRect;

  procedure prSetViewportsInterfaceName;

  //Event Handlers
  procedure DoOnActivate;
  procedure DoOnDeActivate;

  //Properties / Registry
  procedure DoProperties(Read :Boolean); //virtual;
  procedure DoPropertiesDisplay(Read :Boolean);
  procedure DoPropertiesViewports(Read :Boolean; Float :Boolean);
  procedure DoPropertiesDocking(Read :Boolean);
  procedure DoPropertiesDisplayWriteDefaultViewport(Viewport :TQUB_Viewport; Reg :TRegistry);

  //Docking
  procedure DockViewports(Float :Boolean);
  procedure DockViewportsDefault; //virtual
  procedure DockViewportsCommon(vwTop, vwTopRight, vwBottom, vwBottomRight : TQUB_Viewport );


public


  constructor Create(const Name :String; const id :Integer; btn : TSpeedButton ); //virtual;
  destructor Destroy; override;


  procedure Activate;
  procedure DeActivate;

  procedure DoPropertiesDisplayCheckDefault;
  procedure DoPropertiesDisplayWriteDefault;

  property Active                          :Boolean                     read prActive;
  property Floating                        :Boolean                     read prFloating                 write prSetFloating;
  property CanDragDock                     :Boolean                     read prCanDragDock;

  property Name                            :String                      read prName                     write prName;
  property ID                              :Integer                     read prID                       write prID;

  property Events                          :TQUB_Events                 read prEvents;

  property InterfaceButton                 :TSpeedButton                read prInterfaceButton          write prInterfaceButton;

  property WorkSpace                       :TPanel                      read prWorkSpace                write prWorkSpace;

  property ViewportList                    :TQUB_Viewport_List          read prViewportList             write prViewportList;

end;


TQUB_Interface_List = class(TStringList)
protected


  function prGetInterface(Index :Integer) :TQUB_Interface;
  function prGetInterfaceByName(const Name :String) :TQUB_Interface;


public


  destructor Destroy; override;


  procedure AddInterface(InterfaceObject :TQUB_Interface);


  property InterfaceObject[Index      :Integer] :TQUB_Interface read prGetInterface; default;
  property InterfaceByName[const Name :String]  :TQUB_Interface read prGetInterfaceByName;


end;


implementation


uses


QUB_GUIInterface,
QUB_ActionSelector;


//******************************************************************************
constructor TQUB_Interface.Create(const Name :String; const id : Integer; btn : TSpeedButton);
begin
  inherited Create;

  prActive                := False;
  prFloating              := True;
  prCanDragDock           := False;

  prName                  := Name;

  prEvents                := TQUB_Events.Create;

  //Restore specific properties from registry
  DoProperties(True{Read});

  prID := id;

  prInterfaceButton := btn;

end;

destructor TQUB_Interface.Destroy;
begin
  try

    //Store specific properties to registry
    DoProperties(False{Read});

    prEvents.Free;

  finally

    inherited;

  end;
end;


//****************************************************************************//
//                                                                            //
//                         S E T    F L O A T I N G                           //
//                                                                            //
//****************************************************************************//
procedure TQUB_Interface.prSetFloating(NewFloating :Boolean);
begin
  if (NewFloating = Floating) or (not Active) then begin
    prFloating := NewFloating;
    Exit;
  end;

  //Store Display Properties to registry
  DoPropertiesDisplay(False{Read});

  //Change Floating Flag
  prFloating := NewFloating;

  DoPropertiesDisplay(True{Read});

  //This is necessary to force some redrawing ...
  WorkSpace.DockManager.ResetBounds(True);

end;


//****************************************************************************//
//                                                                            //
//                   A C T I V A T E  /  D E A C T I V A T E                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Interface.Activate;
begin

  prActive := True;

  DoOnActivate;

  prSetViewportsInterfaceName;

  //Restore its own properties
  DoPropertiesDisplay(True{Read});

  //This is necessary to force some redrawing ...
  WorkSpace.DockManager.ResetBounds(True);

end;

procedure TQUB_Interface.DeActivate;
begin

  prActive := False;

  prSetViewportsInterfaceName;

  //Store its own properties
  DoPropertiesDisplay(False{Read});

  DoOnDeActivate;

end;


//****************************************************************************//
//                                                                            //
//                         D O C K    V I E W P O R T S                       //
//                                                                            //
//****************************************************************************//
procedure TQUB_Interface.DockViewports(Float :Boolean);
var
  Viewport :TQUB_Viewport;
  i        :Integer;
begin

  for i := 0 to ViewportList.Count - 1 do begin

    Viewport := ViewportList[i];

    if Float then begin
    //Floating

      //Un-Parent Viewport
      if not Viewport.Floating then begin
        Viewport.ManualFloat(Viewport.FloatBounds);
      end;

      Viewport.Parent   := nil;
      Viewport.DragKind := dkDrag;
      Viewport.DragMode := dmManual;

    end else begin
    //Docked

      ///Parent Viewport
      Viewport.Parent   := WorkSpace;
      Viewport.DragKind := dkDock;
      Viewport.DragMode := dmAutomatic;

    end;

  end;

end;

//----- Most interfaces use a top, bottom view split, with another view possibly to the right of each.
procedure TQUB_Interface.DockViewportsCommon(vwTop, vwTopRight, vwBottom, vwBottomRight : TQUB_Viewport );
var
  i :Integer;
begin
  try
    //-- Top
    if vwTop <> nil Then begin
      vwTop.Visible   := True;
      vwTop.ManualDock(Workspace, nil, alTop);
      //-- Top Right
      if vwTopRight <> nil Then Begin
        vwTopRight.Visible := True;
        vwTopRight.ManualDock(Workspace, vwTop, alRight);
      end;
    end;
    //--Bottom
    if vwBottom <> nil Then begin
      vwBottom.Visible   := True;
      vwBottom.ManualDock(Workspace, nil, alBottom);
      //-- Bottom Right
      if vwBottomRight <> nil Then Begin
        vwBottomRight.Visible := True;
        vwBottomRight.ManualDock(Workspace, vwBottom, alRight);
      end;
    end;

    for i := 0 to ViewportList.Count - 1 do begin
      if ViewportList[i].Floating then ViewportList[i].ManualDock(Workspace, nil, alBottom);
    end;
    
  except

  end;
end;

procedure TQUB_Interface.DockViewportsDefault;
var
  Viewport :TQUB_Viewport;
  i        :Integer;
begin
  if          prID = INTERFACE_ACQUISITION   then begin
    DockViewportsCommon(ViewportData, ViewportProtocol, ViewportAcqInput, ViewportAcqOutput);
    ViewportAcqSShot.Visible  := True;
    ViewportAcqSShot.ManualDock(Workspace, ViewportAcqInput, alLeft);
  end else if prID = INTERFACE_SIMULATION    then begin
    DockViewportsCommon(ViewportData, ViewportProtocol, ViewportModel, ViewportReport);
  end else if prID = INTERFACE_PREPROCESSING then begin
    DockViewportsCommon(ViewportData, nil, ViewportHistogram, ViewportReport);
  end else if prID = INTERFACE_MACROSCOPIC   then begin
    DockViewportsCommon(ViewportData, ViewportDataOverlap, ViewportModel, ViewportReport);
  end else if prID = INTERFACE_MODELING      then begin
    DockViewportsCommon(ViewportData, ViewportResults, ViewportModel, ViewportReport);
  end else if prID = INTERFACE_RESULTS       then begin
    DockViewportsCommon(ViewportData, ViewportResults, ViewportModel, ViewportReport);
  end else if prID = INTERFACE_SCRIPTS       then begin
    //ViewportScripts.Visible := True;
    //ViewportScripts.ManualDock(Workspace, nil, alTop);
    DockViewportsCommon(ViewportPython, nil, ViewportReport, nil);
  end else begin
    //-- executes when no Interface - first time run?  no value in reg?
    for i := 0 to ViewportList.Count - 1 do begin

      Viewport := ViewportList[i];

      case i of
        0: Viewport.ManualDock(Workspace, nil, alTop);
        1: Viewport.ManualDock(Workspace, nil, alBottom);

        2: Viewport.ManualDock(Workspace, ViewportList[0], alRight);
        3: Viewport.ManualDock(Workspace, ViewportList[1], alRight);

        4: Viewport.ManualDock(Workspace, ViewportList[0], alBottom);
        5: Viewport.ManualDock(Workspace, ViewportList[2], alBottom);
        6: Viewport.ManualDock(Workspace, ViewportList[1], alBottom);
        7: Viewport.ManualDock(Workspace, ViewportList[3], alBottom);

        else Viewport.ManualDock(Workspace, nil, alBottom);

      end;//case i

    end;//for
  end;//prID

end;


//****************************************************************************//
//                                                                            //
//                              D O     E V E N T S                           //
//                                                                            //
//****************************************************************************//
procedure TQUB_Interface.DoOnActivate;
begin
  Events.RaiseNotifyEvent('OnActivate', Self);
end;

procedure TQUB_Interface.DoOnDeActivate;
begin
  Events.RaiseNotifyEvent('OnDeActivate', Self);
end;

//Properties / Registry ********************************************************

//****************************************************************************//
//                                                                            //
//                  S P E C I F I C    P R O P E R T I E S                    //
//                                                                            //
//****************************************************************************//
//Store/Restore the specific properties, at OnCreate, and OnDestroy
procedure TQUB_Interface.DoProperties(Read :Boolean);
begin
  //override if necessary
end;


//****************************************************************************//
//                                                                            //
//                   D I S P L A Y    P R O P E R T I E S                     //
//                                                                            //
//****************************************************************************//
//Store/Restore the display properties, at OnActivate, OnDeActivate and OnFloating
procedure TQUB_Interface.DoPropertiesDisplay(Read :Boolean);
begin

  prCanDragDock := True;

  try

    if Read then begin
    //Read

      DockViewports(Floating);

    end;

    DoPropertiesViewports(Read, Floating);

  finally

    prCanDragDock := False;

  end;

end;

//****************************************************************************//
//                                                                            //
//                   V I E W P O R T   P R O P E R T I E S                    //
//                                                                            //
//****************************************************************************//
//Store/Restore the viewport properties
procedure TQUB_Interface.DoPropertiesViewports(Read :Boolean; Float :Boolean);
var
  Viewport :TQUB_Viewport;
  i        :Integer;
begin

  WorkSpace.Visible := False;

  try

    //At this point, Viewports have the right parentship (nil / WorkSpace)
    for i := 0 to ViewportList.Count - 1 do begin
      ViewportList[i].DoPropertiesDisplay(Read, Float);
    end;

    if not Float then begin
    //Dock

      //Undock Viewports that are UnDocked
      for i := 0 to ViewportList.Count - 1 do begin
        Viewport := ViewportList[i];

        if Viewport.UnDocked then begin
        //Viewport is not Docked
          if not Viewport.Floating then begin
          //Viewport is not Floating yet
            Viewport.ManualFloat(Viewport.UnDockBounds);
          end;
          Viewport.Parent     := nil;
          Viewport.DragKind   := dkDock;
          Viewport.BoundsRect := Viewport.UnDockBounds;
        end;

      end;

      DoPropertiesDocking(Read);

    end;

  finally

    WorkSpace.Visible := True;

  end;
end;


//****************************************************************************//
//                                                                            //
//                    D O C K I N G    P R O P E R T I E S                    //
//                                                                            //
//****************************************************************************//
//Store/Restore the docking info
procedure TQUB_Interface.DoPropertiesDocking(Read :Boolean);
var
  Stream   :TMemoryStream;
  Reg      :TRegistry;
  DockInfo :Boolean;
begin

  Reg      := TRegistry.Create;
  Stream   := nil;
  DockInfo := False;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Interface ' + Name, True) then begin
        if Read then begin
        //Read

          if ValueExists('Docking Data') then begin
            if ValueExists('Docking Data Size') then begin

              try
                Stream := TMemoryStream.Create;
                Stream.SetSize(ReadInteger('Docking Data Size'));
                ReadBinaryData('Docking Data', Stream.Memory^, Stream.Size);
                Workspace.DockManager.LoadFromStream(Stream);
                //Apparently, loading two times solves some malfunctioning ...
                Stream.Position := 0;
                Workspace.DockManager.LoadFromStream(Stream);
                DockInfo := True;
              except
              end;

            end;
          end;

          if not DockInfo then begin

            DockViewportsDefault;

          end;

        end else begin
        //Write

          try
            Stream := TMemoryStream.Create;
            Workspace.DockManager.SaveToStream(Stream);
            WriteInteger('Docking Data Size', Stream.Size);
            WriteBinaryData('Docking Data', Stream.Memory^, Stream.Size);
          except
          end;

        end;

      end;//Software
      CloseKey;

    end;//with

  finally

    Stream.Free;
    Reg.Free;

  end;

end;


//Default Registry
procedure TQUB_Interface.DoPropertiesDisplayCheckDefault;
var
  Reg :TRegistry;
  f   :Boolean;
begin
  Reg := TRegistry.Create;

  try

    with Reg do begin
      f := False;
      if OpenKey('Software\QUB\Interface ' + Name, True) then begin
        f := ValueExists('Floating Windows');//indicates the transition to floating windows
        WriteBool('Floating Windows', True);
      end;
      CloseKey;

      if not f then DoPropertiesDisplayWriteDefault;

    end;//with

  finally

    Reg.Free;

  end;

end;

procedure TQUB_Interface.DoPropertiesDisplayWriteDefault;
var
  Viewport :TQUB_Viewport;
  Reg      :TRegistry;
  i        :Integer;
begin
  Reg := TRegistry.Create;

  try

    with Reg do begin

      //Delete docking info
      if OpenKey('Software\QUB\Interface ' + Name, True) then begin
        for i := 0 to ViewportList.Count - 1 do begin
          DeleteKey('Viewport ' + ViewportList[i].Name);
        end;
        DeleteValue('Docking Data');
        DeleteValue('Docking Data Size');
      end;
      CloseKey;
      for i := 0 to ViewportList.Count - 1 do begin
         WorkSpace.DockManager.RemoveControl(ViewportList[i]);
      end;

      //Write default registry for viewports
      for i := 0 to ViewportList.Count - 1 do begin
        Viewport := ViewportList[i];
        if OpenKey('Software\QUB\Interface ' + Name + '\Viewport ' + Viewport.Name, True) then begin
          DoPropertiesDisplayWriteDefaultViewport(Viewport, Reg);
        end;
        CloseKey;
      end;

    end;//with

  finally

    Reg.Free;

  end;
end;

//----- Write interface defaults - previously part of descendant classes
procedure TQUB_Interface.DoPropertiesDisplayWriteDefaultViewport(Viewport :TQUB_Viewport; Reg :TRegistry);
var
  lWrite : Boolean;   // Initialize to false, set to true if written as a special case.
  cVw : String;       // View Name to check
begin
  lWrite := False;
  cVw := Viewport.Name;

  //----- Functionality specific to each Interface
  if prID = INTERFACE_ACQUISITION Then Begin
    if cVw = 'Acquisition SShot'  then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0, 50, 15, 50) );
    if cVw = 'Data'               then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0,  0, 60, 50) );
    if cVw = 'Acquisition Protocol' then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 60,  0, 40, 50) );
    if cVw = 'Acquisition Input'  then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(15, 50, 51, 50) );
    if cVw = 'Acquisition Output' then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(66, 50, 34, 50) );
  end else if prID = INTERFACE_SIMULATION Then Begin
    if cVw = 'Data'               then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0,  0, 60, 70) );
    if cVw = 'Histogram'          then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(60,  0, 40, 70) );
    if cVw = 'Model'              then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0, 70, 50, 30) );
    if cVw = 'Report'             then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(50, 70, 50, 30) );
  end else if prID = INTERFACE_PREPROCESSING Then Begin
    if cVw = 'Data'               then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0,  0,100, 80) );
    if cVw = 'Histogram'          then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0, 90, 60, 20) );
    if cVw = 'Report'             then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(60, 80, 40, 20) );
  end else if prID = INTERFACE_MODELING Then Begin
    if cVw = 'Data'               then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0,  0, 70, 70) );
    if cVw = 'Model'              then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0, 70, 70, 30) );
    if cVw = 'Results'            then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(70,  0, 30, 70) );
    if cVw = 'Report'             then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(70, 70, 30, 30) );
  end else if prID = INTERFACE_SCRIPTS Then Begin
    if cVw = 'Scripts'            then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0,  0, 50, 70) );
    if cVw = 'Python Scripts'     then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(50,  0, 50, 70) );
    if cVw = 'Report'             then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0, 70,100, 30) );
  end else if prID = INTERFACE_RESULTS Then Begin
    if cVw = 'Data'               then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0,  0, 70, 70) );
    if cVw = 'Model'              then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0, 70, 70, 30) );
    if cVw = 'Results'            then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(70,  0, 30, 70) );
    if cVw = 'Report'             then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(70, 70, 30, 30) );
  end else if prID = INTERFACE_MACROSCOPIC Then Begin
    if cVw = 'Data'               then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0,  0, 60, 70) );
    if cVw = 'Data Overlap'       then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(60,  0, 40, 70) );
    if cVw = 'Model'              then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds( 0, 70, 60, 30) );
    if cVw = 'Report'             then lWrite := ViewportWriteRegistry( reg, True, prPercentsToBounds(60, 70, 40, 30) );
end;

  //----- if lWrite is still False, none of the cases above applied, use default.
  if not lWrite Then
    ViewportWriteRegistry( reg, False, prPercentsToBounds(40, 40, 20, 20) );
end;

procedure TQUB_Interface.prSetViewportsInterfaceName;
var
  i :Integer;
begin
  for i := 0 to ViewportList.Count - 1 do begin
    ViewportList[i].InterfaceName := Name;
  end;
  for i := 0 to ActionsList.Count - 1 do begin
    ActionsList[i].ActiveLayout   := Name;
  end;
end;

function TQUB_Interface.prPercentsToBounds(l, t, w, h :Single) :TRect;
var
  r :TRect;
begin
  r.TopLeft     := WorkSpace.ClientToScreen(Point(0, 0));
  r.BottomRight := WorkSpace.ClientToScreen(Point(WorkSpace.ClientWidth, WorkSpace.ClientHeight));
  with Result do begin
    Left   := r.Left + Round(l * (r.Right  - r.Left) / 100);
    Top    := r.Top  + Round(t * (r.Bottom - r.Top)  / 100);
    Right  := Left   + Round(w * (r.Right  - r.Left) / 100);
    Bottom := Top    + Round(h * (r.Bottom - r.Top)  / 100);
  end;
end;


//Interface List ***************************************************************
destructor TQUB_Interface_List.Destroy;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      Self[i].Free;
      Delete(i);
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_Interface_List.AddInterface(InterfaceObject :TQUB_Interface);
begin
  AddObject(InterfaceObject.Name, InterfaceObject);
end;

function TQUB_Interface_List.prGetInterface(Index :Integer) :TQUB_Interface;
begin
  if (Index >= 0) and (Index < Count) then Result := TQUB_Interface(Objects[Index]) else Result := nil;
end;

function TQUB_Interface_List.prGetInterfaceByName(const Name :String) :TQUB_Interface;
var
  i      :Integer;
begin
  Result := nil;
  for i := 0 to Count - 1 do begin
    if TQUB_Interface(Objects[i]).Name = Name then begin
      Result := TQUB_Interface(Objects[i]);
      Break;
    end;
  end;
end;

//******************************************************************************
end.

