unit QUB_Kalman;
{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
{

                               Qub Kalman Filter
                                     unit

(the Kalman filter is the baseline-tracking component of IdlBase)

(C) 2002 Lorin Milescu
}


//01-21-2002   Creation date



{$ASSERTIONS OFF}


interface


uses


AllocTypes, Alloc;


const


KALMAN_SMALL        = 1e-30;


type


TQUB_KalmanFilter   = class(TObject)
protected


  prDimX            :Integer;
  prDimY            :Integer;


  function prAllocMemory :Boolean;
  function prFreeMemory :Boolean;


public


  constructor Create(DimX, DimY :Integer); virtual;
  destructor Destroy; override;


  function Forward :Boolean;
  function Backward :Boolean;


end;




//The x and y scalar versions
procedure KalmanForward(out x{post x}, p{post p} :Double; xp{prior x}, pp{prior p}, nm{meas noise}, np{proc noise}, y{data} :Double); forward;


//The x vectorial, y scalar versions
procedure KalmanForward_2StateX(x{post x} :TPADouble; p{post p} :TPMDouble; xp{prior x} :TPADouble; pp{prior p} :TPMDouble; nm{meas noise} :Double; np{proc noise} :TPMDouble; a :TPMDouble; c{measurement}, k{gain} :TPADouble; y{data} :Double); forward;


//Multiplies two 2x2 matrices
procedure MulMat22_22(m1, m2, m3 :TPMDouble); forward;

//Multiplies two 2x2 matrices, the second one being transposed
procedure MulMat22_22T(m1, m2, m3 :TPMDouble); forward;

//multiplies a 2x2 matrix with a 2 vector
procedure MulMat22_Vec2(m1 :TPMDouble; v1, v2 :TPADouble); forward;

//multiplies two 2 vectors, the second being transposed, resulting in a 2x2 matrix
procedure MulVec22T(v1, v2 :TPADouble; m1 :TPMDouble); forward;

//multiplies a 2x2 matrix with a scalar
procedure MulMat22_S(m1 :TPMDouble; s1 :Double; m2 :TPMDouble); forward;

//Adds two 2x2 matrices
procedure AddMat22(m1, m2, m3 :TPMDouble); forward;

//subtracts m2 from m1, two 2x2 matrices
procedure SubMat22(m1, m2, m3 :TPMDouble); forward;

//Adds two 2 vectors
procedure AddVec2(v1, v2, v3 :TPADouble); forward;

//Subtracts two 2 vectors
procedure SubVec2(v1, v2, v3 :TPADouble); forward;

//Inverts a 2x2 matrix
procedure InvMat22(m1, m2 :TPMDouble); forward;


//******************************************************************************
implementation


//******************************************************************************
constructor TQUB_KalmanFilter.Create(DimX, DimY :Integer);
begin
  inherited Create;

  prDimX := DimX;
  prDimY := DimY;

  prAllocMemory;

end;

destructor TQUB_KalmanFilter.Destroy;
begin
  try
    prFreeMemory;
  finally
    inherited;
  end;
end;

function TQUB_KalmanFilter.Forward :Boolean;
begin
  Result := True;
end;

function TQUB_KalmanFilter.Backward :Boolean;
begin
  Result := True;
end;

function TQUB_KalmanFilter.prAllocMemory :Boolean;
begin
  Result := False;
end;

function TQUB_KalmanFilter.prFreeMemory :Boolean;
begin
  Result := False;
end;


//******************************************************************************
procedure MulMat22_22(m1, m2, m3 :TPMDouble);
begin
  m3[0, 0] := m1[0, 0]*m2[0, 0] + m1[0, 1]*m2[1, 0];
  m3[0, 1] := m1[0, 0]*m2[0, 1] + m1[0, 1]*m2[1, 1];
  m3[1, 0] := m1[1, 0]*m2[0, 0] + m1[1, 1]*m2[1, 0];
  m3[1, 1] := m1[1, 0]*m2[0, 1] + m1[1, 1]*m2[1, 1];
end;

procedure MulMat22_22T(m1, m2, m3 :TPMDouble);
begin
  m3[0, 0] := m1[0, 0]*m2[0, 0] + m1[0, 1]*m2[0, 1];
  m3[0, 1] := m1[0, 0]*m2[1, 0] + m1[0, 1]*m2[1, 1];
  m3[1, 0] := m1[1, 0]*m2[0, 0] + m1[1, 1]*m2[0, 1];
  m3[1, 1] := m1[1, 0]*m2[1, 0] + m1[1, 1]*m2[1, 1];
end;

procedure MulMat22_Vec2(m1 :TPMDouble; v1, v2 :TPADouble);
begin
  v2[0] := m1[0, 0]*v1[0] + m1[0, 1]*v1[1];
  v2[1] := m1[1, 0]*v1[0] + m1[1, 1]*v1[1];
end;

procedure MulVec22T(v1, v2 :TPADouble; m1 :TPMDouble);
begin
  m1[0, 0] := v1[0]*v2[0];
  m1[0, 1] := v1[0]*v2[1];
  m1[1, 0] := v1[1]*v2[0];
  m1[1, 1] := v1[1]*v2[1];
end;

procedure MulMat22_S(m1 :TPMDouble; s1 :Double; m2 :TPMDouble);
begin
  m2[0, 0] := m1[0, 0] * s1;
  m2[0, 1] := m1[0, 1] * s1;
  m2[1, 0] := m1[1, 0] * s1;
  m2[1, 1] := m1[1, 1] * s1;
end;

procedure InvMat22(m1, m2 :TPMDouble);
var
  s :Double;
begin
  s := m1[0, 0]*m1[1, 1] - m1[0, 1]*m1[1, 0];
  if s < KALMAN_SMALL then s := KALMAN_SMALL;
  s := 1 / s;
  m2[0, 0] :=  s * m1[1, 1];
  m2[0, 1] := -s * m1[0, 1];
  m2[1, 0] := -s * m1[1, 0];
  m2[1, 1] :=  s * m1[0, 0];
end;

procedure AddMat22(m1, m2, m3 :TPMDouble);
begin
  m3[0, 0] := m1[0, 0] + m2[0, 0];
  m3[0, 1] := m1[0, 1] + m2[0, 1];
  m3[1, 0] := m1[1, 0] + m2[1, 0];
  m3[1, 1] := m1[1, 1] + m2[1, 1];
end;

procedure SubMat22(m1, m2, m3 :TPMDouble);
begin
  m3[0, 0] := m1[0, 0] - m2[0, 0];
  m3[0, 1] := m1[0, 1] - m2[0, 1];
  m3[1, 0] := m1[1, 0] - m2[1, 0];
  m3[1, 1] := m1[1, 1] - m2[1, 1];
end;

procedure AddVec2(v1, v2, v3 :TPADouble);
begin
  v3[0] := v1[0] + v2[0];
  v3[1] := v1[1] + v2[1];
end;

procedure SubVec2(v1, v2, v3 :TPADouble);
begin
  v3[0] := v1[0] - v2[0];
  v3[1] := v1[1] - v2[1];
end;

procedure KalmanForward(out x{post x}, p{post p} :Double; xp{prior x}, pp{prior p}, nm{meas noise}, np{proc noise}, y{data} :Double);
var
  xpp, ppp, k :Double;
begin

  try


    //a = 1
    //xpp = a . xp;
    xpp := xp;

    //ppp = a . pp . a' + np
    ppp := pp + np;

    //k = ppp . c' . (c . ppp . c' + r)-1
    k := ppp / (ppp + nm);

    //x = xpp + k . (y - c . xpp)
    x := xpp + k*(y - xpp);

    //p := ppp - k . c . ppp
    p := ppp - k*ppp;

    //try
      //Assert(p >= 0);
    //except
    //end;

  except

    x := xp;
    p := pp;

  end;

end;

procedure KalmanForward_2StateX(x{post x} :TPADouble; p{post p} :TPMDouble; xp{prior x} :TPADouble; pp{prior p} :TPMDouble; nm{meas noise} :Double; np{proc noise} :TPMDouble; a :TPMDouble; c{measurement}, k{gain} :TPADouble; y{data} :Double);
var
  xpp :Array[0..1]       of Double;
  ppp :Array[0..1, 0..1] of Double;
  m1  :Array[0..1, 0..1] of Double;
  v1  :Array[0..1]       of Double;
  v2  :Array[0..1]       of Double;
  s1  :Double;

begin

  try

    //x = |x |
    //    |x'|

    //xpp = a . xp;
    xpp[0] := a[0, 0]*xp[0] + a[0, 1]*xp[1];
    xpp[1] := a[1, 0]*xp[0] + a[1, 1]*xp[1];
    //           ^-- this is always zero, in the CV model


    //ppp = a . pp . a' + np
    //m1 = a . pp
    m1[0, 0] := a[0, 0]*pp[0, 0] + a[0, 1]*pp[1, 0];
    m1[0, 1] := a[0, 0]*pp[0, 1] + a[0, 1]*pp[1, 1];
    m1[1, 0] := a[1, 0]*pp[0, 0] + a[1, 1]*pp[1, 0];
    m1[1, 1] := a[1, 0]*pp[0, 1] + a[1, 1]*pp[1, 1];
    //ppp = m1 . a'
    ppp[0, 0] := m1[0, 0]*a[0, 0] + m1[0, 1]*a[0, 1];
    ppp[0, 1] := m1[0, 0]*a[1, 0] + m1[0, 1]*a[1, 1];
    ppp[1, 0] := m1[1, 0]*a[0, 0] + m1[1, 1]*a[0, 1];
    ppp[1, 1] := m1[1, 0]*a[1, 0] + m1[1, 1]*a[1, 1];
    //ppp = ppp + np
    ppp[0, 0] := ppp[0, 0] + np[0, 0];
    ppp[0, 1] := ppp[0, 1] + np[0, 1];
    ppp[1, 0] := ppp[1, 0] + np[1, 0];
    ppp[1, 1] := ppp[1, 1] + np[1, 1];



    //k = ppp . c' . (c . ppp . c' + r)-1
    //                         ^ a scalar
    //v1 = ppp . c'                             //c' is column, so v1 is also column
    v1[0] := ppp[0, 0]*c[0] + ppp[0, 1]*c[1];
    v1[1] := ppp[1, 0]*c[0] + ppp[1, 1]*c[1];
    //v2 = c . ppp                              //c is row, so v2 is also row
    v2[0] := c[0]*ppp[0, 0] + c[1]*ppp[1, 0];
    v2[1] := c[0]*ppp[0, 1] + c[1]*ppp[1, 1];
    //s1 = (v2 . c' + nm)-1
    s1 := v2[0]*c[0] + v2[1]*c[1];              //v2 is row, c' is col, so s1 is a scalar
    s1 := 1 / (s1 + nm);

    Assert(s1 >= 0);

    //k = v1 * s1  //v1 is column, so k is column
    k[0] := v1[0]*s1;
    k[1] := v1[1]*s1;



    //x := xpp + k . (y - c . xpp)
    //s1 := c . xpp
    s1 := c[0]*xpp[0] + c[1]*xpp[1];
    //s1 = y - s1
    s1 := y - s1;
    //x = xpp + k . s1
    x[0] := xpp[0] + k[0]*s1;
    x[1] := xpp[1] + k[1]*s1;



    //p := ppp - k . c . ppp
    //m1 := k . c
    m1[0, 0] := k[0]*c[0];
    m1[0, 1] := k[0]*c[1];
    m1[1, 0] := k[1]*c[0];
    m1[1, 1] := k[1]*c[1];
    //p = m1 . ppp
    p[0, 0] := m1[0, 0]*ppp[0, 0] + m1[0, 1]*ppp[1, 0];
    p[0, 1] := m1[0, 0]*ppp[0, 1] + m1[0, 1]*ppp[1, 1];
    p[1, 0] := m1[1, 0]*ppp[0, 0] + m1[1, 1]*ppp[1, 0];
    p[1, 1] := m1[1, 0]*ppp[0, 1] + m1[1, 1]*ppp[1, 1];
    //p = ppp - p
    p[0, 0] := ppp[0, 0] - p[0, 0];
    p[0, 1] := ppp[0, 1] - p[0, 1];
    p[1, 0] := ppp[1, 0] - p[1, 0];
    p[1, 1] := ppp[1, 1] - p[1, 1];


    Assert(p[0, 0] >= 0);
    Assert(p[1, 1] >= 0);


  except
  end;


end;




//******************************************************************************
end.
