unit QUB_LKC_Model;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Classes,
AllocTypes, Alloc,
qubopt_iface,
QUB_Matrices,

QUB_Fit_Model,
QUB_Q_Global, QUB_Exp_Global;


const


IDX_GL = 0;
IDX_VL = 1;


type


TQUB_LKC_Model                 = class(TQUB_Fit_Model)
protected


  procedure prAllocModelStuff; override;
  procedure prInitModelStuff; override;
  procedure prFreeModelStuff; override;

  function prGet_cParCount :Integer; override;


public


  V                          :Double;//Voltage
  V0                         :Double;//Previous voltage


  N                          :Integer;    //N exponentials


  //Leak params ................................................................
  GL                         :Double;
  dSSdGL                     :Double;
  VL                         :Double;
  dSSdVL                     :Double;

  //Cap params .................................................................
  //Tau's ......................................................................
  Tau                        :TPADouble;  //Time constants
  dSSdTau                    :TPADouble;

  //Amp's ......................................................................
  Amp                        :TPADouble;  //PreExponentials
  dSSdAmp                    :TPADouble;


  constructor Create(var OptParams :TQUB_Exp_Opt_Params); override;


  function ResetSSandGR :Boolean; override;

  function Do_Mean_t(var mt :Double; t :Double) :Integer; override;

  function Do_dMean_t_dGL(var dmt :Double; t :Double) :Integer;
  function Do_dMean_t_dVL(var dmt :Double; t :Double) :Integer;
  function Do_dMean_t_dTau(var dmt :Double; t :Double; k :Integer) :Integer;
  function Do_dMean_t_dAmp(var dmt :Double; t :Double; k :Integer) :Integer;

  function Do_fParsToModel :Integer; override;

  function Do_InitcPars :Integer; override;

  function Do_dSSdfPar :Integer; override;
  function Do_Varf2Varc :Integer; override;


end;


implementation


uses


QUB_Node;


//******************************************************************************
constructor TQUB_LKC_Model.Create(var OptParams :TQUB_Exp_Opt_Params);
begin

  Self.N := OptParams.LKCN;

  inherited Create(OptParams);

end;

function TQUB_LKC_Model.ResetSSandGR :Boolean;
var
  k :Integer;
begin

  Result := inherited ResetSSandGR;

  //dSSdGL
  dSSdGL  := 0;
  //dSSdVL
  dSSdVL  := 0;
  //dSSdTau's, dSSdAmp's
  for k := 0 to N - 1 do begin
    dSSdTau[k] := 0;
    dSSdAmp[k] := 0;
  end;

end;


//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_LKC_Model.Do_fParsToModel :Integer;
var
  i :Integer;
begin

  Result := inherited Do_fParsToModel;

  //cParsToModel
  GL := Exp(cPars[IDX_GL]);//Positivity constraint
  VL :=     cPars[IDX_VL];
  for i := 0 to N - 1 do begin
    Tau[i] := Exp(cPars[IDX_VL + 1 + 2*i + 0]);//Positivity constraint
    Amp[i] :=     cPars[IDX_VL + 1 + 2*i + 1];
  end;

end;


//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_LKC_Model.prAllocModelStuff;
begin

  inherited;

  Tau                        := DAlloc1(N);
  Amp                        := DAlloc1(N);

  dSSdTau                    := DAlloc1(N);
  dSSdAmp                    := DAlloc1(N);

  //Temporary quantities
  Vec1                       := DAlloc1(N);
  Vec2                       := DAlloc1(N);
  Mat1                       := DAlloc2(N, N);
  Mat2                       := DAlloc2(N, N);

end;

procedure TQUB_LKC_Model.prInitModelStuff;
var
  i :Integer;
begin

  inherited;

  GL := prOptParams.LKCGL;
  VL := prOptParams.LKCVL;
  for i := 0 to N - 1 do begin
    Tau[i] := prOptParams.LKCTau[i];
    Amp[i] := prOptParams.LKCAmp[i];
  end;
end;


procedure TQUB_LKC_Model.prFreeModelStuff;
begin

  FreeMem(Tau);
  Tau := nil;
  FreeMem(Amp);
  Amp := nil;

  FreeMem(dSSdTau);
  dSSdTau := nil;
  FreeMem(dSSdAmp);
  dSSdAmp := nil;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
function TQUB_LKC_Model.Do_InitcPars :Integer;
var
  i :Integer;
begin

  Result := inherited Do_InitcPars;

  cPars[IDX_GL] := Ln(GL);
  cPars[IDX_VL] :=    VL;
  for i := 0 to N - 1 do begin
    cPars[IDX_VL + 1 + 2*i + 0] := Ln(Tau[i]);
    cPars[IDX_VL + 1 + 2*i + 1] :=    Amp[i];
  end;

end;


//****************************************************************************//
//                                                                            //
//                        L O G  -  L I K E L I H O O D                       //
//                                                                            //
//****************************************************************************//

//Current mean
function TQUB_LKC_Model.Do_Mean_t(var mt :Double; t :Double) :Integer;
var
  i :Integer;
begin

  Result := inherited Do_Mean_t(mt, t);

  //mt = (V-VL)*GL + (V-V0)*Sum(i)[Ampi.exp(-t/Taui)]

  mt := (V-VL)*GL;
  for i := 0 to N - 1 do begin
    mt := mt + (V-V0)*Amp[i]*exp(-t/Tau[i]);
  end;

end;

function TQUB_LKC_Model.Do_dMean_t_dGL(var dmt :Double; t :Double) :Integer;
begin

  Result := 0;

  //mt = (V-VL)*GL + (V-V0)*Sum(i)[Ampi.exp(-t/Taui)]
  //dmt = (V-VL)

  dmt := (V-VL);

end;

function TQUB_LKC_Model.Do_dMean_t_dVL(var dmt :Double; t :Double) :Integer;
begin

  Result := 0;

  //mt = (V-VL)*GL + (V-V0)*Sum(i)[Ampi.exp(-t/Taui)]
  //dmt = -GL

  dmt := -GL;

end;

function TQUB_LKC_Model.Do_dMean_t_dTau(var dmt :Double; t :Double; k :Integer) :Integer;
begin
  Result := 0;

  //mt = (V-VL)*GL + (V-V0)*Sum(i)[Ampi.exp(-t/Taui)]
  //dmt = (V-V0)*Ampi*t/(Tau^2)*exp(-t/Taui)

  dmt := (V-V0)*t*Amp[k]/(Tau[k]*Tau[k])*exp(-t/Tau[k]);

end;

function TQUB_LKC_Model.Do_dMean_t_dAmp(var dmt :Double; t :Double; k :Integer) :Integer;
begin
  Result := 0;

  //mt = (V-VL)*GL + (V-V0)*Sum(i)[Ampi.exp(-t/Taui)]
  //dmt = (V-V0)*exp(-t/Taui)

  dmt := (V-V0)*exp(-t/Tau[k]);

end;

function TQUB_LKC_Model.Do_dSSdfPar :Integer;
var
  k :Integer;
begin

  Result := inherited Do_dSSdfPar;

  //dSSdfPar
  dSSdfPar[IDX_GL] := dSSdGL*exp(fPars[IDX_GL]);//GL=exp(fPar)
  dSSdfPar[IDX_VL] := dSSdVL;
  for k := 0 to N - 1 do begin
    dSSdfPar[IDX_VL + 1 + 2*k + 0] := dSSdTau[k]*exp(fPars[IDX_VL + 1 + 2*k + 0]);//Tau=exp(fPar)
    dSSdfPar[IDX_VL + 1 + 2*k + 1] := dSSdAmp[k];
  end;

end;

function TQUB_LKC_Model.Do_Varf2Varc :Integer;
var
  k :Integer;
begin

  Result := inherited Do_Varf2Varc;

  //var(GL|VL|Tau|Amp) ~= Sqr(d{GL|VL|Tau|Amp}/dz) * var(z)
  //dGL/dz  = dGL /dc * 1 = Exp(c) * 1
  //dVL/dz  =   1     * 1
  //dTau/dz = dTau/dc * 1 = Exp(c) * 1
  //dAmp/dz =   1     * 1

  //GL
  cParsVar[IDX_GL] := Sqr(Exp(cPars[IDX_GL])) * Abs(fParsVar[IDX_GL, IDX_GL]);
  //VL
  cParsVar[IDX_VL] :=                           Abs(fParsVar[IDX_VL, IDX_VL]);
  //Tau, Amp
  k := IDX_VL + 1;
  repeat
    //Tau
    cParsVar[k] := Sqr(Exp(cPars[k])) * Abs(fParsVar[k, k]);
    Inc(k);
    //Amp
    cParsVar[k] :=                      Abs(fParsVar[k, k]);
    Inc(k);
  until (k >= fParCount);

end;

function TQUB_LKC_Model.prGet_cParCount :Integer;
begin
  Result := inherited prGet_cParCount + 2*N + 2;
end;


//******************************************************************************
end.
