unit QUB_Matrix;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


AllocTypes, Alloc;



//******************************************************************************
function SolveLinEqs(A :TPMDouble{n*n}; B :TPADouble{n}; n :Integer; X :TPADouble{n}; Improve :Boolean) :Boolean; forward;

function ludcmp(a :TPMDouble{n*n}; n :Integer; indx :TPAInteger{n}; var d :Double) :Boolean; forward;
function lubksb(a :TPMDouble{n*n}; n :Integer; indx :TPAInteger{n}; b :TPADouble{n}) :Boolean; forward;
function mprove(a, alud :TPMDouble{n*n}; n :Integer; indx :TPAInteger{n}; b :TPADouble{n}; x :TPADouble{n}) : Boolean; forward;



//******************************************************************************
implementation



//******************************************************************************
function SolveLinEqs(A :TPMDouble{n*n}; B :TPADouble{n}; n :Integer; X :TPADouble{n}; Improve :Boolean) :Boolean;
var
  alud   :TPMDouble;
  blud   :TPADouble;
  indx   :TPAInteger;
  d      :Double;
  i, j   :Integer;
begin

  alud   := nil;
  blud   := nil;
  indx   := nil;

  try

    try

      indx := IAlloc1(n);
      alud := DAlloc2(n, n);
      blud := DAlloc1(n);
      for i := 0 to n - 1 do begin
        blud[i] := B[i];
        for j := 0 to n - 1 do begin
          alud[i, j] := A[i, j];
        end;
      end;

      Result := ludcmp(alud, n, indx, d);
      if not Result then Result := lubksb(alud, n, indx, blud);

      for i := 0 to n - 1 do begin
        X[i] := blud[i];
      end;

      if Improve then begin
        if not Result then Result := mprove(A, alud, n, indx, B, X);
      end;


    except

      Result := True;

    end;

  finally

    Free2(TPMChar(alud));
    FreeMem(blud);
    FreeMem(indx);

  end;

end;

function ludcmp(a :TPMDouble{n*n}; n :Integer; indx :TPAInteger{n}; var d :Double) :Boolean;
const
  tiny = 1.0e-20;
var
   i, imax, j, k       :Integer;
   big, dum, sum, temp :Double;
   vv                  :TPADouble;
begin

   Result := False;

   d      := 1.0;
   imax   := -1;//was not initialized in the original

   vv     := nil;

   try

     try

       vv := DAlloc1(n);

       for i := 0 to n - 1 do begin

         big := 0.0;

         for j := 0 to n - 1 do begin
           temp := abs(a[i, j]);
           if temp > big then begin
             big := temp;
           end;
         end;

         if big = 0.0 then begin
           Result := True;
           Exit;//singular matrix
         end;

         vv[i] := 1.0 / big;

       end;

       for j := 0 to n - 1 do begin

         for i := 0 to j - 1 do begin
           sum := a[i, j];
           for k := 0 to i - 1 do begin
             sum := sum - a[i, k] * a[k, j];
           end;
           a[i, j] := sum;
         end;

         big := 0.0;

         for i := j to n - 1 do begin
           sum := a[i, j];
           for k := 0 to j - 1 do begin
             sum := sum - a[i, k] * a[k, j];
           end;
           a[i, j] := sum;
           dum := vv[i] * abs(sum);
           if dum >= big then begin
             big := dum;
             imax := i;
           end;
         end;

         if j <> imax then begin
           for k := 0 to n - 1 do begin
             dum := a[imax, k];
             a[imax, k] := a[j, k];
             a[j, k] := dum;
           end;
           d := -d;
           vv[imax] := vv[j];
         end;

         indx[j] := imax;
         if a[j, j] = 0.0 then a[j, j] := tiny;

         if j <> (n - 1) then begin
           dum := 1.0 / a[j, j];
           for i := j + 1 to n - 1 do begin
             a[i, j] := a[i, j] * dum;
           end;
         end;

       end;

     except

       Result := True;

     end;

  finally

    FreeMem(vv);

  end;

end;

function lubksb(a :TPMDouble{n*n}; n :Integer; indx :TPAInteger{n}; b :TPADouble{n}) :Boolean;
var
  j, ip, ii, i :Integer;
  sum          :Double;
begin

  Result := False;

  ii := -1;//0, in the 1-index based original

  try

    for i := 0 to n - 1 do begin
      ip    := indx[i];
      sum   := b[ip];
      b[ip] := b[i];
      if ii <> -1 then begin
        for j := ii to i - 1 do begin
          sum := sum - a[i, j] * b[j];
        end;
      end else if sum <> 0.0 then begin
        ii := i;
      end;
      b[i] := sum;
    end;

    for i := n - 1 downto 0 do begin
      sum := b[i];
      for j := i + 1 to n - 1 do begin
        sum := sum - a[i, j] * b[j];
      end;
      b[i] := sum / a[i, i];
    end;

  except

    Result := True;

  end;

end;

function mprove(a, alud :TPMDouble{n*n}; n :Integer; indx :TPAInteger{n}; b :TPADouble{n}; x :TPADouble{n}) : Boolean;
var
  j, i :Integer;
  sdp  :Double;
  r    :TPADouble;
begin

  r := nil;

  try

    try

      r := DAlloc1(n);

      for i := 0 to n - 1 do begin
        sdp := -b[i];
        for j := 0 to n - 1 do begin
          sdp := sdp + a[i, j]*x[j];
        end;
        r[i] := sdp;
      end;

      Result := lubksb(alud, n, indx, r);

      for i := 0 to n - 1 do begin
        x[i] := x[i] - r[i];
      end;

    except

      Result := True;

    end;

  finally

    FreeMem(r);

  end;

end;


//******************************************************************************
end.
