{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Node_HEKA;

// utilities for reading HEKA trees.
// QUB trees are used behind-the-scenes for in-memory storage.
// used by QUB_QFS_QDF

interface

uses

Classes, Sysutils,
AllocTypes,
QUB_Node;

{$Z1}

(*  Adapted from Pulsedef.h

      Author:     Stefan. H. Heinemann
                  Abteilung Membranbiophysik
                  Max-Planck-Institut fuer biophys. Chemie
                  D-3400 Goettingen
                  Germany

      Revised:    15-Sep-95  Hubert Affolter
*)

(*
 * Pulsed data file format
 *
 * This module defines the data types to be used for control information on
 * experiments involving pulsed data; it uses the tree structures as defined
 * in module Tree and provides types for a 'PulsedTree'.
 *
 * All the records are defined as PACKED ARRAYs, i.e., no additional padding 
 * bytes are used in the blocks stored in the files.
 *
 * The sizes in bytes of the used variables are:
 *    BYTE           1
 *    CHAR           1
 *    enumeration    1
 *    SET16          2
 *    INT16          2
 *    INT32          4
 *    ADDRESS        4
 *    REAL           4
 *    LONGREAL       8
 *
 * This module is based on proposals by:
 *    Hubert Affolter
 *    Dan Brown
 *    Stefan Heinemann
 *    Erwin Neher
 *    Fred Sigworth
 *    Walter Stuehmer
 *)



// CONST
//   VersionNumber     = 7;

(* Structure of the trees *)

//CONST
//   TreeLevels        = 4;  (* 4 Levels: PHeader, Group, Series, Sweep *)

CONST
   Unit2Size         = 2;
   Unit4Size         = 4;
   SwSpareSize       = 10;
   StringSize        = 14;
   CommentSize       = 80;
   EPC9StateSize     = 104;
   RootTextSize      = 400;

   SeExtraLongReals  = 4;


(*    RecordTypes    : Root, Group, Series, Sweep
 *
 * Sweep
 * SweepRecord
 *
 * A SweepRecord describes the sampled data from a single stimulation/
 * acquisition.
 *)

CONST
   ExtendedBit      = 0;
   ExtendedBitMask  = 1;
   FuraBit          = 1;
   FuraBitMask      = 2;
   LockInBit        = 2;
   LockInBitMask    = 4;
   DataFormatBit    = 3;
   DataFormatBitMask = 8;
   LittleEndianBit  = 4;
   LittleEndianBitMask = 16;
   ScanRateBit      = 15;

TYPE
   StringType       = ARRAY [0..StringSize-1] OF CHAR;
   CommentType      = ARRAY [0..CommentSize-1] OF CHAR;
   RootTextType     = ARRAY [0..RootTextSize-1] OF CHAR;

   DataAbscissaType = ( Time,
                        Amplitude, LnAmplitude, LogAmplitude,
                        Frequency, LnFrequency, LogFrequency );

   DataFormatType   = ( int16, int32, real32, real64 );


TYPE SweepRecord    = RECORD
   Time             : Double;     (* UNIX (seconds)+ Rest            *)
   StimCount        : Integer;        (* relevant entry in StimTree      *)
   SweepCount       : Integer;        (* .. at time of acquisition       *)
   AverageCount     : Integer;        (* number of on-line averages      *)
   Leak             : BOOLEAN;      (* TRUE, if Record has a leaksweep *)
   SecondTrace      : BOOLEAN;      (* TRUE, if Record has a 2. trace  *)
   Lbl              : StringType;
   DataPoints       : Integer;        (* Number of data points in RAM    *)
   Data             : Integer;        (* Offset in raw data file         *)
   DataPointer      : Pointer;        (* Raw data ADDRESS in memory      *)
       (* EPC 7/9 settings *)
   DataFactor1      : Double;       (* Amperes/ADC-unit                *)
   DataFactor2      : Double;       (* Volts/ADC-unit                  *)
   CSlow            : Double;       (* Capacitance, Farad              *)
   GSeries          : Double;       (* Series conductance, Siemens     *)
   RsValue          : Double;       (* Series resistance setting, Ohms *)
   RsFraction       : Double;       (* % RsCompensation                *)
       (* Results of pre-analysis  *)
   ZeroCurrent      : Double;       (* Amperes                         *)
   OnlineYResult    : Double;       (* Param. determined by last anal. *)
   OnlineXResult    : Double;       (* Param. determined by last anal. *)

   TotalPoints      : Integer;        (* Total data points in file       *)
   Offset           : Integer;        (* Offset of loaded data           *)

   SweepKind        : Word;            (* meaning of bits:
                                        ExtendedBit   => extended fields defined
                                        FuraBit       => Fura.Active
                                        LockInBit     => LockIn.Active
                                        DataFormatBit => "DataFormat" field valid
                                        LittleEndianBit => byte sequence
                                                         "Mac" = cleared
                                                         "DOS" = set
                                        ScanRateBit   => set: use scan-rate!
                                     *)

   FuraPoints       : Integer;        (* Number of data points in RAM    *)
   FuraData         : Integer;        (* Offset in raw data file         *)
   FuraPointer      : Pointer;        (* Raw data ADDRESS in memory      *)
   OnlineYResult2   : Double;       (* Param. determined by last anal. *)
   OnlineXResult2   : Double;       (* Param. determined by last anal. *)

   DispFactor1      : Double;       (* reserved by PULSE               *)
   DispFactor2      : Double;       (* reserved by PULSE               *)

     (*
       Since we introduced the new "DataFormat" field in a field
       which before was without importance, we are interpreting it
       ONLY if "SweepKind" has ExtendedBit and DataFormatBit set!
       Thus, the procedure "StimIO.InitializeSweep" scans
       the "*.pul" tree upon loading and enforces that check.
      *)
   DataFormat       : DataFormatType; (* What kind of abscissa, INT or REAL *)
   DataAbscissa     : DataAbscissaType;
   Timer            : Double;

   Spares           : ARRAY[0..SwSpareSize-1] OF CHAR;
END (* RECORD *);

TPSweepRecord      = ^SweepRecord;

(*
 * Series
 * SeriesRecord
 *
 * A SeriesRecord describes a Series of Sweeps; this is characterized
 * by the StimulationRecord, that generates it; also the record in-
 * cludes a number of environmental parameters that are obtained
 * before the series is started.
 *)

TYPE
  RecordingModeType     = ( InOut,
                            OnCell,
                            OutOut,
                            WholeCell,
                            PCClamp,
                            VClamp,
                            NoRec,
                            TestInt,
                            TestExt );

  UserParamType         = RECORD
    Value               : Double;
    Name                : StringType;
    aUnit               : ARRAY [0..Unit2Size-1] OF CHAR;
  END;

  EPC9StateType         = ARRAY[0..EPC9StateSize-1] OF BYTE;

  SeriesRecord          = RECORD
    Time                : Double;       (* seconds                    *)
        (* Patch parameters, usually obtained before series *)
    Bandwidth           : Double;       (* Hertz                      *)
    PipettePotential    : Double;       (* Volts                      *)
    CellPotential       : Double;       (* Volts                      *)
    PipetteResistance   : Double;       (* Ohms                       *)
    SealResistance      : Double;       (* Ohms                       *)
    BackgroundNoise     : Double;       (* Amperes, RMS               *)
    Temperature         : Double;       (* Degrees C                  *)
    PipettePressure     : Double;       (* in cm H20                  *)
    UserParam1          : UserParamType;
    UserParam2          : UserParamType;

    RecordingMode       : RecordingModeType;
    Filler1             : BOOLEAN;
    Comment             : CommentType;
      (* EPC9 state record *)
    EPC9State           : EPC9StateType;
    InternalSolution,
    ExternalSolution    : Integer;        (* solutions according to solution
                                             code as stored in '*.sol'  *)
    ExtraYUnit1         : ARRAY [0..Unit2Size-1] OF CHAR;
    ExtraYUnit2         : ARRAY [0..Unit2Size-1] OF CHAR;

    DispYUnit1          : ARRAY [0..Unit4Size-1] OF CHAR; (* reserved by PULSE    *)
    DispYUnit2          : ARRAY [0..Unit4Size-1] OF CHAR; (* reserved by PULSE    *)

    FuraK               : Double;
    FuraMin             : Double;
    FuraMax             : Double;

    LockInExtPhase      : Double;

    Timer               : Double;

      (* Reserved for extensions *)

    ExtraLongReal       : ARRAY [0..SeExtraLongReals-1] OF Double;

  END (* RECORD *);

  TPSeriesRecord        = ^SeriesRecord;

(*
 * Group
 * GroupRecord
 *
 * A GroupRecord describes a group of series, such as patch and
 * whole cell currents obtained simultanuously, or groups of series
 * obtained in sequence under different sets of conditions
 *)         

TYPE GroupRecord        = RECORD
   Lbl                  : StringType;
   Text                 : CommentType;
   ExperimentNumber     : Integer;
   ExtraLongReal        : Double;
END (* RECORD *);

TYPE TPGroupRecord = ^GroupRecord;

(*
 * Root
 * RootRecord
 *
 * A RootRecord describes a complete experiment.
 *)

TYPE PulseRootRecord    = RECORD
   Version              : SmallInt;
   VersionName          : StringType;
   FileName             : StringType;
   Comments             : RootTextType;
   StartTime            : Double;
END (* RECORD *);

TPPulseRootRecord      = ^PulseRootRecord;


(*
      Author:     Stefan. H. Heinemann
                  Abteilung Membranbiophysik
                  Max-Planck-Institut fuer biophys. Chemie
                  D-3400 Goettingen
                  Germany

      Revised:    15-Sep-95  Hubert Affolter
*)

(*
 * Stim data file format
 *
 * This module defines the data types to be used for the stimulation in
 * experiments involving pulsed data; it uses the tree structures as defined
 * in module Tree and provides types for a 'StimTree'.
 *
 * The sizes in bytes of the used variables are:
 *    BYTE           1
 *    CHAR           1
 *    enumeration    1
 *    SET16          2
 *    INT16          2
 *    INT32          4
 *    ADDRESS        4
 *    REAL           4
 *    LONGREAL       8
 *
 * This module is based on proposals by:
 *    Hubert Affolter
 *    Dan Brown
 *    Stefan Heinemann
 *    Erwin Neher
 *    Fred Sigworth
 *    Walter Stuehmer
 *)


//CONST
   //VersionNumber  = 7;

(* Structure of the trees *)

//CONST
   //TreeLevels     = 3;   (* 3 Levels: SHeader, Stimulation, StimSegment *)

CONST
   //Unit2Size      = 2;
   //StringSize     = 14;
   MacroNameSize  = 16;


(*
 * RecordTypes for the StimTree:  Root, Stimulation, Segment
 *
 * Stimulation
 * StimulationRecord
 *
 * A StimulationRecord describes a stimulation pulse made up of one or
 * more segments. Each segment has a voltage and a duration. The segment
 * can be a constant output voltage (SegmentConstant) or a ramp from the
 * previous segment voltage to the present segment voltage (SegmentRamp).
 * For conditioning pulses a segment "SegmentConditioning" is provided.
 * No P/n is performed for such segments. For long conditioning pulses
 * a possibly less accurate timing is applied.
 * Incrementing voltages or times in a ramped record is allowed, but be
 * sure to think about what will happen; incrementing is done before the
 * ramp parameters are evaluated.
 *
 * When a series of pulses are output, the Delta entries in each
 * StimulationSegment describe the parameter change to perform for each
 * pulse in the series. We have:
 *
 *       Voltage(n)  := Voltage(0) * DeltaVFactor^n + DeltaVIncrement * n;
 *       Duration(n) := Duration(0) * DeltaTFactor^n + DeltaTIncrement * n;
 *
 * The Leak sweeps usually follow the test sweeps with a delay LeakDelay.
 * If LeakDelay is negative then leak sweeps precede test sweeps and Leak
 * Delay is the time between the end of the last leak sweep and the
 * beginning of the test sweep
 *)

TYPE SegmentClass = ( SegmentConstant,
                      SegmentRamp,
                      SegmentConditioning,
                      SegmentContinuous,
                      SegmentConstSine,
                      SegmentRampSine );

TYPE StimSegmentRecord = RECORD
   Cls                  : SegmentClass; (* constant, ramp, or condit.   *)
   IsHolding            : BOOLEAN;      (* TRUE if Voltage is to be set *)
   Voltage              : Double;     (* to holding                   *)
   Duration             : Double;
   DeltaVFactor         : Double;
   DeltaVIncrement      : Double;
   DeltaTFactor         : Double;
   DeltaTIncrement      : Double;
END (* RECORD *);

TPStimSegmentRecord     = ^StimSegmentRecord;


TYPE Trigger = RECORD
   TriggerSegment       : SmallInt;
   TriggerTime          : Double;   (* Seconds, within trig.Segment *)
   TriggerLength        : Double;   (* Length of Trig.Pulse, sec    *)
   TriggerAmplitude     : Double;   (* Amplitude of Trig.Pulse      *)
   TriggerDac           : SmallInt;
END; (* RECORD *)

TYPE IncrementModeType  = ( ModeInc,
                            ModeDec,
                            ModeIncInterleaved,
                            ModeDecInterleaved,
                            ModeAlternate );

TYPE GUpdateType        = ( NoGUpdate,
                            SwGSlow,
                            SwGFast,
                            SwGBoth,
                            SeGSlow,
                            SeGFast,
                            SeGBoth );

TYPE WriteModeType      = ( WriteEnabled,
                            WriteDisabled,
                            NoWriteNoShow,
                            WriteButNoShow );

TYPE ExtTriggerType     = ( TrigNone, TrigSeries, TrigSweep );

TYPE PostVmembType      = ( VmembSweepIncr,
                            VmembValue,
                            VmembSeriesIncr );

TYPE LockInType         = ( Loff, Lnormal, Lpwlinear );

TYPE AmplModeType       = ( AllAmplModes, VCAmplMode, CCAmplMode );


// JTB 7-22-2004 added PACKED here to correct sample interval.
// TODO - look through code for other places this may be an issue.
TYPE StimulationRecord  = PACKED RECORD
   FileName             : ARRAY [0..StringSize-1] OF CHAR; (* Source File     *)
   EntryName            : ARRAY [0..StringSize-1] OF CHAR; (* Identifier      *)
   SampleInterval       : Double;       (* Seconds                          *)
   FilterFactor         : Double;       (* oversampling factor              *)
   SweepInterval        : Double;       (* Repetition-Interval              *)
   NumberSweeps         : Integer;          (* Number of sweeps in Series       *)
   NumberRepeats        : Integer;          (* Number of Seq. repeats           *)
   RepeatWait           : Double;       (* Wait between repeats             *)
   LinkedSequence       : ARRAY [0..StringSize-1] OF CHAR;
   LinkedWait           : Double;       (* Wait between linked sequences    *)

   LeakCount            : Integer;          (* number of leak sweeps            *)
   LeakSize             : Double;       (* rel. amplitude of leak pulse
                                             e.g. 0.25 for P/4 protocol       *)
   LeakHolding          : Double;
   LeakAlternate        : BOOLEAN;        (* norm. or alt. leak protoc.       *)
   AltLeakAveraging     : BOOLEAN;        (* norm. or alt. leak while aver.   *)
   LeakDelay            : Double;       (* Seconds                          *)
   Trig                 : ARRAY [0..2] OF Trigger;
   NumberOfTriggers     : SmallInt;          (* Number actually used; 0<x<4      *)
   RelevantXSegment     : SmallInt;          (* usually that which changes       *)
   RelevantYSegment     : SmallInt;          (* usually that which changes       *)

   WriteMode            : WriteModeType;
   IncrementMode        : IncrementModeType;

   TotalSweepLength     : Integer;          (* Total sweeplength including a    *)
                                          (* possible continuous segment,     *)
                                          (* in units of sample intervals.    *)

   MaxSweepLength       : Integer;          (* max length of sweep to be shown. *)
                                          (* For a pulse without a continuous *)
                                          (* segment it is identical to the   *)
                                          (* max sweeplength from trig #1,    *)
                                          (* in units of sample intervals     *)

   InputChannels        : SmallInt;          (* # input channels. Default        *)
   GUpdate              : GUpdateType;    (* make C-slow/fast bef. each pulse *)
                                          (* only for EPC9                    *)
   RelAbsPot            : BOOLEAN;        (* absolute or relativ potentials   *)

   HasContinuous        : BOOLEAN;
   LogIncrement         : BOOLEAN;

   StimDac              : SmallInt;
   Adc1                 : SmallInt;
   Adc2                 : SmallInt;
   YUnit1               : ARRAY[0..Unit2Size-1] OF CHAR;
   YUnit2               : ARRAY[0..Unit2Size-1] OF CHAR;

   VmembIncrement       : Single;

   ExtTrigger           : ExtTriggerType;
   FileTemplate         : BOOLEAN;

   StimKind             : WORD;
                            (* meaning of bits:
                                 0 => extended fields defined
                                 1 => LockIn.Active
                                 2 => Fura.Active
                                15 => set: use scan-rate!
                             *)

   LockInCycle          : Double;
   LockInAmplitude      : Double;

   FuraOn               : BOOLEAN;
   VmembMode            : PostVmembType;
   FuraTotLength        : Double;
   FuraDelay            : Double;
   FuraLength1          : Double;
   FuraLength2          : Double;
   FuraWaveLength0      : Double;
   FuraWaveLength1      : Double;
   FuraWaveLength2      : Double;
   FuraRepeats          : SmallInt;

   LockInSkip           : Integer;
   LockInVReversal      : Double;
   LockInMode           : LockInType;
   LockInShow           : BOOLEAN;

   ConfigMacro          : ARRAY[0..MacroNameSize-1] OF CHAR;
   EndMacro             : ARRAY[0..MacroNameSize-1] OF CHAR;

   AmplModeKind         : AmplModeType;
   Filler1              : BOOLEAN;
END (* RECORD *);

TPStimulationRecord     = ^StimulationRecord;


TYPE StimRootRecord     = RECORD
   Version              : SmallInt;
END (* RECORD *);

TPStimRootRecord        = ^StimRootRecord;


{ PULSE V9 }
CONST

  SizeByte             = 1;
  SizeChar             = 1;
  SizeEnum             = 1;
  SizeBoolean          = 1;
  SizeInt16            = 2;
  SizeCard16           = 2;
  SizeSet16            = 2;
  SizeInt32            = 4;
  SizeCard32           = 4;
  SizeReal             = 4;
  SizeLongReal         = 8;

  String8Size          = 8;
  String32Size         = 32;
  String80Size         = 80;
  String400Size        = 400;

  ParamDescrSize       = 24;
  SwUserParamNo        = 4;
  SeUserParamNo        = 4;
  GrStimParams         = 10;

  SizeStateVersion     = 8;
  SizeSerialNumber     = 8;
  SizeCalibDate        = 16;


  RootLevel            = 0;
  GroupLevel           = 1;
  SeriesLevel          = 2;
  SweepLevel           = 3;
  TraceLevel           = 4;


  LittleEndianBit9      = 0;
  IsLeak               = 1;
  IsVirtual            = 2;
  IsImon               = 3;
  IsVmon               = 4;
  Clip                 = 5;
{#      DataKind          -> meaning of bits:
#                           - LittleEndianBit => byte sequence
#                             "PowerPC Mac" = cleared
#                             "Windows and Intel Mac" = set
#                           - IsLeak
#                             set if trace is a leak trace
#                           - IsVirtual
#                             set if trace is a virtual trace
#                           - IsImon
#                             -> set if trace was from Imon ADC
#                             -> it flags a trace to be used to
#                                compute LockIn traces from
#                             -> limited to "main" traces, not "leaks"!
#                           - IsVmon
#                             -> set if trace was from Vmon ADC
#                           - Clip
#                             -> set if amplifier of trace was clipping}
     
TYPE
 {  RecordingModeType    =
( InOut,
  OnCell,
  OutOut,
  WholeCell,
  CClamp,
  VClamp,
  NoMode );}

 {  DataFormatType       = 
( int16,
  int32,
  real32,
  real64 );}

   UserParamDescrType = PACKED RECORD
      Name          :Array[0..String32Size-1] of Char;
      AUnit         :Array[0..String8Size-1] of Char;
   end;


   BundleItem = PACKED RECORD
     oStart         :Integer;
     oLength        :Integer;
     oExtension     :Array[0..7] of Char;
   end;

   BundleHeader = PACKED RECORD
     oSignature     :Array[0..7] of Char;
     oVersion       :Array[0..31] of Char;
     oTime          :Double;
     oItems         :Integer;
     oIsLittleEndian:Byte;
     oReserved      :Array[0..10] of Byte;
     oBundleItems   :Array[0..11] of BundleItem;
   end;


   TraceRecord9 = PACKED RECORD
      Mark          :Integer;
      Lbl           :Array[0..String32Size-1] of Char;
      TraceCount    :Integer;
      Data          :Integer;
      DataPoints    :Integer;
      InternalSolution  :Integer;
      AverageCount      :Integer;
      LeakCount         :Integer;
      LeakTraces        :Integer;
      DataKind          :SmallInt;
      UseXStart         :Byte;
      Filler1           :Byte;
      RecordingMode     :RecordingModeType;
      AmplIndex         :Byte;
      DataFormat        :DataFormatType;
      DataAbscissa      :Byte;
      DataScaler        :Double;
      TimeOffset        :Double;
      ZeroData          :Double;
      YUnit             :Array[0..String8Size-1] of Char;
      XInterval         :Double;
      XStart            :Double;
      XUnit             :Array[0..String8Size-1] of Char;
      YRange            :Double;
      YOffset           :Double;
      Bandwidth         :Double;
      PipetteResistance :Double;
      CellPotential     :Double;
      SealResistance    :Double;
      CSlow             :Double;
      GSeries           :Double;
      RsValue           :Double;
      GLeak             :Double;
      MConductance      :Double;
      LinkDAChannel     :Integer;
      ValidYRange       :Byte;
      AdcMode           :Byte;
      AdcChannel        :SmallInt;
      YMin              :Double;
      YMax              :Double;
      SourceChannel     :Integer;
      ExternalSolution  :Integer;
      CM                :Double;
      GM                :Double;
      Phase             :Double;
      DataCRC           :Integer;
      CRC               :Integer;
      GS                :Double;
      SelfChannel       :Integer;
      InterleaveSize    :Integer;
      InterleaveSkip    :Integer;
      ImageIndex        :Integer;
      Markers           :Array[0..9] of Double;
      SECM_X            :Double;
      SECM_Y            :Double;
      SECM_Z            :Double;
end;

TPTraceRecord9 = ^TraceRecord9;

SweepRecord9 = PACKED RECORD
   Mark                 :Integer;
   Lbl                  :Array[0..String32Size-1] of Char;
   AuxDataFileOffset    :Integer;
   StimCount            :Integer;
   SweepCount           :Integer;
   Time                 :Double;
   Timer                :Double;
   UserParams           :Array[0..3] of Double;
   Temperature          :Double;
   OldIntSol            :Integer;
   OldExtSol            :Integer;
   DigitalIn            :SmallInt;
   Filler1              :SmallInt;
   Markers              :Array[0..3] of Double;
   Filler2              :Integer;
   CRC                  :Integer;
end;

TPSweepRecord9 = ^SweepRecord9;

SeriesRecord9 = PACKED RECORD
   Mark                 :Integer;
   Lbl                  :Array[0..String32Size-1] of Char;
   Comment              :Array[0..String80Size-1] of Char;
   SeriesCount          :Integer;
   NumberSweeps         :Integer;
   AmplStateOffset      :Integer;
   AmplStateSeries      :Integer;
   SeriesType           :Byte;
   UseXStart            :Byte;
   Filler1              :Byte;
   Filler2              :Byte;
   Time                 :Double;
   PageWidth            :Double;
   UserParamDescr       :Array[0..3] of UserParamDescrType;
   Filler3              :Array[0..31] of Char;
   UserParams1          :Array[0..3] of Double;
   LockInParams         :Array[0..95] of Char;
   AmplifierState       :Array[0..399] of Char;
   Username             :Array[0..String80Size-1] of Char;
   UserParamDescr1      :Array[0..3] of UserParamDescrType;
   Filler4              :Integer;
   CRC                  :Integer;
   UserParams2          :Array[0..3] of Double;
   UserParamDescr2      :Array[0..3] of UserParamDescrType;
   ScanParams           :Array[0..95] of Char;
end;

TPSeriesRecord9 = ^SeriesRecord9;

GroupRecord9 = PACKED RECORD
   Mark                 :Integer;
   Lbl                  :Array[0..String32Size-1] of Char;
   Text                 :Array[0..String80Size-1] of Char;
   ExperimentNumber     :Integer;
   GroupCount           :Integer;
   CRC                  :Integer;
end;

TPGroupRecord9 = ^GroupRecord9;

PulseRootRecord9 = PACKED RECORD
   Version              :Integer;
   Mark                 :Integer;
   VersionName          :Array[0..String32Size-1] of Char;
   AuxFileName          :Array[0..String80Size-1] of Char;
   RootText             :Array[0..String400Size-1] of Char;
   StartTime            :Double;
   MaxSamples           :Integer;
   CRC                  :Integer;
   Features             :SmallInt;
   Filler1              :SmallInt;
   Filler2              :Integer;
end;

TPPulseRootRecord9 = ^PulseRootRecord9;


function SwappedInt32( Value : Integer ): Integer;

procedure SwapBytesA( x:Pointer; iSize:Integer; iCount:Integer );
procedure Swap2Bytes(x :Pointer);
procedure Swap4Bytes(x :Pointer);
procedure Swap8Bytes(x :Pointer);

function LoadTree( Stream :TStream ): INode; overload;
function LoadTree(
             Stream             : TStream;
             Sizes              : TPAInteger;
             VAR Levels         : LONGINT;
             VAR NeedsByteSwap  : BOOLEAN;
             AsChildrenOf       : INode )
             : Boolean; overload;

// *****************************************************************************

type

THEKA_Tree_Pulse          = class;
THEKA_Tree_Pulse_Group    = class;
THEKA_Tree_Pulse_Series   = class;
THEKA_Tree_Pulse_Sweep    = class;

THEKA_Tree_Pulse          = class(TList)
protected
  prRootNode              :INode;
  prRoot                  :TPPulseRootRecord;

  function prGetGroup(Ix :Integer): THEKA_Tree_Pulse_Group;

public
  constructor Create( Tree :INode );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                   :TPPulseRootRecord            read prRoot;
  property Groups[Ix :Integer]   :THEKA_Tree_Pulse_Group       read prGetGroup;  default;
end;

THEKA_Tree_Pulse_Group    = class(TList)
protected
  prGroupNode                    :INode;
  prGroup                        :TPGroupRecord;

  function prGetSeries(Ix :Integer): THEKA_Tree_Pulse_Series;

public
  constructor Create( GroupNode :INode; NeedsByteSwap :Boolean );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                   :TPGroupRecord                 read prGroup;
  property Series[Ix :Integer]   :THEKA_Tree_Pulse_Series       read prGetSeries;  default;
end;

THEKA_Tree_Pulse_Series   = class(TList)
protected
  prSeriesNode         :INode;
  prSeries             :TPSeriesRecord;

  function prGetSweep(Ix :Integer): THEKA_Tree_Pulse_Sweep;

public
  constructor Create( SeriesNode :INode; NeedsByteSwap :Boolean );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                   :TPSeriesRecord               read prSeries;
  property Sweeps[Ix :Integer]   :THEKA_Tree_Pulse_Sweep       read prGetSweep;    default;
end;

THEKA_Tree_Pulse_Sweep    = class(TList)
protected
  prSweepNode         :INode;
  prSweep             :TPSweepRecord;

public
  constructor Create( SweepNode :INode; NeedsByteSwap :Boolean );

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec        :TPSweepRecord            read prSweep;
end;


function LoadPulseTree( Stream :TStream ): THEKA_Tree_Pulse; // returns nil on error


type

THEKA_Tree_Pulse9          = class;
THEKA_Tree_Pulse9_Group    = class;
THEKA_Tree_Pulse9_Series   = class;
THEKA_Tree_Pulse9_Sweep    = class;
THEKA_Tree_Pulse9_Trace    = class;

THEKA_Tree_Pulse9          = class(TList)
protected
  prRootNode              :INode;
  prRoot                  :TPPulseRootRecord9;

  function prGetGroup(Ix :Integer): THEKA_Tree_Pulse9_Group;

public
  constructor Create( Tree :INode );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                   :TPPulseRootRecord9            read prRoot;
  property Groups[Ix :Integer]   :THEKA_Tree_Pulse9_Group       read prGetGroup;  default;
end;

THEKA_Tree_Pulse9_Group    = class(TList)
protected
  prGroupNode                    :INode;
  prGroup                        :TPGroupRecord9;

  function prGetSeries(Ix :Integer): THEKA_Tree_Pulse9_Series;

public
  constructor Create( GroupNode :INode; NeedsByteSwap :Boolean );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                   :TPGroupRecord9                 read prGroup;
  property Series[Ix :Integer]   :THEKA_Tree_Pulse9_Series       read prGetSeries;  default;
end;

THEKA_Tree_Pulse9_Series   = class(TList)
protected
  prSeriesNode         :INode;
  prSeries             :TPSeriesRecord9;

  function prGetSweep(Ix :Integer): THEKA_Tree_Pulse9_Sweep;

public
  constructor Create( SeriesNode :INode; NeedsByteSwap :Boolean );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                   :TPSeriesRecord9               read prSeries;
  property Sweeps[Ix :Integer]   :THEKA_Tree_Pulse9_Sweep       read prGetSweep;    default;
end;

THEKA_Tree_Pulse9_Sweep    = class(TList)
protected
  prSweepNode         :INode;
  prSweep             :TPSweepRecord9;

  function prGetTrace(Ix :Integer): THEKA_Tree_Pulse9_Trace;

public
  constructor Create( SweepNode :INode; NeedsByteSwap :Boolean );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec        :TPSweepRecord9            read prSweep;
  property Traces[Ix :Integer]   :THEKA_Tree_Pulse9_Trace    read prGetTrace;    default;
end;

THEKA_Tree_Pulse9_Trace     = class(TList)
protected
  prTraceNode         :INode;
  prTrace             :TPTraceRecord9;

public
  constructor Create( TraceNode :INode; NeedsByteSwap :Boolean );

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec        :TPTraceRecord9            read prTrace;
end;

function LoadPulse9Tree( Stream :TStream ): THEKA_Tree_Pulse9; // returns nil on error


// *****************************************************************************

type

THEKA_Tree_StimRoot       = class;
THEKA_Tree_Stimulation    = class;
THEKA_Tree_StimSeg        = class;

THEKA_Tree_StimRoot       = class(TList)
protected
  prRootNode              :INode;
  prRoot                  :TPStimRootRecord;

  function prGetStim(Ix :Integer): THEKA_Tree_Stimulation;

public
  constructor Create( Tree :INode );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                      :TPStimRootRecord             read prRoot;
  property Stimulation[Ix :Integer] :THEKA_Tree_Stimulation       read prGetStim;  default;
end;

THEKA_Tree_Stimulation    = class(TList)
protected
  prStimNode                     :INode;
  prStim                         :TPStimulationRecord;

  function prGetSeg(Ix :Integer): THEKA_Tree_StimSeg;

public
  constructor Create( StimNode :INode; NeedsByteSwap :Boolean );
  destructor Destroy; override;

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec                   :TPStimulationRecord           read prStim;
  property Seg[Ix :Integer]      :THEKA_Tree_StimSeg            read prGetSeg;  default;
end;

THEKA_Tree_StimSeg    = class(TList)
protected
  prSegNode           :INode;
  prSeg               :TPStimSegmentRecord;

public
  constructor Create( SegNode :INode; NeedsByteSwap :Boolean );

  procedure Print( Indent :String; Stream :TStringStream );

  property Rec        :TPStimSegmentRecord            read prSeg;
end;


function LoadStimTree( Stream :TStream ): THEKA_Tree_StimRoot; // returns nil on error



implementation


(*
 * MagicNumber - This is a special value used as a prefix for a tree
 * stored to a file.  The value contains the four byte values of the
 * characters 'Tree' in order.
 *
 * SwappedMagicNumber - This is the MagicNumber written by a CPU
 * which used the opposite byte ordering.
 *)

CONST
   MagicNumber        = $054726565; (* i.e. "Tree" in ASCII *)
   SwappedMagicNumber = $065657254; (* i.e. "eerT" in ASCII *)

   //MagicNumberDAT2    = $044415432; (* i.e. "DAT2" in ASCII *)
   MagicNumberDAT2    = $032544144; (* i.e. "2TAD" in ASCII  (wtf) *)

(*
 * SwappedInt32 - Swaps the byte of a 32 bit long variable.
 *)

function SwappedInt32( Value : Integer ): Integer;
VAR
   Source,
   Target      : TPAByte;
BEGIN
   Source      := @Value;
   Target      := @Result;
   Target^[0]  := Source^[3];
   Target^[1]  := Source^[2];
   Target^[2]  := Source^[1];
   Target^[3]  := Source^[0];
END;


(*
 * LoadOneRecord
 *
 *    Loads one data block.
 *    All "TermIO" statements are for demonstration purpose only.
 *
 *    The variables are
 *           Stream        : the file handle to the open file
 *           FileSize      : the number of bytes the data block has in the 
 *                           file.
 *           MemorySize    : the byte length of the memory block where the
 *                           data is going to be stored.
 *           WhereToStore  : the address of where the data block is going 
 *                           to be stored.
 *
 *    The procedure returns TRUE, if it encountered no errors.
 *)

function LoadOneRecord(
             Stream       : TStream;
             FileSize     : LONGINT;
             MemorySize   : LONGINT;
             WhereToStore : Pointer )
             : BOOLEAN;

VAR
   Excess    : LONGINT;
   FileBytes : LONGINT;

BEGIN

   (* Here we load the next block of data into memory.
    *
    * First, we have to compare the number of bytes we can load from the file
    * with the bytes of allocated memory for that record.
    *
    * There are 3 possibilities:
    *    1. The size of the allocated memory ("MemorySize") equals the number
    *       of bytes of the data block in the file ("FileSize"). Thus, we can
    *       load the complete block.
    *    2. There are fewer bytes in the file than we expect. This can occur,
    *       e.g., when the file has been written by an earlier version which 
    *       used fewer parameters than the present one. In this case, we would
    *       have to zero out those fields which are not filled with data from 
    *       the file.
    *    3. There are more bytes in the file than we expect. This can happen, 
    *       when the program which created that tree file was using more 
    *       parameters than we presently know of. In that case, we would load 
    *       only as much byte as we had reserved RAM for.
    *)

   Excess    := MemorySize - FileSize;

   IF Excess = 0 THEN begin
      (* The file record has as many bytes as there is space in RAM *)

      FileBytes := MemorySize;

   end else if Excess < 0 THEN begin
      (* The file record has more many bytes than there is space in RAM.
       * Load only as many bytes as there is space in RAM.
       *)

      FileBytes := MemorySize;

   end ELSE begin (* i.e., Excess > 0D *)
      (* The file record has fewer bytes than there is space in RAM.
       * Load only as many bytes as there are in the file.
       *)

      FileBytes := FileSize;

      (* Do not forget to clear the remaining fields which are not going
       * to be filled from the file.
       *)

      FillChar( PChar(WhereToStore)[FileSize], Excess, 0 );
      // Buffer.Set( ADDADR( WhereToStore, FileSize ), Excess, 0 );

   END (* IF *);

   Stream.ReadBuffer( WhereToStore^, FileBytes );
   Result := True;
   // IOBytes.Read( Stream, FileBytes, WhereToStore );

END; // LoadOneRecord;


(*
 * LoadOneLevel
 *
 *    Processes the loading of one data block from a "Tree", and all
 *    its "children".
 *    All "TermIO" statements are for demonstration purpose only.
 *
 *    The variables are
 *           Stream        : the file handle to the open file
 *           Sizes         : the array containing the level sizes
 *           Levels        : the number of levels in the tree
 *           NeedsByteSwap : the flag telling, whether byte-swapping is needed
 *           Level         : the actual tree level to load
 *           IsFirst       : a flag telling, whether it is the first child 
 *                           loaded. This is only required for text output!
 *           Parent        : add this "level" (record) as a child of this qub tree
 *           Position      : the variable containing the position in the file.
 *
 *    The procedure returns TRUE, if it encountered no errors.
 *)

function LoadOneLevel(
             Stream        : TStream;
             Sizes         : TPAInteger;
             Levels        : LONGINT;
             NeedsByteSwap : BOOLEAN;
             Level         : LONGINT;
             Parent        : INode;
             VAR Position  : LONGINT )
             : BOOLEAN;

VAR
   Count     : Integer;
   Size      : LONGINT;
   Children  : LONGINT;
   MyNode    : INode;

BEGIN

   MyNode    := Parent.AppendChild('');
   MyNode.SetupData( NODE_DATATYPE_UNKNOWN, 1, Sizes[Level] );

   IF NOT
      LoadOneRecord( Stream, Sizes[Level], MyNode.DataCount, MyNode.DataPtr )
   THEN begin
      // Alert.IOError( '7-Error' );
      // RETURN FALSE;

      Result := False;
      Exit;
   END;

   (* If byte-swapping is required, we would now have to swap the bytes of all 
      fields in the loaded record!       (this is now the consumer's responsibility [see LoadPulseTree -> THEKA_Tree_Pulse.Create])
    *)

   (* Increase the file pointer by "Sizes[Level]" bytes and set the file position
    * just beyond the next data block:
    *)

   INC( Position, Sizes[Level] );

   IF Position <> Stream.Seek( Position, soFromBeginning )
   THEN begin
      // Alert.IOError( '8-Error' );
      // RETURN FALSE;

      Result := False;
      Exit;
   END;


   (* The next 4 bytes contain the number of children of the present level. *)

   Size := SIZEOF( Integer );

   IF Size <> Stream.Read( Count, Size ) then begin
      //Alert.IOError( '9-Error' );
      //RETURN FALSE;

      Result := False;
      Exit;
   END;

   (* The file pointer increased by 4 bytes: *)
   INC( Position, 4 );

   (* And we swap the bytes, if needed: *)

   IF NeedsByteSwap THEN begin Count := SwappedInt32( Count ); END;

   (* Now, we can proceed to load all the children of the present level,
    * if there are any:
    *)

   Children := 0;

   IF Level < Levels THEN begin

      WHILE Children < Count DO begin

         IF NOT
            LoadOneLevel(
               Stream,
               Sizes,
               Levels,
               NeedsByteSwap,
               Level + 1,
               MyNode,
               Position )
         THEN begin
            Result := FALSE;
            Exit;
         END; (* IF *)

         INC( Children );

      END (* WHILE *);

   END (* IF *);

   Result := True;

END; //  LoadOneLevel;




(*
 * LoadTree
 *
 *    Scans a complete Tree.
 *    All "TermIO" statements are for demonstration purpose only.
 *
 *    The variables are
 *           Stream        : the file handle to the open file
 *           Sizes         : the array is returns the level sizes in the Tree
 *                           on disk.
 *           Levels        : the number of levels in the tree
 *           NeedsByteSwap : the flag telling, whether byte-swapping is needed
 *
 *    The procedure returns TRUE, if it encountered no errors.
 *)




function LoadPULFromBundle(
             Stream             : TStream;
             Sizes              : TPAInteger;
             VAR Levels         : LONGINT;
             VAR NeedsByteSwap  : BOOLEAN;
             AsChildrenOf       : INode )
             : Boolean;
VAR
  Bundle    :BundleHeader;
  nRead     :Integer;
  i         :Integer;
begin
     // the PUL info we're looking for is bundled in with data and other resources;
     // we find the offset of PUL, adjust the stream and read like usual
  Result    := True;
  Stream.Position := 0;
  nRead     := Stream.Read(Bundle, Sizeof(BundleHeader));
  if nRead < Sizeof(BundleHeader) then begin
    Exit;
  end;
  for i := 0 to 11 do begin
    if String(Bundle.oBundleItems[i].oExtension) = '.pul' then begin
      Stream.Position := Bundle.oBundleItems[i].oStart;
      Result := LoadTree(Stream, Sizes, Levels, NeedsByteSwap, AsChildrenOf);
      Exit;
    end;
  end;
end;

function LoadTree(
             Stream             : TStream;
             Sizes              : TPAInteger;
             VAR Levels         : LONGINT;
             VAR NeedsByteSwap  : BOOLEAN;
             AsChildrenOf       : INode )
             : Boolean; overload;
VAR
   Value     : Integer;
   Position  : LONGINT;
   Size      : LONGINT;
   i         : INTEGER;
   Success   : BOOLEAN;

BEGIN

   Result := False;

   (* We start at the beginning of the file. We keep the variable
    * "Position" containing the actual position in the file.
    *)

   Position := Stream.Position;


   (* The first 4 bytes should contain the "MagicNumber", see above.
    * a variable of type Integer is a 32-bit long, signed word.
    *)

   Size := SIZEOF( Integer );

   IF Size <> Stream.Read( Value, Size ) then begin
      // Alert.IOError( '2-Error' );
      // RETURN FALSE;

      Result := False;
      Exit;
   END;

   if Value = MagicNumberDAT2 then begin // TODO: any big-endian files out there?
     Result := LoadPULFromBundle(Stream, Sizes, Levels, NeedsByteSwap, AsChildrenOf);
     Exit;
   end;

   IF Value = MagicNumber THEN begin
      NeedsByteSwap := FALSE;
   end else if Value = SwappedMagicNumber THEN begin
      NeedsByteSwap := TRUE;
   end ELSE begin
      // Alert.OK( '3-Error: File does not start with "Tree" !' );
      // RETURN FALSE;

      Result := False;
      Exit;
   END; (* IF *)

   (* The file pointer increased by 4 bytes: *)
   INC( Position, 4 );


   (* Next we load the number of levels in the Tree, which is stored in the
    * next 4 bytes (at offset 4):
    *)

   Size := SIZEOF( Integer );

   IF Size <> Stream.Read( Levels, Size ) then begin
      // Alert.IOError( '4-Error' );
      // RETURN FALSE;

      Result := False;
      Exit;
   END;

   (* The file pointer increased by 4 bytes: *)
   INC( Position, 4 );


   (* If the file originates from a platform with opposite byte ordering,
    * then we have to swap the bytes:
    *)

   IF NeedsByteSwap THEN begin Levels := SwappedInt32( Levels ); END;

   // TermIO.WriteString( '  -> levels: ' );
   // TermIO.WriteLongInt( Levels, 0 );


   (* The next bytes contain the sizes of all levels. Thus, there is
    * one 4-byte variable for each level, totaling in "Levels" times 4
    * bytes.
    *
    * First, we check, if the array "Sizes" passed to this procedure is
    * large enough to contain all level sizes:
    *)

   IF ( Levels <= 0 ) then begin // OR ( Levels > LONGINT(HIGH(Sizes)+1) ) THEN begin
      // Alert.OK( '5-Error: number of level either <= 0 or too large!' );
      // ETURN FALSE;

      Result := False;
      Exit;
   END (* IF *);


   (* Next, we load the "Level Size": *)

   Size := Levels * LONGINT( SIZEOF( Integer ) );

   IF Size <> Stream.Read( Sizes^, Size ) then begin
      // Alert.IOError( '6-Error' );
      // RETURN FALSE;

      Result := False;
      Exit;
   END;

   (* The file pointer increased by "Size" bytes: *)
   INC( Position, Size );

   (* And we swap the bytes, if needed: *)

   IF NeedsByteSwap THEN begin
      FOR i := 0 TO Levels - 1 DO begin
         Sizes[i] := SwappedInt32( Sizes[i] );
      END; (* FOR *)
   END; (* IF *)

   // TermIO.WriteString( ';  sizes: ' );
   // FOR i := 0 TO SHORT( Levels - 1D ) DO
   //    TermIO.WriteLongInt( Sizes[i], 0 );
   //    IF i < SHORT( Levels - 1D ) THEN
   //       TermIO.WriteString( ', ' );
   //    END; (* IF *)
   // END; (* FOR *)

   // TermIO.WriteString( ';  swap: ' );
   // TermIO.WriteBoolean( NeedsByteSwap );
   // TermIO.WriteLn;


   (* Now, the tree data follow.
    * We can load them by a recursive procedure:
    *)

   Success :=
      LoadOneLevel(
         Stream,
         Sizes,
         Levels,
         NeedsByteSwap,
         0,
         AsChildrenOf,
         Position );

   IF Success THEN begin
      Result := True;
   END; (* IF *)

END; // LoadTree;


function LoadTree( Stream :TStream ): INode; overload;
VAR
   Sizes          : ARRAY[0..9] OF Integer;
   Levels         : LONGINT;
   NeedsByteSwap  : BOOLEAN;

BEGIN

   Result         := TNode.Create('HEKA Tree');

   (* Now, load the "Tree" : *)

   if LoadTree( Stream, @(Sizes[0]), Levels, NeedsByteSwap, Result ) then begin
     Result['NeedsByteSwap'].SetData( Integer(NeedsByteSwap) );
   end else begin

   end;

END;


// *****************************************************************************
procedure SwapBytesA( x:Pointer; iSize:Integer; iCount:Integer );
var
  i,j,iHalf : Integer;
  Bytes : TPAByte;
  b : Byte;
begin
  iHalf := iSize DIV 2;
  Bytes := TPAByte(x);
  for i := 0 to iCount-1 do
    for j := 0 to iHalf-1 do Begin
      b:=Bytes[i*iSize+j];
      Bytes[i*iSize+j]:=Bytes[(i+1)*iSize-j-1];
      Bytes[(i+1)*iSize-j-1]:=b;
    end;
end;

procedure Swap2Bytes(x :Pointer);
var
  Temp               :Word;
  Bytes, TempBytes   :TPAByte;
begin
  Bytes              := TPAByte(x);
  TempBytes          := TPAByte(@Temp);
  Move(Bytes^, TempBytes^, 2);
  Bytes[0]           := TempBytes[1];
  Bytes[1]           := TempBytes[0];
end;

procedure Swap4Bytes(x :Pointer);
var
  Temp               :LongWord;
  Bytes, TempBytes   :TPAByte;
begin
  Bytes              := TPAByte(x);
  TempBytes          := TPAByte(@Temp);
  Move(Bytes^, TempBytes^, 4);
  Bytes[0]           := TempBytes[3];
  Bytes[1]           := TempBytes[2];
  Bytes[2]           := TempBytes[1];
  Bytes[3]           := TempBytes[0];
end;

procedure Swap8Bytes(x :Pointer);
var
  Temp               :Double;
  Bytes, TempBytes   :TPAByte;
begin
  Bytes              := TPAByte(x);
  TempBytes          := TPAByte(@Temp);
  Move(Bytes^, TempBytes^, 8);
  Bytes[0]           := TempBytes[7];
  Bytes[1]           := TempBytes[6];
  Bytes[2]           := TempBytes[5];
  Bytes[3]           := TempBytes[4];
  Bytes[4]           := TempBytes[3];
  Bytes[5]           := TempBytes[2];
  Bytes[6]           := TempBytes[1];
  Bytes[7]           := TempBytes[0];
end;


constructor THEKA_Tree_Pulse.Create( Tree :INode );
var
  Child               :INode;
  NeedsByteSwap       :Boolean;
begin
  prRootNode         := Tree[0];
  prRoot             := TPPulseRootRecord( prRootNode.DataPtr );
  NeedsByteSwap      := False;
  try
    NeedsByteSwap      := 0 <> Integer(Tree['NeedsByteSwap'].Data[0]);
  except
  end;

  if NeedsByteSwap then begin
    Swap2Bytes( @(prRoot.Version) );
    Swap8Bytes( @(prRoot.StartTime) );
  end;

  Child               := prRootNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Pulse_Group.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Pulse.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Groups[i].Free;
  end;
end;

procedure THEKA_Tree_Pulse.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'PulseRoot'#13#10 );
  Indent               := Indent + #9;

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Version', Rec.Version]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'VersionName', Rec.VersionName]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'FileName', Rec.FileName]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'Comments', Rec.Comments]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'StartTime', Rec.StartTime]) );

  for i := 0 to Count - 1 do begin
    Groups[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Pulse.prGetGroup(Ix :Integer): THEKA_Tree_Pulse_Group;
begin
  Result := THEKA_Tree_Pulse_Group( Items[Ix] );
end;

constructor THEKA_Tree_Pulse_Group.Create( GroupNode :INode; NeedsByteSwap :Boolean );
var
  Child               :INode;
begin
  prGroupNode        := GroupNode;
  prGroup            := TPGroupRecord( GroupNode.DataPtr );

  if NeedsByteSwap then begin
    Swap4Bytes( @(prGroup.ExperimentNumber) );
    Swap8Bytes( @(prGroup.ExtraLongReal) );
  end;

  Child               := GroupNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Pulse_Series.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Pulse_Group.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Series[i].Free;
  end;
end;

procedure THEKA_Tree_Pulse_Group.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'Group'#13#10 );
  Indent               := Indent + #9;

  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'Lbl', Rec.Lbl]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'Text', Rec.Text]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'ExperimentNumber', Rec.ExperimentNumber]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'ExtraLongReal', Rec.ExtraLongReal]) );

  for i := 0 to Count - 1 do begin
    Series[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Pulse_Group.prGetSeries(Ix :Integer): THEKA_Tree_Pulse_Series;
begin
  Result := THEKA_Tree_Pulse_Series( Items[Ix] );
end;

constructor THEKA_Tree_Pulse_Series.Create( SeriesNode :INode; NeedsByteSwap :Boolean );
var
  Child               :INode;
  i                   :Integer;
begin
  prSeriesNode        := SeriesNode;
  prSeries            := TPSeriesRecord( SeriesNode.DataPtr );

  if NeedsByteSwap then begin
    Swap8Bytes( @(prSeries.Time) );
    Swap8Bytes( @(prSeries.Bandwidth) );
    Swap8Bytes( @(prSeries.PipettePotential) );
    Swap8Bytes( @(prSeries.CellPotential) );
    Swap8Bytes( @(prSeries.PipetteResistance) );
    Swap8Bytes( @(prSeries.SealResistance) );
    Swap8Bytes( @(prSeries.BackgroundNoise) );
    Swap8Bytes( @(prSeries.Temperature) );
    Swap8Bytes( @(prSeries.PipettePressure) );
    Swap8Bytes( @(prSeries.UserParam1.Value) );
    Swap8Bytes( @(prSeries.UserParam2.Value) );
    Swap4Bytes( @(prSeries.InternalSolution) );
    Swap4Bytes( @(prSeries.ExternalSolution) );
    Swap8Bytes( @(prSeries.FuraK) );
    Swap8Bytes( @(prSeries.FuraMin) );
    Swap8Bytes( @(prSeries.FuraMax) );
    Swap8Bytes( @(prSeries.LockInExtPhase) );
    Swap8Bytes( @(prSeries.Timer) );

    for i := 0 to SeExtraLongReals - 1 do begin
      Swap8Bytes( @(prSeries.ExtraLongReal[i]) );
    end;
  end;

  Child               := SeriesNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Pulse_Sweep.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Pulse_Series.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Sweeps[i].Free;
  end;
end;

procedure THEKA_Tree_Pulse_Series.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'Series'#13#10 );
  Indent               := Indent + #9;

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Time', Rec.Time]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Bandwidth', Rec.Bandwidth]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'PipettePotential', Rec.PipettePotential]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'CellPotential', Rec.CellPotential]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'PipetteResistance', Rec.PipetteResistance]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'SealResistance', Rec.SealResistance]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'BackgroundNoise', Rec.BackgroundNoise]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Temperature', Rec.Temperature]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'PipettePressure', Rec.PipettePressure]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'RecordingMode', Integer(Rec.RecordingMode)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Filler1', Integer(Rec.Filler1)]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'Comment', Rec.Comment]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'InternalSolution', Integer(Rec.InternalSolution)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'ExternalSolution', Integer(Rec.ExternalSolution)]) );

  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'DispYUnit1', Rec.DispYUnit1]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'DispYUnit2', Rec.DispYUnit2]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraK', Rec.FuraK]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraMin', Rec.FuraMin]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraMax', Rec.FuraMax]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LockInExtPhase', Rec.LockInExtPhase]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Timer', Rec.Timer]) );

  for i := 0 to Count - 1 do begin
    Sweeps[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Pulse_Series.prGetSweep(Ix :Integer): THEKA_Tree_Pulse_Sweep;
begin
  Result := THEKA_Tree_Pulse_Sweep( Items[Ix] );
end;

constructor THEKA_Tree_Pulse_Sweep.Create( SweepNode :INode; NeedsByteSwap :Boolean );
begin
  prSweepNode         := SweepNode;
  prSweep             := TPSweepRecord( SweepNode.DataPtr );

  if NeedsByteSwap then begin
    Swap8Bytes( @(prSweep.Time) );
    Swap4Bytes( @(prSweep.StimCount) );
    Swap4Bytes( @(prSweep.SweepCount) );
    Swap4Bytes( @(prSweep.AverageCount) );
    Swap4Bytes( @(prSweep.DataPoints) );
    Swap4Bytes( @(prSweep.Data) );
    Swap8Bytes( @(prSweep.DataFactor1) );
    Swap8Bytes( @(prSweep.DataFactor2) );
    Swap8Bytes( @(prSweep.CSlow) );
    Swap8Bytes( @(prSweep.GSeries) );
    Swap8Bytes( @(prSweep.RsValue) );
    Swap8Bytes( @(prSweep.RsFraction) );
    Swap8Bytes( @(prSweep.ZeroCurrent) );
    Swap8Bytes( @(prSweep.OnlineYResult) );
    Swap8Bytes( @(prSweep.OnlineXResult) );
    Swap4Bytes( @(prSweep.TotalPoints) );
    Swap4Bytes( @(prSweep.Offset) );
    Swap2Bytes( @(prSweep.SweepKind) );
    Swap4Bytes( @(prSweep.FuraPoints) );
    Swap4Bytes( @(prSweep.FuraData) );
    Swap8Bytes( @(prSweep.OnlineYResult2) );
    Swap8Bytes( @(prSweep.OnlineXResult2) );
    Swap8Bytes( @(prSweep.DispFactor1) );
    Swap8Bytes( @(prSweep.DispFactor2) );
    Swap8Bytes( @(prSweep.Timer) );
  end;
end;

procedure THEKA_Tree_Pulse_Sweep.Print( Indent :String; Stream :TStringStream );
begin
  Stream.WriteString( Indent + 'Sweep'#13#10 );
  Indent               := Indent + #9;

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Time', Rec.Time]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'StimCount', Rec.StimCount]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'SweepCount', Rec.SweepCount]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'AverageCount', Rec.AverageCount]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Leak', Integer(Rec.Leak)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'SecondTrace', Integer(Rec.SecondTrace)]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'Lbl', Rec.Lbl]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'DataPoints', Rec.DataPoints]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Data', Rec.Data]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'DataPointer', Integer(Rec.DataPointer)]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DataFactor1', Rec.DataFactor1]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DataFactor2', Rec.DataFactor2]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'CSlow', Rec.CSlow]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'GSeries', Rec.GSeries]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'RsValue', Rec.RsValue]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'RsFraction', Rec.RsFraction]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'ZeroCurrent', Rec.ZeroCurrent]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'OnlineYResult', Rec.OnlineYResult]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'OnlineXResult', Rec.OnlineXResult]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'TotalPoints', Rec.TotalPoints]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Offset', Rec.Offset]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'SweepKind', Integer(Rec.SweepKind)]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'FuraPoints', Rec.FuraPoints]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'FuraData', Rec.FuraData]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'FuraPointer', Integer(Rec.FuraPointer)]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'OnlineYResult2', Rec.OnlineYResult2]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'OnlineXResult2', Rec.OnlineXResult2]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DispFactor1', Rec.DispFactor1]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DispFactor2', Rec.DispFactor2]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'DataFormat', Integer(Rec.DataFormat)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'DataAbscissa', Integer(Rec.DataAbscissa)]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Timer', Rec.Timer]) );

end;


function LoadPulseTree( Stream :TStream ): THEKA_Tree_Pulse;
var
  Tree          :INode;
  Sweep         :INode;
begin
  Result        := nil;

  try
    Tree        := LoadTree( Stream );
//    if not Tree.Child.IsNull then begin
// count 4 levels:
    Sweep       := Tree[0].Child.Child.Child;
    if (not Sweep.IsNull) and (Sweep.Child.IsNull) then begin
      Result    := THEKA_Tree_Pulse.Create( Tree );
    end;
  except
  end;
end;



constructor THEKA_Tree_Pulse9.Create( Tree :INode );
var
  Child               :INode;
  NeedsByteSwap       :Boolean;
begin
  prRootNode         := Tree[0];
  prRoot             := TPPulseRootRecord9( prRootNode.DataPtr );
  NeedsByteSwap      := False;
  try
    NeedsByteSwap      := 0 <> Integer(Tree['NeedsByteSwap'].Data[0]);
  except
  end;

  if NeedsByteSwap then begin
    Swap4Bytes( @(prRoot.Version) );
    Swap4Bytes( @(prRoot.Mark) );
    Swap8Bytes( @(prRoot.StartTime) );
    Swap4Bytes( @(prRoot.MaxSamples) );
    Swap4Bytes( @(prRoot.CRC) );
    Swap2Bytes( @(prRoot.Features) );
  end;

  Child               := prRootNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Pulse9_Group.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Pulse9.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Groups[i].Free;
  end;
end;

procedure THEKA_Tree_Pulse9.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'PulseRoot'#13#10 );
  Indent               := Indent + #9;

  for i := 0 to Count - 1 do begin
    Groups[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Pulse9.prGetGroup(Ix :Integer): THEKA_Tree_Pulse9_Group;
begin
  Result := THEKA_Tree_Pulse9_Group( Items[Ix] );
end;

constructor THEKA_Tree_Pulse9_Group.Create( GroupNode :INode; NeedsByteSwap :Boolean );
var
  Child               :INode;
begin
  prGroupNode        := GroupNode;
  prGroup            := TPGroupRecord9( GroupNode.DataPtr );

  if NeedsByteSwap then begin
    Swap4Bytes( @(prGroup.Mark) );
    Swap4Bytes( @(prGroup.ExperimentNumber) );
    Swap4Bytes( @(prGroup.GroupCount) );
    Swap4Bytes( @(prGroup.CRC) );
  end;

  Child               := GroupNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Pulse9_Series.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Pulse9_Group.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Series[i].Free;
  end;
end;

procedure THEKA_Tree_Pulse9_Group.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'Group'#13#10 );
  Indent               := Indent + #9;

  for i := 0 to Count - 1 do begin
    Series[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Pulse9_Group.prGetSeries(Ix :Integer): THEKA_Tree_Pulse9_Series;
begin
  Result := THEKA_Tree_Pulse9_Series( Items[Ix] );
end;

constructor THEKA_Tree_Pulse9_Series.Create( SeriesNode :INode; NeedsByteSwap :Boolean );
var
  Child               :INode;
  i                   :Integer;
begin
  prSeriesNode        := SeriesNode;
  prSeries            := TPSeriesRecord9( SeriesNode.DataPtr );

  if NeedsByteSwap then begin
    Swap4Bytes( @(prSeries.Mark) );
    Swap4Bytes( @(prSeries.SeriesCount) );
    Swap4Bytes( @(prSeries.NumberSweeps) );
    Swap4Bytes( @(prSeries.AmplStateOffset) );
    Swap4Bytes( @(prSeries.AmplStateSeries) );
    Swap8Bytes( @(prSeries.Time) );
    Swap8Bytes( @(prSeries.PageWidth) );
    Swap8Bytes( @(prSeries.UserParams1[0]) );
    Swap8Bytes( @(prSeries.UserParams1[1]) );
    Swap8Bytes( @(prSeries.UserParams1[2]) );
    Swap8Bytes( @(prSeries.UserParams1[3]) );
    Swap4Bytes( @(prSeries.CRC) );
    Swap8Bytes( @(prSeries.UserParams2[0]) );
    Swap8Bytes( @(prSeries.UserParams2[1]) );
    Swap8Bytes( @(prSeries.UserParams2[2]) );
    Swap8Bytes( @(prSeries.UserParams2[3]) );
  end;

  Child               := SeriesNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Pulse9_Sweep.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Pulse9_Series.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Sweeps[i].Free;
  end;
end;

procedure THEKA_Tree_Pulse9_Series.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'Series'#13#10 );
  Indent               := Indent + #9;

  for i := 0 to Count - 1 do begin
    Sweeps[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Pulse9_Series.prGetSweep(Ix :Integer): THEKA_Tree_Pulse9_Sweep;
begin
  Result := THEKA_Tree_Pulse9_Sweep( Items[Ix] );
end;

constructor THEKA_Tree_Pulse9_Sweep.Create( SweepNode :INode; NeedsByteSwap :Boolean );
var
  Child               :INode;
begin
  prSweepNode         := SweepNode;
  prSweep             := TPSweepRecord9( SweepNode.DataPtr );

  if NeedsByteSwap then begin
    Swap4Bytes( @(prSweep.Mark) );
    Swap4Bytes( @(prSweep.AuxDataFileOffset) );
    Swap4Bytes( @(prSweep.StimCount) );
    Swap4Bytes( @(prSweep.SweepCount) );
    Swap8Bytes( @(prSweep.Time) );
    Swap8Bytes( @(prSweep.Timer) );
    Swap8Bytes( @(prSweep.UserParams[0]) );
    Swap8Bytes( @(prSweep.UserParams[1]) );
    Swap8Bytes( @(prSweep.UserParams[2]) );
    Swap8Bytes( @(prSweep.UserParams[3]) );
    Swap8Bytes( @(prSweep.Temperature) );
    Swap4Bytes( @(prSweep.OldIntSol) );
    Swap4Bytes( @(prSweep.OldExtSol) );
    Swap2Bytes( @(prSweep.DigitalIn) );
    Swap8Bytes( @(prSweep.Markers[0]) );
    Swap8Bytes( @(prSweep.Markers[1]) );
    Swap8Bytes( @(prSweep.Markers[2]) );
    Swap8Bytes( @(prSweep.Markers[3]) );
    Swap4Bytes( @(prSweep.CRC) );

  end;

  Child               := SweepNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Pulse9_Trace.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Pulse9_Sweep.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Traces[i].Free;
  end;
end;

procedure THEKA_Tree_Pulse9_Sweep.Print( Indent :String; Stream :TStringStream );
var
  i                    :Integer;
begin
  Stream.WriteString( Indent + 'Sweep'#13#10 );
  Indent               := Indent + #9;


  for i := 0 to Count - 1 do begin
    Traces[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Pulse9_Sweep.prGetTrace(Ix :Integer): THEKA_Tree_Pulse9_Trace;
begin
  Result  := THEKA_Tree_Pulse9_Trace( Items[Ix] );
end;

constructor THEKA_Tree_Pulse9_Trace.Create( TraceNode :INode; NeedsByteSwap :Boolean );
var
  i                   :Integer;
begin
  prTraceNode         := TraceNode;
  prTrace             := TPTraceRecord9( TraceNode.DataPtr );

  if NeedsByteSwap then begin
    Swap4Bytes( @(prTrace.Mark) );
    Swap4Bytes( @(prTrace.TraceCount) );
    Swap4Bytes( @(prTrace.Data) );
    Swap4Bytes( @(prTrace.DataPoints) );
    Swap4Bytes( @(prTrace.InternalSolution) );
    Swap4Bytes( @(prTrace.AverageCount) );
    Swap4Bytes( @(prTrace.LeakCount) );
    Swap4Bytes( @(prTrace.LeakTraces) );
    Swap2Bytes( @(prTrace.DataKind) );
    Swap8Bytes( @(prTrace.DataScaler) );
    Swap8Bytes( @(prTrace.TimeOffset) );
    Swap8Bytes( @(prTrace.ZeroData) );
    Swap8Bytes( @(prTrace.XInterval) );
    Swap8Bytes( @(prTrace.XStart) );
    Swap8Bytes( @(prTrace.YRange) );
    Swap8Bytes( @(prTrace.YOffset) );
    Swap8Bytes( @(prTrace.Bandwidth) );
    Swap8Bytes( @(prTrace.PipetteResistance) );
    Swap8Bytes( @(prTrace.CellPotential) );
    Swap8Bytes( @(prTrace.SealResistance) );
    Swap8Bytes( @(prTrace.CSlow) );
    Swap8Bytes( @(prTrace.GSeries) );
    Swap8Bytes( @(prTrace.RsValue) );
    Swap8Bytes( @(prTrace.gLeak) );
    Swap8Bytes( @(prTrace.MConductance) );
    Swap4Bytes( @(prTrace.LinkDAChannel) );
    Swap2Bytes( @(prTrace.AdcChannel) );
    Swap8Bytes( @(prTrace.YMin) );
    Swap8Bytes( @(prTrace.YMax) );
    Swap4Bytes( @(prTrace.SourceChannel) );
    Swap4Bytes( @(prTrace.ExternalSolution) );
    Swap8Bytes( @(prTrace.CM) );
    Swap8Bytes( @(prTrace.GM) );
    Swap8Bytes( @(prTrace.Phase) );
    Swap4Bytes( @(prTrace.DataCRC) );
    Swap4Bytes( @(prTrace.CRC) );
    Swap8Bytes( @(prTrace.GS) );
    Swap4Bytes( @(prTrace.SelfChannel) );
    Swap4Bytes( @(prTrace.InterleaveSize) );
    Swap4Bytes( @(prTrace.InterleaveSkip) );
    Swap4Bytes( @(prTrace.ImageIndex) );
    for i := 0 to 9 do begin
      Swap8Bytes( @(prTrace.Markers[i]) );
    end;
    Swap8Bytes( @(prTrace.SECM_X) );
    Swap8Bytes( @(prTrace.SECM_Y) );
    Swap8Bytes( @(prTrace.SECM_Z) );

  end;
end;

procedure THEKA_Tree_Pulse9_Trace.Print( Indent :String; Stream :TStringStream );
begin
  Stream.WriteString( Indent + 'Trace'#13#10 );
  Indent               := Indent + #9;

end;


function LoadPulse9Tree( Stream :TStream ): THEKA_Tree_Pulse9;
var
  Tree          :INode;
begin
  Result        := nil;

  try
    Tree        := LoadTree( Stream );
    if not Tree.Child.IsNull then begin
      Result    := THEKA_Tree_Pulse9.Create( Tree );
    end;
  except
  end;
end;


// *****************************************************************************

constructor THEKA_Tree_StimRoot.Create( Tree :INode );
var
  Child               :INode;
  NeedsByteSwap       :Boolean;
begin
  prRootNode         := Tree[0];
  prRoot             := TPStimRootRecord( prRootNode.DataPtr );
  NeedsByteSwap      := False;
  try
    NeedsByteSwap      := 0 <> Integer(Tree['NeedsByteSwap'].Data[0]);
  except
  end;

  {
  if NeedsByteSwap then begin
    Swap8Bytes( @(prRoot.StartTime) );
  end;}

  Child               := prRootNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_Stimulation.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_StimRoot.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Stimulation[i].Free;
  end;
end;

procedure THEKA_Tree_StimRoot.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'StimRoot'#13#10 );
  Indent               := Indent + #9;
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Version', Integer(Rec.Version)]) );


  for i := 0 to Count - 1 do begin
    Stimulation[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_StimRoot.prGetStim(Ix :Integer): THEKA_Tree_Stimulation;
begin
  Result := THEKA_Tree_Stimulation( Items[Ix] );
end;

constructor THEKA_Tree_Stimulation.Create( StimNode :INode; NeedsByteSwap :Boolean );
var
  Child               :INode;
begin
  prStimNode         := StimNode;
  prStim             := TPStimulationRecord( prStimNode.DataPtr );

  if NeedsByteSwap then begin
    Swap8Bytes( @(prStim.SampleInterval) );
    Swap8Bytes( @(prStim.FilterFactor) );
    Swap8Bytes( @(prStim.SweepInterval) );
    Swap4Bytes( @(prStim.NumberSweeps) );
    Swap4Bytes( @(prStim.NumberRepeats) );
    Swap8Bytes( @(prStim.RepeatWait) );
  end;

  Child               := prStimNode.Child;
  while not Child.IsNull do begin
    Add( THEKA_Tree_StimSeg.Create( Child, NeedsByteSwap ) );
    Child             := Child.Sibling;
  end;
end;

destructor THEKA_Tree_Stimulation.Destroy;
var
  i                   :Integer;
begin
  for i := 0 to Count - 1 do begin
    Seg[i].Free;
  end;
end;

procedure THEKA_Tree_Stimulation.Print( Indent :String; Stream :TStringStream );
var
  i                   :Integer;
begin
  Stream.WriteString( Indent + 'Stimulation'#13#10 );
  Indent               := Indent + #9;
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'FileName', Rec.FileName]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'EntryName', Rec.EntryName]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'SampleInterval', Rec.SampleInterval]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FilterFactor', Rec.FilterFactor]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'SweepInterval', Rec.SweepInterval]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'NumberSweeps', Rec.NumberSweeps]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'NumberRepeats', Rec.NumberRepeats]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'RepeatWait', Rec.RepeatWait]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'LinkedSequence', Rec.LinkedSequence]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LinkedWait', Rec.LinkedWait]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'LeakCount', Rec.LeakCount]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LeakSize', Rec.LeakSize]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LeakHolding', Rec.LeakHolding]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'LeakAlternate', Integer(Rec.LeakAlternate)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'AltLeakAveraging', Integer(Rec.AltLeakAveraging)]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LeakDelay', Rec.LeakDelay]) );

  // Trig                 : ARRAY [0..2] OF Trigger;
  Stream.WriteString( Indent + '(Trigger Record omitted)'#13#10 );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'NumberOfTriggers', Rec.NumberOfTriggers]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'RelevantXSegment', Rec.RelevantXSegment]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'RelevantYSegment', Rec.RelevantYSegment]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'WriteMode', Integer(Rec.WriteMode)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'IncrementMode', Integer(Rec.IncrementMode)]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'TotalSweepLength', Rec.TotalSweepLength]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'MaxSweepLength', Rec.MaxSweepLength]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'InputChannels', Rec.InputChannels]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'GUpdate', Integer(Rec.GUpdate)]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'RelAbsPot', Integer(Rec.RelAbsPot)]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'HasContinuous', Integer(Rec.HasContinuous)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'LogIncrement', Integer(Rec.LogIncrement)]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'StimDac', Rec.StimDac]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Adc1', Rec.Adc1]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Adc2', Rec.Adc2]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'YUnit1', Rec.YUnit1]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'YUnit2', Rec.YUnit2]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'VmembIncrement', Rec.VmembIncrement]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'ExtTrigger', Integer(Rec.ExtTrigger)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'FileTemplate', Integer(Rec.FileTemplate)]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'StimKind', Integer(Rec.StimKind)]) );

  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LockInCycle', Rec.LockInCycle]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LockInAmplitude', Rec.LockInAmplitude]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'FuraOn', Integer(Rec.FuraOn)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'VmembMode', Integer(Rec.VmembMode)]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraTotLength', Rec.FuraTotLength]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraDelay', Rec.FuraDelay]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraLength1', Rec.FuraLength1]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraLength2', Rec.FuraLength2]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraWaveLength0', Rec.FuraWaveLength0]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraWaveLength1', Rec.FuraWaveLength1]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'FuraWaveLength2', Rec.FuraWaveLength2]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'FuraRepeats', Rec.FuraRepeats]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'LockInSkip', Rec.LockInSkip]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'LockInVReversal', Rec.LockInVReversal]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'LockInMode', Integer(Rec.LockInMode)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'LockInShow', Integer(Rec.LockInShow)]) );

  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'ConfigMacro', Rec.ConfigMacro]) );
  Stream.WriteString( Format('%s%s'#9'%s'#13#10, [Indent, 'EndMacro', Rec.EndMacro]) );

  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'AmplModeKind', Integer(Rec.AmplModeKind)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Filler1', Integer(Rec.Filler1)]) );

  for i := 0 to Count - 1 do begin
    Seg[i].Print(Indent, Stream);
  end;
end;

function THEKA_Tree_Stimulation.prGetSeg(Ix :Integer): THEKA_Tree_StimSeg;
begin
  Result := THEKA_Tree_StimSeg( Items[Ix] );
end;

constructor THEKA_Tree_StimSeg.Create( SegNode :INode; NeedsByteSwap :Boolean );
begin
  prSegNode         := SegNode;
  prSeg             := TPStimSegmentRecord( prSegNode.DataPtr );
  {
  if NeedsByteSwap then begin
    Swap2Bytes( @(prSeg.Version) );
    Swap8Bytes( @(prSeg.StartTime) );
  end;  }
end;

procedure THEKA_Tree_StimSeg.Print( Indent :String; Stream :TStringStream );
begin
  Stream.WriteString( Indent + 'StimSeg'#13#10 );
  Indent               := Indent + #9;
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'Cls', Integer(prSeg.Cls)]) );
  Stream.WriteString( Format('%s%s'#9'%d'#13#10, [Indent, 'IsHolding', Integer(prSeg.IsHolding)]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Voltage', prSeg.Voltage]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'Duration', prSeg.Duration]) );       
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DeltaVFactor', prSeg.DeltaVFactor]) );   
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DeltaVIncrement', prSeg.DeltaVIncrement]) );
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DeltaTFactor', prSeg.DeltaTFactor]) );   
  Stream.WriteString( Format('%s%s'#9'%f'#13#10, [Indent, 'DeltaTIncrement', prSeg.DeltaTIncrement]) );
end;

function LoadStimTree( Stream :TStream ): THEKA_Tree_StimRoot; // returns nil on error
var
  Tree          :INode;
begin
  Result        := nil;

  try
    Tree        := LoadTree( Stream );
    if not Tree.Child.IsNull then begin
      Result    := THEKA_Tree_StimRoot.Create( Tree );
    end;
  except
  end;
end;


// status so far:
//   error reporting pathetic:
//     LoadTree should put error messages in the tree,
//     but right now a return tree with no children indicates failure
//     no idea how to replace the TermIO/Alert/ReturnFalse
//   byte swapping considered experimental: swap doubles or not?


// *****************************************************************************
end.

