{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_ObjectCache;


// Chris Nicolai // June 2003
{
This is an associative container for no more than N objects,
whose creation and destruction are managed by the container.
When you request the object for a new key:
  * if already at capacity: it destroys the least-recently-accessed object
  * it creates and returns a new object.
The keys and objects can be anything -- integers even.
You provide methods to create and destroy objects, and the maximum
objects to keep at a time.
When the container is freed, it destroys all remaining objects.
}

interface

uses

Classes;


type

TQUB_ObjectCache_CreateFn  = function  (Key :Pointer): Pointer of object;
TQUB_ObjectCache_DestroyFn = procedure (Key, Obj :Pointer)     of object;

TQUB_ObjectCache = class
protected
  prCreate           :TQUB_ObjectCache_CreateFn;
  prDestroy          :TQUB_ObjectCache_DestroyFn;
  prMax              :Integer;
  prKeys             :TList;
  prObjects          :TList;

  function prGetObject(Key :Pointer): Pointer;

public
  constructor Create( CreateFn  :TQUB_ObjectCache_CreateFn;
                      DestroyFn :TQUB_ObjectCache_DestroyFn;
                      MaxObjects :Integer );

  destructor Destroy; override;

  procedure FreeObject(Key :Pointer);

  property Objects[Key :Pointer]: Pointer           read prGetObject; default;

end;


implementation

// this is not as efficient as it could be -- keep MaxObjects small or fix this code


function TQUB_ObjectCache.prGetObject(Key :Pointer): Pointer;
var
  Ix           :Integer;
begin
  Ix           := prKeys.IndexOf(Key);
  if Ix < 0 then begin
    if prKeys.Count = prMax then begin
      prDestroy( prKeys[0], prObjects[0] );
      prObjects.Delete(0);
      prKeys.Delete(0);
    end;

    Result     := prCreate(Key);
    prKeys.Add(Key);
    prObjects.Add(Result);
  end else begin
    Result     := prObjects[Ix];

    if Ix < (prKeys.Count - 1) then begin
      prObjects.Delete(Ix);
      prKeys.Delete(Ix);
      prObjects.Add(Result);
      prKeys.Add(Key);
    end;
  end;
end;

constructor TQUB_ObjectCache.Create( CreateFn  :TQUB_ObjectCache_CreateFn;
                    DestroyFn :TQUB_ObjectCache_DestroyFn;
                    MaxObjects :Integer );
begin
  prCreate     := CreateFn;
  prDestroy    := DestroyFn;
  prMax        := MaxObjects;
  prKeys       := TList.Create;
  prObjects    := TList.Create;
end;

destructor TQUB_ObjectCache.Destroy;
var
  i            :Integer;
begin
  try
    for i := 0 to prObjects.Count - 1 do begin
      prDestroy( prKeys[i], prObjects[i] );
    end;
    prObjects.Free;
    prKeys.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_ObjectCache.FreeObject(Key :Pointer);
var
  Ix          :Integer;
begin
  Ix          := prKeys.IndexOf(Key);
  if Ix >= 0 then begin
    prDestroy( prKeys[Ix], prObjects[Ix] );
    prObjects.Delete(Ix);
    prKeys.Delete(Ix);
  end;
end;


// *****************************************************************************
end.
 
