{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Lorin's custom parser for reading strings and files


unit QUB_Parser;


{$define INDEXED}


interface


const


TOKEN_UNKNOWN    :Integer = 0;
TOKEN_NULL       :Integer = 1;
TOKEN_NAME       :Integer = 2;
TOKEN_NUMBER     :Integer = 3;
TOKEN_STRING     :Integer = 4;
TOKEN_SEPARATOR  :Integer = 5;
TOKEN_COMMENT    :Integer = 6;
TOKEN_TAB        :Integer = 7;
TOKEN_NEWLINE    :Integer = 8;
TOKEN_BAD_CHAR   :Integer = 9;
TOKEN_BAD_NUMBER :Integer = 10;


type


//forward class declarations
TQUB_Token  = class;
TQUB_Parser = class;


TQUB_ParserFormat = procedure(Token :TQUB_Token) of object;


//class declarations
TQUB_Token = class(TObject)
public

  Token           :String;
  TokenType       :Integer;
  TokenNumber     :Variant;
  TokenLen        :Integer;
  TokenPos        :Integer;
  TokenLine       :Integer;
  TokenInLine     :Integer;
  TokenNextLine   :Integer;
  TokenNextInLine :Integer;
  TextLen         :Integer;
  TextPos         :Integer;


  procedure Copy(Source :TQUB_Token);


end;


TQUB_Parser = class(TObject)
protected


  prNames        :String;
  prStrings      :String;
  prCommLine     :String;
  prCommBlkStart :String;
  prCommBlkEnd   :String;
  prToken        :TQUB_Token;
  prErr          :String;
  prFormat       :TQUB_ParserFormat;


  procedure prInit; virtual;
  function prGetError: String;
  function prCheckEnd :Boolean; virtual;
  function prCheckTab(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckNewLine(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckSpaces(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckStrings(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckCommLine(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckCommBlock(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckNumbers(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckSeparators(const Text :String; var i :Integer) :Boolean; virtual;
  function prCheckNames(const Text :String; var i :Integer) :Boolean; virtual;
  procedure prBadChar(const Text :String; var i :Integer); virtual;
  procedure prSetToken(const TokenType :Integer; const Token :String; const TokenNumber :Variant; const TokenPos :Integer;
                       const TokenLine :Integer; const TokenInLine :Integer; const TokenNextLine :Integer);


public


  ReturnComment :Boolean;
  ReturnTab     :Boolean;
  ReturnNewLine :Boolean;


  constructor Create; virtual;
  destructor Destroy; override;


  procedure Start; virtual;
  function Next(const Text :String) :Boolean; virtual;


  property Token  :TQUB_Token        read prToken;
  property Error  :String            read prGetError;
  property Format :TQUB_ParserFormat write prFormat;


end;


var


DecimalPoint :String;
ThousandSep  :String;
AllDigits    :String;
ByteCount    :Integer;


//******************************************************************************
implementation


uses


SysUtils, Windows;


//******************************************************************************
procedure TQUB_Token.Copy(Source :TQUB_Token);
begin
  TokenType       := Source.TokenType;
  Token           := Source.Token;
  TokenNumber     := Source.TokenNumber;
  TokenPos        := Source.TokenPos;
  TokenLen        := Source.TokenLen;
  TokenLine       := Source.TokenLine;
  TokenInLine     := Source.TokenInLine;
  TokenNextLine   := Source.TokenNextLine;
  TokenNextInLine := Source.TokenNextInLine;
  TextLen         := Source.TextLen;
  TextPos         := Source.TextPos;
end;

//******************************************************************************
constructor TQUB_Parser.Create;
begin
  inherited Create;
  prToken := TQUB_Token.Create;
  prInit;
end;

destructor TQUB_Parser.Destroy;
begin
  prToken.Free;
  inherited Destroy;
end;

procedure TQUB_Parser.prSetToken(const TokenType :Integer; const Token :String; const TokenNumber :Variant; const TokenPos :Integer;
                                 const TokenLine :Integer; const TokenInLine :Integer; const TokenNextLine :Integer);
begin
  prToken.TokenType := TokenType;
  prToken.Token := Token;
  prToken.TokenNumber := TokenNumber;
  prToken.TokenPos := TokenPos;
  prToken.TokenLen := Length(prToken.Token);
  if TokenLine <> - 1 then prToken.TokenLine := TokenLine;
  if TokenInLine <> - 1 then prToken.TokenInLine := TokenInLine;
  if TokenNextLine > 0 then begin
    prToken.TokenNextLine := TokenNextLine;
    prToken.TokenNextInLine := 1;
  end else if TokenNextLine < 0 then begin
    prToken.TokenNextInLine := prToken.TokenInLine + prToken.TokenLen;
  end;
  prToken.TextPos := prToken.TokenPos + prToken.TokenLen;
end;

function TQUB_Parser.prGetError: String;
begin
  Result := prErr;
  prErr := '';
end;

procedure TQUB_Parser.prInit;
begin
  prNames        := 'abcdefghijklmnopqrstuvwxyz_ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890';
  prStrings      := '"';
  prCommLine     := '//';
  prCommBlkStart := '/*';
  prCommBlkEnd   := '*/';
  ReturnComment  := False;
  ReturnTab      := False;
  ReturnNewLine  := False;
end;

procedure TQUB_Parser.Start;
begin
  prSetToken(TOKEN_UNKNOWN, '', varEmpty, 1, 1, 1, 1);
end;

function TQUB_Parser.prCheckEnd :Boolean;
begin
  Result := False;
  if (prToken.TextPos > prToken.TextLen) then begin
    prSetToken(TOKEN_NULL, '', varEmpty, prToken.TextPos, -1, -1, 1);
    Result := True;
  end;
end;

function TQUB_Parser.prCheckTab(const Text :String; var i :Integer) :Boolean;
begin
  Result := False;
  if Text[i] = #9 then begin
    prSetToken(TOKEN_TAB, #9, varEmpty, i, -1, -1, prToken.TokenLine + 1);
    Result := True;
  end;
end;

function TQUB_Parser.prCheckNewLine(const Text :String; var i :Integer) :Boolean;
begin
  Result := False;
  if ((Text[i] = #13) and (Text[i + 1] = #10)) then begin
    prSetToken(TOKEN_NEWLINE, #13#10, varEmpty, i, -1, -1, prToken.TokenLine + 1);
    Result := True;
  end;
end;

function TQUB_Parser.prCheckSpaces(const Text :String; var i :Integer) :Boolean;
begin
  Result := False;
  while (Text[i] = ' ') do begin
    Inc(prToken.TokenInLine);
    if (i = prToken.TextLen) then begin
      prSetToken(TOKEN_NULL, '', varEmpty, i + 1, -1, -1, 1);
      Result := True;
      Break;
    end;
    Inc(i);
  end;
end;

//To modify for embedded "" !!!
function TQUB_Parser.prCheckStrings(const Text :String; var i :Integer) :Boolean;
var
  p, j :Integer;
begin
  Result := False;
  if ((prStrings <> '') and (Text[i] = prStrings)) then begin
    p := pos(prStrings, copy(Text, i + 1, prToken.TextLen - (i + 1) + 1));
    if (p = 0) then begin
      p := prToken.TextLen + 1;
    end else begin
      p := p + i;
    end;
    for j := i + Length(prCommBlkStart) To p do begin
      if ((Text[j] = #13) and (Text[j + 1] = #10)) then begin
        Inc(prToken.TokenNextLine);
        prToken.TokenNextInLine := 0;
      end else begin
        Inc(prToken.TokenNextInLine);
      end;
    end;
    prSetToken(TOKEN_STRING, copy(Text, i, p + 1 - i), varEmpty, i, -1, -1, 0);
    Result := True;
  end;
end;

function TQUB_Parser.prCheckCommLine(const Text :String; var i :Integer) :Boolean;
var
  p :Integer;
begin
  Result := False;
  if ((prCommLine <> '') and (copy(Text, i, Length(prCommLine)) = prCommLine)) then begin
    p := pos(#13#10, copy(Text, i + Length(prCommLine), prToken.TextLen - (i + Length(prCommLine)) + 1));
    if (p = 0) then begin
      p := prToken.TextLen + 1;
    end else begin
      p := p + i + Length(prCommLine) - 1
    end;
    prSetToken(TOKEN_COMMENT, copy(Text, i, p - i), varEmpty, i, -1, -1, -1);
    Result := True;
  end;
end;

function TQUB_Parser.prCheckCommBlock(const Text :String; var i :Integer) :Boolean;
var
  p, j :Integer;
begin
  Result := False;
  if ((prCommBlkStart <> '') and (prCommBlkEnd <> '') and (copy(Text, i, Length(prCommBlkStart)) = prCommBlkStart)) then begin {Comm Block}
    p := pos(prCommBlkEnd, copy(Text, i + Length(prCommBlkStart), prToken.TextLen - (i + Length(prCommBlkStart)) + 1));
    if (p = 0) then begin
      p := prToken.TextLen + 1;
    end else begin
      p := p + i + Length(prCommBlkStart) - 1;
    end;
    for j := i + Length(prCommBlkStart) To p do begin
      if ((Text[j] = #13) and (Text[j + 1] = #10)) then begin
        Inc(prToken.TokenNextLine);
        prToken.TokenNextInLine := 0;
      end else begin
        Inc(prToken.TokenNextInLine);
      end;
    end;
    prSetToken(TOKEN_COMMENT, copy(Text, i, p + Length(prCommBlkEnd) - i), varEmpty, i, -1, -1, 0);
    Result := True;
  end;
end;

function TQUB_Parser.prCheckNumbers(const Text :String; var i :Integer) :Boolean;
var
  v :Extended;
  c :Integer;
begin
  Result := False;
  if pos(Text[i], AllDigits) <> 0 then begin
    val(copy(Text, i, 100) + 'a', v, c);
    prToken.TokenNumber := v;
    prToken.Token := copy(Text, i, c - 1);
    if c < 2 then Exit;
    try
      val(prToken.Token, v, c);
      prSetToken(TOKEN_NUMBER, prToken.Token, v, i, -1, -1, -1);
    except
      prSetToken(TOKEN_BAD_NUMBER, prToken.Token, varEmpty, i, -1, -1, -1);
      prErr := 'Invalid number';
    end;
    Result := True;
  end;
end;

function TQUB_Parser.prCheckSeparators(const Text :String; var i :Integer) :Boolean;
var
  t :String;
begin
  Result := False;
  case Text[i] of
    '{', '}', ';', ',', '(', ')', '+', '-', '*', '/', '^', '=', '''', '.', ':', '[', ']':
      begin
        t := Text[i];
        Result := True;
      end;
    '<':
      begin
        if ((Text[i + 1] = '=') or (Text[i + 1] = '>')) then t := Text[i] + Text[i + 1] else t := Text[i];
        Result := True;
      end;
    '>':
      begin
        if Text[i + 1] = '=' then t := Text[i] + Text[i + 1] else t := Text[i];
        Result := True;
      end;
  end;
  if Result then  prSetToken(TOKEN_SEPARATOR, t, varEmpty, i, -1, -1, -1);
end;

function TQUB_Parser.prCheckNames(const Text :String; var i :Integer) :Boolean;
var
  p, f, j :Integer;
  n       :Boolean;
begin
  Result := False;
  n := False;
  p := 0;
  f := i;
  repeat
    if ((pos(Text[f], prNames) <> 0) and (Text[f] <> '_')) then begin
      n := True;
      p := 0;
      for j := f + 1 To prToken.TextLen do begin
        if (pos(Text[j], prNames) = 0) then begin
          p := j;
          Break;
        end;
      end;
      if (p = 0) then begin
        p := prToken.TextLen + 1;
      end;
      f := p;
      if Text[f] <> '.' then begin
        Break;
      end else begin
        Inc(f);
      end;
    end else begin
      Break;
    end;
  until(False);
  if n then begin
    prSetToken(TOKEN_NAME, copy(Text, i, p - i), varEmpty, i, -1, -1, -1);
    Result := True;
  end;
end;

procedure TQUB_Parser.prBadChar(const Text :String; var i :Integer);
begin
  prSetToken(TOKEN_BAD_CHAR, Text[i], varEmpty, i, -1, -1, -1);
  prErr := 'Invalid character';
end;

function TQUB_Parser.Next(const Text :String) :Boolean;
var
  i :Integer;
begin
  Result := False;
  prToken.TextLen := Length(Text);
  repeat
    if prCheckEnd then Exit;
    prToken.TokenLine := prToken.TokenNextLine;
    prToken.TokenInLine := prToken.TokenNextInLine;
    i := prToken.TextPos;
    repeat
      if prCheckSpaces(Text, i) then begin
        Break;
      end else if prCheckTab(Text, i) then begin
        Break;
      end else if prCheckNewLine(Text, i) then begin
        Break;
      end else if prCheckStrings(Text, i) then begin
        Break;
      end else if prCheckCommLine(Text, i) then begin
        Break;
      end else if prCheckCommBlock(Text, i) then begin
        Break;
      end else if prCheckNumbers(Text, i) then begin
        Break;
      end else if prCheckSeparators(Text, i) then begin
        Break;
      end else if prCheckNames(Text, i) then begin
        Break;
      end else begin
        prBadChar(Text, i);
        Break;
      end;
    until(False);
    if Assigned(prFormat) then prFormat(Token);
    if ((Token.TokenType = TOKEN_BAD_NUMBER) or (Token.TokenType = TOKEN_BAD_CHAR)) then begin
      Result := True;
      Break;
    end else if ((Token.TokenType = TOKEN_COMMENT) and (ReturnComment = True)) then begin
      Break;
    end else if ((Token.TokenType = TOKEN_TAB) and (ReturnTab = True)) then begin
      Break;
    end else if ((Token.TokenType = TOKEN_NEWLINE) and (ReturnNewLine = True)) then begin
      Break;
    end else if ((Token.TokenType <> TOKEN_TAB) and (Token.TokenType <> TOKEN_NEWLINE) and (Token.TokenType <> TOKEN_COMMENT)) then begin
      Break;
    end;
  until(False);
end;


initialization


//Find the decimal separator
begin

  DecimalPoint := '.';
  {
  if GetLocaleInfo(LOCALE_SYSTEM_DEFAULT, LOCALE_SDECIMAL, PChar(DecimalPoint), Length(DecimalPoint)) <> 0 then begin
    DecimalPoint := '.';
  end;
  }
  ThousandSep := ',';
  {
  if GetLocaleInfo(LOCALE_SYSTEM_DEFAULT, LOCALE_STHOUSAND, PChar(ThousandSep), Length(ThousandSep)) <> 0 then begin
    ThousandSep := ',';
  end;
  }
  AllDigits := '0123456789' + DecimalPoint + ThousandSep;

end;

//******************************************************************************
end.

