{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Numerical Recipes


unit QUB_Powell;


interface


uses


Classes, Dialogs,
Math,
AllocTypes, Alloc;


const


POWELL_TINY   = 1.0e-25;                                                        //A small number.
POWELL_ITMAX  = 200;                                                            //Maximum allowed iterations.
LINMIN_TOL    = 2.0e-4;                                                         //Tolerance passed to brent.
MNBRAK_GOLD   = 1.618034;                                                       //Default ratio by which successive intervals are magnified
MNBRAK_GLIMIT = 100.0;                                                          //Maximum magnification allowed for a parabolic-fit step
MNBRAK_TINY   = 1.0e-20;
BRENT_ITMAX   = 100;                                                            //Maximum allowed number of iterations;
BRENT_CGOLD   = 0.3819660;                                                      //Golden ratio;
BRENT_ZEPS    = 1.0e-10;                                                        //Small number that protects against trying to achieve
                                                                                //fractional accuracy for a minimum that happens to be exactly zero.

type


//TQUB_Optimize_Func      = function(p :TPADouble) :Double of object;
TQUB_Optimize_Func      = function(Caller :TObject; p :TPADouble) :Double; stdcall;
TQUB_Optimize_Func1Dim  = function(x :Double) :Double of object;


TQUB_Optimize_Powell    = class(TObject)
protected


  prCaller               :TObject;

  prStopped              :Boolean;
  prStop                 :Boolean;
  prOnChange             :TNotifyEvent;
  prError                :Boolean;
  prErrorMsg             :String;

  prNcom                 :Integer;
  prPcom                 :TPADouble;
  prXicom                :TPADouble;
  prNrfunc               :TQUB_Optimize_Func;


  procedure prPowell(p :TPADouble; xi :TPMDouble; n :Integer; ftol :Double; var iter :Integer; var fret :Double; func :TQUB_Optimize_Func);
  function prBrent(ax, bx, cx :Double; f :TQUB_Optimize_Func1Dim; tol :Double; var xmin :Double) :Double;
  procedure prMnbrak(var ax, bx, cx, fa, fb, fc :Double; func :TQUB_Optimize_Func1Dim);
  procedure prLinmin(p :TPADouble; xi :TPADouble; n :Integer; var fret :Double; func :TQUB_Optimize_Func);
  function prF1dim(x :Double) :Double;
  function prSign(a, b :Double) :Double;


public


  constructor Create;
  destructor Destroy; override;

  function Start(Caller :TObject; p :TPADouble; n :Integer; ftol :Double; maxiter :Integer; var iter :Integer; var fret :Double; func :TQUB_Optimize_Func) :Boolean;
  procedure Stop;

  property Stopped       :Boolean        read prStopped;
  property Error         :Boolean        read prError;
  property ErrorMsg      :String         read prErrorMsg;

  property OnChange      :TNotifyEvent   read prOnChange       write prOnChange;


end;


implementation


//******************************************************************************
constructor TQUB_Optimize_Powell.Create;
begin
  inherited;

  prStopped := True;

end;

destructor TQUB_Optimize_Powell.Destroy;
begin
  try

    Stop;

    while not Stopped do;

  finally

    inherited;

  end;
end;

function TQUB_Optimize_Powell.Start(Caller :TObject; p :TPADouble; n :Integer; ftol :Double; maxiter :Integer; var iter :Integer; var fret :Double; func :TQUB_Optimize_Func) :Boolean;
var
  xi         :TPMDouble;
  i, j       :Integer;
begin
  Result := not prStopped;
  if Result then Exit;

  prStopped  := False;
  prError    := False;
  prErrorMsg := '';

  prCaller := Caller;

  xi := nil;

  try

    xi := DAlloc2(n + 1, n + 1);
    for i := 0 to n - 1 do begin
      for j := 0 to n - 1 do begin
        xi[i, j] := 1;//random;
      end;
    end;

    prPowell(p, xi, n, ftol, iter, fret, func);

  finally

    Free2(TPMChar(xi));
    prStopped := True;

  end;

end;

procedure TQUB_Optimize_Powell.Stop;
begin
  prStop := True;
end;


function TQUB_Optimize_Powell.prBrent(ax, bx, cx :Double; f :TQUB_Optimize_Func1Dim; tol :Double; var xmin :Double) :Double;
(*
Given a function f, and given a bracketing triplet of abscissas ax, bx, cx (such that bx is
between ax and cx, and f(bx) is less than both f(ax) and f(cx)), this routine isolates
the minimum to a fractional precision of about tol using Brent's method. The abscissa of
the minimum is returned as xmin, and the minimum function value is returned as brent, the
returned function value.
*)
var
  a, b, d,
  etemp,
  fu, fv, fw, fx,
  p, q, r,
  tol1, tol2,
  u, v, w,
  x, xm          :Double;
  e              :Double;//This will be the distance moved on the step before last.
  iter           :Integer;

begin

  e := 0.0;

  d := 0;//d is not initialized in the original C routine

  if ax < cx then a := ax else a := cx;
  //a and b must be in ascending order, but input abscissas need not be.
  if ax > cx then b := ax else b := cx;

  //Initializations...
  v := bx;
  w := bx;
  x := bx;

  fx := f(x);
  fv := fx;
  fw := fx;

  for iter := 1 to BRENT_ITMAX do begin
  //Main program loop.

    xm := 0.5 * (a + b);
    tol1 := tol * abs(x) + BRENT_ZEPS;
    tol2 := 2.0 * tol1;

    if abs(x - xm) <= (tol2 - 0.5 * (b - a)) then begin
    //Test for done here.

      xmin := x;
      Result := fx;
      Exit;

    end;

    if abs(e) > tol1 then begin
    //Construct a trial parabolic fit.

      r := (x - w) * (fx - fv);
      q := (x - v) * (fx - fw);
      p := (x - v) * q - (x - w) * r;
      q := 2.0 * (q - r);

      if q > 0.0 then p := -p;
      q := abs(q);

      etemp := e;
      e := d;

      if (abs(p) >= abs(0.5 * q * etemp)) or (p <= (q * (a - x))) or (p >= (q * (b - x))) then begin

        if x >= xm then e := a - x else e := b - x;
        d := BRENT_CGOLD * e;

        //The above conditions determine the acceptability of the parabolic fit.
        //Here we take the golden section step into the larger of the two segments.

      end else begin

        d := p / q;
        //Take the parabolic step.

        u := x + d;

        if ((u - a) < tol2) or ((b-u) < tol2) then begin
          d := prSign(tol1, xm - x);
        end;

      end;

    end else begin

      if x >= xm then e := a - x else e := b - x;
      d := BRENT_CGOLD * e;

    end;

    if abs(d) >= tol1 then u := x + d else u := x + prSign(tol1, d);
    fu := f(u);
    //This is the one function evaluation per iteration.

    if fu <= fx then begin
    //Now decide what to do with our function evaluation.

      if u >= x then a := x else b := x;

      //Housekeeping follows:
      v := w;
      w := x;
      x := u;

      fv := fw;
      fw := fx;
      fx := fu;

    end else begin

      if u < x then a := u else b := u;

      if (fu <= fw) or (w = x) then begin

        v := w;
        w := u;
        fv := fw;
        fw := fu;

      end else if (fu <= fv) or (v = x) or (v = w) then begin

        v := u;
        fv := fu;

      end;

    end;
    //Done with housekeeping.

    //Back for another iteration.

  end;

  prError    := True;
  prErrorMsg := 'Too many iterations in brent';

  xmin := x;//Never get here.

  Result := fx;

end;

procedure TQUB_Optimize_Powell.prMnbrak(var ax, bx, cx, fa, fb, fc :Double; func :TQUB_Optimize_Func1Dim);
(*
Given a function func, and given distinct initial points ax and bx, this routine searches in
the downhill direction (defined by the function as evaluated at the initial points) and returns
new points ax, bx, cx that bracket a minimum of the function. Also returned are the function
values at the three points, fa, fb, and fc.
*)
var
  ulim, u, r, q, fu, dum :Double;
begin

  fa := func(ax);
  fb := func(bx);

  if fb > fa then begin
  //Switch roles of a and b so that we can go downhill in the direction from a to b.

    dum := ax;
    ax  := bx;
    bx  := dum;

    dum := fb;
    fb  := fa;
    fa  := dum;

  end;

  cx := bx + MNBRAK_GOLD *(bx - ax);
  //First guess for c.

  fc := func(cx);

  while (fb > fc) do begin
  //Keep returning here until we bracket.

    r := (bx - ax) * (fb - fc);
    //Compute u by parabolic extrapolation from a; b; c.
    //TINY is used to prevent any possible division by zero.

    q := (bx - cx) * (fb - fa);

    u := bx - ((bx - cx) * q - (bx - ax)*r) / (2.0 * prSign(Max(Abs(q - r), MNBRAK_TINY), q - r));

    ulim := bx + MNBRAK_GLIMIT * (cx - bx);
    //We won't go farther than this. Test various possibilities:

    if ((bx - u) * (u - cx)) > 0.0 then begin
    //Parabolic u is between b and c: try it.

      fu := func(u);

      if fu < fc then begin
      //Got a minimum between b and c.

        ax := bx;
        bx := u;
        fa := fb;
        fb := fu;

        Exit;

      end else if fu > fb then begin
      //Got a minimum between between a and u.

        cx := u;
        fc := fu;

        Exit;

      end;

      u := cx + MNBRAK_GOLD * (cx - bx);
      //Parabolic fit was no use. Use default magnification.

      fu := func(u);

    end else if ((cx - u) * (u - ulim)) > 0.0 then begin
    //Parabolic fit is between c and its allowed limit.

      fu := func(u);

      if fu < fc then begin

        bx := cx;
        cx := u;
        u  := cx + MNBRAK_GOLD * (cx - bx);

        fb := fc;
        fc := fu;
        fu := func(u);

      end;

    end else if ((u - ulim) * (ulim - cx)) >= 0.0 then begin
    //Limit parabolic u to maximum allowed value.

      u  := ulim;
      fu := func(u);

    end else begin
    //Reject parabolic u, use default magnification.

      u  := cx + MNBRAK_GOLD * (cx - bx);
      fu := func(u);

    end;

    ax := bx;
    bx := cx;
    cx := u;
    //Eliminate oldest point and continue.

    fa := fb;
    fb := fc;
    fc := fu;

  end;

end;

function TQUB_Optimize_Powell.prF1dim(x :Double) :Double;
//Must accompany linmin.
var
  xt :TPADouble;
  j  :Integer;
begin

  xt := DAlloc1(prNcom + 1);

  try

    for j := 0 to prNcom - 1 do begin

      xt[j] := prPcom[j] + x * prXicom[j];

    end;

    Result := prNrfunc(prCaller, xt);

  finally

    FreeMem(xt);

  end;

end;

procedure TQUB_Optimize_Powell.prLinmin(p :TPADouble; xi :TPADouble; n :Integer; var fret :Double; func :TQUB_Optimize_Func);
(*
Given an n-dimensional point p[1..n] and an n-dimensional direction xi[1..n], moves and
resets p to where the function func(p) takes on a minimum along the direction xi from p,
and replaces xi by the actual vector displacement that p was moved. Also returns as fret
the value of func at the returned location p. This is actually all accomplished by calling the
routines mnbrak and brent.
*)
var
  xx, xmin, fx, fb, fa, bx, ax :Double;
  j                            :Integer;
begin

  prNcom := n;                                                                  //Define the global variables.

  prPcom  := DAlloc1(n + 1);
  prXicom := DAlloc1(n + 1);
  prNrfunc := func;

  try


    for j := 0 to n - 1 do begin

      prPcom[j]  := p[j];
      prXicom[j] := xi[j];

    end;

    ax := 0.0;                                                                  //Initial guess for brackets.
    xx := 1.0;

    prMnbrak(ax, xx, bx, fa, fx, fb, prF1dim);

    fret := prBrent(ax, xx, bx, prF1dim, LINMIN_TOL, xmin);

    for j := 0 to n - 1 do begin                                                    //Construct the vector results to return.

      xi[j] := xi[j] * xmin;
      p[j]  := p[j]  + xi[j];

    end;

  finally

    FreeMem(prXicom);
    FreeMem(prPcom);

  end;

end;

procedure TQUB_Optimize_Powell.prPowell(p :TPADouble; xi :TPMDouble; n :Integer; ftol :Double; var iter :Integer; var fret :Double; func :TQUB_Optimize_Func);
(*
Minimization of a function func of n variables. Input consists of an initial starting point
p[1..n]; an initial matrix xi[1..n][1..n], whose columns contain the initial set of di-
rections (usually the n unit vectors); and ftol, the fractional tolerance in the function value
such that failure to decrease by more than this amount on one iteration signals doneness. On
output, p is set to the best point found, xi is the then-current direction set, fret is the returned
function value at p, and iter is the number of iterations taken. The routine linmin is used.
*)
var
  i, j, ibig       :Integer;
  del, fp, fptt, t :Double;
  pt, ptt, xit     :TPADouble;
  iteration        :Integer;
begin

  pt  := DAlloc1(n + 1);
  ptt := DAlloc1(n + 1);
  xit := DAlloc1(n + 1);

  iteration := 0;

  try

    fret := func(prCaller, p);

    for j := 0 to n - 1 do begin
      pt[j] := p[j];// Save the initial point.
    end;

    for iteration := 1 to POWELL_ITMAX do begin

      //Exit if stopped by user or Error
      if (Stopped) or (prError) then Break;

      fp    := fret;
      ibig  := 0;

      del   := 0.0;                                                             //Will be the biggest function decrease.

      for i := 0 to n - 1 do begin                                                  //In each iteration, loop over all directions in the set.

        for j := 0 to n - 1 do xit[j] := xi[j, i];                                  //Copy the direction,

        fptt := fret;

        prLinmin(p, xit, n, fret, func);                                        //minimize along it,

        if (fptt - fret) > del then begin                                       //and record it if it is the largest decrease so far.

          del := fptt - fret;
          ibig := i;

        end;

      end;

      if Abs{not in the original}(2.0 * (fp - fret)) <= (ftol * (abs(fp) + abs(fret)) + POWELL_TINY) then begin

        Exit;

      end;

      //Construct the extrapolated point and the average direction moved. Save the old starting point.
      for j := 0 to n - 1 do begin

        ptt[j] := 2.0 * p[j] - pt[j];
        xit[j] := p[j] - pt[j];
        pt[j]  := p[j];

      end;

      fptt := func(prCaller, ptt);                                                        //Function value at extrapolated point.

      if fptt < fp then begin

        t := 2.0 * (fp - 2.0 * fret + fptt) * SQR(fp - fret - del) - del * SQR(fp - fptt);

        if t < 0.0 then begin

          prLinmin(p, xit, n, fret, func);                                      //Move to the minimum of the new direction, and save the new direction.

          for j := 0 to n - 1 do begin

            xi[j, ibig] := xi[j, n];
            xi[j, n]    := xit[j];

          end;

        end;

      end;

    end;//iter

    if not Stopped then begin
      prError := True;
      prErrorMsg := 'Powell exceeding maximum iterations';
    end;

  finally

    iter := iteration;

    FreeMem(pt);
    FreeMem(ptt);
    FreeMem(xit);

  end;

end;

function TQUB_Optimize_Powell.prSign(a, b :Double) :Double;
begin
  if b >= 0.0 then Result := abs(a) else Result := - abs(a);
end;


//******************************************************************************
end.


