unit QUB_Profiles;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Sep 2005 Chris
//
// Preferences management
// Centralizes the disk access.
// To play along:
//     maintain your current prefs in a QUB_NODE
//     come up with a category Name
//     init: update my prefs from ProfileMgr[Name].Active
//           ProfileMgr[Name].Active := my prefs tree
//           ProfileMgr[Name].OnSet := something that updates my prefs tree
//     don't forget to unset OnSet before you are destroyed
//     qub will take care of saving and restoring Active across sessions
//     for multiple documents sharing a category:
//           only the front one should be active
//           (owning Active and OnSet)
//           the init procedure above will give new docs the same profile as the last viewed doc
//     when showing doc properties, switch active and onset over to the dialog
//     the gui components provided handle loading,saving,switching
// QUB_MAIN instantiates the Mgr and stores it in QUB_GUIInterface



interface

uses

Sysutils, Classes, Windows, Controls, StdCtrls, Menus, Dialogs, FileCtrl, Forms,

QUB_Global, QUB_Node


;



type


TQUB_ProfileMgr = class;
TQUB_ProfileCat = class;

TQUB_Profile_OnSet = procedure (Sender :TQUB_ProfileCat; Updates :INode) of object;

//******************************************************************************

TQUB_ProfileCat = class
protected
  prMgr         :TQUB_ProfileMgr;
  prName        :String;
  prActive      :INode;


public
  OnSet         :TQUB_Profile_OnSet;

  constructor Create(Mgr :TQUB_ProfileMgr; const Name :String);
  destructor Destroy; override;

  procedure SetProperties(Updates :INode);

  function ListProfiles(Results :TStrings): Boolean;
  function SaveProfile(const Nm :String): Boolean;
  function LoadProfile(const Nm :String): Boolean;
  function ExportProfile(const Path :String): Boolean;
  function ImportProfile(const Path :String): Boolean;

  function SaveActive: Boolean;

  property Name        :String        read prName;
  property Active      :INode         read prActive        write prActive;

end;


//******************************************************************************

TQUB_ProfileMgr = class
protected
  prSystemPath         :String;
  prUserPath           :String;

  prCats               :TStringList;

  function prGetCategory(Nm :String): TQUB_ProfileCat;

public

  constructor Create(const SystemPath, UserPath :String);
  destructor Destroy; override;

  procedure SaveAll;

  function ListProfiles(const Cat :String; Results :TStrings): Boolean;
  function SaveProfile(Props :INode; const Cat, Nm :String): Boolean;
  function LoadProfile(const Cat, Nm :String): INode;

  property Category[Nm :String]: TQUB_ProfileCat    read prGetCategory; default;

end;

//******************************************************************************

TQUB_ProfileReadBox = class(TCustomComboBox)
protected
  prCat        :String;
  prProperties :INode;

  procedure prSetCat(CatName :String);

  procedure prOnChange(Sender :TObject);

public
  OnChangeProfile  :TNotifyEvent;

  constructor Create(AOwner :TComponent); override;
  procedure Refresh;
  
  property Cat        :String        read prCat        write prSetCat;
  property Properties :INode         read prProperties;
end;



//******************************************************************************

TQUB_ProfileEditBox = class(TCustomComboBox)
protected
  prCatName    :String;
  prCat        :TQUB_ProfileCat;

  procedure prSetCat(CatName :String);

  procedure prOnDropDown(Sender :TObject);
  procedure prOnChange(Sender :TObject);

public
  constructor Create(AOwner :TComponent); override;

  property Cat        :String        read prCatName        write prSetCat;
end;




function RetardedBooleanCast(i :Integer): Boolean;



var ProfileMgr :TQUB_ProfileMgr;


//******************************************************************************


implementation



function RetardedBooleanCast(i :Integer): Boolean;
begin
  Result := i <> 0;
end;


//******************************************************************************

constructor TQUB_ProfileCat.Create(Mgr :TQUB_ProfileMgr; const Name :String);
begin
  prMgr         := Mgr;
  prName        := Name;
  prActive      := Mgr.LoadProfile(Name, '__active');
  if prActive.IsNull then prActive := TNode.Create('Properties');
end;

destructor TQUB_ProfileCat.Destroy;
begin
  inherited;
end;

procedure TQUB_ProfileCat.SetProperties(Updates :INode);
begin
  if Assigned(OnSet) then begin
    OnSet(Self, Updates);
  end;
end;

function TQUB_ProfileCat.ListProfiles(Results :TStrings): Boolean;
begin
  Result := prMgr.ListProfiles(Name, Results);
end;

function TQUB_ProfileCat.SaveProfile(const Nm :String): Boolean;
begin
  Result := prMgr.SaveProfile(Active, Name, Nm);
end;

function TQUB_ProfileCat.LoadProfile(const Nm :String): Boolean;
var
  Prof   :INode;
begin
  Prof   := prMgr.LoadProfile(Name, Nm);
  Result := Prof.IsNull;
  if not Result then SetProperties(Prof);
end;

function TQUB_ProfileCat.ExportProfile(const Path :String): Boolean;
begin
  Result := not Active.Clone(True).SaveAs(Path);
end;

function TQUB_ProfileCat.ImportProfile(const Path :String): Boolean;
var
  Prof   :INode;
begin
  Prof   := TNode.OpenReadOnly(Path);
  Prof.Close();
  Result := Prof.IsNull;
  if not Result then SetProperties(Prof);
end;

function TQUB_ProfileCat.SaveActive: Boolean;
begin
  Result := prMgr.SaveProfile(Active, Name, '__active');
end;


//******************************************************************************

constructor TQUB_ProfileMgr.Create(const SystemPath, UserPath :String);
var
  Fi                 :TextFile;
begin
  prSystemPath       := SystemPath;
  prUserPath         := UserPath;
  prCats             := TStringList.Create;
  prCats.Sorted      := True;
  prCats.Duplicates  := dupIgnore;
  try
    if FileExists(ExtractFilePath(Application.ExeName)+'profile_log.txt') then begin
      AssignFile(Fi, ExtractFilePath(Application.ExeName)+'profile_log.txt');
      Erase(Fi);
    end;
  except
  end;
end;

destructor TQUB_ProfileMgr.Destroy;
var
  Ix                 :Integer;
begin
  try
    for Ix := 0 to prCats.Count - 1 do begin
      TQUB_ProfileCat(prCats.Objects[Ix]).Free;
    end;
    prCats.Free;
  finally
    inherited;
  end;
end;

function TQUB_ProfileMgr.prGetCategory(Nm :String): TQUB_ProfileCat;
var
  Ix         :Integer;
begin
  Ix         := prCats.IndexOf(Nm);
  if Ix < 0 then begin
    Ix       := prCats.AddObject(Nm, TQUB_ProfileCat.Create(Self, Nm));
  end;
  Result     := TQUB_ProfileCat(prCats.Objects[Ix]);
end;

procedure TQUB_ProfileMgr.SaveAll;
var
  Ix         :Integer;
begin
  for Ix := 0 to prCats.Count - 1 do begin
    TQUB_ProfileCat(prCats.Objects[Ix]).SaveActive;
  end;
end;

function TQUB_ProfileMgr.ListProfiles(const Cat :String; Results :TStrings): Boolean;
var
  Sorted     :TStringList;

  procedure ListDir(const Path :String);
  var
    F          :TSearchRec;
  begin
    if 0 = FindFirst(Path+'\*.qpr', 0, F) then begin
      repeat
        if Copy(F.Name,0,2) <> '__'
        then Sorted.Add(ChangeFileExt(F.Name, ''));
      until 0 <> FindNext(F);
    end;
    Sysutils.FindClose(F);
  end;

begin
  Sorted     := TStringList.Create;
  Sorted.Sorted := True;
  Sorted.Duplicates := dupIgnore;

  ListDir(prUserPath+Cat);
  ListDir(prSystemPath+Cat);
  Results.Assign(Sorted);
  Sorted.Free;
  Result     := False;
end;

function TQUB_ProfileMgr.SaveProfile(Props :INode; const Cat, Nm :String): Boolean;
begin
  if not DirectoryExists(prUserPath+Cat+'\') then CreateDir(prUserPath+Cat+'\');
  Result := not Props.Clone(True).SaveAs(prUserPath+Cat+'\'+Nm+'.qpr');
end;

function TQUB_ProfileMgr.LoadProfile(const Cat, Nm :String): INode;
var
  Path  :String;
  Fi    :TextFile;
begin
  Result := TNode.OpenReadOnly(prUserPath+Cat+'\'+Nm+'.qpr');
  if Result.IsNull then Result := TNode.OpenReadOnly(prSystemPath+Cat+'\'+Nm+'.qpr');
  Result.Close();

  try
    try
      Path := ExtractFilePath(Application.ExeName)+'profile_log.txt';
      AssignFile(Fi, Path);
      if FileExists(Path) then begin
        try
          Append(Fi);
        except
          ReWrite(Fi);
        end;
      end else begin
        ReWrite(Fi);
      end;
      Write(Fi, Format('%s: %s'#13#10, [Cat, Nm]));
      Write(Fi, Result.AsString);
      Write(Fi, #13#10);
    finally
      CloseFile(Fi);
    end;
  except
  end;
end;


//******************************************************************************

constructor TQUB_ProfileReadBox.Create(AOwner :TComponent);
begin
  inherited;

  Style        := csDropDownList;
  OnChange     := prOnChange;
end;

procedure TQUB_ProfileReadBox.Refresh;
begin
  ProfileMgr[Cat].ListProfiles(Items);
  Items.Insert(0, '<none>');
  ItemIndex    := 0;
  prProperties := TNode.Create('');
end;

procedure TQUB_ProfileReadBox.prSetCat(CatName :String);
begin
  prCat        := CatName;
  Refresh;
end;

procedure TQUB_ProfileReadBox.prOnChange(Sender :TObject);
begin
  if Text = '<none>' then begin
    prProperties := TNode.Create('');
  end else begin
    prProperties := ProfileMgr.LoadProfile(Cat, Text);
  end;
  if Assigned(OnChangeProfile) then OnChangeProfile(Self);
end;



//******************************************************************************

constructor TQUB_ProfileEditBox.Create(AOwner :TComponent);
begin
  inherited;

  prCat        := nil;
  Style        := csDropDownList;
  OnDropDown   := prOnDropDown;
  OnChange     := prOnChange;
end;

procedure TQUB_ProfileEditBox.prSetCat(CatName :String);
begin
  prCatName    := CatName;
  prCat        := ProfileMgr[CatName];
  Items.Add('Presets...');
  ItemIndex    := 0;
end;

procedure TQUB_ProfileEditBox.prOnDropDown(Sender :TObject);
begin
  ProfileMgr[Cat].ListProfiles(Items);
  Items.Insert(0, '* Import...');
  Items.Insert(1, '* Export...');
  Items.Insert(2, '* Add to menu...');
  Items.Insert(3, '* Help...');
end;


procedure TQUB_ProfileEditBox.prOnChange(Sender :TObject);
var
  Name         :String;
  ODlg         :TOpenDialog;
  SDlg         :TSaveDialog;
begin
  if Text = '* Import...' then begin
    ODlg := TOpenDialog.Create(Self);
    ODlg.Filter := 'QUB presets (*.qpr)|*.qpr|' +
                   'All files (*.*)|*.*';
    ODlg.Options := [ofPathMustExist, ofFileMustExist];
    if ODlg.Execute then begin
      if prCat.ImportProfile(ODlg.Files[0])
      then MessageHelpDlg('Can''t open '+ODlg.Files[0], mtError, [mbOK], 0, 'Error:Open_File');
    end;
    ODlg.Free;
  end else if Text = '* Export...' then begin
    SDlg := TSaveDialog.Create(Self);
    SDlg.Filter := 'QUB presets (*.qpr)|*.qpr|' +
                   'All files (*.*)|*.*';

    SDlg.Options := [ofPathMustExist{, ofOverwritePrompt}];
    SDlg.Title := 'Save presets as';
    SDlg.FileName := prCatName+' Presets';
    if SDlg.Execute then begin
      if prCat.ExportProfile(ChangeFileExt(SDlg.FileName, '.qpr'))
      then MessageDlg('Can''t save '+SDlg.Filename, mtError, [mbOK], 0);
    end;
  end else if Text = '* Add to menu...' then begin
    Name := 'New Presets';
    if InputQuery('Add to menu...', 'Presets name:', Name)
    then if prCat.SaveProfile(Name)
         then MessageHelpDlg('Can''t save presets '+Name, mtError, [mbOK], 0, 'Error:Save_Presets');
  end else if Text = '* Help...' then begin
    ShowWiki('Interface_Theory#Presets_and_Defaults');
  end else begin
    prCat.LoadProfile(Text);
  end;

  Items.Clear;
  Items.Add('Presets...');
  ItemIndex    := 0;
end;




//******************************************************************************

end.
