{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Properties;

// Allows you to connect various ui components to nodes in a qubtree.
// Nodes are referenced by path (e.g. '["a"]["b"]') within .Properties,
// and Properties may be replaced at any time.

// QUB_Task_Properties and QUB_QVO_Properties use this framework to work with QUB_Profiles.
// The connection is done using:
//   SetProperties(INode)      -- copies common items into .Properties
//   prUpdateComponents        -- copies values from .Properties into ui components

// see QUB_Task_Data_Properties for a basic example
// and QUB_Task_Mil_Properties for an example of not all variables managed


interface


uses

Classes, Controls, StdCtrls, Windows, Messages, SysUtils, Graphics, Forms, Dialogs,
ComCtrls, ExtCtrls, Registry,

{$IFDEF DELPHI6_OR_HIGHER}
Variants,
{$ENDIF}

PythonEngine,
QUB_Node, RegistryInterface;


type


TQUB_Property_RegType = (REGTYPE_NONE, REGTYPE_BOOL, REGTYPE_INT, REGTYPE_FLOAT, REGTYPE_STRINGFLOAT, REGTYPE_STRING);


// Abstract base for objects which tie together specific gui component(s)
// with a node in the properties tree (identified by its path)
// and its default value and (obsolete) registry location

// Clients: you create one of these per property you want managed
//          and pass it off to QUB_PropertiesForm, which handles the rest including
//           - reading the registry or using the defaults, as needed
//           - displaying changes to the value (e.g. changing profile)
//           - reading and validating user input

TQUB_Property = class
protected
  prPath           :String;
  prDefVal         :Variant;
  prRegName        :String;
  prRegType        :TQUB_Property_RegType;
  prIx             :Integer;
  prEnsuredDim     :Boolean;


  function prGetNode(Root :INode): INode;
  // to be called only when holding python GIL

  procedure prSetIx(x :Integer);
  procedure prEnsureDim(Node :INode);

public
  constructor CreateBase(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant); overload;
  constructor CreateBase(const Path :String); overload;

  procedure Init(Root :INode; var Reg :TRegistry; const RegBase :String); virtual; abstract;
  procedure Display(Root :INode); virtual; abstract;
  procedure Accept(Root :INode); virtual; abstract;
  // accept: copy displayed info to node
  //         if err, copy node to display and (re)raise exception

  property Path    :String                read prPath;
  property DefVal  :Variant               read prDefVal;
  property RegName :String                read prRegName;
  property RegType :TQUB_Property_RegType read prRegType;

  property Ix      :Integer               read prIx          write prSetIx;
  // def 0; see TQUB_Property_Indexer below
end;


// Acceptor: a function which converts user-input strings to the desired value type
// possibly raising an exception e.g. bad format, out of bounds
// There are some acceptors below e.g. AcceptIntBetween(0, 10)
// or you can write your own as python expressions which evaluate to function of string

// Format: tells how to write the node's data in the TEdit
//         either something printf-style like '%.6g'
//         or a python expression evaluating to function of node returning string (it must begin with "lambda")
//         or '' (empty string) to just use the default string representation

TQUB_Property_Edit = class(TQUB_Property)
protected
  prEdit           :TCustomEdit;
  prAcceptor       :PPyObject;
  prFormat         :PPyObject;

  prOldOnKeyPress  :TKeyPressEvent;
  prOldOnExit      :TNotifyEvent;

  procedure prCreate(Edit :TCustomEdit; const Acceptor, Format :String);
  procedure prDisplay(Node :PPyObject);

public
  constructor Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                     Edit :TCustomEdit; const Acceptor, Format :String); overload;
  constructor Create(const Path :String;
                     Edit :TCustomEdit; const Acceptor, Format :String); overload;
  // acceptor, format may be '' (pass-thru)
  constructor CreatePy(const Path :String; Edit :TCustomEdit; Acceptor, Format :PPyObject);
  destructor Destroy; override;

  procedure Init(Root :INode; var Reg :TRegistry; const RegBase :String); override;
  procedure Display(Root :INode); override;
  procedure Accept(Root :INode); override; // pump Edit.Text thru acceptor
  // propagates exceptions for you to show user
  // if exception, first replaces text with Format(node-value)

  property Edit          :TCustomEdit      read prEdit;
  property Acceptor      :PPyObject        read prAcceptor;
  property Format        :PPyObject        read prFormat;

  property OldOnKeyPress :TKeyPressEvent   read prOldOnKeyPress  write prOldOnKeyPress;
  property OldOnExit     :TNotifyEvent     read prOldOnExit      write prOldOnExit;

end;


// for a TUpDown associated to a TEdit
// you can still make the TEdit readonly

TQUB_Property_UpDown = class(TQUB_Property)
protected
  prUpDown         :TUpDown;
  prEdit           :TEdit;
  prAcceptor       :PPyObject;
  prOldOnExit      :TNotifyEvent;
  prOldOnUpDownClick  :TUDClickEvent;

  procedure prCreate(UpDown :TUpDown; Edit :TEdit; const Acceptor :String);

public
  constructor Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                     UpDown :TUpDown; Edit :TEdit; const Acceptor :String); overload;
  constructor Create(const Path :String;
                     UpDown :TUpDown; Edit :TEdit; const Acceptor :String); overload;
  destructor Destroy; override;

  procedure Init(Root :INode; var Reg :TRegistry; const RegBase :String); override;
  procedure Display(Root :INode); override;
  procedure Accept(Root :INode); override;
  procedure AcceptText(Root :INode);

  property UpDown      :TUpDown          read prUpDown;
  property Edit        :TEdit            read prEdit;
  property Acceptor    :PPyObject        read prAcceptor;

  property OldOnExit        :TNotifyEvent    read prOldOnExit         write prOldOnExit;
  property OldOnUpDownClick :TUDClickEvent   read prOldOnUpDownClick  write prOldOnUpDownClick;
end;


// property node values are Integer, either 1 or 0
// so should your defaults be

TQUB_Property_CheckBox = class(TQUB_Property)
protected
  prCheckBox         :TCheckBox;
  prEnables          :TList;

  prOldOnClick       :TNotifyEvent;

  procedure prDoEnables;

public
  constructor Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                     CheckBox :TCheckBox); overload;
  constructor Create(const Path :String; CheckBox :TCheckBox); overload;
  destructor Destroy; override;

  procedure Init(Root :INode; var Reg :TRegistry; const RegBase :String); override;
  procedure Display(Root :INode); override;
  procedure Accept(Root :INode); override;

  function Enables(X :TControl): TQUB_Property_CheckBox;

  property CheckBox    :TCheckBox        read prCheckBox;
  property OldOnClick  :TNotifyEvent     read prOldOnClick write prOldOnClick;

end;


// create radiobuttons to attach to radiogroup (below)

TQUB_Property_RadioButton = class
protected
  prEnables        :TList;

  prOldOnClick       :TNotifyEvent;

  procedure prCreate(RadioButton  :TRadioButton; Value :Variant);

public
  RadioButton      :TRadioButton;
  SetValue         :PPyObject;
  RegValue         :Variant;

  constructor Create(RadioButton  :TRadioButton; Value, RegValue :Variant); overload;
  constructor Create(RadioButton  :TRadioButton; Value :Variant); overload;
  destructor Destroy; override;

  procedure SetDataOf(Node :INode; Ix :Integer);
  function DataEquals(Node :INode; Ix :Integer): Boolean;

  function Enables(X :TControl): TQUB_Property_RadioButton;

  procedure DoEnables;

  property OldOnClick  :TNotifyEvent     read prOldOnClick write prOldOnClick;

  end;


// old-style registry values can be string or int
// and new-style node values can (independently) also be either
// your default should be in the node-value-type (Either string or int)

// for easy construction, you can chain calls to AddButton:
//   RadioGroup := TQUB_Property_RadioGroup.Create( ... ).AddButton( ... ).AddButton( ... );

TQUB_Property_RadioGroup = class(TQUB_Property)
protected

  procedure prSetValFromRegVal(Node :INode; RegVal :Variant);

public
  Buttons        :Array of TQUB_Property_RadioButton;

  constructor Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant); overload;
  constructor Create(const Path :String); overload;
  destructor Destroy; override;

  procedure Init(Root :INode; var Reg :TRegistry; const RegBase :String); override;
  procedure Display(Root :INode); override;
  procedure Accept(Root :INode); override;

  function AddButton(Btn :TQUB_Property_RadioButton): TQUB_Property_RadioGroup;
  // returns self so you can chain calls

end;


TQUB_Property_ColorPanel = class(TQUB_Property)
protected
  prPanel        :TPanel;

public
  constructor Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                     Panel :TPanel); overload;
  constructor Create(const Path :String; Panel :TPanel); overload;

  procedure Init(Root :INode; var Reg :TRegistry; const RegBase :String); override;
  procedure Display(Root :INode); override;
  procedure Accept(Root :INode); override;

  property Panel :TPanel   read prPanel;
end;


TQUB_Property_Indexer = class
protected
  prOffset       :Integer;
  prProperties   :Array of TQUB_Property;

public
  UpDown       :TUpDown;

  constructor Create(AUpDown :TUpDown; AOffset :Integer); // undefined if Position is not equal to Offset

  function Add(Prop :TQUB_Property): TQUB_Property_Indexer; // returns Self

  procedure OnChangingEx(Sender: TObject; var AllowChange: Boolean; NewValue: SmallInt; Direction: TUpDownDirection; Root :INode);
end;


const

         AcceptString  :String = 'acceptString';
         AcceptInt     :String = 'acceptInt';
         AcceptFloat   :String = 'acceptFloat';
function AcceptIntGreaterThan(lb :Integer): String;
function AcceptIntGreaterThanOrEqualTo(lb :Integer): String;
function AcceptIntLessThan(ub :Integer): String;
function AcceptIntLessThanOrEqualTo(ub :Integer): String;
function AcceptIntBetween(lb, ub :Integer): String;
function AcceptFloatGreaterThan(lb :Double): String;
function AcceptFloatGreaterThanOrEqualTo(lb :Double): String;
function AcceptFloatLessThan(ub :Double): String;
function AcceptFloatLessThanOrEqualTo(ub :Double): String;
function AcceptFloatBetween(lb, ub :Double): String;
function AcceptFloatAbsBetween(lb, ub :Double): String;
function AcceptIntList(Acceptor :String): String;
function AcceptFloatList(Acceptor :String): String;
function AcceptDimList(AcceptList :String; Dim :Integer; Msg :String): String;


type


TQUB_AProperty_Edit        = Array of TQUB_Property_Edit;
TQUB_AProperty_UpDown      = Array of TQUB_Property_UpDown;
TQUB_AProperty_CheckBox    = Array of TQUB_Property_CheckBox;
TQUB_AProperty_RadioGroup  = Array of TQUB_Property_RadioGroup;
TQUB_AProperty_ColorPanel  = Array of TQUB_Property_ColorPanel;
TQUB_AProperty_Indexer     = Array of TQUB_Property_Indexer;


TQUB_PropertiesForm                       = class(TForm)


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure FormKeyPress(Sender: TObject; var Key: Char);


protected


  prProperties                            :INode;
  prRegKeyName                            :String;

  prEditItems                             :TQUB_AProperty_Edit;
  prUpDownItems                           :TQUB_AProperty_UpDown;
  prCheckBoxItems                         :TQUB_AProperty_CheckBox;
  prRadioGroupItems                       :TQUB_AProperty_RadioGroup;
  prColorPanelItems                       :TQUB_AProperty_ColorPanel;
  prIndexers                              :TQUB_AProperty_Indexer;

  prDlgColor                              :TColorDialog;


  procedure prOnKeyPress(Sender: TObject; var Key: Char);
  procedure prOnExit(Sender :TObject);
  procedure prOnExitUpDownText(Sender :TObject);
  procedure prOnUpDownClick(Sender :TObject; Button: TUDBtnType);
  procedure prOnCheckClick(Sender :TObject);
  procedure prOnRadioClick(Sender :TObject);
  procedure prOnColorClick(Sender :TObject);
  procedure prOnIndexerChangingEx(Sender: TObject; var AllowChange: Boolean; NewValue: SmallInt; Direction: TUpDownDirection);

  procedure prSetInitialProperties(Props :INode); virtual;
  // override: inherited; fill missing properties from registry or defaults;
  // (QVO_Properties doesn't call this because the QVO takes care of defaults and such)

  procedure prUpdateComponents; virtual;
  // override: put all your property values into the form
  // (supersedes MyProperty := MyProperty; in FormShow)
  // also called after SetProperties e.g. when user changes profile

  // first register your items // we adopt responsibility for their destruction
  procedure AddEditProperty(Item :TQUB_Property_Edit);
  procedure AddUpDownProperty(Item :TQUB_Property_UpDown);
  procedure AddCheckBoxProperty(Item :TQUB_Property_CheckBox);
  procedure AddRadioGroupProperty(Item :TQUB_Property_RadioGroup);
  procedure AddColorPanelProperty(Item :TQUB_Property_ColorPanel);
  procedure AddIndexer(Indexer :TQUB_Property_Indexer);


  // after registering items, set this, to get e.g. threadcount and dataprocessed correctly from legacy registry
  property RegKeyName                     :String                read prRegKeyName                write prRegKeyName;


public


  procedure SetProperties(Props :INode);
  // sets only those properties which are in Props


  property Properties                     :INode                 read prProperties;

  property EditItems        :TQUB_AProperty_Edit                 read prEditItems;
  property UpDownItems      :TQUB_AProperty_UpDown               read prUpDownItems;
  property CheckBoxItems    :TQUB_AProperty_CheckBox             read prCheckBoxItems;
  property RadioGroupItems  :TQUB_AProperty_RadioGroup           read prRadioGroupItems;
  property ColorPanelItems  :TQUB_AProperty_ColorPanel           read prColorPanelItems;


end;


implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Viewport_Python,
QUB_Python_Thread,
QUB_Node_Python;


//******************************************************************************
procedure TQUB_PropertiesForm.FormCreate(Sender: TObject);
begin
  while not PythonThread.PythonReady do Sleep(100);
  PythonThread.BeginPython;
  GetPythonEngine.Run_CommandAsString('from Dialogs import *', single_input);
  PythonThread.EndPython;

  prProperties                            := TNode.Create('Properties');
end;

procedure TQUB_PropertiesForm.FormDestroy(Sender: TObject);
var
  i         :Integer;
begin
  try
    for i := 0 to Length(prEditItems) - 1 do begin
      prEditItems[i].Free;
    end;
    for i := 0 to Length(prUpDownItems) - 1 do begin
      prUpDownItems[i].Free;
    end;
    for i := 0 to Length(prCheckBoxItems) - 1 do begin
      prCheckBoxItems[i].Free;
    end;
    for i := 0 to Length(prRadioGroupItems) - 1 do begin
      prRadioGroupItems[i].Free;
    end;
    for i := 0 to Length(prColorPanelItems) - 1 do begin
      prColorPanelItems[i].Free;
    end;
    for i := 0 to Length(prIndexers) - 1 do begin
      prIndexers[i].Free;
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Reg                    :TRegistry;
  i                      :Integer;
begin
  prProperties           := Props;

  Reg                    := nil; // will be created as needed by ReadXRegistry*
  try
    for i := 0 to Length(prEditItems) - 1 do begin
      try prEditItems[i].Init(prProperties, Reg, RegKeyName); except end;
    end;
    for i := 0 to Length(prUpDownItems) - 1 do begin
      try prUpDownItems[i].Init(prProperties, Reg, RegKeyName); except end;
    end;
    for i := 0 to Length(prCheckBoxItems) - 1 do begin
      try prCheckBoxItems[i].Init(prProperties, Reg, RegKeyName); except end;
    end;
    for i := 0 to Length(prRadioGroupItems) - 1 do begin
      try prRadioGroupItems[i].Init(prProperties, Reg, RegKeyName); except end;
    end;
    for i := 0 to Length(prColorPanelItems) - 1 do begin
      try prColorPanelItems[i].Init(prProperties, Reg, RegKeyName); except end;
    end;
  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_PropertiesForm.prUpdateComponents;
var
  i                      :Integer;
begin
  for i := 0 to Length(prEditItems) - 1 do begin
    prEditItems[i].Display(prProperties);
  end;
  for i := 0 to Length(prUpDownItems) - 1 do begin
    prUpDownItems[i].Display(prProperties);
  end;
  for i := 0 to Length(prCheckBoxItems) - 1 do begin
    prCheckBoxItems[i].Display(prProperties);
  end;
  for i := 0 to Length(prRadioGroupItems) - 1 do begin
    prRadioGroupItems[i].Display(prProperties);
  end;
  for i := 0 to Length(prColorPanelItems) - 1 do begin
    prColorPanelItems[i].Display(prProperties);
  end;
end;

procedure TQUB_PropertiesForm.SetProperties(Props :INode);
  procedure Recur(Mine, Theirs :INode);
  // copies data from Theirs to Mine for all nodes existing in both
  var
    MyChild, TheirChild :INode;
  begin
    TheirChild := Theirs.Child;
    while not TheirChild.IsNull do begin
      MyChild := Mine.Find(TheirChild.Name);
      if not MyChild.IsNull then begin
        MyChild.EmulateData(TheirChild);
        Recur(MyChild, TheirChild);
      end;
      TheirChild := TheirChild.Sibling;
    end;
  end;
begin
  Recur(Properties, Props);
end;


procedure TQUB_PropertiesForm.AddEditProperty(Item :TQUB_Property_Edit);
var
  i        :Integer;
begin
  i := Length(prEditItems);
  SetLength(prEditItems, i+1);
  prEditItems[i] := Item;

  //with Item.Edit as Item.Edit.ClassType do begin
  //TCustomEdit has OnKeyPress and OnExit protected, but the actual derived classes don't
  //  OnKeyPress := prOnKeyPress;
  //  OnExit     := prOnExit;
  //end;

  //This should fix it
  if Item.Edit is TEdit then begin
    Item.OldOnKeyPress          := TEdit(Item.Edit).OnKeyPress;
    Item.OldOnExit              := TEdit(Item.Edit).OnExit;
    TEdit(Item.Edit).OnKeyPress := prOnKeyPress;
    TEdit(Item.Edit).OnExit     := prOnExit;
  end else if Item.Edit is TMemo then begin
    Item.OldOnKeyPress          := TMemo(Item.Edit).OnKeyPress;
    Item.OldOnExit              := TMemo(Item.Edit).OnExit;
    TMemo(Item.Edit).OnKeyPress := prOnKeyPress;
    TMemo(Item.Edit).OnExit     := prOnExit;
  end else if Item.Edit is TRichEdit then begin
    Item.OldOnKeyPress          := TRichEdit(Item.Edit).OnKeyPress;
    Item.OldOnExit              := TRichEdit(Item.Edit).OnExit;
    TRichEdit(Item.Edit).OnKeyPress := prOnKeyPress;
    TRichEdit(Item.Edit).OnExit     := prOnExit;
  end;


  // Chris sez: i dunno. the problem is that with your style prOnExit is never called.

  //Item.Edit.OnKeyPress := prOnKeyPress;
  //Item.Edit.OnExit     := prOnExit;
end;

procedure TQUB_PropertiesForm.AddUpDownProperty(Item :TQUB_Property_UpDown);
var
  i        :Integer;
begin
  i := Length(prUpDownItems);
  SetLength(prUpDownItems, i+1);
  prUpDownItems[i] := Item;

  Item.OldOnUpDownClick := Item.UpDown.OnClick;
  Item.UpDown.OnClick   := prOnUpDownClick;
  Item.Edit.OnKeyPress  := prOnKeyPress;
  Item.OldOnExit        := Item.Edit.OnExit;
  Item.Edit.OnExit      := prOnExitUpDownText;
end;

procedure TQUB_PropertiesForm.AddCheckBoxProperty(Item :TQUB_Property_CheckBox);
var
  i        :Integer;
begin
  i := Length(prCheckBoxItems);
  SetLength(prCheckBoxItems, i+1);
  prCheckBoxItems[i] := Item;

  Item.OldOnClick := Item.CheckBox.OnClick;
  Item.CheckBox.OnClick := prOnCheckClick;
end;

procedure TQUB_PropertiesForm.AddRadioGroupProperty(Item :TQUB_Property_RadioGroup);
var
  i        :Integer;
begin
  i := Length(prRadioGroupItems);
  SetLength(prRadioGroupItems, i+1);
  prRadioGroupItems[i] := Item;

  for i := 0 to Length(Item.Buttons) - 1 do begin
    Item.Buttons[i].OldOnClick := Item.Buttons[i].RadioButton.OnClick;
    Item.Buttons[i].RadioButton.OnClick := prOnRadioClick;
  end;
end;

procedure TQUB_PropertiesForm.AddColorPanelProperty(Item :TQUB_Property_ColorPanel);
var
  i        :Integer;
begin
  i := Length(prColorPanelItems);
  SetLength(prColorPanelItems, i+1);
  prColorPanelItems[i] := Item;

  Item.Panel.OnClick := prOnColorClick;

  if i = 0 then prDlgColor := TColorDialog.Create(Self);
end;

procedure TQUB_PropertiesForm.AddIndexer(Indexer :TQUB_Property_Indexer);
var
  i        :Integer;
begin
  i := Length(prIndexers);
  SetLength(prIndexers, i+1);
  prIndexers[i] := Indexer;
  Indexer.UpDown.OnChangingEx := prOnIndexerChangingEx;
end;

procedure TQUB_PropertiesForm.prOnKeyPress(Sender: TObject; var Key: Char);
var
  i :Integer;
begin
  inherited;//Is this call right?! LM
  for i := 0 to Length(prEditItems) - 1 do begin
    if Sender = prEditItems[i].Edit then begin
      if Key = #13 then begin
        if Sender is TEdit then begin
          Key := #0;
          TEdit(Sender).OnExit(Sender);
        end else if Sender is TMemo then begin
          TMemo(Sender).OnExit(Sender);
        end else if Sender is TRichEdit then begin
          TRichEdit(Sender).OnExit(Sender);
        end;
      end else begin
        if Sender is TEdit then begin
          TEdit(Sender).Font.Color := clRed;
        end else if Sender is TMemo then begin
          TMemo(Sender).Font.Color := clRed;
        end;
      end;
      if Assigned(prEditItems[i].OldOnKeyPress) then prEditItems[i].OldOnKeyPress(prEditItems[i].Edit, Key);
    end;
  end;
end;

procedure TQUB_PropertiesForm.prOnExit(Sender :TObject);
var
  i :Integer;
begin
  for i := 0 to Length(prEditItems) - 1 do begin
    if Sender = prEditItems[i].Edit then begin
      if Sender is TEdit then begin
        TEdit(Sender).Font.Color := clWindowText;
      end else if Sender is TMemo then begin
        TMemo(Sender).Font.Color := clWindowText;
      end;
      try
        prEditItems[i].Accept(Properties);
      except
        on E: Exception do begin
          MessageDlg(E.Message, mtError, [mbOK], 0);
          TCustomEdit(Sender).SetFocus;
        end;
      end;
      if Assigned(prEditItems[i].OldOnExit) then prEditItems[i].OldOnExit(prEditItems[i]);
      Break;
    end;
  end;
end;

procedure TQUB_PropertiesForm.prOnExitUpDownText(Sender :TObject);
var
  i        :Integer;
begin
  for i := 0 to Length(prUpDownItems) - 1 do begin
    if Sender = prUpDownItems[i].Edit then begin
      TEdit(Sender).Font.Color := clWindowText;
      try
        prUpDownItems[i].AcceptText(Properties);
      except
        on E: Exception do begin
          MessageDlg(E.Message, mtError, [mbOK], 0);
          TEdit(Sender).SetFocus;
        end;
      end;
      if Assigned(prUpDownItems[i].OldOnExit) then prUpDownItems[i].OldOnExit(prEditItems[i]);
      Break;
    end;
  end;
end;

procedure TQUB_PropertiesForm.prOnUpDownClick(Sender :TObject; Button: TUDBtnType);
var
  i        :Integer;
begin
  for i := 0 to Length(prUpDownItems) - 1 do begin
    if Sender = prUpDownItems[i].UpDown then begin
      prUpDownItems[i].Accept(Properties);
      prUpDownItems[i].Edit.Font.Color := clWindowText;
      if Assigned(prUpDownItems[i].OldOnUpDownClick) then prUpDownItems[i].OldOnUpDownClick(Sender, Button);
      Break;
    end;
  end;
end;

procedure TQUB_PropertiesForm.prOnCheckClick(Sender :TObject);
var
  i        :Integer;
begin
  for i := 0 to Length(prCheckBoxItems) - 1 do begin
    if Sender = prCheckBoxItems[i].CheckBox then begin
      prCheckBoxItems[i].Accept(Properties);
      if Assigned(prCheckBoxItems[i].OldOnClick) then prCheckBoxItems[i].OldOnClick(Sender);
      Break;
    end;
  end;
end;

procedure TQUB_PropertiesForm.prOnRadioClick(Sender :TObject);
var
  i, j     :Integer;
begin
  for i := 0 to Length(prRadioGroupItems) - 1 do begin
    for j := 0 to Length(prRadioGroupItems[i].Buttons) - 1 do begin
      if Sender = prRadioGroupItems[i].Buttons[j].RadioButton then begin
        prRadioGroupItems[i].Accept(Properties);
        if Assigned(prRadioGroupItems[i].Buttons[j].OldOnClick) then prRadioGroupItems[i].Buttons[j].OldOnClick(Sender);
        Break;
      end;
    end;
  end;
end;

procedure TQUB_PropertiesForm.prOnColorClick(Sender :TObject);
var
  i        :Integer;
begin
  for i := 0 to Length(prColorPanelItems) - 1 do begin
    if Sender = prColorPanelItems[i].Panel then begin
      if prDlgColor.Execute then begin
        prColorPanelItems[i].Panel.Color := prDlgColor.Color;
        prColorPanelItems[i].Accept(Properties);
      end;
      Break;
    end;
  end;
end;

procedure TQUB_PropertiesForm.prOnIndexerChangingEx(Sender: TObject; var AllowChange: Boolean; NewValue: SmallInt; Direction: TUpDownDirection);
var
  i         :Integer;
begin
  for i := 0 to Length(prIndexers) - 1 do begin
    if Sender = prIndexers[i].UpDown then begin
      AllowChange := (NewValue >= prIndexers[i].UpDown.Min)
                 and (NewValue <= prIndexers[i].UpDown.Max);
      if AllowChange then begin
        prIndexers[i].OnChangingEx(Sender, AllowChange, NewValue, Direction, Properties);
        prUpdateComponents;
      end;
      Break;
    end;
  end;
end;


//******************************************************************************

constructor TQUB_Property.CreateBase(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant);
begin
  prPath       := Path;
  prRegName    := RegName;
  prRegType    := RegType;
  prDefVal     := DefVal;
  prIx         := 0;
  prEnsuredDim := False;
end;

constructor TQUB_Property.CreateBase(const Path :String);
begin
  prPath       := Path;
  prRegName    := '';
  prRegType    := REGTYPE_NONE;
  prDefVal     := 0;
  prIx         := 0;
  prEnsuredDim := False;
end;

function TQUB_Property.prGetNode(Root :INode): INode;
var
  PyNode         :PPyObject;
  Getter, Args   :PPyObject;
begin
  with GetPythonEngine do begin
    Getter       := EvalString('lambda root: root'+Path);
    Args         := Py_BuildValue('(N)', [PyNode_FromNode(Root)]);
    PyNode       := PyObject_CallObject(Getter, Args);
    if PyNode = nil then begin
      PyErr_Print;
      Result     := QUB_NODE_NULL;
    end else begin
      Result     := PyNode_ToNode(PyNode);
      Py_DECREF(PyNode);
    end;
    Py_DECREF(Getter);
    Py_DECREF(Args);
  end;
end;

procedure TQUB_Property.prSetIx(x :Integer);
begin
  prIx         := x;
  prEnsuredDim := False;
end;

procedure TQUB_Property.prEnsureDim(Node :INode);
var
  At           :Integer;
begin
  if prEnsuredDim then Exit;
  prEnsuredDim := True;
  
  if (Node.DataType <> NODE_DATATYPE_STRING) and (LongInt(Node.DataCount) <= Ix) then begin
    At := Node.DataCount;
    if Node.DataCount = 0 then Node.SetData(DefVal); // to initialize data type
    Node.ResizeData(Ix+1);
    while At <= Ix do begin
      Node.Data[At] := DefVal;
      Inc(At);
    end;
  end;
end;

//******************************************************************************

constructor TQUB_Property_Edit.Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                                      Edit :TCustomEdit; const Acceptor, Format :String);
begin
  inherited CreateBase(Path, RegName, RegType, DefVal);
  prCreate(Edit, Acceptor, Format);
end;

constructor TQUB_Property_Edit.Create(const Path :String;
                                      Edit :TCustomEdit; const Acceptor, Format :String);
begin
  inherited CreateBase(Path);
  prCreate(Edit, Acceptor, Format);
end;

procedure TQUB_Property_Edit.prCreate(Edit :TCustomEdit; const Acceptor, Format :String);
begin
  prEdit         := Edit;

  try
    PythonThread.BeginPython;
    with GetPythonEngine do begin
      prAcceptor       := nil;
      try
        if Acceptor <> '' then prAcceptor := EvalString('NodeAcceptor('+Acceptor+', '''+Path+''')');
      except
      end;
      if prAcceptor = nil then prAcceptor := EvalString('NodeAcceptor(acceptString, '''+Path+''')');

      prFormat         := nil;
      try
        if Copy(Format, 0, 6) = 'lambda' then begin
          prFormat := EvalString(Format);
        end else if Format <> '' then begin
          prFormat := EvalString('NodeFormat("'+Format+'")');
        end;
      except
      end;
      if prFormat = nil then prFormat := EvalString('NodeFormat("%s")');
    end;
  finally
    PythonThread.EndPython;
  end;
end;

constructor TQUB_Property_Edit.CreatePy(const Path :String; Edit :TCustomEdit; Acceptor, Format :PPyObject);
begin
  inherited CreateBase(Path, '', REGTYPE_NONE, 0{default is never used if python:Buttons has set up category first});
  prEdit         := Edit;

  try
    PythonThread.BeginPython;
    with GetPythonEngine do begin
      prAcceptor       := Acceptor;
      Py_INCREF(prAcceptor);

      prFormat         := Format;
      if prFormat = nil then begin
        prFormat := EvalString('NodeFormat("%s")');
      end else begin
        Py_INCREF(prFormat);
      end;
    end;
  finally
    PythonThread.EndPython;
  end;
end;

destructor TQUB_Property_Edit.Destroy;
begin
  try
    if not PythonOK then Exit;
    PythonThread.BeginPython;
    with GetPythonEngine do begin
      Py_XDECREF(prAcceptor);
      Py_XDECREF(prFormat);
    end;
  finally
    if PythonOK then PythonThread.EndPython;
    inherited;
  end;
end;

procedure TQUB_Property_Edit.Init(Root :INode; var Reg :TRegistry; const RegBase :String);
var
  Node         :INode;
  iVal         :Integer;
  dVal         :Double;
  sVal         :String;
begin
  PythonThread.BeginPython;
  Node         := prGetNode(Root);
  PythonThread.EndPython;

  if Node.DataCount = 0 then begin
    try
      if RegType = REGTYPE_INT then begin
        if ReadXRegistryInteger(Reg, RegBase, RegName, iVal)
        then Node.SetData(iVal);
      end else if RegType = REGTYPE_FLOAT then begin
        if ReadXRegistryFloat(Reg, RegBase, RegName, dVal)
        then Node.SetData(dVal);
      end else if RegType = REGTYPE_STRINGFLOAT then begin
        if ReadXRegistryStringFloat(Reg, RegBase, RegName, dVal)
        then Node.SetData(dVal);
      end else if RegType = REGTYPE_STRING then begin
        if ReadXRegistryString(Reg, RegBase, RegName, sVal)
        then Node.SetData(sVal);
      end else if RegType = REGTYPE_NONE then begin
                           // good reg type
      end else begin
        raise Exception.Create('bad reg type');
      end;
    except
    end;

    if Node.DataCount = 0 then begin
      Node.SetData(DefVal);
    end;
  end;
end;

procedure TQUB_Property_Edit.prDisplay(Node :PPyObject);
var
  Args          :PPyObject;
  PyResult      :PPyObject;
begin
  with GetPythonEngine do begin
    Args          := Py_BuildValue('(O, i)', [Node, Ix]);
    PyResult      := PyObject_CallObject(prFormat, Args);
    if PyResult = nil then begin
      PyErr_Print;
      PyErr_Clear;
    end else begin
      Edit.Text   := PyString_AsString(PyResult);
      Py_DECREF(PyResult);
    end;
    Py_DECREF(Args);
  end;
end;

procedure TQUB_Property_Edit.Display(Root :INode);
var
  Node            :INode;
  PyNode          :PPyObject;
begin
  with GetPythonEngine do begin
    PyNode          := nil;
    try
      PythonThread.BeginPython;
      Node            := prGetNode(Root);
      prEnsureDim(Node);
      PyNode          := PyNode_FromNode(Node);
      prDisplay(PyNode);
    finally
      Py_XDECREF(PyNode);
      PythonThread.EndPython;
    end;
  end;
end;

procedure TQUB_Property_Edit.Accept(Root :INode);
var
  PyRoot        :PPyObject;
  PyNode        :PPyObject;
  PyResult      :PPyObject;
  Args          :PPyObject;
  Msg           :String;
begin
  with GetPythonEngine do begin
    PyRoot        := nil;
    PyNode        := nil;
    PyResult      := nil;
    Args          := nil;
    try
      PythonThread.BeginPython;
      PyRoot    := PyNode_FromNode(Root);
      PyNode    := PyNode_FromNode(prGetNode(Root));
      with GetPythonEngine do begin
        Args    := Py_BuildValue('(s, O, i)', [PChar(prEdit.Text), PyRoot, Ix]);
        PyResult   := PyObject_CallObject( Acceptor, Args );
        if PyResult = nil then begin
          Msg      := PyErr_ShortMessage;
          PyErr_Print;
          prDisplay(PyNode);
          raise Exception.Create('Invalid value; ' + Copy(Msg, 60, Length(Msg)));//30 is too short
          // 30 is the length of "Dialogs.FieldAcceptException: "
        end;
      end;
    finally
      Py_XDECREF(PyRoot);
      Py_XDECREF(PyNode);
      Py_XDECREF(PyResult);
      Py_XDECREF(Args);
      PythonThread.EndPython;
    end;
  end;
end;


//******************************************************************************

constructor TQUB_Property_UpDown.Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                   UpDown :TUpDown; Edit :TEdit; const Acceptor :String);
begin
  inherited CreateBase(Path, RegName, RegType, DefVal);
  prCreate(UpDown, Edit, Acceptor);
end;

constructor TQUB_Property_UpDown.Create(const Path :String;
                   UpDown :TUpDown; Edit :TEdit; const Acceptor :String);
begin
  inherited CreateBase(Path);
  prCreate(UpDown, Edit, Acceptor);
end;

procedure TQUB_Property_UpDown.prCreate(UpDown :TUpDown; Edit :TEdit; const Acceptor :String);
begin
  prUpDown       := UpDown;
  prEdit         := Edit;

  try
    PythonThread.BeginPython;
    with GetPythonEngine do begin
      prAcceptor       := nil;
      try
        if Acceptor <> '' then prAcceptor := EvalString('NodeAcceptor('+Acceptor+', '''+Path+''')');
      except
      end;
      if prAcceptor = nil then prAcceptor := EvalString('NodeAcceptor(acceptString, '''+Path+''')');
    end;
  finally
    PYthonThread.EndPython;
  end;
end;

destructor TQUB_Property_UpDown.Destroy;
begin
  try
    if not PythonOK then Exit;
    PythonThread.BeginPython;
    with GetPythonEngine do begin
      Py_XDECREF(prAcceptor);
    end;
  finally
    if PythonOK then PythonThread.EndPython;
    inherited;
  end;
end;


procedure TQUB_Property_UpDown.Init(Root :INode; var Reg :TRegistry; const RegBase :String);
var
  Node         :INode;
  iVal         :Integer;
begin
  PythonThread.BeginPython;
  Node         := prGetNode(Root);
  PythonThread.EndPython;

  if Node.DataCount = 0 then begin
    try
      if RegType = REGTYPE_INT then begin
        if ReadXRegistryInteger(Reg, RegBase, RegName, iVal)
        then Node.SetData(iVal);
      end else if RegType = REGTYPE_NONE then begin
                           // good reg type
      end else begin
        raise Exception.Create('bad reg type');
      end;
    except
    end;

    if Node.DataCount = 0 then begin
      Node.SetData(DefVal);
    end;
  end;
end;

procedure TQUB_Property_UpDown.Display(Root :INode);
var
  Node             :INode;
begin
  PythonThread.BeginPython;
  Node             := prGetNode(Root);
  PythonThread.EndPython;
  prEnsureDim(Node);
  if   UpDown.Position <> Node.Data[Ix]
  then UpDown.Position := Node.Data[Ix]
  else Edit.Text       := IntToStr(Node.Data[Ix]);
end;

procedure TQUB_Property_UpDown.Accept(Root :INode);
var
  Node             :INode;
begin
  PythonThread.BeginPython;
  Node             := prGetNode(Root);
  PythonThread.EndPython;
  prEnsureDim(Node);
  Node.Data[Ix]     := UpDown.Position;
end;

procedure TQUB_Property_UpDown.AcceptText(Root :INode);
var
  PyNode        :PPyObject;
  PyResult      :PPyObject;
  Args          :PPyObject;
  Msg           :String;
begin
  with GetPythonEngine do begin
    PyNode        := nil;
    PyResult      := nil;
    Args          := nil;
    try
      PythonThread.BeginPython;
      PyNode    := PyNode_FromNode(Root);
      with GetPythonEngine do begin
        Args       := Py_BuildValue('(s, O, i)', [PChar( prEdit.Text ), PyNode, Ix]);
        PyResult   := PyObject_CallObject( Acceptor, Args );
        if PyResult = nil then begin
          Msg      := PyErr_ShortMessage;
          raise Exception.Create('Invalid value; ' + Copy(Msg, 30, Length(Msg)));
          // 30 is the length of "Dialogs.FieldAcceptException: "
        end;
      end;
    finally
      Py_XDECREF(PyNode);
      Py_XDECREF(PyResult);
      Py_XDECREF(Args);
      PythonThread.EndPython;
      Display(Root);
    end;
  end;
end;


//******************************************************************************

constructor TQUB_Property_CheckBox.Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                                          CheckBox :TCheckBox);
begin
  inherited CreateBase(Path, RegName, RegType, DefVal);

  prCheckBox   := CheckBox;
  prEnables    := TList.Create;
end;

constructor TQUB_Property_CheckBox.Create(const Path :String; CheckBox :TCheckBox);
begin
  inherited CreateBase(Path);

  prCheckBox   := CheckBox;
  prEnables    := TList.Create;
end;

destructor TQUB_Property_CheckBox.Destroy;
begin
  try
    prEnables.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Property_CheckBox.Init(Root :INode; var Reg :TRegistry; const RegBase :String);
var
  Node         :INode;
  bVal         :Boolean;
  iVal         :Integer;
begin
  PythonThread.BeginPython;
  Node         := prGetNode(Root);
  PythonThread.EndPython;

  if Node.DataCount = 0 then begin
    try
      if RegType = REGTYPE_BOOL then begin
        if ReadXRegistryBoolean(Reg, RegBase, RegName, bVal)
        then Node.SetData( Integer(bVal) );
      end else if RegType = REGTYPE_INT then begin
        if ReadXRegistryInteger(Reg, RegBase, RegName, iVal)
        then Node.SetData(iVal);
      end else if RegType = REGTYPE_NONE then begin
                           // good reg type
      end else begin
        raise Exception.Create('bad reg type');
      end;
    except
    end;

    if Node.DataCount = 0 then begin
      Node.SetData(DefVal);
    end;
  end;
end;

procedure TQUB_Property_CheckBox.Display(Root :INode);
var
  Node             :INode;
  Chk              :Boolean;
begin
  PythonThread.BeginPython;
  Node             := prGetNode(Root);
  PythonThread.EndPython;

  prEnsureDim(Node);
  Chk              := 0<>Node.Data[Ix];
  if   CheckBox.Checked <> Chk
  then CheckBox.Checked := Chk;

  prDoEnables;
end;

procedure TQUB_Property_CheckBox.Accept(Root :INode);
var
  Node             :INode;
begin
  PythonThread.BeginPython;
  Node             := prGetNode(Root);
  PythonThread.EndPython;

  prEnsureDim(Node);
  Node.Data[Ix]     := Integer(CheckBox.Checked);

  prDoEnables;
end;

function TQUB_Property_CheckBox.Enables(X :TControl): TQUB_Property_CheckBox;
begin
  prEnables.Add(X);
  Result := Self;
end;

procedure TQUB_Property_CheckBox.prDoEnables;
var
  i        :Integer;
begin
  for i := 0 to prEnables.Count - 1 do begin
    TControl(prEnables[i]).Enabled := CheckBox.Enabled and CheckBox.Checked;
  end;
end;


//******************************************************************************

constructor TQUB_Property_RadioButton.Create(RadioButton  :TRadioButton; Value, RegValue :Variant);
begin
  Self.RegValue    := RegValue;
  prCreate(RadioButton, Value);
end;

constructor TQUB_Property_RadioButton.Create(RadioButton  :TRadioButton; Value :Variant);
begin
  Self.RegValue       := 0;
  prCreate(RadioButton, Value);
end;

procedure TQUB_Property_RadioButton.prCreate(RadioButton  :TRadioButton; Value :Variant);
var
  Dummy       :INode;
  PyDum       :PPyObject;
  PyDat       :PPyObject;
  PyVal       :PPyObject;

  function PyValueOfVariant(V :Variant): PPyObject;
  begin
    with GetPythonEngine do begin
      Dummy.SetData(V);
      if (Dummy.DataType = NODE_DATATYPE_STRING)
      or (Dummy.DataCount = 0) then begin
        Result  := PyObject_Str(PyDat);
      end else begin
        Result  := PySequence_GetItem(PyDat, 0);
      end;
    end;
  end;

  function MakeNodeSetter(PyVal :PPyObject): PPyObject;
  var
    Ctor       :PPyObject;
    Args       :PPyObject;
  begin
    with GetPythonEngine do begin
      Ctor     := EvalString('NodeSetter');
      Args     := Py_BuildValue('(O)', [PyVal]);
      Result   := PyObject_CallObject(Ctor, Args);
      if Result = nil then PyErr_Print;
      Py_DECREF(Ctor);
      Py_DECREF(Args);
    end;
  end;

begin
  Self.RadioButton := RadioButton;

  Dummy                := TNode.Create('');
  PyDum                := nil;
  PyDat                := nil;
  PyVal                := nil;
  with GetPythonEngine do begin
    try
      PythonThread.BeginPython;
      PyDum            := PyNode_FromNode(Dummy);
      PyDat            := PyObject_GetAttrString(PyDum, 'data');
      PyVal            := PyValueOfVariant(Value);
      Self.SetValue    := MakeNodeSetter(PyVal);
    finally
      Py_XDECREF(PyDum);
      Py_XDECREF(PyDat);
      Py_XDECREF(PyVal);
      PythonThread.EndPython;
    end;
  end;

  prEnables            := TList.Create;
end;

destructor TQUB_Property_RadioButton.Destroy;
begin
  try
    if not PythonOK then Exit;
    PythonThread.BeginPython;
    GetPythonEngine.Py_XDECREF(SetValue);
    prEnables.Free;
  finally
    if PythonOK then PythonThread.EndPython;
    inherited;
  end;
end;

procedure TQUB_Property_RadioButton.SetDataOf(Node :INode; Ix :Integer);
var
  Args         :PPyObject;
  PyResult      :PPyObject;
begin
  with GetPythonEngine do begin
    Args     := Py_BuildValue('(N, i)', [PyNode_FromNode(Node), Ix]);
    PyResult := PyObject_CallObject(SetValue, Args);
    if PyResult = nil then begin
      PyErr_Print;
    end else begin
      Py_DECREF(PyResult);
    end;
    Py_DECREF(Args);
  end;
end;

function TQUB_Property_RadioButton.DataEquals(Node :INode; Ix :Integer): Boolean;
var
  Dummy    :INode;
begin
  Dummy    := TNode.Create('');
  SetDataOf(Dummy, Ix);
  if Node.DataType = NODE_DATATYPE_STRING then begin
    Result := Node.DataAsString = Dummy.DataAsString;
  end else if (Node.DataCount > 0) and (Dummy.DataCount > 0) then begin
    Result := Node.Data[Ix] = Dummy.Data[0];
  end else begin
    Result := Node.DataCount = Dummy.DataCount;
  end;
end;

function TQUB_Property_RadioButton.Enables(X :TControl): TQUB_Property_RadioButton;
begin
  prEnables.Add(X);
  Result := Self;
end;

procedure TQUB_Property_RadioButton.DoEnables;
var
  i        :Integer;
begin
  for i := 0 to prEnables.Count - 1 do begin
    TControl(prEnables[i]).Enabled := RadioButton.Enabled and RadioButton.Checked;
  end;
end;



constructor TQUB_Property_RadioGroup.Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant);
begin
  inherited CreateBase(Path, RegName, RegType, DefVal);
end;

constructor TQUB_Property_RadioGroup.Create(const Path :String);
begin
  inherited CreateBase(Path);
end;

destructor TQUB_Property_RadioGroup.Destroy;
var
  i        :Integer;
begin
  try
    for i := 0 to Length(Buttons) - 1 do begin
      Buttons[i].Free;
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_Property_RadioGroup.prSetValFromRegVal(Node :INode; RegVal :Variant);
var
  i            :Integer;
begin
  for i := 0 to Length(Buttons) - 1 do begin
    if RegVal = Buttons[i].RegValue then begin
      try
        PythonThread.BeginPython;
        Buttons[i].SetDataOf(Node, Ix);
      finally
        PythonThread.EndPython;
      end;
      Break;
    end;
  end;
end;

procedure TQUB_Property_RadioGroup.Init(Root :INode; var Reg :TRegistry; const RegBase :String);
var
  Node         :INode;
  iVal         :Integer;
  sVal         :String;
begin
  PythonThread.BeginPython;
  Node         := prGetNode(Root);
  PythonThread.EndPython;

  if Node.DataCount = 0 then begin
    try
      if RegType = REGTYPE_INT then begin
        if ReadXRegistryInteger(Reg, RegBase, RegName, iVal)
        then prSetValFromRegVal(Node, iVal);
      end else if RegType = REGTYPE_STRING then begin
        if ReadXRegistryString(Reg, RegBase, RegName, sVal)
        then prSetValFromRegVal(Node, sVal);
      end else if RegType = REGTYPE_NONE then begin
                             // good reg type
      end else begin
        raise Exception.Create('bad reg type');
      end;
    except
    end;

    if Node.DataCount = 0 then begin
      Node.SetData(DefVal);
    end;
  end;
end;

procedure TQUB_Property_RadioGroup.Display(Root :INode);
var
  Node          :INode;
  i             :Integer;
  Chk           :Boolean;
begin
  try
    PythonThread.BeginPython;
    Node          := prGetNode(Root);
    prEnsureDim(Node);
    for i := 0 to Length(Buttons) - 1 do begin
      Chk            := Buttons[i].DataEquals(Node, Ix);
      PythonThread.EndPython;
      if   Buttons[i].RadioButton.Checked <> Chk
      then Buttons[i].RadioButton.Checked := Chk;
      Buttons[i].DoEnables;
      PythonThread.BeginPython;
    end;
  finally
    PythonThread.EndPython;
  end;
end;

procedure TQUB_Property_RadioGroup.Accept(Root :INode);
var
  Node          :INode;
  i             :Integer;
begin
  try
    PythonThread.BeginPython;
    Node          := prGetNode(Root);
    prEnsureDim(Node);
    for i := 0 to Length(Buttons) - 1 do begin
      if Buttons[i].RadioButton.Checked then begin
        Buttons[i].SetDataOf(Node, Ix);
      end;
      Buttons[i].DoEnables;
    end;
  finally
    PythonThread.EndPython;
  end;
end;

function TQUB_Property_RadioGroup.AddButton(Btn :TQUB_Property_RadioButton): TQUB_Property_RadioGroup;
var
  i       :Integer;
begin
  i       := Length(Buttons);
  SetLength(Buttons, i+1);
  Buttons[i] := Btn;
  Result  := Self;
end;


//******************************************************************************

constructor TQUB_Property_ColorPanel.Create(const Path, RegName :String; RegType :TQUB_Property_RegType; DefVal :Variant;
                                            Panel :TPanel);
begin
  inherited CreateBase(Path, RegName, RegType, DefVal);
  prPanel := Panel;
end;

constructor TQUB_Property_ColorPanel.Create(const Path :String; Panel :TPanel);
begin
  inherited CreateBase(Path);
  prPanel := Panel;
end;

procedure TQUB_Property_ColorPanel.Init(Root :INode; var Reg :TRegistry; const RegBase :String);
var
  Node         :INode;
  iVal         :Integer;
begin
  PythonThread.BeginPython;
  Node         := prGetNode(Root);
  PythonThread.EndPython;

  if Node.DataCount = 0 then begin
    try
      if RegType = REGTYPE_INT then begin
        if ReadXRegistryInteger(Reg, RegBase, RegName, iVal)
        then Node.SetData(iVal);
      end else if RegType = REGTYPE_NONE then begin
                             // good reg type
      end else begin
        raise Exception.Create('bad reg type');
      end;
    except
    end;

    if Node.DataCount = 0 then begin
      Node.SetData(DefVal);
    end;
  end;
end;

procedure TQUB_Property_ColorPanel.Display(Root :INode);
var
  Node             :INode;
begin
  PythonThread.BeginPython;
  Node             := prGetNode(Root);
  prEnsureDim(Node);
  PythonThread.EndPython;
  prPanel.Color  := TColor(Node.Data[Ix]);
end;

procedure TQUB_Property_ColorPanel.Accept(Root :INode);
var
  Node             :INode;
begin
  PythonThread.BeginPython;
  Node             := prGetNode(Root);
  prEnsureDim(Node);
  PythonThread.EndPython;
  Node.Data[Ix]     := Integer(prPanel.Color);
end;



//******************************************************************************
constructor TQUB_Property_Indexer.Create(AUpDown :TUpDown; AOffset :Integer);
begin
  UpDown          := AUpDown;
  prOffset          := AOffset;
end;

function TQUB_Property_Indexer.Add(Prop :TQUB_Property): TQUB_Property_Indexer; // returns Self
var
  i                 :Integer;
begin
  i                 := Length(prProperties);
  SetLength(prProperties, i+1);
  prProperties[i]   := Prop;
  Result            := Self;
end;

procedure TQUB_Property_Indexer.OnChangingEx(Sender: TObject; var AllowChange: Boolean; NewValue: SmallInt; Direction: TUpDownDirection; Root :INode);
var
  i                 :Integer;
begin
  AllowChange       := True;
  for i := 0 to Length(prProperties) - 1 do begin
    try
      prProperties[i].Accept(Root);
    except
    end;
    prProperties[i].Ix := NewValue - prOffset;
  end;
end;


//******************************************************************************
//Chris, you must add >= and <= !!! LM.
//.....................................
//Done
function AcceptIntGreaterThan(lb :Integer): String;
begin
  Result := Format('acceptIntGreaterThan(%d)', [lb]);
end;

function AcceptIntGreaterThanOrEqualTo(lb :Integer): String;
begin
  Result := Format('acceptIntGreaterThanOrEqualTo(%d)', [lb]);
end;

function AcceptIntLessThan(ub :Integer): String;
begin
  Result := Format('acceptIntLessThan(%d)', [ub]);
end;

function AcceptIntLessThanOrEqualTo(ub :Integer): String;
begin
  Result := Format('acceptIntLessThanOrEqualTo(%d)', [ub]);
end;

function AcceptIntBetween(lb, ub :Integer): String;
begin
  Result := Format('acceptIntBetween(%d, %d)', [lb, ub]);
end;

function Accept_FormatFloat(x :Double): String;
begin
  Result := StringReplace(Format('%g', [x]), ',', '.', [rfReplaceAll]);  // fix euro decimals
end;

function AcceptFloatGreaterThan(lb :Double): String;
begin
  Result := Format('acceptFloatGreaterThan(%s)', [Accept_FormatFloat(lb)]);
end;

function AcceptFloatGreaterThanOrEqualTo(lb :Double): String;
begin
  Result := Format('acceptFloatGreaterThanOrEqualTo(%s)', [Accept_FormatFloat(lb)]);
end;

function AcceptFloatLessThan(ub :Double): String;
begin
  Result := Format('acceptFloatLessThan(%s)', [Accept_FormatFloat(ub)]);
end;

function AcceptFloatLessThanOrEqualTo(ub :Double): String;
begin
  Result := Format('acceptFloatLessThanOrEqualTo(%s)', [Accept_FormatFloat(ub)]);
end;

function AcceptFloatBetween(lb, ub :Double): String;
begin
  Result := Format('acceptFloatBetween(%s, %s)', [Accept_FormatFloat(lb),
                                                  Accept_FormatFloat(ub)]);
end;

function AcceptFloatAbsBetween(lb, ub :Double): String;
begin
  Result := Format('acceptFloatAbsBetween(%s, %s)', [Accept_FormatFloat(lb),
                                                     Accept_FormatFloat(ub)]);
end;

function AcceptIntList(Acceptor :String): String;
begin
  Result := Format('acceptIntList(%s)', [Acceptor]);
end;

function AcceptFloatList(Acceptor :String): String;
begin
  Result := Format('acceptFloatList(%s)', [Acceptor]);
end;

function AcceptDimList(AcceptList :String; Dim :Integer; Msg :String): String;
begin
  Result := Format('acceptDimList(%s, %d, "%s")', [AcceptList, Dim, Msg]);
end;

procedure TQUB_PropertiesForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Word(Key) = VK_ESCAPE then begin
    Key := #0;
    ModalResult := mrCancel;
  end;
end;


//******************************************************************************
end.
