{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2002
// thread which processes a queue of python requests, usually stuff typed in Viewport_Python.
// can also interrupt python and get/release the Global Interpreter Lock


unit QUB_Python_Thread;


interface


uses


SysUtils, Classes, SyncObjs, Windows, Forms,
PythonEngine;

// requests (interpret, evaluate) from other threads are enqueued then executed in order


const

DEF_TIMEOUT = 3000;


type


TQUB_Proc_Method = procedure of object;

// something that is part of the flow of a delphi/python object, but must be synchronized,
// can be passed to the PyThread for synchronization
TQUB_Python_SyncTask = class(TObject)
public

  procedure Execute; virtual; abstract;

end;

TQUB_Python_SyncProc = class(TQUB_Python_SyncTask)
protected
  prProc         :TQUB_Proc_Method;
public
  constructor Create(proc :TQUB_Proc_Method);

  procedure Execute; override;
end;


TQUB_Python_RequestMode = (QP_INTERPRET, QP_EVALUATE, QP_STOP);

TQUB_Python_Request = class
  // these three
  Mode              :TQUB_Python_RequestMode;
  Command           :String;
  Silent            :Boolean;
  Output            :PPyObject; // evaluation result
  // or this one
  SyncTask          :TQUB_Python_SyncTask;    // if this ain't nil, ignore Command, Silent, Output
  // synchronous if notify ain't nil
  Notify            :TEvent; // if nil: request is autodeleted, else deleted by notified party (e.g. in evaluate)
  Done              :Boolean;//True after the python thread finishes with it
end;

TQUB_Python_RequestQueue = class
protected

  prQ               :TList;
  prLock            :TCriticalSection;
  prRqAvail         :TEvent; // wakes up Accept()

public

  constructor Create;
  destructor Destroy; override;

  procedure Add(Request :TQUB_Python_Request);
  function Rescind(Request :TQUB_Python_Request): Boolean; // false if it wasn't in the queue (i.e. is running or done)
  function Accept(TimeoutMS :Integer): TQUB_Python_Request; // or nil if timed out

end;


TQUB_Python_Thread = class(TThread)
protected


  // properties
  prEngine              :TPythonEngine;
  prOnSendData          :TSendDataEvent;

  // request queue and notification
  prRequestQ            :TQUB_Python_RequestQueue;

  prInterruptThreads    :TList; // of PPyThreadState
  prPythonThreadRunning :Boolean;

  // synchronized output string
  prOutput              :String;

  //procedure prSyncOutput;

  //procedure prSyncGUITask;

  procedure prInitialize;

  procedure Execute; override;


public
  PythonReady           :Boolean;

  constructor Create(CreateSuspended :Boolean);
  destructor  Destroy; override;

  function RequestTerminate(TimeoutMS :Integer): Boolean;


  function Interpret(cmd :String; var PyOutput :PPyObject; Mode :TQUB_Python_RequestMode; Asynch :Boolean; Silent :Boolean; TimeOut :Integer) :Boolean;
  procedure InterpretFromEvent(Sender: TObject; const Data : String);
  procedure InterpretSilently(Data : String);
  procedure InterpretNonSilently(Data : String);

  // run something from the Python thread in the gui thread
  // (typically from a delphi object called from inside python)
  //procedure SyncGUITask(task :TQUB_Python_SyncTask);
  // run something from another thread in the python thread
  // (typically initialization or conversion)
  // if it's Asynch (return before it finishes or maybe even starts):
  //   task will be auto-freed after running
  // else:
  //   you're responsible for freeing it
  function SyncPyTask(task :TQUB_Python_SyncTask; Asynch: Boolean; Silent :Boolean; TimeOut :Integer): Boolean;
  // wrappers for method pointers
  procedure SyncGUIProc(proc: TQUB_Proc_Method);    // ya might as well call PythonViewport.SyncGUIProc since that's what this one does now
  function SyncPyProc(proc: TQUB_Proc_Method; Asynch: Boolean; Silent :Boolean; TimeOut :Integer): Boolean;

  procedure SneakPyProc(proc :TQUB_Proc_Method); // get the GIL and run proc in the current thread
  // procedure DoneSneaking;                        // clean up thread-specific sneaky storage

  procedure BeginPython;       // get the GIL
  procedure EndPython;         // release the GIL
  procedure DonePython;        // clean up per-thread storage of thread state


  procedure EnableInterruptsThisThread;
  procedure DisableInterruptsThisThread;
  procedure DoInterrupt;


  // IO.OnSendData; synchronize: redirect to property OnSendData
  procedure PythonPrint(Sender: TObject; const Data : String );

  property Engine          :TPythonEngine      read prEngine          write prEngine;

  // OnSendData: e.g. PythonReadEvalPrint.PythonPrint
  property OnSendData      :TSendDataEvent     read prOnSendData      write prOnSendData;

end;


TQUB_RunInterruptible = class(TPyObject)
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;


var
  guiPyThreadState                :PPyThreadState;
  pythonPyThreadState             :PPyThreadState;
  QUB_Python_Interrupt            :PPyObject;
  QUB_RunInterruptiblePyType      :TPythonType;


threadvar
  myPyThreadState                 :PPyThreadState;
// this should be valid for the event thread (creator of python engine)
// and the python thread (sets up thread state at beginning of Execute)
// as well as any thread that uses PythonThread.Begin/EndPython or SneakPyProc,
// and any thread that messes with it directly.

// stuff called from python could be on a python-created thread,
// so you're better off using PyThreadState_Get each time (see the bottom of QUB_Interface_Modeling for examples of yielding the GIL during an intensive task)

// from   http://www.linuxjournal.com/article.php?sid=3641
// Be certain to delete all your thread-state objects with PyThreadState_Clear and PyThreadState_Delete before calling Py_Finalize.

// For the most part, it's just the python and event threads that have thread state,
// and I don't really care if they get cleaned up prior to finalization.
// After all, it seems to work just fine without it.
// However, if you're using SneakPyProc from QUB_Task threads, that could turn into a substantial leak.
// Make sure you call PythonThread.DonePython before the task exits.


implementation


uses

QUB_Main,
QUB_GUIInterface,
//QUB_Interface_Simulation,
QUB_Node_Model,
qubopt_iface,
QUB_Task_Fit_Dlog,
QUB_Python_Wizard,
QUB_Baseline,
QUB_Viewport_Python,
QUB_Node_Python,
QUB_Python_Prompt,
QUB_Python_ModalDialog,
QUB_Task_Python_Properties,
QUB_Profiles_Python,
QUB_Python_Eigen,
QUBFits;



constructor TQUB_Python_RequestQueue.Create;
begin
  prQ          := TList.Create;
  prLock       := TCriticalSection.Create;
  prRqAvail    := TEvent.Create(nil{security}, False{auto-reset}, False{init:off}, ''{unique});
end;

destructor TQUB_Python_RequestQueue.Destroy;
begin
  try
    prQ.Free;
    prLock.Free;
    prRqAvail.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Python_RequestQueue.Add(Request :TQUB_Python_Request);
begin
  prLock.Acquire;
  prQ.Add(Request);
  prLock.Release;
  prRqAvail.SetEvent;
end;

function TQUB_Python_RequestQueue.Rescind(Request :TQUB_Python_Request): Boolean; // false if it wasn't in the queue (i.e. is running or done)
begin
  prLock.Acquire;
  Result := prQ.Remove(Request) <> -1;
  prLock.Release;
end;

function TQUB_Python_RequestQueue.Accept(TimeoutMS :Integer): TQUB_Python_Request; // or nil if timed out
begin
  Result := nil;

  prLock.Acquire;
  if prQ.Count > 0 then begin
    Result := TQUB_Python_Request( prQ[0] );
    prQ.Delete(0);
  end;
  prLock.Release;

  if Result = nil then begin // Q was empty
    prRqAvail.WaitFor( TimeoutMS );
    prLock.Acquire;
    if prQ.Count > 0 then begin
      Result := TQUB_Python_Request( prQ[0] );
      prQ.Delete(0);
    end;
    prLock.Release;
  end;

  prRqAvail.ResetEvent;
end;


//******************************************************************************
constructor TQUB_Python_Thread.Create(CreateSuspended :Boolean);
begin
  inherited Create(CreateSuspended);

  prRequestQ      := TQUB_Python_RequestQueue.Create;
  prInterruptThreads := TList.Create;
  prPythonThreadRunning := False;
  PythonReady           := False;
end;

destructor  TQUB_Python_Thread.Destroy;
begin
  prRequestQ.Free;
  prInterruptThreads.Free;

  inherited;
end;

function TQUB_Python_Thread.RequestTerminate(TimeoutMS :Integer): Boolean;
var
  StopRequest  :TQUB_Python_Request;
begin
  Result                 := False;

  StopRequest            := TQUB_Python_Request.Create;
  StopRequest.Mode       := QP_STOP;
  StopRequest.SyncTask   := nil;
  StopRequest.Notify     := TEvent.Create(nil{security}, True{manual reset}, False{init: off}, ''{unique});

  prRequestQ.Add( StopRequest );

  if TimeOutMS < 0 then begin
    StopRequest.Notify.WaitFor( INFINITE );
  end else begin
    StopRequest.Notify.WaitFor( TimeOutMS );
  end;

  if prRequestQ.Rescind( StopRequest ) then begin // successfully rescinded
    Result := True;
  end else begin // running or done
    while not StopRequest.Done do begin
      StopRequest.Notify.WaitFor(1000);
    end;
  end;

  StopRequest.Notify.Free;
  StopRequest.Free;
end;

procedure TQUB_Python_Thread.prInitialize;
var
  pubpath                       :String;
  libpath                       :String;
begin
  with GetPythonEngine do begin
    // put pythonscripts first
    // (so it loads qubtree.py rather than qubtree.dll)
    libpath := ExtractFileDir(Application.ExeName);
    pubpath := Copy(libpath, 0, 2){drive letter} + '\\Users\\Public';
    Run_CommandAsString('sys.path.insert(1,''' + pubpath + '\\QUB\\PythonScripts'')', single_input);
    Run_CommandAsString('sys.path.insert(2,''' + libpath + '\\PythonScripts'')', single_input);
    Run_CommandAsString('sys.path.insert(3, ''' + libpath + '\\site-packages'')', single_input);
    Run_CommandAsString('sys.path.insert(4, ''' + libpath + '\\site-packages\\win32'')', single_input);
    Run_CommandAsString('sys.path.insert(5, ''' + libpath + '\\site-packages\\win32\\lib'')', single_input);

    Run_CommandAsString('import pywintypes', single_input); // important for starting via dbl-click document

    PyNode_InitCallbacks;
    qubopt_pyinit;
    InitProfiles_Py;
    InitQUBFits(ExtractFileDir(Application.ExeName) + '\Curve Plugins\');

    Run_CommandAsString('from math import *', single_input);
    Run_CommandAsString('import random', single_input);
    Run_CommandAsString('import qubtree', single_input);
    Run_CommandAsString('import QUB', single_input);   // delphi objects
    Run_CommandAsString('import QUBTypes', single_input);
    Run_CommandAsString('QUB.ReportView = QUBTypes.CreateReportViewport()', single_input);
    Run_CommandAsString('ReportView = QUB.ReportView', single_input);
    Run_CommandAsString('ReportNode = lambda node: QUB.ReportView.Write( str( node ) )', single_input);
    Run_CommandAsString('QUB.Report = lambda node: QUB.ReportView.Write( str( node ) )', single_input);
    try
      Run_CommandAsString('import QUBB', single_input);    // qubfolder/QUBB.py
      Run_CommandAsString('from QUBB import *', single_input);
    except
    end;
    try
      Run_CommandAsString('import Commands', single_input);
      Run_CommandAsString('import qubopt', single_input);
      Run_CommandAsString('import CurveFitting', single_input); // so QUB_Task_Fit_Dlog can easily look up the fitting modules
      Run_CommandAsString('import Reports', single_input); // so QUB_Reports_* can easily interact with Python layer
      Run_CommandAsString('import AcquisitionProtocol', single_input); // so protocol can access it
      InitModelScripts;
      InitCurveFits;
      InitPythonPrompt;
      InitPythonModalDialog;
      InitPythonTaskProperties;
    // PyNode_PublishNodeFunc( TNodeFunc.Create(TestMplexIdl), 'MplexIdl' );
      QUB_Python_Wizard_Init;
      QUB_Baseline_Py_Init;
      //Eigen_Scipy_Init;
    except
    end;
  end;
end;


procedure TQUB_Python_Thread.Execute;
var
  Request        :TQUB_Python_Request;
  NotStopped     :Boolean;
  RunInterruptible :PPyObject;
begin

  BeginPython;

  pythonPyThreadState := myPyThreadState;
  QUB_Python_Interrupt := GetPythonEngine.PyErr_NewException('QUB.GraphicalInterrupt', nil, nil);
  // should now add it to module QUB?

  QUB_RunInterruptiblePyType := TPythonType.Create(nil);
  QUB_RunInterruptiblePyType.Engine := GetPythonEngine;
  QUB_RunInterruptiblePyType.Module := PythonViewport.pmQUBTypes;
  QUB_RunInterruptiblePyType.TypeName := 'RunInterruptibleType';
  QUB_RunInterruptiblePyType.Services.Basic := [bsCall];
  QUB_RunInterruptiblePyType.PyObjectClass := TQUB_RunInterruptible;
  QUB_RunInterruptiblePyType.Initialize;

  RunInterruptible := QUB_RunInterruptiblePyType.CreateInstance;
  PyNode_PublishPyObj( RunInterruptible, 'RunInterruptible' );
  GetPythonEngine.Py_DECREF( RunInterruptible );

  prInitialize;

  EndPython;
  PythonReady := True;

  NotStopped := True;
  while NotStopped do begin

    Request := prRequestQ.Accept(10000);

    if Request = nil then Continue;

    BeginPython;

    Engine.RedirectIO := not Request.Silent;

    prPythonThreadRunning := True;
    try
      if Request.SyncTask <> nil then begin
        Request.SyncTask.Execute;
      end else if Request.Mode = QP_INTERPRET then begin
        prEngine.Run_CommandAsString(Request.Command, single_input);
      end else if Request.Mode = QP_EVALUATE then begin
        Request.Output := prEngine.EvalString(Request.Command);
      end else if Request.Mode = QP_STOP then begin
        NotStopped := False;
      end;
    except
    end;
    prPythonThreadRunning := False;

    EndPython;

    if Request.Notify <> nil then begin
      Request.Done := True;
      Request.Notify.SetEvent;
    end else begin
      try
        Request.SyncTask.Free;
        Request.Free;
      except
      end;
    end;

  end;

  //DonePython;     -- python interpreter may already be cleaned up, and qub is closing so who cares anyway

  // after stop, clear out the q (lie and say they're all done)
  while True do begin
    Request := prRequestQ.Accept(1);
    if Request = nil then Break;

    Request.Done := True;
    if Request.Notify <> nil then begin
      Request.Notify.SetEvent;
    end else begin
      try
        Request.SyncTask.Free;
        Request.Free;
      except
      end;
    end;
  end;

end;

//Interpret
{
  The caller want it to start running, asynchronously or not, after at most TimeOut, else Error
  Returns False for Succes and True for Error
  if Mode = Evaluate then PyOutput returns Request.Output
}
function TQUB_Python_Thread.Interpret(cmd :String; var PyOutput :PPyObject; Mode :TQUB_Python_RequestMode; Asynch :Boolean; Silent :Boolean; TimeOut :Integer) :Boolean;
var
  Request          :TQUB_Python_Request;
begin
  Result           := False;

  Request          := TQUB_Python_Request.Create;
  Request.Mode     := Mode;
  Request.Command  := cmd;
  Request.Silent   := Silent;
  Request.Output   := nil;
  Request.SyncTask := nil;
  Request.Notify   := nil;
  Request.Done     := False;

  if not Asynch then begin
    Request.Notify := TEvent.Create(nil{security}, True{manual reset}, False{init: off}, ''{unique});
  end;

  prRequestQ.Add( Request );

  if not Asynch then begin
    if TimeOut < 0 then begin
      Request.Notify.WaitFor( INFINITE );
    end else begin
      Request.Notify.WaitFor( TimeOut );
    end;

    if prRequestQ.Rescind( Request ) then begin // successfully rescinded
      Result := True;
    end else begin // running or done
      while not Request.Done do begin
        Request.Notify.WaitFor(1000);
      end;
    end;

    PyOutput := Request.Output;
    Request.Notify.Free;
    Request.Free;
  end;
end;

procedure TQUB_Python_Thread.InterpretFromEvent(Sender :TObject; const Data :String);
var
  p :PPyObject;
begin
  Interpret(Data, p, QP_INTERPRET, True{Asynch}, False{Silent}, DEF_TIMEOUT);
end;


procedure TQUB_Python_Thread.InterpretSilently(Data :String);
var
  p :PPyObject;
begin
  Interpret(Data, p, QP_INTERPRET, True{Asynch}, True{Silent}, DEF_TIMEOUT);
end;

procedure TQUB_Python_Thread.InterpretNonSilently(Data :String);
var
  p :PPyObject;
begin
  Interpret(Data, p, QP_INTERPRET, True{Asynch}, False{Silent}, DEF_TIMEOUT);
end;


//******************************************************************************
type
TQUB_Python_Print = class(TQUB_Runnable)
public
  Data      :String;
  constructor Create(AData :String);
  procedure Run; override;
end;

constructor TQUB_Python_Print.Create(AData :String);
begin
  Data := AData;
end;

procedure TQUB_Python_Print.Run;
begin
  PythonThread.OnSendData(PythonThread, Data);
end;

procedure TQUB_Python_Thread.PythonPrint(Sender: TObject; const Data : String );
begin
  Asynchronize(TQUB_Python_Print.Create(Data));
  {with GetPythonEngine do begin
    EndPython;

    prOutput := Data;
    QUB_Main.SyncGUIProc(prSyncOutput);

    BeginPython;
  end;}
end;

{procedure TQUB_Python_Thread.prSyncOutput;
begin
  try
    prOnSendData(Self, prOutput);
  except
  end;
end;}

procedure TQUB_Python_Thread.SyncGUIProc(proc :TQUB_Proc_Method);
begin
  EndPython;
  QUB_MainForm.DoSyncGUIProc(proc);
  BeginPython;
end;

function TQUB_Python_Thread.SyncPyTask(task :TQUB_Python_SyncTask; Asynch :Boolean; Silent :Boolean; TimeOut :Integer): Boolean;
var
  Request          :TQUB_Python_Request;
begin
  Result           := False;

  Request          := TQUB_Python_Request.Create;
  Request.Output   := nil;
  Request.SyncTask := task;
  Request.Notify   := nil;
  Request.Done     := False;
  Request.Silent   := Silent;

  if not Asynch then begin
    Request.Notify := TEvent.Create(nil{security}, True{manual reset}, False{init: off}, ''{unique});
  end;

  prRequestQ.Add( Request );

  if not Asynch then begin
    if TimeOut < 0 then begin
      while wrTimeout = Request.Notify.WaitFor(100) do Application.ProcessMessages;
      //Request.Notify.WaitFor( INFINITE );
    end else begin
      Request.Notify.WaitFor( TimeOut );
    end;

    if prRequestQ.Rescind( Request ) then begin // successfully rescinded
      Result := True;
    end else begin // running or done
      while not Request.Done do begin
        Request.Notify.WaitFor(1000);
      end;
    end;

    Request.Notify.Free;
    Request.Free;
  end;

end;

function TQUB_Python_Thread.SyncPyProc(proc :TQUB_Proc_Method; Asynch :Boolean; Silent :Boolean; TimeOut :Integer): Boolean;
var
  Task        :TQUB_Python_SyncProc;
begin
  Task        := TQUB_Python_SyncProc.Create( proc );
  Result      := SyncPyTask( Task, Asynch, Silent, TimeOut );
  if not Asynch then Task.Free;
end;

procedure TQUB_Python_Thread.SneakPyProc(proc :TQUB_Proc_Method); // get the GIL and run proc in the current thread
begin
  BeginPython;
  try
    proc;
  finally
    EndPython;
  end;
end;

procedure TQUB_Python_Thread.BeginPython;
begin
  while guiPyThreadState = nil do Sleep(300);
  with GetPythonEngine do begin
    PyEval_AcquireLock;

    if myPyThreadState = nil then begin
      myPyThreadState := PyThreadState_New( guiPyThreadState.interp );
    end;

    PyThreadState_Swap( myPyThreadState );
  end;
end;

procedure TQUB_Python_Thread.EndPython;
begin
  with GetPythonEngine do begin
    if myPyThreadState = nil then begin
      myPyThreadState := PyThreadState_Get;
    end;

    PyThreadState_Swap( nil );
    PyEval_ReleaseLock;
  end;
end;

procedure TQUB_Python_Thread.DonePython;
begin
  if myPyThreadState = nil then Exit; // nothing to clean up
  if myPyThreadState = guiPyThreadState then Exit; // you really want to keep this one (e.g. to instantiate new thread states)

  try
    with GetPythonEngine do begin
      PyEval_AcquireLock();
      PyThreadState_Swap(nil);
      PyThreadState_Clear(myPyThreadState);
      PyThreadState_Delete(myPyThreadState);
      myPyThreadState := nil;
      PyEval_ReleaseLock();
    end;
  except
    // the final DonePython sometimes falls after the engine is destroyed
  end;
end;

procedure TQUB_Python_Thread.EnableInterruptsThisThread;
var
  i           :Integer;
begin
  if myPyThreadState = nil then begin
    myPyThreadState := GetPythonEngine.PyThreadState_Get;
  end;
  if myPyThreadState = guiPyThreadState then Exit;
  for i := 0 to prInterruptThreads.Count - 1 do begin
    if myPyThreadState = prInterruptThreads[i] then Exit;
  end;

  prInterruptThreads.Add(myPyThreadState);
end;

procedure TQUB_Python_Thread.DisableInterruptsThisThread;
var
  i           :Integer;
begin
  if myPyThreadState = nil then begin
    myPyThreadState := GetPythonEngine.PyThreadState_Get;
  end;
  if myPyThreadState = guiPyThreadState then Exit;
  for i := 0 to prInterruptThreads.Count - 1 do begin
    if myPyThreadState = prInterruptThreads[i] then begin
      prInterruptThreads.Delete(i);
      Exit;
    end;
  end;
end;

procedure TQUB_Python_Thread.DoInterrupt;
var
  i               :Integer;
begin
  BeginPython;
  with GetPythonEngine do begin
    if prPythonThreadRunning then {i :=} PyThreadState_SetAsyncExc(pythonPyThreadState.thread_id, QUB_Python_Interrupt);
    //dbg to check rtn value// if i <> 1 then begin EndPython; Exit; end;
    for i := 0 to prInterruptThreads.Count - 1 do begin
      if 1 < PyThreadState_SetAsyncExc(PPyThreadState(prInterruptThreads[i]).thread_id, QUB_Python_Interrupt)
      then   PyThreadState_SetAsyncExc(PPyThreadState(prInterruptThreads[i]).thread_id, nil);
    end;
  end;
  EndPython;
end;

constructor TQUB_Python_SyncProc.Create(proc :TQUB_Proc_Method);
begin
  prProc := proc;
end;

procedure TQUB_Python_SyncProc.Execute;
begin
  prProc;
end;

function TQUB_RunInterruptible.Call( args, kw : PPyObject) : PPyObject;
var
  pyfunc   :PPyObject;
  args2     :PPyObject;
begin
  with GetPythonEngine do begin
    if ( PyArg_ParseTuple( args, 'O:RunInterruptible', [@pyfunc] ) = 0 ) then begin
      Result := nil
    end else begin
      if PyCallable_Check(pyfunc) = 0 then begin
        PyErr_SetString( PyExc_TypeError^, PChar('RunInterruptible: argument must be callable') );
        Result := nil;
      end else begin
        PythonThread.EnableInterruptsThisThread;
        args2 := PyTuple_New(0);
        Result := PyObject_CallObject( pyfunc, args2 );
        Py_DECREF( args2 );
        PythonThread.DisableInterruptsThisThread;
      end;
    end;
  end;
end;


//******************************************************************************
end.
