{*******************************************************}
{                                                       }
{                   Qub File System                     }
{           Copyright (C) 2000 Lorin Milescu            }
{                  Created 01-31-2000                   }
{                                                       }
{*******************************************************}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


unit QUB_QFS;


//******************************************************************************
interface


uses


SysUtils, Classes,

AllocTypes, Alloc,

QUB_Global,

QUB_Node;


//******************************************************************************
const


QUB_QFS_MAGIC                                  = 'QUB_(;-)_QFS';
QUB_QFS_MAGIC_LEN                              = Length(QUB_QFS_MAGIC);
QUB_QFS_RECORDS_MAX                            = 65535;


//******************************************************************************

//The new Node Structures
{$Z1} //1-byte enumerated data type
{ moved to QUB_Node
type


TQUB_QFS_Node_DataType = (NODE_DATATYPE_EMPTY,    NODE_DATATYPE_UNKNOWN,
                          NODE_DATATYPE_POINTER,
                          NODE_DATATYPE_STRING,   //DataSize = 1; for WideString is 2, but we don't check it -
                          NODE_DATATYPE_BYTE,     NODE_DATATYPE_SHORTINT,
                          NODE_DATATYPE_WORD,     NODE_DATATYPE_SMALLINT,
                          NODE_DATATYPE_LONGWORD, NODE_DATATYPE_LONGINT,
                          NODE_DATATYPE_UINT64,   NODE_DATATYPE_INT64,
                          NODE_DATATYPE_SINGLE,   NODE_DATATYPE_DOUBLE,   NODE_DATATYPE_EXTENDED
                         );

TQUB_QFS_Node_Flags    = Word;

TQUB_QFS_NodeStream    = record

  Flags                :TQUB_QFS_Node_Flags;
  Flags2               :Byte;

  DataType             :TQUB_QFS_Node_DataType;
  DataSize             :LongInt;
  DataCount            :LongInt;
  DataStreamPos        :LongInt; //  0 should mean NONE

  ChildStreamPos       :LongInt; //       -,,-
  SiblingStreamPos     :LongInt; //       -,,-

  Reserved             :Array[0..6] of Byte;

  NameLen              :Byte;

end;
TQUB_QFS_PNodeStream   = ^TQUB_QFS_NodeStream;

TQUB_QFS_PNodeMem      = ^TQUB_QFS_NodeMem;
TQUB_QFS_NodeMem       = record

  NodeStream           :TQUB_QFS_NodeStream;

  RefCount             :Integer;

  Name                 :PChar;

  ParentNode           :TQUB_QFS_PNodeMem;
  ChildNode            :TQUB_QFS_PNodeMem;
  SiblingNode          :TQUB_QFS_PNodeMem;

  DataLoadSel          :TQUB_FirstLast;
  Data                 :Pointer;

  Stream               :TStream;
  StreamPos            :LongInt;

  Mutex                :Pointer;
  ReaderCount          :Integer;

end;


const


//Flags ************************************************************************
//On Stream
NODE_FLAG_UNSIGNED                              = 1;
NODE_FLAG_PRELOAD                               = 2;
NODE_FLAG_DATA_IN_NODE                          = 4;
NODE_FLAG_MATRIX                                = 8;
NODE_FLAG_NONWRITABLE                           = 16;

//In Memory only
NODE_FLAG_CHANGED                               = $1000;
NODE_FLAG_CHILDCHANGED                          = $2000;


//Flag Masks *******************************************************************
NODE_FLAG_ALL_MASK                              = High(TQUB_QFS_Node_Flags);
NODE_FLAG_UNSIGNED_MASK                         = NODE_FLAG_ALL_MASK - NODE_FLAG_UNSIGNED;
NODE_FLAG_PRELOAD_MASK                          = NODE_FLAG_ALL_MASK - NODE_FLAG_PRELOAD;
NODE_FLAG_DATA_IN_NODE_MASK                     = NODE_FLAG_ALL_MASK - NODE_FLAG_DATA_IN_NODE;
NODE_FLAG_MATRIX_MASK                           = NODE_FLAG_ALL_MASK - NODE_FLAG_MATRIX;
NODE_FLAG_CHANGED_MASK                          = NODE_FLAG_ALL_MASK - NODE_FLAG_CHANGED;
NODE_FLAG_CHILDCHANGED_MASK                     = NODE_FLAG_ALL_MASK - NODE_FLAG_CHILDCHANGED;
NODE_FLAG_NONWRITABLE_MASK                      = NODE_FLAG_ALL_MASK - NODE_FLAG_NONWRITABLE;


//Default flags ****************************************************************
NODE_FLAGS_DEFAULT                              = NODE_FLAG_PRELOAD or NODE_FLAG_CHANGED;
NODE_FLAGS_MEMONLY                              = NODE_FLAG_CHANGED or NODE_FLAG_CHILDCHANGED;
}


type


//Forward class declarations ***************************************************
TQUB_QFS_Record                                = class;


//Object
TQUB_QFS_Record                                = class(TObject)
protected


  //What is read/written from/to the Stream
  prNodeMem                                    :TQUB_QFS_NodeMem;     //as a pointer  ???

  prName                                       :String;

  prParent                                     :TQUB_QFS_Record;

  prRecords                                    :TStringList;
  prRecordCount                                :Integer;


  //Helper flags
  prPermanent                                  :Boolean;
  prModified                                   :Boolean;

  prReading                                    :Boolean;
  prWriting                                    :Boolean;


  function prGetRecord(Index :Integer) :TQUB_QFS_Record;
  function prGetRecordByName(const Index :String) :TQUB_QFS_Record;

  procedure prSetDataType(t :TQUB_QFS_Node_DataType); virtual;
  procedure prSetDataSize(s :LongInt); virtual;
  procedure prSetDataCount(c :LongInt); virtual;
  procedure prSetData(d :Pointer); virtual;


  function prGetDataValue :Variant; virtual;
  procedure prSetDataValue(v :Variant); virtual;
  function prGetDataArrayValue(Index :Integer) :Variant; virtual;
  procedure prSetDataArrayValue(Index :Integer; v :Variant); virtual;

  function prGetFlagUnsigned :Boolean;
  function prGetFlagPreload :Boolean;
  procedure prSetFlagPreload(f :Boolean);
  function prGetFlagDataInNode :Boolean;
  procedure prSetFlagDataInNode(f :Boolean);
  function prGetFlagNonWritable :Boolean;
  procedure prSetFlagNonWritable(f :Boolean);

  procedure prIncDataSizeCount(NewDataSize, NewDataCount :Integer);

  function prWaitToBeReady(MSecs :Integer) :Boolean;
  //Replace it with Mutex or similar ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); virtual;
  constructor CreateAsValue(const Name :String; Parent :TQUB_QFS_Record; Permanent :Boolean; Value :Variant; ValueType :TQUB_QFS_Node_DataType; ValueCount :Integer); virtual;
  destructor Destroy; override;


  procedure Reset; virtual;
  procedure ClearToFree; virtual;


  function AddRecord(r :TQUB_QFS_Record) :Boolean; overload; virtual;
  function AddRecord(r :TQUB_QFS_Record; Index :Integer) :Boolean; overload; virtual;
  function AddRecordFromNode(n :INode): Boolean;
  function DelRecord(r :TQUB_QFS_Record) :Boolean; virtual;
  procedure RemoveFromParent; virtual;
  function RecordIndex(r :TQUB_QFS_Record) :Integer;

  procedure Swap(i, j :Integer);

  function Read(ReadStream :TStream) :Boolean; virtual;
  function Write(WriteStream :TStream) :Boolean; virtual;

  function ReadFromNode(n :INode) :Boolean; virtual;    // copy qubtree

  function CloneAsNode( Deep :Boolean; AppendTo :INode ): INode;

  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; virtual;

  function SizeOnStream(Recursive :Boolean = True) :LongInt;


  function PreLoadData :Boolean; virtual;

  function LoadData(First, Last :Integer) :Boolean; virtual;
  function LoadDataInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean; virtual;

  function UnloadData :Boolean; virtual;
  function ClearData :Boolean; virtual;


  property Parent                              :TQUB_QFS_Record           read prParent;

  property Name                                :String                    read prName;

  property Stream                              :TStream                   read prNodeMem.Stream                       write prNodeMem.Stream;
  property StreamPos                           :LongInt                   read prNodeMem.StreamPos                    write prNodeMem.StreamPos;

  property Permanent                           :Boolean                   read prPermanent                            write prPermanent;
  property Modified                            :Boolean                   read prModified;

  property Records[Index :Integer]             :TQUB_QFS_Record           read prGetRecord;                           default;
  property RecordsByName[const Index :String]  :TQUB_QFS_Record           read prGetRecordByName;
  property RecordCount                         :Integer                   read prRecordCount;

  //Flags
  property Flags                               :TQUB_QFS_Node_Flags       read prNodeMem.NodeStream.Flags             write prNodeMem.NodeStream.Flags;
  //Helper Flags
  property FlagUnsigned                        :Boolean                   read prGetFlagUnsigned;
  property FlagPreload                         :Boolean                   read prGetFlagPreload                       write prSetFlagPreload;
  property FlagDataInNode                      :Boolean                   read prGetFlagDataInNode                    write prSetFlagDataInNode;
  property FlagNonWritable                     :Boolean                   read prGetFlagNonWritable                   write prSetFlagNonWritable;

  //Data Descriptor
  property DataType                            :TQUB_QFS_Node_DataType    read prNodeMem.NodeStream.DataType          write prSetDataType;
  property DataSize                            :LongInt                   read prNodeMem.NodeStream.DataSize          write prSetDataSize;
  property DataCount                           :LongInt                   read prNodeMem.NodeStream.DataCount         write prSetDataCount;
  property DataStreamPos                       :LongInt                   read prNodeMem.NodeStream.DataStreamPos     write prNodeMem.NodeStream.DataStreamPos;
  property DataLoadSel                         :TQUB_FirstLast            read prNodeMem.DataLoadSel                  write prNodeMem.DataLoadSel;
  property Data                                :Pointer                   read prNodeMem.Data                         write prSetData;


  property DataValue                           :Variant                   read prGetDataValue                         write prSetDataValue;
  property DataArrayValue[Index :Integer]      :Variant                   read prGetDataArrayValue                    write prSetDataArrayValue;


  property SiblingStreamPos                    :LongInt                   read prNodeMem.NodeStream.SiblingStreamPos;


end;
//..............................................................................



//******************************************************************************
implementation


{$IFDEF DELPHI6_OR_HIGHER}
uses

Variants;
{$ENDIF}


//****************************************************************************//
//                                                                            //
//                              TQUB_QFS_Record                               //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_Record.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited Create;

  //Change if dynamic nodes
  prNodeMem.RefCount := 1;

  //Initialize the record
  with prNodeMem.NodeStream do begin
    DataStreamPos    := 0;
    ChildStreamPos   := 0;
    SiblingStreamPos := 0;
  end;

  prParent := Parent;
  if not (prParent is TQUB_QFS_Record) then begin
    prParent := nil;
  end;
  prName := Name;
  prNodeMem.NodeStream.NameLen := Length(Name);
  prNodeMem.Name := PChar(prName);

  prNodeMem.DataLoadSel := QUB_FirstLast(-1, -1);

  Reset;

  if prParent <> nil then begin
    prParent.AddRecord(Self);
  end;

end;

constructor TQUB_QFS_Record.CreateAsValue(const Name :String; Parent :TQUB_QFS_Record; Permanent :Boolean; Value :Variant; ValueType :TQUB_QFS_Node_DataType; ValueCount :Integer);
begin
  Create(Name, Parent);

  Flags          := NODE_FLAG_PRELOAD;
  Self.Permanent := Permanent;
  DataCount      := ValueCount;
  DataType       := ValueType;//this will allocate memory, if PRELOAD
  DataValue      := Value;

end;


destructor TQUB_QFS_Record.Destroy;
begin
  try

    RemoveFromParent;

    ClearToFree;

  finally

    inherited;

  end;

end;

procedure TQUB_QFS_Record.RemoveFromParent;
begin
  if prParent <> nil then begin
    prParent.DelRecord(Self);
    prParent := nil;
  end;
end;

procedure TQUB_QFS_Record.Reset;
var
  i :Integer;
begin
  if prRecords <> nil then begin
    try
      for i := prRecords.Count - 1 downto 0 do begin
        if not Records[i].Permanent then begin
          prRecords.Objects[i].Free;
        end else begin
          Records[i].Reset;
        end;
      end;
    finally
      if RecordCount = 0 then begin
        prRecords.Free;
        prRecords := nil;
      end;
    end;
  end;

  UnloadData;

  prModified   := False;

end;

procedure TQUB_QFS_Record.ClearToFree;
var
  i :Integer;
begin
  if prRecords <> nil then begin
    try
      for i := prRecords.Count - 1 downto 0 do begin
        prRecords.Objects[i].Free;
      end;
    finally
      prRecords.Free;
      prRecords := nil;
    end;
  end;

  ClearData;

  prModified   := False;

end;

//******************************************************************************
procedure CopyQFSDataToNode( qfs: TQUB_QFS_Record; node: INode );
var
  ldCount, rem              :Integer;
  ldBounds                  :TQUB_FirstLast;
  dCount                    :Integer;
  LastChar                  :Char;
begin
  node.Preload := qfs.FlagPreload or (qfs.DataCount <= 0);

  dCount                    := qfs.DataCount;
  if qfs.DataType = NODE_DATATYPE_STRING then begin
    try
      qfs.LoadDataInBuffer(@LastChar, dCount-1, dCount-1);
      if LastChar = #0 then Dec(dCount);
    except
      Dec(dCount);
    end;
  end;

  if dCount > 0 then begin

    node.SetupData( qfs.DataType, qfs.DataSize, dCount );

    ldCount             := QUB_NODE_MAX_LOAD_FOR_COPY;
    ldBounds.First      := 0;
    rem                 := dCount;

    while rem > 0 do begin
      if rem < ldCount then ldCount := rem;
      ldBounds.Last := ldBounds.First + ldCount - 1;
      node.LoadRows( ldBounds, False );
      qfs.LoadDataInBuffer( node.DataPtr, ldBounds.First, ldBounds.Last );

      ldBounds.First := ldBounds.First + ldCount;
      rem := rem - ldCount;
    end;

    if node.Preload then begin
      ldBounds.First := 0;
      ldBounds.Last := node.Rows - 1;
      node.LoadRows( ldBounds, True );
    end else begin
      node.UnloadRows( True );
    end;

  end;
end;

function TQUB_QFS_Record.CloneAsNode( Deep :Boolean; AppendTo :INode ): INode;
var
  i          :Integer;
begin
  if AppendTo.IsNull then begin
    Result := TNode.Create( Name );
  end else begin
    Result := AppendTo.AppendChild( Name );
  end;
  
  CopyQFSDataToNode( Self, Result );

  if Deep then begin
    for i := 0 to RecordCount - 1 do begin
      Records[i].CloneAsNode( Deep, Result );
    end;
  end;
end;


//******************************************************************************
function TQUB_QFS_Record.AddRecord(r :TQUB_QFS_Record) :Boolean;
var
  PrevRec :TQUB_QFS_Record;
begin
  Result := False;
  if prRecords = nil then begin
    prRecords := TStringList.Create;
  end;
  PrevRec := Records[RecordCount - 1];
  prRecords.AddObject('', r);
  if PrevRec <> nil then begin
    PrevRec.prNodeMem.SiblingNode := @(r.prNodeMem);
  end;
  r.prParent := Self;
  r.prNodeMem.ParentNode := @prNodeMem;
  prRecordCount := prRecords.Count;
  if RecordCount = 0 then begin
    prNodeMem.ChildNode := nil;
  end else begin
    prNodeMem.ChildNode := @(Records[0].prNodeMem);
  end;
end;

function TQUB_QFS_Record.AddRecord(r :TQUB_QFS_Record; Index :Integer) :Boolean;
var
  PrevRec, NextRec :TQUB_QFS_Record;
begin
  Result := False;
  if prRecords = nil then begin
    prRecords := TStringList.Create;
  end;
  PrevRec := Records[Index - 1];
  NextRec := Records[Index];
  prRecords.InsertObject(Index, '', r);
  if PrevRec <> nil then begin
    PrevRec.prNodeMem.SiblingNode := @(r.prNodeMem);
  end;
  if NextRec <> nil then begin
    r.prNodeMem.SiblingNode := @(NextRec.prNodeMem);
  end;
  r.prParent := Self;
  r.prNodeMem.ParentNode := @prNodeMem;
  prRecordCount := prRecords.Count;
  if RecordCount = 0 then begin
    prNodeMem.ChildNode := nil;
  end else begin
    prNodeMem.ChildNode := @(Records[0].prNodeMem);
  end;
end;

function TQUB_QFS_Record.AddRecordFromNode(n :INode) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_Record.DelRecord(r :TQUB_QFS_Record) :Boolean;
var
  PrevRec, NextRec :TQUB_QFS_Record;
  i                :Integer;
begin
  Result := False;
  if prRecords = nil then Exit;
  i := prRecords.IndexOfObject(r);
  if i = -1 then Exit;
  PrevRec := Records[i - 1];
  NextRec := Records[i + 1];
  prRecords.Delete(i);
  if PrevRec <> nil then begin
    if NextRec <> nil then begin
      PrevRec.prNodeMem.SiblingNode := @(NextRec.prNodeMem);
    end else begin
      PrevRec.prNodeMem.SiblingNode := nil;
    end;
  end;
  prRecordCount := prRecords.Count;
  if prRecords.Count = 0 then begin
    prRecords.Free;
    prRecords := nil;
  end;
  if RecordCount = 0 then begin
    prNodeMem.ChildNode := nil;
  end else begin
    prNodeMem.ChildNode := @(Records[0].prNodeMem);
  end;
end;


//******************************************************************************
function TQUB_QFS_Record.SizeOnStream(Recursive :Boolean = True) :LongInt;
var
  c :Integer;
begin
  Result := SizeOf(prNodeMem.NodeStream);
  Inc(Result, prNodeMem.NodeStream.NameLen);
  if not FlagDataInNode then Inc(Result, DataSize * DataCount);
  if not Recursive then Exit;

  for c := 0 to RecordCount - 1 do begin
    Inc(Result, Records[c].SizeOnStream);
  end;
end;


//******************************************************************************
//This function can be overriden, to read ahead and, depending on the node's name
//return a different type, descendent from Record
function TQUB_QFS_Record.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
begin
  Result := TQUB_QFS_Record.Create('', Self);
end;


//******************************************************************************
procedure TQUB_QFS_Record.prIncDataSizeCount(NewDataSize, NewDataCount :Integer);
begin
  if not FlagPreload then Exit;

  try

    if FlagDataInNode then begin
    //Data In Node

      if (NewDataCount * NewDataSize) > SizeOf(prNodeMem.NodeStream.DataStreamPos) then begin
      //No longer fits in DataStreamPos field

        //Create separate data
        prNodeMem.Data := AllocMem(NewDataCount * NewDataSize);
        Move(prNodeMem.NodeStream.DataStreamPos, prNodeMem.Data^, SizeOf(prNodeMem.NodeStream.DataStreamPos));
        prNodeMem.NodeStream.DataStreamPos := 0;

        //Clear the flag
        FlagDataInNode := False;

      end;

    end else begin

      ReAllocMem(prNodeMem.Data, NewDataCount * NewDataSize);

    end;

    prNodeMem.DataLoadSel := QUB_FirstLast(0, NewDataCount - 1);

    prNodeMem.NodeStream.DataCount := NewDataCount;
    DataSize  := NewDataSize;

  except
  end;
end;


//Flags
function TQUB_QFS_Record.prGetFlagUnsigned :Boolean;
begin
  Result := False;
  case DataType of
    NODE_DATATYPE_BYTE, NODE_DATATYPE_WORD, NODE_DATATYPE_LONGWORD:
      Result := True;
  end;//case
end;

function TQUB_QFS_Record.prGetFlagPreload :Boolean;
begin
  Result := (prNodeMem.NodeStream.Flags and NODE_FLAG_PRELOAD) > 0;
end;

procedure TQUB_QFS_Record.prSetFlagPreload(f :Boolean);
begin
  if f then prNodeMem.NodeStream.Flags := NODE_FLAG_PRELOAD or (prNodeMem.NodeStream.Flags and NODE_FLAG_PRELOAD_MASK) else prNodeMem.NodeStream.Flags := (prNodeMem.NodeStream.Flags and NODE_FLAG_PRELOAD_MASK);
end;

function TQUB_QFS_Record.prGetFlagDataInNode :Boolean;
begin
  Result := (prNodeMem.NodeStream.Flags and NODE_FLAG_DATA_IN_NODE) > 0;
end;

procedure TQUB_QFS_Record.prSetFlagDataInNode(f :Boolean);
begin
  if f then prNodeMem.NodeStream.Flags := NODE_FLAG_DATA_IN_NODE or (prNodeMem.NodeStream.Flags and NODE_FLAG_DATA_IN_NODE_MASK) else prNodeMem.NodeStream.Flags := (prNodeMem.NodeStream.Flags and NODE_FLAG_DATA_IN_NODE_MASK);
end;

function TQUB_QFS_Record.prGetFlagNonWritable :Boolean;
begin
  Result := (prNodeMem.NodeStream.Flags and NODE_FLAG_NONWRITABLE) > 0;
end;

procedure TQUB_QFS_Record.prSetFlagNonWritable(f :Boolean);
begin
  if f then prNodeMem.NodeStream.Flags := NODE_FLAG_NONWRITABLE or (prNodeMem.NodeStream.Flags and NODE_FLAG_NONWRITABLE_MASK) else prNodeMem.NodeStream.Flags := (prNodeMem.NodeStream.Flags and NODE_FLAG_NONWRITABLE_MASK);
end;



//****************************************************************************//
//                                                                            //
//                                 V A L U E                                  //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_Record.prGetDataValue :Variant;
begin
  Result := DataArrayValue[0];
end;

procedure TQUB_QFS_Record.prSetDataValue(v :Variant);
begin
  DataArrayValue[0] := v;
end;

function TQUB_QFS_Record.prGetDataArrayValue(Index :Integer) :Variant;
begin
  Result := varEmpty;
  if FlagPreload then begin
    if (Index < 0) then Exit;
    if (Index >= DataCount) then prIncDataSizeCount(DataSize, Index + 1);
    case prNodeMem.NodeStream.DataType of
      NODE_DATATYPE_POINTER:
        begin
          Result := Integer(TPAPointer(Data)[Index]);
        end;
      NODE_DATATYPE_STRING:
        begin
          Result := String(PChar(Data));
        end;
      NODE_DATATYPE_BYTE:
        begin
          Result := TPAByte(Data)[Index];
        end;
      NODE_DATATYPE_SHORTINT:
        begin
          Result := TPAShort(Data)[Index];
        end;
      NODE_DATATYPE_WORD:
        begin
          Result := TPAWord(Data)[Index];
        end;
      NODE_DATATYPE_SMALLINT:
        begin
          Result := TPASmall(Data)[Index];
        end;
      {
      NODE_DATATYPE_INTEGER:
        begin
          Result := Variant(TPAInteger(Data)[Index]);
        end;
      }
      NODE_DATATYPE_LONGWORD:
        begin
          Result := Variant(TPALongWord(Data)[Index]);
        end;
      NODE_DATATYPE_LONGINT:
        begin
          Result := TPALongInt(Data)[Index];
        end;
      NODE_DATATYPE_INT64:
        begin
          Result := Variant(Integer(TPAInt64(Data)[Index]));
        end;
      NODE_DATATYPE_UINT64:
        begin
          Result := Variant(Integer(TPAUInt64(Data)[Index]));
        end;
      NODE_DATATYPE_SINGLE:
        begin
          Result := TPASingle(Data)[Index];
        end;
      NODE_DATATYPE_DOUBLE:
        begin
          Result := TPADouble(Data)[Index];
        end;
      NODE_DATATYPE_EXTENDED:
        begin
          Result := TPAExtended(Data)[Index];
        end;
    end;//case
  end;
end;

procedure TQUB_QFS_Record.prSetDataArrayValue(Index :Integer; v :Variant);
var
  l :Integer;
begin
  try
    if FlagPreload then begin
      if (Index < 0) then Exit;
      if (Index >= DataCount) then prIncDataSizeCount(DataSize, Index + 1);
      case prNodeMem.NodeStream.DataType of
        NODE_DATATYPE_POINTER:
          begin
            TPAPointer(Data)[Index] := Pointer(Integer(v));
          end;
        NODE_DATATYPE_STRING:
          begin
            l := Length(VarAsType(v, varString)) + SizeOf(Char);//allocate room for one extra byte ...
            if l > DataCount then prIncDataSizeCount(DataSize, l);
            Move(PChar(String(v))^, Data^, DataCount - SizeOf(Char));//... copy all except the extra byte...
            TPAByte(Data)[DataCount - 1] := 0;//... and zero it, per Chris' request
          end;
        NODE_DATATYPE_BYTE:
          begin
            TPAByte(Data)[Index] := Byte(v);
          end;
        NODE_DATATYPE_SHORTINT:
          begin
            TPAShort(Data)[Index] := ShortInt(v);
          end;
        NODE_DATATYPE_WORD:
          begin
            TPAWord(Data)[Index] := Word(v);
          end;
        NODE_DATATYPE_SMALLINT:
          begin
            TPASmall(Data)[Index] := SmallInt(v);
          end;
        {
        NODE_DATATYPE_INTEGER:
          begin
            TPAInteger(Data)[Index] := Integer(v);
          end;
        }
        NODE_DATATYPE_LONGWORD:
          begin
            TPALongWord(Data)[Index] := LongWord(v);
          end;
        NODE_DATATYPE_LONGINT:
          begin
            TPALongInt(Data)[Index] := LongInt(v);
          end;
        NODE_DATATYPE_INT64:
          begin
            TPAInt64(Data)[Index] := Int64(Integer(v));///
          end;
        NODE_DATATYPE_UINT64:
          begin
            TPAUInt64(Data)[Index] := UInt64(Integer(v));///
          end;
        NODE_DATATYPE_SINGLE:
          begin
            TPASingle(Data)[Index] := Single(v);
          end;
        NODE_DATATYPE_DOUBLE:
          begin
            TPADouble(Data)[Index] := Double(v);
          end;
        NODE_DATATYPE_EXTENDED:
          begin
            TPAExtended(Data)[Index] := Extended(v);
          end;
      end;//case
    end;
  except
  end;
end;

//****************************************************************************//
//                                                                            //
//                                   R E A D                                  //
//                                                                            //
//****************************************************************************//
//function TQUB_QFS_Record.ReadDll(ReadStream :TStream) :Boolean;
//var
//  RootNode :TQUB_QFS_PNodeMem;
//begin

  //Result := False;
  {

  //Register DllMethodTable and StreamPointer
  // . . .

  RootNode := ReadTreeFromStream(StreamMethodsHandle);
  //The reference count here is 1, but I own the reference !!!
  Result := RootNode = nil;
  if Result then Exit;


  prNodeMem := RootNode;

  Result := ReadFromTree;
  }

//end;


function TQUB_QFS_Record.ReadFromNode(n :INode) :Boolean;
begin

  Result := False;
  //make and read children (if neccessary), of corresponding objects types,
  //and have each child ReadFromNode

end;


function TQUB_QFS_Record.Read(ReadStream :TStream) :Boolean;
var
  ChildPos :LongInt;
  Child    :TQUB_QFS_Record;
  ReadingChildren :Boolean;
begin

  //The parent or the overriding method must protect the stream from concurent access
  Result := False;
  ReadingChildren := False;

  Reset;

  Stream := ReadStream;

  //Read itself first:
  try

    try

      //Save position in FileStream
      StreamPos := Stream.Position;

      //Read NodeStream
      Stream.ReadBuffer(prNodeMem.NodeStream, SizeOf(prNodeMem.NodeStream));

      //Read Name
      SetLength(prName, prNodeMem.NodeStream.NameLen);
      prNodeMem.Name := PChar(prName);
      Stream.ReadBuffer(prNodeMem.Name^, prNodeMem.NodeStream.NameLen);

      //Preload Data
      if FlagPreload then begin
        PreLoadData;
      end else begin
        UnloadData;
      end;

      //Read Children
      ReadingChildren := True;
      ChildPos := prNodeMem.NodeStream.ChildStreamPos;
      while ChildPos > 0 do begin

        Stream.Position := ChildPos;
        Child           := NewChild(ReadStream);
        Child.Read(ReadStream);
        ChildPos        := Child.SiblingStreamPos;

      end;


    except

      Result := True;

    end;


  finally

    if Result then begin
      if not ReadingChildren then Reset;
    end;


  end;


end;



//****************************************************************************//
//                                                                            //
//                                  W R I T E                                 //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_Record.Write(WriteStream :TStream) :Boolean;
var
  WriteNodeStream  :TQUB_QFS_NodeStream;
  WNode            :TQUB_QFS_PNodeMem;
  WritePos         :LongInt;
  Buffer           :Pointer;
  BufferCount      :Integer;
  DataCountWritten :integer;// love ya
  OldPos           :Integer;
  c                :Integer;
begin

  //The parent or the overriding method must protect the stream from concurent access

  Result := False;

  try

    WritePos := WriteStream.Position;

    //Copy the WriteNodeStream
    WriteNodeStream := prNodeMem.NodeStream;

    //Set the Stream Pointers
    if (DataSize * DataCount > 0) then begin
      if not FlagDataInNode then begin
        WriteNodeStream.DataStreamPos  := WritePos + SizeOf(prNodeMem.NodeStream) + prNodeMem.NodeStream.NameLen;
      end;
    end else begin
      WriteNodeStream.DataStreamPos    := 0;
    end;

    WriteNodeStream.ChildStreamPos     := WritePos + SizeOnStream(False);
    WriteNodeStream.SiblingStreamPos   := WriteNodeStream.ChildStreamPos;

    if prNodeMem.ChildNode = nil then begin
    //There is no Child
      WriteNodeStream.ChildStreamPos   := 0;
    end else begin
      WNode := prNodeMem.ChildNode;
      while (WNode.NodeStream.Flags and NODE_FLAG_NONWRITABLE) <> 0 do begin
        WNode := WNode.SiblingNode;
        if WNode = nil then Break;
      end;
      if WNode = nil then begin
        WriteNodeStream.ChildStreamPos := 0;
      end;                   
    end;

    if prNodeMem.SiblingNode = nil then begin
    //There is no Sibling
      WriteNodeStream.SiblingStreamPos := 0;
    end else begin
      //Find the next Writable Node
      WNode := prNodeMem.SiblingNode;
      while (WNode.NodeStream.Flags and NODE_FLAG_NONWRITABLE) <> 0 do begin
        WNode := WNode.SiblingNode;
        if WNode = nil then Break;
      end;
      if WNode <> nil then begin
        //Adjust Sibling position to account for the Children
        for c := 0 to RecordCount - 1 do begin
          if not Records[c].FlagNonWritable then Inc(WriteNodeStream.SiblingStreamPos, Records[c].SizeOnStream);
        end;
      end else begin
        WriteNodeStream.SiblingStreamPos := 0;
      end;
    end;

    //Write NodeStream
    WriteStream.WriteBuffer(WriteNodeStream, SizeOf(WriteNodeStream));

    //Write Name
    WriteStream.WriteBuffer(prNodeMem.Name^, WriteNodeStream.NameLen);

    //Write Data, if any
    if (DataSize * DataCount > 0) and (not FlagDataInNode) then begin

      if (Data <> nil) and (DataLoadSel.First = 0) and (DataLoadSel.Last = (DataCount-1)) then begin // if Preload then begin
      //Write the Data directly

        WriteStream.WriteBuffer(Data^, DataSize * DataCount);

      end else begin

        OldPos := 0;
        BufferCount := ONE_MEGABYTE;
        Buffer := nil;
        try
          try
            Buffer := AllocMem(BufferCount * DataSize);
            DataCountWritten := 0;
            while DataCountWritten < DataCount do begin
              if (DataCountWritten + BufferCount) > DataCount then BufferCount := DataCount - DataCountWritten;
              if WriteStream = Stream then begin
                OldPos := WriteStream.Position;
              end;
              try
                LoadDataInBuffer(Buffer, DataCountWritten, DataCountWritten + BufferCount - 1);
              finally
                if WriteStream = Stream then begin
                  WriteStream.Position := OldPos;
                end;
              end;
              WriteStream.WriteBuffer(Buffer^, DataSize * BufferCount);
              Inc(DataCountWritten, BufferCount);
            end;
          except
          end;
        finally
          FreeMem(Buffer);
        end;

      end;

    end;


    //Write Children
    for c := 0 to RecordCount - 1 do begin
      if not Records[c].FlagNonWritable then begin
        Records[c].Write(WriteStream);
      end;
    end;


  except

    Result := True;

  end;

end;

function TQUB_QFS_Record.PreLoadData :Boolean;
var
  OldPos :Integer;
begin
  Result := False;

  if not FlagPreload then Exit;

  if (FlagDataInNode) and (DataSize * DataCount <= SizeOf(prNodeMem.NodeStream.DataStreamPos))then begin

    prNodeMem.Data := @prNodeMem.NodeStream.DataStreamPos;
    DataLoadSel := QUB_FirstLast(0, DataCount - 1);

  end else if (prNodeMem.NodeStream.DataStreamPos > 0) then begin   // if havedata added by Chris june 2002

    try

      prNodeMem.Data := AllocMem(DataCount * DataSize);

      OldPos := Stream.Position;
      try
        Stream.Position := DataStreamPos;
        Stream.ReadBuffer(Data^, DataCount * DataSize);
        DataLoadSel := QUB_FirstLast(0, DataCount - 1);
      finally
        Stream.Position := OldPos;
      end;

    except

      Result := True;
      DataLoadSel := QUB_FirstLast(-1, -1);

    end;

  end;
end;

function TQUB_QFS_Record.LoadData(First, Last :Integer) :Boolean;
var
  OldPos :Integer;
begin
  Result := False;

  if FlagPreload then Exit;

  try

    if ((DataLoadSel.Last - DataLoadSel.First) <> (Last - First)) or (DataLoadSel.First = -1) or (DataLoadSel.Last = -1) then begin
      ReallocMem(prNodeMem.Data, (Last - First + 1) * DataSize);
    end;

    OldPos := Stream.Position;
    try
      Stream.Position := DataStreamPos + First * DataSize;
      Stream.ReadBuffer(Data^, (Last - First + 1) * DataSize);
      DataLoadSel := QUB_FirstLast(First, Last);
    finally
      Stream.Position := OldPos;
    end;

  except

    Result := True;
    DataLoadSel := QUB_FirstLast(-1, -1);

  end;
end;

// warning: if data is partially loaded and the portion in memory has changed since reading,
//          then loaddatainbuffer will load the version from disk ignoring what's in mem

function TQUB_QFS_Record.LoadDataInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean;
var
  OldPos :Integer;
begin
  Result := False;

  try

    if (First >= DataLoadSel.First) and (First <= DataLoadSel.Last) and (Last >= DataLoadSel.First) and (Last <= DataLoadSel.Last) and (Data <> nil) then begin
    //Data is already loaded, copy it

      Move((@TPAByte(Data)[DataSize * (First - DataLoadSel.First)])^, Buffer^, (Last - First + 1) * DataSize);

    end else begin

      OldPos := Stream.Position;
      try
        Stream.Position := DataStreamPos + First * DataSize;
        Stream.ReadBuffer(Buffer^, (Last - First + 1) * DataSize);
      finally
        Stream.Position := OldPos;
      end;

    end;

  except

    Result := True;

  end;
end;

function TQUB_QFS_Record.UnloadData :Boolean;
begin
  Result := False;
  if FlagPreload then begin
    if not FlagDataInNode then begin
      try
        ReAllocMem(prNodeMem.Data, DataCount * DataSize);
      except
        Result := True;
      end;
    end;
    DataLoadSel := QUB_FirstLast(0, DataCount - 1);
  end else begin
    if not FlagDataInNode then begin
      try
        FreeMem(Data);
      except
        Result := True;
      end;
      Data := nil;
    end;
    DataLoadSel := QUB_FirstLast(-1, -1);
  end;
end;

function TQUB_QFS_Record.ClearData :Boolean;
begin
  Result := False;
  if not FlagDataInNode then begin
    try
      FreeMem(Data);
    except
      Result := True;
    end;
  end;
  Data := nil;
  DataLoadSel := QUB_FirstLast(-1, -1);
end;

//******************************************************************************
procedure TQUB_QFS_Record.prSetDataType(t :TQUB_QFS_Node_DataType);
var
  s :Integer;
begin
  prNodeMem.NodeStream.DataType := t;
  s := 0;
  case DataType of
    NODE_DATATYPE_EMPTY:        s := 0;
    NODE_DATATYPE_UNKNOWN:      s := 0;
    NODE_DATATYPE_POINTER:      s := SizeOf(Pointer);
    NODE_DATATYPE_STRING:       s := SizeOf(Char);
    NODE_DATATYPE_BYTE:         s := SizeOf(Byte);
    NODE_DATATYPE_SHORTINT:     s := SizeOf(ShortInt);
    NODE_DATATYPE_WORD:         s := SizeOf(Word);
    NODE_DATATYPE_SMALLINT:     s := SizeOf(SmallInt);
    //NODE_DATATYPE_INTEGER:      s := SizeOf(Integer);
    NODE_DATATYPE_LONGWORD:     s := SizeOf(LongWord);
    NODE_DATATYPE_LONGINT:      s := SizeOf(LongInt);
    NODE_DATATYPE_INT64:        s := SizeOf(Int64);
    NODE_DATATYPE_UINT64:       s := SizeOf(UInt64);
    NODE_DATATYPE_SINGLE:       s := SizeOf(Single);
    NODE_DATATYPE_DOUBLE:       s := SizeOf(Double);
    NODE_DATATYPE_EXTENDED:     s := SizeOf(Extended);
  end;//case

  if FlagPreload then begin
    prIncDataSizeCount(s, DataCount);
  end;

end;

procedure TQUB_QFS_Record.prSetDataSize(s :LongInt);
begin
  prNodeMem.NodeStream.DataSize := s;
end;

procedure TQUB_QFS_Record.prSetDataCount(c :LongInt);
begin
  if (c > DataCount) then prIncDataSizeCount(DataSize, c);
  prNodeMem.NodeStream.DataCount := c;
end;

procedure TQUB_QFS_Record.prSetData(d :Pointer);
begin
  prNodeMem.Data := d;       // holy memory leak, batman!
end;


//******************************************************************************
function TQUB_QFS_Record.prGetRecord(Index :Integer) :TQUB_QFS_Record;
begin
  Result := nil;
  if prRecords <> nil then begin
    if (Index >= 0) and (Index < prRecordCount) then begin
      Result := TQUB_QFS_Record(prRecords.Objects[Index]);
    end;
  end;
end;

function TQUB_QFS_Record.prGetRecordByName(const Index :String) :TQUB_QFS_Record;
var
  i :Integer;
begin
  Result := nil;
  if prRecords <> nil then begin
    for i := 0 to prRecordCount - 1 do begin
      if TQUB_QFS_Record(prRecords.Objects[i]).Name = Index then begin
        Result := TQUB_QFS_Record(prRecords.Objects[i]);
        Exit;
      end;
    end;
  end;
end;

function TQUB_QFS_Record.RecordIndex(r :TQUB_QFS_Record) :Integer;
begin
  if prRecords = nil then Result := -1 else Result := prRecords.IndexOfObject(r);
end;

procedure TQUB_QFS_Record.Swap(i, j :Integer);
  procedure Relink(k :Integer);
  begin
    if k = 0 then begin
      prNodeMem.ChildNode := @ Records[0].prNodeMem;
    end else begin
      Records[k-1].prNodeMem.SiblingNode := @ Records[k].prNodeMem;
    end;
    if (k+1) < prRecordCount then begin
      Records[k].prNodeMem.SiblingNode := @ Records[k+1].prNodeMem;
    end else begin
      Records[k].prNodeMem.SiblingNode := nil;
    end;
  end;
begin
  if (i < prRecordCount) and (j < prRecordCount) then begin
    if( i<j ) Then Begin
      prRecords.Move(i,j);
      prRecords.Move(j-1,i);
    end else begin
      prRecords.Move(j,i);
      prRecords.Move(i-1,j);
    end;
    Relink(i);
    Relink(j);
  end;
end;


//******************************************************************************
function TQUB_QFS_Record.prWaitToBeReady(MSecs :Integer) :Boolean;
var
  Start :Comp;
begin
  Result := True;

  try
    Start := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  except
    Exit;
  end;

  repeat
    try
      if (TimeStampToMSecs(DateTimeToTimeStamp(Now)) - Start) > MSecs then Break;
    except
      Break;
    end;
  until( not ((prReading) or (prWriting)));

  if not ((prReading) or (prWriting)) then begin
    //prReading := True;
    Result := False;
  end;

end;


initialization

//Fill DLLMethodTable


//******************************************************************************
end.

