{*******************************************************}
{                                                       }
{                   Qub File System                     }
{           Copyright (C) 2000 Lorin Milescu            }
{                  Created 01-31-2000                   }
{                                                       }
{*******************************************************}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


unit QUB_QFS_File;


//******************************************************************************
interface


uses


Windows, SysUtils, Classes, FileCtrl, Dialogs, Graphics, Forms, SyncObjs,

Alloc,

QUB_Global,
QUB_Console, QUB_Rich_Console,
QUB_QFS, QUB_QFS_Edit;


//******************************************************************************
type


TQUB_EditOperation = (EDIT_OPER_COPY, EDIT_OPER_CUT, EDIT_OPER_PASTE, EDIT_OPER_DELETE,
                      EDIT_OPER_ERASE, EDIT_OPER_FILL, EDIT_OPER_EDITPOINT, EDIT_OPER_BREAK, EDIT_OPER_UNDO,
                      EDIT_OPER_CLEAR, EDIT_OPER_APPEND,
                      EDIT_OPER_UPDATE,
                      EDIT_OPER_JOIN, EDIT_OPER_IBREAK,
                      EDIT_OPER_INSERT_SEG, EDIT_OPER_INSERT_INTOSEG,
                      EDIT_OPER_DEL_SEG, EDIT_OPER_DEL_SEL
                     );

TQUB_EditEvent = procedure (Sender :TObject; EditOper :TQUB_EditOperation; var Param; var Param2) of object;
//TQUB_EditEvent = procedure (Sender :TObject; EditOper :TQUB_EditOperation; const Sel :TQUB_FirstLast) of object;


//Forward class declarations ***************************************************
TQUB_QFS_File                                  = class;


//File
TQUB_QFS_File                                  = class(TQUB_QFS_Record)
protected


  //Edit
  prEditStream                                 :TStream;
  prEditFileName                               :String;
  prEditOperations                             :TQUB_QFS_EditOperations;

  prFileName                                   :String;

  prOpened                                     :Boolean;
  prCanEdit                                    :Boolean;
  prSaved                                      :Boolean;
  prEditing                                    :Boolean;

  prLockWriteCount                             :Integer;
  prLockReadWriteCount                         :Integer;
  prLockRecordCount                            :Integer;
  prLockPlaybackCount                          :Integer;

  prEditCritSection                            :TCriticalSection;

  prOnOpen                                     :TNotifyEvent;
  prOnClose                                    :TNotifyEvent;

  prOnEdit                                     :TQUB_EditEvent;

  prOpenSilently                               :Boolean;
  prSaveSilently                               :Boolean;

  prOpenedFromStream                           :Boolean;


  function prGetLockedAny :Boolean;
  function prGetLockedWrite :Boolean;
  function prGetLockedReadWrite :Boolean;
  function prGetLockedRecord :Boolean;
  function prGetLockedPlayback :Boolean;

  function prReadMagic(ReadStream :TStream) :Boolean; virtual;
  function prWriteMagic(WriteStream :TStream) :Boolean; virtual;

  function prGetFileNameForSave(const BaseName, BaseExt :String; IncName :Boolean; MaxInc :Integer) :String;

  function prOpenEdit :Boolean; virtual;
  procedure prSetCanEdit(e :Boolean); virtual;

  //procedure prDoEditChange(RaiseEvent :Boolean); virtual;


  //Low level File *************************************************************
  function prNew :Boolean; virtual; abstract;
  function prOpen(const FileName :String) :Boolean; virtual; abstract;
  function prSave(const FileName :String) :Boolean; virtual; abstract;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  destructor Destroy; override;


  //File
  function New(const FileName :String) :Boolean; virtual;
  function Open(const FileName :String) :Boolean; overload; virtual;
  function Open(const FileName :String; OpenStream :TStream) :Boolean; overload; virtual;
  function Save(const FileName :String) :Boolean; virtual;
  function Close :Boolean; virtual;

  function Read(ReadStream :TStream) :Boolean; override;
  function Write(WriteStream :TStream) :Boolean; override;

  procedure OnGetPropertiesNew(Sender :TObject); virtual;
  procedure OnPutPropertiesNew(Sender :TObject); virtual;

  procedure LockAny;
  procedure UnLockAny;
  procedure LockWrite;
  procedure UnLockWrite;
  procedure LockReadWrite;
  procedure UnLockReadWrite;
  procedure LockRecord;
  procedure UnLockRecord;
  procedure LockPlayback;
  procedure UnLockPlayback;


  property FileName :String read prFileName;

  //Flags
  property LockedAny                           :Boolean             read prGetLockedAny;
  property LockedWrite                         :Boolean             read prGetLockedWrite;
  property LockedReadWrite                     :Boolean             read prGetLockedReadWrite;
  property LockedRecord                        :Boolean             read prGetLockedRecord;
  property LockedPlayback                      :Boolean             read prGetLockedPlayback;

  //File Flags
  property Opened                              :Boolean             read prOpened;
  property Saved                               :Boolean             read prSaved;

  //Can Edit
  property EditStream                          :TStream             read prEditStream          write prEditStream;
  property CanEdit                             :Boolean             read prCanEdit             write prSetCanEdit;

  //Events
  property OnOpen                              :TNotifyEvent        read prOnOpen              write prOnOpen;
  property OnClose                             :TNotifyEvent        read prOnClose             write prOnClose;
  property OnEdit                              :TQUB_EditEvent      read prOnEdit              write prOnEdit;

  property OpenSilently                        :Boolean             read prOpenSilently        write prOpenSilently;
  property SaveSilently                        :Boolean             read prSaveSilently        write prSaveSilently;


end;


//******************************************************************************
implementation


//****************************************************************************//
//                                                                            //
//                              TQUB_QFS_File                                 //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_File.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited;

  prEditCritSection := TCriticalSection.Create;

  prCanEdit         := True;

  prEditOperations  := TQUB_QFS_EditOperations.Create;

end;

destructor TQUB_QFS_File.Destroy;
begin
  try

    prEditOperations.Free;

    prEditCritSection.Free;

  finally

    inherited;

  end;
end;


//****************************************************************************//
//                                                                            //
//                            F I L E    M E T H O D S                        //
//                                                                            //
//****************************************************************************//


//****************************************************************************//
//                                                                            //
//                                    N e w                                   //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_File.New(const FileName :String) :Boolean;
begin

  Result               := Close;
  if Result then Exit;

  prOpenedFromStream   := False;

  try

    Stream             := TFileStream.Create(FileName, fmCreate or fmShareDenyWrite);

  except

    Stream.Free;
    Stream := nil;
    Result := True;
    Console.NewEntry('Cannot create new file ' + FileName, COLOR_ERROR);
    if not prOpenSilently then MessageHelpDlg('Cannot create new file', mtError, [mbOK], 0, 'Error:Cannot_Create_New_File');
    Exit;

  end;

  prFileName           := FileName;

  //Temporary EditFile
  if CanEdit then begin

    prOpenEdit;

  end;

  prOpened             := True;

  prModified           := False;
  prSaved              := False;

  prLockWriteCount     := 0;
  prLockReadWriteCount := 0;
  prLockRecordCount    := 0;
  prLockPlaybackCount  := 0;

  Result               := prNew;

  if Assigned(OnOpen) then OnOpen(Self);

end;


//****************************************************************************//
//                                                                            //
//                                  O p e n                                   //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_File.Open(const FileName :String) :Boolean;
begin

  Result              := Close;
  if Result then Exit;

  prOpenedFromStream  := False;

  Result              := prOpen(FileName);
  if Result then Exit;

  prFileName          := FileName;

  try

    Stream            := TFileStream.Create(prFileName, fmOpenRead or fmShareDenyWrite);

  except

    Stream.Free;
    Stream := nil;
    Result := True;
    Console.NewEntry('Cannot open file ' + prFileName, COLOR_ERROR);
    // obnoxiously duplicated by viewport_whatever:
    // if not prOpenSilently then MessageHelpDlg('Cannot open file', mtError, [mbOK], 0, 'Error:OpenFile');
    Exit;

  end;

  //Temporary EditFile
  if CanEdit then begin

    prOpenEdit;

  end;

  Result              := Read(Stream);

  prOpened            := True;
  prModified          := False;
  prSaved             := True;

  if Result then begin

    Close;

  end else begin

    prLockWriteCount     := 0;
    prLockReadWriteCount := 0;
    prLockRecordCount    := 0;
    prLockPlaybackCount  := 0;

    if Assigned(OnOpen) then OnOpen(Self);

  end;

end;


//****************************************************************************//
//                                                                            //
//                      O p e n    F r o m    S t r e a m                     //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_File.Open(const FileName :String; OpenStream :TStream) :Boolean;
begin

  Result              := Close;
  if Result then Exit;


  prOpenedFromStream  := True;

  Result              := prOpen(FileName);
  if Result then Exit;

  prFileName          := FileName;

  Stream              := OpenStream;

  //Temporary EditFile
  if CanEdit then begin

    prOpenEdit;

  end;

  Result              := Read(Stream);

  prOpened            := True;
  prModified          := False;
  prSaved             := True;

  if Result then begin

    Close;

  end else begin

    prLockWriteCount     := 0;
    prLockReadWriteCount := 0;
    prLockRecordCount    := 0;
    prLockPlaybackCount  := 0;

    if Assigned(OnOpen) then OnOpen(Self);

  end;

end;


//****************************************************************************//
//                                                                            //
//                                  S a v e                                   //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_File.Save(const FileName :String) :Boolean;
begin

  Result := (LockedAny) or (LockedRecord) or (LockedPlayback);

  if Result then Exit;

  LockAny;
  try

    Result := prSave(FileName);

  finally

    UnlockAny;

  end;

end;



//****************************************************************************//
//                                                                            //
//                                  C l o s e                                 //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_File.Close :Boolean;
begin

  Result := False;

  if not prOpened then Exit;

  try
    prEditOperations.Clear;
  except
  end;

  Reset;

  if Stream <> nil then begin
    Stream.Free;
    Stream := nil;
  end;

  if not prSaved then begin
    //New files which were never saved are deleted
    try
      DeleteFile(prFileName);
    except
    end;
  end;

  if prEditStream <> nil then begin
    prEditStream.Free;
    prEditStream := nil;
    try
      DeleteFile(prEditFileName);
    except
    end;
  end;

  prOpened    := False;
  prModified  := False;

  prLockWriteCount     := 0;
  prLockReadWriteCount := 0;
  prLockRecordCount    := 0;
  prLockPlaybackCount  := 0;

  if Assigned(OnClose) then OnClose(Self);

end;


procedure TQUB_QFS_File.OnGetPropertiesNew(Sender :TObject);
begin
end;

procedure TQUB_QFS_File.OnPutPropertiesNew(Sender :TObject);
begin
end;


//******************************************************************************
function TQUB_QFS_File.Read(ReadStream :TStream) :Boolean;
begin

  Result := prReadMagic(ReadStream);
  if not Result then begin
    Result := inherited Read(ReadStream);
  end;

end;

function TQUB_QFS_File.Write(WriteStream :TStream) :Boolean;
begin
  Result := prWriteMagic(WriteStream);
  if not Result then begin
    Result := inherited Write(WriteStream);
  end;
end;

function TQUB_QFS_File.prReadMagic(ReadStream :TStream) :Boolean;
var
  Magic :Array[0..QUB_QFS_MAGIC_LEN - 1] of Char;
begin

  try

    ReadStream.ReadBuffer(Magic, QUB_QFS_MAGIC_LEN);
    Result := StrLComp(Magic, QUB_QFS_MAGIC, QUB_QFS_MAGIC_LEN) <> 0;

  except

    Result := True;

  end;

end;

function TQUB_QFS_File.prWriteMagic(WriteStream :TStream) :Boolean;
begin

  Result := False;

  try

    WriteStream.WriteBuffer(QUB_QFS_MAGIC, QUB_QFS_MAGIC_LEN);

  except

    Result := True;

  end;

end;

//Lock
function TQUB_QFS_File.prGetLockedAny :Boolean;
begin
  Result := (LockedWrite) or (LockedReadWrite);
end;

function TQUB_QFS_File.prGetLockedWrite :Boolean;
begin
  Result := prLockWriteCount > 0;
end;

function TQUB_QFS_File.prGetLockedReadWrite :Boolean;
begin
  Result := prLockReadWriteCount > 0;
end;

function TQUB_QFS_File.prGetLockedRecord :Boolean;
begin
  Result := prLockRecordCount > 0;
end;

function TQUB_QFS_File.prGetLockedPlayback :Boolean;
begin
  Result := prLockPlaybackCount > 0;
end;

procedure TQUB_QFS_File.LockAny;
begin
  LockWrite;
  LockReadWrite;
end;

procedure TQUB_QFS_File.UnLockAny;
begin
  UnLockWrite;
  UnLockReadWrite;
end;

procedure TQUB_QFS_File.LockWrite;
begin
  Inc(prLockWriteCount);
end;

procedure TQUB_QFS_File.UnLockWrite;
begin
  Dec(prLockWriteCount);
  if prLockWriteCount < 0 then prLockWriteCount := 0;
end;

procedure TQUB_QFS_File.LockReadWrite;
begin
  Inc(prLockReadWriteCount);
end;

procedure TQUB_QFS_File.UnLockReadWrite;
begin
  Dec(prLockReadWriteCount);
  if prLockReadWriteCount < 0 then prLockReadWriteCount := 0;
end;

procedure TQUB_QFS_File.LockRecord;
begin
  Inc(prLockRecordCount);
end;

procedure TQUB_QFS_File.UnLockRecord;
begin
  Dec(prLockRecordCount);
  if prLockRecordCount < 0 then prLockRecordCount := 0;
end;

procedure TQUB_QFS_File.LockPlayback;
begin
  Inc(prLockPlaybackCount);
end;

procedure TQUB_QFS_File.UnLockPlayback;
begin
  Dec(prLockPlaybackCount);
  if prLockPlaybackCount < 0 then prLockPlaybackCount := 0;
end;

function TQUB_QFS_File.prGetFileNameForSave(const BaseName, BaseExt :String; IncName :Boolean; MaxInc :Integer) :String;
var
  i :Integer;
begin

  i := 0;
  repeat
    if IncName then begin
      Result := ChangeFileExt(BaseName + IntToStr(i), BaseExt);
    end else begin
      Result := ChangeFileExt(BaseName, BaseExt + IntToStr(i));
    end;
    if not FileExists(Result) then begin
      Break;
    end else begin
      Result := '';
      Inc(i);
    end;
  until((i > MaxInc) or (Result <> ''));

end;


//******************************************************************************
procedure TQUB_QFS_File.prSetCanEdit(e :Boolean);
begin
  if not prOpened then begin
    prCanEdit := e;
  end else begin
    if (prEditStream = nil) and (e) then begin
      prOpenEdit;
    end else if (prEditStream <> nil) and (not e) then begin
      prCanEdit := e;
      prEditStream.Free;
      prEditStream := nil;
      try
        DeleteFile(prEditFileName);
      except
      end;
    end;
  end;
end;

function TQUB_QFS_File.prOpenEdit;
var
  i :Integer;
  l :Integer;
  t :String;
  d :String;
  f :String;
begin
  Result := False;
  i := 0;
  if prEditStream <> nil then begin
    prEditStream.Free;
    prEditStream := nil;
    try
      DeleteFile(prEditFileName);
    except
    end;
  end;

  try
    d := ExtractFilePath(Application.ExeName) + 'Temp\';
    if not DirectoryExists(d) then begin
      if not CreateDirectory(PChar(d), nil) then begin
        d := ExtractFilePath(prFileName);
      end;
    end;
  except
    d := ExtractFilePath(prFileName);
  end;

  f := d + ExtractFileName(prFileName);

  repeat
    try
      prEditFileName := ChangeFileExt(f, '.~' + IntToStr(i));
      if FileExists(prEditFileName) then begin
        Inc(i);
        prEditStream.Free;
        prEditStream   := nil;
        prEditFileName := '';
      end else begin
        try
          prEditStream := TFileStream.Create(prEditFileName, fmCreate or fmShareDenyWrite);
        except
          l := 1024;
          SetLength(t, l);
          l := GetTempPath(l, PChar(t));
          if l = 0 then Break;
          SetLength(t, l);
          try
            prEditStream := TFileStream.Create(t + ExtractFileName(prEditFileName), fmCreate or fmShareDenyWrite);
            prEditFileName := t + ExtractFileName(prEditFileName);
          except
            Break;
          end;
        end;
      end;
    except
      Inc(i);
      prEditStream.Free;
      prEditStream   := nil;
      prEditFileName := '';
    end;
  until((i = 100) or (prEditStream <> nil));

  prCanEdit := prEditStream <> nil;
  if not prCanEdit then begin
    Console.NewEntry('File ' + FileName + ' was opened Read-Only.', COLOR_WARNING);
  end;

end;


//******************************************************************************
end.

