{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_QtiPlot;

interface

uses

Classes, SysUtils, ScktComp, WinSock, Windows, FileCtrl, Math, Forms,


Alloc, AllocTypes;


type


TQUB_QtiPlot_Connection = class(TObject)
protected

  prSock           :TClientSocket;
  prStream         :TWinSocketStream;
  prBuf            :PChar;
  prReadBytes      :Integer;
  prUsedBytes      :Integer;
  prLine           :PChar;

  prVerb           :String;
  prSData          :String;
  prBytes          :PChar;
  prByteCount      :Integer;

  procedure prReadBuf;
  function  prReadLine: String;
  procedure prReadMsg;
  procedure prVersionHandshake;
  function  prFindQtiPlot: Word;
  function  prStartQtiPlot: Word;

public

  constructor Create(Host :String = '127.0.0.1');
  destructor Destroy; override;

  procedure BringToFront;
  procedure Exec(Command :String);
  procedure ExecFile(Path :String);
  procedure ExecScript(Category, Name :String);
  function  Eval(Expr :String): String;
  procedure SetArray(Buf :Pointer; Rows, Cols :Integer; DType, Lbl :String);
  procedure GetArray(Buf :Pointer; var Rows, Cols :Integer; var DType, Lbl :String);


  class function HaveQtiPlot: Boolean;
  class function LocateQtiPlot: String;
  class procedure ListScripts(Category :String; Lst :TStrings);

end;


function QtiFormat(X :Double): String;
// Formats a decimal appropriate for python scripting glue

var

  ScriptDir        :String;


const

MAX_MSG_BYTES                = 1048575;
LINE_BYTES                   = 2048;
READ_BYTES                   = 64;
TIMEOUT_MS                   = 10000;
QTIPLOT_START_TIMEOUT_SEC    = 30;

implementation

uses

Registry,
RegistryInterface,
ShellAPI,
SystemPath,
PythonEngine,
QUB_Python_Thread,
QUB_Viewport_Python;


procedure UnpackArrayHeader(SData :String; var Rows, Cols :Integer; var DType, Lbl :String);
var
  Tab1, Tab2        :Integer;
  Fields            :Array of String;
  iField            :Integer;
begin
  Tab2              := 0;
  iField            := 0;
  SetLength(Fields, 4);
  while (iField < 4) and (Tab2 <= Length(SData)) do begin
    Tab1            := Tab2;
    while (Tab2 <= Length(SData)) and (SData[Tab2] <> #9) do Inc(Tab2);
    Fields[iField]  := Copy(SData, Tab1+1, Tab2-Tab1-1);
    Inc(iField);
  end;
  if iField > 0 then Rows   := StrToInt(Fields[0]);
  if iField > 1 then Cols   := StrToInt(Fields[1]);
  if iField > 2 then DType  := Fields[2];
  if iField > 3 then Lbl    := Fields[3];
end;


function SizeOfDType(DType :String): Integer;
begin
  if Pos('64', DType) > 0 then begin
    Result := 8;
  end else if Pos('32', DType) > 0 then begin
    Result := 4;
  end else if Pos('16', DType) > 0 then begin
    Result := 2;
  end else if Pos('8', DType) > 0 then begin
    Result := 1;
  end else begin
    raise Exception.Create('QtiPlot: Unknown dtype: '+DType);
  end;
end;

constructor TQUB_QtiPlot_Connection.Create(Host :String = '127.0.0.1');
var
  Port              :Word;
begin
  prBuf             := AllocMem(READ_BYTES);
  prBytes           := AllocMem(MAX_MSG_BYTES+1);
  prReadBytes       := READ_BYTES;
  prUsedBytes       := READ_BYTES;
  prLine            := AllocMem(LINE_BYTES+1);

  prSock            := TClientSocket.Create(nil);
  prSock.ClientType := ctBlocking;
  prSock.Address    := Host; // this one must be numeric string?  prSock.Host is DNS
  try
    Port            := prFindQtiPlot;
    prSock.Port     := Port;
    prSock.Open;
  except
    Port            := prStartQtiPlot;
    prSock.Port     := Port;
    prSock.Open;
  end;
  prStream          := TWinSocketStream.Create(prSock.Socket, TIMEOUT_MS);

  prVersionHandshake;
end;

destructor TQUB_QtiPlot_Connection.Destroy;
begin
  try
    prStream.Free;
    prSock.Close;
    prSock.Free;
    FreeMem(prBuf);
    FreeMem(prBytes);
    FreeMem(prLine);
  finally
    inherited;
  end;
end;

procedure TQUB_QtiPlot_Connection.BringToFront;
var
  HandleStr   :String;
  Handle      :HWnd;
begin
  HandleStr   := Eval('int(app.winId())');
  Handle      := HWnd(StrToInt(HandleStr));

  if IsIconic(Handle) then begin
    // give the user some peace // ShowWindow(Handle, SW_RESTORE);
  end else begin
    BringWindowToTop(Handle);
  end;
end;

procedure TQUB_QtiPlot_Connection.prReadBuf;
begin
  prReadBytes := prStream.Read(prBuf^, READ_BYTES);
  if prReadBytes = 0 then begin
    raise Exception.Create('QtiPlot server timeout');
  end;
  prUsedBytes := 0;
end;

function TQUB_QtiPlot_Connection.prReadLine: String;
var
  iLine     :Integer;
  c         :Char;
begin
  iLine     := 0;
  repeat begin
    if prReadBytes = prUsedBytes then prReadBuf;
    c := prBuf[prUsedBytes];
    Inc(prUsedBytes);
    if iLine = LINE_BYTES then begin
      raise Exception.Create('QtiPlot server line longer than '+IntToStr(LINE_BYTES));
    end;
    prLine[iLine] := c;
    Inc(iLine);
  end until c = #10;
  prLine[iLine-1] := #0;

  Result := String(prLine);
end;

procedure TQUB_QtiPlot_Connection.prReadMsg;
var
  Line    :String;
  Tab     :Integer;
  Rows, Cols :Integer;
  DType, Lbl :String;
  iByte       :Integer;
begin
  Line    := prReadLine;
  Tab     := Pos(#9, Line);
  if Tab = 0 then Tab := Length(Line) + 1;
  prVerb  := Copy(Line, 1, Tab-1);
  if Tab < Length(Line) then begin
    prSData := Copy(Line, Tab, Length(Line)-Tab);
  end else begin
    prSData := '';
  end;

  if (prVerb = 'VAL') or (prVerb = 'ERROR') then begin
    while True do begin
      Line := prReadLine;
      if Line = '.' then break;
      if prSData <> '' then prSData := prSData + #13#10;
      prSData := prSData + Line;
    end;
  end else if prVerb = 'ARRAY' then begin

    UnpackArrayHeader(prSData, Rows, Cols, DType, Lbl);
    prByteCount := Rows*Cols*SizeOfDType(DType);
    if prByteCount > MAX_MSG_BYTES then begin
      raise Exception.Create('QtiPlot: Array too large for buffer');
    end;
    iByte       := 0;
    while (iByte < prByteCount) and (prUsedBytes <> prReadBytes) do begin
      prBytes[iByte] := prBuf[prUsedBytes];
      Inc(prUsedBytes);
      Inc(iByte);
    end;
    while (iByte < prByteCount) do begin
      prReadBuf;
      prUsedBytes := Min(prByteCount-iByte, prReadBytes);
      Move(prBuf[0], prBytes[iByte], prUsedBytes);
      iByte := iByte + prUsedBytes;
    end;
    repeat Line := prReadLine until Line = '.';

  end else if prVerb = 'DONE' then begin     // nothing more to read
  end else if prVerb = 'SET' then begin
  end else if prVerb = 'VERSION' then begin
  end;

  if prVerb = 'ERROR' then begin
    raise Exception.Create('QtiPlot server: '+prSData);
  end;
end;

procedure TQUB_QtiPlot_Connection.prVersionHandshake;
var
  Msg               :String;
begin
  Msg               := 'VERSION'#9'1.0'#10;
  prStream.Write(Msg[1], Length(Msg));
  prReadMsg;
  if prVerb <> 'VERSION' then begin
    raise Exception.Create('QtiPlot server error (VERSION = '+prVerb+')');
  end;
end;

procedure TQUB_QtiPlot_Connection.Exec(Command :String);
var
  Msg               :String;
begin
  Msg               := 'EXEC'#10+Command+#10'.'#10;
  prStream.Write(Msg[1], Length(Msg));
  prReadMsg; // ERROR or DONE
end;

procedure TQUB_QtiPlot_Connection.ExecFile(Path :String);
var
  Fi                :TFileStream;
  Script            :PChar;
begin
  Script            := nil;
  Fi                := TFileStream.Create(Path, fmOpenRead or fmShareDenyNone);
  try
    Script          := AllocMem(Fi.Size + 1);
    Script[Fi.Size] := #0;
    Fi.Read(Script^, Fi.Size);
    Self.Exec(String(Script));
  finally
    FreeMem(Script);
    Fi.Free;
  end;
end;

procedure TQUB_QtiPlot_Connection.ExecScript(Category, Name :String);
begin
  ExecFile(ScriptDir+Category+'\'+Name);
end;

function  TQUB_QtiPlot_Connection.Eval(Expr :String): String;
var
  Msg               :String;
begin
  Msg               := 'EVAL'#9+Expr+#10;
  prStream.Write(Msg[1], Length(Msg));
  prReadMsg;
  Result            := prSData;
end;

procedure TQUB_QtiPlot_Connection.SetArray(Buf :Pointer; Rows, Cols :Integer;
                                           DType, Lbl :String);
var
  ASize             :Integer;
  Hdr               :String;
  Msg               :PChar;
begin
  Hdr               := Format('ARRAY'#9'%d'#9'%d'#9'%s'#9'%s'#10, [Rows, Cols, DType, Lbl]);
  ASize             := Rows * Cols * SizeOfDType(DType);
  Msg               := AllocMem(Length(Hdr) + ASize + 3);
  try
    Move(Hdr[1], Msg^, Length(Hdr));
    Move(Buf^, Msg[Length(Hdr)], ASize);
    Msg[Length(Hdr)+ASize  ] := #10;
    Msg[Length(Hdr)+ASize+1] := '.';
    Msg[Length(Hdr)+ASize+2] := #10;
    prStream.Write(Msg^, Length(Hdr)+ASize+3);
    prReadMsg;
  finally
    FreeMem(Msg);
  end;
end;

procedure TQUB_QtiPlot_Connection.GetArray(Buf :Pointer; var Rows, Cols :Integer;
                                           var DType, Lbl :String);
var
  Msg               :String;
begin
  Msg               := 'GET ARRAY'#9+Lbl+#10;
  prStream.Write(Msg[1], Length(Msg));
  prReadMsg;
  Move(prBytes^, Buf^, Min(prByteCount, Rows*Cols*SizeOfDType(DType)));
  UnpackArrayHeader(prSData, Rows, Cols, DType, Lbl);
end;

function  TQUB_QtiPlot_Connection.prFindQtiPlot: Word;
var
  Fi     :TextFile;
  Line   :String;
  Delim  :Integer;
  QtiPath :String;
begin
  PythonThread.BeginPython;
  GetPythonEngine.ExecString('import tempfile');
  QtiPath := GetPythonEngine.EvalStringAsStr('tempfile.gettempdir()');
  PythonThread.EndPython;
  AssignFile(Fi, QtiPath+'\qtiplot.sockid');
  Reset(Fi);
  ReadLn(Fi, Line);
  CloseFile(Fi);
  Delim  := Pos(#13, Line);
  if Delim = 0 then Delim := Pos(#10, Line);
  if Delim = 0 then Delim := Length(Line)+1;
  Result := StrToInt(Copy(Line, 1, Delim-1));
end;

function  TQUB_QtiPlot_Connection.prStartQtiPlot: Word;
var
  QtiPath        :String;
  Seconds        :Integer;
begin
  Result         := 0;
  PythonThread.BeginPython;
  GetPythonEngine.ExecString('import tempfile');
  QtiPath := GetPythonEngine.EvalStringAsStr('tempfile.gettempdir()');
  PythonThread.EndPython;
  QtiPath := QtiPath+'\qtiplot.sockid';
  if FileExists(QtiPath) then Sysutils.DeleteFile(QtiPath);

  QtiPath        := LocateQtiPlot;
  if QtiPath = '' then begin
    raise Exception.Create('QtiPlot is available from http://soft.proindependent.com/qtiplot.html');
  end;

  ShellExecute(Application.Handle, 'open', PChar(QtiPath + '\QtiPlot.exe'), nil, PChar(QtiPath), SW_SHOWNORMAL);

  Seconds := 0;
  while Seconds < QTIPLOT_START_TIMEOUT_SEC do begin
    try
      Result := prFindQtiPlot;
      Exit;
    except
      Sleep(1000);
      Inc(Seconds);
    end;
  end;
end;

class function TQUB_QtiPlot_Connection.HaveQtiPlot: Boolean;
begin
  Result := LocateQtiPlot <> '';
end;

class function TQUB_QtiPlot_Connection.LocateQtiPlot: String;
var
  Reg            :TRegistry;
  StartupScripts :String;
  AppPath        :String;
  PyPath         :String;
  PyVersion      :String;
begin
  Result         := '';
  StartupScripts := '';
  Reg            := TRegistry.Create;
  // first look for qub custom qtiplot distro, server already enabled
  Reg.RootKey  := HKEY_LOCAL_MACHINE;
  if not (ReadXRegistryString(Reg, 'SOFTWARE\QtiPlot', 'Install_Dir', Result)
          and FileExists(Result+'\qtiplot.exe')) then begin

    // next best: official QtiPlot distro:
    Reg.RootKey := HKEY_CURRENT_USER;
    if ReadXRegistryString(Reg, 'SOFTWARE\ProIndependent\QtiPlot\General\Paths', 'ScriptsDir', Result) then begin
      Result := StringReplace(Result, '/', '\', [rfReplaceAll]);
      if not FileExists(Result+'\qtiplot.exe') then begin
        Result   := '';
      end else begin
        if FileExists(Result+'\Python26.dll') then begin
          PyVersion := '2.6';
        end else if FileExists(Result+'\Python27.dll') then begin
          PyVersion := '2.7';
        end else begin
          Result := '';
        end;
      end;
    end;

    if Result <> '' then begin // Found official QtiPlot distro
      Reg.RootKey  := HKEY_LOCAL_MACHINE;
      if not ReadXRegistryString(Reg, 'SOFTWARE\Python\PythonCore\'+PyVersion+'\InstallPath', '', PyPath) then begin
        Reg.RootKey  := HKEY_CURRENT_USER;
        if not ReadXRegistryString(Reg, 'SOFTWARE\Python\PythonCore\'+PyVersion+'\InstallPath', '', PyPath) then begin
          raise Exception.Create('Please install Python '+PyVersion+' from'#13#10+'http://www.python.org/download/releases/'+PyVersion+'/');
        end;
      end;
      if not DirectoryExists(PyPath+'lib\site-packages\numpy') then begin
        raise Exception.Create('Please install numpy (for Python '+PyVersion+') from'#13#10+'http://sourceforge.net/projects/numpy/files/');
      end;
      Reg.RootKey := HKEY_CURRENT_USER;
      if ReadXRegistryString(Reg, 'SOFTWARE\ProIndependent\QtiPlot\General\Paths', 'StartupScripts', StartupScripts) then begin
        StartupScripts := StringReplace(StartupScripts, '/', '\', [rfReplaceAll]);
        // if we find one that's new enough to support startup scripts,
        if not DirectoryExists(StartupScripts) then CreateDir(StartupScripts);
        // copy in the remote control server and its startup script,
        // and set the default scripting lang to Python.
        AppPath := ExtractFilePath(Application.ExeName);
        if not FileExists(StartupScripts + '\qtiplot_start_remote_ctl.py') then begin
          CopyFile(PChar(AppPath + 'qtiplot_start_remote_ctl.py'),
                   PChar(StartupScripts + '\qtiplot_start_remote_ctl.py'), False);
        end;
        if not (FileExists(Result + '\qtiplot_remote_ctl.py') and FileExists(Result + '\remote_ctl.py')) then begin
          CopyFile(PChar(AppPath + 'qtiplot_remote_ctl.py'),
                   PChar(Result + '\qtiplot_remote_ctl.py'), False);
          CopyFile(PChar(AppPath + 'remote_ctl.py'),
                   PChar(Result + '\remote_ctl.py'), False);
        end;
        WriteXRegistryString(Reg, 'SOFTWARE\ProIndependent\QtiPlot\General', 'ScriptingLang', 'Python');
      end;
    end;
  end;

  Reg.CloseKey;
  Reg.Free;
end;


class procedure TQUB_QtiPlot_Connection.ListScripts(Category :String; Lst :TStrings);
var
  sr        :TSearchRec;
begin
  if 0 = FindFirst(ScriptDir+Category+'\*.py', 0, sr) then begin
    repeat
      Lst.Add(sr.Name);
    until 0 <> FindNext(sr);
  end;
end;



function QtiFormat(X :Double): String;
begin
  Result := StringReplace(Format('%.9g', [X]), ',', '.', [rfReplaceAll]);
end;


// *****************************************************************************
initialization

  ScriptDir         := ExtractFilePath(Application.ExeName)+'QtiPlotScripts\';



// *****************************************************************************
end.
