{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// non-visual scaffold of Results tree, used by QUB Visual Results


unit QUB_Results;


interface


uses


AllocTypes, Alloc, Classes, SysUtils, Math, IEEE754, FMath, Forms,

QUB_Global,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, PythonEngine,
QUB_Events;


const

SELECTVAR_BINCOUNT       = 16;
SELECTVAR_MIN_BINCOUNT   = 8;
SELECTVAR_MAX_BINCOUNT   = 128;
QUB_SELECT_COLOR_COUNT   = 10;


type

TQUB_Results_Point = class;
TQUB_Results_Var = class;
TQUB_SelVarPrefs = class;
TQUB_Results_Model = class;
TQUB_Results_Models = class;

TQUB_Results = class
protected

  prNode                :INode;
  prFileName            :String;

  prRefCount            :Integer;

  prSegments            :array of INode;
  prSegCount            :Integer;
  prSelColors           :array of INode; // segment['SelectedColor']
  prSelVars             :TStringList; // 'var i' -> TQUB_Results_Var
  prColorCount          :Integer;
  prColorCounts         :array of Integer;
  prModels              :TQUB_Results_Models;
  prHistGroups          :array of INode;

  prVarPrefs            :INode;
  prVsVisPrefs          :INode;
  prVsDefaultPrefs      :INode;
  prVsRecentPrefs       :INode;

  prEvents              :TQUB_Events;
  prDoUpdates           :Boolean;  // can suppress events e.g. when python or someone is multi-selecting

  procedure prInitialize;

  procedure prBuildSelVars;  // synchronized with python
  procedure prCleanupSelVars;
  procedure prAddPoint( Point :TQUB_Results_Point; SegIx :Integer );
  procedure prFillMissingPoints( SegIx :Integer );
  procedure prModelSearch(Node :INode);
  procedure prModelSearchHelper(PNode :TQUB_QFS_PNodeMem; ModelName, TrialName :Pointer);
  procedure prHistSearch(Node :INode);
  procedure prHistSearchHelper(PNode :TQUB_QFS_PNodeMem; HistName :Pointer);

  procedure prEnsureColor(i :Integer);
  procedure prSetFileName(NewName :String);
  function prGetModified: Boolean;
  function prGetFileIsOpen: Boolean;

  function prGetSegment(i :Integer): INode;
  function prGetSegColor(i :Integer): Integer;
  procedure prSetSegColor(i :Integer; c :Integer);
  function prGetSegsOfColor(i :Integer) :Integer;

  function prGetSelVarCount: Integer;
  function prGetSelVar(i :Integer): TQUB_Results_Var;
  function prGetSelVarByName(s :String): TQUB_Results_Var;

  function prGetHistGroupCount: Integer;
  function prGetHistGroup(i :Integer): INode;

  function prGetResultType: String;
  function prGetDateTime: TDateTime;

  procedure prSetDoUpdates(x :Boolean);


public

  constructor Create(ResultsNode :INode); overload;
  constructor Create(ResultsFileName :String; ReadOnly :Boolean); overload;
  destructor Destroy; override;

  procedure Incref;
  procedure Decref;

  function Save: Boolean;
  function SaveAs(NewName :String): Boolean;

  procedure UpdateSelection; // when DoUpdates was false, then you set it to true, UpdateSelection in the event thread


  property Node                    :INode                  read prNode;
  property FileName                :String                 read prFileName           write prSetFileName;
  property Modified                :Boolean                read prGetModified;
  property FileIsOpen              :Boolean                read prGetFileIsOpen;
  property ResultType              :String                 read prGetResultType;
  property DateTime                :TDateTime              read prGetDateTime;

  property Segment[i :Integer]     :INode                  read prGetSegment;
  property SegCount                :Integer                read prSegCount;
  property SegColor[i :Integer]    :Integer                read prGetSegColor        write prSetSegColor;
  property SegsOfColor[i :Integer] :Integer                read prGetSegsOfColor;

  property ColorCount              :Integer                read prColorCount;
  property SelVarCount             :Integer                read prGetSelVarCount;
  property SelVar[i :Integer]      :TQUB_Results_Var       read prGetSelVar;
  property SelVarByName[s :String] :TQUB_Results_Var       read prGetSelVarByName; // actually "VarNameIx" ('var i')

  property VsVisPrefs              :INode                  read prVsVisPrefs;     // which 'vs' charts were visible last time
  property VsDefaultPrefs          :INode                  read prVsDefaultPrefs; // which 'vs' charts are indicated by ResultNode['DefaultChartsVs']
  property VsRecentPrefs           :INode                  read prVsRecentPrefs;

  property Models                  :TQUB_Results_Models    read prModels;

  property HistGroupCount          :Integer                read prGetHistGroupCount;
  property HistGroup[i :Integer]   :INode                  read prGetHistGroup;

  property Events                  :TQUB_Events            read prEvents;
  // 'SelectionChanged' on write SegColor[i]
  property DoUpdates               :Boolean                read prDoUpdates          write prSetDoUpdates;


end;


TQUB_Results_Point = class
protected

  prYLog                           :Boolean;    // advisory for subclasses' prGetYValue

  function prGetColor: Integer; virtual; abstract;
  procedure prSetColor( x :Integer ); virtual; abstract;
  function prGetXValue: Double; virtual; abstract;
  procedure prSetXValue( x :Double ); virtual; abstract;
  function prGetYValue: Double; virtual; abstract;
  procedure prSetYValue( x :Double ); virtual; abstract;
  function prGetName: String; virtual; abstract;
  function prGetUnits: String; virtual; abstract;

public

  constructor Create;

  property YLog                    :Boolean                read prYLog              write prYLog;

  property Color                   :Integer                read prGetColor          write prSetColor;
  property XValue                  :Double                 read prGetXValue         write prSetXValue;
  property YValue                  :Double                 read prGetYValue         write prSetYValue;
  property Name                    :String                 read prGetName;
  property Units                   :String                 read prGetUnits;

end;


TQUB_Results_SegPoint = class( TQUB_Results_Point )
protected

  prColorNode                      :INode;
  prXValue                         :Double; // == Ix unless someone sets it to start time

  function prGetColor: Integer; override;
  procedure prSetColor( x :Integer ); override;
  function prGetXValue: Double; override;
  procedure prSetXValue( x :Double ); override;

public

  constructor Create( Ix :Integer; SegNode :INode );

end;


TQUB_Results_SegNullPoint = class( TQUB_Results_SegPoint )
protected

  prName           :String;

  function prGetYValue: Double; override;
  procedure prSetYValue( x :Double ); override;
  function prGetName: String; override;
  function prGetUnits: String; override;

public

  constructor Create( Ix :Integer; SegNode :INode; Name :String );

end;


TQUB_Results_SegVarPoint = class( TQUB_Results_SegPoint )
protected

  prValNode                        :INode;
  prValIx                          :Integer;

  function prGetYValue: Double; override;
  procedure prSetYValue( x :Double ); override;
  function prGetName: String; override;
  function prGetUnits: String; override;

public

  constructor Create( Ix :Integer; ValNode :INode; ValIx :Integer );

end;


TQUB_Results_SegRatePoint = class( TQUB_Results_SegPoint )
protected

  prRateNode                       :INode;    // Rates['Rate']['k0'] or ['k1']
  prRateDir                        :Integer; // 0 there, 1 back

  function prGetYValue: Double; override;
  procedure prSetYValue( x :Double ); override;
  function prGetName: String; override;
  function prGetUnits: String; override;

public

  constructor Create( Ix :Integer; RateNode :INode; RateDir :Integer ); // assumes Segment['ModelFile']['Rates']['Rate']['k0'] or k1

end;


TQUB_Results_Line = class                   // in Delphi, a class var can be like a record pointer but without nasty pascal pointer syntax
public
  Value           :Double;                  // these are owned by a Results_Var and taken up by a SelectChart
  Color           :Integer;
  Width           :Integer;
  Visible         :Boolean;
end;


TQUB_Results_XStyle  = (XSTYLE_INDEX, XSTYLE_START, XSTYLE_HIST);

TQUB_Results_Var = class(TPyObject)
protected
  prResults          :TQUB_Results;
  prPrefs            :INode;
  prPoints           :TList;

  prDoStats          :Boolean;
  prMean             :array of Double;      // per color
  prSD               :array of Double;
  prMedian           :array of Double;
  prHistData         :INode;

  prLowerSDLines     :TList;
  prMeanLines        :TList;
  prUpperSDLines     :TList;
  prNumSDForLines    :Double;

  prXStyle           :TQUB_Results_XStyle;

  prEvents           :TQUB_Events;

  function prGetVarName: String;
  function prGetUnits: String;

  function prGetXValue(ix :Integer) :Double;
  function prGetYValue(ix :Integer) :Double;
  procedure prSetYValue(ix :Integer; val :Double);
  function prGetCount: Integer;

  function  prGetYLog: Boolean;
  procedure prSetYLog(NewYLog :Boolean);
  procedure prSetXStyle(Style :TQUB_Results_XStyle);
  function prGetVisible: Boolean;
  procedure prSetVisible(x :Boolean);
  procedure prSetupHist;

  function prGetPoint(i :Integer): TQUB_Results_Point;
  function prGetLowerSDLine(i :Integer): TQUB_Results_Line;
  function prGetMeanLine(i :Integer): TQUB_Results_Line;
  function prGetUpperSDLine(i :Integer): TQUB_Results_Line;
  procedure prSetNumSDForLines(NumSD :Double);

  procedure prOnSelectionChanged(Sender :TObject);

public


  // consider it mandatory like a constructor (PythonForDelphi makes actual constructors too confusing)
  procedure Initialize( ResultsObj :TQUB_Results; PrefsNode :INode );
  procedure Add( Point :TQUB_Results_Point );
  procedure BeginStats;

  destructor Destroy; override;

  procedure OnHistBinsChanged(Sender :TObject);
  function GetHistAsNode: INode;
  function GetXAsNode: INode;
  function GetYAsNode: INode;

  // Python interface:
  function  GetAttr(key : PChar) : PPyObject; override;
  //function  SetAttr(key : PChar; value : PPyObject) : Integer; override;
  function  SqLength : Integer; override;
  function  SqItem( idx : Integer ) : PPyObject; override;
  class procedure RegisterMethods( PythonType : TPythonType ); override;
  function MeanPy( args :PPyObject ): PPyObject; cdecl;
  function SDPy( args :PPyObject ): PPyObject; cdecl;
  function MedianPy( args :PPyObject ): PPyObject; cdecl;
  function CountPy( args :PPyObject ): PPyObject; cdecl;

  // delphi interface:
  function Mean( color_ix :Integer ): Double;
  function SD( color_ix :Integer ): Double;
  function Median( color_ix :Integer): Double;

  property Results               :TQUB_Results  read prResults;
  property XValue[segi :Integer] :Double        read prGetXValue;
  property YValue[segi :Integer] :Double        read prGetYValue      write prSetYValue;   default;
  property Count                 :Integer       read prGetCount;

  property HistData              :INode         read prHistData; // has children X(bins) and Y(count)

  property Name                  :String        read prGetVarName;
  property Units                 :String        read prGetUnits;

  property YLog                  :Boolean       read prGetYLog         write prSetYLog;
  property XStyle                :TQUB_Results_XStyle read prXStyle    write prSetXStyle;
  property Visible               :Boolean       read prGetVisible      write prSetVisible;

  property Points[i :Integer]      :TQUB_Results_Point read prGetPoint;

  property LowerSDLine[i :Integer] :TQUB_Results_Line  read prGetLowerSDLine;
  property MeanLine[i :Integer]    :TQUB_Results_Line  read prGetMeanLine;
  property UpperSDLine[i :Integer] :TQUB_Results_Line  read prGetUpperSDLine;

  property NumSDForLines           :Double      read prNumSDForLines   write prSetNumSDForLines;

  property Events                  :TQUB_Events            read prEvents;
  //OnValuesChanged   :TNotifyEvent;  // e.g. when YLog changes
  //OnStatsChanged    :TNotifyEvent;  // e.g. when selection changes, and after OnValuesChanged
  //OnAxesChanged     :TNotifyEvent;

end;


TQUB_SelVarPrefs = class
protected
  tree       :INode;

  function prGetTypeNode(const ResultType :String): INode;
  procedure prInitialize;
public
  constructor Create;

  property TypeNode[const t :String]: INode               read prGetTypeNode; default;
end;

TQUB_Results_MdlPoint     = class( TQUB_Results_Point )
protected
  prName           :String;

  function prGetColor: Integer; override;
  procedure prSetColor( x :Integer ); override;
  function prGetXValue: Double; override;
  procedure prSetXValue( x :Double ); override;
  function prGetName: String; override;

public
  constructor Create(Name :String);
end;

TQUB_Results_MdlNullPoint = class( TQUB_Results_MdlPoint )
protected
  prValue         :Double;

  function prGetYValue: Double; override;
  procedure prSetYValue( x :Double ); override;
  function prGetUnits: String; override;

public
  constructor Create( Name :String );
  constructor CreateValued( Name :String; Val :Double );
end;


TQUB_Results_MdlRatePoint = class( TQUB_Results_MdlPoint )
protected
  prRateNode                       :INode;    // Rates['Rate']['k0'] or ['k1']
  prRateDir                        :Integer; // 0 there, 1 back

  function prGetYValue: Double; override;
  procedure prSetYValue( x :Double ); override;
  function prGetUnits: String; override;

public
  constructor Create( RateNode :INode; RateDir :Integer );
end;

TQUB_Results_MdlTrialPoint = class( TQUB_Results_MdlPoint )
protected
  prValNode                       :INode;
  prValIx                         :Integer;

  function prGetYValue: Double; override;
  procedure prSetYValue( x :Double ); override;
  function prGetUnits: String; override;

public
  constructor Create( ValNode :INode; ValIx :Integer );
end;


TQUB_Results_Model = class
protected

  prResults                     :TQUB_Results;
  prNode                        :INode;

  prVars                        :TStringList;

  function prGetValueCount: Integer;
  function prGetValue(Ix :Integer): Double;
  procedure prSetValue(Ix :Integer; Val :Double);
  function prGetValueByName(Ix :String): Double;
  procedure prSetValueByName(Ix :String; Val :Double);
  function prHasValue(Ix :String): Boolean;

public

  OnValueChanged                :procedure(Sender :TQUB_Results_Model; VarIx :Integer);

  constructor Create(Results :TQUB_Results; ModelNode :INode);
  destructor Destroy; override;

  procedure AddPoint(Point :TQUB_Results_Point);

  property ValueCount             :Integer                   read prGetValueCount;
  property Value[Ix :Integer]     :Double                    read prGetValue          write prSetValue;
  property ValueByName[Ix :String]:Double                    read prGetValueByName    write prSetValueByName;
  property HasValue[Ix :String]   :Boolean                   read prHasValue;
  property Node                   :INode                     read prNode;

end;

TQUB_Results_Models = class
protected

  prResults                     :TQUB_Results;
  prModels                      :TList;
  prVarNames                    :TStringList;
  prTrialCount                  :Integer;

  procedure prAddRatePoints(Model :TQUB_Results_Model);
  procedure prAddTrialPoints(Model :TQUB_Results_Model; Trial :INode);
  procedure prAddPoint(Model :TQUB_Results_Model; Point :TQUB_Results_Point);

  function prGetCount: Integer;
  function prGetModel(Ix :Integer): TQUB_Results_Model;

  function prGetVarCount: Integer;
  function prGetVarName(Ix :Integer): String;

public

  OnModelsSorted                :TNotifyEvent;

  CompareModelsBy               :Integer;
  SortModelsIncreasing          :Boolean;

  constructor Create(Results :TQUB_Results);
  destructor Destroy; override;

  procedure AddModel(ModelNode :INode);
  procedure AddModelTrial(ModelTrial :INode);
  procedure FinishAdding;

  procedure SortBy(VarName :String; Increasing :Boolean); overload;
  procedure SortBy(VarName :String); overload;

  property Count                :Integer                     read prGetCount;
  property Model[Ix :Integer]   :TQUB_Results_Model          read prGetModel;       default;

  property VarCount             :Integer                     read prGetVarCount;
  property VarName[Ix :Integer] :String                      read prGetVarName;

end;

procedure QUB_Results_SetHistBins(BC :Integer);
procedure QUB_Results_AddHistVar(HVar :TQUB_Results_Var);
procedure QUB_Results_RemHistVar(HVar :TQUB_Results_Var);

var

  ResultsVarPyType                      :TPythonType;

  SelVarPrefs                           :TQUB_SelVarPrefs;
  // SelVarPrefs[ResultType][VarNameIx] = Visible { XStyle = x; YLog = y }

  SelHistBinCount                       :Integer;
  SelHistBinEvents                      :TQUB_Events;
  SelHistVarCount                       :Integer;


implementation


uses


QUB_GUIInterface,
QUB_Profiles,
QUB_QFS_Channels,
QUB_Node_Model;


const

  SELVARS_FILENAME                         = 'qubselectprefs.qtr';


constructor TQUB_Results.Create(ResultsNode :INode);
begin
  prNode       := ResultsNode;
  prFileName   := ResultsNode.Name;
  prDoUpdates  := True;
  prRefCount   := 0;

  prInitialize;
end;

constructor TQUB_Results.Create(ResultsFileName :String; ReadOnly :Boolean);
begin

  if ReadOnly then begin
    prNode     := TNode.OpenReadOnly( ResultsFileName );
  end else begin
    prNode     := TNode.Open( ResultsFileName );
  end;

  prFileName := ResultsFileName;
  prRefCount := 0;

  prInitialize;
end;

procedure TQUB_Results.prInitialize;
var
  DS, Seg         :INode;
  i, j            :Integer;
  Prefs           :INode;
begin
  TreeNonFinitesToZero(prNode);

  prEvents        := TQUB_Events.Create;

  Prefs           := SelVarPrefs[ ResultType ];
  prVarPrefs      := Prefs['Vars'];
  Prefs           := Prefs['Versus'];
  prVsVisPrefs    := Prefs['Visible'];
  prVsRecentPrefs := Prefs['Recent'];
  prVsDefaultPrefs := Node.Find('DefaultChartsVs');

  prSegCount      := 0;
  DS              := prNode.Find('DataSet');
  while not DS.IsNull do begin
    Seg           := DS.Find('Segment');
    while not Seg.IsNull do begin
      prSegCount  := prSegCount + 1;
      Seg         := Seg.SiblingSameName;
    end;
    DS            := DS.SiblingSameName;
  end;

  prColorCount := QUB_SELECT_COLOR_COUNT;
  SetLength( prColorCounts, prColorCount );
  for i := 0 to prColorCount - 1 do begin
    prColorCounts[i] := 0;
  end;

  SetLength( prSegments, prSegCount );
  SetLength( prSelColors, prSegCount );

  i               := 0;
  DS              := prNode.Find('DataSet');
  while not DS.IsNull do begin
    Seg           := DS.Find('Segment');
    while not Seg.IsNull do begin
      prSegments[i]     := Seg;
      prSelColors[i]    := Seg['SelectedColor'];
      if prSelColors[i].DataCount = 0 then begin
        prSelColors[i].SetData( 1 ); // red
      end;

      j                 := Integer(prSelColors[i].Data[0]);
      prEnsureColor(j);
      prColorCounts[j]  := prColorCounts[j] + 1;

      i                 := i + 1;
      Seg               := Seg.SiblingSameName;
    end;
    DS            := DS.SiblingSameName;
  end;

  prSelVars            := TStringList.Create;
  prSelVars.Sorted     := True;
  prSelVars.Duplicates := dupError;

  try

    //while PythonThread.SyncPyProc( prBuildSelVars, False, 50 ) do Application.HandleMessage;
    PythonThread.SneakPyProc( prBuildSelVars );

    prModels        := TQUB_Results_Models.Create(Self);
    prModelSearch( prNode );
    prModels.FinishAdding;

    prHistSearch( prNode );

  except
  end;

end;

procedure TQUB_Results.prBuildSelVars;
var
  DS, Seg     :INode;
  SegIx       :Integer;
  i           :Integer;
  VarNode     :INode;
  RateNode    :INode;
begin
  SegIx                := 0;
  DS                   := prNode.Find('DataSet');
  while not DS.IsNull do begin
    Seg                := DS.Find('Segment');
    while not Seg.IsNull do begin
      VarNode          := Seg.Child;
      while not VarNode.IsNull do begin
        if (VarNode.DataCount > 0) and (VarNode.DataCount < 200) and (VarNode.DataType <> NODE_DATATYPE_STRING) and (VarNode.Child.Sibling.IsNull) and (VarNode.Name <> 'SelectedColor') then begin
          for i := 0 to VarNode.DataCount - 1 do begin
            prAddPoint( TQUB_Results_SegVarPoint.Create( SegIx, VarNode, i ), SegIx );
          end;
        end;

        VarNode        := VarNode.Sibling;
      end;

      VarNode          := Seg.Find('ModelFile').Find('Rates').Find('Rate');
      while not VarNode.IsNull do begin
        RateNode       := VarNode['k0'];
        prAddPoint( TQUB_Results_SegRatePoint.Create( SegIx, RateNode, 0 ), SegIx );
        prAddPoint( TQUB_Results_SegRatePoint.Create( SegIx, RateNode, 1 ), SegIx );

        RateNode       := VarNode['k1'];
        if RateNode.Data[0] <> 0.0 then prAddPoint( TQUB_Results_SegRatePoint.Create( SegIx, RateNode, 0 ), SegIx );
        if RateNode.Data[1] <> 0.0 then prAddPoint( TQUB_Results_SegRatePoint.Create( SegIx, RateNode, 1 ), SegIx );

        VarNode        := VarNode.SiblingSameName;
      end;

      prFillMissingPoints( SegIx );

      SegIx            := SegIx + 1;
      Seg              := Seg.SiblingSameName;
    end;

    DS                 := DS.SiblingSameName;
  end;

  for i := 0 to SelVarCount - 1 do begin
    SelVar[i].BeginStats;
  end;
end;

procedure TQUB_Results.prAddPoint( Point :TQUB_Results_Point; SegIx :Integer );
var
  SelVar               :TQUB_Results_Var;
  PyInst               :PPyObject;
  ix                   :Integer;
  Name                 :String;
begin
  Name                 := Point.Name;
  SelVar               := SelVarByName[ Name ];
  if SelVar = nil then begin
    PyInst             := ResultsVarPyType.CreateInstance;
    SelVar             := TQUB_Results_Var( PythonToDelphi(PyInst) );
    SelVar.Initialize( Self, prVarPrefs[ Name ] );
    ix                 := prSelVars.Add( Name );
    prSelVars.Objects[ix] := SelVar;
  end;

  for ix := SelVar.Count to SegIx - 1 do begin
    SelVar.Add( TQUB_Results_SegNullPoint.Create(ix, Segment[ix], Name) );
  end;

  SelVar.Add( Point );
end;

procedure TQUB_Results.prFillMissingPoints( SegIx :Integer );
var
  i                    :Integer;
begin
  for i := 0 to SelVarCount - 1 do begin
    while SelVar[i].Count <= SegIx do begin
      SelVar[i].Add( TQUB_Results_SegNullPoint.Create( SelVar[i].Count, Segment[SelVar[i].Count], SelVar[i].Name ) );
    end;
  end;
end;
(*
procedure TQUB_Results.prHistSearch(Node :INode);
var
  ch      :INode;
  ix      :Integer;
begin
  ch := Node.Child;
  while not ch.IsNull do begin
    if ch.Name = 'HistogramGroup' then begin
      ix := Length(prHistGroups);
      SetLength( prHistGroups, ix + 1 );
      prHistGroups[ix] := ch;
    end else if ch.Name = 'ModelFile' then begin
      // don't bother searching it
    end else begin
      prHistSearch( ch );
    end;

    ch := ch.Sibling;
  end;
end;
*)
procedure TQUB_Results.prHistSearch(Node :INode);
begin
  if Node.IsNull then Exit;
  prHistSearchHelper(Node.GetNodeMem, QTR_LookupName('HistogramGroup'));
end;

procedure TQUB_Results.prHistSearchHelper(PNode :TQUB_QFS_PNodeMem; HistName :Pointer);
var
  PChild     :TQUB_QFS_PNodeMem;
  ix         :Integer;
  Hist       :INode;
  Valid      :Boolean;
begin
  PChild     := PNode.ChildNode;
  while PChild <> nil do begin
    if Pointer(PChild.Name) = HistName then begin
      Hist := TNode.CreateUsing(PChild);
      Hist := Hist.Find('Histogram');
      Valid := not Hist.IsNull;
      while not Hist.IsNull do begin
        if (Hist.Find('Bins').DataCount < 2) or (Hist.Find('Bars').DataCount < 2) then begin
          Valid := False;
          Break;
        end;
        Hist := Hist.SiblingSameName;
      end;
      if Valid then begin
        ix := Length(prHistGroups);
        SetLength( prHistGroups, ix + 1 );
        prHistGroups[ix] := TNode.CreateUsing(PChild);
      end;
    end else begin
      prHistSearchHelper(PChild, HistName);
    end;

    PChild   := PChild.SiblingNode;
  end;
end;

procedure TQUB_Results.prModelSearch(Node :INode);
begin
  prModelSearchHelper(Node.GetNodeMem, QTR_LookupName('ModelFile'), QTR_LookupName('ModelTrial'));
end;

procedure TQUB_Results.prModelSearchHelper(PNode :TQUB_QFS_PNodeMem; ModelName, TrialName :Pointer);
var
  PChild     :TQUB_QFS_PNodeMem;
begin
  if PNode = nil then Exit;
  PChild     := PNode.ChildNode;
  while PChild <> nil do begin
    if Pointer(PChild.Name) = ModelName then begin
      prModels.AddModel( TNode.CreateUsing(PChild) );
    end else if Pointer(PChild.Name) = TrialName then begin
      prModels.AddModelTrial( TNode.CreateUsing(PChild) );
    end else begin
      prModelSearchHelper(PChild, ModelName, TrialName);
    end;

    PChild   := PChild.SiblingNode;
  end;
end;
(*
procedure TQUB_Results.prModelSearch(Node :INode);
var
  ch      :INode;
begin
  ch := Node.Child;
  while not ch.IsNull do begin
    if ch.Name = 'HistogramGroup' then begin
      // don't bother searching it
    end else if ch.Name = 'ModelTrial' then begin
      prModels.AddModelTrial( ch );
    end else if ch.Name = 'ModelFile' then begin
      prModels.AddModel( ch );
    end else begin
      prModelSearch( ch );
    end;

    ch := ch.Sibling;
  end;
end;
*)
destructor TQUB_Results.Destroy;
begin
  try
    if FileIsOpen then Save;
  except
  end;
  
  try
    //while PythonThread.SyncPyProc( prCleanupSelVars, False, 50 ) do Application.HandleMessage;
    PythonThread.SneakPyProc( prCleanupSelVars );

    prEvents.Free;

    prSegCount := 0;
    SetLength( prSegments, 0 );
    SetLength( prSelColors, 0 );

    prNode := nil;
  finally
    inherited;
  end;
end;

procedure TQUB_Results.prCleanupSelVars;
var
  i          :Integer;
  SelVar     :TQUB_Results_Var;
begin
  with GetPythonEngine do begin
    for i := 0 to prSelVars.Count - 1 do begin
      SelVar := TQUB_Results_Var( prSelVars.Objects[i] );
      Py_DECREF( SelVar.GetSelf );
    end;
  end;
  prSelVars.Clear;
end;

procedure TQUB_Results.Incref;
begin
  Inc( prRefCount );
end;

procedure TQUB_Results.Decref;
begin
  Dec( prRefCount );
  if prRefCount < 1 then Self.Free;
end;

function TQUB_Results.Save: Boolean;
begin
  Result := prNode.Save;
end;

function TQUB_Results.SaveAs(NewName :String): Boolean;
begin
  FileName := NewName;

  if ExtractFileExt(FileName) = '.qrt' then begin
    prNode.Close;
    prNode.SaveAsText(FileName);
    Result := False;
  end else begin
    Result := not prNode.SaveAs( NewName );
  end;
end;

procedure TQUB_Results.prSetFileName(NewName :String);
begin
  prNode.Close;
  prFileName := NewName;
end;

function TQUB_Results.prGetModified: Boolean;
begin
  Result := prNode.Changed or prNode.ChildChanged;
end;

function TQUB_Results.prGetFileIsOpen: Boolean;
begin
  Result := prNode.IsInFile;
end;

procedure TQUB_Results.prEnsureColor(i :Integer);
begin
  if i >= prColorCount then begin
    SetLength(prColorCounts, i+1);
    prColorCount := i+1;
  end;
end;

function TQUB_Results.prGetSegment(i :Integer): INode;
begin
  if (i >= 0) and (i < prSegCount) then begin
    Result := prSegments[i];
  end else begin
    Result := TNode.CreateNull;
  end;
end;

function TQUB_Results.prGetSegColor(i :Integer): Integer;
begin
  Result := 0;
  if (i >= 0) and (i < prSegCount) then begin
    Result := Integer(prSelColors[i].Data[0]);
  end;
end;

procedure TQUB_Results.prSetSegColor(i :Integer; c :Integer);
begin
  if (i >= 0) and (i < prSegCount) then begin
    prEnsureColor(c);
    prColorCounts[ Integer(prSelColors[i].Data[0]) ] := prColorCounts[ Integer(prSelColors[i].Data[0]) ] - 1;
    prColorCounts[ c ] := prColorCounts[ c ] + 1;
    prSelColors[i].Data[0] := c;
    prSelColors[i].Changed := True;
    if DoUpdates then Events.RaiseNotifyEvent('SelectionChanged', Self);
  end;
end;

function TQUB_Results.prGetSegsOfColor(i :Integer) :Integer;
begin
  Result := 0;
  if (i >= 0) and (i < prColorCount) then begin
    Result := prColorCounts[i];
  end;
end;

function TQUB_Results.prGetSelVarCount: Integer;
begin
  Result := prSelVars.Count;
end;

function TQUB_Results.prGetSelVar(i :Integer): TQUB_Results_Var;
begin
  Result := nil;
  if (i >= 0) and (i < prSelVars.Count) then begin
    Result := TQUB_Results_Var( prSelVars.Objects[i] );
  end;
end;

function TQUB_Results.prGetSelVarByName(s :String): TQUB_Results_Var;
begin
  Result := SelVar[ prSelVars.IndexOf(s) ];
end;

function TQUB_Results.prGetHistGroupCount: Integer;
begin
  Result := Length(prHistGroups);
end;

function TQUB_Results.prGetHistGroup(i :Integer): INode;
begin
  if (i >= 0) and (i < HistGroupCount) then begin
    Result := prHistGroups[i];
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results.prGetResultType: String;
begin
  Result := Node['ResultType'].DataAsString;
end;

function TQUB_Results.prGetDateTime: TDateTime;
begin
  try
    Result := StrToDateTime(Node['Date'].DataAsString);
  except
    Result := Now;
  end;
end;

procedure TQUB_Results.prSetDoUpdates(x :Boolean);
begin
  prDoUpdates := x;
end;

procedure TQUB_Results.UpdateSelection;
begin
  Events.RaiseNotifyEvent('SelectionChanged', Self);
end;

// *****************************************************************************

constructor TQUB_Results_Point.Create;
begin
  prYLog   := False;
end;

constructor TQUB_Results_SegPoint.Create( Ix :Integer; SegNode :INode );
begin
  inherited Create;

  prColorNode  := SegNode['SelectedColor'];
  if prColorNode.DataCount = 0 then begin
    prColorNode.SetData( 1 );
  end;

  prXValue     := Ix + 1;
end;

function TQUB_Results_SegPoint.prGetColor: Integer;
begin
  Result := Integer( prColorNode.Data[0] );
end;

procedure TQUB_Results_SegPoint.prSetColor( x :Integer );
begin
  prColorNode.Data[0] := x;
end;

function TQUB_Results_SegPoint.prGetXValue: Double;
begin
  Result := prXValue;
end;

procedure TQUB_Results_SegPoint.prSetXValue( x :Double );
begin
  prXValue := x;
end;

constructor TQUB_Results_SegNullPoint.Create( Ix :Integer; SegNode :INode; Name :String );
begin
  inherited Create( Ix, SegNode );

  prName := Name;
end;

function TQUB_Results_SegNullPoint.prGetYValue: Double;
begin
  Result := 0.0;
end;

procedure TQUB_Results_SegNullPoint.prSetYValue( x :Double );
begin
end;

function TQUB_Results_SegNullPoint.prGetName: String;
begin
  Result := prName;
end;

function TQUB_Results_SegNullPoint.prGetUnits: String;
begin
  if YLog then Result := 'log10' else Result := '';
end;

constructor TQUB_Results_SegVarPoint.Create( Ix :Integer; ValNode :INode; ValIx :Integer );
begin
  inherited Create( Ix, ValNode.Parent );

  prValNode       := ValNode;
  prValIx         := ValIx;
end;

function TQUB_Results_SegVarPoint.prGetYValue: Double;
begin
  try
    if prValNode.DataType = NODE_DATATYPE_DOUBLE then begin
      Result      := TPADouble(prValNode.DataPtr)[prValIx];
    end else if prValNode.DataType = NODE_DATATYPE_LONGINT then begin
      Result      := TPAInteger(prValNode.DataPtr)[prValIx];
    end else begin
      Result        := Double( prValNode.Data[ prValIx ] );
    end;
    if YLog and (Result > 0.0) then Result := Log10( Result );
  except
    Result          := 0.0;
  end;
  if IsNan(Result) or IsInfinity(Result) then Result := 0;
end;

procedure TQUB_Results_SegVarPoint.prSetYValue( x :Double );
begin
  try
    prValNode.Data[ prValIx ] := x;
  except
  end;
end;

function TQUB_Results_SegVarPoint.prGetName: String;
begin
  if prValNode.Cols > 1 then begin
    Result := Format('%s %d,%d', [prValNode.Name, (prValIx div LongInt(prValNode.Cols)) + 1, (prValIx mod LongInt(prValNode.Cols)) + 1]);
  end else if prValNode.DataCount > 1 then begin
    Result := Format('%s %d', [prValNode.Name, prValIx + 1]);
  end else begin
    Result := prValNode.Name;
  end;
end;

function TQUB_Results_SegVarPoint.prGetUnits: String;
begin
  Result := prValNode.LineComment;
  if YLog then Result := 'log10 ' + Result;
end;

constructor TQUB_Results_SegRatePoint.Create( Ix :Integer; RateNode :INode; RateDir :Integer );
begin
  inherited Create( Ix, RateNode.Parent.Parent.Parent.Parent );  // k0 -> Rate -> Rates -> ModelFile -> Segment

  prRateNode        := RateNode;
  prRateDir         := RateDir;
end;

function TQUB_Results_SegRatePoint.prGetYValue: Double;
begin
  try
    Result        := Double( prRateNode.Data[ prRateDir ] );
    if YLog then Result := Log10( Result );
  except
    Result          := 0.0;
  end;
  if IsNan(Result) or IsInfinity(Result) then Result := 0;
end;

procedure TQUB_Results_SegRatePoint.prSetYValue( x :Double );
begin
  try
    prRateNode.Data[ prRateDir ] := x;
  except
  end;
end;

function TQUB_Results_SegRatePoint.prGetName: String;
var
  States   :INode;
begin
  Result := prRateNode.Name;

  States   := prRateNode.Parent['States'];
  if prRateDir = 0 then begin
    Result := Result + Format(' %d->%d', [Integer(States.Data[0])+1, Integer(States.Data[1])+1]);
  end else begin
    Result := Result + Format(' %d->%d', [Integer(States.Data[1])+1, Integer(States.Data[0])+1]);
  end;
end;

function TQUB_Results_SegRatePoint.prGetUnits: String;
begin
  if YLog then Result := 'log10';
end;

// *****************************************************************************

procedure TQUB_Results_Var.Initialize( ResultsObj :TQUB_Results; PrefsNode :INode );
var
  i                       :Integer;
  Node                    :INode;
begin
  prResults               := ResultsObj;
  prPoints                := TList.Create;
  prDoStats               := False;
  prHistData              := TNode.Create('Hist');

  prPrefs                 := PrefsNode;
  if prPrefs.IsNull then prPrefs := TNode.Create('');
  if prPrefs.DataCount = 0 then prPrefs.SetData( 0 ); // not visible
  Node                    := prPrefs['XStyle'];
  if Node.DataCount = 0 then Node.SetData( Integer(XSTYLE_INDEX) );
  Node                    := prPrefs['YLog'];
  if Node.DataCount = 0 then Node.SetData( 0 );

  SetLength( prMean, prResults.ColorCount );
  SetLength( prSD, prResults.ColorCount );
  SetLength( prMedian, prResults.ColorCount );

  prXStyle                := XStyle;

  prLowerSDLines          := TList.Create;
  prMeanLines             := TList.Create;
  prUpperSDLines          := TList.Create;
  prNumSDForLines         := 1.0;

  for i := 0 to QUB_SELECT_COLOR_COUNT - 1 do begin
    prLowerSDLines.Add( TQUB_Results_Line.Create );
    prMeanLines.Add( TQUB_Results_Line.Create );
    prUpperSDLines.Add( TQUB_Results_Line.Create );

    LowerSDLine[i].Color := i;
    MeanLine[i].Color    := i;
    UpperSDLine[i].Color := i;

    LowerSDLine[i].Width := 1;
    MeanLine[i].Width    := 2;
    UpperSDLine[i].Width := 1;
  end;

  prResults.Events.AddEvent('SelectionChanged', prOnSelectionChanged);

  prEvents               := TQUB_Events.Create;
end;

procedure TQUB_Results_Var.Add( Point :TQUB_Results_Point );
begin
  Point.YLog := YLog;
  prPoints.Add( Point );
  if prDoStats then prOnSelectionChanged(Self);
end;

procedure TQUB_Results_Var.BeginStats;
begin
  prDoStats        := True;
  prOnSelectionChanged(Self);
end;

destructor TQUB_Results_Var.Destroy;
var
  i        :Integer;
begin
  try
    if XStyle = XSTYLE_HIST then QUB_Results_RemHistVar(Self);

    for i := 0 to Count - 1 do begin
      TQUB_Results_Point(prPoints[i]).Free;
    end;
    prPoints.Free;

    for i := 0 to QUB_SELECT_COLOR_COUNT - 1 do begin
      TQUB_Results_Line(LowerSDLine[i]).Free;
      TQUB_Results_Line(MeanLine[i]).Free;
      TQUB_Results_Line(UpperSDLine[i]).Free;
    end;
    prLowerSDLines.Free;
    prMeanLines.Free;
    prUpperSDLines.Free;

    prEvents.Free;
  finally
    inherited;
  end;
end;

function BureaucraticNightmare(Item1, Item2: Pointer): Integer;
// See discussion/denunciation in next function
var
  a :Double;
  b :Double;
begin

  a := TQUB_Results_Point(Item1).YValue;
  b := TQUB_Results_Point(Item2).YValue;

  if a < b then begin
    Result := -1;
  end else if a = b then begin
    Result := 0;
  end else begin
    Result := 1;
  end;

end;

procedure TQUB_Results_Var.prOnSelectionChanged;
var
  i, j, ColorCount                             :Integer;
  diff                                         :Double;
  // I want to compute the median for each color
  // by listing values of color c, sorting, and
  // taking the central (/average of two central) value.
  // Unfortunately Delphi is so hideously incomplete that
  // we need a custom function and pointer tricks:
  Sorter                                       :TList;
  //function StupidDelphiCompare(Item1, Item2: Pointer): Integer;
  //begin
  //  Result := Float(Item1) - Float(Item2);
  //end;
  // Never mind; Delphi won't even allow those type casts
  //function BureaucraticNightmare(Item1, Item2: Pointer): Integer;
  //var a, b :Double;
  //begin
  //  a := TQUB_Results_Point(Item1).YValue;
  //  b := TQUB_Results_Point(Item2).YValue;
  //  if a < b then begin
  //    Result := -1;
  //  end else if a = b then begin
  //    Result := 0;
  //  end else begin
  //    Result := 1;
  //  end;
  //end;
  // Never-mind * 2: Error: Local procedure/function assigned to procedure variable.  FUck M4e
begin
  if Length(prMean) < Results.ColorCount then begin
    SetLength(prMean, Results.ColorCount);
    SetLength(prSD,   Results.ColorCount);
    SetLength(prMedian, Results.ColorCount);
  end;

  Sorter := TList.Create;
  for i := 0 to Results.ColorCount - 1 do begin
    if 0 = Results.SegsOfColor[i] then begin
      prMedian[i] := 0.0;
    end else begin
      Sorter.Clear;
      for j := 0 to Count - 1 do begin
        if i = Results.SegColor[j] then begin
          Sorter.Add(prPoints[j]);
        end;
      end;
      Sorter.Sort(BureaucraticNightmare);
      j := Sorter.Count div 2;
      if (Sorter.Count mod 2) = 0 then begin
        prMedian[i] := (TQUB_Results_Point(Sorter[j]).YValue + TQUB_Results_Point(Sorter[j-1]).YValue) / 2;
      end else begin
        prMedian[i] := TQUB_Results_Point(Sorter[j]).YValue;
      end;
    end;
  end;
  Sorter.Free;

  for i := 0 to Results.ColorCount - 1 do begin
    prMean[i] := 0.0;
    prSD[i]   := 0.0;
  end;

  for i := 0 to Count - 1 do begin
    j := Results.SegColor[i];
    prMean[j] := prMean[j] + YValue[i];
  end;

  for i := 0 to Results.ColorCount - 1 do begin
    ColorCount := prResults.SegsOfColor[i];
    if ColorCount > 0 then begin
      prMean[i] := prMean[i] / ColorCount;
    end;
  end;

  // TODO 7-6-2004 Overflow for albert on sleep kinetics (diff = +/- 10^200)
  // Open model file, open .txt, rt click - open list - same .txt file
  // mil 100x , 'together' - overflows on prSD below
  for i := 0 to Count - 1 do begin
    j := Results.SegColor[i];
    diff    := prMean[j] - YValue[i];
    prSD[j] := prSD[j] + diff * diff;
  end;

  for i := 0 to Results.ColorCount - 1 do begin
    ColorCount := prResults.SegsOfColor[i];
    if ColorCount > 0 then prSD[i] := Sqrt( prSD[i] / ColorCount );
  end;

  for i := 0 to QUB_SELECT_COLOR_COUNT - 1 do begin
    LowerSDLine[i].Value   := prMean[i] - prNumSDForLines * prSD[i];
    LowerSDLine[i].Visible := prResults.SegsOfColor[i] > 0;

    MeanLine[i].Value      := prMean[i];
    MeanLine[i].Visible    := LowerSDLine[i].Visible;

    UpperSDLine[i].Value   := prMean[i] + prNumSDForLines * prSD[i];
    UpperSDLine[i].Visible := LowerSDLine[i].Visible;
  end;

  prEvents.RaiseNotifyEvent('OnStatsChanged', Self);
end;

function TQUB_Results_Var.prGetVarName: String;
begin
  Result := '';
  if Count > 0 then Result := Points[0].Name;
end;

function TQUB_Results_Var.prGetUnits: String;
begin
  Result := '';
  if Count > 0 then Result := Points[0].Units;
end;

function TQUB_Results_Var.prGetXValue(ix :Integer) :Double;
begin
  Result := Points[ix].XValue;
end;

function TQUB_Results_Var.prGetYValue(ix :Integer) :Double;
begin
  Result := Points[ix].YValue;
end;

procedure TQUB_Results_Var.prSetYValue(ix :Integer; val :Double);
begin
  Points[ix].YValue := val;
  prEvents.RaiseNotifyEvent('OnValuesChanged', Self);
  prOnSelectionChanged(Self);
end;

function TQUB_Results_Var.prGetCount: Integer;
begin
  Result := prPoints.Count;
end;

function TQUB_Results_Var.prGetYLog: Boolean;
begin
  Result := 0 <> prPrefs['YLog'].Data[0];
end;

procedure TQUB_Results_Var.prSetYLog(NewYLog :Boolean);
var
  i       :Integer;
begin
  if NewYLog <> YLog then begin
    if NewYLog then begin
      prPrefs['YLog'].Data[0] := 1;
    end else begin
      prPrefs['YLog'].Data[0] := 0;
    end;

    for i := 0 to Count - 1 do begin
      Points[i].YLog := NewYLog;
    end;
    
    prEvents.RaiseNotifyEvent('OnValuesChanged', Self);
    prOnSelectionChanged(Self);
    prEvents.RaiseNotifyEvent('OnAxesChanged', Self);
  end;
end;

procedure MakeBins(Data :TPADouble; Bins :TPADouble; BinOfData :TPAInteger; DataCount, BinCount :Integer);
var
  i, b                :Integer;
  imax, imin          :Double;
  binwidth, halfwidth :Double;
begin
  imax := Data[0];
  imin := Data[0];
  for i := 1 to DataCount - 1 do begin
    if imax < Data[i] then imax := Data[i];
    if imin > Data[i] then imin := Data[i];
  end;

  binwidth := (imax - imin) / BinCount;
  if binwidth = 0.0 then binwidth := 1.0; // avoid divide by 0
  halfwidth := binwidth / 2.0;

  for i := 0 to BinCount - 1 do begin
    Bins[i] := imin + halfwidth + i * binwidth;
  end;

  for i := 0 to DataCount - 1 do begin
    b := Floor( (Data[i] - imin) / binwidth );
    if b <  0        then b := 0;
    if b >= BinCount then b := BinCount - 1;
    BinOfData[i] := b;
  end;
end;

procedure TQUB_Results_Var.prSetXStyle(Style :TQUB_Results_XStyle);
var
  i                :Integer;
  Point            :TQUB_Results_Point;
begin
  if Style <> prXStyle then begin
    if prXStyle = XSTYLE_HIST then QUB_Results_RemHistVar(Self);
    if Style    = XSTYLE_HIST then QUB_Results_AddHistVar(Self);

    prXStyle := Style;
    prPrefs['XStyle'].Data[0] := Integer( Style );

    if Style = XSTYLE_INDEX then begin
      for i := 0 to Count - 1 do begin
        Point := Points[i];
        Point.XValue := i + 1;
      end;
    end else if Style = XSTYLE_START then begin
      for i := 0 to Count - 1 do begin
        try
          Points[i].XValue := 1.0e-3 * Double(Results.Segment[i]['start'].Data[0]);
        except
        end;
      end;
    end else if Style = XSTYLE_HIST then begin
      prSetupHist;
    end;

    //if Assigned(OnValuesChanged) then OnValuesChanged(Self);
    //if Assigned(OnAxesChanged) then OnAxesChanged(Self);
    prEvents.RaiseNotifyEvent('OnValuesChanged', Self);
    prEvents.RaiseNotifyEvent('OnAxesChanged', Self);
  end;
end;

procedure TQUB_Results_Var.prSetupHist;
var
  i, j             :Integer;
  vals             :array of Double;
  binOfSeg         :array of Integer;
  HistX, HistY     :INode;
begin
  SetLength( vals, Count );
  SetLength( binOfSeg, Count );

  HistX := prHistData['X'];
  HistY := prHistData['Y'];
  if LongInt(HistX.DataCount) <> SelHistBinCount then begin
    HistX.SetupNumData( NODE_DATATYPE_DOUBLE, SelHistBinCount, 1 );
    HistY.SetupNumData( NODE_DATATYPE_DOUBLE, SelHistBinCount, 1 );
  end;

  for i := 0 to Count - 1 do begin
    vals[i] := YValue[i];
  end;

  MakeBins(@(vals[0]), HistX.DataPtr{@(bins[0])}, @(binOfSeg[0]), Count, SelHistBinCount);

  for i := 0 to SelHistBinCount - 1 do begin
    HistY.Data[i] := 0; // counts[i] := 0;
  end;
  for i := 0 to Count - 1 do begin
    j   := binOfSeg[i];
    HistY.Data[j] := HistY.Data[j] + 1; // counts[j] := counts[j] + 1;

    Points[i].XValue     := HistY.Data[j] - 0.25; // counts[j];
    // points[i].yvalue?
  end;
end;

procedure TQUB_Results_Var.OnHistBinsChanged(Sender :TObject);
begin
  prSetupHist;
  prEvents.RaiseNotifyEvent('OnValuesChanged', Self);
end;

function TQUB_Results_Var.GetHistAsNode: INode;
begin
  Result := TNode.Create('Histogram');
  Result['Title'].SetData( Name );
  Result['XLabel'].SetData( Units );
  Result['BinCount'].SetData( SelHistBinCount );
  Result.AppendClone( HistData['X'], False ).Name := 'Bins';
  Result.AppendClone( HistData['Y'], False ).Name := 'Bars';
end;

function TQUB_Results_Var.GetXAsNode: INode;
var
  i          :Integer;
begin
  Result     := TNode.Create(Name);
  Result.SetupNumData(NODE_DATATYPE_SINGLE, Count, 1);
  for i := 0 to Count - 1 do begin
    Result.Data[i] := XValue[i];
  end;
end;

function TQUB_Results_Var.GetYAsNode: INode;
var
  i          :Integer;
begin
  Result     := TNode.Create(Name);
  Result.SetupNumData(NODE_DATATYPE_SINGLE, Count, 1);
  for i := 0 to Count - 1 do begin
    Result.Data[i] := YValue[i];
  end;
end;

function TQUB_Results_Var.prGetVisible: Boolean;
begin
  Result := 0 <> prPrefs.Data[0];
end;

procedure TQUB_Results_Var.prSetVisible(x :Boolean);
begin
  if x then begin
    prPrefs.Data[0] := 1;
  end else begin
    prPrefs.Data[0] := 0;
  end;
end;

function TQUB_Results_Var.prGetPoint(i :Integer): TQUB_Results_Point;
begin
  if (i >= 0) and (i < Count) then begin
    Result := TQUB_Results_Point( prPoints[i] );
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results_Var.prGetLowerSDLine(i :Integer): TQUB_Results_Line;
begin
  if (i >= 0) and (i < QUB_SELECT_COLOR_COUNT) then begin
    Result := TQUB_Results_Line( prLowerSDLines[i] );
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results_Var.prGetMeanLine(i :Integer): TQUB_Results_Line;
begin
  if (i >= 0) and (i < QUB_SELECT_COLOR_COUNT) then begin
    Result := TQUB_Results_Line( prMeanLines[i] );
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results_Var.prGetUpperSDLine(i :Integer): TQUB_Results_Line;
begin
  if (i >= 0) and (i < QUB_SELECT_COLOR_COUNT) then begin
    Result := TQUB_Results_Line( prUpperSDLines[i] );
  end else begin
    Result := nil;
  end;
end;

procedure TQUB_Results_Var.prSetNumSDForLines(NumSD :Double);
begin
  prNumSDForLines := NumSD;
  prOnSelectionChanged(Self);
end;

function TQUB_Results_Var.GetAttr(key : PChar) : PPyObject;
var
  oneOrZero             :Integer;
  str                   :String;
begin
  with GetPythonEngine do begin
    if key = 'yIsLog' then begin
      if YLog then oneOrZero := 1 else oneOrZero := 0;
      Result := PyInt_FromLong( oneOrZero );
    end else if key = 'xStyle' then begin
      Result := PyInt_FromLong( Integer( XStyle ) );
    end else if key = 'units' then begin
      str    := '';
      if Count > 0 then str := Points[0].Units;
      Result := PyString_FromString( PChar( str ) );
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

function TQUB_Results_Var.SqLength : Integer;
begin
  Result := prPoints.Count
end;

function TQUB_Results_Var.SqItem( idx : Integer ) : PPyObject;
begin
  with GetPythonEngine do begin
    if (idx < 0) or (idx >= SqLength) then begin
      PyErr_SetString( PyExc_IndexError^, PChar('index out of range') );
      Result := nil;
    end else begin
      Result := PyFloat_FromDouble( YValue[idx] );
    end;
  end;
end;

class procedure TQUB_Results_Var.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do begin
    AddMethod( 'mean',  @TQUB_Results_Var.MeanPy,  'SelectVar.mean( color )' );
    AddMethod( 'sd',    @TQUB_Results_Var.SDPy,    'SelectVar.sd( color )' );
    AddMethod( 'median',@TQUB_Results_Var.MedianPy,'SelectVar.median( color )' );
    AddMethod( 'count', @TQUB_Results_Var.CountPy, 'SelectVar.count( color )' );
  end;
end;

function TQUB_Results_Var.MeanPy( args :PPyObject ): PPyObject;
var
  col    :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:SelectVar.mean( color )', [@col] ) <> 0 then begin
      Result := PyFloat_FromDouble( Mean(col) );
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Results_Var.SDPy( args :PPyObject ): PPyObject; cdecl;
var
  col    :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:SelectVar.sd( color )', [@col] ) <> 0 then begin
      Result := PyFloat_FromDouble( SD(col) );
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Results_Var.MedianPy( args :PPyObject ): PPyObject;
var
  col    :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:SelectVar.median( color )', [@col] ) <> 0 then begin
      Result := PyFloat_FromDouble( Median(col) );
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Results_Var.CountPy( args :PPyObject ): PPyObject; cdecl;
var
  col    :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:SelectVar.count( color )', [@col] ) <> 0 then begin
      Result := PyInt_FromLong( Results.SegsOfColor[col] );
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Results_Var.Mean( color_ix :Integer ): Double;
begin
  Result := prMean[ color_ix ];
end;

function TQUB_Results_Var.SD( color_ix :Integer ): Double;
begin
  Result := prSD[ color_ix ];
end;

function TQUB_Results_Var.Median( color_ix :Integer ): Double;
begin
  Result := prMedian[ color_ix ];
end;

//******************************************************************************

constructor TQUB_SelVarPrefs.Create;
begin
  tree := nil;
end;

procedure TQUB_SelVarPrefs.prInitialize;
begin
  tree := ProfileMgr['Select'].Active;
end;

function TQUB_SelVarPrefs.prGetTypeNode(const ResultType :String): INode;
begin
  if tree = nil then prInitialize;

  Result := tree[ResultType];
end;


//******************************************************************************

constructor TQUB_Results_MdlPoint.Create(Name :String);
begin
  inherited Create;
  prName := Name;
end;

function TQUB_Results_MdlPoint.prGetColor: Integer;
begin
  Result := 0;
end;

procedure TQUB_Results_MdlPoint.prSetColor( x :Integer );
begin
end;

function TQUB_Results_MdlPoint.prGetXValue: Double;
begin
  Result := 0.0;
end;

procedure TQUB_Results_MdlPoint.prSetXValue( x :Double );
begin
end;

function TQUB_Results_MdlPoint.prGetName: String;
begin
  Result := prName;
end;

constructor TQUB_Results_MdlNullPoint.Create( Name :String );
begin
  inherited Create(Name);
  prValue := 0.0;
end;

constructor TQUB_Results_MdlNullPoint.CreateValued( Name :String; Val :Double );
begin
  inherited Create(Name);
  prValue := Val;
end;

function TQUB_Results_MdlNullPoint.prGetYValue: Double;
begin
  Result := prValue;
end;

procedure TQUB_Results_MdlNullPoint.prSetYValue( x :Double );
begin
  prValue := x;
end;

function TQUB_Results_MdlNullPoint.prGetUnits: String;
begin
  if YLog then Result := 'log10';
end;

constructor TQUB_Results_MdlRatePoint.Create( RateNode :INode; RateDir :Integer );
var
  Name      :String;
  States   :INode;
begin
  prRateNode := RateNode;
  prRateDir  := RateDir;

  Name := prRateNode.Name;

  States   := RateNode.Parent['States'];
  if RateDir = 0 then begin
    Name := Name + Format(' %d->%d', [Integer(States.Data[0])+1, Integer(States.Data[1])+1]);
  end else begin
    Name := Name + Format(' %d->%d', [Integer(States.Data[1])+1, Integer(States.Data[0])+1]);
  end;

  inherited Create(Name);
end;

function TQUB_Results_MdlRatePoint.prGetYValue: Double;
begin
  try
    Result        := Double( prRateNode.Data[ prRateDir ] );
    if YLog then Result := Log10( Result );
  except
    Result          := 0.0;
  end;
end;

procedure TQUB_Results_MdlRatePoint.prSetYValue( x :Double );
begin
  try
    prRateNode.Data[ prRateDir ] := x;
  except
  end;
end;

function TQUB_Results_MdlRatePoint.prGetUnits: String;
begin
  if YLog then Result := 'log10';
end;

constructor TQUB_Results_MdlTrialPoint.Create( ValNode :INode; ValIx :Integer );
var
  Name                :String;
begin
  prValNode           := ValNode;
  prValIx             := ValIx;

  if prValNode.DataCount > 1 then begin
    Name := Format('%s %d', [prValNode.Name, prValIx + 1]);
  end else begin
    Name := prValNode.Name;
  end;

  inherited Create(Name);
end;

function TQUB_Results_MdlTrialPoint.prGetYValue: Double;
begin
  try
    Result        := Double( prValNode.Data[ prValIx ] );
    if YLog and (Result > 0.0) then Result := Log10( Result );
  except
    Result          := 0.0;
  end;
end;

procedure TQUB_Results_MdlTrialPoint.prSetYValue( x :Double );
begin
  try
    prValNode.Data[ prValIx ] := x;
  except
  end;
end;

function TQUB_Results_MdlTrialPoint.prGetUnits: String;
begin
  Result := prValNode.LineComment;
  if YLog then Result := 'log10 ' + Result;
end;

constructor TQUB_Results_Model.Create(Results :TQUB_Results; ModelNode :INode);
begin
  prResults                   := Results;
  prNode                      := ModelNode;
  prVars                      := TStringList.Create;
  prVars.Sorted               := True;
end;

destructor TQUB_Results_Model.Destroy;
var
  i                           :Integer;
begin
  try
    for i := 0 to ValueCount - 1 do begin
      TQUB_Results_Point( prVars.Objects[i] ).Free;
    end;
    prVars.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Results_Model.AddPoint(Point :TQUB_Results_Point);
var
  Ix      :Integer;
begin
  Ix := prVars.Add( Point.Name );
  prVars.Objects[Ix] := Point;
end;

function TQUB_Results_Model.prGetValueCount: Integer;
begin
  Result := prVars.Count;
end;

function TQUB_Results_Model.prGetValue(Ix :Integer): Double;
begin
  if (Ix >= 0) and (Ix < ValueCount) then begin
    Result := TQUB_Results_Point( prVars.Objects[Ix] ).YValue;
  end else begin
    raise ERangeError.CreateFMT('Results_Model value index %d is out of bounds [%d - %d]', [Ix, 0, ValueCount-1]);
  end;
end;

procedure TQUB_Results_Model.prSetValue(Ix :Integer; Val :Double);
begin
  if (Ix >= 0) and (Ix < ValueCount) then begin
    TQUB_Results_Point( prVars.Objects[Ix] ).YValue := Val;
    if Assigned(OnValueChanged) then OnValueChanged(Self, Ix);
  end else begin
    raise ERangeError.CreateFMT('Results_Model value index %d is out of bounds [%d - %d]', [Ix, 0, ValueCount-1]);
  end;
end;

function TQUB_Results_Model.prGetValueByName(Ix :String): Double;
begin
  Result := prGetValue( prVars.IndexOf(Ix) );
end;


procedure TQUB_Results_Model.prSetValueByName(Ix :String; Val :Double);
begin
  prSetValue( prVars.IndexOf(Ix), Val );
end;

function TQUB_Results_Model.prHasValue(Ix :String): Boolean;
begin
  Result := prVars.IndexOf(Ix) <> -1;
end;

constructor TQUB_Results_Models.Create(Results :TQUB_Results);
begin
  prResults               := Results;
  prModels                := TList.Create;
  prVarNames              := TStringList.Create;
  prVarNames.Sorted       := True;
  CompareModelsBy         := -1;
  prTrialCount            := 0;
end;

destructor TQUB_Results_Models.Destroy;
var
  i             :Integer;
begin
  for i := 0 to Count - 1 do begin
    Model[i].Free;
  end;
  prModels.Free;
  prVarNames.Free;
end;

procedure TQUB_Results_Models.AddModel(ModelNode :INode);
var
  Model             :TQUB_Results_Model;
begin
  Model             := TQUB_Results_Model.Create(prResults, ModelNode);
  prAddRatePoints(Model);
  prModels.Add(Model);
end;


procedure TQUB_Results_Models.AddModelTrial(ModelTrial :INode);
var
  Model             :TQUB_Results_Model;
  ModelNode         :INode;
  NextModel         :INode;
begin
  ModelNode         := ModelTrial.Find('ModelFile');
  if ModelNode.IsNull then Exit;

  prTrialCount      := prTrialCount + 1;

  NextModel         := ModelNode.SiblingSameName;
  while not NextModel.IsNull do begin
    AddModel( ModelNode );
    prAddPoint( Self.Model[Count-1], TQUB_Results_MdlNullPoint.CreateValued('Trial', prTrialCount) );

    ModelNode       := NextModel;
    NextModel       := NextModel.SiblingSameName;
  end;

  Model             := TQUB_Results_Model.Create(prResults, ModelNode);
  prAddRatePoints(Model);
  prAddTrialPoints(Model, ModelTrial);
  prAddPoint( Model, TQUB_Results_MdlNullPoint.CreateValued('Trial', prTrialCount) );
  prModels.Add(Model);
end;

procedure TQUB_Results_Models.FinishAdding;
var
  i, j             :Integer;
begin
  for i := 0 to Count - 1 do begin
    if Model[i].ValueCount <> prVarNames.Count then begin
      for j := 0 to prVarNames.Count - 1 do begin
        if not Model[i].HasValue[ prVarNames[j] ] then begin
          Model[i].AddPoint( TQUB_Results_MdlNullPoint.Create(prVarNames[j]) );
        end;
      end;
    end;
  end;
end;

procedure TQUB_Results_Models.prAddRatePoints(Model :TQUB_Results_Model);
var
  VarNode          :INode;
  RateNode         :INode;
begin

  VarNode          := Model.Node.Find('Rates').Find('Rate');
  while not VarNode.IsNull do begin
    RateNode       := VarNode['k0'];
    prAddPoint( Model, TQUB_Results_MdlRatePoint.Create( RateNode, 0 ) );
    prAddPoint( Model, TQUB_Results_MdlRatePoint.Create( RateNode, 1 ) );

    RateNode       := VarNode['k1'];
    try
      if RateNode.Data[0] <> 0.0 then
        prAddPoint( Model, TQUB_Results_MdlRatePoint.Create( RateNode, 0 ) );
    except
    end;

    try
      if RateNode.Data[1] <> 0.0 then
        prAddPoint( Model, TQUB_Results_MdlRatePoint.Create( RateNode, 1 ) );
    except
    end;

    VarNode        := VarNode.SiblingSameName;
  end;

end;

procedure TQUB_Results_Models.prAddTrialPoints(Model :TQUB_Results_Model; Trial :INode);
var
  VarNode          :INode;
  i                :Integer;
begin
  VarNode          := Trial.Child;
  while not VarNode.IsNull do begin
    if (VarNode.DataCount > 0) and (VarNode.DataCount < 200) and (VarNode.DataType <> NODE_DATATYPE_STRING) and (VarNode.Child.Sibling.IsNull) and (VarNode.Name <> 'SelectedColor') then begin
      for i := 0 to VarNode.DataCount - 1 do begin
        prAddPoint( Model, TQUB_Results_MdlTrialPoint.Create( VarNode, i ) );
      end;
    end;

    VarNode        := VarNode.Sibling;
  end;
end;

procedure TQUB_Results_Models.prAddPoint(Model :TQUB_Results_Model; Point :TQUB_Results_Point);
var
  Name         :String;
begin
  Model.AddPoint( Point );
  Name         := Point.Name;
  if prVarNames.IndexOf( Name ) = -1 then prVarNames.Add( Name );
end;

var
  gModelsToSort           :TQUB_Results_Models;

function CompareModels(Item1, Item2 :Pointer): Integer;
var
  Diff       :Double;
begin
  with gModelsToSort do begin
    Diff       := TQUB_Results_Model(Item1).Value[ CompareModelsBy ]
                - TQUB_Results_Model(Item2).Value[ CompareModelsBy ];
    if Diff > 0.0 then begin
      Result := 1;
    end else if Diff = 0.0 then begin
      Result := 0;
    end else begin
      Result := -1;
    end;

    if not SortModelsIncreasing then Result := - Result;
  end;
end;

procedure TQUB_Results_Models.SortBy(VarName :String; Increasing :Boolean);
begin
  CompareModelsBy             := prVarNames.IndexOf(VarName);
  SortModelsIncreasing        := not Increasing;
  SortBy(VarName);
end;
// make SortBy(VarName) think the models are already sorted by VarName, in the opposite direction, so it reverses them.

procedure TQUB_Results_Models.SortBy(VarName :String);
var
  SortBy                      :Integer;
begin
  SortBy                      := prVarNames.IndexOf(VarName);
  if SortBy <> -1 then begin
    if SortBy = CompareModelsBy then begin
      SortModelsIncreasing      := not SortModelsIncreasing;
    end else begin
      SortModelsIncreasing      := False;
    end;
    CompareModelsBy             := SortBy;

    gModelsToSort               := Self;
    prModels.Sort(CompareModels);
  end;
  
  if Assigned(OnModelsSorted) then OnModelsSorted(Self);
end;

function TQUB_Results_Models.prGetCount: Integer;
begin
  Result := prModels.Count;
end;

function TQUB_Results_Models.prGetModel(Ix :Integer): TQUB_Results_Model;
begin
  if (Ix >= 0) and (Ix < Count) then begin
    Result := TQUB_Results_Model( prModels[Ix] );
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results_Models.prGetVarCount: Integer;
begin
  Result := prVarNames.Count;
end;

function TQUB_Results_Models.prGetVarName(Ix :Integer): String;
begin
  if (Ix >= 0) and (Ix < VarCount) then begin
    Result := prVarNames[Ix];
  end else begin
    Result := '';
  end;
end;


//******************************************************************************

procedure QUB_Results_SetHistBins(BC :Integer);
begin
  SelHistBinCount  := BC;
  SelHistBinEvents.RaiseNotifyEvent('BinsChanged', nil);
end;

procedure QUB_Results_AddHistVar(HVar :TQUB_Results_Var);
begin
  SelHistBinEvents.AddEvent('BinsChanged', HVar.OnHistBinsChanged);
  Inc(SelHistVarCount);
  if SelHistVarCount = 1 then begin
    SelHistBinEvents.RaiseNotifyEvent('HaveHist', nil);
  end;
end;

procedure QUB_Results_RemHistVar(HVar :TQUB_Results_Var);
begin
  SelHistBinEvents.DelEvent('BinsChanged', HVar.OnHistBinsChanged);
  Dec(SelHistVarCount);
  if SelHistVarCount = 0 then begin
    SelHistBinEvents.RaiseNotifyEvent('HaveNoHist', nil);
  end;
end;

//******************************************************************************

initialization

  SelVarPrefs      := TQUB_SelVarPrefs.Create;
  SelHistBinCount  := SELECTVAR_BINCOUNT;
  SelHistBinEvents := TQUB_Events.Create;
  SelHistVarCount  := 0;


//******************************************************************************
end.

