{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
//September 4th, 1998;
// used by the Spectrum (FFT) histogram

unit QUB_Spectrum;


interface


uses


AllocTypes, Alloc;


procedure FFT(Real, Imag :TPADouble; N, lnN :Integer; sign :Integer;
              var StopFlag :Boolean); forward;

function Power2(p :Integer) :Integer; forward;

// January 31, 2003;

type
  TFFT_WindowFunc = function(i, N :Integer): Double;
  TFFT_DataFunc = function(NextN :Integer; Buf :Pointer): Integer of object;
  // gets next n points as doubles into Buf; returns # points read in case N aren't available
  // data should already have avg value subtracted

function FFT_RectWindow(i, N :Integer): Double;
function FFT_HanningWindow(i, N :Integer): Double;
function FFT_HammingWindow(i, N :Integer): Double;

procedure WindowAvgSpectrum( DataFunc :TFFT_DataFunc;
                             DataCount :Integer;
                             WindowSize :Integer;
                             WindowFunc :TFFT_WindowFunc;
                             Overlap :Integer;
                             SpectrumOut :TPADouble;
                             var FrameCount :Integer;
                             var StopFlag :Boolean );

procedure FFTSpectrum(       DataFunc :TFFT_DataFunc;
                             DataCount :Integer;
                             WindowSize :Integer;              // must be >= DataCount
                             // WindowFunc :TFFT_WindowFunc;    // always rectangular
                             // Overlap :Integer;
                             SpectrumOut :TPADouble;
                             var StopFlag :Boolean );


implementation


uses


SysUtils, Classes, Math;


//*************************** M O D E L ****************************************
procedure FFT(Real, Imag :TPADouble; N, lnN :Integer; sign :Integer; var StopFlag :Boolean);
var
  divN, s, ur, ur1, ui, wr, wi, tr, ti  :Double;
  nd2, i, j, k, l, le, le1, ip          :Integer;
begin

  divN  := 1 / sqrt(n * 1.0);
  nd2   := N div 2;

  j     := 1;
  for i := 1 to N - 1 do begin

    if i < j then begin
      s := Real[i]; Real[i] := Real[j]; Real[j] := s;
      s := Imag[i]; Imag[i] := Imag[j]; Imag[j] := s;
    end;

    k := nd2;

    while k < j do begin
      j := j - k;
      k := k div 2;
    end;

    j := j + k;

  end;

  for l := 1 to lnN do begin

    le  := Power2(l);
    le1 := le div 2;
    ur  := 1;
    ui  := 0;
    wr  := cos(pi / le1);
    wi  := sign * sin(pi / le1);

    for j:= 1 to le1 do begin

      i := j;

      while (i <= N) do begin
        ip := i + le1;
        tr := Real[ip]*ur - Imag[ip]*ui;
        ti := Real[ip]*ui + Imag[ip]*ur;
        Real[ip] := Real[i] - tr;
        Imag[ip] := Imag[i] - ti;
        Real[i]  := Real[i] + tr;
        Imag[i]  := Imag[i] + ti;
        i := i + le;
      end;

      ur1 := ur*wr - ui*wi;
      ui  := ur*wi + ui*wr;
      ur  := ur1;

    end;

  end;

  for i := 1 to N do begin

    Real[i] := Real[i]*divN;
    Imag[i] := Imag[i]*divN;

  end;

end;

function Power2(p :Integer) :Integer;
var
  n, i :Integer;
begin
  n := 1;
  for i := 1 to p do begin
    n := n*2;
  end;
  Result := n;
end;


//******************************************************************************
procedure WindowAvgSpectrum( DataFunc :TFFT_DataFunc;
                             DataCount :Integer;
                             WindowSize :Integer;
                             WindowFunc :TFFT_WindowFunc;
                             Overlap :Integer;
                             SpectrumOut :TPADouble;
                             var FrameCount :Integer;
                             var StopFlag :Boolean );
var
  DataIn             :TPADouble; // circular input buffer, pre-window
  DataFirstOff       :Integer; // start of oldest frame ('Increment' long) in DataIn
  DataLastOff        :Integer; // start of newest frame ('Increment' long) in DataIn
  DataRemain         :Integer;
  Re, Im             :TPADouble;
  FullWin            :TPADouble;
  WinSizePow2        :Integer;
  Increment          :Integer;
  PointsUsed         :Integer;
  Pad                :Integer;
  i, j               :Integer;

  procedure InitData;
  var
    di               :Integer;
  begin
    DataIn           := DAlloc1(WindowSize+1);
    Re               := DAlloc1(WindowSize+1);
    Im               := DAlloc1(WindowSize+1);

    for di := 0 to WindowSize - 1 do begin
      DataIn[di] := 0.0;
    end;
  end;

  procedure FreeData;
  begin
    FreeMem(DataIn);
    FreeMem(Re);
    FreeMem(Im);
  end;

  function NextData: Integer; // returns # points read (max 'Increment')
  begin
    DataFirstOff     := (((DataFirstOff + Increment)-1) mod WindowSize)+1;
    DataLastOff      := (((DataLastOff + Increment)-1) mod WindowSize)+1;
    Result           := DataFunc(Increment, @(DataIn[DataLastOff]));
    DataRemain       := DataRemain - Result;
  end;

  procedure DataToImRe;
  var
    di               :Integer;
    shift            :Integer;
  begin
    shift            := - DataFirstOff;
    for di := DataFirstOff to WindowSize - 1 do begin
      Re[di + shift + 1] := DataIn[di];
      Im[di + 1]         := 0.0;
    end;
    shift            := WindowSize - DataFirstOff;
    for di := 0 to DataFirstOff - 1 do begin
      Re[di + shift + 1] := DataIn[di];
      Im[di + 1]         := 0.0;
    end;
  end;

  procedure AddAvg;
  var
    di               :Integer;
    Weight           :Double;
  begin
    Weight           := PointsUsed / WindowSize;
    for di := 0 to WindowSize - 1 do begin
      SpectrumOut[di] := SpectrumOut[di] + Weight * (Re[di+1] * Re[di+1] + Im[di+1] * Im[di+1]);
    end;
  end;

  procedure FinishAvg;
  var
    di               :Integer;
  begin
    for di := 0 to WindowSize - 1 do begin
      SpectrumOut[di] := SpectrumOut[di] / FrameCount;
    end;
  end;

begin
  DataIn             := nil;
  Re                 := nil;
  Im                 := nil;
  FullWin            := nil;

  try
    Increment        := WindowSize div Overlap;
    DataLastOff      := 1;
    DataFirstOff     := Increment;
    DataRemain       := DataCount;
    FrameCount       := 0; // Ceil(DataCount / Increment) + (Overlap - 1) - 1;

    WinSizePow2      := 0;
    i                := WindowSize;
    while i > 1 do begin
      WinSizePow2    := WinSizePow2 + 1;
      i              := i div 2;
    end;

    InitData;

    FullWin          := DAlloc1(WindowSize);
    for j := 0 to WindowSize - 1 do begin
      SpectrumOut[j] := 0.0;
      FullWin[j]     := WindowFunc(j, WindowSize);
    end;

    // frames padded on the left with zeros
    for i := 1 to Overlap - 1 do begin
      if StopFlag then Break;

      FrameCount     := FrameCount + 1;

      NextData;
      DataToImRe;
      PointsUsed     := i * Increment;
      Pad            := WindowSize - PointsUsed;
      for j := 1 to Pad do begin
        Re[j]        := 0.0;
      end;
      for j := Pad to WindowSize - 1 do begin
        Re[j+1]      := Re[j+1] * WindowFunc(j - Pad, PointsUsed);
      end;

      FFT(Re, Im, WindowSize, WinSizePow2, -1, StopFlag);
      AddAvg;
    end;

    // frames in the middle that are full
    PointsUsed        := WindowSize;
    while not StopFlag do begin // Overlap to FrameCount - Overlap do begin
      FrameCount     := FrameCount + 1;

      Pad               := Increment - NextData;
      DataToImRe;
      if DataRemain = 0 then begin
        // the last frame in the middle, may end a few points shy
        PointsUsed        := WindowSize - Pad;
        for j := 0 to PointsUsed - 1 do begin
          Re[j+1]         := Re[j+1] * WindowFunc(j, PointsUsed);
        end;
        for j := PointsUsed + 1 to WindowSize do begin
          Re[j]           := 0.0;
        end;

        FFT(Re, Im, WindowSize, WinSizePow2, -1, StopFlag);
        AddAvg;

        Break;
      end else begin
        for j := 0 to WindowSize - 1 do begin
          Re[j+1]       := Re[j+1] * FullWin[j];
        end;

        FFT(Re, Im, WindowSize, WinSizePow2, -1, StopFlag);
        AddAvg;
      end;
    end;

    // frames that are padded on the right with zeros
    for i := 1 to Overlap - 1 do begin
      if StopFlag then Break;

      FrameCount     := FrameCount + 1;

      DataFirstOff    := (DataFirstOff + Increment) mod WindowSize;
      DataToImRe;
      PointsUsed      := PointsUsed - Increment;
      for j := 0 to PointsUsed do begin
        Re[j+1]       := Re[j+1] * Windowfunc(j, PointsUsed);
      end;
      for j := PointsUsed + 1 to WindowSize do begin
        Re[j]         := 0.0;
      end;

      FFT(Re, Im, WindowSize, WinSizePow2, -1, StopFlag);
      AddAvg;
    end;

    if not StopFlag then FinishAvg;

  finally
    FreeData;
    FreeMem(FullWin);
  end;

end;

procedure FFTSpectrum(       DataFunc :TFFT_DataFunc;
                             DataCount :Integer;
                             WindowSize :Integer;              // must be >= DataCount
                             // WindowFunc :TFFT_WindowFunc;    // always rectangular
                             // Overlap :Integer;
                             SpectrumOut :TPADouble;
                             var StopFlag :Boolean );
var
  Re, Im             :TPADouble;
  WinSizePow2        :Integer;
  i, j               :Integer;

begin
  Re                 := nil;
  Im                 := nil;

  try
    Re               := DAlloc1(WindowSize+1);
    Im               := DAlloc1(WindowSize+1);

    WinSizePow2      := 0;
    i                := WindowSize;
    while i > 1 do begin
      WinSizePow2    := WinSizePow2 + 1;
      i              := i div 2;
    end;

    for j := 0 to WindowSize - 1 do begin
      SpectrumOut[j] := 0.0;
    end;

    DataFunc(DataCount, @(Re[1]));
    for i := DataCount+1 to WindowSize do begin
      Re[i] := 0;
    end;
    for i := 1 to WindowSize do begin
      Im[i] := 0;
    end;

    FFT(Re, Im, WindowSize, WinSizePow2, -1, StopFlag);

    for i := 0 to WindowSize - 1 do begin
      SpectrumOut[i] := Re[i+1] * Re[i+1] + Im[i+1] * Im[i+1];
    end;

  finally
    FreeMem(Re);
    FreeMem(Im);
  end;

end;


function FFT_RectWindow(i, N :Integer): Double;
begin
  Result := 1.0;
end;

function FFT_HanningWindow(i, N :Integer): Double;
begin
  Result := 0.5 - 0.5 * cos(2 * PI * i / N);
end;

function FFT_HammingWindow(i, N :Integer): Double;
begin
  Result := 0.54 - 0.46 * cos(2 * PI * i / N)
end;

//******************************************************************************
end.
