{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Task_BaselineMT;


{
            Multi-threading Data Baseline correction
see also QUB_Thread_Baseline
}


interface


uses


Windows, SysUtils, Classes, SyncObjs, Graphics,
AllocTypes, Alloc,

QUB_Global,
QUB_Task, QUB_Task_Data, QUB_Task_DataMT,
QUB_Task_ApplyBaseline_Properties,
QUB_Thread, QUB_Thread_Baseline,
QUB_Thread_Socket, QUB_Thread_Socket_Baseline,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_Baseline_Global,
QUB_Stats;


const


DEF_DATA_PROCESSED           = DATA_PROCESSED_ASSUCH;


type


TQUB_Task_BaselineMT  = class(TQUB_Task_DataMT)
protected


  //Baseline Parameters
  prParams                  :TQUB_Baseline_Params;


  function prDoExecuteInitialize :Boolean; override;

  procedure prResultsToData; override;

  procedure prEventLoadData(Event :TQUB_Thread_Event_PRec); override;

  //Thread
  function  prNewThread :TQUB_Thread; override;
  function  prNewSocket(const RemoteServer :String; Port :Integer) :TQUB_Thread_Socket; override;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;


public


  constructor Create(Suspended :Boolean); override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Parameters
  property Params                    :TQUB_Baseline_Params       read prParams                   write prParams;


end;


implementation


//****************************************************************************//
//                                                                            //
//                     T Q U B _ T a s k B a s e l i n e M T                  //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_BaselineMT.Create(Suspended :Boolean);
begin

  inherited;

  prName := 'Correct Baseline';

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Task_BaselineMT.prDoExecuteInitialize :Boolean;
var
  Event :TQUB_Thread_Event_RequestData;
begin
  Result := False;

  prRemote := False;
  
  //This is where to calculate the initial parameters

  Event.DataSel := TQUB_QVO_QVD(Data).DataBaselineSel;
  if (Event.DataSel.Last - Event.DataSel.First + 1) > prChunkSize then begin
    Event.DataSel.Last := Event.DataSel.First + prChunkSize - 1;
  end;
  if Event.DataSel.Last = Event.DataSel.First then begin
    prErrorMsg := 'No initial data selection';
    prWikiTopic := 'Error:No_Initial_Data_Selection';
    prError := True;
    Result := True;
    Exit;
  end;

  Event.Data := IAlloc1(Event.DataSel.Last - Event.DataSel.First + 1);

  try

    prEventLoadData(@Event);

    if Params.Method = METHOD_LINEAR then begin
      CalculateWindow(Event.Data, Event.DataSel, prParams.Init.xAvg, prParams.Init.Avg, prParams.Init.Std, prParams.Init.yStdR, prParams.Init.Slope, prParams.Init.Intercept, prParams.Init.SlopeStd, False, prParams.Init.yVar);
    end else if Params.Method = METHOD_ADAPTIVE then begin
      CalculateWindowAvgStd(Event.Data, Event.DataSel, prParams.Init.Avg, prParams.Init.Std);
    end;

    if not Params.CorrectDCOffset then begin
      prParams.DCOffset := Round(Params.Init.Avg);
    end else begin
      prParams.DCOffset := 0;
    end;

    if Params.Method = METHOD_ADAPTIVE then begin
      prParams.StartPoint := (TQUB_QVO_QVD(Data).DataBaselineSel.First + TQUB_QVO_QVD(Data).DataBaselineSel.Last) div 2;
    end else begin
      prParams.StartPoint := 0;
    end;

    prParams.Init.N   := TQUB_QVO_QVD(Data).DataBaselineSel.Last - TQUB_QVO_QVD(Data).DataBaselineSel.First + 1;
    prParams.Sampling := TQUB_QVO_QVD(Data).Sampling;
    prParams.Scaling  := TQUB_QVO_QVD(Data).Scaling;

  finally

    FreeMem(Event.Data);

  end;

end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_BaselineMT.prChunkData :Boolean;
var
  Chunk    :TQUB_Thread_PChunk;
  BasChunk :TQUB_Thread_Baseline_PChunk;
  s, i     :Integer;
begin

  Result := inherited prChunkData;

  if Result then Exit;

  s := (TQUB_QVO_QVD(Data).DataBaselineSel.First + TQUB_QVO_QVD(Data).DataBaselineSel.Last) div 2;

  for i := 0 to prChunks.Count - 1 do  begin

    Chunk    := TQUB_Thread_PChunk(prChunks.List[i]);
    BasChunk := TQUB_Thread_Baseline_PChunk(Chunk);

    if (s >= Chunk.Sel.First) and (s <= Chunk.Sel.Last) then begin
      BasChunk.Start := s;
    end else begin
      BasChunk.Start := (Chunk.Sel.First + Chunk.Sel.Last) div 2;
    end;

  end;

end;

procedure TQUB_Task_BaselineMT.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  BasChunk :TQUB_Thread_Baseline_PChunk;
begin

  BasChunk := TQUB_Thread_Baseline_PChunk(Chunk);

  FreeMem(BasChunk.Nodes);
  FreeMem(BasChunk.NodesDBL);

  FreeMem(BasChunk);

end;

function TQUB_Task_BaselineMT.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
var
  Chunk    :TQUB_Thread_PChunk;
  BasChunk :TQUB_Thread_Baseline_PChunk;
  First    :Integer;
  s        :Integer;
begin

  Result := nil;

  BasChunk := AllocMem(SizeOf(TQUB_Thread_Baseline_Chunk));

  if BasChunk = nil then Exit;

  Chunk                 := TQUB_Thread_PChunk(BasChunk);

  Chunk.StartTime       := 0;
  for s := 0 to TQUB_QVO_QVD(Data).DataSegmentCount - 1 do begin
    if Sel.First <= (TQUB_QVO_QVD(Data).DataBreakpoints[s] - 1) then begin
      if s = 0 then First := 0 else First := TQUB_QVO_QVD(Data).DataBreakpoints[s - 1];
      Chunk.StartTime := TQUB_QVO_QVD(Data).DataObject.Segments[s].StartTime + TQUB_QVO_QVD(Data).Sampling * (Sel.First - First);
      Break;
    end;
  end;

  BasChunk.NodeCount  := 0;
  BasChunk.Nodes      := nil;
  BasChunk.NodesDBL   := nil;

  Result := Chunk;

end;


//****************************************************************************//
//                                                                            //
//                               T H R E A D S                                //
//                                                                            //
//****************************************************************************//
function TQUB_Task_BaselineMT.prNewThread :TQUB_Thread;
var
  t :TQUB_Thread_Baseline;
begin

  t                 := TQUB_Thread_Baseline.Create(False{Remote}, True{Suspended});

  t.Params          := Params;

  t.OnEvent         := prOnThreadEvent;

  Result            := TQUB_Thread(t);

end;

function TQUB_Task_BaselineMT.prNewSocket(const RemoteServer :String; Port :Integer) :TQUB_Thread_Socket;
var
  t :TQUB_Thread_Socket_Baseline;
begin

  t                 := TQUB_Thread_Socket_Baseline.Create(RemoteServer, Port);

  t.Params          := Params;

  t.OnEvent         := prOnThreadEvent;

  Result            := TQUB_Thread_Socket(t);

end;


//****************************************************************************//
//                                                                            //
//                                 E V E N T S                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_BaselineMT.prEventLoadData(Event :TQUB_Thread_Event_PRec);
begin
  case DataProcessed of
    DATA_PROCESSED_NONE:
      TQUB_QVO_QVD(Data).LoadDataChannel(TQUB_Thread_Event_PRequestData(Event).Data, ChannelIndex,        TQUB_Thread_Event_PRequestData(Event).DataSel, False{Process}, False, False);
    DATA_PROCESSED_ASDISPLAYED:
      TQUB_QVO_QVD(Data).LoadDataChannelProcess(TQUB_Thread_Event_PRequestData(Event).Data, ChannelIndex, TQUB_Thread_Event_PRequestData(Event).DataSel, TQUB_QVO_QVD(Data).FilterOn[ChannelIndex], TQUB_QVO_QVD(Data).FilterFreq[ChannelIndex], TQUB_QVO_QVD(Data).BaselineOn[ChannelIndex]);
    DATA_PROCESSED_ASSUCH:
      TQUB_QVO_QVD(Data).LoadDataChannelProcess(TQUB_Thread_Event_PRequestData(Event).Data, ChannelIndex, TQUB_Thread_Event_PRequestData(Event).DataSel, ProcessFilter, ProcessFilterFreq, ProcessBaseline);
  end;
end;



//****************************************************************************//
//                                                                            //
//                               R E S U L T S                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_BaselineMT.prResultsToData;
var
  NodeCount  :Integer;
  Nodes      :TPAInteger;
  NodesDBL   :TPAInteger;
  BasChunk   :TQUB_Thread_Baseline_PChunk;
  Chunk      :TQUB_Thread_PChunk;
  i, f, l    :Integer;
begin

  NodeCount  := 0;
  Nodes      := nil;
  NodesDBL   := nil;

  for i := 0 to prChunks.Count - 1 do begin

    Chunk    := prChunks.List[i];
    BasChunk := TQUB_Thread_Baseline_PChunk(Chunk);

    if Chunk.Status = CHUNK_STATUS_DONE then begin

      if BasChunk.NodeCount > 0 then begin

        ReallocMem(Nodes,    SizeOf(Integer) * (NodeCount + BasChunk.NodeCount));
        ReallocMem(NodesDBL, SizeOf(Integer) * (NodeCount + BasChunk.NodeCount));

        Move(BasChunk.Nodes   [0], Nodes   [NodeCount], SizeOf(BasChunk.Nodes   [0]) * BasChunk.NodeCount);
        Move(BasChunk.NodesDBL[0], NodesDBL[NodeCount], SizeOf(BasChunk.NodesDBL[0]) * BasChunk.NodeCount);

        Inc(NodeCount, BasChunk.NodeCount);

      end;

    end;

    //Clear Chunk - for memory efficiency
    FreeMem(BasChunk.Nodes);
    FreeMem(BasChunk.NodesDBL);
    BasChunk.NodeCount := 0;
    BasChunk.Nodes     := nil;
    BasChunk.NodesDBL  := nil;

  end;

  if NodeCount > 0 then begin

    f := 0;
    l := 0;
    //Add the two marginal nodes ?!
    if Nodes[0] > TQUB_QVO_QVD(Data).DataPoint.First then begin
      f := 1;
    end;
    if Nodes[NodeCount - 1] < TQUB_QVO_QVD(Data).DataPoint.Last then begin
      l := 1;
    end;
    if (f + l) > 0 then begin
      ReAllocMem(Nodes,    SizeOf(Integer) * (NodeCount + f + l));
      ReAllocMem(NodesDBL, SizeOf(Integer) * (NodeCount + f + l));
      if f > 0 then begin
        for i := NodeCount downto 1 do begin
          Nodes[i]    := Nodes[i - 1];
          NodesDBL[i] := NodesDBL[i - 1];
        end;
        Nodes[0]      := TQUB_QVO_QVD(Data).DataPoint.First;
        Inc(NodeCount);
      end;
      if l > 0 then begin
        Nodes[NodeCount]    := TQUB_QVO_QVD(Data).DataPoint.Last;
        NodesDBL[NodeCount] := NodesDBL[NodeCount - 1];
        Inc(NodeCount);
      end;
    end;

    TQUB_QVO_QVD(Data).ApplyBaselineNodes(ChannelIndex, NodeCount, Nodes, NodesDBL, ProcessBaseline, Params.DCOffset);
          
  end;

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_BaselineMT.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_ApplyBaseline_PropertiesForm do begin

    Self.ChannelIndex                  := Properties['ChannelIndex'].Data[0];

    //Data Source
    Self.DataSource                    := DataSource;

    //Data Processed
    Self.DataProcessed                 := StringToDataProcessed(Properties['ProcessData'].DataAsString);
    Self.ProcessFilter                 := 0<>Properties['ProcessData']['Filter'].Data[0];
    Self.ProcessFilterFreq             := Properties['ProcessData']['FilterFreq'].Data[0];
    Self.ProcessBaseline               := 0<>Properties['ProcessData']['Baseline'].Data[0];

    //Client/Server
    Self.ThreadCount                   := Properties['ThreadCount'].Data[0];
    Self.Remote                        := 0<>Properties['Remote'].Data[0];

    //Method
    Self.Params                        := Params;

  end;//with
end;

procedure TQUB_Task_BaselineMT.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_ApplyBaseline_PropertiesForm do begin
  //  ChannelIndex                       := Self.ChannelIndex;
  //end;//with
end;


//******************************************************************************
end.

