{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Task_Data;


{
              The base class for tasks operating on data
                  running in single-threading mode
}


interface


uses


SysUtils, Classes, Graphics, Forms,

QUB_Global,

QUB_Task,
QUB_Thread,
QUB_QVO;


const


DEF_CHUNK_SIZE                   =      ONE_MEGABYTE;
MAX_CHUNK_SIZE                   = 16 * ONE_MEGABYTE;
MIN_CHUNK_SIZE                   =      ONE_MEGABYTE div 16;


type


TQUB_Task_Data  = class(TQUB_Task)
protected


  //Data
  prData                   :TQUB_QVO;
  prDataSource             :TQUB_Task_Data_Source;
  prDataProcessed          :TQUB_Task_Data_Processed;
  prProcessFilter          :Boolean;
  prProcessFilterFreq      :Double;
  prProcessBaseline        :Boolean;
  prProcessResample        :Boolean;
  prProcessResampleFactor  :Double;
  prChannelIndex           :Integer;

  //Chunks
  prChunks                 :TList;
  prChunkSize              :Integer;
  prChunkSegmentation      :Boolean;
  prChunkCount             :Integer;

  //Remote and MultiThread stuff
  prThreadCount            :Integer;
  prRemote                 :Boolean;
  //prServers                :TQUB_ServerViewForm;

  prMovieMake              :Boolean;
  

  procedure prDoExecute; virtual;
  procedure prDoExecuteStart; virtual;
  procedure prDoExecuteEnd; virtual;
  function prDoExecuteInitialize :Boolean; virtual;
  procedure prDoExecuteFinalize; virtual;

  procedure prDoTask; virtual; //The specifics

  procedure prResultsToData; virtual;

  //Chunks
  procedure prSetChunkParameters; virtual;
  function prChunkData :Boolean; virtual;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; virtual; abstract;
  procedure prFreeChunks;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); virtual; abstract;

  procedure prSetData(Data :TQUB_QVO);

  procedure prCheckStatus;


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Execute; override;

  procedure UpdateProps; virtual;


  property Data                      :TQUB_QVO                   read prData                    write prSetData;
  property DataSource                :TQUB_Task_Data_Source      read prDataSource              write prDataSource;
  property DataProcessed             :TQUB_Task_Data_Processed   read prDataProcessed           write prDataProcessed;
  property ProcessFilter             :Boolean                    read prProcessFilter           write prProcessFilter;
  property ProcessFilterFreq         :Double                     read prProcessFilterFreq       write prProcessFilterFreq;
  property ProcessBaseline           :Boolean                    read prProcessBaseline         write prProcessBaseline;
  property ProcessResample           :Boolean                    read prProcessResample         write prProcessResample;
  property ProcessResampleFactor     :Double                     read prProcessResampleFactor   write prProcessResampleFactor;
  property ChannelIndex              :Integer                    read prChannelIndex            write prChannelIndex;


  //property Servers                   :TQUB_ServerViewForm        read prServers                 write prServers;

  property Remote                    :Boolean                    read prRemote                  write prRemote;
  property ThreadCount               :Integer                    read prThreadCount             write prThreadCount;

  property MovieMake                 :Boolean                    read prMovieMake               write prMovieMake;


end;


implementation

uses

QUB_QVO_QVD, QUB_QVO_QVD_Global;


//******************************************************************************
constructor TQUB_Task_Data.Create(Suspended :Boolean);
begin
  inherited;

  Priority              := tpNormal;

  prChunkCount          := 0;
  prChunkSize           := DEF_CHUNK_SIZE;
  prChunkSegmentation   := True;
  prChunks              := TList.Create;

end;

destructor TQUB_Task_Data.Destroy;
begin

  try

    prFreeChunks;
    prChunks.Free;

  finally

    inherited;

  end;

end;

procedure TQUB_Task_Data.UpdateProps;
begin
end;

//Execute
procedure TQUB_Task_Data.Execute;
begin

  try

    if Terminated then Exit;

    prStatus := 'Initializing';
    prPercent := 0;
    prETA := '0:00:00';

    Synchronize(prDoOnChange);

    prDoExecuteStart;

    try
      prDoExecute;
    except
     on e :Exception do begin
       prErrorMsg := e.Message;
       prError    := True;
     end;
    end;

    if prError then begin
      prStatus := 'Error';
      Synchronize(prDoOnChange);
      Synchronize(ShowMessage);
    end;

    if not SilentOutput then Beep;

    //Terminate;

    prStatus := 'Terminated';
    prPercent := 0;
    prETA := '0:00:00';

    Synchronize(prDoOnChange);

  finally

    prDoExecuteEnd;

  end;

end;

procedure TQUB_Task_Data.prDoExecute;
begin

  //Reset error
  prErrorMsg     := '';
  prError        := False;

  prChunkCount   := 0;
  prStopped      := False;

  prStTime       := 0;

  try

    //Initialize section; to be overriden
    if prDoExecuteInitialize then Exit;

    //Chunk data to be used by threads
    if prChunkData then Exit;

    //Start here
    prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
    prStatus := 'Running';
    Synchronize(prDoOnChange);

    prDoTask;//The specifics


  finally


    prUpdateStatus(100, prStTime);

    if not prError then begin

      if prStopped then begin
        prStatus := 'Stopping';
      end else begin
        prStatus := 'Terminating';
      end;
      Synchronize(prDoOnChange);
      
      Synchronize(prResultsToData);

    end else if prError then begin

      prStatus := 'Error';
      Synchronize(prDoOnChange);

      Synchronize(ShowMessage);

    end;

    prDoExecuteFinalize;

    prErrorMsg := '';
    prError    := False;

  end;

end;

function TQUB_Task_Data.prDoExecuteInitialize;
begin
  Result := False;
end;

procedure TQUB_Task_Data.prDoExecuteFinalize;
begin
end;

procedure TQUB_Task_Data.prDoExecuteStart;
begin
end;

procedure TQUB_Task_Data.prDoExecuteEnd;
begin
  if Data <> nil then Data.UnlockWrite;
end;

procedure TQUB_Task_Data.prDoTask;
begin
end;

//Results;
procedure TQUB_Task_Data.prResultsToData;
begin
end;

//******************************************************************************
procedure TQUB_Task_Data.prCheckStatus;
var
  p  :Single;
  c :Integer;
begin

  if prChunks.Count = 0 then Exit;

  p := 0;
  for c := 0 to prChunks.Count - 1 do begin
    p := p + TQUB_Thread_PChunk(prChunks[c]).Completion;
  end;

  c := Round(p / prChunks.Count);
  if c <> prPercent then begin
    prUpdateStatus(c, prStTime);
  end;

end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Data.prFreeChunks;
var
  i :Integer;
begin
  for i := 0 to prChunks.Count - 1 do begin
    prFreeChunk(prChunks.List[i]);
  end;
end;

procedure TQUB_Task_Data.prSetChunkParameters;
begin
  prChunkSize         := MAX_CHUNK_SIZE;
  prChunkSegmentation := True;
end;

function TQUB_Task_Data.prChunkData :Boolean;
var
  ISels            :TList;
  OSels            :TList;
  Sel              :TQUB_PFirstLast;
  CurrList         :TQUB_QVO_QVD_SelectionList;
  ListIter         :TQUB_QVO_QVD_SelectionListIter;
  Chunk            :TQUB_Thread_PChunk;
  l, s, p          :Integer;
begin

  Result           := False;

  prSetChunkParameters;

  ISels            := nil;
  OSels            := nil;

  try

    ISels          := TList.Create;
    OSels          := TList.Create;

    case DataSource of

      DATA_SOURCE_WHOLEFILE:
        begin
          New(Sel);
          ISels.Add(Sel);
          Sel^ := TQUB_QVO_QVD(Data).DataPoint;
        end;

      DATA_SOURCE_SELECTIONLIST:
        begin
          if (TQUB_QVO_QVD(Data).SelectionLists.ItemIndex >= 0) and (TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0) then begin
            CurrList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
            ListIter := CurrList.Find(0, TQUB_QVO_QVD(Data).DataPoint.Last, True{ActiveOnly});
            if not ListIter.Done then repeat
              New(Sel);
              ISels.Add(Sel);
              Sel^ := ListIter.Sel;
            until ListIter.Next;
            ListIter.Free;
          end;
        end;

      DATA_SOURCE_SELECTION:
        begin
          if TQUB_QVO_QVD(Data).DataSel.First < TQUB_QVO_QVD(Data).DataSel.Last then begin
            New(Sel);
            ISels.Add(Sel);
            Sel^ := TQUB_QVO_QVD(Data).DataSel;
          end;
        end;
    end;

    if ISels.Count < 1 then begin
      Result     := True;
      prError    := True;
      prErrorMsg := 'No selection active or empty selection list or empty selection';
      prWikiTopic := 'Error:No_Selection';
      Exit;
    end;

    //Careful ! segments artificially chunked according to MaxSize may have to be joined later !

    if (TQUB_QVO_QVD(Data).SegmentData(ISels, OSels, prChunkSegmentation, prChunkSize)) or (OSels.Count < 1) then begin
      Result     := True;
      prError    := True;
      prErrorMsg := 'Data could not be segmented';
      prWikiTopic := 'Error:Data_could_not_be_segmented';
      Exit;
    end;

    for l := 0 to OSels.Count - 1 do begin

      Sel                   := TQUB_PFirstLast(OSels.List[l]);
      Chunk                 := prNewChunk(Sel^);
      if Chunk <> nil then begin
        prChunks.Add(Chunk);

        Chunk.Sel           := Sel^;
        Chunk.Status        := CHUNK_STATUS_EMPTY;
        Chunk.Trials        := 0;
        Chunk.Index         := l;

        Chunk.StartTime     := TQUB_QVO_QVD(Data).DataObject.Segments[0].StartTime + TQUB_QVO_QVD(Data).DataObject.Sampling * Sel.First;
        p := 0;
        for s := 0 to TQUB_QVO_QVD(Data).DataObject.Segments.SegmentCount - 1 do begin
          if Sel.First <= p then begin
            if s > 0 then begin
              Chunk.StartTime := TQUB_QVO_QVD(Data).DataObject.Segments[s].StartTime + TQUB_QVO_QVD(Data).DataObject.Sampling * (Sel.First - p);
            end;
            Break;
          end;
          Inc(p, TQUB_QVO_QVD(Data).DataObject.Segments[s].Channels.DataCount);
        end;

      end;

    end;

    if prChunks.Count = 0 then begin
      Result     := True;
      prError    := True;
      prErrorMsg := 'Data could not be segmented';
      prWikiTopic := 'Error:Data_could_not_be_segmented';
      Exit;
    end;

  finally

    if ISels <> nil then begin
      for s := 0 to ISels.Count - 1 do begin
        FreeMem(TQUB_PFirstLast(ISels.List[s]));
      end;
      ISels.Free;
    end;
    if OSels <> nil then begin
      for s := 0 to OSels.Count - 1 do begin
        FreeMem(TQUB_PFirstLast(OSels.List[s]));
      end;
      OSels.Free;
    end;

  end;

end;


//Data
procedure TQUB_Task_Data.prSetData(Data :TQUB_QVO);
begin
  prData := Data;
  if Data <> nil then TQUB_QVO_QVD(Data).LockWrite;
end;


//******************************************************************************
end.
