{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Task_DataMT;


{
              The base class for tasks operating on data
                  running in multi-threading mode
}


interface


uses


SysUtils, Classes, SyncObjs, Graphics, Forms,

Alloc, AllocTypes,

QUB_Global,

QUB_Task_Data,
QUB_Socket,
QUB_Thread,
QUB_Thread_Socket,
QUB_QVO_QVD, QUB_QVO_QVD_Global;


const


TASK_DATAMT_WAITFOR              = 10000;


type


TQUB_Task_DataMT  = class(TQUB_Task_Data)
protected


  //Multi-threading Signaling
  prSignal                 :TSimpleEvent;

  //Lists
  prThreads                :TList;
  prEvents                 :TThreadList;


  procedure prDoExecute; override;

  //Events
  procedure prEventLoadData(Event :TQUB_Thread_Event_PRec); virtual;
  procedure prEventMessage(Event :TQUB_Thread_Event_PRec); virtual;
  procedure prCheckEvents; virtual;

  //Threads
  procedure prSpawnThreads;
  procedure prSpawnNewThreads;
  procedure prReSpawnThread(t :Integer);
  procedure prSpawnAnotherThread(t :Integer);
  procedure prFreeThreads;
  procedure prFreeThread(t :Integer);

  function  prNewThread :TQUB_Thread; virtual; abstract;
  function  prNewSocket(const RemoteServer :String; Port :Integer) :TQUB_Thread_Socket; virtual; abstract;

  //I/O
  procedure prReadThreadOutput(Thread :TQUB_Thread); virtual;
  procedure prWriteThreadInput(Thread :TQUB_Thread; Chunk :TQUB_Thread_PChunk); virtual;
  procedure prWriteThreadInputs(Thread :TQUB_Thread; c1, c2 :Integer); virtual;

  procedure prReadSocketOutput(Socket :TQUB_Thread_Socket); virtual;
  procedure prWriteSocketInput(Socket :TQUB_Thread_Socket; Chunk :TQUB_Thread_PChunk); virtual;

  procedure prOnThreadEvent(Sender :TObject; Event :TQUB_Thread_Event_PRec);


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Stop; override;


  property Signal                    :TSimpleEvent               read prSignal;


end;


implementation


//******************************************************************************
constructor TQUB_Task_DataMT.Create(Suspended :Boolean);
begin
  inherited;

  prThreadCount         := 1;
  prRemote              := False;

  prSignal              := TSimpleEvent.Create;

  prEvents              := TThreadList.Create;
  prEvents.Duplicates   := dupAccept;

  prThreads             := TList.Create;

end;

destructor TQUB_Task_DataMT.Destroy;
begin

  try

    prFreeThreads;

    prSignal.Free;

    prEvents.Free;

    prThreads.Free;

  finally

    inherited;

  end;

end;


//Execute
procedure TQUB_Task_DataMT.prDoExecute;
var
  SignalType     :TWaitResult;
begin

  //Reset error
  prErrorMsg     := '';
  prError        := False;

  prChunkCount   := 0;
  prStopped      := False;

  prStTime       := 0;

  try

    //Initialize section; to be overriden
    if prDoExecuteInitialize then Exit;

    //Chunk data to be used by threads
    if prChunkData then Exit;

    //Allocate and start threads
    prSpawnThreads;

    //Start here
    prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
    prStatus := 'Running';
    Synchronize(prDoOnChange);


    //**********************************************************************//
    //                                                                      //
    //                              Main loop                               //
    //                                                                      //
    //**********************************************************************//
    repeat
    //Main loop

      //Stop thread if user-stopped
      if prStopped then begin
        prStatus := 'Stopping';
        Synchronize(prDoOnChange);
        Exit;
      end;

      //The JOB thread now goes to sleep, since it has nothing to do, unless the Signal is set
      //or times out.
      //the running threads will have to SetEvent when they want something from JOB thread.
      SignalType := prSignal.WaitFor(TASK_DATAMT_WAITFOR);


      //Clear the Signal
      prSignal.Acquire;
      try
        prSignal.ResetEvent;
      finally
        prSignal.Release;
      end;


      //Stop thread if user-stopped
      if prStopped then begin
        prStatus := 'Stopping';
        Synchronize(prDoOnChange);
        Exit;
      end;


      if SignalType = wrSignaled then begin
      //was Signaled, analyze the request list
        prCheckEvents;
      end;

      //Spawn new threads
      prSpawnNewThreads;

    until ((prChunkCount = prChunks.Count) or (prStopped) or (prError));
    //Main loop


  finally


    prUpdateStatus(100, prStTime);

    if (not prStopped) and (not prError) then begin

      prStatus := 'Terminating';
      Synchronize(prDoOnChange);

      Synchronize(prResultsToData);

    end else if prError then begin

      prStatus := 'Error';
      Synchronize(prDoOnChange);

      Synchronize(ShowMessage);

    end else if prStopped then begin

      prStatus := 'Stopping';
      Synchronize(prDoOnChange);

    end;

    prDoExecuteFinalize;

    prErrorMsg := '';
    prError    := False;

  end;

end;


//Stop
procedure TQUB_Task_DataMT.Stop;
begin
  inherited;

  prSignal.Acquire;
  try
    prSignal.SetEvent;
  finally
    prSignal.Release;
  end;

end;


//****************************************************************************//
//                                                                            //
//                         T H R E A D    E V E N T S                         //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_DataMT.prCheckEvents;
var
  Event :TQUB_Thread_Event_PRec;
  Chunk :TQUB_Thread_PChunk;
  List  :TList;
begin

  repeat

    //Application.ProcessMessages;  ???

    if prStopped then begin

      Break;

    end;

    List := prEvents.LockList;

    if List.Count = 0 then begin

      prEvents.UnlockList;
      Exit;

    end;

    Event := List.List[0];
    List.Delete(0);
    prEvents.UnlockList;

    if TObject(Event.Thread) is TQUB_Thread then begin
      Chunk := Event.Thread.Chunk[0];
    end else begin
      Chunk := TQUB_Thread_Socket(Event.Thread).Chunk;
    end;

    try

      if Event.Event = THREAD_EVENT_REQUEST_DATA then begin
      //The Thread needs Data

        prEventLoadData(Event);

        if TObject(Event.Thread) is TQUB_Thread then begin
          Event.Thread.SetEvent;
        end else if TObject(Event.Thread) is TQUB_Thread_Socket then begin
          TQUB_Thread_Socket(Event.Thread).SetEvent;
        end;

      end else if Event.Event = THREAD_EVENT_REPORT_STATUS then begin
      //the Thread reports Status (Job Completion Percent)

        prCheckStatus;

      end else if Event.Event = THREAD_EVENT_REPORT_MESSAGE then begin
      //the Thread reports some Message

        prEventMessage(Event);

      end else if Event.Event = THREAD_EVENT_REPORT_TERMINATION then begin
      //the Thread has Terminated

        if Chunk.Status = CHUNK_STATUS_DONE then begin
        //the Chunk was Done OK

          if TObject(Event.Thread) is TQUB_Thread then begin
            prReadThreadOutput(Event.Thread);
          end else begin
            prReadSocketOutput(TQUB_Thread_Socket(Event.Thread));
          end;

        end else if (Chunk.Status = CHUNK_STATUS_ERROR) or (Chunk.Status = CHUNK_STATUS_STOPPED) then begin
        //the Chunk was Stopped or Error

          prFreeThread(Chunk.Index);

          if (Chunk.Trials <= CHUNK_MAX_TRIALS) and (not Event.Thread.Remote) then begin
            prReSpawnThread(Chunk.Index);
          end else begin
            prError := True;
            prErrorMsg := 'Failed to execute one or more threads';
            prWikiTopic := 'Error:Failed_to_execute_threads';
          end;

        end;

      end else if Event.Event = THREAD_EVENT_REPORT_ERROR then begin
      //Thread/Socket Error

        if TObject(Event.Thread) is TQUB_Thread then begin
          prErrorMsg := 'Thread error';
          prWikiTopic := 'Error:Thread_Error';
        end else if TObject(Event.Thread) is TQUB_Thread_Socket then begin
          prErrorMsg := 'Socket error';
          prWikiTopic := 'Error:Thread_Error';
        end;

        if (Chunk.Trials <= CHUNK_MAX_TRIALS) and (not Event.Thread.Remote) then begin
          prReSpawnThread(Chunk.Index);
          prErrorMsg := '';
        end else begin
          prError := True;
        end;

      end;

    finally

      FreeMem(Event);

    end;

  until (False);

end;

procedure TQUB_Task_DataMT.prEventLoadData(Event :TQUB_Thread_Event_PRec);
begin
end;

procedure TQUB_Task_DataMT.prEventMessage(Event :TQUB_Thread_Event_PRec);
begin
end;

//******************************************************************************
procedure TQUB_Task_DataMT.prOnThreadEvent(Sender :TObject; Event :TQUB_Thread_Event_PRec);
begin

  prSignal.Acquire;

  try

    prEvents.Add(Event);
    prSignal.SetEvent;

  finally

    prSignal.Release;

  end;

end;


//****************************************************************************//
//                                                                            //
//                              T H R E A D S                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_DataMT.prSpawnThreads;
var
  Thread     :TQUB_Thread;
  ServList   :TPAPointer;
  ServCount  :Integer;
  Count      :Integer;
  ChPerTh    :Integer;
  TotCap     :Integer;
  SvrCap     :Integer;
  t, s       :Integer;
  p          :Pointer;
begin

  (*//Remote
  if prRemote then begin

    ServList := nil;
    TotCap   := 0;

    try

      Servers.ServersCritSection.Acquire;

      //Find the servers
      try

        ServCount := Servers.ServerCount;

        if ServCount > 0 then begin

          ServList := AllocMem(ServCount * SizeOf(Pointer));
          s := 0;
          for t := 0 to ServCount - 1 do begin
            if Servers.Server(t + 1).Used then begin
              ServList[s] := Servers.Server(t + 1);
              Inc(s);
            end;
          end;
          ServCount := s;

          if ServCount = 0 then begin
          //No server is checked
            ServCount := Servers.ServerCount;
            for s := 0 to ServCount - 1 do begin
              ServList[s] := Servers.Server(s + 1);
            end;
          end;

          if ServCount > 0 then begin

            if ServCount > ThreadCount then begin
            //Sort according to capacity - running - pending
              for t := 0 to ServCount - 1 do begin
                for s := 0 to ServCount - 2 do begin
                  if (TQUB_PServer(ServList[s]).Capacity - TQUB_PServer(ServList[s]).Running - TQUB_PServer(ServList[s]).Pending) < (TQUB_PServer(ServList[s + 1]).Capacity - TQUB_PServer(ServList[s + 1]).Running - TQUB_PServer(ServList[s + 1]).Pending) then begin
                    p := ServList[s]; ServList[s] := ServList[s + 1]; ServList[s + 1] := p;
                  end;
                end;
              end;
              ServCount := prThreadCount;
            end else begin
              prThreadCount := ServCount;
            end;

            TotCap := 0;
            for s := 0 to ServCount - 1 do begin
              SvrCap := TQUB_PServer(ServList[s]).Capacity - TQUB_PServer(ServList[s]).Running - TQUB_PServer(ServList[s]).Pending;
              Inc(TotCap, SvrCap);
            end;

          end else begin
            prRemote := False;
          end;

        end else begin
          prRemote := False;
        end;

        //Still Remote ?
        if prRemote then begin

          if prChunks.Count < prThreadCount then prThreadCount := prChunks.Count;

          s := 0;
          for t := 0 to prThreadCount - 1 do begin

            SvrCap        := TQUB_PServer(ServList[t]).Capacity - TQUB_PServer(ServList[t]).Running - TQUB_PServer(ServList[t]).Pending;
            ChPerTh       := Round(SvrCap * prChunks.Count / TotCap);

            if ChPerTh < 1 then ChPerTh := 1;
            if (s + ChPerTh) > prChunks.Count then ChPerTh := prChunks.Count - s;
            if t = prThreadCount - 1 then ChPerTh := prChunks.Count - s;

            if ChPerTh = 0 then Continue;

            Thread        := prNewThread;
            Thread.Remote := True;
            Thread.Host   := TQUB_PServer(ServList[t]).Host;
            Thread.Port   := TQUB_PServer(ServList[t]).Port;

            prWriteThreadInputs(Thread, s, s + ChPerTh - 1);//The range of Chunks

            Inc(s, ChPerTh);

            prThreads.Add(Thread);
            Thread.Resume;

          end;

          prThreadCount := prThreads.Count;

        end;

      except

        prRemote := False;

      end;


    finally

      FreeMem(ServList);
      Servers.ServersCritSection.Release;

    end;

  end;

  //Local
  if not prRemote then begin
    *)
    Count := prThreadCount;
    if Count > prChunks.Count then Count := prChunks.Count;
    if Count < 1 then Count := 1;

    for t := 0 to Count - 1 do begin

      Thread := prNewThread;
      prWriteThreadInput(Thread, prChunks.List[t]);
      prThreads.Add(Thread);
      Thread.Resume;

    end;

 // end;

end;

procedure TQUB_Task_DataMT.prSpawnNewThreads;
var
  Chunk             :TQUB_Thread_PChunk;
  c, cd, ce, cr, cs :Integer;
begin
  cd := 0;
  ce := 0;
  //Count how many Chunks are done and how many are empty
  for c := 0 to prChunks.Count - 1 do begin
    Chunk := TQUB_Thread_PChunk(prChunks.List[c]);
    if Chunk.Status = CHUNK_STATUS_EMPTY then Inc(ce);
    if Chunk.Status = CHUNK_STATUS_DONE  then Inc(cd);
  end;
  //Obviously, all - (done + empty) = running
  cr := prChunks.Count - (cd + ce);

  if cr < prThreadCount then begin
  //Fewer than necessary are still running
    //how many to start
    cs := prThreadCount - cr;
    for c := 0 to prChunks.Count - 1 do begin
      Chunk := TQUB_Thread_PChunk(prChunks.List[c]);
      if Chunk.Status = CHUNK_STATUS_EMPTY then begin
        try
          prSpawnAnotherThread(Chunk.Index);
        finally
          Dec(cs);
        end;
        if cs = 0 then Break;
      end;
    end;
  end;

end;

procedure TQUB_Task_DataMT.prReSpawnThread(t :Integer);
var
  Thread :TQUB_Thread;
  Socket :TQUB_Thread_Socket;
  SName  :String;
begin

  //if not prRemote then begin

    Thread := prNewThread;
    prWriteThreadInput(Thread, prChunks.List[t]);
    prThreads.Insert(t, Thread);
    Inc(Thread.Chunk[0].Trials);
    Thread.Resume;
   (*
  end else begin

    //It should never get here, for now

    SName := prServers.Server(t).Host;

    if SName <> '' then begin

      Socket := prNewSocket(SName, QUB_SOCKET_PORT);
      prWriteSocketInput(Socket, prChunks.List[t]);
      prThreads.Insert(t, Socket);
      Inc(Socket.Chunk.Trials);
      Synchronize(Socket.Open);

    end else begin

      Thread := prNewThread;
      prWriteThreadInput(Thread, prChunks.List[t]);
      prThreads.Insert(t, Thread);
      Inc(Thread.Chunk[0].Trials);
      Thread.Resume;

    end;

  end;
        *)
  Dec(prChunkCount);

end;

procedure TQUB_Task_DataMT.prSpawnAnotherThread(t :Integer);
var
  Thread :TQUB_Thread;
  Socket :TQUB_Thread_Socket;
  SName  :String;
begin

  //if not prRemote then begin

    Thread := prNewThread;
    prWriteThreadInput(Thread, prChunks.List[t]);
    prThreads.Insert(t, Thread);
    Thread.Resume;
      (*
  end else begin

    SName := prServers.Server(t).Host;

    if SName <> '' then begin

      Socket := prNewSocket(SName, QUB_SOCKET_PORT);
      prWriteSocketInput(Socket, prChunks.List[t]);
      prThreads.Insert(t, Socket);
      Synchronize(Socket.Open);

    end else begin

      Thread := prNewThread;
      prWriteThreadInput(Thread, prChunks.List[t]);
      prThreads.Insert(t, Thread);
      Thread.Resume;

    end;

  end;
       *)
end;

procedure TQUB_Task_DataMT.prFreeThreads;
var
  Thread :TQUB_Thread;
  Socket :TQUB_Thread_Socket;
  i      :Integer;
begin

  for i := 0 to prThreads.Count - 1 do begin

    if TObject(prThreads.List[i]) is TQUB_Thread then begin

      Thread := TQUB_Thread(prThreads.List[i]);

      if (not Thread.Terminated) and (not Thread.Suspended) then begin

        if Thread.Suspended then Thread.Resume;// ???
        Thread.Abort;
        Thread.WaitFor;

      end;

      Thread.Free;

    end else if TObject(prThreads.List[i]) is TQUB_Thread_Socket then begin

      Socket := TQUB_Thread_Socket(prThreads.List[i]);

      if Socket.Connected then begin

        Synchronize(Socket.Close);

      end;

      Synchronize(Socket.Free);

    end;

  end;

end;

procedure TQUB_Task_DataMT.prFreeThread(t :Integer);
var
  Thread :TQUB_Thread;
  Socket :TQUB_Thread_Socket;
begin

  if TObject(prThreads.List[t]) is TQUB_Thread then begin

    Thread := TQUB_Thread(prThreads.List[t]);

    if (not Thread.Terminated) and (not Thread.Suspended) then begin

      if Thread.Suspended then Thread.Resume;// ???
      Thread.Abort;
      Thread.WaitFor;

    end;

    Thread.Free;

  end else if TObject(prThreads.List[t]) is TQUB_Thread_Socket then begin

    Socket := TQUB_Thread_Socket(prThreads.List[t]);

    if Socket.Connected then begin

      Synchronize(Socket.Close);

    end;

    Synchronize(Socket.Free);

  end;

  prThreads.Delete(t);

  Inc(prChunkCount);

end;

//****************************************************************************//
//                                                                            //
//                   T H R E A D S   D A T A    I / O                         //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_DataMT.prReadThreadOutput(Thread :TQUB_Thread);
var
  i :Integer;
begin

  for i := 0 to Thread.ChunkCount - 1 do begin
    if Thread.Chunk[i].Status = CHUNK_STATUS_DONE then Inc(prChunkCount);
  end;

end;

procedure TQUB_Task_DataMT.prWriteThreadInput(Thread :TQUB_Thread; Chunk :TQUB_Thread_PChunk);
begin

  Thread.Chunk[0]     := Chunk;
  Chunk.Status     := CHUNK_STATUS_RUNNING;
  Chunk.Completion := 0;

end;

procedure TQUB_Task_DataMT.prWriteThreadInputs(Thread :TQUB_Thread; c1, c2 :Integer);
var
  i :Integer;
begin

  Thread.ChunkCount := c2 - c1 + 1;
  for i := c1 to c2 do begin
    Thread.Chunk[i - c1] := prChunks[i];
    Thread.Chunk[i - c1].Status := CHUNK_STATUS_RUNNING;
    Thread.Chunk[i - c1].Completion := 0;
  end;

end;

procedure TQUB_Task_DataMT.prReadSocketOutput(Socket :TQUB_Thread_Socket);
begin

  Inc(prChunkCount);

end;

procedure TQUB_Task_DataMT.prWriteSocketInput(Socket :TQUB_Thread_Socket; Chunk :TQUB_Thread_PChunk);
begin

  Socket.Chunk     := Chunk;
  Chunk.Status     := CHUNK_STATUS_RUNNING;
  Chunk.Completion := 0;

end;


//******************************************************************************
end.
