{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// contains Actions: IdlWav and DataWav
// creates a wav file with frequencies determined by the amplitude of each point in the chosen data


unit QUB_Task_DataWav;


interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,

QUB_Viewport_Data, QUB_Viewport_Model,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_DataWav_Properties,

QUB_QVO_QVM, QUB_QFS_QMF, QUB_QFS_Channels,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python,
QUB_GUIInterface;


type


TQUB_Task_DataWav  = class(TQUB_Task_Data)
protected


  prModel                  :TQUB_QFS_QMF;

  prScale                  :Double;

  prBuffer                 :TPASmallInt;
  prBufIx                  :Integer;
  prBeginOfSeg             :Boolean;
  prStartTime              :Double;


  function prSetupSource: Integer; virtual; abstract; // returns seg count
  function prNextDwell(var Amp, Tm :Double): Boolean; virtual; abstract; // returns True if last in seg
  procedure prCleanupSource; virtual;

  procedure prSetupAmpTransform(var Slope, Icept :Double);

  // data output file: setup, buffered append, finalization
  procedure prSetupData;
  procedure prAddPoint(Point :SmallInt);
  procedure prFlushBuffer;
  procedure prNextSeg;
  procedure prFinishData;

  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  //Chunks    -- we don't use 'em
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  property Scale             :Double                    read prScale             write prScale;


end;


TQUB_Task_SampleDataWav = class(TQUB_Task_DataWav)
protected
  prDataSet                     :INode;
  prSeg                         :INode;
  prLoadSeg                     :INode;
  prLoadedData                  :TPADouble;
  prLoadedCount                 :Integer;
  prAt                          :Integer;
  prSampleTm                    :Double;

  function prSetupSource: Integer; override; // returns seg count
  function prNextDwell(var Amp, Tm :Double): Boolean; override; // returns True if last in seg
end;

TQUB_Task_IdealWav = class(TQUB_Task_DataWav)
protected
  prDataSet                     :INode;
  prSeg                         :INode;
  prCls, prTm                   :INode;
  prDwellCount                  :Integer;
  prIndex                       :Integer;

  function prSetupSource: Integer; override; // returns seg count
  function prNextDwell(var Amp, Tm :Double): Boolean; override; // returns True if last in seg
end;


TQUB_DataWav_WaveGen = class
protected

  prTable                       :array of SmallInt;
  prTableSize                   :Integer;
  prCurrEntry                   :Double;
  prEntryIncr                   :Double;

  prFreq                        :Double;
  prSampling                    :Double;
  prScaling                     :Double;

  procedure prSetFreq( NewFreq :Double );
  procedure prSetSampling( NewSamp :Double );
  procedure prSetScaling( NewScal  :Double );

public

  constructor Create(TableSize :Integer);

  function NextPoint: SmallInt;
  procedure Reset; // start from phase 0

  property Freq                  :Double                read prFreq              write prSetFreq;
  property Sampling              :Double                read prSampling          write prSetSampling;
  property Scaling               :Double                read prScaling           write prSetScaling;

end;



implementation


uses


QUB_Node_Model,
QUB_Python_Data;


const


DATAWAV_TABLE_SIZE              = 10240;
DATAWAV_BUFSIZE                 = 1000000;
QUB_TASK_DATAWAV_BASEFREQ       = 220;


//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _I D L W A V                   //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_DataWav.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'Data->WAV';

  prBuffer     := AllocMem( DATAWAV_BUFSIZE * SizeOf(SmallInt) );
  prBufIx      := 0;
  prBeginOfSeg := True;
  prStartTime  := 0.0;

end;

destructor TQUB_Task_DataWav.Destroy;
begin
  FreeMem( prBuffer );

  inherited;
end;

// *****************************************************************************
procedure TQUB_Task_DataWav.prSetupData;
begin
  // new data file (lock it)
  if ViewportData.New('', False{Silent}) then begin
    Exit;
  end;
  Data := ViewportData.CurrData;  // do this twice and face the wrath (it locks, with the matching unlock in Task_Data.ExecuteEnd)

  // instead of Silent=False, can we just set it to (WAV, 44.1KHz)?
end;

procedure TQUB_Task_DataWav.prAddPoint(Point :SmallInt);
begin
  prBuffer[ prBufIx ] := Point;

  prBufIx             := prBufIx + 1;
  if prBufIx = DATAWAV_BUFSIZE then begin
    prFlushBuffer;
  end;
end;

procedure TQUB_Task_DataWav.prFlushBuffer;
begin
  if prBufIx > 0 then begin
    TQUB_QVO_QVD(Data).AppendData(prBuffer, 1 * SizeOf(SmallInt), prBufIx, prBeginOfSeg, prStartTime, False{CanRefresh - Doit}, False{Acq});
    prStartTime  := prStartTime + prBufIx * TQUB_QVO_QVD(Data).Sampling;

    prBufIx      := 0;
    prBeginOfSeg := False;
  end;
end;

procedure TQUB_Task_DataWav.prNextSeg;
begin
  prFlushBuffer;
  prBeginOfSeg := True;
end;

procedure TQUB_Task_DataWav.prFinishData;
begin
  TQUB_QVO_QVD(Data).AppendData( nil, 0, 0, False, 0, True, False{Acq}); // update
end;

procedure TQUB_Task_DataWav.prSetupAmpTransform(var Slope, Icept :Double);
var
  Amp              :Double;
  MinAmp, MaxAmp   :Double;
  NClass           :Integer;   // for now cheat and say it's max class ix + 1
  i                :Integer;
begin
  NClass           := 0;

  if prModel.States.StateCount > 0 then begin
    MinAmp         := 1.0e20;
    MaxAmp         := -1.0e20;
    for i := 0 to prModel.States.StateCount - 1 do begin
      if prModel.States.State[i].Cls > NClass then NClass := prModel.States.State[i].Cls;

      Amp          := prModel.States.State[i].Amp;
      if Amp < MinAmp then MinAmp := Amp;
      if Amp > MaxAmp then MaxAmp := Amp;
    end;
    if MinAmp = MaxAmp then MaxAmp := MinAmp + 1.0;
  end else begin
    MinAmp         := 0.0;
    MaxAmp         := 1.0;
  end;

  NClass           := NClass + 1;
  if NClass < 2 then NClass := 2;

  Slope            := (NClass - 1) / (MaxAmp - MinAmp);
  Icept            := - (NClass - 1) * MinAmp / (MaxAmp - MinAmp);
end;
{
procedure TQUB_Task_DataWav.prSetupAmpTransform(var Slope, Icept, BaseFreq :Double);
var
  Model            :TQUB_QFS_QMF;
  Amp              :Double;
  MinAmp, MaxAmp   :Double;
  NClass           :Integer;   // for now cheat and say it's max class ix + 1
  BaseFreq         :Double;
  i                :Integer;
begin
  NClass           := 0;

  if prModel.States.StateCount > 0 then begin
    MinAmp         := 1.0e20;
    MaxAmp         := -1.0e20;
    for i := 0 to prModel.States.StateCount - 1 do begin
      if prModel.States.State[i].Cls > NClass then NClass := prModel.States.State[i].Cls;

      Amp          := prModel.States.State[i].Amp;
      if Amp < MinAmp then MinAmp := Amp;
      if Amp > MaxAmp then MaxAmp := Amp;
    end;
    if MinAmp = MaxAmp then MaxAmp := MinAmp + 1.0;
  end else begin
    MinAmp         := 0.0;
    MaxAmp         := 1.0;
  end;

  NClass           := NClass + 1;
  if NClass < 2 then NClass := 2;

  BaseFreq         := 2 * QUB_TASK_DATAWAV_CENTERFREQ / (NClass + 1);
  Slope            := BaseFreq * (NClass - 1) / (MaxAmp - MinAmp);
  Icept            := BaseFreq * (1 - (NClass - 1) * MinAmp / (MaxAmp - MinAmp));
end;
}
procedure TQUB_Task_DataWav.prCleanupSource;
begin
end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_DataWav.prDoTask;
var
  DwellLen         :Double; // output length in points
  Samples          :Integer;
  WaveGen          :TQUB_DataWav_WaveGen;
  SegCount         :Integer;
  Slope, Icept     :Double; // to transform amp into freq
  Amp, Tm          :Double;
  SegDone          :Boolean;
  j                :Integer;
begin
  SegCount := prSetupSource;
  if SegCount <= 0 then Exit;

  Synchronize( prSetupData );
  if prData = nil then Exit;

  prSetupAmpTransform(Slope, Icept);

  WaveGen          := TQUB_DataWav_WaveGen.Create( DATAWAV_TABLE_SIZE );
  WaveGen.Scaling  := 20000.0; // about 2/3 of WAV-max instead of TQUB_QVO_QVD(Data).Scaling;
  WaveGen.Sampling := TQUB_QVO_QVD(Data).Sampling;
  DwellLen         := 0.0;

  while SegCount > 0 do begin
    SegDone        := prNextDwell(Amp, Tm);

    WaveGen.Freq   := QUB_TASK_DATAWAV_BASEFREQ * Power(2, Amp * Slope + Icept);
    DwellLen       := DwellLen + Tm * Scale / TQUB_QVO_QVD(Data).Sampling;
    Samples        := Round(DwellLen);
    DwellLen       := DwellLen - Samples;

    for j := 0 to Samples - 1 do begin
      prAddPoint( WaveGen.NextPoint );
    end;

    if SegDone then begin
      prNextSeg;
      SegCount    := SegCount - 1;
    end;
  end;

  WaveGen.Free;
  prFinishData;

end;

procedure TQUB_Task_DataWav.prDoExecuteStart;
begin

  inherited;

end;

procedure TQUB_Task_DataWav.prDoExecuteEnd;
begin

  inherited;

end;


function TQUB_Task_DataWav.prDoExecuteInitialize :Boolean;
var
  Model         :TQUB_QVO_QVM;
begin
  Model         := ViewportModel.CurrModel;
  if Model <> nil then begin
    prModel     := Model.ModelObject;
    Result      := False;
  end else begin
    prModel     := nil;
    Result      := True;
  end;
end;

//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_DataWav.prChunkData :Boolean;
begin
  Result := False;
end;

procedure TQUB_Task_DataWav.prFreeChunk(Chunk :TQUB_Thread_PChunk);
begin
end;

function TQUB_Task_DataWav.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := nil;

end;

//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_DataWav.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_DataWav_PropertiesForm do begin

    Self.Scale                  := Properties['TimeScaling'].Data[0];

  end;//with
end;

procedure TQUB_Task_DataWav.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_DataWav_PropertiesForm do begin
  //  Scale                := Self.Scale;
  //end;//with
end;


//******************************************************************************
{  prDataSet                     :INode;
  prSeg                         :INode;
  prLoadSeg                     :INode;
  prLoadedData                  :TPADouble;
  prLoadedCount                 :Integer;
  prAt                          :Integer;
  prSampleTm                    :Double;
  // ...
}
function TQUB_Task_SampleDataWav.prSetupSource: Integer;
// returns seg count
var
  Seg               :INode;
begin
  prDataSet         := DataNoder.GetChosenData(TNode.CreateNull)[0];
  prDataSet['ProcessData'].SetData('AsDisplayed');

  Result := 0;
  Seg               := prDataSet.Find('Segment');
  while not Seg.IsNull do begin
    Result          := Result + 1;
    Seg             := Seg.SiblingSameName;
  end;

  prSeg             := prDataSet.Find('Segment');
  prLoadSeg         := prDataSet.InsertChild(TNode.CreateNull, 'Segment');
  prLoadSeg.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );

  if not prSeg.IsNull then begin
    prLoadSeg.Data[0] := prSeg.Data[0];
    prLoadSeg.Data[1] := Min( prSeg.Data[1], prLoadSeg.Data[0] + DATAWAV_BUFSIZE - 1 );
    DataNoder.FillData( prLoadSeg );
    prLoadedData      := TPADouble( prLoadSeg['Channel'].DataPtr );
    prLoadedCount     := prLoadSeg.Data[1] - prLoadSeg.Data[0] + 1;
    prAt              := 0;
  end;

  prSampleTm          := Double(prDataSet['sampling'].Data[0]) * 1.0e-3;

end;

function TQUB_Task_SampleDataWav.prNextDwell(var Amp, Tm :Double): Boolean;
// returns True if last in seg
begin
  Result            := False;
  if prSeg.IsNull then Exit;

  Amp               := prLoadedData[ prAt ];
  Tm                := prSampleTm;
  prAt              := prAt + 1;
  
  if prAt = prLoadedCount then begin
    prLoadSeg.RemoveChild( prLoadSeg['Channel'] );

    if prLoadSeg.Data[1] = prSeg.Data[1] then begin
      Result        := True;
      prSeg         := prSeg.SiblingSameName;
      if not prSeg.IsNull then begin
        prLoadSeg.Data[0] := prSeg.Data[0];
      end;
    end else begin
      prLoadSeg.Data[0]   := prLoadSeg.Data[1] + 1;
    end;

    if not prSeg.IsNull then begin
      prLoadSeg.Data[1] := Min( prSeg.Data[1], prLoadSeg.Data[0] + DATAWAV_BUFSIZE - 1 );
      DataNoder.FillData( prLoadSeg );
      prLoadedData      := TPADouble( prLoadSeg['Channel'].DataPtr );
      prLoadedCount     := prLoadSeg.Data[1] - prLoadSeg.Data[0] + 1;
      prAt              := 0;
    end;
  end;
end;

//******************************************************************************

function TQUB_Task_IdealWav.prSetupSource: Integer;
// returns seg count
var
  Seg               :INode;
begin
  prDataSet         := DataNoder.GetChosenIdealization(TNode.CreateNull)[0];
  Result            := 0;
  Seg               := prDataSet.Find('Segment');
  while not Seg.IsNull do begin
    Result          := Result + 1;
    Seg             := Seg.SiblingSameName;
  end;

  prSeg             := prDataSet.Find('Segment');
  prCls             := prSeg.Find('Classes');
  prTm              := prSeg.Find('Durations');
  prDwellCount      := Integer(prSeg['DwellCount'].Data[0]);
  prIndex           := 0;
end;

function TQUB_Task_IdealWav.prNextDwell(var Amp, Tm :Double): Boolean;
// returns True if last in seg
begin
  Result            := False;
  if prSeg.IsNull then Exit;

  Amp               := prModel.Amp[ Integer( prCls.Data[ prIndex ] ) ];
  Tm                := 1.0e-3 * Double(prTm.Data[ prIndex ]);
  prIndex           := prIndex + 1;
  if prIndex = prDwellCount then begin
    Result          := True;
    prSeg           := prSeg.SiblingSameName;
    if not prSeg.IsNull then begin
      prCls           := prSeg.Find('Classes');
      prTm            := prSeg.Find('Durations');
      prDwellCount    := Integer(prSeg['DwellCount'].Data[0]);
      prIndex         := 0;
    end;
  end;
end;



//******************************************************************************
constructor TQUB_DataWav_WaveGen.Create(TableSize :Integer);
begin
  prTableSize       := TableSize;
  SetLength( prTable, TableSize );
  prCurrEntry       := 0.0;

  prSampling        := 1.0 / 44100.0;
  Scaling           := 320; // (re-)fills table

  Freq              := 440.0; // sets prEntryIncr
end;

function TQUB_DataWav_WaveGen.NextPoint: SmallInt;
begin
  Result            := prTable[ Floor( prCurrEntry ) ];

  prCurrEntry       := prCurrEntry + prEntryIncr;
  while prCurrEntry >= prTableSize do begin
    prCurrEntry     := prCurrEntry - prTableSize;
  end;
end;

procedure TQUB_DataWav_WaveGen.Reset;
begin
  prCurrEntry := 0.0;
end;

procedure TQUB_DataWav_WaveGen.prSetFreq( NewFreq :Double );
begin
  prFreq      := NewFreq;
  prEntryIncr := prFreq * prTableSize * prSampling;
end;

procedure TQUB_DataWav_WaveGen.prSetSampling( NewSamp :Double );
begin
  prSampling := NewSamp;
  Freq       := Freq; // re-set EntryIncr
end;

procedure TQUB_DataWav_WaveGen.prSetScaling( NewScal :Double );
var
  i          :Integer;
begin
  prScaling := NewScal;

  for i := 0 to prTableSize - 1 do begin
    prTable[i] := Smallint( Round( prScaling * cos( 2.0 * PI * i / prTableSize ) ) );
  end;
end;

//******************************************************************************
end.

