{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Lorin's exponential fitter for macroscopic data
{
  New function added for fitting and removal (in ExtractSelection)
  of leak and capacitance artifacts in macroscopic data (LKC).
  12-04. LM

  10-21-06 LM. The LKC correction - as done in Extract - was modified to follow the voltage changes.
  It's no longer limited to just one voltage step.
}


unit QUB_Task_Exp;


interface


uses


SysUtils, Classes, Windows,

Alloc, AllocTypes,
qubopt_iface,

QUB_Global,
QUB_Q_Global, QUB_Fit_Model, QUB_Exp_Model, QUB_LKC_Model, QUB_Sin_Model, QUB_Exp_Global,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_Thread,
QUB_Task, QUB_Task_Data,
QUB_QFS_QDF,
QUB_IdlCond,
QUB_ExpCond_Parameters,
QUB_Task_Exp_Properties;

//KPL_Parser;


type


TQUB_Task_Exp  = class(TQUB_Task_Data)
protected


  //PropertiesForm
  prPropForm               :TQUB_Task_Exp_PropertiesForm;

  //Source Objects
  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  //Optimizer Parameters
  prOptParams              :TQUB_Exp_Opt_Params;

  //Models
  prFitModel               :TQUB_Fit_Model;
  prExpModel               :TQUB_Exp_Model;
  prLKCModel               :TQUB_LKC_Model;
  prSinModel               :TQUB_Sin_Model;

  //Previous SS
  prPSS                    :Double;

  //Best Grads
  prfParsGrads             :TPADouble;

  //OptSegments
  prOptSegsChunks          :TList;
  prOptSegsOutput          :String;

  prQuietOutput            :Boolean;

  prUpdateParams           :Boolean;

  prFitPreCleared          :Boolean;

  prFitDataListName        :String;
  prGroupList              :String;
  prBatchCount             :Integer;
  prBatchIdx               :Integer;

  prDataSetCount           :Integer;

  prQuietEntry             :Integer;

  prFitList                :TPAPoint;
  prFitListCount           :Integer;

  prMethod                 :Integer;


  //Execute
  procedure prDoTask; override;

  procedure prDoExecute; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;

  //Chunks
  procedure prSetChunkParameters; override;
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  //Update visuals
  procedure prResetModel;
  procedure prUpdateData;
  procedure prShowData;
  procedure prRefreshZoom;

  procedure prDoOverlapData;

  function DataCallBack(ExpChunk :TQUB_Exp_PChunk) :Integer;

  procedure prExp;
  function prSSGrad(CompGrad :Integer) :Integer;

  procedure prPrintOptParams;
  procedure prPrintParams;
  procedure prPrintGrads;
  procedure prDoUpdateParams;
  

public


  constructor Create(Suspended :Boolean); override;


  function DFPFunc(Pars :TPADouble; Grad :TPADouble; var SS :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
  function DFPCheck(Iters, nf, ndf :Integer; SS :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Parameters
  property OptParams         :TQUB_Exp_Opt_Params          read prOptParams         write prOptParams;

  property QuietOutput       :Boolean                      read prQuietOutput       write prQuietOutput;
  property UpdateParams      :Boolean                      read prUpdateParams      write prUpdateParams;

  property FitDataListName   :String                       read prFitDataListName   write prFitDataListName;


end;


implementation


uses


QUB_Main, QUB_GUIInterface, QUB_Viewport_Data, QUB_Python_Data, QUB_FitCurves,
QUB_Node;


//DFP Callbacks ****************************************************************
function DFPFunc(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
begin
  Result := TQUB_Task_Exp(call_obj).DFPFunc(Pars, Grad, LL, nf, ndf, CompGrad);
end;

function DFPCheck(call_obj :TObject; Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
begin
  Result := TQUB_Task_Exp(call_obj).DFPCheck(Iters, nf, ndf, LL, Pars, Grads);
end;



//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _ E x p                        //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_Exp.Create(Suspended :Boolean);
begin

  inherited;

  prName := 'Exp optimizer';

end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Exp.prDoTask;
begin

  inherited;

  prExp;

end;

procedure TQUB_Task_Exp.prExp;
var
  Done             :Boolean;
  Cnt              :Integer;
  m, n             :Integer;
  BatchFirst       :Integer;
  BatchIter        :Integer;
  BatchCountCount  :Integer;
  BatchCountList   :TPAInteger;
begin


  prOptSegsChunks := TList.Create;

  BatchCountCount := 1;
  BatchCountList  := nil;


  if prOptParams.OptSegments = OPTSEGS_GROUPLIST then begin
    BatchCountCount := 100;
    BatchCountList  := IAlloc1(BatchCountCount);
    StrToArrayOfInt(prGroupList, BatchCountList, BatchCountCount);
    if BatchCountCount = 0 then begin
      prOptParams.OptSegments := OPTSEGS_GROUP;
      BatchCountCount := 1;
    end;
  end;


  prFitList      := nil;
  prFitListCount := 0;


  try


    for BatchIter := 0 to BatchCountCount - 1 do begin



      if prStopped then Break;

      //Save them all
      for m := 0 to prChunks.Count - 1 do begin
        prOptSegsChunks.Add(prChunks[m]);
      end;

      //Initialize pointers in list
      BatchFirst := 0;
      if          prOptParams.OptSegments = OPTSEGS_TOGETHER{ALL} then begin
        prBatchCount := prChunks.Count;
      end else if prOptParams.OptSegments = OPTSEGS_GROUP then begin
        prBatchCount := prOptParams.OptSegmentsGroupCount;
      end else begin
        prBatchCount := Abs(BatchCountList[BatchIter]);
      end;


      Synchronize(prPrintOptParams);


      prBatchIdx := 0;


      while (prBatchIdx < prOptParams.OptSegmentsMaxBatch) and (BatchFirst < prOptSegsChunks.Count) and (not prStopped) do begin
      //Batches


        Inc(prBatchIdx);


        if (prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST) then begin
          Synchronize(prResetModel);
        end;

        if (BatchFirst + prBatchCount - 1) >= prOptSegsChunks.Count then begin
          prBatchCount := prOptSegsChunks.Count - BatchFirst;
        end;


        //Next Batch
        prChunks.Clear;


        for m := BatchFirst to BatchFirst + prBatchCount - 1 do begin
          prChunks.Add(prOptSegsChunks[m]);
        end;


        //MUX files
        if prOptParams.OptMUXFiles = 1 then begin
          for m := BatchFirst + prBatchCount to prOptSegsChunks.Count - 1 do begin
            if TQUB_Exp_PChunk(prOptSegsChunks[BatchFirst]).QVD <> TQUB_Exp_PChunk(prOptSegsChunks[m]).QVD then begin
              for n := m + BatchFirst to m + BatchFirst + prBatchCount - 1 do begin
                if n >= prOptSegsChunks.Count then Break;
                prChunks.Add(prOptSegsChunks[n]);
              end;
              Break;
            end;
          end;
        end;


        Synchronize(prShowData);


        //**********************************************************************
        Cnt  := 0;

        repeat
        //restart if Grads > some threshold...


          if Cnt > 0 then begin
            prConsoleMsg  := 'Restarting...';
            prConsoleCode := MESSAGE_WARNING;
            if not prQuietOutput then Synchronize(prWriteNewConsole);
            prpSS := 1e10;
          end;

          Done := True;


          prfParsGrads := DAlloc1(prFitModel.fParCount);
          for m := 0 to prFitModel.fParCount - 1 do prfParsGrads[m] := 0;


          try


            prFitModel.Do_fParsToModel;


            if prOptParams.OptMode = MODE_OPTIMIZE then begin
            //Optimize

              if prQuietOutput then begin
                prConsoleMsg  := Format('Batch index %d: starting ...', [prBatchIdx]);
                prConsoleCode := MESSAGE_WARNING;
                Synchronize(prWriteNewConsole);
                prQuietEntry := prConsoleEntry;
              end;

              try
                DFP_MIN(Self, prFitModel.fParCount, prFitModel.fPars, prFitModel.fParsVar,
                        QUB_Task_Exp.DFPFunc, QUB_Task_Exp.DFPCheck,
                        OptParams.OptMaxIter, OptParams.OptSSConv, OptParams.OptGradConv, OptParams.OptMaxStep,
                        prFitModel.Error);
              except
              end;

              prFitModel.Do_Varf2Varc;


            end else begin
            //Evaluate Sum-Of-Squares


              prSSGrad(1{CompGrad});

              prFitModel.Do_Varf2Varc;


              prPrintGrads;


            end;


            prPrintParams;


          finally


            //Check Grads:
            for m := 0 to prFitModel.fParCount - 1 do begin
              if Abs(prfParsGrads[m]) >= 1 then begin
                Done := False;
                Break;
              end;
            end;

            FreeMem(prfParsGrads);


          end;


          Inc(Cnt);


          prFitModel.Iterations := 0;

          if (OptParams.OptRestart = 0) and (Cnt > 0) then Break;

        until ((Done) or (Cnt >= 2) or (prStopped) or (prError));
        //************************************************************************


        prConsoleMsg  := prOptSegsOutput;
        prConsoleCode := MESSAGE_MESSAGE;
        if prQuietOutput then begin
          prConsoleEntry := prQuietEntry;
          Synchronize(prWriteSetConsole);
        end else begin
          Synchronize(prWriteNewConsole);
        end;


        //Next batch
        Inc(BatchFirst, prBatchCount);


      end;


      if ((prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST)) and (prBatchCount > 1) then Synchronize(prResetModel);

      prChunks.Clear;

      for m := 0 to prOptSegsChunks.Count - 1 do begin
        prChunks.Add(prOptSegsChunks[m]);
      end;

      prOptSegsChunks.Clear;


    end;//BatchIter


  finally

    prOptSegsChunks.Free;
    FreeMem(BatchCountList);

    FreeMem(prFitList);

  end;

end;


procedure TQUB_Task_Exp.prPrintOptParams;
var
  t :String;
begin

  with prOptParams do begin
    t :=     #13#10'Exp parameters for:';
    t := t + #13#10#9'data  '    + Choice(prData  <> nil, ExtractFileName(prData.FileName),  '<none>');
    t := t +         ' SegCount ' + IntToStr(prBatchCount) + '.';
    t := t +         ' V = ' + FloatToStr(TQUB_Exp_PChunk(prChunks[0]).V) + ' V0 = ' + FloatToStr(TQUB_Exp_PChunk(prChunks[0]).V0);
    t := t + #13#10;
  end;

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
end;


procedure TQUB_Task_Exp.prPrintParams;
var
  SS :Double;
  t  :String;
  k  :Integer;
begin

  t :=     #13#10'================================================================';
  prOptSegsOutput :=     '';

  SS := prFitModel.SS;

  try

    //METHOD_EXP
    if          prMethod = METHOD_EXP then begin

      t := t + Format(#13#10'Exp final results:  SS = %-10.6g Iterations = %d'#13#10 +
                      #9'--------------------------------------------------'#13#10 +
                      #9'Ieq       '#9'Std       ',
                     [SS, prFitModel.Iterations]);
      for k := 0 to prExpModel.StateCount - 2 do begin
        t := t + Format(#9'Tau %d    '#9'Std       '#9'Amp %d    '#9'Std       ', [k + 1, k + 1]);
      end;

    //METHOD_LKC
    end else if prMethod = METHOD_LKC then begin

      t := t + Format(#13#10'LKC final results:  SS = %-10.6g Iterations = %d'#13#10 +
                      #9'--------------------------------------------------'#13#10 +
                      #9'GL        '#9'Std       '#9'VL [mV]   '#9'Std       ',
                     [SS, prFitModel.Iterations]);
      for k := 0 to prLKCModel.N - 1 do begin
        t := t + Format(#9'Tau %d[ms]'#9'Std       '#9'Amp %d    '#9'Std       ', [k + 1, k + 1]);
      end;

    //METHOD_SIN
    end else if prMethod = METHOD_SIN then begin

      t := t + Format(#13#10'Sine final results:  SS = %-10.6g Iterations = %d'#13#10 +
                      #9'--------------------------------------------------'#13#10 +
                      #9'Ieq       '#9'Std       '#9'Amp       '#9'Std       '#9'Freq      '#9'Std       '#9'Phase     '#9'Std       ',
                     [SS, prFitModel.Iterations]);
    end;

    t := t + #13#10;

  except
  end;


  try

    //METHOD_EXP
    if prMethod = METHOD_EXP then begin

      for k := 0 to prFitModel.cParCount - 1 do begin
        if Odd(k) then begin
        //Tau
          t := t + Format(#9'%-10.6g'#9'%-10.6g ', [Exp(prExpModel.cPars[k]), Sqrt(prExpModel.cParsVar[k])]);
          prOptSegsOutput := prOptSegsOutput +
                             Trim(Format('%10.6g', [Exp (prExpModel.cPars   [k])])) + #9 +
                             Trim(Format('%10.6g', [Sqrt(prExpModel.cParsVar[k])])) + #9;
        end else begin
        //Amp
          t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prExpModel.cPars[k],      Sqrt(prExpModel.cParsVar[k])]);
          prOptSegsOutput := prOptSegsOutput +
                             Trim(Format('%10.6g', [     prExpModel.cPars   [k]]))  + #9 +
                             Trim(Format('%10.6g', [Sqrt(prExpModel.cParsVar[k])])) + #9;
        end;
      end;

    //METHOD_LKC
    end else if prMethod = METHOD_LKC then begin

      k := 0;
      t := t + Format(#9'%-10.6g'#9'%-10.6g ', [Exp(prLKCModel.cPars[k]), Sqrt(prLKCModel.cParsVar[k])]);
      prOptSegsOutput := prOptSegsOutput +
                         Trim(Format('%10.6g', [Exp (prLKCModel.cPars   [k])])) + #9 +
                         Trim(Format('%10.6g', [Sqrt(prLKCModel.cParsVar[k])])) + #9;
      Inc(k);
      t := t + Format(#9'%-10.6g'#9'%-10.6g ', [    prLKCModel.cPars[k],  Sqrt(prLKCModel.cParsVar[k])]);
      prOptSegsOutput := prOptSegsOutput +
                         Trim(Format('%10.6g', [     prLKCModel.cPars   [k]]))  + #9 +
                         Trim(Format('%10.6g', [Sqrt(prLKCModel.cParsVar[k])])) + #9;
      Inc(k);
      for k := k to prLKCModel.cParCount - 1 do begin
        if not Odd(k) then begin
        //Tau
          t := t + Format(#9'%-10.6g'#9'%-10.6g ', [1e3*Exp(prLKCModel.cPars[k]), 1e3*Sqrt(prLKCModel.cParsVar[k])]);
          prOptSegsOutput := prOptSegsOutput +
                             Trim(Format('%10.6g', [1e3*Exp (prLKCModel.cPars   [k])])) + #9 +
                             Trim(Format('%10.6g', [1e3*Sqrt(prLKCModel.cParsVar[k])])) + #9;
        end else begin
        //Amp
          t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prLKCModel.cPars[k], Sqrt(prLKCModel.cParsVar[k])]);
          prOptSegsOutput := prOptSegsOutput +
                             Trim(Format('%10.6g', [     prLKCModel.cPars   [k]]))  + #9 +
                             Trim(Format('%10.6g', [Sqrt(prLKCModel.cParsVar[k])])) + #9;
        end;
      end;

    //METHOD_SIN
    end else if prMethod = METHOD_SIN then begin

      //Ieq
      if not prSinModel.FixedIeq then begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prSinModel.cPars[prSinModel.IdxIeq], Sqrt(prSinModel.cParsVar[prSinModel.IdxIeq])]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [     prSinModel.cPars   [prSinModel.IdxIeq]]))  + #9 +
                           Trim(Format('%10.6g', [Sqrt(prSinModel.cParsVar[prSinModel.IdxIeq])])) + #9;
      end else begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prSinModel.Ieq, 1.0]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [prSinModel.Ieq]))  + #9 +
                           Trim(Format('%10.6g', [1.0]))             + #9;
      end;
      //Amp
      if not prSinModel.FixedAmp then begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prSinModel.cPars[prSinModel.IdxAmp], Sqrt(prSinModel.cParsVar[prSinModel.IdxAmp])]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [     prSinModel.cPars   [prSinModel.IdxAmp]]))  + #9 +
                           Trim(Format('%10.6g', [Sqrt(prSinModel.cParsVar[prSinModel.IdxAmp])])) + #9;
      end else begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prSinModel.Amp, 1.0]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [prSinModel.Amp]))  + #9 +
                           Trim(Format('%10.6g', [1.0]))             + #9;
      end;
      //Freq
      if not prSinModel.FixedFreq then begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [Exp(prSinModel.cPars[prSinModel.IdxFreq]), Sqrt(prSinModel.cParsVar[prSinModel.IdxFreq])]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [Exp (prSinModel.cPars   [prSinModel.IdxFreq])])) + #9 +
                           Trim(Format('%10.6g', [Sqrt(prSinModel.cParsVar[prSinModel.IdxFreq])])) + #9;
      end else begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prSinModel.Freq, 1.0]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [prSinModel.Freq]))  + #9 +
                           Trim(Format('%10.6g', [1.0]))              + #9;
      end;
      //Phase
      if not prSinModel.FixedPhase then begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prSinModel.cPars[prSinModel.IdxPhase], Sqrt(prSinModel.cParsVar[prSinModel.IdxPhase])]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [     prSinModel.cPars   [prSinModel.IdxPhase]]))  + #9 +
                           Trim(Format('%10.6g', [Sqrt(prSinModel.cParsVar[prSinModel.IdxPhase])])) + #9;
      end else begin
        t := t + Format(#9'%-10.6g'#9'%-10.6g ', [prSinModel.Phase, 1.0]);
        prOptSegsOutput := prOptSegsOutput +
                           Trim(Format('%10.6g', [prSinModel.Phase]))  + #9 +
                           Trim(Format('%10.6g', [1.0]))               + #9;
      end;

    end;

  except
  end;

  prOptSegsOutput := prOptSegsOutput + Trim(Format('%14.10g', [SS]));


  t := t + #13#10#9'--------------------------------------------------';

  //prBatchOutput := #13#10 + prBatchOutput;


  if OptParams.OptMode = MODE_OPTIMIZE then begin
    t := t + #13#10'Free pars grads: ';
    for k := 0 to prFitModel.fParCount - 1 do begin
      try
        t := t + Format('%-12.8g ', [-prfParsGrads[k]]);
      except
      end;
    end;
  end;

  t := t + #13#10 + '================================================================';

  //UpdateParams
  if prUpdateParams then begin
    prDoUpdateParams;
  end;

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);
end;

procedure TQUB_Task_Exp.prDoUpdateParams;
var
  Props :INode;
  k     :Integer;
  N     :Integer;
begin

  Props := TNode.Create('Properties');

  case prMethod of

    //EXP
    METHOD_EXP:
      begin
        Props['Ieq'].SetData(prExpModel.Ieq);
        N         := prPropForm.Properties['StateCount'].Data[0] - 1;
        for k := 0 to N - 1 do begin
          prPropForm.Tau[k] := prExpModel.Tau[k];
          prPropForm.Amp[k] := prExpModel.Amp[k];
        end;
      end;

    //LKC
    METHOD_LKC:
      begin
        Props['LKCGL'].SetData(prLKCModel.GL);
        Props['LKCVL'].SetData(prLKCModel.VL);
        N         := prPropForm.Properties['LKCN'].Data[0];
        for k := 0 to N - 1 do begin
          prPropForm.LKCTau[k] := prLKCModel.Tau[k];
          prPropForm.LKCAmp[k] := prLKCModel.Amp[k];
        end;
      end;

    //SIN
    METHOD_SIN:
      begin
        Props['SineIeq'].SetData(prSinModel.Ieq);
        Props['SineAmp'].SetData(prSinModel.Amp);
        Props['SineFreq'].SetData(prSinModel.Freq);
        Props['SinePhase'].SetData(prSinModel.Phase);
      end;

  end;//case

  prPropForm.SetProperties(Props);

end;

procedure TQUB_Task_Exp.prPrintGrads;
var
  SS  :Double;
  dSS :Double;
  t   :String;
  k   :Integer;
begin

  SS  := prFitModel.SS;

  prConsoleMsg  := Format('Batch %d-%d, SS: %-14.10g, Gr: ', [prBatchIdx, prFitModel.Iterations, SS]);

  t := Format('Iteration %d, SS = %-14.10g, Func calls = %d, Grad calls = %d',
              [prFitModel.Iterations, SS, prFitModel.FuncCalls, prFitModel.GradCalls]);

  t := t + #13#10#9'Free pars grads: ';
  for k := 0 to prFitModel.fParCount - 1 do begin
    dSS := prFitModel.dSSdfPar[k];
    try
      t := t + Format('%-12.8g ', [dSS]);
      prConsoleMsg  := prConsoleMsg + Format('%-12.8g ', [dSS]);
    except
    end;
  end;

  //t must be assigned here
  if prQuietOutput then begin
    prConsoleCode := MESSAGE_WARNING;
    Synchronize(prWriteSetConsole);
  end;

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);

end;


//******************************************************************************
function TQUB_Task_Exp.DFPFunc(Pars :TPADouble; Grad :TPADouble; var SS :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
var
  i :Integer;
begin

  Result := 0;

  if prFitModel.Iterations > 0 then begin

    //Copy the new pars into the Fit model
    for i := 0 to prFitModel.fParCount - 1 do begin
      prFitModel.fPars[i] := Pars[i];
    end;

    prFitModel.Do_fParsToModel;

  end;


  try
    prSSGrad(CompGrad);
  except
  end;


  SS := prFitModel.SS;


  if CompGrad <> 0 then begin
    for i := 0 to prFitModel.fParCount - 1 do begin
      Grad[i] := prFitModel.dSSdfPar[i];
    end;
    Move(Grad[0], prfParsGrads[0], prFitModel.fParCount * SizeOf(Double));
  end;


  Inc(prFitModel.Iterations);
  Inc(nf);
  if CompGrad <> 0 then Inc(ndf);
  prFitModel.FuncCalls := nf;
  prFitModel.GradCalls := ndf;

end;

function TQUB_Task_Exp.DFPCheck(Iters, nf, ndf :Integer; SS :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
var
  SSConverged :Boolean;
  GRConverged :Boolean;
  k           :Integer;
begin

  Result := 0;

  prFitModel.Iterations := Iters;

  if Iters >= OptParams.OptMaxIter then Result := 1;

  SSConverged := Abs(SS - prpSS) < OptParams.OptSSConv;
  prpSS := SS;

  GRConverged := True;
  for k := 0 to prFitModel.fParCount - 1 do begin
    if Abs(Grads[k]) > OptParams.OptGradConv then GRConverged := False;
  end;


  //Synchronize(prUpdateModel);
  Synchronize(prUpdateData);


  prPrintGrads;


  if (SSConverged) and{or} (GRConverged) then Result := 1;
  if prStopped then Result := 1;

  //if Result = 1 then begin
    //Change the parameters to the last good ones (???)
    //Copy the new pars into the Fit model
    for k := 0 to prFitModel.fParCount - 1 do begin
      prFitModel.fPars[k] := Pars[k];
    end;
    prFitModel.Do_fParsToModel;

    //Only the fParsGrads
    Move(Grads[0], prfParsGrads[0], prFitModel.fParCount * SizeOf(Double));
  //end;

end;



//****************************************************************************//
//                                                                            //
//                             S S   +   G R A D                              //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Exp.prSSGrad(CompGrad :Integer) :Integer;
var
  ExpChunk            :TQUB_Exp_PChunk;
  FitChunk            :TQUB_Exp_PChunk;
  Chunk               :TQUB_Thread_PChunk;
  DataCount           :Integer;
  IData               :TPAInteger;
  FData               :Double;
  cSS, cGR            :Double;
  c, k, l, t          :Integer;
  Scale               :Double;
  dt                  :Double;
  mt, dmt             :Double;
  FitThisPoint        :Boolean;
  mc                  :Integer;
  FitDataList         :TQUB_QVO_QVD_SelectionList;
  FitDataListCount    :Integer;
  FitListIdx          :Integer;

procedure ChunkData(c :Integer);
begin
  ExpChunk  := prChunks[c];
  Chunk     := prChunks[c];
  if ExpChunk.Data = nil then begin
    DataCallback(ExpChunk);
  end;
  IData     := ExpChunk.Data;
  DataCount := TQUB_Thread_PChunk(ExpChunk).Sel.Last - TQUB_Thread_PChunk(ExpChunk).Sel.First + 1;
  Scale     := 1 / ExpChunk.Scaling;
  dt        := ExpChunk.Sampling;
end;

procedure UnlockAndNullifyFitDataList;
begin
  if FitDataList <> nil then begin
    FitDataList.Unlock;
    FitDataList := nil;
  end;
end;

//The fit list must be sorted, and selections should not overlap
procedure AssignAndLockFitDataList(ChunkIndex :Integer);
var
  f, s :Integer;
begin
  FitDataList := TQUB_Exp_PChunk(prChunks[ChunkIndex]).QVD.SelectionLists.ListByName[FitDataListName];
  if FitDataList <> nil then begin
    FitDataList.Lock;
    FitDataListCount := FitDataList.Items.Count;
    if prFitListCount < FitDataListCount then begin
      prFitListCount := FitDataListCount;
      ReAllocMem(prFitList, prFitListCount * SizeOf(TPoint));
    end;
    f := TQUB_Thread_PChunk(prChunks[ChunkIndex]).Sel.First;
    for s := 0 to FitDataListCount - 1 do begin
      prFitList[s].x := FitDataList.Selection[s].Sel.First - f;
      prFitList[s].y := FitDataList.Selection[s].Sel.Last  - f;
    end;
    FitListIdx := 0;
  end;
end;


begin


  Result      := 0;


  FitDataList := nil;


  //Reset SS and derivatives ***************************************************
  prFitModel.ResetSSandGR;
  //****************************************************************************


  if OptParams.OptIdenticalChunks = 1 then begin
  //Identical chunks: faster, but it loads all data
  //Obviously, use only with identical chunks, whether in the same file or not


    //Set FitDataList
    if (prOptParams.OptFitDataMode = FITDATA_LIST) then begin
      UnlockAndNullifyFitDataList;
      AssignAndLockFitDataList(0);
    end;


    try


      //Load data
      for c := prChunks.Count - 1 downto 0 do begin
        ChunkData(c);
      end;


      //For each DataPoint
      for t := 0 to DataCount - 1 do begin


        //Fit this point ?
        FitThisPoint := True;
        if (prOptParams.OptSkip = 1) and ((t*dt) <= (prOptParams.OptSkipTime*1e-3)) then begin
          FitThisPoint := False;
        end;
        if (FitThisPoint) and (FitDataList <> nil) then begin
          FitThisPoint := False;
          for l := FitListIdx to FitDataListCount - 1 do begin
            if (t >= prFitList[l].x) and (t <= prFitList[l].y) then begin
              FitThisPoint := True;
              FitListIdx := l;
              Break;
            end;
          end;
        end;


        //**********************************************************************//
        //                                                                      //
        //                             Fit this point !                         //
        //                                                                      //
        //**********************************************************************//
        if FitThisPoint then begin


          //METHOD_EXP
          if prMethod = METHOD_EXP then begin

            //I Mean
            prFitModel.Do_Mean_t(mt, t*dt);

            //Sum-Of-Squares .......................................................
            for c := 0 to prChunks.Count - 1 do begin
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;//Does it include ChannelScale?!

              cSS := Sqr(FData - mt);

              prFitModel.SS := prFitModel.SS + cSS;//Use minus sign in order to minimize SS
            end;

            //Ieq Gradient .........................................................
            prExpModel.Do_dMean_t_dIeq(dmt, t*dt);

            for c := 0 to prChunks.Count - 1 do begin

              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;

              cGR := -2*(FData - mt)*dmt;

              prExpModel.dSSdIeq := prExpModel.dSSdIeq + cGR;

            end;//c


            //Tau gradient
            for k := 0 to prExpModel.StateCount - 2 do begin

              prExpModel.Do_dMean_t_dTau(dmt, t*dt, k);

              for c := 0 to prChunks.Count - 1 do begin

                FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;

                cGR := -2*(FData - mt)*dmt;

                prExpModel.dSSdTau[k] := prExpModel.dSSdTau[k] + cGR;

              end;//c

            end;//k

            //Amp gradient
            for k := 0 to prExpModel.StateCount - 2 do begin

              prExpModel.Do_dMean_t_dAmp(dmt, t*dt, k);

              for c := 0 to prChunks.Count - 1 do begin

                FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;

                cGR := -2*(FData - mt)*dmt;

                prExpModel.dSSdAmp[k] := prExpModel.dSSdAmp[k] + cGR;

              end;//c

            end;//k


          //METHOD_LKC @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
          end else if prMethod = METHOD_LKC then begin

            //Each Chunk presumably has a different V or V0
            for c := 0 to prChunks.Count - 1 do begin

              prLKCModel.V  := TQUB_Exp_PChunk(prChunks[c]).V;
              prLKCModel.V0 := TQUB_Exp_PChunk(prChunks[c]).V0;

              //I Mean
              prFitModel.Do_Mean_t(mt, t*dt);

              //Sum-Of-Squares .......................................................
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;//Does it include ChannelScale?!
              cSS := Sqr(FData - mt);
              prFitModel.SS := prFitModel.SS + cSS;//Use minus sign in order to minimize SS

              //GL Gradient ........................................................
              prLKCModel.Do_dMean_t_dGL(dmt, t*dt);
              cGR := -2*(FData - mt)*dmt;
              prLKCModel.dSSdGL := prLKCModel.dSSdGL + cGR;

              //VL Gradient ........................................................
              prLKCModel.Do_dMean_t_dVL(dmt, t*dt);
              cGR := -2*(FData - mt)*dmt;
              prLKCModel.dSSdVL := prLKCModel.dSSdVL + cGR;

              //Tau gradient .......................................................
              for k := 0 to prLKCModel.N - 1 do begin
                prLKCModel.Do_dMean_t_dTau(dmt, t*dt, k);
                cGR := -2*(FData - mt)*dmt;
                prLKCModel.dSSdTau[k] := prLKCModel.dSSdTau[k] + cGR;
              end;//k

              //Amp gradient .......................................................
              for k := 0 to prLKCModel.N - 1 do begin
                prLKCModel.Do_dMean_t_dAmp(dmt, t*dt, k);
                cGR := -2*(FData - mt)*dmt;
                prLKCModel.dSSdAmp[k] := prLKCModel.dSSdAmp[k] + cGR;
              end;//k

            end;//c

          //METHOD_SIN @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
          end else if prMethod = METHOD_SIN then begin

            //I Mean
            prFitModel.Do_Mean_t(mt, t*dt);

            //Sum-Of-Squares .......................................................
            for c := 0 to prChunks.Count - 1 do begin
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;//Does it include ChannelScale?!

              cSS := Sqr(FData - mt);

              prFitModel.SS := prFitModel.SS + cSS;//Use minus sign in order to minimize SS
            end;

            //Ieq Gradient .......................................................
            prSinModel.Do_dMean_t_dIeq(dmt, t*dt);
            for c := 0 to prChunks.Count - 1 do begin
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdIeq := prSinModel.dSSdIeq + cGR;
            end;//c

            //Amp Gradient .......................................................
            prSinModel.Do_dMean_t_dAmp(dmt, t*dt);
            for c := 0 to prChunks.Count - 1 do begin
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdAmp := prSinModel.dSSdAmp + cGR;
            end;//c

            //Freq Gradient ......................................................
            prSinModel.Do_dMean_t_dFreq(dmt, t*dt);
            for c := 0 to prChunks.Count - 1 do begin
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdFreq := prSinModel.dSSdFreq + cGR;
            end;//c

            //Phase Gradient .....................................................
            prSinModel.Do_dMean_t_dPhase(dmt, t*dt);
            for c := 0 to prChunks.Count - 1 do begin
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdPhase := prSinModel.dSSdPhase + cGR;
            end;//c

          end;//METHOD


        end;//Fit this point!



        //Show fitting curve ?
        if (OptParams.OptShowFit <> 0) then begin

          for c := 0 to prChunks.Count - 1 do begin

            FitChunk := TQUB_Exp_PChunk(prChunks[c]);


            //METHOD_LKC @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
            if prMethod = METHOD_LKC then begin

              prLKCModel.V  := TQUB_Exp_PChunk(prChunks[c]).V;
              prLKCModel.V0 := TQUB_Exp_PChunk(prChunks[c]).V0;

            end;

            //I Mean
            prFitModel.Do_Mean_t(mt, t*dt);

            //************************************************************
            if FitThisPoint then begin
              if          OptParams.OptShowFitWhat = SHOWFIT_AVG then begin
                FitChunk.FitCurve[t] := Round((mt-FitChunk.Offset) * FitChunk.Scaling);
              end else if OptParams.OptShowFitWhat = SHOWFIT_RES then begin
                FitChunk.FitCurve[t] := Round((mt-FitChunk.Offset) * FitChunk.Scaling - TQUB_Exp_PChunk(prChunks[c]).Data[t]);
              end;
            end else begin
              if          OptParams.OptShowFitWhat = SHOWFIT_AVG then begin
                FitChunk.FitCurve[t] := TQUB_Exp_PChunk(prChunks[c]).Data[t];
              end else if OptParams.OptShowFitWhat = SHOWFIT_RES then begin
                FitChunk.FitCurve[t] := 0;
              end;
            end;

          end;//c

        end;//ShowFit


      end;//of t


    finally

      UnlockAndNullifyFitDataList;

    end;


  end else begin
  //Separate treatment of each chunk
  //It still loads all data


    //For each chunk *************************************************************
    for c := 0 to prChunks.Count - 1 do begin


      ChunkData(c);


      //Set FitDataList
      if (prOptParams.OptFitDataMode = FITDATA_LIST) then begin
        UnlockAndNullifyFitDataList;
        AssignAndLockFitDataList(c);
      end;


      try


        //For each DataPoint
        for t := 0 to DataCount - 1 do begin


          //Fit this point ?
          FitThisPoint := True;
          if (prOptParams.OptSkip = 1) and (t*dt <= prOptParams.OptSkipTime*1e-3) then begin
            FitThisPoint := False;
          end;
          if (FitThisPoint) and (prOptParams.OptFitDataMode = FITDATA_LIST) and (FitDataList <> nil) then begin
            FitThisPoint := False;
            for l := 0 to FitDataList.Items.Count - 1 do begin
              if (t >= FitDataList.Selection[l].Sel.First) and (t <= FitDataList.Selection[l].Sel.Last) then begin
                FitThisPoint := True;
                Break;
              end;
            end;
          end;



          //**********************************************************************//
          //                                                                      //
          //                             Fit this point !                         //
          //                                                                      //
          //**********************************************************************//
          if FitThisPoint then begin


            //METHOD_EXP
            if prMethod = METHOD_EXP then begin

              //I Mean
              prFitModel.Do_Mean_t(mt, t*dt);

              //Sum-Of-Squares .......................................................
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;//Does it include ChannelScale?!
              cSS := Sqr(FData - mt);
              prFitModel.SS := prFitModel.SS + cSS;//Use minus sign in order to minimize SS

              //Ieq Gradient .........................................................
              prExpModel.Do_dMean_t_dIeq(dmt, t*dt);
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prExpModel.dSSdIeq := prExpModel.dSSdIeq + cGR;

              //Tau gradient
              for k := 0 to prExpModel.StateCount - 2 do begin
                prExpModel.Do_dMean_t_dTau(dmt, t*dt, k);
                FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
                cGR := -2*(FData - mt)*dmt;
                prExpModel.dSSdTau[k] := prExpModel.dSSdTau[k] + cGR;
              end;//k

              //Amp gradient
              for k := 0 to prExpModel.StateCount - 2 do begin
                prExpModel.Do_dMean_t_dAmp(dmt, t*dt, k);
                FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
                cGR := -2*(FData - mt)*dmt;
                prExpModel.dSSdAmp[k] := prExpModel.dSSdAmp[k] + cGR;
              end;//k


            //METHOD_LKC @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
            end else if prMethod = METHOD_LKC then begin


              prLKCModel.V  := TQUB_Exp_PChunk(prChunks[c]).V;
              prLKCModel.V0 := TQUB_Exp_PChunk(prChunks[c]).V0;

              //I Mean
              prFitModel.Do_Mean_t(mt, t*dt);

              //Sum-Of-Squares .......................................................
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;//Does it include ChannelScale?!
              cSS := Sqr(FData - mt);
              prFitModel.SS := prFitModel.SS + cSS;//Use minus sign in order to minimize SS

              //GL Gradient ........................................................
              prLKCModel.Do_dMean_t_dGL(dmt, t*dt);
              cGR := -2*(FData - mt)*dmt;
              prLKCModel.dSSdGL := prLKCModel.dSSdGL + cGR;

              //VL Gradient ........................................................
              prLKCModel.Do_dMean_t_dVL(dmt, t*dt);
              cGR := -2*(FData - mt)*dmt;
              prLKCModel.dSSdVL := prLKCModel.dSSdVL + cGR;

              //Tau gradient .......................................................
              for k := 0 to prLKCModel.N - 1 do begin
                prLKCModel.Do_dMean_t_dTau(dmt, t*dt, k);
                cGR := -2*(FData - mt)*dmt;
                prLKCModel.dSSdTau[k] := prLKCModel.dSSdTau[k] + cGR;
              end;//k

              //Amp gradient .......................................................
              for k := 0 to prLKCModel.N - 1 do begin
                prLKCModel.Do_dMean_t_dAmp(dmt, t*dt, k);
                cGR := -2*(FData - mt)*dmt;
                prLKCModel.dSSdAmp[k] := prLKCModel.dSSdAmp[k] + cGR;
              end;//k


            //METHOD_SIN @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
            end else if prMethod = METHOD_SIN then begin

              //I Mean
              prFitModel.Do_Mean_t(mt, t*dt);

              //Sum-Of-Squares .......................................................
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;//Does it include ChannelScale?!
              cSS := Sqr(FData - mt);
              prFitModel.SS := prFitModel.SS + cSS;//Use minus sign in order to minimize SS

              //Ieq Gradient .......................................................
              prSinModel.Do_dMean_t_dIeq(dmt, t*dt);
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdIeq := prSinModel.dSSdIeq + cGR;

              //Amp Gradient .......................................................
              prSinModel.Do_dMean_t_dAmp(dmt, t*dt);
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdAmp := prSinModel.dSSdAmp + cGR;

              //Freq Gradient ......................................................
              prSinModel.Do_dMean_t_dFreq(dmt, t*dt);
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdFreq := prSinModel.dSSdFreq + cGR;

              //Phase Gradient .....................................................
              prSinModel.Do_dMean_t_dPhase(dmt, t*dt);
              FData := TQUB_Exp_PChunk(prChunks[c]).Data[t]*Scale;
              cGR := -2*(FData - mt)*dmt;
              prSinModel.dSSdPhase := prSinModel.dSSdPhase + cGR;

            end;//METHOD


          end;//Fit this point!


          //Show fitting curve ?
          if (OptParams.OptShowFit <> 0) then begin
            FitChunk := TQUB_Exp_PChunk(prChunks[c]);

            //************************************************************
            if FitThisPoint then begin

              //METHOD_LKC @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
              if prMethod = METHOD_LKC then begin

                prLKCModel.V  := TQUB_Exp_PChunk(prChunks[c]).V;
                prLKCModel.V0 := TQUB_Exp_PChunk(prChunks[c]).V0;

              end;

              //I Mean
              prFitModel.Do_Mean_t(mt, t*dt);
              if          OptParams.OptShowFitWhat = SHOWFIT_AVG then begin
                FitChunk.FitCurve[t] := Round((mt-FitChunk.Offset) * FitChunk.Scaling);
              end else if OptParams.OptShowFitWhat = SHOWFIT_RES then begin
                FitChunk.FitCurve[t] := Round((mt-FitChunk.Offset) * FitChunk.Scaling - TQUB_Exp_PChunk(prChunks[c]).Data[t]);
              end;
            end else begin
              if          OptParams.OptShowFitWhat = SHOWFIT_AVG then begin
                FitChunk.FitCurve[t] := TQUB_Exp_PChunk(prChunks[c]).Data[t];
              end else if OptParams.OptShowFitWhat = SHOWFIT_RES then begin
                FitChunk.FitCurve[t] := 0;
              end;
            end;

          end;//ShowFit


        end;//of t


      finally

        UnlockAndNullifyFitDataList;

      end;


    end;//of Chunk
    //****************************************************************************


  end;//IdenticalChunks


  //Calculate free pars gradients
  prFitModel.Do_dSSdfPar;


  Synchronize(prDoOverlapData);


end;


procedure TQUB_Task_Exp.prDoOverlapData;
var
  FitIter         :TQUB_FitCurveIter;
  Chunk           :TQUB_Exp_PChunk;
  i               :Integer;
begin
  if OptParams.OptShowFit = 0 then Exit;

  if not prFitPreCleared then begin
    for i := 0 to prChunks.Count - 1 do begin
      TQUB_Exp_PChunk(prChunks[i]).QVD.FitCurves.Clear;//this may be repetitive. Is there any penalty? LM
    end;
  end;

  for i := 0 to prChunks.Count - 1 do begin
    Chunk         := TQUB_Exp_PChunk(prChunks[i]);
    FitIter       := nil;
    try
      FitIter       := Chunk.QVD.FitCurves.Find(Chunk.Chunk.Sel.First);
      if (0 <> FitIter.Done) or (FitIter.Curve.First <> Chunk.Chunk.Sel.First) or (FitIter.Curve.Last <> Chunk.Chunk.Sel.Last) then begin
        FitIter     := Chunk.QVD.FitCurves.Add(Chunk.Chunk.Sel.First, Chunk.Chunk.Sel.Last);
      end;
      Move(Chunk.FitCurve^, FitIter.Curve.Data^, (Chunk.Chunk.Sel.Last - Chunk.Chunk.Sel.First + 1) * SizeOf(Integer));
      FitIter.MarkChanged;
    finally
      FitIter.Free;
    end;
  end;

end;


//******************************************************************************
procedure TQUB_Task_Exp.prResetModel;
var
  i :Integer;
begin

  if OptParams.OptPropagateParams = 0 then begin
    //METHOD_EXP
    if prMethod = METHOD_EXP then begin
      prExpModel.Ieq := prOptParams.Ieq;
      for i := 0 to prOptParams.StateCount - 2 do begin
        prExpModel.Tau[i] := prOptParams.Tau[i];
        prExpModel.Amp[i] := prOptParams.Amp[i];
      end;
    //METHOD_LKC
    end else if prMethod = METHOD_LKC then begin
      prLKCModel.GL := prOptParams.LKCGL;
      prLKCModel.VL := prOptParams.LKCVL;
      for i := 0 to prOptParams.LKCN - 1 do begin
        prLKCModel.Tau[i] := prOptParams.LKCTau[i];
        prLKCModel.Amp[i] := prOptParams.LKCAmp[i];
      end;
    //METHOD_EXP
    end else if prMethod = METHOD_SIN then begin
      prSinModel.Ieq   := prOptParams.SineIeq;
      prSinModel.Amp   := prOptParams.SineAmp;
      prSinModel.Freq  := prOptParams.SineFreq;
      prSinModel.Phase := prOptParams.SinePhase;
    end;
  end;

  prFitModel.Do_InitcPars;
  prFitModel.Do_InitfPars;
end;

procedure TQUB_Task_Exp.prUpdateData;
begin
  //TODO :Update all data sets, for FileList. LM
  if Data <> nil then begin
    if OptParams.OptShowFit <> 0 then begin
      TQUB_QVO_QVD(Data).Zoom.Draw;
    end;
  end;
end;

procedure TQUB_Task_Exp.prShowData;
begin
  //TODO :Do it for FileList. LM
  if Data <> nil then begin
    //TO DO: Make it an option. LM
    if TQUB_QVO_QVD(Data).DrawListOverlapped then begin
      //TQUB_QVO_QVD(Data).MoveSelectionIntoWindow(QUB_FirstLast(0, TQUB_Thread_PChunk(prChunks[0]).Sel.Last - TQUB_Thread_PChunk(prChunks[0]).Sel.First), TQUB_Thread_PChunk(prChunks[0]).Sel.First);
    end else begin
      //TQUB_QVO_QVD(Data).MoveSelectionIntoWindow(QUB_FirstLast(TQUB_Thread_PChunk(prChunks[0]).Sel.First, TQUB_Thread_PChunk(prChunks[prChunks.Count - 1]).Sel.Last), TQUB_Thread_PChunk(prChunks[0]).Sel.First);
    end;
  end;
end;

procedure TQUB_Task_Exp.prRefreshZoom;
begin
  //TODO :Do it for FileList. LM
  if Data <> nil then begin
    TQUB_QVO_QVD(Data).Zoom.Draw;
  end;
end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Exp.prDoExecute;
begin
  prConsoleMsg  := 'Exp optimizer running ... ';
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
  inherited;
end;

procedure TQUB_Task_Exp.prDoExecuteStart;
begin

  inherited;

  //todo :Do it for FileList. LM
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

end;


procedure TQUB_Task_Exp.prDoExecuteEnd;
begin
  if prSelectionList <> nil then prSelectionList.UnLock;

  //Free Model
  prFitModel.Free;

  Synchronize(prRefreshZoom);

  inherited;
end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Exp.prChunkData :Boolean;
var
  DataSets         :INode;
  DS, Seg          :INode;

  QVD              :TQUB_QVO_QVD;
  QDF              :TQUB_QFS_QDF;

  FileIndex        :Integer;

  Chunk            :TQUB_Thread_PChunk;
  ExpChunk         :TQUB_Exp_PChunk;
  ChunkBounds      :TList;
  FirstChunk       :Integer;

  Sel              :TQUB_PFirstLast;
  Dat              :TQUB_FirstLast;

  c, l             :Integer;
begin

  Result           := False;

  prSetChunkParameters;


  try


    prDataSetCount := 0;

    DataSets       := DataNoder.GetChosenData( TNode.CreateNull );
    DS             := DataSets.Find('DataSet');
    FileIndex      := 0;


    //EXP
    if prMethod = METHOD_EXP then begin
      prExpModel   := TQUB_Exp_Model.Create(prOptParams);
      prFitModel   := prExpModel;
    //LKC
    end else if prMethod = METHOD_LKC then begin
      prLKCModel   := TQUB_LKC_Model.Create(prOptParams);
      prFitModel   := prLKCModel;
    //SIN
    end else if prMethod = METHOD_SIN then begin
      prSinModel   := TQUB_Sin_Model.Create(prOptParams);
      prFitModel   := prSinModel;
    end;


    while not DS.IsNull do begin

      Inc(prDataSetCount);

      FirstChunk   := prChunks.Count;
      Seg          := DS.Find('Segment');

      if not Seg.IsNull then begin

        QVD         := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[ DS['FileName'].DataAsString]);
        QDF         := QVD.DataObject;
        ChunkBounds := TList.Create;


        while not Seg.IsNull do begin

          Dat                      := QUB_FirstLast(Seg.Data[0], Seg.Data[1]);
          Chunk                    := prNewChunk(Dat);

          ExpChunk                 := TQUB_Exp_PChunk(Chunk);

          prChunks.Add(ExpChunk);

          Chunk.StartTime          := Double(Seg['start'].Data[0]) * 1.0e-3;

          ExpChunk.QVD             := QVD;
          ExpChunk.QDF             := QDF;

          ExpChunk.Sampling        := QDF.Sampling;
          ExpChunk.Scaling         := QDF.Scaling * QVD.DataChannelScaling[QVD.ActiveChannel];
          ExpChunk.Offset          := QVD.DataChannelOffset[QVD.ActiveChannel];

          if prMethod = METHOD_LKC then begin

            //For LKC
            if prOptParams.LKCVChannel < TQUB_QVO_QVD(Data).DataObject.ADChannelCount then begin
              //V
              ExpChunk.V               := ExpChunk.QVD.Average(prOptParams.LKCVChannel{StimCh}, Dat.First + prOptParams.LKCSkipPoints, Dat.Last - prOptParams.LKCSkipPoints);
              //V0
              ExpChunk.V0              := ExpChunk.QVD.Average(prOptParams.LKCVChannel{StimCh}, Dat.First - prOptParams.LKCV0Points + prOptParams.LKCSkipPoints, Dat.First - prOptParams.LKCSkipPoints);
            end else begin
              //V
              ExpChunk.V               := 0;
              //V0
              ExpChunk.V0              := 0;
            end;

          end;
          
          ExpChunk.Data            := nil;

          New(Sel);
          Sel.First                := Seg.Data[0]; // Dat.First + Skp.First;
          Sel.Last                 := Seg.Data[1]; // Dat.Last - Skp.Last;
          ChunkBounds.Add(Sel);

          ExpChunk.FitModel        := prFitModel;

          ExpChunk.FileIndex       := FileIndex;

          Seg                      := Seg.SiblingSameName;

        end;//while not Seg.IsNull


        for l := FirstChunk to prChunks.Count - 1 do begin
          //ExpChunk                   := TQUB_Exp_PChunk(prChunks[l]);
          Dispose(TQUB_PFirstLast(ChunkBounds[l - FirstChunk]));
        end;

        ChunkBounds.Free;

        Inc(FileIndex);

      end;

      DS           := DS.SiblingSameName;

    end;


    //**************************************************************************


    for c := 0 to prChunks.Count - 1 do begin

      Chunk               := prChunks[c];
      Chunk.Status        := CHUNK_STATUS_EMPTY;
      Chunk.Trials        := 0;
      Chunk.Index         := c;

    end;

    if prChunks.Count = 0 then begin

      Result     := True;
      prError    := True;

      if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
        prErrorMsg := 'The file list is empty';
      end else begin
        prErrorMsg := 'No Data';
      end;
      prWikiTopic := 'Error:No_Data';

      Exit;

    end;


  finally

  end;

end;

procedure TQUB_Task_Exp.prSetChunkParameters;
begin
  inherited;
  prChunkSize := 0;
  prChunkSegmentation := True;
end;

procedure TQUB_Task_Exp.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  ExpChunk :TQUB_Exp_PChunk;
begin

 ExpChunk := TQUB_Exp_PChunk(Chunk);

  if ExpChunk.Data <> nil then begin
    FreeMem(ExpChunk.Data);
    ExpChunk.Data := nil;
  end;

  FreeMem(ExpChunk.FitCurve);
  ExpChunk.FitCurve := nil;

  FreeMem(ExpChunk);

end;

function TQUB_Task_Exp.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result           := AllocMem(SizeOf(TQUB_Exp_Chunk));

  Result.Sel       := Sel;

  TQUB_Exp_PChunk(Result).FitCurve := IAlloc1(Sel.Last - Sel.First + 1);

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Exp.OnGetProperties(Sender :TObject);
var
  k :Integer;
begin
  with Sender as TQUB_Task_Exp_PropertiesForm do begin

    prPropForm                  := TQUB_Task_Exp_PropertiesForm(Sender);

    Self.ChannelIndex           := {prPropForm.}Properties['ChannelIndex'].Data[0];//ChannelIndex

    //Data Processed
    //Self.DataProcessed          := DataProcessed;
    //Self.ProcessFilter          := ProcessFilter;
    //Self.ProcessFilterFreq      := ProcessFilterFreq;
    Self.DataProcessed          := StringToDataProcessed(Properties['ProcessData'].DataAsString);
    Self.ProcessFilter          := 0<>Properties['ProcessData']['Filter'].Data[0];
    Self.ProcessFilterFreq      := Properties['ProcessData']['FilterFreq'].Data[0];

    //Client/Server
    Self.ThreadCount            := ThreadCount;
    Self.Remote                 := Remote;

    Self.QuietOutput            := QuietOutput;

    Self.UpdateParams           := UpdateParams;

    Self.OptParams              := Params;
    //To force updating params
    //Exp
    Self.prOptParams.Ieq        := Ieq;
    for k := 0 to StateCount - 2 do begin
      Self.prOptParams.Tau[k]   := Tau[k];
      Self.prOptParams.Amp[k]   := Amp[k];
    end;
    //LKC
    Self.prOptParams.LKCGL      := LKCGL;
    Self.prOptParams.LKCGL      := LKCGL;
    for k := 0 to LKCN - 1 do begin
      Self.prOptParams.LKCTau[k]:= LKCTau[k];
      Self.prOptParams.LKCAmp[k]:= LKCAmp[k];
    end;
    //Sin
    Self.prOptParams.SineIeq    := SineIeq;
    Self.prOptParams.SineAmp    := SineAmp;
    Self.prOptParams.SineFreq   := SineFreq;
    Self.prOptParams.SinePhase  := SinePhase;

    Self.FitDataListName        := lstFitData.Items.Strings[FitDataListIdx];
    prGroupList                 := txtOptSegmentsGroupList.Text;

    prMethod                    := pageMethod.ActivePage.Tag;

  end;//with
end;

procedure TQUB_Task_Exp.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_Exp_PropertiesForm do begin
  //  ChannelIndex                := Self.ChannelIndex;
  //end;//with
end;


function TQUB_Task_Exp.DataCallBack(ExpChunk :TQUB_Exp_PChunk) :Integer;
var
  BaseChunk :TQUB_Thread_PChunk;
begin
  Result := 0;

  try

    BaseChunk := TQUB_Thread_PChunk(ExpChunk);

    if ExpChunk.Data = nil then begin
    //Allocate

      ExpChunk.Data := AllocMem(SizeOf(Integer) * (BaseChunk.Sel.Last - BaseChunk.Sel.First + 1));

      //todo. LM
      //Chris, this is bad!!! The DataSource must be local and frozen during the task execution.

      if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin

        //case DataProcessed of
          //DATA_PROCESSED_NONE:
            ExpChunk.QDF.ReadData(ExpChunk.Data, False{Raw}, ChannelIndex, BaseChunk.Sel.First, BaseChunk.Sel.Last);
        //end;

      end else begin

        case DataProcessed of
          DATA_PROCESSED_NONE:
            with ExpChunk.QVD do LoadDataChannel       (ExpChunk.Data, ChannelIndex, BaseChunk.Sel, False{Process}, False, False);
          DATA_PROCESSED_ASDISPLAYED:
            //use QVD's properties
            with ExpChunk.QVD do LoadDataChannelProcess(ExpChunk.Data, ChannelIndex, BaseChunk.Sel, FilterOn[ChannelIndex], FilterFreq[ChannelIndex], BaselineOn[ChannelIndex]);
          DATA_PROCESSED_ASSUCH:
            with ExpChunk.QVD do LoadDataChannelProcess(ExpChunk.Data, ChannelIndex, BaseChunk.Sel, ProcessFilter,          ProcessFilterFreq,        ProcessBaseline);
        end;

      end;

    end else begin
    //DeAllocate

      FreeMem(ExpChunk.Data);
      ExpChunk.Data := nil;

    end;

  except

    Result := 1;

  end;

end;


//******************************************************************************
end.
