{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// TODO:
//  list of frequencies? not now -- will make the curve-fit stage tougher
//  crashing?

unit QUB_Task_Kalman;


interface

uses

Classes, Controls, Dialogs, Forms, SysUtils, Math,

Alloc, AllocTypes,
maxill_iface,
QUB_Global,
QUB_QVO_QVD,
QUB_QVO_QVD_Global,
QUB_Task,
QUB_Node;


type

TQUB_Task_Kalman = class(TQUB_Task)
protected
  prData        :TQUB_QVO_QVD;
  prSelectionList :TQUB_QVO_QVD_SelectionList;
  prProperties  :INode;
  prSampling    :Double;
  prOutData     :INode;
  prOutName     :String;

  prPreviewD    :TPADouble;
  prPreviewT    :TPADouble;
  prPreviewF    :TPADouble;
  prPreviewN    :Integer;

  procedure prSetData(Data :TQUB_QVO_QVD);
  procedure prSetupOutData;
  procedure prOpenOutData;
  procedure prDoFilter;
  procedure prInitPeriodic(Buf :TPADouble; Count :Integer; var Freq :Double; F :TQUB_Kalman_Filter);

public

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Execute; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  property Data          :TQUB_QVO_QVD          read prData         write prSetData;
  property Properties    :INode                 read prProperties   write prProperties;
  property Sampling      :Double                read prSampling     write prSampling;
  property OutName       :String                read prOutName      write prOutName;
  property PreviewD      :TPADouble             read prPreviewD     write prPreviewD;
  property PreviewT      :TPADouble             read prPreviewT     write prPreviewT;
  property PreviewF      :TPADouble             read prPreviewF     write prPreviewF;
  property PreviewN      :Integer               read prPreviewN     write prPreviewN;

end;


implementation


uses

IEEE754,
QUBFits,
QUB_GUIInterface,
QUB_Task_Kalman_Properties,
QUB_QFS_QDF;


constructor TQUB_Task_Kalman.Create(Suspended :Boolean);
begin
  inherited;

  prName     := 'Kalman filter';

  prData     := nil;
  prSelectionList := nil;
  prOutData  := TNode.CreateNull;
  prOutName  := '';
  prPreviewD := nil;
  prPreviewT := nil;
  prPreviewF := nil;
  prPreviewN := 0;

end;

destructor TQUB_Task_Kalman.Destroy;
begin
  try


  finally
    inherited;
  end;
end;

procedure TQUB_Task_Kalman.prSetData(Data :TQUB_QVO_QVD);
begin
  prData := Data;
  prSampling := Data.DataObject.Sampling;

  Data.LockWrite;
  prSelectionList    := nil;
  if Data.SelectionLists.Items.Count > 0 then begin
    if Data.SelectionLists.ItemIndex <> -1 then begin
      prSelectionList := Data.SelectionLists[Data.SelectionLists.ItemIndex];
      prSelectionList.Lock;
    end;
  end;
end;

procedure TQUB_Task_Kalman.Execute;
begin

  try

    if Terminated then Exit;

    prStatus := 'Initializing';
    prPercent := 0;
    prETA := '0:00:00';

    Synchronize(prDoOnChange);

    try

      if prData <> nil then begin
        Synchronize(prSetupOutData);
        if prOutData.IsNull then begin
          raise Exception.Create('Can''t save '+prOutName);
        end;
      end;

      prDoFilter;

      if prData <> nil then begin
        prOutData.Save;
        prOutData := nil;
        Synchronize(prOpenOutData);
      end;

    except
    end;

    prStatus := 'Terminated';
    prPercent := 0;
    prETA := '0:00:00';

    Synchronize(prDoOnChange);

  finally

    if Data <> nil then Data.UnLockWrite;
    if prSelectionList    <> nil then prSelectionList.UnLock;

  end;
end;

procedure TQUB_Task_Kalman.prSetupOutData;
begin
  if FileExists(prOutName) then begin
    if mrNo = MessageDlg('File ' + prOutName + ' already exists. Do you want to ovewrite it ?', mtConfirmation, [mbYes, mbNo], 0) then begin
      Exit;
    end;
  end;
  if FileExists(ChangeFileExt(prOutName, '.qsf')) then DeleteFile(ChangeFileExt(prOutName, '.qsf'));
  prOutData := TNode.Create('DataFile');
  prOutData['Sampling'].SetData( prData.DataObject.Sampling );
  prOutData['Scaling'].SetData( prData.DataObject.Scaling );
  prOutData['ADDataSize'].SetData( prData.DataObject.ADDataSize );
  prOutData['ADDataType'].SetData( prData.DataObject.ADDataType );
  prOutData['ADChannelCount'].SetData( prData.DataObject.ADChannelCount );
  prOutData.AppendChild(prData.CloneSession);
  // remove stuff that would need to be edited for DataSource: sel or list
  prOutData['SessionFile'].RemoveChild(prOutData['SessionFile']['Lists']);
  prOutData['SessionFile'].RemoveChild(prOutData['SessionFile']['Idealization']);
  prOutData['SessionFile'].RemoveChild(prOutData['SessionFile']['Results']);
  prOutData['SessionFile'].RemoveChild(prOutData['SessionFile']['Baseline']);
  prOutData['SessionFile'].RemoveChild(prOutData['SessionFile']['Markers']);
  prOutData['SessionFile'].RemoveChild(prOutData['SessionFile']['TimeMarkers']);
  if not prOutData.SaveAs(prOutName) then prOutData := TNode.CreateNull;
end;

procedure TQUB_Task_Kalman.prOpenOutData;
begin
  ViewportData.Open(prOutName, False{Silent});
end;

const
  BUFFER_CAP       = 16384;

procedure TQUB_Task_Kalman.prDoFilter;
var
  StTime           :Comp;
  F                :TQUB_Kalman_Filter;
  RemoveBaseline   :Boolean;
  RemovePeriodic   :Boolean;
  Std              :Double;
  Jump, JumpOff    :Double;
  LastZ            :Double;
  Freq             :Double;
  i, c             :Integer;
  z, s, o          :Double;
  Buf              :TPADouble;
  IBuf             :TPAInteger;
  Sel, Rows        :TQUB_FirstLast;
  DS, Seg, Ch      :INode;
  SegOut           :INode;
  BufIx            :Integer;
  BufCount         :Integer;
begin
  StTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  F                := nil;
  Buf              := nil;
  try
    RemoveBaseline := 0<>prProperties['Baseline'].Data[0];
    RemovePeriodic := 0<>prProperties['Periodic'].Data[0];
    Jump           := prProperties['Jump'].Data[0];
    if RemovePeriodic then begin
      F            := TQUB_Kalman_Filter.Create(3, 1);
      Freq         := prProperties['PeriodicFreq'].Data[0];
      F.Rec.H[0,0] := 1.0; // baseline
      F.Rec.H[0,1] := 1.0; // real
      F.Rec.H[0,2] := 0.0; // imag
    end else begin
      F            := TQUB_Kalman_Filter.Create(1, 1);
      F.Rec.H[0,0] := 1.0;
    end;
    Std          := prProperties['BaselineStd'].Data[0];
    F.Rec.Q[0,0] := Sqrt(Std);
    Std            := prProperties['MeasStd'].Data[0];
    F.Rec.R[0,0]   := Sqrt(Std);

    if prPreviewN > 0 then begin
      if RemovePeriodic then begin
        prInitPeriodic(prPreviewD, prPreviewN, Freq, F);
      end else begin
        F.Rec.X[0] := prPreviewD[0];
      end;
      prPreviewF[0] := prPreviewD[0];
      LastZ         := prPreviewD[0];
      JumpOff       := 0.0;
      F.Reset;

      for i := 1 to prPreviewN - 1 do begin
        z          := prPreviewD[i];
        if Abs(z - LastZ) >= Jump then begin
          JumpOff  := JumpOff + (z - LastZ);
        end;
        LastZ      := z;
        z          := z - JumpOff;
        F.Next(@z);
        prPreviewT[i] := F.Rec.Z[0] + JumpOff;

        z          := prPreviewD[i];
        if RemoveBaseline then z := z - F.Rec.X[0];
        if RemovePeriodic then z := z - F.Rec.X[1];
        prPreviewF[i] := z;
        
        if prStopped then Break;
      end;
    end else begin
      c      := Data.ActiveChannel;
      s      := Data.DataObject.Scaling * Data.DataChannelScaling[c];
      o      := Data.DataChannelOffset[c];
      Buf    := DAlloc1(BUFFER_CAP);
      IBuf   := TPAInteger(@(Buf[BUFFER_CAP div 2]));
      DS     := Data.GetChosenData;
      Seg    := DS.Find('Segment');
      BufCount := 0;
      while not Seg.IsNull do begin
        Inc(BufCount, Ceil((Seg.Data[1]-Seg.Data[0] + 1) / BUFFER_CAP));
        Seg  := Seg.SiblingSameName;
      end;
      Seg    := DS.Find('Segment');
      BufIx  := 0;
      SegOut := TNode.CreateNull;
      while not (prStopped or Seg.IsNull) do begin
        Sel.First := Seg.Data[0];
        Sel.Last  := Seg.Data[1];
        SegOut := prOutData['Segments'].InsertChild(SegOut, 'Segment');
        SegOut['StartTime'].SetData( Seg['start'].Data[0] );
        Ch   := SegOut['Channels'];
        Ch.Preload := False;
        Ch.SetupNumData(Data.DataObject.ADDataType, Sel.Last - Sel.First + 1, Data.DataObject.ADChannelCount);
        Sel.Last := Min(Sel.Last, Sel.First + BUFFER_CAP - 1);
        Rows.First := 0;
        Rows.Last  := Rows.First + (Sel.Last - Sel.First);
        repeat

          Ch.LoadRows(Rows, False);
          prData.DataObject.ReadData(Ch.DataPtr, True, -1, Sel.First, Sel.Last);
          prData.LoadDataChannel(IBuf, c, Sel, True, True, True, True);
          for i := 0 to (Sel.Last - Sel.First + 1) - 1 do begin
            Buf[i] := IBuf[i] / s + o;
          end;
          if Sel.First = Seg.Data[0] then begin
            if RemovePeriodic then begin
              prInitPeriodic(Buf, Sel.Last-Sel.First+1, Freq, F);
            end else begin
              F.Rec.X[0] := Buf[0];
            end;
            LastZ        := Buf[0];
            JumpOff      := 0.0;
            F.Reset;
          end else begin
            z         := Buf[0];
            if Abs(z - LastZ) >= Jump then begin
              JumpOff := JumpOff + (z - LastZ);
            end;
            LastZ     := Buf[0];
            z         := z - JumpOff;
            F.Next(@z);
          end;
          if RemoveBaseline then Buf[0] := Buf[0] - F.Rec.X[0];
          if RemovePeriodic then Buf[0] := Buf[0] - F.Rec.X[1];
          for i := 1 to (Sel.Last - Sel.First + 1) - 1 do begin
            z      := Buf[i];
            if Abs(z - LastZ) >= Jump then begin
              JumpOff := JumpOff + (z - LastZ);
            end;
            LastZ  := z;
            z      := z - JumpOff;
            F.Next(@z);
            if RemoveBaseline then Buf[i] := Buf[i] - F.Rec.X[0];
            if RemovePeriodic then Buf[i] := Buf[i] - F.Rec.X[1];
          end;
          for i := 0 to (Sel.Last - Sel.First + 1) - 1 do begin
            Ch.MatrixData[Rows.First+i, c] := Integer(Round( s * (Buf[i] - o) ));
          end;
          Ch.UnloadRows(True);
          Sel.First  := Sel.Last + 1;
          Sel.Last   := Min(Sel.First + BUFFER_CAP - 1, Seg.Data[1]);
          Rows.First := Rows.Last + 1;
          Rows.Last  := Rows.First + (Sel.Last - Sel.First);

          Inc(BufIx);
          prUpdateStatus((100 * BufIx) div BufCount, StTime);

        until prStopped or (Sel.First > Seg.Data[1]);

        Seg  := Seg.SiblingSameName;
      end;
    end;

  finally
    F.Free;
    FreeMem(Buf);
  end;
end;

const
  FIT_PERIODS = 5;

procedure TQUB_Task_Kalman.prInitPeriodic(Buf :TPADouble; Count :Integer; var Freq :Double; F :TQUB_Kalman_Filter);
var
  XX, WW :Array of Double;
  Data   :TQF_Data;
  PData  :TPQF_Data;
  lo, hi :Double;
  Curve  :TQUB_Curve;
  Fitter :TQUB_Fitter;
  Amp, Phase :Double;
  DP, Std    :Double;
  i      :Integer;
begin
  Curve  := nil;
  Fitter := nil;
  Freq   := prProperties['PeriodicFreq'].Data[0];
  SetLength(XX, Count);
  SetLength(WW, Count);
  lo     := Buf[0];
  hi     := lo;
  for i := 0 to Count - 1 do begin
    XX[i] := i * prSampling;
    WW[i] := 1.0;
    if Buf[i] < lo then lo := Buf[i];
    if Buf[i] > hi then hi := Buf[i];
  end;
  Data.N       := Min(Count, Round(FIT_PERIODS / (Sampling * Freq)));
  Data.XX      := @(XX[0]);
  Data.YY      := @(Buf[0]);
  Data.WW      := @(WW[0]);
  Data.NV      := 0;
  PData        := @Data;
  try
  // fit Buf, i*Sampling to  'Amp * Cos( (2*pi*Freq*x) + Phase ) + Baseline';
    Curve      := CreateCurve('Custom');
    Curve.SetEqn('Amp * cos( (2*pi*Freq*x) + Phase ) + Baseline');
    Curve.SetParamVal(0, (hi - lo)/2, 0); // amp
    Curve.SetParamConst(0, 0);
    Curve.SetParamVal(1, Freq, 0);
    Curve.SetParamConst(1, 1);  // fix freq at first
    Curve.SetParamVal(2, 0.0, 0); // phase (?)
    Curve.SetParamConst(2, 0);
    Curve.SetParamVal(3, (hi + lo)/2, 0); // baseline
    Curve.SetParamConst(3, 0);
    Fitter     := CreateFitter('Levenburg-Marquardt');
    Fitter.Fit(Curve, @PData, 1);
    if 0<>prProperties['PeriodicFit'].Data[0] then begin
      Curve.SetParamConst(1, 0); // allow freq to change
      Fitter.Fit(Curve, @PData, 1);
    end;
    Amp     := Curve.GetParamVal(0);
    if IsNAN(Amp) or IsInfinity(Amp) or (Amp = 0.0) then begin
      Amp   := prProperties['BaselineStd'].Data[0];
    end else if Amp < 0 then begin
      Amp   := - Amp;
    end;
    Freq    := Curve.GetParamVal(1);
    if IsNAN(Freq) or IsInfinity(Freq) or (Freq = 0.0) then begin
      Freq  := prProperties['PeriodicFreq'].Data[0];
    end else if Freq < 0 then begin
      Freq  := - Freq;
    end;
    Phase   := Curve.GetParamVal(2);
    F.Rec.X[0]    := Curve.GetParamVal(3);
    F.Rec.X[1]    := Amp * Cos(Phase);
    F.Rec.X[2]    := Amp * Sin(Phase);

    DP      := 2 * PI * Sampling * Freq;
    F.Rec.A[1,2] := -DP;
    F.Rec.A[2,1] := DP;
    Std     := prProperties['PeriodicStd'].Data[0];
    F.Rec.P[1,1] := Sqrt(Std*Amp);
    F.Rec.Q[1,1] := Sqrt(Std*Amp);

  finally
    Curve.Free;
    Fitter.Free;
  end;
end;

procedure TQUB_Task_Kalman.OnGetProperties(Sender :TObject);
var
  Form    :TQUB_Task_Kalman_PropertiesForm;
begin
  Form    := TQUB_Task_Kalman_PropertiesForm(Sender);
  prProperties := Form.Properties;
  prOutName    := Form.OutName;
end;

procedure TQUB_Task_Kalman.OnPutProperties(Sender :TObject);
begin
end;


end.
