unit QUB_Task_MSL;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses

Classes, SysUtils, Math,
Optim, Matrices, FMath,
Alloc, AllocTypes,
QUB_Global, QUB_Globals, QUB_Node,
Maxill_Utils, Maxill_iface,
QUB_Thread, QUB_Task, QUB_Task_Data,
QUB_Viewport_Data, QUB_QVO_QVD, QUB_QVO_QVD_Global, QUB_QFS_QDF, QUB_Idealization,
QUB_Viewport_Model, QUB_QVO_QVM, QUB_QFS_QMF, QUB_QFS_Channels,
QUB_Viewport_Results, QUB_QVO_QVR,
QUB_Task_MSL_Properties;


// TODO:
//   text outputs



type

TQUB_Task_MSL = class(TQUB_Task_Data)
protected
  prModel               :TQUB_QVO_QVM;
  prSelectionList       :TQUB_QVO_QVD_SelectionList;
  prProperties          :INode;
  prMaxIter             :Integer;
  prShowIterations      :Boolean;
  prDataDescription     :INode;
  prDataSets            :array of INode;
  prResults             :INode;
  prMSL                 :TMaxill_Subi_Run;
  prIteration           :Integer;
  prLL                  :Double;

  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize: Boolean; override;

  procedure prPublishResults;

  function prChunkData: Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast): TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetModel(Model :TQUB_QVO_QVM);
  procedure prUpdateModel;

public

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Stop; override;

  function MsgOut(Msg :String) :Integer;
  function Eval(X : PVector) :Double;
  function Check(X : PVector; Final :Boolean) :Integer;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  property Model         :TQUB_QVO_QVM           read prModel      write prSetModel;
  property Properties    :INode                  read prProperties write prProperties;

end;



implementation

uses

QUB_GUIInterface, QUB_Main,
QUB_Node_Model,
QUB_Python_Data;


function Task_MSL_MsgOut(Msg :PChar; Data :Pointer): Integer; stdcall;
begin
  Result := TQUB_Task_MSL(Data).MsgOut(String(Msg));
end;

function Task_MSL_EvalSimplex(Caller :TObject; X : PVector) :Double; stdcall;
begin
  Result := TQUB_Task_MSL(Caller).Eval(X);
end;

function Task_MSL_CheckSimplex(Caller :TObject; X : PVector; F :Double) :Integer; stdcall;
begin
  Result := TQUB_Task_MSL(Caller).Check(x, False);
end;

constructor TQUB_Task_MSL.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'MSL';
  prProperties := TNode.Create('Properties');
  prDataDescription := TNode.CreateNull;
  prResults    := TNode.Create('Results');
  prMSL        := nil;
end;

destructor TQUB_Task_MSL.Destroy;
begin
  inherited;
end;

procedure TQUB_Task_MSL.Stop;
begin
  inherited;
end;

procedure TQUB_Task_MSL.prDoTask;
var
  ExpCond         :INode;
  MslModel        :TMaxill_Model;
  MslData         :TMaxill_Subi_Data;
  di              :Integer;
  i               :Integer;
  Z               :Array of Float;
  Zv              :PVector; // alias
  Fmin            :Float;
  DS, Seg, Segout :INode;
begin
  MslModel        := nil;
  prLL            := 0.0;
  prIteration     := 0;

  try
    MslModel           := TMaxill_Model.Create(Model.GetAsNode);
    prMSL              := TMaxill_Subi_Run.Create(MslModel);
    prMSL.MsgOut       := Task_MSL_MsgOut;
    prMSL.MsgOutData   := Self;
    prMSL.SearchLimit  := prProperties['SearchLimit'].Data[0];
    prMSL.UsePeq       := 0<>prProperties['UsePeq'].Data[0];
    prMSL.StopFlag     := @prStopped;

    for di := 0 to Length(prDataSets) - 1 do begin
      DS := prResults.AppendChild('DataSet');
      DS.AppendClone(prDataSets[di]['FileName'], True);
      Seg          := prDataSets[di].Find('Signal').Find('Idealization').Find('Segment');
      Segout       := TNode.CreateNull;
      while not Seg.IsNull do begin
        Segout     := DS.InsertClone(Segout, Seg, False);
        Segout.AppendClone(Seg['start'], False);
        Seg        := Seg.SiblingSameName;
      end;

      ExpCond      := prDataSets[di]['ExpCond'];
      MslData      := TMaxill_Subi_Data.Create(prDataSets[di]);
      MslData.Multiplex(MslModel);
      prMSL.AddData(MslData);
    end;

    if prMSL.Setup then begin

      // more guidance about constraints?
      // (maxill's callback has Reported its problems)

    end else begin

      SetLength(Z, prMSL.Np);
      Zv             := PVector(@Z[0]);
      for i := 0 to prMSL.Np - 1 do begin
        Z[i]         := prMSL.S[i];
      end;

      if prProperties['Mode'].DataAsString <> 'evaluate' then begin

        prMaxIter := prProperties['Simplex']['MaxIterations'].Data[0];
        if (prMaxIter > 0) and not prStopped then begin
          {foo := } SimplexObject(Self{Caller}, prStopped, prMSL.Np, Zv,
                                  prMaxIter,
                                  prProperties['Simplex']['Tolerance'].Data[0],
                                  prProperties['Simplex']['Step'].Data[0],
                                  prIteration, Fmin, @Task_MSL_EvalSimplex, @Task_MSL_CheckSimplex);
        end;
      end;

      if not prStopped then begin
        prLL      := - Eval(Zv);
        Check(Zv, True);
        prResults['LL'].SetData(prLL);
        Synchronize(prPublishResults);
      end;

    end;
  except on ex :Exception do
    begin
      prError := True;
      prErrorMsg := 'MSL: '+ex.Message;
      prWikiTopic := 'Error:Task';
    end;
  end;

  if MslModel <> nil then MslModel.Free;
  if prMSL <> nil then prMSL.Free;
end;

function TQUB_Task_MSL.Eval(X : PVector) :Double;
var
  i   :Integer;
begin
  for i := 0 to prMSL.Np - 1 do begin
    prMSL.S[i] := X[i];
  end;

  prLL   := prMSL.LL(prMSL.S);
  Result := - prLL;
end;

function TQUB_Task_MSL.Check(X : PVector; Final :Boolean) :Integer;
var
  i, j   :Integer;
  DSout  :INode;
  Segout :INode;
  LL     :Double;
  SegLL  :Double;
  Nd, Ns :Integer;
  SegNd  :Integer;
  SegNs  :Integer;
begin
  for i := 0 to prMSL.Np - 1 do begin
    prMSL.S[i] := X[i];
  end;

  prMSL.SetPars(prMSL.S);
  prMSL.Model.UpdateModelRates;
  Synchronize(prUpdateModel);

  prResults['Iterations'].SetData(prIteration);
  LL     := 0;
  Nd     := 0;
  Ns     := 0;

  i      := 0;
  DSout  := prResults.Find('DataSet');
  while not DSout.IsNull do begin
    j      := 0;
    Segout := DSout.Find('Segment');
    while not Segout.IsNull do begin
      SegLL   := prMSL.Datas[i].SegLL[j];
      SegNd   := prMSL.Datas[i].DwellCountses[j][0]; // original dwells
      SegNs   := prMSL.Datas[i].DwellCounts[j];      // sub-intervals
      Segout['LL'].SetData(SegLL);
      Segout['DwellCount'].SetData(SegNd);
      Segout['SubDwellCount'].SetData(SegNs);
      if SegNs > 0 then Segout['LL per subevent'].SetData(SegLL/SegNs);

      LL      := LL + SegLL;
      Nd      := Nd + SegNd;
      Ns      := Ns + SegNs;

      Inc(j);
      Segout := Segout.SiblingSameName;
    end;

    Inc(i);
    DSout := DSout.SiblingSameName;
  end;

  if Final then begin
    MsgOut(Format('MSL LL: %12.6g Iterations: %4d Sub-intervals: %d', [LL, prIteration, Ns]));
    // so that prIter/prMaxIter = 100%, and prIteration, if valid, stays valid:
    if prIteration <= 0 then prIteration := 1;
    if prMaxIter <= 0 then prMaxIter := prIteration;
  end else if prShowIterations then begin
    MsgOut(Format('MSL LL: %12.6g', [LL]));
  end;

  prUpdateStatus( (prIteration * 100) div prMaxIter, prStTime );

  Result := Integer(prStopped);
end;

function TQUB_Task_MSL.MsgOut(Msg :String): Integer;
begin
  Result := 0;
  prConsoleMsg := Msg;
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
end;

procedure TQUB_Task_MSL.prUpdateModel;
begin
  SetRatesFromNode( prModel.ModelObject, prMSL.Model.Model['Rates'] );
end;

procedure TQUB_Task_MSL.prPublishResults;
begin
  prResults.Name := 'MSL ' + ChangeFileExt(ExtractFileName(Data.FileName), '');
  prResults['ResultType'].SetData('MSL');
  if not prDataDescription.IsNull then begin
    prResults.AppendChild(prDataDescription);
  end;
  if prSelectionList <> nil then begin
    prSelectionList.Unlock;
    prSelectionList := nil;
  end;
  ViewportResults.AddResults( prResults, False );
end;

procedure TQUB_Task_MSL.prDoExecuteStart;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex >= 0 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  inherited;
end;

procedure TQUB_Task_MSL.prDoExecuteEnd;
begin
  inherited;

  if prSelectionList <> nil then begin
    prSelectionList.Unlock;
  end;

  if Model <> nil then begin
    Model.UnlockAny;
  end;
end;

function TQUB_Task_MSL.prDoExecuteInitialize: Boolean;
var
  Mdl       :INode;
begin
  Result    := False;

  Mdl       := Model.GetAsNode;
  Mdl.SetData('Initial');
  if Properties.Find('ModelFile').IsNull then Properties.AppendChild(Mdl);
end;

function TQUB_Task_MSL.prChunkData: Boolean;
var
  DataSets        :INode;
  DS              :INode;
begin

  if prScripting then begin
    DataSets         := DataNoderScript.GetChosenIdlSignals(TNode.CreateNull);
  end else begin
    DataSets         := DataNoder.GetChosenIdlSignals(TNode.CreateNull);
  end;
  prDataDescription := DataSets.Find('DataDescription');

  DS               := DataSets.Find('DataSet');
  while not DS.IsNull do begin
    SetLength( prDataSets, Length(prDataSets) + 1 );
    prDataSets[ Length(prDataSets) - 1 ] := DS;

    DS             := DS.SiblingSameName;
  end;

  Result               := False;
end;

procedure TQUB_Task_MSL.prFreeChunk(Chunk :TQUB_Thread_PChunk);
begin

  FreeMem(Chunk);

end;

function TQUB_Task_MSL.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Chunk));

  Result.Sel := Sel;

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MSL.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_MSL_PropertiesForm do begin

    prProperties                := Properties.Clone(True);
    prShowIterations            := 0<>prProperties['ShowIterations'].Data[0];
  end;//with
end;

procedure TQUB_Task_MSL.OnPutProperties(Sender :TObject);
begin
end;


//******************************************************************************
procedure TQUB_Task_MSL.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then prModel.LockAny;
end;




//******************************************************************************
end.
