unit QUB_Task_Protocol_To_Data;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

// mod 8-3-5 Chris:
//   set BaseChannel to generate into a sub-range of ADChannels
//   other channels will be filled with junk


interface


uses

Windows, SyncObjs, Math, Sysutils,

Alloc, AllocTypes,

QUB_Task,
QUB_QFS_QPF, QUB_QFS_QDF,
QUB_QVO, QUB_QVO_QVP, QUB_QVO_QVD,
QUB_Node;


type

TQUB_Task_Protocol_To_Data = class(TQUB_Task)
protected

  prProtocol                     :TQUB_QVO_QVP;
  prData                         :TQUB_QVO_QVD;
  prBaseChannel                  :Integer;

  procedure prSetProtocol(x :TQUB_QVO_QVP);
  procedure prSetData(x :TQUB_QVO_QVD);


public

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  procedure Execute; override;


  property Protocol        :TQUB_QVO_QVP        read prProtocol       write prSetProtocol;
  property Data            :TQUB_QVO_QVD        read prData           write prSetData;
  property BaseChannel     :Integer             read prBaseChannel    write prBaseChannel;

end;



implementation


uses

QUB_QFS_QDF_MPX;


constructor TQUB_Task_Protocol_To_Data.Create(Suspended :Boolean);
begin
  inherited;

  prName                    := 'Protocol -> Data';
  prBaseChannel             := 0;
end;

destructor TQUB_Task_Protocol_To_Data.Destroy;
begin
  try
    Protocol       := nil;
    Data           := nil;
  finally
    inherited;
  end;
end;

procedure TQUB_Task_Protocol_To_Data.Execute;
var
  iRep, iReps : Integer;      // Repetitions of stage to be eval'd
  aiMap : array of Integer;         // Randomly selected rep mapping
  iMap : Integer;             // Mapped repetition number

  st, sh, c, dc, i, s :Integer;
  LastVal             :Integer;
  Stage               :TQUB_QFS_QPF_Stage;
  Channel             :TQUB_QFS_QPF_Channel;
  Shape               :TQUB_QFS_QPF_Shape;
  ChanMin, ChanMax    :Double;
  OutOfRange          :array of Boolean;
  ProtoChanCount      :Integer;

  StageBufferMPX      :Pointer;
  StageBuffer         :TPMInteger;
  StageSamples        :Integer;
  NextSample          :Integer;
  ShapeSamples        :INode;
  ShapeSamplesPtr     :TPADouble;
  Scaling             :Double;
  Offset              :Double;
  StartTime           :Double;

  Tm                  :Double;

begin
  if Protocol = nil then begin
    prError           := True;
    prErrorMsg        := 'Protocol is unavailable or in use';
    prWikiTopic       := 'Error:Protocol_is_unavailable';
  end;

  if Data = nil then begin
    prError           := True;
    prErrorMsg        := 'No data chosen for output acquisition, or data in use';
    prWikiTopic       := 'Error:No_Data_For_Output';
  end;

  if (Data <> nil) and (Data.DataObject.ADChannelCount < (BaseChannel + Protocol.Protocol.ChannelCount)) then begin
    prWarning         := True;
    prWarningMsg      := 'Data file has fewer D/A channels than protocol';
    prWikiTopic       := 'Error:Data_file_has_fewer_channels';
  end;
  ProtoChanCount      := Min(Data.DataObject.ADChannelCount-BaseChannel, Protocol.Protocol.ChannelCount);

  StageBuffer         := nil;
  StartTime           := 0.0;
  StageBufferMPX      := nil;

  prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  prStatus := 'Running';
  Synchronize(prDoOnChange);

  try
    if not prError then begin
      SetLength(OutOfRange, ProtoChanCount);
      for c := 0 to ProtoChanCount - 1 do begin
        OutOfRange[c] := False;
      end;

      for st := 0 to Protocol.Protocol.StageCount - 1 do begin

        if prStopped then Exit;

        Stage := Protocol.Protocol.Stages[st];
        if Stage.RepeatMode = QUB_QFS_QPF_REPEAT_TIMES then
          iReps := Stage.RepeatTimes
        else
          iReps := 1;

        //-- If Randomize selected, build a mapping from 0..iReps to random
        SetLength( aiMap , iReps );
        if Stage.Randomize Then Begin
          for iRep := 0 to iReps-1 do
            aiMap[iRep] := -1;
          for iRep := 0 to iReps-1 do Begin
            Repeat                    // hunt randomly for an empty spot ( perhaps slow )
              iMap := Random( iReps );
            until aiMap[iMap]=-1;
            aiMap[iMap]:=iRep;
          end;
        end else begin
          for iRep := 0 to iReps-1 do
            aiMap[iRep] := iRep;
        end;

        for iRep := 0 to iReps-1 do begin
          iMap := aiMap[iRep];        // Get next randomly selected rep.

          if prStopped then Exit;

          prUpdateStatus(Round(100.0 * (st + iRep/iReps) / Protocol.Protocol.StageCount), prStTime);

          StageSamples   := Round(Stage.Duration(iMap) / Data.Sampling);
          if StageSamples = 0 then Continue;

          StageBuffer    := IAlloc2( Data.DataObject.ADChannelCount, StageSamples );
          StageBufferMPX := AllocMem(Data.DataObject.ADDataSize * Data.DataObject.ADChannelCount * StageSamples);

          for c := 0 to ProtoChanCount - 1 do begin
            dc           := BaseChannel + c; // "data" channel index
            Channel      := Stage.Channels[c];
            Scaling      := Data.Scaling * Data.DataChannelScaling[dc];
            Offset       := Data.DataChannelOffset[dc];
            ChanMin      := - Power(2, 8*Data.DataObject.ADDataSize-1) / Scaling + Offset;
            ChanMax      := (Power(2, 8*Data.DataObject.ADDataSize-1) - 1) / Scaling + Offset;

            NextSample   := 0;
            Tm           := 0.0;

            for sh := 0 to Channel.ShapeCount - 1 do begin
              if prStopped then Exit;

              Shape        := Channel.Shapes[sh];
              ShapeSamples := Shape.Generate(0, Shape.SampleCount(Tm, Data.Sampling, iMap) - 1,
                                             Tm, Data.Sampling, iMap);
              if not ShapeSamples.Find('Err').IsNull then begin
                prError           := True;
                prErrorMsg        := ShapeSamples['Err'].DataAsString;
                prWikiTopic       := 'Error:Protocol_to_data';
                Exit;
              end;
              ShapeSamplesPtr := TPADouble(ShapeSamples.DataPtr);
              Tm           := Tm + Shape.Duration(iMap);

              for s := 0 to ShapeSamples.DataCount - 1 do begin
                if (ShapeSamplesPtr[s] < ChanMin) or (ShapeSamplesPtr[s] > ChanMax) then OutOfRange[c] := True;
                StageBuffer[dc][s+NextSample] := Round( (ShapeSamplesPtr[s] - Offset) * Scaling );
              end;
              NextSample   := NextSample + LongInt(ShapeSamples.DataCount);
            end;

            LastVal := 0;
            if NextSample > 0 then LastVal := StageBuffer[dc][NextSample-1];
            for i := NextSample to StageSamples - 1 do begin
              StageBuffer[dc][i] := LastVal;
            end;
          end;

          MPXData(StageBuffer, StageBufferMPX, Data.DataObject.ADChannelCount, -1{all ch}, Data.DataObject.ADDataSize, Data.DataObject.ADDataType, 0, StageSamples-1, True{MPX});
          Data.AppendData(StageBufferMPX, Data.DataObject.ADChannelCount * Data.DataObject.ADDataSize, StageSamples, True{NewSegment}, StartTime, False{CanRefresh - Doit}, False{Acq});
          StartTime := StartTime + Data.Sampling * StageSamples;

          Free2(TPMChar(StageBuffer));
          StageBuffer := nil;
          FreeMem(StageBufferMPX);
          StageBufferMPX := nil;
        end;
      end;
    end;
  finally

    prUpdateStatus( 100, prStTime );

    if not prError then begin

      for c := 0 to Protocol.Protocol.ChannelCount - 1 do begin
        if OutOfRange[c] then begin
          prWarning    := True;
          prWarningMsg := prWarningMsg + Format('AO Channel %s went out of range.  Try adjusting its Scaling.'#13#10, [Data.DataChannelName[BaseChannel+c]]);
          prWikiTopic  := 'Error:Out_of_range';
        end;
      end;

      if prStopped then begin
        prStatus := 'Stopping';
      end else begin
        prStatus := 'Terminating';
      end;
      Synchronize(prDoOnChange);

      if prWarning then begin
        Synchronize(ShowMessage);
      end;
      
    end else if prError then begin

      prStatus := 'Error';
      Synchronize(prDoOnChange);

      Synchronize(ShowMessage);

    end;

    Free2(TPMChar(StageBuffer));
    FreeMem(StageBufferMPX);

    if Data <> nil then Data.AppendData(nil, 0, 0, False, 0, True, False{Acq});

  end;

end;

procedure TQUB_Task_Protocol_To_Data.prSetProtocol(x :TQUB_QVO_QVP);
begin
  if Protocol <> nil then Protocol.UnlockAny;
  if (x <> nil) and x.LockedAny then x := nil;
  prProtocol := x;
  if Protocol <> nil then Protocol.LockAny;
end;

procedure TQUB_Task_Protocol_To_Data.prSetData(x :TQUB_QVO_QVD);
begin
  if Data <> nil then Data.UnlockAny;
  if (x <> nil) and x.LockedAny then x := nil;
  prData := x;
  if Data <> nil then Data.LockAny;
end;

procedure TQUB_Task_Protocol_To_Data.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_Protocol_To_Data.OnPutProperties(Sender :TObject);
begin
end;


// *****************************************************************************
end.
 
