unit QUB_Task_Sim_DoseResponse;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


interface


uses

Windows, SyncObjs, Math, Sysutils, Classes,

Alloc, AllocTypes,

QUB_GUIInterface,
QUB_Task,
QUB_QFS_QDF, QUB_QFS_QDF_Global,
QUB_QVO, QUB_QVO_QVD, QUB_QVO_QVM,
QUB_Node;


const

MACROSCOPIC_THRESHOLD = 512;

type

TQUB_Task_Sim_DoseResponse = class(TQUB_Task)
protected

  prData                         :TQUB_QVO_QVD;
  prModel                        :TQUB_QVO_QVM;

  prStimulus                     :Array of Double;
  prStimulus2                    :Array of Double;
  prStimCount                    :Integer;
  prVarName                      :String;
  prVarName2                     :String;
  prProperties                   :INode;

  prSignal                       :TEvent;
  prSubtask                      :TQUB_Task;
  prScaling                      :Double;
  prStimData                     :TQUB_QVO_QVD;
  prStartTime                    :Double;
  prBufferMPX                    :Pointer;
  prIsNewSegment                 :Boolean;
  prSavedChannelCount            :Integer;
  prResults                      :INode;

  procedure prSetData(x :TQUB_QVO_QVD);

  procedure prNewStimData;
  procedure prAddStimSeg;
  procedure prDoneStimSegs;
  procedure prSimulate;
  procedure prAfterSimulate;
  procedure prPublishResults;

  procedure prOnSubtaskDone(Sender :TObject);

  procedure prFitPeaks(N :Integer; XX, YY, WW :TPADouble; var A, B, Tau :Double; var FitState :INode);
  function prReshapeResults: Boolean;

public

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  procedure Execute; override;
  procedure Stop; override;


  property Data            :TQUB_QVO_QVD        read prData           write prSetData;
  property Model           :TQUB_QVO_QVM        read prModel          write prModel;

end;


implementation


uses

QUB_Global,
QUB_QFS_QDF_MPX,
QUB_ExpCond_Parameters,
QUBFits,
QUB_Viewport_Results,
QUB_QVO_QVR,
QUB_Results_Select,
QUB_Task_Simulate,
QUB_Task_Sim_DoseResponse_Properties,
FileCtrl,
SystemPath;


constructor TQUB_Task_Sim_DoseResponse.Create(Suspended :Boolean);
begin
  inherited;

  prName                    := 'DoseResponse';
  prSignal                  := TEvent.Create(nil, False, False, '');
end;

destructor TQUB_Task_Sim_DoseResponse.Destroy;
begin
  try
    prSignal.Free;
    Data           := nil;
  finally
    inherited;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse.Execute;
var
  i, s, r              :Integer;
  Repeats              :Integer;
  StartRep             :Integer;
  Resting              :Double;
  Pulses               :Array of Double;
  StartRep2            :Integer;
  Resting2             :Double;
  Pulses2              :Array of Double;
  Node                 :INode;
  dy, hi, lo           :Double;
  Scaling, Scaling2    :Double;
  Offset               :Double;
  PulseHeight          :Double;

  StageBuffer          :TPAInteger;

  Sel                  :TQUB_FirstLast;
  FindPeak             :Boolean;
  PeakTime             :Double;
  Current, Peak        :Double;
  MaxPeak              :Double;
  DS, Seg              :INode;
  SegVar, SegVar2      :INode;
  SegCurrent, SegPeakTime :INode;
  SegNorm              :INode;
  Times                :INode;
  Weights              :Array of Double;
  FitA, FitB, FitTau   :Double;
  FitState             :INode;
  VarToShow            :String;
  Normalize            :Boolean;
  F                    :File;
  Text                 :String;
  Model                :TQUB_QVO_QVM;
  Sign                 :Double;
  Integral             :Double;
  SegIntegral          :INode;
begin
  StageBuffer         := nil;
  prStartTime         := 0.0;
  prBufferMPX         := nil;

  prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  prStatus := 'Running';
  Synchronize(prDoOnChange);

  if Terminated then Exit;

  try
    prScaling := 327.68;
    Synchronize(prNewStimData);
    if prError then Exit;

    Sign                           := 1.0; // to find positive peak
    Model                          := ViewportModel.CurrModel;
    if Model <> nil then begin
      if Model.ModelObject.Amp[1] < Model.ModelObject.Amp[0] then begin
        Sign                       := -1;
      end;
    end;

    Repeats                        := prProperties['Repeats'].Data[0];
    StartRep                       := prProperties['StartRep'].Data[0];

    // setup levels and scaling
    Resting                        := prProperties['RestingLevel'].Data[0];
    lo := Resting;   hi := Resting;
    Node                           := prProperties['PulseLevels'];
    SetLength(Pulses, Node.DataCount);
    for i := 0 to Node.DataCount - 1 do begin
      Pulses[i] := Node.Data[i];
      if Pulses[i] < lo then lo := Pulses[i];
      if Pulses[i] > hi then hi := Pulses[i];
    end;
    dy := Max(1e-8, 1.2 * Max(Abs(lo), Abs(hi)) );
    Scaling := 100.0 / dy;
    prStimData.DataChannelScaling[0] := Scaling;
    prStimData.DataChannelOffset[0]  := 0.0;
    Scaling := Scaling * prScaling;

    Resting2 := 0;
    Scaling2 := 0;
    if 0<>prProperties['Variable2'].Data[0] then begin
      StartRep2                      := prProperties['StartRep2'].Data[0];
      Resting2                       := prProperties['RestingLevel2'].Data[0];
      lo := Resting2;   hi := Resting2;
      Node                           := prProperties['PulseLevels2'];
      SetLength(Pulses2, Node.DataCount);
      for i := 0 to Node.DataCount - 1 do begin
        Pulses2[i] := Node.Data[i];
        if Pulses2[i] < lo then lo := Pulses2[i];
        if Pulses2[i] > hi then hi := Pulses2[i];
      end;
      dy := Max(1e-8, 1.2 * Max(Abs(lo), Abs(hi)) );
      Scaling2 := 100.0 / dy;
      prStimData.DataChannelScaling[1] := Scaling2;
      prStimData.DataChannelOffset[1]  := 0.0;
      Scaling2 := Scaling2 * prScaling;
    end;

    // write each pulse to prStimData
    StageBuffer    := AllocMem(prStimCount * SizeOf(Integer));
    prBufferMPX    := AllocMem(prStimData.DataObject.ADDataSize * prStimData.DataObject.ADChannelCount * prStimCount);

    for i := 0 to Max(Length(Pulses), Length(Pulses2)) - 1 do begin
      if prStopped then Exit;
      prUpdateStatus(Round(50.0 * (i / Max(Length(Pulses), Length(Pulses2)))), prStTime);

      for r := 0 to Repeats - 1 do begin
        if  ((r+1) >= StartRep) and (i < Length(Pulses)) then begin
          PulseHeight  := Pulses[i] - Resting;
          for s := 0 to prStimCount - 1 do begin
            StageBuffer[s] := Round( (prStimulus[s] * PulseHeight + Resting) * Scaling );
          end;
        end else begin
          for s := 0 to prStimCount - 1 do begin
            StageBuffer[s] := 0; // Round(Resting * Scaling);
          end;
        end;
        MPXData(StageBuffer, prBufferMPX, prStimData.DataObject.ADChannelCount, 0, prStimData.DataObject.ADDataSize, prStimData.DataObject.ADDataType, 0, prStimCount-1, True{MPX});

        if ((r+1) >= StartRep2) and (i < Length(Pulses2)) then begin
          PulseHeight  := Pulses2[i] - Resting2;
          for s := 0 to prStimCount - 1 do begin
            StageBuffer[s] := Round( (prStimulus2[s] * PulseHeight + Resting2) * Scaling2 );
          end;
          MPXData(StageBuffer, prBufferMPX, prStimData.DataObject.ADChannelCount, 1, prStimData.DataObject.ADDataSize, prStimData.DataObject.ADDataType, 0, prStimCount-1, True{MPX});
        end else if 0<>prProperties['Variable2'].Data[0] then begin
          for s := 0 to prStimCount - 1 do begin
            StageBuffer[s] := 0; // Round(Resting2 * Scaling2);
          end;
          MPXData(StageBuffer, prBufferMPX, prStimData.DataObject.ADChannelCount, 1, prStimData.DataObject.ADDataSize, prStimData.DataObject.ADDataType, 0, prStimCount-1, True{MPX});
        end;

        prIsNewSegment := r = 0;
        Synchronize(prAddStimSeg);
        prStartTime := prStartTime + prStimData.Sampling * prStimCount;
      end;
    end;

    FreeMem(StageBuffer);
    StageBuffer := nil;
    FreeMem(prBufferMPX);
    prBufferMPX := nil;

    Synchronize(prDoneStimSegs);
    if prError then Exit;

    Synchronize(prSimulate);
    if prError then Exit;

    while (not prStopped) and (not prError) and (prSubtask <> nil) do begin
      prSignal.WaitFor(1000);
    end;

    Synchronize(prAfterSimulate);
    if prError then Exit;

    // measure segments building result
    FindPeak       := 0 = prProperties['Measure'].Data[0];
    PeakTime       := prProperties['MeasureTime'].Data[0];
    StageBuffer    := AllocMem(prStimCount * SizeOf(Integer));
    Scaling        := Data.DataObject.Scaling * Data.DataChannelScaling[0];
    Offset         := Data.DataChannelOffset[0];

    prResults := TNode.Create('DoseResponse');
    prResults['ResultType'].SetData( 'DRsp' );
    DS := prResults.AppendChild( Data.CreateDataSet );
    Seg := TNode.CreateNull;
    MaxPeak := 0.0;
    for i := 0 to Max(Length(Pulses), Length(Pulses2)) - 1 do begin
      if prStopped then Exit;
      prUpdateStatus(Round(50.0 + 50.0 * (i / Max(Length(Pulses), Length(Pulses2)))), prStTime);

      Seg := DS.InsertChild(Seg, 'Segment');
      Seg.SetupNumData(NODE_DATATYPE_LONGINT, 2, 1);
      Sel.First   := i*prStimCount*Repeats;
      Sel.Last    := Sel.First + prStimCount*Repeats - 1;
      Seg.Data[0] := Sel.First;
      Seg.Data[1] := Sel.Last;
      Seg['start'].SetData(Data.DataObject.Sampling * 1e3 * i * prStimCount);
      if Repeats = 1 then begin
        if i < Length(Pulses) then begin
          Seg[prVarName].SetData(Pulses[i]);
        end else begin
          Seg[prVarName].SetData(0.0/1.0);
        end;
        if i < Length(Pulses2) then begin
          Seg[prVarName2].SetData(Pulses2[i]);
        end else if 0<>prProperties['Variable2'].Data[0] then begin
          Seg[prVarName2].SetData(0.0/1.0);
        end;

        Data.LoadDataChannel(StageBuffer, 0, Sel, Data.FilterOn[0], True, True, True);

        if FindPeak then begin
          Peak := -1e20;
          for s := 0 to prStimCount - 1 do begin
            Current    := Sign * (StageBuffer[s] / Scaling + Offset);
            if Current > Peak then begin
              Peak     := Current;
              PeakTime := s * Data.DataObject.Sampling * 1e3;
            end;
          end;
        end else begin
          Peak      := StageBuffer[Round(PeakTime/(Data.DataObject.Sampling * 1e3))] / Scaling + Offset;
        end;
        MaxPeak := Max(MaxPeak, Abs(Peak));
        Seg['Current'].SetData(Sign*Peak);
        Seg['PeakTime'].SetData(PeakTime);

        Integral := 0;
        for s := 0 to prStimCount - 1 do begin
          Integral := Integral + Sign * (StageBuffer[s] / Scaling + Offset);
        end;
        Seg['Current Integral'].SetData(Integral * Data.DataObject.Sampling);
      end else begin
        SegVar := Seg[prVarName];
        SegVar.SetupNumData(NODE_DATATYPE_DOUBLE, Repeats, 1);
        if (i < Length(Pulses2)) or (0<>prProperties['Variable2'].Data[0]) then begin
          SegVar2 := Seg[prVarName2];
          SegVar2.SetupNumData(NODE_DATATYPE_DOUBLE, Repeats, 1);
        end else begin
          SegVar2 := TNode.CreateNull;
        end;
        SegCurrent := Seg['Current'];
        SegCurrent.SetupNumData(NODE_DATATYPE_DOUBLE, Repeats, 1);
        SegPeakTime := Seg['PeakTime'];
        SegPeakTime.SetupNumData(NODE_DATATYPE_DOUBLE, Repeats, 1);
        SegIntegral := Seg['Current Integral'];
        SegIntegral.SetupNumData(NODE_DATATYPE_DOUBLE, Repeats, 1);

        for r := 0 to Repeats - 1 do begin
          Sel.First := Seg.Data[0] + r*prStimCount;
          Sel.Last  := Sel.First + prStimCount - 1;

          if ((r+1) >= StartRep) and (i < Length(Pulses)) then begin
            SegVar.Data[r] := Pulses[i];
          end else begin
            SegVar.Data[r] := 0.0;
          end;

          if ((r+1) >= StartRep2) and ((i < Length(Pulses2)) or (0<>prProperties['Variable2'].Data[0])) then begin
            SegVar2.Data[r] := Pulses2[i];
          end else if LongInt(SegVar2.DataCount) > r then begin
            SegVar2.Data[r] := 0.0;
          end;

          Data.LoadDataChannel(StageBuffer, 0, Sel, Data.FilterOn[0], True, True, True);

          if FindPeak then begin
            Peak := -1e20;
            for s := 0 to prStimCount - 1 do begin
              Current    := Sign * (StageBuffer[s] / Scaling + Offset);
              if Current > Peak then begin
                Peak     := Current;
                PeakTime := s * Data.DataObject.Sampling * 1e3;
              end;
            end;
          end else begin
            Peak      := StageBuffer[Round(PeakTime/(Data.DataObject.Sampling * 1e3))] / Scaling + Offset;
          end;
          MaxPeak := Max(MaxPeak, Abs(Peak));
          SegCurrent.Data[r] := Sign*Peak;
          SegPeakTime.Data[r] := PeakTime;

          Integral := 0;
          for s := 0 to prStimCount - 1 do begin
            Integral := Integral + Sign * (StageBuffer[s] / Scaling + Offset);
          end;
          SegIntegral.Data[r] := Integral * (prStimCount * Data.DataObject.Sampling);
        end;
      end;
    end;

    Normalize := (0<>prProperties['Normalize'].Data[0]) and (0.0 <> MaxPeak);
    if Normalize then begin
      Seg := DS.Find('Segment');
      while not Seg.IsNull do begin
        SegCurrent := Seg['Current'];
        SegNorm    := Seg['Normalized Current'];
        SegNorm.SetupNumData(NODE_DATATYPE_DOUBLE, Repeats, 1);
        for r := 0 to Repeats - 1 do begin
          SegNorm.Data[r] := Sign * SegCurrent.Data[r] / MaxPeak;
        end;
        Seg := Seg.SiblingSameName;
      end;
    end;

    // if repeats > 1 time to curve fit
    if Repeats > 1 then begin
      SetLength(Weights, Repeats);
      for r := 0 to Repeats - 1 do begin
        Weights[r] := 1.0;
      end;
      Seg := DS.Find('Segment');
      while not Seg.IsNull do begin
        Times := Seg['AbsPeakTime'];
        Times.SetupNumData(NODE_DATATYPE_DOUBLE, Repeats, 1);
        SegCurrent := Seg['Normalized Current'];
        if LongInt(SegCurrent.DataCount) < Repeats then begin
          SegCurrent := Seg['Current'];
        end;
        SegPeakTime := Seg['PeakTime'];
        for r := 0 to Repeats - 1 do begin
          Times.Data[r] := SegPeakTime.Data[r] + r * 1e3*prData.Sampling * prStimCount;
        end;

        prFitPeaks(Repeats, TPADouble(Times.DataPtr), TPADouble(SegCurrent.DataPtr), TPADouble(@Weights[0]),
                   FitA, FitB, FitTau, FitState);
        Seg['Initial Current'].SetData(FitA);
        Seg['Final Current'].SetData(FitB);
        Seg['Tau'].SetData(FitTau);
        Seg.AppendChild(FitState);

        Seg := Seg.SiblingSameName;
      end;
    end;

    if 0<>prProperties['SaveMeas'].Data[0] then begin
      try
        AssignFile(f, GetBaseStampedFileName(prProperties['MeasFileBase'].DataAsString, '.txt'));
        Rewrite(f, 1);

        if 0 <> prProperties['Variable2'].Data[0] then begin
          Text := prVarName+#9+prVarName2;
        end;
        if Repeats = 1 then begin
          Text := Text + #9'Current'#9;
        end else begin
          Text := Text + #9;
          for r := 0 to Repeats - 1 do begin
            Text := Text + Format('Current %d'#9, [r+1]);
          end;
        end;
        if Normalize then Text := Text + 'Normalized Current'#9;
        if Repeats = 1 then begin
          Text := Text + 'PeakTime'#13#10;
        end else begin
          Text := Text + 'Initial Current'#9'Final Current'#9'Tau'#13#10;
        end;
        BlockWrite(f, PChar(Text)^, Length(Text));

        Seg := DS.Find('Segment');
        while not Seg.IsNull do begin
          Text := Format('%.6g', [Double(Seg[prVarName].Data[0])]);
          if 0<>prProperties['Variable2'].Data[0] then begin
            Text := Text + Format(#9'%.6g', [Double(Seg[prVarName2].Data[0])]);
          end;
          for r := 0 to Repeats - 1 do begin
            Text := Text + Format(#9'%.6g', [Double(Seg['Current'].Data[r])]);
          end;
          if Normalize then Text := Text + Format(#9'%.6g', [Double(Seg['Normalized Current'].Data[0])]);
          if Repeats = 1 then begin
            Text := Text + Format(#9'%.6g'#13#10, [Double(Seg['PeakTime'].Data[0])]);
          end else begin
            Text := Text + Format(#9'%.6g'#9'%.6g'#9'%.6g'#13#10, [Double(Seg['Initial Current'].Data[0]), Double(Seg['Final Current'].Data[0]), Double(Seg['Tau'].Data[0])]);
          end;
          BlockWrite(f, PChar(Text)^, Length(Text));
          Seg := Seg.SiblingSameName;
        end;
      except
        on ex: Exception do begin
          prWarning := True;
          prWarningMsg := 'Failed to write measurements file: ' + ex.Message;
        end;
      end;
    end;

    if Repeats > 1 then begin
      VarToShow := 'Tau';
    end else if Normalize then begin
      VarToShow := 'Normalized Current';
    end else begin
      VarToShow := 'Current';
    end;
    if Repeats > 1 then begin
      prResults['DefaultChartsVs'].AppendChild(prVarName+' '+IntToStr(Repeats)).SetData(VarToShow);
      if 0 <> prProperties['Variable2'].Data[0] then begin
        prResults['DefaultChartsVs'].AppendChild(prVarName2+' '+IntToStr(Repeats)).SetData(VarToShow);
      end;
    end else begin
      prResults['DefaultChartsVs'].AppendChild(prVarName).SetData(VarToShow);
      if 0 <> prProperties['Variable2'].Data[0] then begin
        prResults['DefaultChartsVs'].AppendChild(prVarName2).SetData(VarToShow);
      end;
    end;

    if prReshapeResults then begin // avoid situation with just one result segment
      if Normalize then begin
        VarToShow := 'Normalized Current';
      end else begin
        VarToShow := 'Current';
      end;
      prResults['DefaultChartsVs'].AppendChild('AbsPeakTime').SetData(VarToShow);
    end;
    
    prResults.AppendClone(prProperties, True).AppendChild(prModel.GetAsNode).Name := ExtractFileName(prModel.FileName);

    Synchronize(prPublishResults);

  finally

    try
      CloseFile(f);
    except
    end;

    prUpdateStatus( 100, prStTime );

    if not prError then begin

      if prStopped then begin
        prStatus := 'Stopping';
      end else begin
        prStatus := 'Terminating';
      end;
      Synchronize(prDoOnChange);

      if prWarning then begin
        Synchronize(ShowMessage);
      end;

    end else if prError then begin

      prStatus := 'Error';
      Synchronize(prDoOnChange);

      Synchronize(ShowMessage);

    end;

    FreeMem(StageBuffer);
    FreeMem(prBufferMPX);

    if prStimData <> nil then begin
      ViewportData.ListObjects.SelQVO(prStimData);
      ViewportData.Close;
      ViewportData.ListObjects.SelQVO(Data);
    end;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse.prFitPeaks(N :Integer; XX, YY, WW :TPADouble;
                                                var A, B, Tau :Double; var FitState :INode);
var
  Vars                  :TStringList;
  Data                  :TQF_Data;
  PData                 :TPQF_Data;
  Curve                 :TQUB_Curve;
  Fitter                :TQUB_Fitter;
begin
  QF_Data_Init(Data);
  Curve       := nil;
  Fitter      := nil;
  Vars        := nil;

  A           := YY[0];
  B           := YY[N-1];
  Tau         := (XX[N-1] - XX[0]) / 4;
  FitState    := TNode.Create('FittingState');

  try
    Data.N                := N;
    Data.XX               := XX;
    Data.YY               := YY;
    Data.WW               := WW;
    Vars                  := TStringList.Create;
    Vars.AddObject('X', TObject(XX));
    Vars.AddObject('Y', TObject(YY));
    QF_Data_SetVars(Data, Vars);

    Curve                 := CreateCurve('Declining Exponential');
    Curve.SetParamVal(0, A, 1);
    Curve.SetParamVal(1, B, 1);
    Curve.SetParamVal(2, Tau, 1);
    Curve.SetParamVal(3, 0, 1);
    Curve.SetParamVal(4, 0, 1);
    Curve.SetParamConst(0, 1);
    Curve.SetParamConst(1, 1);
    Curve.SetParamConst(2, 0);
    Curve.SetParamConst(3, 1);
    Curve.SetParamConst(4, 1);

    Fitter                := CreateFitter('Levenburg-Marquardt');
    PData                 := @Data;
    Fitter.Fit(Curve, TPAQF_Data(@PData), 1);

    Curve.SetParamConst(0, 0);
    Curve.SetParamConst(1, 0);
    Fitter.Fit(Curve, TPAQF_Data(@PData), 1);

    if Curve.GetParamVal(1) < 0 then begin
      Curve.SetParamVal(1, 0, 1);
      Curve.SetParamConst(1, 1);
      Fitter.Fit(Curve, TPAQF_Data(@PData), 1);
    end;

    A := Curve.GetParamVal(0);
    B := Curve.GetParamVal(1);
    Tau := Curve.GetParamVal(2);
    FitState.AppendClone(TNode.CreateUsing(Curve.GetState), True).Name := 'Curve';
    FitState.AppendClone(TNode.CreateUsing(Fitter.GetState), True).Name := 'Fitter';
  except
  end;

  if Curve <> nil then Curve.Free;
  if Fitter <> nil then Fitter.Free;
  if Vars <> nil then Vars.Free;
  QF_Data_FreeVars(Data);
end;

function TQUB_Task_Sim_DoseResponse.prReshapeResults: Boolean;
var
  i            :Integer;
  Repeats      :Integer;
  DS, Seg      :INode;
  Child        :INode;
  First        :Integer;
  Start,Dur    :Double;
begin
  Result       := False;
  Repeats      := prProperties['Repeats'].Data[0];
  DS           := prResults.Find('DataSet');
  Seg          := DS.Find('Segment');
  if Seg.IsNull or (not Seg.SiblingSameName.IsNull) or (Repeats <= 1) then Exit;
  // we're still here because there is only one segment, with multi-reps
  First := Seg.Data[0];
  Start := Seg['start'].Data[0];
  Dur   := 1e3 * prData.Sampling * prStimCount;
  i := 0;
  repeat                                // multiply and adjust bounds
    Seg.Data[0] := First;
    First := First + prStimCount;
    Seg.Data[1] := First - 1;
    Seg['start'].SetData(Start);
    Start := Start + Dur;

    Inc(i);
    if i < Repeats then Seg := DS.InsertClone(Seg, Seg, True);
  until i = Repeats;

  Seg := DS.Find('Segment');             // distribute rep data
  i   := 0;
  while not Seg.IsNull do begin
    Child := Seg.Child;
    while not Child.IsNull do begin
      if LongInt(Child.DataCount) = Repeats then begin
        Child.SetData(Child.Data[i]);
      end;
      Child := Child.Sibling;
    end;
    Seg := Seg.SiblingSameName;
    Inc(i);
  end;
  Result := True;
end;

procedure TQUB_Task_Sim_DoseResponse.prOnSubtaskDone(Sender :TObject);
begin
  prSubtask := nil;
  prSignal.SetEvent;
end;

procedure TQUB_Task_Sim_DoseResponse.Stop;
begin
  inherited;
  if prSubtask <> nil then prSubtask.Stop;
  prSignal.SetEvent;
end;

procedure TQUB_Task_Sim_DoseResponse.prNewStimData;
var
  Base  :String;
  FName :String;
begin
  with ViewportData.NewPropertiesForm do begin
    Base  := GetSystemPath(SystemPath.Personal) + 'DoseResponse\';
    if not DirectoryExists(Base) then CreateDir(Base);
    FName := Base + GetBaseStampedFileName('DoseResponse-stimulus', '.qdf');
    if FileExists(FName) then FName := '';
    FileFormat     := QFS_QDF_TYPE_QDF;
    if 0<>prProperties['Variable2'].Data[0] then begin
      ADChannelCount := 2;
    end else begin
      ADChannelCount := 1;
    end;
    Sampling       := prProperties['Sampling'].Data[0];
    ADDataSize     := 2;
    Scaling        := prScaling;
  end;
  if ViewportData.New(FName, True{Silent}) then begin
    prError        := True;
    prErrorMsg     := 'Can''t create temporary stimulus file.';
    prWikiTopic    := 'Error:Cannot_Create_New_File';
    Exit;
  end;
  prStimData       := ViewportData.CurrData;
  prStimData.DataChannelName[0] := prVarName;
  if 0<>prProperties['Variable2'].Data[0] then begin
    prStimData.DataChannelName[1] := prVarName2;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse.prAddStimSeg;
begin
  prStimData.AppendData(prBufferMPX, prStimData.DataObject.ADChannelCount * prStimData.DataObject.ADDataSize, prStimCount, prIsNewSegment, prStartTime, False{CanRefresh - Doit}, False{Acq});
end;

procedure TQUB_Task_Sim_DoseResponse.prDoneStimSegs;
begin
  if prStimData <> nil then prStimData.AppendData(nil, 0, 0, False, 0, True, False{Acq});
end;

procedure TQUB_Task_Sim_DoseResponse.prSimulate;
var
  Task                     :TQUB_Task_Simulate;
  SimProperties            :INode;
  i                        :Integer;
  ExpCond                  :TQUB_ExpCond_Parameters;
begin
  Task                     := TQUB_Task_Simulate.Create(True{Suspended});
  prSubtask                := Task;
  prSubtask.OnTerminate    := prOnSubtaskDone;

  try
    ViewportData.ListObjects.SelQVO(Data);
    Task.Data              := prData;
    Task.StimData          := prStimData;

    prSavedChannelCount    := prModel.ModelObject.ChannelCount;
    prModel.ModelObject.ChannelCount   := prProperties['ChannelCount'].Data[0];
    Task.Models.AddObject('', prModel);

    ExpCond := ViewportModel.DefExpCond;
    for i := 0 to ExpCond.Count - 1 do begin
      ExpCond.ChangeParChannel(i, -1);
    end;
    i := ExpCond.IndexOf(prVarName);
    if i >= 0 then begin
      ExpCond.ChangeParChannel(i, 1);
    end else begin
      ExpCond.AddPar(prVarName, 1.0, 1, False, 1);
    end;
    if 0<>prProperties['Variable2'].Data[0] then begin
      i := ExpCond.IndexOf(prVarName2);
      if i >= 0 then begin
        ExpCond.ChangeParChannel(i, 2);
      end else begin
        ExpCond.AddPar(prVarName2, 1.0, 2, False, 2);
      end;
    end;
    Task.ModelParameters   := ExpCond;

    SimProperties          := ProfileMgr['Sim'].Active;
    with Task.Properties do begin
      if prModel.ModelObject.ChannelCount >= MACROSCOPIC_THRESHOLD then begin
        Response               := SIM_RESPONSE_DETERMINISTIC
      end else begin
        Response               := SIM_RESPONSE_STOCHASTIC;
      end;
      Stimulation              := SIM_STIM_FILE;
      UseCondVoltage           := prModel.IeqFv;
      CondVoltage              := SimProperties['Simulation']['CondVoltage'].Data[0];
      ILeak                    := 0<>SimProperties['Simulation']['ILeak'].Data[0];
      GLeak                    := SimProperties['Simulation']['GLeak'].Data[0];
      VLeak                    := SimProperties['Simulation']['VLeak'].Data[0];
      IdealizedData            := 0<>SimProperties['Data']['IdealizedData'].Data[0];
      IDLOutput                := Integer(OUT_CLASS);
      Randomize                := 0<>SimProperties['Data']['Randomize'].Data[0];
      RandomSeed               := SimProperties['Data']['RandomSeed'].Data[0];
      ApplyDeadTime            := 0<>SimProperties['Data']['ApplyDeadTime'].Data[0];
      DeadTime                 := SimProperties['Data']['DeadTime'].Data[0];
      FilterOn                 := 0<>SimProperties['FilterOn'].Data[0];
      FilterFreq               := SimProperties['FilterFreq'].Data[0];
      TransitionStates[1]      := SimProperties['Data']['TransitionState1'].Data[0];
      TransitionStates[2]      := SimProperties['Data']['TransitionState2'].Data[0];
      Flux                     := 0<>SimProperties['Flux'].Data[0];
      if SimProperties['FluxStates'].DataCount = 2 then begin
        FluxFrom               := SimProperties['FluxStates'].Data[0];
        FluxTo                 := SimProperties['FluxStates'].Data[1];
      end;
      Staircase                := 0<>SimProperties['Simulation']['Staircase'].Data[0];
      StaircaseUseSteps        := 0<>SimProperties['Simulation']['StaircaseUseSteps'].Data[0];
      StaircaseSteps           := SimProperties['Simulation']['StaircaseSteps'].DataAsString;
      StaircaseStepStd         := SimProperties['Simulation']['StaircaseStepStd'].Data[0];
      StaircaseNoJumpClass     := 0<>SimProperties['Simulation']['StaircaseNoJumpClass'].Data[0];
      BaselineFluct            := 0<>SimProperties['Baseline']['Fluctuations'].Data[0];
      BaselineFluctLifetime    := SimProperties['Baseline']['Fluctuations']['Lifetime'].Data[0];
      BaselineFluctMaxAmp      := SimProperties['Baseline']['Fluctuations']['MaxAmp'].Data[0];
      BaselineFluctN           := SimProperties['Baseline']['Fluctuations']['N'].Data[0];
      BaselineDrift            := 0<>SimProperties['Baseline']['LinearDrift'].Data[0];
      BaselineDriftSlope       := SimProperties['Baseline']['LinearDrift']['Slope'].Data[0];
      BaselineDriftIntercept   := SimProperties['Baseline']['LinearDrift']['Intercept'].Data[0];
      BaselineJumps            := 0<>SimProperties['Baseline']['Jumps'].Data[0];
      BaselineJumpsAmpAvg      := SimProperties['Baseline']['Jumps']['AmpAvg'].Data[0];
      BaselineJumpsAmpStd      := SimProperties['Baseline']['Jumps']['AmpStd'].Data[0];
      BaselineJumpsLifetime    := SimProperties['Baseline']['Jumps']['Lifetime'].Data[0];
      BaselineJumpsFastW       := SimProperties['Baseline']['Jumps']['Fast w'].Data[0];
      BaselineJumpsFast        := SimProperties['Baseline']['Jumps']['Fast rec'].Data[0];
      BaselineJumpsSlow        := SimProperties['Baseline']['Jumps']['Slow rec'].Data[0];
      BaselinePeriodic         := 0<>SimProperties['Baseline']['Periodic'].Data[0];
      BaselinePeriodicAmp      := SimProperties['Baseline']['Periodic']['Amp'].Data[0];
      BaselinePeriodicFreq     := SimProperties['Baseline']['Periodic']['Freq'].Data[0];
      BaselinePeriodicPhase    := SimProperties['Baseline']['Periodic']['Phase'].Data[0];
      BaselineLGM              := 0<>SimProperties['Baseline']['LinearGaussian'].Data[0];
      BaselineLGMProcStd       := SimProperties['Baseline']['LinearGaussian']['ProcessStd'].Data[0];
      BaselineLGMX0            := SimProperties['Baseline']['LinearGaussian']['X0'].Data[0];
      BaselineLGMX0p           := SimProperties['Baseline']['LinearGaussian']['X0p'].Data[0];
      BaselineLGM2State        := SimProperties['Baseline']['LinearGaussian']['2 state model'].Data[0];
      MakeMovie                := SimProperties['MakeMovie'].Data[0];
      MakeMovieSaveFrames      := SimProperties['MakeMovie']['SaveFrames'].Data[0];
    end;

    Task.OnChange          := TaskView.OnTaskChange;
    Task.SilentOutput      := True;

    Data.UnlockAny;

    if prData.Clear then begin
      prError := True;
      prErrorMsg := 'Can''t clear new DoseResponse data file';
      prSubtask.Terminate;
      Exit;
    end;

  finally

    prSubtask.Resume;

  end;
end;

procedure TQUB_Task_Sim_DoseResponse.prAfterSimulate;
begin
  prModel.ModelObject.ChannelCount := prSavedChannelCount;

  Data.UnlockAny;

  if prStimData <> nil then begin
    ViewportData.ListObjects.SelQVO(prStimData);
    ViewportData.Close;
    prStimData := nil;

    ViewportData.ListObjects.SelQVO(Data);
    ViewportData.CurrData.FilterOn[0] := 0<>prProperties['CurrentFilter'].Data[0];
    if ViewportData.CurrData.FilterOn[0] then begin
      ViewportData.CurrData.FilterFreq[0] := 1e3 * prProperties['CurrentFilterFreq'].Data[0];
    end;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse.prPublishResults;
var
  QVR         :TQUB_QVO_QVR;
  Chart       :TQUB_Results_Select2Chart;
  i           :Integer;
  Seg         :INode;
begin
  ViewportResults.AddResults( prResults, False );

  QVR := ViewportResults.CurrResults;
  Seg := prResults['DataSet']['Segment'];
  for i := 0 to QVR.Select.Chart2Count - 1 do begin
    Chart := QVR.Select.Charts2[i];
    if Chart.XSelVar.Name = 'AbsPeakTime' then begin
      Chart.SetupFit(Seg['FittingState']);
      Break;
    end;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse.prSetData(x :TQUB_QVO_QVD);
begin
  if Data <> nil then Data.UnlockAny;
  if (x <> nil) and x.LockedAny then x := nil;
  prData := x;
  if Data <> nil then Data.LockAny;
end;

procedure TQUB_Task_Sim_DoseResponse.OnGetProperties(Sender :TObject);
var
  Props :TQUB_Task_Sim_DoseResponse_PropertiesForm;
begin
  Props := TQUB_Task_Sim_DoseResponse_PropertiesForm(Sender);
  SetLength(prStimulus, Props.StimCount);
  Move(Props.Stimulus[0], prStimulus[0], Props.StimCount*SizeOf(Double));
  SetLength(prStimulus2, Props.StimCount);
  Move(Props.Stimulus2[0], prStimulus2[0], Props.StimCount*SizeOf(Double));
  prStimCount := Props.StimCount;
  prVarName   := Props.VarName;
  prVarName2  := Props.VarName2;
  prProperties := Props.Properties.Clone(True);
end;

procedure TQUB_Task_Sim_DoseResponse.OnPutProperties(Sender :TObject);
begin
end;


// *****************************************************************************
end.

