unit QUB_Task_Survey;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math, SyncObjs, Controls,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,

QUB_Viewport_Data, QUB_Viewport_Model,

QUB_Thread,
QUB_Task,

QUB_Task_MIL_Properties,
QUB_Task_MIL,
QUB_Task_MSL_Properties,
QUB_Task_MSL,
QUB_Task_MIP_Properties,
QUB_Task_MIP,
QUB_Task_Mac_Properties,
QUB_Task_Mac,

QUB_QVO_QVM, QUB_QFS_QMF, QUB_QFS_Channels,

QUB_Node,
QUB_GUIInterface;


type


TQUB_Task_Survey  = class(TQUB_Task)
protected

  prSignal              :TEvent;
  prRunning             :Boolean;
  prTask                :TQUB_Task;

  prModel               :TQUB_QVO_QVM;
  prModeSet             :INode;
  prResults             :INode;
  prSteps               :array of Double;
  prModelTree           :INode;
  prLastModel           :INode;
  prNode                :INode;
  prIx                  :Integer;
  prAlgo                :String;
  prMode                :String;
  prVary                :String;
  prTrialLbl            :String;
  prRestore             :Boolean;

  procedure prSetModel;
  procedure prSetMode;
  procedure prRunMIL;
  procedure prRunMSL;
  procedure prRunMIP;
  procedure prRunMac;
  procedure prOnTaskDone(Sender :TObject);

  procedure prSetupResults;
  procedure prHarvestResults;
  procedure prPublishResults;

public

  MILProperties          :TQUB_Task_MIL_PropertiesForm;
  MSLProperties          :TQUB_Task_MSL_PropertiesForm;
  MIPProperties          :TQUB_Task_MIP_PropertiesForm;
  MacProperties          :TQUB_Task_Mac_PropertiesForm;

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Execute; override;
  procedure Stop; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


end;


implementation



uses

QUB_Profiles,
QUB_Task_Survey_Properties,
QUB_ActionSelector_Modeling;


//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _M a c L i v e F i t           //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_Survey.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'Survey';

  prSignal              := TEvent.Create(nil, False, False, '');
  prRunning             := False;
  prTask                := nil;

  prModel               := ViewportModel.CurrModel;

end;

destructor TQUB_Task_Survey.Destroy;
begin
  prSignal.Free;

  inherited;
end;

//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Survey.OnGetProperties(Sender :TObject);
var
  List          :INode;
  i,s           :Integer;
  a,b,d         :Double;
begin
  with Sender as TQUB_Task_Survey_PropertiesForm do begin
    prModelTree := Model;
    prNode      := Node;
    prIx        := Ix;
    prAlgo      := Properties['Algorithm'].DataAsString;
    prMode      := Properties['Mode'].DataAsString;
    prVary      := Properties['Vary'].DataAsString;
    prRestore   := 0<>Properties['Restore'].Data[0];

    if Properties['In'].DataAsString = 'Range' then begin
      a   := Properties['From'].Data[0];
      b   := Properties['To'].Data[0];
      s   := Properties['Steps'].Data[0];
      SetLength(prSteps, s);
      prSteps[0]   := a;
      prSteps[s-1] := b;
      d   := b - a;
      if d > 0 then begin
        for i := 1 to s-2 do begin
          prSteps[i] := a + d * (exp(i/s/1.1) - 1);
        end;
      end else begin
        for i := 1 to s-2 do begin
          prSteps[i] := a + d * (1 - exp(-i/s/0.3));
        end;
      end;
    end else begin // 'List'
      List := Properties['List'];
      SetLength(prSteps, List.DataCount);
      for i := 0 to High(prSteps) do begin
        prSteps[i] := List.Data[i];
      end;
    end;
  end;
end;

procedure TQUB_Task_Survey.OnPutProperties(Sender :TObject);
begin
end;


//****************************************************************************//


procedure TQUB_Task_Survey.Execute;
var
  Val0    :Double;
  Mode0   :String;
  i       :Integer;
begin
  if Terminated then Exit;

  prStatus := 'Initializing';
  prPercent := 0;
  prETA := '0:00:00';
  Synchronize(prDoOnChange);

  Val0    := prNode.Data[prIx];
  Mode0   := ProfileMgr[prAlgo].Active['Mode'].DataAsString;
  prModeSet := TNode.Create('Properties');
  prModeSet['Mode'].SetData(prMode);
  Synchronize(prSetMode);

  prSetupResults;

  for i := 0 to High(prSteps) do begin
    if prStopped then Break;

    prStatus := 'Running';
    prPercent := (i * 100) div Length(prSteps);
    prETA := '0:00:00';
    Synchronize(prDoOnChange);

    prNode.Data[prIx] := prSteps[i];
    Synchronize(prSetModel);

    prRunning := True;
    if prAlgo = 'MIL' then begin
      Synchronize(prRunMIL);
    end else if prAlgo = 'MSL' then begin
      Synchronize(prRunMSL);
    end else if prAlgo = 'MIP' then begin
      Synchronize(prRunMIP);
    end else {if prAlgo = 'Mac' then} begin
      Synchronize(prRunMac);
    end;
    while prRunning do begin
      prSignal.WaitFor(1000);
    end;

    prTrialLbl := Format('%s %f', [prVary, prSteps[i]]);
    Synchronize(prHarvestResults);
  end;

  if prRestore then begin
    prNode.Data[prIx] := Val0;
    Synchronize(prSetModel);
  end;
  prModeSet['Mode'].SetData(Mode0);
  Synchronize(prSetMode);

  Synchronize(prPublishResults);

  prStatus := 'Terminated';
  prPercent := 0;
  prETA := '0:00:00';
  Synchronize(prDoOnChange);
end;

procedure TQUB_Task_Survey.prSetMode;
begin
  ProfileMgr[prAlgo].SetProperties(prModeSet);
end;

function TranscribeOneDetail(Tree1, Tree2, Detail1 :INode; DetailIx :Integer): Boolean;
// for lack of lambda...
var
  Ch1, Ch2   :INode;
begin
  Result     := False;
  if Tree1.Equals(Detail1) then begin
    Tree2.Data[DetailIx] := Tree1.Data[DetailIx];
    Result   := True;
  end;
  Ch1        := Tree1.Child;
  Ch2        := Tree2.Child;
  while not (Ch1.IsNull or Ch2.IsNull) do begin
    Result := TranscribeOneDetail(Ch1, Ch2, Detail1, DetailIx);
    if Result then Exit;
    Ch1      := Ch1.Sibling;
    Ch2      := Ch2.Sibling;
  end;
end;

procedure TQUB_Task_Survey.prSetModel;
begin
  if prRestore then begin
    prModel.SetAsNode(prModelTree);
  end else begin
    TranscribeOneDetail(prModelTree, prLastModel, prNode, prIx);
    prModel.SetAsNode(prLastModel);
  end;
end;

procedure TQUB_Task_Survey.prRunMIL;
begin
  TQUB_ActionSelector_Modeling(ActionsMod).ActionByName['MIL'].OnLClick(prOnTaskDone);
end;

procedure TQUB_Task_Survey.prRunMSL;
begin
  TQUB_ActionSelector_Modeling(ActionsMod).ActionByName['MSL'].OnLClick(prOnTaskDone);
end;

procedure TQUB_Task_Survey.prRunMIP;
begin
  TQUB_ActionSelector_Modeling(ActionsMod).ActionByName['MIP'].OnLClick(prOnTaskDone);
end;

procedure TQUB_Task_Survey.prRunMac;
//var
//  MacTask         :TQUB_Task_Mac;
begin
  TQUB_ActionSelector_Modeling(ActionsMod).ActionByName['Mac'].OnLClick(prOnTaskDone);
end;
//  MacTask := TQUB_Task_Mac.Create(True{Suspended});
//  prTask  := MacTask;
//  MacTask.OnTerminate    := prOnTaskDone;
//
//  MacProperties.OnGetProperties := MacTask.OnGetProperties;
//  MacProperties.OnPutProperties := MacTask.OnPutProperties;
//
//  MacProperties.FormShow(MacProperties);
//  MacProperties.ModalResult := mrOK;
//  MacProperties.btnOKClick(MacProperties);
//
//  MacTask.OnChange := TaskView.OnTaskChange;
//  MacTask.Data     := ViewportData.CurrData;
//  MacTask.Model    := prModel;
//
//  MacTask.SilentOutput          := True;
//
//  MacTask.Resume;
//
//end;

procedure TQUB_Task_Survey.prOnTaskDone(Sender :TObject);
begin
  prRunning   := False;
  prTask      := nil;
  prSignal.SetEvent;
end;

procedure TQUB_Task_Survey.Stop;
begin
  inherited;
  if prTask <> nil then prTask.Stop;
  prSignal.SetEvent;
end;

procedure TQUB_Task_Survey.prSetupResults;
begin
  prResults   := TNode.Create('Results');
  prResults['ResultType'].SetData('Surv');
  prResults['Hold'].SetData(1);
  prLastModel := prModelTree;
end;

procedure TQUB_Task_Survey.prHarvestResults;
var
  LastResult :INode;
  Trial      :INode;
  Mdl        :INode;
  DataSet    :INode;
begin
  LastResult := ViewportResults.CurrResults.Results.Node;
  Trial      := prResults.AppendChild('ModelTrial');
  Trial.AppendClone(LastResult['LL'], True).Name := ' LL';
  if not LastResult.Find('Gradient').IsNull then begin
    Trial.AppendClone(LastResult['Gradient'], True);
  end else begin
    Trial.AppendClone(LastResult['Free pars grads'], True).Name := 'Gradient';
  end;
  prLastModel := LastResult.Find('ModelFile');
  if prLastModel.IsNull then prLastModel := LastResult.Find('Properties').Find('ModelFile');
  if prLastModel.IsNull then prLastModel := ViewportModel.CurrModel.GetAsNode;
  prLastModel := prLastModel.Clone(True);
  Mdl        := Trial.AppendClone(prLastModel, True);
  Mdl.SetData(prTrialLbl);
  DataSet    := LastResult.Find('DataSet');
  while not DataSet.IsNull do begin
    Trial.AppendClone(DataSet, True);
    DataSet  := DataSet.SiblingSameName;
  end;
end;

procedure TQUB_Task_Survey.prPublishResults;
begin
  ViewportResults.AddResults( prResults, False );
end;


//******************************************************************************
end.

