unit QUB_Thread_Baseline;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


SysUtils, Classes, Windows, Forms, SyncObjs,
AllocTypes, Alloc,

QUB_Global,
QUB_QVO_QVD_Global,
QUB_Thread,
QUB_Thread_Data,
QUB_Baseline_Global,
QUB_Stats;


const


NODE_INCREMENT                    = 10000;


type


TQUB_Thread_Baseline_ProcessState = (PROCESSSTATE_NOTHING,
                                     PROCESSSTATE_READING_CHUNK, PROCESSSTATE_READING_PARAMS,
                                     PROCESSSTATE_WRITING_ACK
                                    );


TQUB_Thread_Baseline              = class;


TQUB_Thread_Baseline_Chunk        = record
  Chunk                           :TQUB_Thread_Chunk;
  Start                           :Integer;
  NodeCount                       :Integer;
  Nodes                           :TPAInteger;
  NodesDBL                        :TPAInteger;
end;
TQUB_Thread_Baseline_PChunk       = ^TQUB_Thread_Baseline_Chunk;


TQUB_Thread_Baseline              = class(TQUB_Thread_Data)
protected


  //Baseline Parameters
  prParams                        :TQUB_Baseline_Params;


  procedure prExecute_Linear;
  procedure prExecute_Adaptive;
  procedure prExecute_CapTrans;


public


  constructor Create(Remote :Boolean; Suspended :Boolean); override;

  
  procedure Execute; override;


  //Parameters
  property Params                 :TQUB_Baseline_Params       read prParams      write prParams;


end;


implementation


//******************************************************************************
constructor TQUB_Thread_Baseline.Create(Remote :Boolean; Suspended :Boolean);
begin

  inherited Create(Remote, Suspended);

  Priority := tpHigher;

end;


//Execute **********************************************************************
procedure TQUB_Thread_Baseline.Execute;
begin

  if Params.Method = METHOD_ADAPTIVE then begin
    prExecute_Adaptive;
  end else if Params.Method = METHOD_LINEAR then begin
    prExecute_Linear;
  end else begin
    prExecute_CapTrans;
  end;

end;


//****************************************************************************//
//                                                                            //
//                     L I N E A R     S E G M E N T S                        //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Baseline.prExecute_Linear;
var
  ExtChunk          :TQUB_FirstLast;

  Win               :TQUB_FirstLast;
  Ext               :TQUB_FirstLast;
  Reg               :TQUB_FirstLast;

  xAvg              :Double;
  yAvg              :Double;
  yStd              :Double;
  yStdR             :Double;
  Slope             :Double;
  Intercept         :Double;
  SlopeStd          :Double;

  ExtAvg            :Double;
  ExtStd            :Double;
  ExtStdR           :Double;
  ExtSlope          :Double;
  ExtSlopeStd       :Double;
  ExtIntercept      :Double;

  PredAvg           :Double;

  LastGood          :Integer;

  GoodFound         :Boolean;
  GoodIter          :Boolean;

  Nodes             :TPAInteger;
  NodesDBl          :TPAInteger;
  NodeCount         :Integer;
  NodeMaxCount      :Integer;
  OrigNodeCount     :Integer;

  x                 :Double;

  RegAdv            :Integer;

  f, l              :Boolean;
  s                 :Integer;
  i                 :Integer;
  p                 :Integer;

  dy                :Double;

  SegFound          :Boolean;

  MaxBaseShift      :Double;

  Alpha             :Double;

  Segment           :TQUB_Baseline_PSegment;
  Prev, Next        :TQUB_Baseline_PSegment;
  Sel               :TQUB_PFirstLast;
  Accept            :Boolean;
  y, y1, y2         :Double;

  ActualMaxSlope    :Double;

  dNode             :Double;

  SegIndex          :Integer;
  SegFirst          :Integer;
  SegLast           :Integer;
  Seg1, Seg2        :TQUB_Baseline_PSegment;
  Xa, Xb            :Double;
  Xa2, Xb2          :Double;
  Xa3, Xb3          :Double;
  Ya, Yb            :Double;
  dYa, dYb          :Double;
  detM              :Double;
  detA              :Double;
  detB              :Double;
  detC              :Double;
  a, b, c, d        :Double;
  SmallCount        :Integer;
  FirstPoint        :Integer;
  LastPoint         :Integer;
  IncCount          :Integer;

  SegList           :TList;

  DataSize          :Integer;

  NewCompletion     :Integer;
  TotalCount        :Integer;
  ActualCount       :Integer;

  Error             :Boolean;

  BasChunk          :TQUB_Thread_Baseline_PChunk;

begin

  BasChunk            := TQUB_Thread_Baseline_PChunk(Chunk[0]);

  Chunk[0].Status     := CHUNK_STATUS_RUNNING;

  prNeedData          := nil;

  Error               := False;

  NodeCount           := 0;
  NodeMaxCount        := 0;
  Nodes               := nil;
  NodesDBl            := nil;

  SegList             := nil;

  try

    //Find the Buffer Size
    if prRemote then begin
      DataSize := ONE_MEGABYTE div 16;
    end else begin
      DataSize := ONE_MEGABYTE;
    end;
    if DataSize > (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1) then DataSize := (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1);

    try

      //Allocate Buffer
      prNeedData := IAlloc1(DataSize);

    except

      Error := True;
      Exit;

    end;

    SegList := TList.Create;

    //Some local parameters
    RegAdv := Round(Params.Linear.RegressionSegment * Params.Linear.RegressionAdvance / 100);
    if RegAdv = 0 then RegAdv := 1;
    MaxBaseShift := Params.Linear.MaxBaselineShift * Params.Scaling * Params.Init.yStdR;
    ActualMaxSlope := Params.Linear.MaxSlope * Params.Init.yStdR / 10;

    //Calculate the Total Count
    TotalCount := Chunk[0].Sel.Last - Chunk[0].Sel.First + 1;

    {
    Three windows are used:

      - Win: holds the good segment
      - Reg: holds the regression segment
      - Ext: holds the segment to be tryed

    }

    GoodFound := False;

    Win.First := Chunk[0].Sel.First;
    Win.Last  := Win.First;

    prHasDataSel.Last := Win.Last - 1;

    //**********************************************************************//
    //                                                                      //
    //                              Main loop                               //
    //                                                                      //
    //**********************************************************************//
    repeat
    //Main loop

      Ext.First := Win.First;
      Ext.Last  := Ext.First + DataSize - 1;
      if Ext.Last > Chunk[0].Sel.Last then Ext.Last := Chunk[0].Sel.Last;

      prCheckNeedsLoadData(Ext, +1, DataSize);

      Reg.First := Win.First;
      Reg.Last  := Reg.First;

      //**********************************************************************//
      //                                                                      //
      //                     Step 1 - Identify segments                       //
      //                                                                      //
      //**********************************************************************//
      repeat
      //Iteration loop - each iteration processes one ChunkBuffer


       //Stop thread if user-stopped
        if prStopped then begin
          Exit;
        end;

        ActualCount   := Win.First - Chunk[0].Sel.First;
        NewCompletion := Round(100 * ActualCount / TotalCount);
        if NewCompletion <> prCompletion then begin
          prCompletion := NewCompletion;
          Chunk[0].Completion := prCompletion;
          prReportStatus;
        end;

        //Update Ext
        if Win.First = Win.Last then begin
          Ext.First := Win.First;
          Ext.Last  := Ext.First + Params.Linear.ExtensionSegment - 1;
        end else begin
          Ext.First := Win.Last + 1;
          Ext.Last  := Ext.First + Params.Linear.ExtensionSegment - 1;
        end;
        if Ext.Last > Chunk[0].Sel.Last then Ext.Last := Chunk[0].Sel.Last;


        //Step 1 Segment acceptance criteria: **********************************
        //  - variance
        //  - outliers
        //  - coincidence, if it is added to previous
        //**********************************************************************


        //Variance *************************************************************
        CalculateVariance(@prNeedData[Ext.First - prHasDataSel.First], Ext, xAvg, yAvg, yStdR, Slope, Intercept);
        GoodIter := SampleVarianceLikePopulation(yStdR, Params.Init.yStdR, (Ext.Last - Ext.First), Params.Linear.VarianceConfidence, True);

        //Outliers *************************************************************
        if GoodIter then begin

          GoodIter  := not HasOutliers(@prNeedData[Ext.First - prHasDataSel.First], Ext, xAvg, yAvg, Slope, Params.Linear.OutliersTolerance * yStdR, p, dy);
          if not GoodIter then begin
            if p > (Ext.First + 1) then begin
            //Salvage whatever possible
              Ext.Last := p - 1;
              //Adjust calculations
              CalculateVariance(@prNeedData[Ext.First - prHasDataSel.First], Ext, xAvg, yAvg, yStdR, Slope, Intercept);
              GoodIter := True;
            end;
          end;

        end;

        //Coincidence **********************************************************
        if GoodIter then begin

          if Reg.Last > Reg.First then begin
          //Prev is non-null
            GoodIter := SegsCoincide(@prNeedData[Reg.First - prHasDataSel.First], @prNeedData[Ext.First - prHasDataSel.First], Reg, Ext, 1 - Params.Linear.CoincidenceConfidence, Params.Init.yStdR);
          end else begin
          //Prev is null, therefore Next coincides with itself
            GoodIter := True;
          end;

        end;


        //Accept Segment *******************************************************
        if GoodIter then begin
        //Iteration was successful, keep going

          GoodFound := True;

          Win.Last := Ext.Last;

          if (Ext.Last - Win.First + 1) < Params.Linear.RegressionSegment{Change name} then begin
          //if shorter than regression, then Extend

            Reg.Last := Ext.Last;

          end;

          if ((Win.Last + Params.Linear.ExtensionSegment) > prHasDataSel.Last) or
          //if next extension would cross the limits of the available buffer
              (Win.Last > (Chunk[0].Sel.Last - 3))
           //Not much has remained
          then begin

            GoodIter := False;

          end;

        end;//of GoodIter ******************************************************


        //Unacceptable Segment or Buffer exhausted *****************************
        if not GoodIter then begin

          if GoodFound then begin
          //If a Segment was found

            if (Win.Last - Win.First + 1) >= Params.Linear.MinSegment then begin
            //If the Segment is long enough

              CalculateVariance(@prNeedData[Win.First - prHasDataSel.First], Win, xAvg, yAvg, yStdR, Slope, Intercept);

              //Add Segment ****************************************************
              New(Segment);
              Segment.Sel       := Win;
              Segment.xAvg      := xAvg;
              Segment.yAvg      := yAvg;
              Segment.yStdR     := yStdR;
              Segment.Slope     := Slope;
              Segment.Intercept := Intercept;
              SegList.Add(Segment);
              //****************************************************************

            end;

            Win.First := Win.Last + 1;

          end else begin
          //No Segment was found, advance the window

            Inc(Win.First, RegAdv);

          end;

          Win.Last := Win.First;
          Reg := Win;

          GoodFound := False;

        end;//of not GoodIter **************************************************

      until (((Win.Last + Params.Linear.ExtensionSegment) > prHasDataSel.Last) or ((Win.Last + Params.Linear.ExtensionSegment) > (Chunk[0].Sel.Last - 3)));
      //Iteration loop


      //**********************************************************************//
      //                                                                      //
      //                     Step 2 - Process segments                        //
      //                                                                      //
      //**********************************************************************//
      if SegList.Count > 0 then begin

        for s := 0 to SegList.Count - 1 do begin

          Segment := SegList.List[s];

          //Slope ****************************************************************
          Accept := Abs(Segment.Slope) <= Abs(ActualMaxSlope);

          if Accept then begin

            if (Segment.Sel.Last - Segment.Sel.First + 1) < Params.Linear.MinBaseline then begin
            //Length not satisfied, try to recover

              //Find Prev and Next
              Prev := nil;
              Next := nil;
              for i := s - 1 downto 0 do begin
                Prev := SegList.List[i];
                if (Prev.Sel.Last - Prev.Sel.First + 1) >= Params.Linear.MinBaseline then begin
                  Break;
                end else begin
                  Prev := nil;
                end;
              end;
              for i := s + 1 to SegList.Count - 1 do begin
                Next := SegList.List[i];
                if (Next.Sel.Last - Next.Sel.First + 1) >= Params.Linear.MinBaseline then begin
                  Break;
                end else begin
                  Next := nil;
                end;
              end;

              if (Prev = nil) and (Next = nil) then begin
                Accept := False;
              end;

              if Accept then begin
              //Segment is between two large segments, Prev and Next

                if Prev <> nil then begin
                  y1 := Prev.Intercept + Prev.Sel.Last  * Prev.Slope;
                end;
                if Next <> nil then begin
                  y2 := Next.Intercept + Next.Sel.First * Next.Slope;
                end;

                if Prev = nil then begin
                  y1 := y2;
                end else if Next = nil then begin
                  y2 := y1;
                end else begin
                  if y1 > y2 then begin
                    y := y1;
                    y1 := y2;
                    y2 := y;
                  end;
                end;

                if (Segment.yAvg < (y1 - MaxBaseShift)) or (Segment.yAvg > (y2 + MaxBaseShift)) then begin

                  Accept := False;

                end;

              end;

            end;

          end;//of Accept

          Segment.Accept := Accept;
          //********************************************************************

        end;//SegList


        //Third order interpolation
        if Params.Linear.ThirdOrderInterpolation then begin

          SegIndex := 0;
          repeat

            SegFirst := -1;
            SegLast := -1;

            Seg1 := nil;
            Seg2 := nil;

            //Find first segment
            for s := SegIndex to SegList.Count - 1 do begin
              Seg1 := SegList.List[s];
              if Seg1.Accept then begin
                SegFirst := s;
                Break;
              end;
            end;
            if SegFirst = -1 then Break;
            for s := SegFirst + 1 to SegList.Count - 1 do begin
              Seg2 := SegList.List[s];
              if Seg2.Accept then begin
                SegLast := s;
                Break;
              end;
            end;
            if SegLast = -1 then Break;

            SegIndex := SegLast;

            if (Seg2.Sel.First - Seg1.Sel.Last) < 3 then Continue;

            Xa := Seg1.Sel.Last  - Seg1.Sel.Last;
            Xb := Seg2.Sel.First - Seg1.Sel.Last;
            Ya := (Seg1.Intercept + Seg1.Sel.Last  * Seg1.Slope);
            Yb := (Seg2.Intercept + Seg2.Sel.First * Seg2.Slope);
            dYa := Seg1.Slope;
            dYb := Seg2.Slope;

            Xa2 := Xa*Xa;
            Xa3 := Xa2*Xa;
            Xb2 := Xb*Xb;
            Xb3 := Xb2*Xb;

            detM := (Xb3 - Xa3)*2*(Xa     - Xb)     - (Xb2 - Xa2)*3*(Xa2     - Xb2)     + (Xb - Xa)*2*3*(Xa2*Xb  - Xb2*Xa);
            detA := (Yb - Ya)  *2*(Xa     - Xb)     - (Xb2 - Xa2)*  (dYa     - dYb)     + (Xb - Xa)*2  *(dYa*Xb  - dYb*Xa);
            detB := (Xb3 - Xa3)*  (dYa    - dYb)    - (Yb - Ya)  *3*(Xa2     - Xb2)     + (Xb - Xa)  *3*(Xa2*dYb - Xb2*dYa);
            detC := (Xb3 - Xa3)*2*(Xa*dYb - Xb*dYa) - (Xb2 - Xa2)*3*(Xa2*dYb - Xb2*dYa) + (Yb - Ya)*2*3*(Xa2*Xb  - Xb2*Xa);

            a := detA / detM;
            b := detB / detM;
            c := detC / detM;
            d := Ya - a*Xa3 - b*Xa2 - c*Xa;

            //Divide it in smaller linear, accepted segments ...
            SmallCount := Round((Seg2.Sel.First - Seg1.Sel.Last) / 10);
            if SmallCount < 2 then SmallCount := 2;
            FirstPoint := Seg1.Sel.Last + 1;
            IncCount := 0;

            repeat

              LastPoint := FirstPoint + SmallCount - 1;
              if LastPoint >= Seg2.Sel.First then LastPoint := Seg2.Sel.First - 1;

              if LastPoint = FirstPoint then Break;

              Xa := (FirstPoint - Seg1.Sel.Last);
              Xb := (LastPoint  - Seg1.Sel.Last);

              try
                Ya := a*Xa*Xa*Xa + b*Xa*Xa + c*Xa + d;
                Yb := a*Xb*Xb*Xb + b*Xb*Xb + c*Xb + d;
              except
                Ya := (Seg1.Intercept + Seg1.Sel.Last  * Seg1.Slope);
                Yb := (Seg2.Intercept + Seg2.Sel.First * Seg2.Slope);
              end;

              New(Segment);
              Segment.Sel.First := FirstPoint;
              Segment.Sel.Last  := LastPoint;
              Segment.xAvg      := (FirstPoint + LastPoint) / 2;
              Segment.yAvg      := (Ya + Yb) / 2;
              Segment.yStdR     := 0;
              Segment.Slope     := (Yb - Ya) / (LastPoint - FirstPoint);
              Segment.Intercept := Segment.yAvg - Segment.Slope * Segment.xAvg;
              Segment.Accept    := True;
              SegList.Insert(SegLast + IncCount, Segment);

              Inc(IncCount);

              FirstPoint := LastPoint + 1;

            until(FirstPoint >= Seg2.Sel.First);

            SegIndex := SegLast + IncCount;

          until((SegFirst = -1) or (SegLast = -1));

        end;

        for s := 0 to SegList.Count - 1 do begin
          Segment := SegList.List[s];
          if Segment.Accept then begin
          //Add Nodes
            Inc(NodeCount, 2);

            if NodeCount > NodeMaxCount then begin
              Inc(NodeMaxCount, 100);
              ReallocMem(Nodes,    SizeOf(Integer) * NodeMaxCount);
              ReallocMem(NodesDBl, SizeOf(Integer) * NodeMaxCount);
            end;

            //This is not compatible with interpolation
            dNode := (1 / (1 + Exp(((Segment.Sel.Last - Segment.Sel.First + 1) - 50) / 15)));
            dNode := dNode * ((Segment.Sel.Last - Segment.Sel.First + 1) / 2);
            Nodes[NodeCount - 2]    := Round(Segment.Sel.First + dNode);
            Nodes[NodeCount - 1]    := Round(Segment.Sel.Last  - dNode);
            if Nodes[NodeCount - 1] = Nodes[NodeCount - 2] then begin
              Nodes[NodeCount - 1] := Nodes[NodeCount - 1] + 1;
              Nodes[NodeCount - 2] := Nodes[NodeCount - 2] - 1;
            end;

            NodesDBl[NodeCount - 2] := Params.DCOffset - Round(Segment.Intercept + Nodes[NodeCount - 2] * Segment.Slope);
            NodesDBl[NodeCount - 1] := Params.DCOffset - Round(Segment.Intercept + Nodes[NodeCount - 1] * Segment.Slope);

          end;
        end;

        //Now clear the SegList ************************************************
        for i := SegList.Count - 1 downto 0 do begin
          Segment := SegList.List[i];
          Dispose(Segment);
        end;
        SegList.Clear;
        //**********************************************************************

      end;

    until ((Win.Last + Params.Linear.ExtensionSegment) > (Chunk[0].Sel.Last - 3));
    //Main loop


  finally

    SegList.Free;

    FreeMem(prNeedData);
    prNeedData := nil;

    if (not prStopped) and (not Error) then begin

      Chunk[0].Status := CHUNK_STATUS_DONE;

      f := False;
      l := False;
      if NodeCount > 0 then begin
        OrigNodeCount := NodeCount;
        if Nodes[NodeCount - 1] <> Chunk[0].Sel.Last  then begin
          Inc(NodeCount);
          l := True;
        end;
        if Nodes[0]             <> Chunk[0].Sel.First then begin
          Inc(NodeCount);
          f := True;
        end;
        if (NodeCount <> OrigNodeCount) or (NodeCount <> NodeMaxCount) then begin
          ReallocMem(Nodes,    SizeOf(Integer) * NodeCount);
          ReallocMem(NodesDBl, SizeOf(Integer) * NodeCount);
        end;
        if f then begin
          for i := OrigNodeCount downto 1 do begin
            Nodes   [i] := Nodes   [i - 1];
            NodesDBl[i] := NodesDBl[i - 1];
          end;
          Nodes[0] := Chunk[0].Sel.First;
        end;
        if l then begin
          Nodes   [NodeCount - 1] := Chunk[0].Sel.Last;
          NodesDBl[NodeCount - 1] := NodesDBl[NodeCount - 2];
        end;

      end else begin
      //No nodes found - create the two marginal nodes

        NodeCount := 2;
        ReallocMem(Nodes,    SizeOf(Integer) * NodeCount);
        ReallocMem(NodesDBl, SizeOf(Integer) * NodeCount);

        Nodes[0]    := Chunk[0].Sel.First;
        Nodes[1]    := Chunk[0].Sel.Last;
        NodesDBL[0] := Params.DCOffset;
        NodesDBL[1] := Params.DCOffset;

      end;

      BasChunk.NodeCount := NodeCount;
      BasChunk.Nodes     := Nodes;
      BasChunk.NodesDBL  := NodesDBL;

    end else if (prStopped) or (Error) then begin
    //Error or Stopped

      if prStopped then begin
        Chunk[0].Status := CHUNK_STATUS_STOPPED;
      end else begin
        Chunk[0].Status := CHUNK_STATUS_ERROR;
      end;

      //Reset output
      FreeMem(BasChunk.Nodes);
      FreeMem(BasChunk.NodesDBL);
      BasChunk.NodeCount := 0;
      BasChunk.Nodes     := nil;
      BasChunk.NodesDBL  := nil;

    end;

    prReportTermination;

    Terminate;

  end;

end;


//****************************************************************************//
//                                                                            //
//                       A D A P T I V E    W I N D O W                       //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Baseline.prExecute_Adaptive;
const
  MIN_INSIDE          = 5;
var
  DataSize            :Integer;

  Win                 :TQUB_FirstLast;
  Ext                 :TQUB_FirstLast;
  //ExtChunk            :TQUB_FirstLast;

  NodeCount           :Integer;
  NodeMaxCount        :Integer;
  OrigNodeCount       :Integer;
  NodeCountForward    :Integer;
  NodeCountBackward   :Integer;
  Nodes               :TPAInteger;
  NodesDBl            :TPAInteger;
  BackNodes           :TPAInteger;
  BackNodesDBl        :TPAInteger;

  DWin                :Double;

  ThisAvg             :Double;
  ThisSum             :Double;
  PrevAvg             :Double;

  Sense               :Integer;

  NeedSum             :Boolean;

  NewCompletion       :Integer;
  TotalCount          :Integer;
  ActualCount         :Integer;

  ToBeSub             :Integer;
  ToBeAdd             :Integer;

  N                   :Integer;

  s                   :Integer;
  i                   :Integer;
  f, l                :Boolean;

  Error               :Boolean;

  BasChunk            :TQUB_Thread_Baseline_PChunk;

begin

  BasChunk            := TQUB_Thread_Baseline_PChunk(Chunk[0]);

  Chunk[0].Status     := CHUNK_STATUS_RUNNING;

  prNeedData          := nil;

  Error               := False;

  NodeCount           := 0;
  NodeCountForward    := 0;
  NodeMaxCount        := 0;
  Nodes               := nil;
  NodesDBl            := nil;

  try

    //Find the Buffer Size
    if prRemote then begin
      DataSize := ONE_MEGABYTE div 16;
    end else begin
      DataSize := ONE_MEGABYTE;
    end;
    if DataSize > (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1) then DataSize := (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1);

    try

      //Allocate Buffer
      prNeedData := IAlloc1(DataSize);

    except

      Error := True;
      Exit;

    end;

    Win.First := BasChunk.Start - (Params.Adaptive.WindowWidth div 2) + 1;
    if Win.First < Chunk[0].Sel.First then Win.First := Chunk[0].Sel.First;
    Win.Last  := Win.First + Params.Adaptive.WindowWidth - 1;
    if Win.Last > Chunk[0].Sel.Last then Win.Last := Chunk[0].Sel.Last;
    if (Win.Last - Win.First + 1) > DataSize then begin
      Win.Last := Win.First + DataSize - 1;
    end;

    //It will need it !!!
    prHasDataSel.Last := Win.Last - 1;

    DWin := Params.Adaptive.WindowHeight * Params.Init.Std;

    //Calculate the Total Count
    TotalCount := (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1) - Params.Adaptive.WindowWidth;
    ActualCount := 0;


    Sense := +1;
    //Loop once in each sense
    for s := 1 to 2 do begin

      Ext := Win;

      prCheckNeedsLoadData(Ext, Sense, DataSize);

      NeedSum := False;
      PrevAvg := Params.Init.Avg;
      CalculateAdaptiveWindowAvg(@prNeedData[Ext.First - prHasDataSel.First], Ext, (PrevAvg - DWin), (PrevAvg + DWin), ThisAvg, ThisSum, N);

      repeat

        //Stop thread if user-stopped
        if prStopped then begin
          Exit;
        end;

        NewCompletion := Round(100 * ActualCount / TotalCount);
        if NewCompletion <> prCompletion then begin
          prCompletion := NewCompletion;
          Chunk[0].Completion := prCompletion;
          prReportStatus;
        end;

        if Sense = +1 then begin
        //Forward
          ToBeSub := prNeedData[Ext.First - prHasDataSel.First];
        end else begin
        //Backword
          ToBeSub := prNeedData[Ext.Last - prHasDataSel.First];
        end;

        Inc(Ext.First, Sense);
        Inc(Ext.Last,  Sense);

        if (Ext.First < Chunk[0].Sel.First) or (Ext.Last > Chunk[0].Sel.Last) then Break;

        Inc(ActualCount);

        prCheckNeedsLoadData(Ext, Sense, DataSize);

        if Sense = +1 then begin
        //Forward
          ToBeAdd := prNeedData[Ext.Last - prHasDataSel.First];
        end else begin
        //Backword
          ToBeAdd := prNeedData[Ext.First - prHasDataSel.First];
        end;


        if NeedSum then begin
        //Re-Sum

          CalculateAdaptiveWindowAvg(@prNeedData[Ext.First - prHasDataSel.First], Ext, (PrevAvg - DWin), (PrevAvg + DWin), ThisAvg, ThisSum, N);
          NeedSum := False;

        end else begin
        //Use Sub/Add

          //Sub First Point ?
          if N > 0 then begin
            if ((PrevAvg - DWin) <= ToBeSub) and ((PrevAvg + DWin) >= ToBeSub) then begin
              ThisSum := ThisSum - ToBeSub;
              Dec(N);
            end;
          end;

          //Add Next Point ?
          if   ((PrevAvg - DWin) <= ToBeAdd) and ((PrevAvg + DWin) >= ToBeAdd) then begin
            ThisSum := ThisSum + ToBeAdd;
            Inc(N);
          end;

          if N > 0 then begin
            ThisAvg := ThisSum / N;
          end;

        end;


        if N >= MIN_INSIDE then begin

          if Abs(ThisAvg - PrevAvg) >= Params.Adaptive.MinDifference then begin
          //Add Node

            Inc(NodeCount);
            if NodeCount > NodeMaxCount then begin
              Inc(NodeMaxCount, NODE_INCREMENT);
              try
                ReallocMem(Nodes,    SizeOf(Integer) * NodeMaxCount);
                ReallocMem(NodesDBl, SizeOf(Integer) * NodeMaxCount);
              except
                Dec(NodeCount);
                Exit;
              end;
            end;
            Nodes   [NodeCount - 1] := (Ext.First + Ext.Last) div 2;
            NodesDBl[NodeCount - 1] := Params.DCOffset - Round(ThisAvg);

            NeedSum := True;
            PrevAvg := ThisAvg;

          end;

        end;

      until(False);//of Ext advance

      Sense := -Sense;

      if Sense = -1 then begin
        NodeCountForward := NodeCount;
      end;

    end;//for sense


  finally

    FreeMem(prNeedData);
    prNeedData := nil;

    if (not prStopped) and (not Error) then begin

      Chunk[0].Status := CHUNK_STATUS_DONE;

      f := False;
      l := False;

      if NodeCount > 0 then begin

        NodeCountBackward := NodeCount - NodeCountForward;
        if NodeCountBackward > 0 then begin
        //Reorder Nodes

          //Make a reversed copy of backwards Nodes
          BackNodes    := nil;
          BackNodesDBl := nil;

          try

            try

              ReallocMem(BackNodes,    SizeOf(Integer) * NodeCountBackward);
              ReallocMem(BackNodesDBl, SizeOf(Integer) * NodeCountBackward);

              for i := 0 to NodeCountBackward - 1 do begin
                BackNodes   [i] := Nodes   [(NodeCount - 1) - i];
                BackNodesDBl[i] := NodesDBl[(NodeCount - 1) - i];
              end;

              for i := NodeCount - 1 downto NodeCount - NodeCountForward do begin
                Nodes   [i] := Nodes   [i - NodeCountBackward];
                NodesDBl[i] := NodesDBl[i - NodeCountBackward];
              end;

              for i := 0 to NodeCountBackward - 1 do begin
                Nodes   [i] := BackNodes   [i];
                NodesDBl[i] := BackNodesDBl[i];
              end;

            except

              NodeCount := NodeCountForward;

            end;

          finally

            FreeMem(BackNodes);
            FreeMem(BackNodesDBl);

          end;

        end;

        OrigNodeCount := NodeCount;
        if Nodes[NodeCount - 1] <> Chunk[0].Sel.Last  then begin
          Inc(NodeCount);
          l := True;
        end;
        if Nodes[0]             <> Chunk[0].Sel.First then begin
          Inc(NodeCount);
          f := True;
        end;
        if (NodeCount <> OrigNodeCount) or (NodeCount <> NodeMaxCount) then begin
          ReallocMem(Nodes,    SizeOf(Integer) * NodeCount);
          ReallocMem(NodesDBl, SizeOf(Integer) * NodeCount);
        end;
        if f then begin
          for i := OrigNodeCount downto 1 do begin
            Nodes   [i] := Nodes   [i - 1];
            NodesDBl[i] := NodesDBl[i - 1];
          end;
          Nodes[0] := Chunk[0].Sel.First;
        end;
        if l then begin
          Nodes   [NodeCount - 1] := Chunk[0].Sel.Last;
          NodesDBl[NodeCount - 1] := NodesDBl[NodeCount - 2];
        end;

      end else begin
      //No nodes found - create the two marginal nodes

        NodeCount := 2;
        ReallocMem(Nodes,    SizeOf(Integer) * NodeCount);
        ReallocMem(NodesDBl, SizeOf(Integer) * NodeCount);

        Nodes[0]    := Chunk[0].Sel.First;
        Nodes[1]    := Chunk[0].Sel.Last;
        NodesDBL[0] := Params.DCOffset;
        NodesDBL[1] := Params.DCOffset;

      end;

      BasChunk.NodeCount := NodeCount;
      BasChunk.Nodes     := Nodes;
      BasChunk.NodesDBL  := NodesDBL;

    end else if (prStopped) or (Error) then begin
    //Error or Stopped

      if prStopped then begin
        Chunk[0].Status := CHUNK_STATUS_STOPPED;
      end else begin
        Chunk[0].Status := CHUNK_STATUS_ERROR;
      end;

      //Reset output
      FreeMem(BasChunk.Nodes);
      FreeMem(BasChunk.NodesDBL);
      BasChunk.NodeCount := 0;
      BasChunk.Nodes     := nil;
      BasChunk.NodesDBL  := nil;

    end;

    prReportTermination;

    Terminate;

  end;

end;


//****************************************************************************//
//                                                                            //
//                C A P A C I T A N C E   T R A N S I E N T S                 //
//          (for now just a copy of _Linear)                                  //
//****************************************************************************//
procedure TQUB_Thread_Baseline.prExecute_CapTrans;
const
  MIN_INSIDE          = 5;
var
  DataSize            :Integer;

  Win                 :TQUB_FirstLast;
  Ext                 :TQUB_FirstLast;
  
  NodeCount           :Integer;
  NodeMaxCount        :Integer;
  OrigNodeCount       :Integer;
  NodeCountForward    :Integer;
  NodeCountBackward   :Integer;
  Nodes               :TPAInteger;
  NodesDBl            :TPAInteger;
  BackNodes           :TPAInteger;
  BackNodesDBl        :TPAInteger;

  DWin                :Double;

  ThisAvg             :Double;
  ThisSum             :Double;
  PrevAvg             :Double;

  Sense               :Integer;

  NeedSum             :Boolean;

  NewCompletion       :Integer;
  TotalCount          :Integer;
  ActualCount         :Integer;

  ToBeSub             :Integer;
  ToBeAdd             :Integer;

  N                   :Integer;

  s                   :Integer;
  i                   :Integer;
  f, l                :Boolean;

  Error               :Boolean;

  BasChunk            :TQUB_Thread_Baseline_PChunk;

begin

  BasChunk            := TQUB_Thread_Baseline_PChunk(Chunk[0]);

  Chunk[0].Status     := CHUNK_STATUS_RUNNING;

  prNeedData          := nil;

  Error               := False;

  NodeCount           := 0;
  NodeCountForward    := 0;
  NodeMaxCount        := 0;
  Nodes               := nil;
  NodesDBl            := nil;

  try

    //Find the Buffer Size
    if prRemote then begin
      DataSize := ONE_MEGABYTE div 16;
    end else begin
      DataSize := ONE_MEGABYTE;
    end;
    if DataSize > (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1) then DataSize := (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1);

    try

      //Allocate Buffer
      prNeedData := IAlloc1(DataSize);

    except

      Error := True;
      Exit;

    end;

    Win.First := BasChunk.Start - (Params.Adaptive.WindowWidth div 2) + 1;
    if Win.First < Chunk[0].Sel.First then Win.First := Chunk[0].Sel.First;
    Win.Last  := Win.First + Params.Adaptive.WindowWidth - 1;
    if Win.Last > Chunk[0].Sel.Last then Win.Last := Chunk[0].Sel.Last;
    if (Win.Last - Win.First + 1) > DataSize then begin
      Win.Last := Win.First + DataSize - 1;
    end;

    //It will need it !!!
    prHasDataSel.Last := Win.Last - 1;

    DWin := Params.Adaptive.WindowHeight * Params.Init.Std;

    //Calculate the Total Count
    TotalCount := (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1) - Params.Adaptive.WindowWidth;
    ActualCount := 0;


    Sense := +1;
    //Loop once in each sense
    for s := 1 to 2 do begin

      Ext := Win;

      prCheckNeedsLoadData(Ext, Sense, DataSize);

      NeedSum := False;
      PrevAvg := Params.Init.Avg;
      CalculateAdaptiveWindowAvg(@prNeedData[Ext.First - prHasDataSel.First], Ext, (PrevAvg - DWin), (PrevAvg + DWin), ThisAvg, ThisSum, N);

      repeat

        //Stop thread if user-stopped
        if prStopped then begin
          Exit;
        end;

        NewCompletion := Round(100 * ActualCount / TotalCount);
        if NewCompletion <> prCompletion then begin
          prCompletion := NewCompletion;
          Chunk[0].Completion := prCompletion;
          prReportStatus;
        end;

        if Sense = +1 then begin
        //Forward
          ToBeSub := prNeedData[Ext.First - prHasDataSel.First];
        end else begin
        //Backword
          ToBeSub := prNeedData[Ext.Last - prHasDataSel.First];
        end;

        Inc(Ext.First, Sense);
        Inc(Ext.Last,  Sense);

        if (Ext.First < Chunk[0].Sel.First) or (Ext.Last > Chunk[0].Sel.Last) then Break;

        Inc(ActualCount);

        prCheckNeedsLoadData(Ext, Sense, DataSize);

        if Sense = +1 then begin
        //Forward
          ToBeAdd := prNeedData[Ext.Last - prHasDataSel.First];
        end else begin
        //Backword
          ToBeAdd := prNeedData[Ext.First - prHasDataSel.First];
        end;


        if NeedSum then begin
        //Re-Sum

          CalculateAdaptiveWindowAvg(@prNeedData[Ext.First - prHasDataSel.First], Ext, (PrevAvg - DWin), (PrevAvg + DWin), ThisAvg, ThisSum, N);
          NeedSum := False;

        end else begin
        //Use Sub/Add

          //Sub First Point ?
          if N > 0 then begin
            if ((PrevAvg - DWin) <= ToBeSub) and ((PrevAvg + DWin) >= ToBeSub) then begin
              ThisSum := ThisSum - ToBeSub;
              Dec(N);
            end;
          end;

          //Add Next Point ?
          if   ((PrevAvg - DWin) <= ToBeAdd) and ((PrevAvg + DWin) >= ToBeAdd) then begin
            ThisSum := ThisSum + ToBeAdd;
            Inc(N);
          end;

          if N > 0 then begin
            ThisAvg := ThisSum / N;
          end;

        end;


        if N >= MIN_INSIDE then begin

          if Abs(ThisAvg - PrevAvg) >= Params.Adaptive.MinDifference then begin
          //Add Node

            Inc(NodeCount);
            if NodeCount > NodeMaxCount then begin
              Inc(NodeMaxCount, NODE_INCREMENT);
              try
                ReallocMem(Nodes,    SizeOf(Integer) * NodeMaxCount);
                ReallocMem(NodesDBl, SizeOf(Integer) * NodeMaxCount);
              except
                Dec(NodeCount);
                Exit;
              end;
            end;
            Nodes   [NodeCount - 1] := (Ext.First + Ext.Last) div 2;
            NodesDBl[NodeCount - 1] := Params.DCOffset - Round(ThisAvg);

            NeedSum := True;
            PrevAvg := ThisAvg;

          end;

        end;

      until(False);//of Ext advance

      Sense := -Sense;

      if Sense = -1 then begin
        NodeCountForward := NodeCount;
      end;

    end;//for sense


  finally

    FreeMem(prNeedData);
    prNeedData := nil;

    if (not prStopped) and (not Error) then begin

      Chunk[0].Status := CHUNK_STATUS_DONE;

      f := False;
      l := False;

      if NodeCount > 0 then begin

        NodeCountBackward := NodeCount - NodeCountForward;
        if NodeCountBackward > 0 then begin
        //Reorder Nodes

          //Make a reversed copy of backwards Nodes
          BackNodes    := nil;
          BackNodesDBl := nil;

          try

            try

              ReallocMem(BackNodes,    SizeOf(Integer) * NodeCountBackward);
              ReallocMem(BackNodesDBl, SizeOf(Integer) * NodeCountBackward);

              for i := 0 to NodeCountBackward - 1 do begin
                BackNodes   [i] := Nodes   [(NodeCount - 1) - i];
                BackNodesDBl[i] := NodesDBl[(NodeCount - 1) - i];
              end;

              for i := NodeCount - 1 downto NodeCount - NodeCountForward do begin
                Nodes   [i] := Nodes   [i - NodeCountBackward];
                NodesDBl[i] := NodesDBl[i - NodeCountBackward];
              end;

              for i := 0 to NodeCountBackward - 1 do begin
                Nodes   [i] := BackNodes   [i];
                NodesDBl[i] := BackNodesDBl[i];
              end;

            except

              NodeCount := NodeCountForward;

            end;

          finally

            FreeMem(BackNodes);
            FreeMem(BackNodesDBl);

          end;

        end;

        OrigNodeCount := NodeCount;
        if Nodes[NodeCount - 1] <> Chunk[0].Sel.Last  then begin
          Inc(NodeCount);
          l := True;
        end;
        if Nodes[0]             <> Chunk[0].Sel.First then begin
          Inc(NodeCount);
          f := True;
        end;
        if (NodeCount <> OrigNodeCount) or (NodeCount <> NodeMaxCount) then begin
          ReallocMem(Nodes,    SizeOf(Integer) * NodeCount);
          ReallocMem(NodesDBl, SizeOf(Integer) * NodeCount);
        end;
        if f then begin
          for i := OrigNodeCount downto 1 do begin
            Nodes   [i] := Nodes   [i - 1];
            NodesDBl[i] := NodesDBl[i - 1];
          end;
          Nodes[0] := Chunk[0].Sel.First;
        end;
        if l then begin
          Nodes   [NodeCount - 1] := Chunk[0].Sel.Last;
          NodesDBl[NodeCount - 1] := NodesDBl[NodeCount - 2];
        end;

      end else begin
      //No nodes found - create the two marginal nodes

        NodeCount := 2;
        ReallocMem(Nodes,    SizeOf(Integer) * NodeCount);
        ReallocMem(NodesDBl, SizeOf(Integer) * NodeCount);

        Nodes[0]    := Chunk[0].Sel.First;
        Nodes[1]    := Chunk[0].Sel.Last;
        NodesDBL[0] := Params.DCOffset;
        NodesDBL[1] := Params.DCOffset;

      end;

      BasChunk.NodeCount := NodeCount;
      BasChunk.Nodes     := Nodes;
      BasChunk.NodesDBL  := NodesDBL;

    end else if (prStopped) or (Error) then begin
    //Error or Stopped

      if prStopped then begin
        Chunk[0].Status := CHUNK_STATUS_STOPPED;
      end else begin
        Chunk[0].Status := CHUNK_STATUS_ERROR;
      end;

      //Reset output
      FreeMem(BasChunk.Nodes);
      FreeMem(BasChunk.NodesDBL);
      BasChunk.NodeCount := 0;
      BasChunk.Nodes     := nil;
      BasChunk.NodesDBL  := nil;

    end;

    prReportTermination;

    Terminate;

  end;

end;


//******************************************************************************
end.

