unit QUB_Thread_Socket;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses

Classes, SysUtils, ScktComp, WinSock, Windows,
AllocTypes, Alloc,

QUB_Socket,
QUB_Thread;



type


TQUB_Thread_Socket_State        = (STATE_UNINITIALIZED,
                                   STATE_EXPECT_SERVER_ID,
                                   STATE_CONNECTED
                                  );

TQUB_Thread_Socket_ThreadState  = (THREADSTATE_UNINITIALIZED,
                                   THREADSTATE_AVAILABLE,
                                   THREADSTATE_ACK_INITIALIZED, THREADSTATE_INITIALIZED,
                                   THREADSTATE_ACK_STARTED, THREADSTATE_STARTED,
                                   THREADSTATE_TERMINATED
                                  );




TQUB_Thread_Socket            = class(TClientWinSocket)
protected


  //Chunk
  prChunk                     :TQUB_Thread_PChunk;

  //Server
  prHost                      :String;
  prPort                      :Word;

  prStopped                   :Boolean;

  //State
  prState                     :TQUB_Thread_Socket_State;
  prThreadState               :TQUB_Thread_Socket_ThreadState;

  //Identification
  prIdentification            :TQUB_Socket_Identification;
  prServerID                  :TQUB_Socket_Identification;

  prLastBlockMsg              :TQUB_Socket_Message;

  //Events
  prOnEvent                   :TQUB_Thread_NotifyEvent;

  prMessageCache              :Pointer;
  prMessageState              :TQUB_Socket_Message_State;
  prMessageSize               :Integer;
  prMessageCurrSize           :Integer;


  procedure prReadCache;

  procedure prRead(var Msg :TQUB_Socket_Message); virtual;
  procedure prWrite(Buffer :Pointer; BufferSize :Integer); virtual;

  procedure prSend_Identification;

  function prValidate_Server(ID :TQUB_Socket_Identification) :Boolean;

  procedure prReportError;

  procedure Error(Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer); override;
  procedure Event(Socket: TCustomWinSocket; SocketEvent: TSocketEvent); override;



public


  constructor Create(const RemoteServer :String; Port :Word); virtual;

  procedure Open;

  procedure SetEvent; virtual; abstract;

  property Chunk             :TQUB_Thread_PChunk            read prChunk        write prChunk;

  property OnEvent           :TQUB_Thread_NotifyEvent       read prOnEvent      write prOnEvent;


end;



implementation


//******************************************************************************
constructor TQUB_Thread_Socket.Create(const RemoteServer :String; Port :Word);
begin

  inherited Create(INVALID_SOCKET);

  prIdentification.Task   := QUB_SOCKET_ID_TASK_NULL;
  prIdentification.Thread := QUB_SOCKET_ID_THREAD_NULL;

  prHost                  := RemoteServer;
  prPort                  := Port;

  prStopped               := False;

  prState                 := STATE_UNINITIALIZED;
  prThreadState           := THREADSTATE_UNINITIALIZED;

end;



//****************************************************************************//
//                                                                            //
//                                 O P E N                                    //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.Open;
begin

  try
    inherited Open(prHost, '', '', prPort, False{Block});
  except
  end;

end;



//****************************************************************************//
//                                                                            //
//                                E V E N T                                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.Event(Socket: TCustomWinSocket; SocketEvent: TSocketEvent);
begin

  case SocketEvent of

    //Has connected to server
    seConnect:
      begin
        prState := STATE_EXPECT_SERVER_ID;
        prSend_Identification;
      end;

    //Has disconnected from server
    seDisconnect:
      begin
        //prReportError;
      end;

    //Something in the pipeline
    seRead:
      begin
        while prMessageState = MSGSTATE_UPDATING do begin
        //Wait to finish updating
        end;
        prReadCache;
      end;

  end;//SocketEvent

  inherited;

end;



//****************************************************************************//
//                                                                            //
//                            R E A D     C A C H E                           //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.prReadCache;
var
  PMsg :Pointer;
  SMsg :Integer;
begin


  try

    case prMessageState of

      MSGSTATE_UNINITIALIZED:
        begin

          prMessageState := MSGSTATE_UPDATING;

          try
            SMsg := ReceiveLength;
            PMsg := AllocMem(SMsg);
          except
            SMsg := SizeOf(TQUB_Socket_Message);
            PMsg := AllocMem(SMsg);
          end;

          try
            ReceiveBuf(PMsg^, SMsg);
          except
            //Error;
          end;

          prMessageCurrSize := SMsg;
          prMessageSize     := TQUB_Socket_PMessage(PMsg).Size;
          prMessageCache    := AllocMem(prMessageSize);

          try
            Move(PMsg^, prMessageCache^, SMsg);
          finally
            FreeMem(PMsg);
          end;

        end;

      MSGSTATE_READING:
        begin

          prMessageState := MSGSTATE_UPDATING;

          try
            SMsg := ReceiveLength;
            PMsg := AllocMem(SMsg);
          except
            SMsg := SizeOf(TQUB_Socket_Message);
            PMsg := AllocMem(SMsg);
          end;

          try
            ReceiveBuf(PMsg^, SMsg);
          except
            //Error;
          end;

          Move(PMsg^, TPAByte(prMessageCache)[prMessageCurrSize], SMsg);
          Inc(prMessageCurrSize, SMsg);

          try
            Move(PMsg^, prMessageCache^, SMsg);
          finally
            FreeMem(PMsg);
          end;

        end;

    end;//case MessageState

  finally

    if prMessageCurrSize >= prMessageSize then begin

      try

        prRead(TQUB_Socket_Message(prMessageCache^));

      finally

        FreeMem(prMessageCache);
        prMessageCache    := nil;
        prMessageSize     := 0;
        prMessageCurrSize := 0;
        prMessageState    := MSGSTATE_UNINITIALIZED;

      end;

    end else begin

      prMessageState := MSGSTATE_READING;

    end;

  end;

end;



//****************************************************************************//
//                                                                            //
//                                 R E A D                                    //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.prRead(var Msg :TQUB_Socket_Message);
begin

  if prState = STATE_UNINITIALIZED then begin
  //This State does not allow any reading

    //Error

  end else if prState = STATE_EXPECT_SERVER_ID then begin
  //In this State, the only possible Message is QUB_SOCKET_MESSAGE_SERVER_ID

    if Msg.Message = QUB_SOCKET_MESSAGE_SERVER_ID then begin

      if prValidate_Server(TQUB_Socket_PIdentification(@Msg.Data)^) then begin
      //Server validated
      //From this moment on, the messages will be processed by the descendent class

        prState       := STATE_CONNECTED;
        prThreadState := THREADSTATE_AVAILABLE;

      end else begin
      //Unsupported, Error

        //Error

      end;

    end else begin

      //Error

    end;

  end;

end;



//****************************************************************************//
//                                                                            //
//                   S E N D     I D E N T I F I C A T I O N                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.prSend_Identification;
var
  Msg :TQUB_Socket_Message;
begin

  Msg.Message := QUB_SOCKET_MESSAGE_CLIENT_ID;
  Msg.Size    := SizeOf(TQUB_Socket_Message);
  TQUB_Socket_PIdentification(@Msg.Data)^ := prIdentification;

  prWrite(@Msg, SizeOf(Msg));

end;



//****************************************************************************//
//                                                                            //
//                      V A L I D A T E     S E R V E R                       //
//                                                                            //
//****************************************************************************//
function TQUB_Thread_Socket.prValidate_Server(ID :TQUB_Socket_Identification) :Boolean;
begin

  Result := False;

  if (ID.Task = QUB_SOCKET_ID_TASK_SERVER) and (ID.Thread = QUB_SOCKET_ID_THREAD_SERVER) then begin

    Result := True;

  end;

  if Result then begin

    prServerID := ID;

  end;

end;


//****************************************************************************//
//                                                                            //
//                                E R R O R                                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.Error(Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin

  inherited;

  ErrorCode := 0;

  prReportError;

end;



//****************************************************************************//
//                                                                            //
//                                W R I T E                                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.prWrite(Buffer :Pointer; BufferSize :Integer);
var
  First, Last :Integer;
  Packet      :Integer;
  p           :TPAByte;
begin

  Packet := ReceiveBuf(nil^, -1);
  if Packet <= 0 then Packet := 8192;
  if Packet > BufferSize then Packet := BufferSize;

  First := 0;
  p     := Buffer;

  repeat

    Last := First + Packet - 1;
    if Last > (BufferSize - 1) then Last := (BufferSize - 1);

    repeat

      if SendBuf(p[First], (Last - First + 1)) <> -1 then begin
        Break;
      end else begin
        Sleep(10);
      end;

    until(prStopped);

    First := Last + 1;

  until((prStopped) or (First > (BufferSize - 1)));

end;




//****************************************************************************//
//                                                                            //
//                          R E P O R T    E R R O R                          //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket.prReportError;
var
  Rec :TQUB_Thread_Event_PRec;
begin

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_Rec));

    //Set the Event
    Rec.Thread := TQUB_Thread(Self);
    Rec.Event  := THREAD_EVENT_REPORT_ERROR;

    prOnEvent(Self, Rec);

  end;

end;

//******************************************************************************
end.




