unit QUB_Version;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, Forms, StdCtrls;


type


TVersionResources = (vrCompanyName, vrFileDescription, vrFileVersion,
                     vrInternalName, vrLegalCopyright, vrOriginalFilename,
                     vrProductName, vrProductVersion, vrComments, vrFlags);


TQUB_Version = class(TLabel)
private


  prVersionResource    :TVersionResources;
  prInfoPrefix         :String;
  prShowInfoPrefix     :Boolean;
  prVersionResourceKey :String;
  prLangCharset        :String;


  procedure prSetupInfoPrefix;
  procedure prSetupResourceKey;
  function prGetStringFileInfo(Buffer: Pchar; size: integer) :String;
  function prGetFixedFileInfo(Buffer: PChar; size: integer) :String;
  function prGetInfo :String;
  procedure prSetupCaption;


protected


  procedure prSetInfoPrefix(Value :String);
  function prGetInfoPrefix :String;
  procedure prSetVersionResource(Value :TVersionResources);
  procedure prSetShowInfoPrefix(Value :Boolean);
  procedure prSetVersionResourceKey(Value :String);
  procedure prSetLangCharset(Value :String);


public


  constructor Create(AOwner :TComponent); override;


published


  property VersionResource     :TVersionResources read prVersionResource    write prSetVersionResource;
  property VersionResourceKey  :String            read prVersionResourceKey write prSetVersionResourceKey;
  property InfoPrefix          :String            read prGetInfoPrefix      write prSetInfoPrefix;
  property ShowInfoPrefix      :Boolean           read prShowInfoPrefix     write prSetShowInfoPrefix;
  property LangCharset         :String            read prLangCharset        write prSetLangCharset;


end;


//******************************************************************************
const

//The order of this array must be the same as the VersionResources enum type as that is used for the index lookup

VersionLookup: array[TVersionResources, 0..1] of string = (
                    ('CompanyName', 'Company Name:'),
                    ('FileDescription', 'File Description:'),
                    ('FileVersion', 'File Version:'),
                    ('InternalName', 'Internal Name:'),
                    ('LegalCopyright', 'Legal Copyright:'),
                    ('OriginalFilename', 'Original Filename:'),
                    ('ProductName', 'Product Name:'),
                    ('ProductVersion', 'Product Version:'),
                    ('Comments', 'Comments:'),
                    ('Flags', 'Flags:'));

procedure Register;


implementation


//******************************************************************************
procedure Register;
begin
  RegisterComponents('QUB', [TQUB_Version]);
end;


//******************************************************************************
constructor TQUB_Version.Create(AOwner :TComponent);
begin
  inherited Create(AOwner);
  WordWrap := False;
  Autosize := True;
  ShowInfoPrefix := True;
  LangCharset := '-1'; //-1 = auto detect
  VersionResource := vrFileVersion;
end;

procedure TQUB_Version.prSetVersionResource(Value: TVersionResources);
begin
  prVersionResource := Value;
  prSetupResourceKey;
  prSetupInfoPrefix;
end;

procedure TQUB_Version.prSetupInfoPrefix;
var
  s :String;
begin
  s := VersionLookup[prVersionResource, 1];
  InfoPrefix := s;
end;

procedure TQUB_Version.prSetupResourceKey;
var
  s :String;
begin
  s := VersionLookup[prVersionResource, 0];
  VersionResourceKey := s;
end;

function TQUB_Version.prGetFixedFileInfo(Buffer :PChar; size :Integer) :String;
var
  ValLen        :Integer;
  FixedFileInfo :PVSFixedFileInfo;
begin
  if VerQueryValue(buffer, '\', Pointer(FixedFileInfo), Longword(ValLen)) then begin
    Result := '';
    if (ValLen > 1) then begin
      if FixedFileInfo.dwFileFlags and VS_FF_DEBUG        <> 0 then Result := Result+', Debug Build';
      if FixedFileInfo.dwFileFlags and VS_FF_PRERELEASE   <> 0 then Result := Result+', Pre-Release Build';
      if FixedFileInfo.dwFileFlags and VS_FF_PATCHED      <> 0 then Result := Result+', Patched Build';
      if FixedFileInfo.dwFileFlags and VS_FF_PRIVATEBUILD <> 0 then Result := Result+', Private Build';
      if FixedFileInfo.dwFileFlags and VS_FF_INFOINFERRED <> 0 then Result := Result+', InfoInferred';
      if FixedFileInfo.dwFileFlags and VS_FF_SPECIALBUILD <> 0 then Result := Result+', Special Build';
      if Result <> '' then begin
        if Result[1] = ',' then Delete(Result, 1, 2);
      end;
    end;
  end else begin
    Result := '< Error retrieving version info >';
  end
end;

function TQUB_Version.prGetStringFileInfo(Buffer :Pchar; size :Integer) :String;
var
  Vallen, Translen            :Integer;
  VersionPointer, TransBuffer :PChar;
  Temp                        :Integer;
  CalcLangCharSet             :String;
begin
  if prLangCharSet = '-1' then begin
    VerQueryValue(Buffer, '\VarFileInfo\Translation', Pointer(TransBuffer), LongWord(TransLen));
    if TransLen >= 4 then begin
      StrLCopy(@Temp, TransBuffer, 2);
      CalcLangCharSet := IntToHex(Temp, 4);
      StrLCopy(@Temp, TransBuffer + 2, 2);
      CalcLangCharSet := CalcLangCharSet + IntToHex(Temp, 4);
      prLangCharSet := CalcLangCharSet;
    end else begin
      Result := '< Error retrieving translation info >';
      Exit;
    end;
  end;
  if VerQueryValue(Buffer, PChar('\StringFileInfo\' + prLangCharSet + '\' + VersionResourceKey), Pointer(VersionPointer), LongWord(Vallen)) then begin
    if (Vallen > 1) then begin
      SetLength(Result, Vallen);
      StrLCopy(PChar(Result), VersionPointer, Vallen);
    end else begin
      Result := '< No Version Info >';
    end;
  end else begin
    Result := '< Error retrieving version info >';
  end;
end;

function TQUB_Version.prGetInfo :String;
var
  Dump, Size :Integer;
  Buffer     :PChar;
begin
  if csDesigning in Self.ComponentState then begin
    Result := '< No design info >';
  end else begin
    Size := GetFileVersionInfoSize(PChar(Application.ExeName), LongWord(dump));
    if Size = 0 then begin
      Result := '< No Data Available >';
    end else begin
      Buffer := StrAlloc(Size + 1);
      try
        if not GetFileVersionInfo(Pchar(Application.Exename), 0, Size, Buffer) then begin
          Result := '< Error retrieving version info >'
        end else begin
          if prVersionResource = vrFlags then begin
            Result := prGetFixedFileInfo(Buffer, Size)
          end else begin
            Result := prGetStringFileInfo(Buffer, Size);
          end;
        end;
      finally
        StrDispose(Buffer);
      end;
    end;
  end;
  if ShowInfoPrefix then begin
    Result := InfoPrefix + ' ' + Result;
  end;
end;

procedure TQUB_Version.prSetInfoPrefix(Value :String);
begin
  if prInfoPrefix = Value then Exit;
  prInfoPrefix := Value;
  //The caption needs to be recalculated everytime the prefix is changed, otherwise the detaults override the user specified one
  prSetupCaption;
end;

procedure TQUB_Version.prSetVersionResourceKey(Value :String);
begin
  if prVersionResourceKey = Value then Exit;
  prVersionResourceKey := Value;
  InfoPrefix := Value;
end;

function TQUB_Version.prGetInfoPrefix :String;
begin
  Result := prInfoPrefix;
end;

procedure TQUB_Version.prSetShowInfoPrefix(Value :Boolean);
begin
  if prShowInfoPrefix = Value then Exit;
  prShowInfoPrefix := Value;
  prSetupCaption;
end;

procedure TQUB_Version.prSetLangCharset(Value :String);
begin
  if prLangCharSet = Value then Exit;
  prLangCharSet := Value;
  prSetupCaption;
end;

procedure TQUB_Version.prSetupCaption;
begin
  Caption := prGetInfo;
end;


//******************************************************************************
end.
