"""General-purpose UI components."""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import pygtk
pygtk.require('2.0')
import gtk
from gtk import gdk
from gtk import keysyms
import gobject
import pango
import traceback
import sys
import datetime
from math import *
from qub.types import *
from qub.accept import *

class InputDialog(gtk.Dialog):
    """InputDialog(title='', parent=None, message='', text='', buttons={'Cancel', gtk.RESPONSE_REJECT, 'OK', gtk.RESPONSE_ACCEPT) ):
    Prompts the user to write or edit a string.

    title      title of the dialog
    parent     parent gtk.Window, for centering
    message    label to the left of the string
    text       initial contents of the string
    buttons    same as the gtk.Dialog constructor argument
"""
    def __init__(self, title='', parent=None, message='', text='', buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                                            gtk.STOCK_OK, gtk.RESPONSE_ACCEPT) ):
        gtk.Dialog.__init__(self, title, parent, gtk.DIALOG_MODAL, buttons=buttons)
        self.set_default_response(gtk.RESPONSE_ACCEPT)
        hbox = gtk.HBox()
        lbl = gtk.Label(message)
        hbox.pack_start(lbl)
        lbl.show()
        self.entry = gtk.Entry()
        hbox.pack_start(self.entry)
        self.entry.show()
        self.text = text
        self.entry.set_text(text)
        self.vbox.pack_start(hbox)
        hbox.show()
    def run(self):
        response = gtk.Dialog.run(self)
        self.text = self.entry.get_text()
        return response

MAX_EVENTS_PER_WAIT = 5

class BusyDialog(gtk.Dialog):
    """BusyDialog(title, func, stopCoords=None, parent=None, maxEventsPerUpdate=5):
    Modal dialog with progress bar and stop button; supervises a long-running computation.

    title:              text for the title bar
    func:               long-running computation func(update)
                        should periodically call update( fraction )
                        and stop immediately if update returns False
    stopCoords          screen coordinates of the Stop button, for repositioning
    parent              parent gtk.Window
    maxEventsPerUpdate  process up to this many gui events each time you call update
"""
    def __init__(self, title, func, stopCoords=None, parent=None, maxEventsPerUpdate=MAX_EVENTS_PER_WAIT):
        # inside your long-task callback:
        #    func( update )
        # call the update function regularly to give progress and find out if stopped.
        #    update( fraction ) -> bool should_continue
        gtk.Dialog.__init__(self, title, parent, gtk.DIALOG_MODAL)
        self.prog = gtk.ProgressBar()
        self.vbox.pack_start(self.prog)
        self.prog.show()
        self.btnStop = gtk.Button('Stop')
        self.btnStop.connect('clicked', self.onStop)
        self.action_area.pack_end(self.btnStop)
        self.btnStop.show()
        self.func = func
        self.maxEvents = maxEventsPerUpdate
        self._stopped = False
        self.stopCoords = stopCoords
        self.realize()
    def run(self):
        """Shows the dialog and runs func."""
        self.show()
        gobject.idle_add(self.start)
        result = gtk.Dialog.run(self)
        if self.exc:
            raise self.exc
        return result
    def start(self):
        if not (self.stopCoords is None):
            x, y = self.stopCoords
            xb, yb = self.btnStop.translate_coordinates(self, 0, 0)
            self.move(x-xb, y-yb)
            # print x,xb, y,yb
        try:
            self.exc = None
            self.func(self.update)
        except Exception, e:
            self.exc = e
        if self._stopped:
            self.response(gtk.RESPONSE_CANCEL)
        else:
            self.response(gtk.RESPONSE_OK)
    def onStop(self, widget):
        self._stopped = True
    def update(self, fraction):
        self.prog.set_fraction(min(1.0, fraction))
        for i in xrange(self.maxEvents):
            if gtk.events_pending():
                gtk.main_iteration(False)
            else:
                break
        return not self._stopped

HARMLESS_KEYS = [keysyms.Left, keysyms.Right, keysyms.Up, keysyms.Down, keysyms.Home, keysyms.End, keysyms.Insert]

class NumEntry(gtk.Entry):
    """NumEntry(value, accept=None, format='%s'):
    Originally intended for numbers, this gtk.Entry can parse, validate and display
any suitable data type.

    value        initial value
    accept       f(string) returns value or raises Exception
                 see qub.accept for stock examples
                 default is str
    format       either a %-style format string
                 or f(value) -> string

Properties:
    accept       as above
    value        the last accepted value
    text         the last accepted string
    cGood        gdk.Color of accepted text
    cDiff        gdk.Color of not-yet-accepted text
    good         whether the text is accepted (what color to use)

Events:
    OnEdit(NumEntry, txt)        user has typed, text is now txt
    OnChange(NumEntry, val)      text accepted; new value is val
    OnExit(NumEntry)             focus-out
    OnReject(NumEntry, txt, exc) txt not accepted due to exception exc
"""    
    def __init__(self, value, accept=None, format='%s'):
        gtk.Entry.__init__(self)
        # self.set_size_request(80, -1)
        self.format = acceptFormat(format)
        self.connect("key_press_event", self.onKey)
        self.connect("focus_out_event", self.onExit)
        self.OnEdit = WeakEvent()   # OnEdit(NumEntry, txt)
        self.OnChange = WeakEvent() # OnChange(NumEntry, val)
        self.OnExit = WeakEvent()   # OnExit(NumEntry)
        self.OnReject = WeakEvent() # OnReject(NumEntry, txt, exception)
        self.OnReset = WeakEvent()  # OnReset(NumEntry, val)
        self._cGood = gdk.color_parse("#000000")
        self._cDiff = gdk.color_parse("#ff0000")
        self._good = True
        self._accept = accept or (isinstance(value, str) and str) or (lambda s: type(value)(eval(s)))
        self.value = value
    def setAccept(self, accept):
        self._accept = accept
        self.onParse()
    accept = property(lambda s: s._accept, setAccept)
    def setValue(self, value, repr=None, good=True):
        self._value = value
        if repr == None:
            self._str = self.format(value)
        else:
            self._str = repr
        self.good = good
        self.set_text(self._str)
    value = property(lambda s: s._value, setValue)
    text = property(lambda s: s._str)
    def setGoodColor(self, c):
        self._cGood = c
        self._good = not self._good # force update
        self.good = not self.good
    cGood = property(lambda s: s._cGood, setGoodColor)
    def setDiffColor(self, c):
        self._cDiff = c
        self._good = not self._good # force update
        self.good = not self.good
    cDiff = property(lambda s: s._cDiff, setDiffColor)
    def setGood(self, good):
        if self._good != good:
            self._good = good
            self.modify_text(gtk.STATE_NORMAL, self._good and self._cGood or self._cDiff)
            self.modify_text(gtk.STATE_SELECTED, self._good and self._cGood or self._cDiff)
    good = property(lambda s: s._good, setGood)
    def onKey(self, txt, event):
        if event.keyval == keysyms.Escape:
            self.set_text(self._str)
            self.select_region(0, -1)
            self.good = True
            self.OnReset(self, self.value)
        elif event.keyval == keysyms.Return:
            if self.get_text() == self._str:
                self.OnChange(self, self.value)
                self.good = True
            else:
                self.onParse()
        elif (not (event.keyval in HARMLESS_KEYS)):
            self.OnEdit(self, self.get_text())
            self.good = False
    def onExit(self, txt, event):
        self.onParse()
        self.OnExit(self)
    def onParse(self):
        s = self.get_text()
        if s == self._str:
            self.good = True
            self.OnReset(self, self.value)
        else:
            try:
                self.setValue(self.accept(s), s)
                self.OnChange(self, self.value)
                self.good = True
                self.select_region(0, -1)
            except Exception, e:
                self.OnReject(self, s, e)
                # self.set_text(self._str) # self.format(self.value))

class V2Scale(gtk.Frame):
    """V2Scale(octaves=4, divs=8, minor_zoom=8):
    Widget with two logarithmic sliders: coarse and fine.

    octaves:    coarse ranges from 1/(2^octaves) to 2^octaves
    divs:       number of sub-divisions per octave
    minor_zoom: fine range is 1/minor_zoom of coarse range

Events:
    OnScale(V2Scale, factor)    dragging slider, now at factor
    OnSnap(V2Scale)             released slider, it pops back to center
"""
    def __init__(self, octaves=4, divs=8, minor_zoom=8):
        gtk.Frame.__init__(self)
        self.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
        self.set_size_request(-1, 50)
        self.octs = octaves
        self.divs = divs
        self.zoom = minor_zoom
        self.onScale = WeakEvent() # onScale(V2Scale, factor)
        self.onSnap = WeakEvent() # onSnap(V2Scale)
        # emit 2^(value/(divs*zoom))
        maxval = octaves * divs
        h = gtk.HBox(True, 0)
        self.maj = gtk.Adjustment( value=0, lower=-maxval, upper=maxval, step_incr=1, page_incr=divs, page_size=1 )
        self.maj.connect("value_changed", self.onSlide, 1)
        self.majs = gtk.VScale(self.maj)
        self.majs.connect("button_release_event", self.onEndSlide, self.maj)
        self.majs.set_draw_value(False)
        h.pack_start(self.majs, True, True)
        self.majs.show()
        self.min = gtk.Adjustment( value=0, lower=-maxval, upper=maxval, step_incr=1, page_incr=divs, page_size=1 )
        self.min.connect("value_changed", self.onSlide, minor_zoom)
        self.mins = gtk.VScale(self.min)
        self.mins.connect("button_release_event", self.onEndSlide, self.min)
        self.mins.set_draw_value(False)
        h.pack_start(self.mins, True, True)
        self.mins.show()
        self.snapState = 0
        self.sliding = False
        self.add(h)
        h.show()
        # we want it to snap back to 0 when released, emitting no further events,
        # so we have to ignore some inconvenient events:
        # snapState: 0: normal (sliding or not)
        #            1: button-up (ignoring our setting to 0)
        #            2: overriding bogus setting from mouse-up
        #            3: ignoring our override
        # TODO: this style disables pageup/dn single-clicks
    def onSlide(self, adj, zoom):
        if self.snapState == 0: 
            self.sliding = True
            self.onScale(self, pow(2.0, - adj.value / (self.divs * zoom)))
        elif self.snapState == 1:
            self.snapState = 2
        elif self.snapState == 2:
            adj.value = 0
            self.snapState = 3
        else:
            self.onSnap(self)
            self.snapState = 0
            self.sliding = False
    def onEndSlide(self, slide, evt, adj):
        self.snapState = 1
        adj.set_value(0)
    def reverse(self, factor):
        """If factor is modifying a negative number, you want up to mean abs(smaller)."""
        if factor == 0.0: return 0.0
        power = log(factor) / log(2.0)
        return pow(2.0, - power)

class NumSliders(gtk.VBox):
    """NumSliders(format='%s'):
    Widget containing V2Scale and NumEntry, for kbd and log-slider editing of a number.

    format:   see NumEntry

Events:
    OnChange(val)

Properties:
    txt       the NumEntry
    slide     the V2Scale
    center    value corresponding to the central slider position;
              center == txt.value except when dragging a slider
"""
    def __init__(self, format="%s"):
        gtk.VBox.__init__(self, False, 0)
        self.slide = V2Scale()
        self.hOnScale = self.onScale
        self.slide.onScale += self.hOnScale,"onScale"
        self.hOnSnap = self.onSnap
        self.slide.onSnap += self.hOnSnap,"onSnap"
        self.pack_start(self.slide, True, True)
        self.slide.show()
        self.center = 0.0
        self.txt = NumEntry(self.center, format=format)
        self.txt.set_width_chars(7)
        self.hOnNumChange = self.onNumChange
        self.txt.OnChange += self.hOnNumChange,"onNumChange"
        self.pack_start(self.txt, False, True)
        self.txt.show()
        self.OnChange = WeakEvent() # (value)
    def set_value(self, value):
        if not self.slide.sliding:
            self.txt.value = value
            self.center = value
            # self.OnChange(value)
    def onScale(self, scale, factor):
        if self.center == 0.0:
            self.txt.value = factor - 1.0
        elif self.center < 0:
            self.txt.value = scale.reverse(factor) * self.center
        else:
            self.txt.value = factor * self.center
        self.OnChange(self.txt.value)
    def onSnap(self, scale):
        self.center = self.txt.value
    def onNumChange(self, txt, value):
        self.center = value
        self.OnChange(value)


class DateBox(gtk.HBox):
    """Widget for entering/editing legitimate datetime.datetime values.

Events:
    OnChange(DateBox, date)

Properties
    date        datetime.datetime

"""
    def __init__(self, *args, **kw):
        gtk.HBox.__init__(self, *args, **kw)
        self.day = NumEntry(1, self._acceptDay, "%d")
        self.hOnChangeDay = self._onChangeDay
        self.day.OnChange += self.hOnChangeDay
        self.day.set_width_chars(2)
        self.pack_start(self.day, False, False)
        self.day.show()
        self.month = gtk.combo_box_new_text()
        for m in xrange(1, 13):
            self.month.append_text(datetime.datetime(1999, m, 1).strftime("%b"))
        self.month.connect('changed', self._onChangeMonth)
        self.pack_start(self.month, False, False)
        self.month.show()
        self.year = NumEntry(1999, int, "%d")
        self.hOnChangeYear = self._onChangeYear
        self.year.OnChange += self.hOnChangeYear
        self.year.set_width_chars(4)
        self.pack_start(self.year, False, False)
        self.year.show()
        self.OnChange = WeakEvent() # (DateBox, datetime)
        self.set_date(datetime.datetime.now())
    def set_date(self, date):
        self._updating = True
        self._date = date
        self.day.setValue(date.day)
        self.month.set_active(date.month-1)
        self.year.setValue(date.year)
        self._updating = False
    date = property(lambda s: s._date, set_date)
    def _acceptDay(self, s):
        self.date.replace(day=int(s))
        return int(s)
    def _onChangeDay(self, txt, val):
        if self._updating: return
        self.date = self.date.replace(day=val)
        self.OnChange(self, self.date)
    def _onChangeMonth(self, mnu):
        if self._updating: return
        while True: # decrement day until it fits into new month
            try:
                self.date = self.date.replace(month=mnu.get_active()+1)
                self.OnChange(self, self.date)
                return
            except:
                self._updating = True
                self.date = self.date.replace(day=self.date.day-1)
                self.day.setValue(self.date.day)
                self._updating = False
    def _onChangeYear(self, txt, val):
        if self._updating: return
        self.date = self.date.replace(year=val)
        self.OnChange(self, self.date)


class DynamicComboBox(gtk.EventBox):
    """DynamicComboBox():
    A menu which is re-built each time it drops down.

Events:
    OnPopulate(add)        when dropping down; call add(item_text) for each desired item
    OnChanged(DynamicComboBox, active_text)
                           when a different choice is made

Properties:
    active_i         index of the chosen item
    active_text      text of the chosen item
"""
    def __init__(self):
        gtk.EventBox.__init__(self)
        self._box = gtk.combo_box_new_text()
        self.add(self._box)
        self._box.show()
        self._changed_handler = self._box.connect('changed', self.combo_changed)
        self.OnPopulate = WeakEvent() # add(item-text)
        self.OnChanged = WeakEvent()  # ComboBox, active-text
        self._count = 0
        self._active_text = ''
        self._active_i = -1
        self.indexOf = {}
        self.connect('enter-notify-event', self.enter_notify_event)
    active_i = property(lambda s: s._active_i)
    active_text = property(lambda s: s._active_text)
    def _add(self, text):
        self._box.append_text(text)
        self._count += 1
        if text == self._active_text:
            self._box.set_active(self._count - 1)
        self.indexOf[text] = self._count - 1
    def choose(self, txt):
        """Selects the item with txt, adding it if necessary."""
        self._box.handler_block(self._changed_handler)
        if not self.indexOf.has_key(txt):
            self._add(txt)
        self._box.set_active(self.indexOf[txt])
        self._box.handler_unblock(self._changed_handler)
    def enter_notify_event(self, widget, event):
        self.repopulate()
    def repopulate(self):
        self._active_text = self._box.get_active_text()
        self._box.handler_block(self._changed_handler)
        for i in xrange(self._count):
            self._box.remove_text(0)
        self._count = 0
        self.indexOf = {}
        self.OnPopulate(self._add)
        self._box.handler_unblock(self._changed_handler)
    def combo_changed(self, combo):
        active_i = combo.get_active()
        if active_i < 0:
            combo.set_active(self._active_i)
        else:
            self._active_i = active_i
            self._active_text = combo.get_active_text()
            self.OnChanged(combo, self._active_text)

class SimpleList(gtk.TreeView):
    """SimpleList(header=None, sortable=True):
    A gtk.TreeView specialized as a single-select single-column list.

    header      caption for the top row
    sortable    default True: the user can click header to sort

Events:
    OnSelect(SimpleList, ix)     selected item graphically changed to ix'th
    OnDblClick(SimpleList, ix)   ix'th item was double-clicked (or <enter> key)

Properties:
    popup           a gtk.Menu if they right-click, default None
    model           gtk.ListStore
    index           (unsorted) index of selected item
    active          text of the selected item, or None
"""
    def __init__(self, header=None, sortable=True):
        gtk.TreeView.__init__(self)
        self.set_headers_visible(bool(header))
        self.get_selection().set_mode(gtk.SELECTION_SINGLE)
        self._index = -1 # pre-sorted
        self.OnSelect = WeakEvent() # SimpleList, ix # graphical select only
        self.OnDblClick = WeakEvent() # SimpleList, ix
        self.connect("cursor_changed", self._onCursorChanged)
        self.connect("row_activated", self._onRowActivated)
        self.connect('button_press_event', self._onButtonPress)
        self._popup = None
        self._model = gtk.ListStore(str)
        self._smodel = gtk.TreeModelSort(self._model)
        self.set_model(self._smodel)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(header or "", renderer, text=0)
        if sortable:
            column.set_sort_column_id(0)
            self._smodel.set_sort_column_id(0, gtk.SORT_ASCENDING)
        self.append_column(column)
    popup = property(lambda s: s._popup, lambda s,x: s._setPopup(x))
    def _setPopup(self, x):
        self._popup = x
    model = property(lambda s: s._model)
    index = property(lambda s: s._index, lambda s,x: s._setIndex(x))
    def _setIndex(self, ix):
        try:
            if ix >= 0:
                self.set_cursor(self._smodel.convert_child_path_to_path((ix,)))
            self._index = ix
            # self.get_selection().unselect_all()
            if ix >= 0:
                self.get_selection().select_path(self._smodel.convert_child_path_to_path((ix,)))
            else:
                self.get_selection().unselect_all()
        except:
            traceback.print_exc()
    active = property(lambda s: s.getActive(), lambda s,x: s.setActive(x))
    def getActive(self):
        if self._index >= 0:
            return self._model.get(self._model.get_iter((self._index,)), 0)[0]
        else:
            return None
    def setActive(self, x):
        self.index = self.indexOf(x)
    def indexOf(self, x):
        """Returns the index of the first item with text x"""
        iter = self._model.get_iter_first()
        i = 0
        while iter:
            if self._model.get(iter, 0)[0] == x:
                return i
            iter = self._model.iter_next(iter)
            i += 1
        return -1
    def clear(self):
        """Removes all items from the list."""
        self._model.clear()
        self.index = -1
        self.OnSelect(self, self.index)
    def append(self, x):
        """Adds x at the end of the list."""
        self._model.append((x,))
    def edit(self, index, newtext):
        """Replaces the text at index with newtext."""
        self._model.set(self._model.get_iter((index,)), 0, newtext)
    def remove(self, index):
        """Removes the item at index."""
        self._model.remove(self._model.get_iter((index,)))
        if index == self.index:
            if index == len(self._model):
                index -= 1
            self.index = index
            self.OnSelect(self, self.index)
    def select(self, i, s=True):
        """Selects the item at index i, or unselects if s==False."""
        path = self._smodel.convert_child_path_to_path((i,))
        if s:
            self.get_selection().select_path(path)
        else:
            self.get_selection().unselect_path(path)
    def _onCursorChanged(self, tree):
        sel = self.get_selection()
        smodel, iter = sel.get_selected()
        if iter:
            ix = smodel.convert_path_to_child_path(smodel.get_path(iter))[0]
            if ix != self._index:
                self._index = ix
                self.OnSelect(self, self._index)
        else:
            if self._index >= 0:
                self._index = -1
                self.OnSelect(self, -1)
    def _onRowActivated(self, tree, spath, col):
        path = self._smodel.convert_path_to_child_path(spath)
        if self._index != path[0]:
            self._index = path[0]
            self.OnSelect(self, self._index)
        self.OnDblClick(self, self._index)
    def _onButtonPress(self, view, event):
        if self._popup and event.button == 3:
            spath, col, cellx, celly = view.get_path_at_pos(int(event.x), int(event.y))
            path = self._smodel.convert_path_to_child_path(spath)
            # just in case the view doesn't alreay have focus
            view.grab_focus()
            selection = view.get_selection()
            # if this row isn't already selected, then select it before popup
            if not selection.path_is_selected(spath):
                view.set_cursor( spath, col, 0)
                self._index = path[0]
            self._popup.popup(None, None, None, event.button, event.time)
            return True
    
class MultiList(gtk.TreeView):
    """MultiList(header=None, sortable=True):
    A gtk.TreeView specialized as a multi-select, single-column list.

    header      caption for the top row
    sortable    default True: the user can click header to sort

Properties:
    model           gtk.ListStore
    count           len(model)
"""
    def __init__(self, header=None, sortable=True):
        gtk.TreeView.__init__(self)
        self.set_rubber_banding(True)
        self.set_headers_visible(bool(header))
        self.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
        self._count = 0
        self._model = gtk.ListStore(str)
        self._smodel = gtk.TreeModelSort(self._model)
        self.set_model(self._smodel)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(header or "", renderer, text=0)
        if sortable:
            column.set_sort_column_id(0)
            self._smodel.set_sort_column_id(0, gtk.SORT_ASCENDING)
        self.append_column(column)
    model = property(lambda s: s._model)
    count = property(lambda s: s._count)
    def sortedToUn(self, ix):
        """Returns the unsorted index of the sorted ix'th item."""
        return self._smodel.convert_path_to_child_path((ix,))[0]
    def unToSorted(self, ix):
        """Returns the sorted index of the unsorted ix'th item."""
        return self._smodel.convert_child_path_to_path((ix,))[0]
    def _setIndex(self, ix):
        self._index = ix
        # self.get_selection().unselect_all()
        if ix >= 0:
            self.get_selection().select_path((self.unToSorted(ix),))
        else:
            self.get_selection().unselect_all()
    def indexOf(self, x):
        """Returns the unsorted index of the first item with text x."""
        iter = self._model.get_iter_first()
        i = 0
        while iter:
            if self._model.get(iter, 0)[0] == x:
                return i
            iter = self._model.iter_next(iter)
            i += 1
        return -1
    def clear(self):
        """Removes all items from the list."""
        self._model.clear()
        self._count = 0
    def append(self, x, s=True):
        """Adds item x to the end of the list; selects it if s == True."""
        self._model.append((x,))
        self._count += 1
        self.select(self._count-1, s)
    def selected(self, i):
        """Returns True if unsorted item i if selected."""
        return self.get_selection().path_is_selected((self.unToSorted(i),))
    def select(self, i, s=True):
        """Selects unsorted item i; or unselects if s == False."""
        if s:
            self.get_selection().select_path((self.unToSorted(i),))
        else:
            self.get_selection().unselect_path((self.unToSorted(i),))

def tree_model_sort_order(*cols):
    """tree_model_sort_order(i, j, k, ...):
    Returns a gtk.TreeModel sort_func which sorts first by column i, then column j, ..."""
    def sort_cmp(model, a, b):
        for c in cols:
            ac, bc = [model.get(iter, c)[0] for iter in [a, b]]
            if ac < bc:
                return -1
            elif ac > bc:
                return 1
        return 0
    return sort_cmp

class DictTable(gtk.TreeView):
    """DictTable(fixed, string_cap=10, float_cap=256):
    Specialization of a flat gtk.TreeView that can dynamically add columns for text or number.

    fixed       list of (name, type) of initially visible columns
    string_cap  max number of dynamically added string columns
    float_cap   max number of dynamically added number columns

Events:
    OnDblClick(DictTable, ix)

Properties:
    popup       gtk.Menu for right-click, default None
    model       gtk.ListStore
    index       unsorted index of selected row
    count       len(model)
"""
    def __init__(self, fixed, string_cap=10, float_cap=256):
        # one arg per fixed column: ('name', type)
        gtk.TreeView.__init__(self)
        self.set_rubber_banding(True)
        self.set_headers_visible(True)
        self.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
        self._count = 0
        self._index = -1 # pre-sorted
        self.OnDblClick = WeakEvent() # DictTable, ix
        self.connect("row_activated", self._onRowActivated)
        self.connect('button_press_event', self._onButtonPress)
        self.connect('key_press_event', self._onKeyPress)
        self._popup = None
        self._col_of_name = {}
        self._name_of_col = {}
        self._type_of_col = {}
        self._used_col = []
        self._spare_str = [x for x in xrange(len(fixed), len(fixed)+string_cap)]
        self._spare_str.reverse()
        self._spare_float = [x for x in xrange(len(fixed)+string_cap, len(fixed)+string_cap+float_cap)]
        self._spare_float.reverse()
        types = [typ for nm, typ in fixed] + string_cap*[str] + float_cap*[float]
        self._model = gtk.ListStore(*types)
        self._smodel = gtk.TreeModelSort(self._model)
        self.set_model(self._smodel)
        self.columns = []
        for i, fix in enumerate(fixed):
            nm, typ = fix
            renderer = gtk.CellRendererText()
            column = gtk.TreeViewColumn(nm, renderer, text=i)
            column.set_sort_column_id(i)
            self.append_column(column)
            self._col_of_name[nm] = i
            self._name_of_col[i] = nm
            self._type_of_col[i] = typ
            self._used_col.append(i)
            self.columns.append(column)
        self._smodel.set_sort_column_id(0, gtk.SORT_ASCENDING)
    popup = property(lambda s: s._popup, lambda s,x: s._setPopup(x))
    model = property(lambda s: s._model)
    index = property(lambda s: s._getIndex(), lambda s,x: s._setIndex(x))
    count = property(lambda s: s._count)
    def sortedToUn(self, ix):
        """Returns the unsorted index of the ix'th sorted row."""
        return self._smodel.convert_path_to_child_path((ix,))[0]
    def unToSorted(self, ix):
        """Returns the sorted index of the ix'th unsorted row."""
        return self._smodel.convert_child_path_to_path((ix,))[0]
    def _setPopup(self, x):
        self._popup = x
    def _getIndex(self):
        if self._index >= 0 and self.selected(self._index):
            return self._index
        for i in xrange(self._count):
            if self.selected(i):
                self._index = i
                return i
        self._index = -1
        return -1
    def _setIndex(self, ix):
        self._index = ix
        # self.get_selection().unselect_all()
        if ix >= 0:
            self.get_selection().select_path(self._smodel.convert_child_path_to_path((ix,)))
        else:
            self.get_selection().unselect_all()
    def clear(self):
        """Removes all rows."""
        self._model.clear()
        self._count = 0
    def append(self, d):
        """Adds a row described by dictionary d, where
    d[column_name] = column_value."""
        entry = [None] * (len(self._used_col) + len(self._spare_str) + len(self._spare_float))
        for i in self._used_col:
            entry[i] = self._type_of_col[i]()
        for i in self._spare_str:
            entry[i] = ""
        for i in self._spare_float:
            entry[i] = 0.0
        for nm in sorted(d.iterkeys()):
            val = d[nm]
            try:
                entry[self._col_of_name[nm]] = val
            except:
                if isinstance(val, str):
                    typ = str
                    ix = self._spare_str.pop()
                else:
                    typ = float
                    ix = self._spare_float.pop()
                renderer = gtk.CellRendererText()
                column = gtk.TreeViewColumn(nm, renderer, text=ix)
                column.set_sort_column_id(ix)
                self.append_column(column)
                self._col_of_name[nm] = ix
                self._name_of_col[ix] = nm
                self._type_of_col[ix] = typ
                self._used_col.append(ix)
                self.columns.append(column)
                entry[ix] = val
        self._model.append(entry)
        self._count += 1
    def remove(self, i):
        """Removes the item with unsorted index i."""
        self._model.remove(self._model.get_iter((i,)))
        self._count -= 1
    def selected(self, i):
        """Returns True if the i'th unsorted row is selected."""
        return self.get_selection().path_is_selected(self._smodel.convert_child_path_to_path((i,)))
    def select(self, i, s=True):
        """Selects the i'th unsorted row, or unselect if s == False."""
        path = self._smodel.convert_child_path_to_path((i,))
        if s:
            self.get_selection().select_path(path)
        else:
            self.get_selection().unselect_path(path)
    def select_all(self):
        """Selects all rows."""
        self.get_selection().select_all()
    def unselect_all(self):
        """Unselects all rows."""
        self.get_selection().unselect_all()
    def copy(self):
        """Copies all non-empty columns of all selected rows as tab-selected text, to clipboard."""
        sel = [i for i in xrange(self._count) if self.selected(i)]
        #def has_nonzero(c):
        #    return True in [bool(self._model.get(self._model.get_iter((i)), c)[0]) for i in sel]
        col = [i for i in self._used_col] # if ((self._type_of_col[i] == str) or has_nonzero(i))]
        # copy selected rows, str #and zero# and nonzero columns, to clipboard
        lines = []
        toks = []
        for c in col:
            toks.append(self._name_of_col[c])
        lines.append(toks)
        for i in xrange(self._count):
            r = self.sortedToUn(i)
            if r in sel:
                toks = []
                for c in col:
                    toks.append(str(self._model.get(self._model.get_iter((r)), c)[0]))
                lines.append(toks)
        lines.append([])
        text = '\n'.join(['\t'.join(toks) for toks in lines])
        clipboard = gtk.clipboard_get(gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(text)
    def _onRowActivated(self, tree, spath, col):
        path = self._smodel.convert_path_to_child_path(spath)
        if self._index != path[0]:
            self._index = path[0]
        self.OnDblClick(self, self._index)
    def _onButtonPress(self, view, event):
        if self._popup and event.button == 3:
            spath, col, cellx, celly = view.get_path_at_pos(int(event.x), int(event.y))
            path = self._smodel.convert_path_to_child_path(spath)
            # just in case the view doesn't alreay have focus
            view.grab_focus()
            selection = view.get_selection()
            # if this row isn't already selected, then select it before popup
            if not selection.path_is_selected(spath):
                selection.select_path(spath)
                self._index = path[0]
            self._popup.popup(None, None, None, event.button, event.time)
            return True
    def _onKeyPress(self, view, event):
        # apparently ctrl-a is already select-all
        if event.state & gdk.CONTROL_MASK:
            if event.keyval == ord('c') or event.keyval == ord('C'):
                self.copy()

class ZoomBar(gtk.HBox):
    """ZoomBar(zoom_min=100.0):
    Widget combining H scrollbar with zoom in/out buttons.

    zoom_min      smallest allowed zoom value

Properties:
    range         tuple (lo, hi)  the portion of (0, 100) that should be shown
    zoom          percent magnification, where 100% means all-on-screen
    adjustment    lower = range[0], upper = range[1]; connect to 'changed' and 'value_changed'
"""
    def __init__(self, zoom_min=100.0):
        gtk.HBox.__init__(self)
        self.zoom_min = zoom_min
        self._range = (0.0, 100.0)
        self._zoom = 100.0
        self.r = Reffer()
        dist = self._range[1] - self._range[0]
        self.adjustment = gtk.Adjustment( value=0, lower=self._range[0], upper=self._range[1],
                                          step_incr=dist/1000, page_incr=dist, page_size=dist )
        self.txtPct = NumEntry(100.0, self._acceptPct, "%.0f")
        self.txtPct.set_width_chars(4)
        self.txtPct.OnChange += self.r(self._onChangePct)
        self.pack_start(self.txtPct, False, True)
        self.txtPct.show()
        self.btnMinus = gtk.Button('-')
        self.btnMinus.connect('clicked', self._onPressMinus)
        self.pack_start(self.btnMinus, False, True)
        self.btnMinus.show()
        self.btnPlus = gtk.Button('+')
        self.btnPlus.connect('clicked', self._onPressPlus)
        self.pack_start(self.btnPlus, False, True)
        self.btnPlus.show()
        self.scroll = gtk.HScrollbar(self.adjustment)
        self.pack_start(self.scroll, True, True)
        self.scroll.show()
        self._setting = False
    def set_range(self, r):
        self._range = r
        self._recalc()
    def set_zoom(self, z):
        self._zoom = self._acceptPct(z)
        self._setting = True
        self.txtPct.setValue(self._zoom)
        self._setting = False
        self._recalc()
    range = property(lambda s: s._range, set_range)
    zoom = property(lambda s: s._zoom, set_zoom)
    def _recalc(self):
        lo, hi = self._range
        d = hi - lo
        z = self._zoom
        a = self.adjustment
        a.lower, a.upper = lo, hi
        a.step_increment = d/(10*z)
        a.page_increment = d*100/z
        a.page_size = d*100/z
        a.emit('changed')
        a.set_value(min(a.value, a.upper-a.page_size))
    def _acceptPct(self, pctStr):
        pct = float(pctStr)
        if pct < self.zoom_min:
            raise ValueError('zoom percent must be >= %f' % self._zoom_min)
        return pct
    def _onChangePct(self, txt, pct):
        if self._setting: return
        self.zoom = pct
    def _onPressMinus(self, widget):
        z = self.zoom / 2
        if z >= self.zoom_min:
            self.zoom = z
    def _onPressPlus(self, widget):
        self.zoom *= 2

class Triangle(gtk.DrawingArea):
    """Triangle(angle=-pi/2, color=(0.2, 1.0, 0.2)):
    Shows a triangle; animates when you change angle.

Properties:
    angle        in radians
    color        (r, g, b) in [0..1]
"""
    def __init__(self, angle=-pi/2, color=(0.2, 1.0, 0.2)):
        gtk.DrawingArea.__init__(self)
        self.set_size_request(12, 12)
        self.connect('expose_event', self._expose)
        self._angle = angle
        self._color = color
        self._goal = angle
    def set_angle(self, angle):
        self._goal = angle
        self.animate()
    angle = property(lambda s: s._angle, set_angle)
    def set_color(self, color):
        self._color = color
        self.invalidate()
    color = property(lambda s: s._color, set_color)
    def invalidate(self):
        """Forces redraw."""
        alloc = self.get_allocation()
        rect = gdk.Rectangle(0, 0, alloc.width, alloc.height)
        self.window.invalidate_rect(rect, True)
    def animate(self):
        if self._angle != self._goal:
            sign = (self._goal > self._angle) and 1.0 or -1.0
            # self._angle += sign * min(pi/20, abs(self._goal - self._angle))
            self._angle += 0.2 * (self._goal - self._angle)
            self.invalidate()
        #if self._angle != self._goal:
        if abs(self._angle - self._goal) < 1e-1:
            self._angle = self._goal
        else:
            gobject.timeout_add(20, self.animate)
    def _expose(self, widget, event):
        w,h = self.window.get_size()
        rad = min(w,h)/2.0
        # law of sines: side/sin(opposite) = circumscribing diameter
        side = 2*rad * sin(pi/3)
        corner_y = side/2
        # for what angle is normalized half a side the x coord (cos)?
        corner_theta = pi - asin(corner_y/rad)
        corner_x = cos(corner_theta) * rad
        cr = self.window.cairo_create()
        cr.translate(w/2.0, h/2.0)
        cr.rotate(- self._angle)
        if self._goal == self._angle:
            cr.set_source_rgb(*self._color)
        else:
            cr.set_source_rgba(*(list(self._color)+[.7]))
        cr.move_to(rad, 0)
        cr.line_to(corner_x, corner_y)
        cr.line_to(corner_x, -corner_y)
        cr.fill_preserve()
        cr.set_source_rgb(0,0,0)
        cr.set_line_width(0.5)
        cr.stroke()

class TriangleButton(gtk.Button):
    """gtk.Button decorated with a Triangle."""
    def __init__(self, angle=-pi/2, color=(0.2, 1.0, 0.2)):
        gtk.Button.__init__(self)
        self.triangle = Triangle(angle, color)
        self.triangle.show()
        self.add(self.triangle)


class TextViewAppender(object):
    """File-like object which appends to a gtk.TextView and scrolls to the end."""
    def __init__(self, tv):
        self.tv = tv
    def write(self, msg):
        buf = self.tv.get_buffer()
        buf.insert(buf.get_end_iter(), msg)# [:end]+'\n')
        self.tv.scroll_to_mark(buf.get_insert(), 0)
    def __call__(self, msg):
        self.write(msg)

def SetFixedWidth(textView):
    """Changes the font of a gtk.TextView to Monospace 10."""
    pangoFont = pango.FontDescription('monospace 10')
    textView.modify_font(pangoFont)

def SetClipboardCustom(clipboard, target, bytes):
    """Puts custom data on the specified clipboard, identified by target string."""
    targets = [ (target, 0, 0) ]
    def get_func(clipboard, selectiondata, info, data):
        selectiondata.set_text(data)
        return
    def clear_func(clipboard, data):
        del data
        return
    clipboard.set_with_data(targets, get_func, clear_func, bytes)


if __name__ == '__main__':
    from qub.accept import *
    from math import *
    from itertools import *
    from random import *
    #d = FunctionArgsDialog(['x'])
    #d.run('broken')
    #print 'f(%s) = %s' % (', '.join(d.args), d.expr)
    #d = IterExpressionDialog()
    #if gtk.RESPONSE_ACCEPT == d.run('count()'):
    #    print d.expr
    #d.destroy()
    

