"""Functions for parsing strings into various data types, ranging from
string (a no-op) to a system of differential equations.

An "accept" function takes a string, and returns a value of the desired type.
If the conversion is impossible, it raises a descriptive exception.
"""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import re
from qub.types import *
from math import *
import __main__
main_globals = __main__.__dict__

class FieldAcceptException(Exception):
    def __init__(self, badStr, msg):
        self.badStr = badStr
        self.msg = msg
    def __str__(self):
        return self.msg # + ' (' + self.badStr + ')'

def acceptNothing(x):
    raise FieldAcceptException(x, 'not editable')

def acceptString(x):
    return x

def acceptStringValue() :
    return lambda x:acceptString(x)

def acceptInt(x):
    try:
        return int(x)
    except:
        raise FieldAcceptException(x, "must be an integer")

def acceptFloat(x):
    try:
        return float(x)
    except:
        raise FieldAcceptException(x, "must be a number")

def acceptFloatOrUnset(x):
    if x == '':
        return UNSET_VALUE
    return acceptFloat(x)

def formatFloatOrUnset(fmt):
    fmt_func = acceptFormat(fmt)
    def format(x):
        if abs(x - UNSET_VALUE) < 1e-6:
            return ''
        return fmt_func(x)
    return format
        
def acceptBool(x):
    try:
        if x.lower() == 'false':
            return False
    except:
        pass
    return bool(x)

def acceptCmp(convert, compare, errmsg):
    def accept(s):
        try:
            x = convert(s)
            if not compare(x):
                raise FieldAcceptException(s, errmsg)
            return x
        except ValueError:
            raise FieldAcceptException(s, errmsg)
    return accept

errmsg_cmp = lambda cmpname: lambda typename, bound: "must be %s %s %s" % (typename, cmpname, bound)
errmsg_gt = errmsg_cmp("greater than")
errmsg_ge = errmsg_cmp("greater than or equal to")
errmsg_lt = errmsg_cmp("less than")
errmsg_le = errmsg_cmp("less than or equal to")
name_int = "an integer"
name_float = "a number"

def acceptIntGreaterThan(lb):
    return acceptCmp(int, lambda x:x>lb, errmsg_gt(name_int, lb))
def acceptIntGreaterThanOrEqualTo(lb):
    return acceptCmp(int, lambda x:x>=lb, errmsg_ge(name_int, lb))
def acceptIntLessThan(ub):
    return acceptCmp(int, lambda x:x<ub, errmsg_lt(name_int, ub))
def acceptIntLessThanOrEqualTo(ub):
    return acceptCmp(int, lambda x:x<=ub, errmsg_le(name_int, ub))
def acceptIntBetween(lb, ub):
    return acceptCmp(int, lambda x:lb<=x<=ub, "must be %s between %s and %s" % (name_int, lb, ub))
def acceptFloatGreaterThan(lb):
    return acceptCmp(float, lambda x:x>lb, errmsg_gt(name_float, lb))
def acceptFloatGreaterThanOrEqualTo(lb):
    return acceptCmp(float, lambda x:x>=lb, errmsg_ge(name_float, lb))
def acceptFloatLessThan(ub):
    return acceptCmp(float, lambda x:x<ub, errmsg_lt(name_float, ub))
def acceptFloatLessThanOrEqualTo(ub):
    return acceptCmp(float, lambda x:x<=ub, errmsg_le(name_float, ub))
def acceptFloatBetween(lb, ub):
    return acceptCmp(float, lambda x:lb<=x<=ub, "must be %s between %s and %s" % (name_float, lb, ub))
def acceptFloatAbsBetween(lb, ub):
    return acceptCmp(float, lambda x:(lb<=x<=ub) or ((-ub)<=x<=(-lb)), "must be %s between +/- %s and %s" % (name_float, lb, ub))


# eats all the spaces commas and dashes, so don't accept on those
def acceptAcceptList(s, accept, dashrange, description):
    "e.g. '1, 2, 3-5 7' -> [1, 2, 3, 4, 5, 7]"

    if s == '':
        return []
    
    # replace \d+ ?- ?\d+ with \d+-\d+
    s = re.sub(r'([^ -]+) ?- ?([^ -]+)', r'\1-\2', s)
    # split along [, ]+
    tokens = re.split(r'[ ,\t]+', s)
    # foreach token:  if \d+ append int; elif \d+-\d+ append ints
    result = []
    for tok in tokens:
        match = re.search(r'([^ -]+)-([^ -]+)', tok)
        try:
            if match:
                a, b = accept(match.group(1)), accept(match.group(2))
                result[len(result):] = dashrange(a, b)
            else:
                result.append( accept(tok) )
        except FieldAcceptException:
            raise FieldAcceptException(s, 'must be a list of %s, separated by "," and "-"' % description)
    return result

def acceptList(acceptor, dashrange, description):
    return lambda x: acceptAcceptList(x, acceptor, dashrange, description)

def acceptIntList(acceptor=acceptInt, description='integers'):
    def dashrange(i, j):
        if i <= j:
            return xrange(i, j+1)
        else:
            return xrange(i, j-1, -1)
    return acceptList(acceptor, dashrange, description)

def acceptFloatList(acceptor=acceptFloat, description='numbers'):
    def dashrange(i, j):
        raise FieldAcceptException('%f-%f', 'How shall I enumerate a continuous range?')
    return acceptList(acceptor, dashrange, description)

def acceptDimList(acceptList, dim, msg):
    def checkDim(lst):
        if dim != len(lst):
            raise FieldAcceptException(s, msg)
        return lst
    return lambda x: checkDim( acceptList(x) )


def MissingName(msg):
    m = re.search("'([^']+)'", msg)
    if not m:
        raise NameError, msg
    return m.group(1)

def acceptF(static=["x"], avail=[], custom=False, typ=float, default="0.0"):
    def accept(eqn):
        args = list(static[:])
        valid = False
        func = lambda : 0.0
        while not valid:
            try:
                func = eval( 'lambda ' + join(args, ',') + ': ' + (eqn or default), main_globals )
                fval = typ( func( *map(lambda z:1.0, range(len(args))) ) )
                valid = 1
            except NameError, msg:
                missing = MissingName(str(msg))
                if custom or (missing in avail):
                    args.append(missing)
                else:
                    raise
        return Anon(f=func, args=args)
    return accept

def Union(ls):
    have = {}
    out = []
    for l in ls:
        for x in l:
            if (not (x in have)):
                have[x] = 1
                out.append(x)
    return out

def acceptODEs(static=["x"], avail=[], custom=False):
    def accept(expr):
        nospace = re.sub(r"[ \t]+", "", expr)
        eqns = [x for x in re.split(r";", nospace) if x]
        lrs = [re.split(r"'=", eqn) for eqn in eqns]
        for i,x in enumerate(lrs):
            if len(x) != 2:
                raise Exception("Malformed differential equation: %s" % eqns[i])
        if len(dict(lrs)) < len(lrs):
            raise Exception("Duplicate variable name.")
        vars = [l for l,r in lrs]
        if any((v in static) or (v in avail) for v in vars):
            raise Exception("Differential variables can't have the same name as data series.")
        # static_vars = static + vars
        # ok now we acceptF(static_vars, avail, custom)(each r)
        # to get each F.args.  We combine them into a new static_custom_vars, new acceptF
        # and re-accept to get derivs[] with the same signature
        accept1 = acceptF(static+vars, avail, custom)
        static_vars = static+vars
        customs = [x for x in Union([accept1(r).args for l,r in lrs]) if not (x in static_vars)]
        accept2 = acceptF(static+customs+vars, [], False)
        derivs = [accept2(r).f for l,r in lrs]
        def deriv(*args, **kw):
            return [d(*args, **kw) for d in derivs]
        return Anon(dy=deriv, args=static+customs, ynames=vars)
        # dy(static, ...customs..., ...y...) -> dydx[]
    return accept


def acceptODEorF(static=["x"], avail=[], custom=False):
    aO = acceptODEs(static, avail, custom)
    aF = acceptF(static, avail, custom)
    def accept(x):
        if re.search("'", x):
            return aO(x)
        else:
            return aF(x)
    return accept


# for uniform handling of "format" arguments to qub stuff:
def acceptFormat(format):
    if callable(format):
        return format
    else:
        return lambda v: format % v
