"""Support for user scripting, including
    output injection, multi-re-direction and rotating logs
    well-defined global namespace
    persistent additions to sys.path
    persistent two-stage startup script

The official global scope is module __main__.
To set a global variable: env.globals['x'] = 1
To run user code, exec(code_string, env.globals)
To evaluate,      eval(code_string, env.globals)
"""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import os
import sys
import traceback
import qub.types
import __main__

# We provide OnOutput callback, which could do potentially non-thread-safe
# things like write to the gui.  Typically in this situation there is a
# way to request execution on the gui thread, e.g. gobject.idle_add.
# For simpler programs, we can call OnOutput directly, in-thread:
def call_now(func, *args):
    func(*args)

class PythonEnvironment(object):
    """Python environment management for user scripting.

properties:
    path    list of additions to sys.path
    script  string to execute on startup/reload

Don't construct this directly; call Init() and use global "env"
"""
    def __init__(self, folder, call_later=call_now):
        global env
        env = self
        self.folder = folder
        self.call_later = call_later
        self.rotate_logs()
        self.log = open(os.path.join(folder, 'output.log'), 'w')
        self.OnOutput = qub.types.WeakEvent("PythonEnvironment.OnOutput") # (str)
        self.silence = False
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self
        self._path = []
        path = []
        self.script = """from math import *

def after_initialize():
    pass

"""
        try:
            prefs = open(os.path.join(folder, 'env.py'), 'r')
            try:
                path = eval(prefs.readline())
            except:
                pass
            self.script = prefs.read()
        except:
            pass
        
        self.globals = __main__.__dict__
        self.path = path
        self.reload()
    def set_path(self, path):
        for p in self._path:
            sys.path.remove(p)
        self._path = path
        for p in self._path:
            sys.path.append(p)
    path = property(lambda s: s._path, set_path)
    def reload(self):
        """executes the startup script"""
        try:
            exec(self.script, self.globals)
        except:
            traceback.print_exc()
    def after_initialize(self):
        """calls global after_initialize().  call this when your program is totally running."""
        try:
            self.globals['after_initialize']()
        except:
            traceback.print_exc()
    def save(self):
        """writes path and script to disk"""
        try:
            prefs = open(os.path.join(self.folder, 'env.py'), 'w')
            prefs.write('%s\n'%self.path)
            prefs.write(self.script)
            if (not self.script) or (self.script[-1] != '\n'):
                prefs.write('\n')
        except:
            traceback.print_exc()
    def write(self, x):
        """Aappends x to sys.stdout and the log-file, and passes x to any OnOutput listeners."""
        if self.silence: return
        self.log.write(x)
        self.stdout.write(x)
        self.call_later(self.OnOutput, x)
    def flush(self):
        self.stdout.flush()
    def rotate_logs(self):
        """moves the older log files along toward oblivion; called automatically on construction."""
        MAXLOG = 5
        base = os.path.join(self.folder, 'output.log')
        name = lambda i: "%s.%i" % (base, i)
        names = [name(i) for i in reversed(xrange(MAXLOG))] + [base]
        for i in xrange(len(names)-1):
            try:
                if os.path.exists(names[i]):
                    os.remove(names[i])
                if os.path.exists(names[i+1]):
                    os.rename(names[i+1], names[i])
            except:
                traceback.print_exc()

def Init(folder, call_later=call_now):
    """Sets up the global variable "env" (class PythonEnvironment)

    folder:     path where settings and logs are saved
    call_later: e.g. gobject.idle_add if your OnOutput callback calls on gtk
"""
    global env
    if not os.path.exists(folder):
        os.makedirs(folder)
    env = PythonEnvironment(folder, call_later)
