"""Classes for reading and writing all QuB binary files.

A QuB Tree is a hierarchy of Nodes.  Each node may have a name and some text or numeric data.
Specific QuB file types, e.g. QMF (qub model files), expect specific heirarchies and data types.

History:

In the late 1990s Dr. Lorin Milescu defined the binary layout and used it for QuB data files (QDF).
He wrote a class in Delph pascal called "QFS" to read and modify open files on disk.

By the early 2000s we were collaborating on a DLL interface between his GUI, written in Delphi, and
older algorithms written in C.  We found ourselves spending a lot of time defining struct- and pointer-based
containers for all the varieties of information shared between GUI and DLL.  At the same time,
folks were asking us for new extensible formats for models and idealized data.  Lorin naturally proposed
QFS.  I was feeling some pressure to use human-readable ASCII text, so I proposed a file that could
be stored as binary or text, depending whether efficiency or transparency was foremost.  Then I went
ahead and implemented it in C/C++, with bindings to Delphi.  I never really got the text part right, but
no-one really cared.

So now we had two implementations: QFS, and "qubtree" or "native".  We had some idea of changing
everything over to the newer one, but it would have meant a ton of extra work re-implementing the core of QuB.

When I made the Delphi bindings, I had our DLL quagmire in mind.  qubtree.dll took care of memory management
and reference counting so you could pass trees willy-nilly between the EXE and DLL without headaches.  And now
that everything boiled down to one or two ref-counted qubtree pointers, we weren't constantly breaking binary
compatibility between DLL and EXE.  Similarly, I added Python bindings and rudimentary callback support, so we
could do a lot of stuff across the three languages without too much intricate boilerplate.

Now, in the late 2000s, I'd like to write pure-python programs which read and write QuB files.  I'm willing to
sacrifice the cross-language magic.  And through the power of numpy and mmap, it might even make pure-python
users more efficient than with the compiled library.  This 'numpy' flavor behaves exactly like the 'native' one
except:
  - it doesn't even try to ReadString or ReadText; like I said, always broken anyway
  - it ignores lock/unlock since python's single-threaded for ref-counting purposes
  - you can't edit string data one character at a time, due to a limitation in module mmap
  - you can get direct access to data as numpy.array: node.storage.data
  - you can resize a node with empty data, by specifying a default type


Flavors:

By default this module still uses the 'native' flavor, so there are no surprises for existing programs.
If you don't have the compiled _qubtree library, it falls back on the 'numpy' flavor.  Of course,
you must have numpy, available separately.  If you have both, you can switch the default for new nodes:
     qub.tree.CHOOSE_FLAVOR('numpy')
and even convert a tree between the two flavors:
     as_numpy = qub.tree.AS_FLAVOR(native_node, 'numpy')


Limitations:

32 bits, except when signed/unsigned are treated inconsistently; then it's more like 31.
(max safe file size 2GB - 1)

Supported by no-one else

Not very efficient when you have lots of nodes.


Chris Nicolai 2009

"""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import struct
import array
import weakref
import mmap
import cStringIO as StringIO
import operator
import time
import traceback
import os
from itertools import izip
import collections


FLAVOR_DEFAULT = 'native'
FLAVORS_AVAIL = []
FLAVOR = FLAVOR_DEFAULT

try:
    import _qubtree
    have_lib = True
    FLAVORS_AVAIL.append('native')
    Node = _qubtree.Node
    NullNode = _qubtree.NullNode
    Open = _qubtree.Open
    ReadBytes = _qubtree.ReadBytes
except:
    have_lib = False

try:
    import numpy
    have_numpy = True
    FLAVORS_AVAIL.append('numpy')
except:
    have_numpy = False

def CHOOSE_FLAVOR(flav):
    """Selects the default flavor for new Nodes, either 'native' (default) or 'numpy'."""
    global Node, NullNode, Open, ReadBytes, FLAVOR
    if flav == 'native':
        Node = _qubtree.Node
        NullNode = _qubtree.NullNode
        Open = _qubtree.Open
        ReadBytes = _qubtree.ReadBytes
    elif flav == 'numpy':
        Node = Node_numpy
        NullNode = NullNode_numpy
        Open = Open_numpy
        ReadBytes = ReadBytes_numpy
    FLAVOR = flav

def AS_FLAVOR(node, flav, copyAnyway=False):
    """Returns node in the requested flavor.  If it is already, we can return the original (default) or copyAnyway."""
    if have_lib and (0 <= str(type(node)).find('QUBTree')):
        if flav == 'native':
            if copyAnyway:
                return node.clone()
            else:
                return node
        elif flav == 'numpy':
            if node.isNull:
                return NullNode_numpy()
            else:
                return ReadBytes_numpy(node.getBytes())
    elif have_numpy and isinstance(node, NullNode_numpy):
        if flav == 'numpy':
            return node
        elif flav == 'native':
            return _qubtree.NullNode()
    elif have_numpy and isinstance(node, Node_numpy):
        if flav == 'numpy':
            if copyAnyway:
                return node.clone()
            else:
                return node
        elif flav == 'native':
            return _qubtree.ReadBytes(node.getBytes())

# node.data.type constants

QTR_TYPE_EMPTY, \
    QTR_TYPE_UNKNOWN, \
    QTR_TYPE_POINTER, \
    QTR_TYPE_STRING, \
    QTR_TYPE_UCHAR, \
    QTR_TYPE_CHAR, \
    QTR_TYPE_USHORT, \
    QTR_TYPE_SHORT, \
    QTR_TYPE_UINT, \
    QTR_TYPE_INT, \
    QTR_TYPE_ULONG, \
    QTR_TYPE_LONG, \
    QTR_TYPE_FLOAT, \
    QTR_TYPE_DOUBLE, \
    QTR_TYPE_LDOUBLE = range(15)
TYPES = (QTR_TYPE_EMPTY,
         QTR_TYPE_UNKNOWN,
         QTR_TYPE_POINTER,
         QTR_TYPE_STRING,
         QTR_TYPE_UCHAR,
         QTR_TYPE_CHAR,
         QTR_TYPE_USHORT,
         QTR_TYPE_SHORT,
         QTR_TYPE_UINT,
         QTR_TYPE_INT,
         QTR_TYPE_ULONG,
         QTR_TYPE_LONG,
         QTR_TYPE_FLOAT,
         QTR_TYPE_DOUBLE,
         QTR_TYPE_LDOUBLE)

# for human consumption:
TYPENAMES =  {QTR_TYPE_EMPTY: "empty",
              QTR_TYPE_UNKNOWN: "unknown",
              QTR_TYPE_POINTER : "pointer",
              QTR_TYPE_STRING : "string",
              QTR_TYPE_UCHAR : "unsigned 8-bit int",
              QTR_TYPE_CHAR : "signed 8-bit int",
              QTR_TYPE_USHORT : "unsigned 16-bit int",
              QTR_TYPE_SHORT : "signed 16-bit int",
              QTR_TYPE_UINT : "unsigned int",
              QTR_TYPE_INT : "int",
              QTR_TYPE_ULONG : "unsigned 64-bit int",
              QTR_TYPE_LONG : "signed 64-bit int",
              QTR_TYPE_FLOAT : "single-precision float",
              QTR_TYPE_DOUBLE : "float",
              QTR_TYPE_LDOUBLE : "extended-precision float"
              }

# for resizing:
TYPE_DEFAULTS = (None, None, None, ' ', 0, 0, 0, 0, 0, 0, 0, 0, 0.0, 0.0, 0.0)

# size of one element, in bytes
SIZE_OF_TYPE = (0, 0, 4, 1, 1, 1, 2, 2, 4, 4, 8, 8, 4, 8, 10)

# mapping between numpy and qubtree types:
TYPES_NUMPY = [None, None, 'uint32', 'uint8', 'uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64', 'float32', 'float64', 'float80']

TYPE_OF_NUMPY = collections.defaultdict(lambda:QTR_TYPE_DOUBLE)
TYPE_OF_NUMPY['uint8'] = QTR_TYPE_UCHAR
TYPE_OF_NUMPY['int8'] = QTR_TYPE_CHAR
TYPE_OF_NUMPY['uint16'] = QTR_TYPE_USHORT
TYPE_OF_NUMPY['int16'] = QTR_TYPE_SHORT
TYPE_OF_NUMPY['uint32'] = QTR_TYPE_UINT
TYPE_OF_NUMPY['int32'] = QTR_TYPE_INT
TYPE_OF_NUMPY['uint64'] = QTR_TYPE_ULONG
TYPE_OF_NUMPY['int64'] = QTR_TYPE_LONG
TYPE_OF_NUMPY['float32'] = QTR_TYPE_FLOAT
TYPE_OF_NUMPY['float64'] = QTR_TYPE_DOUBLE
TYPE_OF_NUMPY['float80'] = QTR_TYPE_LDOUBLE

# iteration, since it's too late to build in as a method

def children(tree, name=None):
    """Yields all children of tree, matching name if specified."""
    if name == None:
        child = tree.child
    else:
        child = tree.find(name)
    while not child.isNull:
        yield child
        if name == None:
            child = child.sibling
        else:
            child = child.nextSameName()


# reading raw nodes:

framedef = "bbbbiiiiiibbbb"
framelen = struct.calcsize(framedef)
frame_dataoff = 12

QTR_FLAG_PRELOAD = 0x2
QTR_FLAG_DATA_IN_NODE = 0x4
QTR_MAGIC = "QUB_(;-)_QFS"


class dtree(object):
    """A light intermediary between a file and a reader of qubtree nodes."""

    def __init__(self, fi):
        """Sets up a dtree reader given an open file fi."""
        self.fi = fi
    def rawframe(self, off):
        """Returns the tuple of all fields of a qubtree node, located at offset off, except the name.

Returns: (flags, a, b, type, size, count, dataPos, childPos, siblingPos, c, d, e, f, nameLen)
         where a thru f are reserved (ignored) bytes.
"""
        self.fi.seek(off)
        dat = self.fi.read(framelen)
        fdef = framedef
        while fdef:
            try:
                return struct.unpack(fdef, dat)
            except:
                fdef[:-1] = []
        return tuple()
    def frame(self, off):
        """Returns the tuple of all fields of a qubtree.Node including the name, with flags upacked.

Returns: (preload, data_in_node, type, size, count, dataPos, childPos, siblingPos, name)
"""
        raw = list(self.rawframe(off))
        raw = raw + [None] * (14-len(raw))
        flags, a, b, type, size, count, dPos, cPos, sPos, c, d, e, f, nameLen = raw
        preload = bool(flags and (flags & 2))
        dataIn = bool(flags and (flags & 4))
        if type is not None:
            type = [None, None, None, "s", "B", 'b', 'H', 'h', 'I', 'i', 'L', 'l', 'f', 'd', None][type]
        name = None
        if nameLen == 0:
            name = ""
        elif nameLen:
            try:
                self.fi.seek(off+framelen)
                name = self.fi.read(nameLen)
            except:
                pass
        return (preload, dataIn, type, size, count, dPos, cPos, sPos, name)
    def hier(self, off=12, indent=""):
        """Prints the tree of frames rooted at offset off, with indentation to show hierarchy."""
        frame = self.frame(off)
        print indent,frame
        if frame[6]:
            self.hier(frame[6], indent+'   ')
        if frame[7]:
            self.hier(frame[7], indent)
    def data(self, off, count, format):
        """Prints the array of count number of items at offset off, with type given as a format string such as in module struct."""
        self.fi.seek(off)
        arr = array.array(format)
        arr.fromfile(self.fi, count)
        return arr


# numpy flavor:

if have_numpy:

    # node.data when node.isNull:
    class NullData(object):
        """Represents the data of a null node, which is empty and will stay empty."""
        def __repr__(self):
            return '<QUBTree NULL data>'
        def __str__(self):
            return ''
        def __len__(self):
            return 0
        def __getitem__(self, key):
            raise IndexError()
        def __setitem__(self, key, val):
            raise IndexError()
        def get_node(self):
            return NullNode_numpy_instance
        node = property(get_node, doc="the null node whose data this is.")
        def set_none(self, x):
            pass
        preload = property(lambda self: True, set_none)
        count = property(__len__)
        rows = property(lambda self: 0)
        cols = property(lambda self: 0)
        type = property(lambda self: QTR_TYPE_EMPTY)
        size = property(lambda self: 0)
        loaded = property(lambda self: (-1, -1))
        loadedRows = property(lambda self: (-1, -1))
        def setup(self, type, rows, cols):
            pass
        def resize(self, nr):
            pass
        def clear(self):
            pass
        def loadRows(self, first, last, do_read=True):
            pass
        def unloadRows(self, do_write=True):
            pass
        def row(self, i):
            raise IndexError()
        def col(self, j):
            raise IndexError()

    class NullNode_numpy(object):
        """Represents the absence of a qubtree node.

Has all the properties and methods, so you can e.g. ask blithely if it has any data or children (no).
"""
        def __init__(self):
            self.__data = NullData()
        def __repr__(self):
            return '<QUBTree Node : NULL>'
        def __str__(self):
            return '\n'
        def __cmp__(self, other):
            try:
                if other.isNull:
                    return 0
                else:
                    return -1
            except:
                return -1
        def __nonzero__(self):
            return False # a null node is Falsy
        def __len__(self):
            return 0
        def __getitem__(self, key):
            return self
        def set_none(self, x):
            pass
        name = property(lambda self: '', set_none)
        data = property(lambda self: self.__data, set_none)
        lineComment = property(lambda self: '', set_none)
        isNull = property(lambda self: True)
        child = property(lambda self: self)
        sibling = property(lambda self: self)
        parent = property(lambda self: self)
        path = property(lambda self: '')
        modified = property(lambda self: False)
        root = property(lambda self: self)
        def clone(self, deep=True):
            return self
        def saveAs(self, path):
            return False
        def saveAsText(self, path):
            return False
        def save(self):
            return False
        def close(self):
            return False
        def getBytes(self):
            return ''
        def list(self):
            return []
        def find(self, name):
            return self
        def next(self, name):
            return self
        def nextSameName(self):
            return self
        def append(self, childOrName):
            return self
        def appendClone(self, other, deep=True):
            return self
        def insert(self, child, after=None):
            return self
        def insertClone(self, other, deep=True, after=None):
            return self
        def remove(self, child):
            pass
        def setChanged(self):
            pass
        def lock(self, timeoutMS=None):
            pass
        def unlock(self):
            pass

    # from here on out we try to use this instance instead of constructing a flock of identical Nulls.
    NullNode_numpy_instance = NullNode_numpy()

    # generic data wrappers
    # These three (Data, DataRow, DataCol) help read and manipulate node data, with
    # the same API as the 'native' flavor.
    # The real work happens in the storage and Node classes below.
    class Data(object):
        """A handler for the data in one qubtree node.  Don't construct one for yourself; use node.data"""
        def __init__(self, node):
            """Constructs a data handler for a qubtree node.  Don't call this directly."""
            self.__node = weakref.ref(node)
        def __repr__(self):
            return '<QUBTree data for %s>' % self.node
        def __str__(self):
            if self.type == QTR_TYPE_STRING:
                if self.node.storage.data[0] == '\0':
                    return ''
                else:
                    return self.node.storage.data[:]
            elif self.cols > 1 and self.rows > 1:
                return '(%s)' % '\\\n'.join([' '.join([x for x in self.row(r)]) for r in xrange(self.rows)])
            elif self.count:
                return ' '.join(str(x) for x in self)
            else:
                return ''
        def __len__(self):
            return self.node.storage.dataCount
        def __getitem__(self, key):
            s = self.node.storage
            n = s.dataCount
            if isinstance(key, slice):
                return [s.get_data_item(i) for i in xrange(*key.indices(n))]
            elif -n <= key < 0:
                return s.get_data_item(n+key)
            elif 0 <= key < n:
                return s.get_data_item(key)
            else:
                raise IndexError()
        def __setitem__(self, key, val):
            s = self.node.storage
            n = s.dataCount
            if isinstance(key, slice):
                for i, v in izip(xrange(*key.indices(n)), val):
                    s.set_data_item(i, v)
            elif -n <= key < 0:
                return s.set_data_item(n+key, val)
            elif 0 <= key < n:
                return s.set_data_item(key, val)
            else:
                raise IndexError()
        def get_node(self):
            try:
                return self.__node()
            except:
                node = NullNode_numpy_instance
                self.__node = weakref.ref(node)
                return node
        node = property(get_node, doc="the qubtree node whose data this is")
        preload = property(lambda self: self.node.storage.preload, lambda self, x: self.node.storage.set_preload(x),
                           doc="whether all data is loaded into memory on file open")
        count = property(__len__, doc="total data elements (rows * cols)")
        rows = property(lambda self: self.node.storage.rows, doc="number of rows")
        cols = property(lambda self: self.node.storage.cols, doc="number of columns")
        type = property(lambda self: self.node.storage.dataType, doc="type of data elements, in QTR_TYPE_*")
        size = property(lambda self: self.node.storage.dataSize, doc="size of one data element in bytes")
        loaded = property(lambda self: self.node.storage.loaded, doc="(first, last) element loaded into memory")
        loadedRows = property(lambda self: self.node.storage.loadedRows, doc="(first, last) row loaded into memory")
        def setup(self, type, rows, cols):
            """Discards any data and sets up space for rows*cols of the specified type."""
            self.node.storage.setup_data(type, rows, cols)
        def resize(self, nr, type_if_none=QTR_TYPE_DOUBLE):
            """Changes the number of rows allocated."""
            self.node.storage.resize_data(nr, type_if_none)
        def clear(self):
            """Discards any data."""
            self.node.storage.clear_data()
        def loadRows(self, first=-2, last=-2, do_read=True):
            """Loads the rows from first up to and including last into memory.  By default loads all.

In this numpy flavor, all data is always loaded (memory mapped), and loadRows just adjusts .loadedRows."""
            if first == -2:
                if self.rows:
                    f = 0
                else:
                    f = -1
            else:
                f = first
            if last == -2:
                if self.count:
                    l = self.rows - 1
                else:
                    l = -1
            else:
                l = last
            self.node.storage.load_rows(f, l, do_read)
        def unloadRows(self, do_write=True):
            """Unloads any loaded data.

In this numpy flavor, all data is always loaded (memory mapped), and unloadRows just adjusts .loadedRows."""
            self.node.storage.unload_rows(do_write)
        def read_rows(self, first, last):
            """Returns a copy of the data in rows first..last"""
            return self.node.storage.read_rows(first, last)
        def write_rows(self, first, last, data):
            """Replaces the data in rows first..last with the given data."""
            self.node.storage.write_rows(first, last, data)
        def get_rows(self, first, last):
            """Returns a slice into the first..last rows of data.  If possible, it will be a direct, mutable reference.
This may not be possible with some string data."""
            return self.node.storage.get_rows(first, last)
        def row(self, i):
            """Returns an object that can manipulate the data in row i, and acts like a list."""
            if 0 <= i < self.rows:
                return DataRow(self, i)
            raise IndexError()
        def col(self, j):
            """Returns an object that can manipulate the data in column j, and acts like a list."""
            if 0 <= self.cols:
                return DataCol(self, j)
            raise IndexError()

    class DataRow(object):
        """A handler for the data in one row of a qubtree node.  Don't construct one for yourself; use node.data.row(i)"""
        def __init__(self, data, r):
            """Don't call this directly."""
            self.data = data
            self.r = r
        def __repr__(self):
            return '<QUBTree data for %s, row %i>' % (self.data.node.name, self.r)
        def __str__(self):
            return ' '.join([x for x in self])
        def __len__(self):
            return self.data.cols
        def __getitem__(self, key):
            r = self.r
            d = self.data
            l = d.cols
            s = r * l
            if isinstance(key, slice):
                return [d[s+i] for i in xrange(*key.indices(l))]
            elif -l <= key < 0:
                return d[s+l+key]
            elif 0 <= key < l:
                return d[s+key]
            else:
                raise IndexError()
        def __setitem__(self, key, val):
            r = self.r
            d = self.data
            l = d.cols
            s = r * l
            if isinstance(key, slice):
                for i,v in izip(xrange(*key.indices(l)), val):
                    d[s+i] = v
            elif -l <= key < 0:
                d[s+l+key] = val
            elif 0 <= key < l:
                d[s+key] = val
            else:
                raise IndexError()

    class DataCol(object):
        """A handler for the data in one column of a qubtree node.  Don't construct one for yourself; use node.data.col(j)"""
        def __init__(self, data, c):
            """Don't call this directly."""
            self.data = data
            self.c = c
        def __repr__(self):
            return '<QUBTree data for %s, col %i>' % (self.data.node.name, self.c)
        def __str__(self):
            return ' '.join([x for x in self])
        def __len__(self):
            return self.data.rows
        def __getitem__(self, key):
            c = self.c
            d = self.data
            l = d.rows
            nc = d.cols
            if isinstance(key, slice):
                return [d[i*nc+c] for i in xrange(*key.indices(l))]
            elif -l <= key < 0:
                return d[(l+key)*nc+c]
            elif 0 <= key < l:
                return d[key*nc+c]
            else:
                raise IndexError()
        def __setitem__(self, key, val):
            c = self.c
            d = self.data
            l = d.rows
            nc = d.cols
            if isinstance(key, slice):
                for i,v in izip(xrange(*key.indices(l)), val):
                    d[i*nc+c] = v
            elif -l <= key < 0:
                d[(l+key)*nc+c] = val
            elif 0 <= key < l:
                d[key*nc+c] = val
            else:
                raise IndexError()

    # clean vs modified:
    # - in the native flavor's C code, these are the flags "changed" vs "child_changed"
    # - node.storage.clean tells if the node is saved on disk, and completely up to date.  It's true only:
    #     + in an open file
    #     + after Opening
    #     + after saving
    #     + before changing this node's data, child, or sibling
    # - actually, it may be fraudulently True if e.g. the child's name was changed
    #     + name change usually means the node has to move
    #     + this node doesn't get the new child address until the "pre-save" linkup
    #     + pre-save happens automatically, right before saving, and when checking node.modified
    # - node.modified tells if this node or any descendent is not clean.
    # - "storage" doesn't have access to the Node object or its parent, so it signals "newly unclean" by a callback event
    # - the callback is weak-referenced to avoid circular reference
    # - if there is no callback provided, it calls this function:
    def ignore_dirtied(storage):
        pass

    # abstract storage
    class NodeStorage(object):
        """Abstract base class for managing a concrete representation of one qubtree node.

Subclasses NodeInFile and NodeInMem are instantiated as needed by class Node."""
        clean = property(lambda self: self.__clean, lambda self, x: self.set_clean(x),
                         doc="whether this node has an up-to-date representation on disk.")
        OnDirtied = property(lambda self: self.get_ondirtied(), lambda self, x: self.set_ondirtied(x),
                             "client-provided f(storage) is called when clean becomes False.")
        name = property(lambda self: self.get_name(), lambda self, x: self.set_name(x), doc="the node's name")
        preload = property(lambda self: self.get_preload(), lambda self, x: self.set_preload(x),
                           doc="whether to load all data elements on file open")
        dataCount = property(lambda self: self.get_dataCount(), doc="total number of data elements; rows*cols")
        dataSize = property(lambda self: self.get_dataSize(), doc="size of one data element, in bytes")
        dataType = property(lambda self: self.get_dataType(), doc="type of data, in QTR_TYPE_*")
        dataPos = property(lambda self: self.get_dataPos(), doc="position of data in a file, in bytes")
        childPos = property(lambda self: self.get_childPos(), lambda self, x: self.set_childPos(x),
                            doc="position of child node in file, in bytes, or 0")
        siblingPos = property(lambda self: self.get_siblingPos(), lambda self, x: self.set_siblingPos(x),
                              doc="position of sibling node in file, in bytes, or 0")
        rows = property(lambda self: self.get_rowcount(), doc="number of rows of data")
        cols = property(lambda self: self.get_colcount(), doc="number of columns of data")
        loaded = property(lambda self: self.get_loaded(), doc="indices of (first, last) loaded rows")
        loadedRows = property(lambda self: self.get_loadedRows(), doc="indices of (first, last) loaded data elements")
        data = property(lambda self: self.get_data(), doc="the actual data, mutable if possible.  Depending on data type and storage type, can be an mmap.mmap, a slice of an mmap.mmap, a numpy.array, a numpy array onto a slice of an mmap.mmap, or a plain string.")
        file = property(lambda self: self.get_file(), doc="the NodeFile containing this node, or None")
        offset = property(lambda self: self.get_offset(), doc="the location of this node (frame) in file, in bytes")
        def __init__(self):
            self.__clean = False # dirty unless it's on disk and up-to-date
            self.OnDirtied = None
        def set_clean(self, x):
            """Called by subclass methods when they make a change to the node structure from what's on disk."""
            if x == self.__clean: return
            self.__clean = x
            if not x:
                self.OnDirtied(self)
        def get_ondirtied(self): # the OnDirtied callback, for when it flips from clean to unclean
            return self.__dirtied() or ignore_dirtied
        def set_ondirtied(self, x):
            if x:
                self.__dirtied = weakref.ref(x)
            else:
                self.__dirtied = weakref.ref(ignore_dirtied)
        def get_data_item(self, i):
            pass
        def set_data_item(self, i, x):
            pass
        def get_dataPos(self):
            return 0 # position in bytes of the start of data, or 0 if none; only relevant in a file
        def get_childPos(self):
            return 0 # position in bytes of the child's node structure; only relevant in a file
        def set_childPos(self, x):
            pass
        def get_siblingPos(self):
            return 0 # position in bytes of the sibling's node structure; only relevant in a file
        def set_siblingPos(self, x):
            pass
        def setup_data(self, type, rows, cols):
            pass # allocate storage, and put an array, mmap-slice or string in self.data
        def resize_data(self, nr, type_if_none):
            pass # change number of rows
        def clear_data(self):
            pass
        def set_string_data(self, s):
            pass # special bypass since you can't first allocate then slice-assign string data
        def load_rows(self, first, last, do_read):
            pass # or since we'll have them all loaded anyway, just change self.loaded*
        def unload_rows(self, do_write):
            pass # make self.loaded* = (-1, -1)
        def read_rows(self, first, last):
            pass # return a copy of the first..last rows
        def write_rows(self, first, last, arr):
            pass # copy arr into the first..last rows
        def get_rows(self, first, last):
            pass # return a slice into the actual rows if possible, or a copy
        def get_file(self):
            return None # NodeFile, if on disk
        def get_offset(self):
            return 0 # position in bytes of the node structure; only relevant in a file
        def read(self):
            pass # if on disk, read node structure and set up self.data
        def write(self):
            pass # if on disk, write node structure
        def measure(self):
            """Returns the number of bytes required for this node and its data (not children etc)."""
            return framelen + len(self.name) \
                + ((not can_data_in_node(self.dataType, self.dataCount)) and 1 or 0) * \
                    (SIZE_OF_TYPE[self.dataType]*self.rows*self.cols + \
                         ((self.dataType == QTR_TYPE_STRING) and 1 or 0))
        
    class NodeInMem(NodeStorage):
        """Manages the attributes and data of a Node that's not in a file."""
        def __init__(self, init):
            """Instantiated by Node with either a NodeStorage template to copy, or the name for a fresh new one.

Basically it uses self.__whatever for the attributes, and keeps data in either numpy.array or mmap.mmap (for strings)."""
            NodeStorage.__init__(self)
            if isinstance(init, NodeStorage):
                self.__name = init.name
                self.__dataType = init.dataType
                self.__dataSize = init.dataSize
                self.__dataCount = init.dataCount
                self.__rows = init.rows
                self.__cols = init.cols
                self.__preload = init.preload
                bounds = init.loadedRows
                self.__loadedFirst = bounds[0]
                self.__loadedLast = bounds[1]
                self.__data = None
                if self.__dataCount:
                    other_data = init.get_rows(0, init.rows-1)
                    if init.dataType == QTR_TYPE_STRING:
                        self.__data = mmap.mmap(-1, len(other_data), access=mmap.ACCESS_WRITE)
                        self.__data[:] = other_data
                    else:
                        self.__data = other_data.copy()
            elif isinstance(init, str):
                self.__name = init
                self.__dataType = QTR_TYPE_EMPTY
                self.__dataSize = 0
                self.__dataCount = 0
                self.__rows = 0
                self.__cols = 0
                self.__preload = True
                self.__loadedFirst = -1
                self.__loadedLast = -1
                self.__data = None
        def get_name(self):
            return self.__name
        def set_name(self, x):
            self.__name = x
        def get_preload(self):
            return self.__preload
        def set_preload(self, x):
            self.__preload = x
        def get_data(self):
            return self.__data
        def get_dataCount(self):
            return self.__dataCount
        def get_dataSize(self):
            return self.__dataSize
        def get_dataType(self):
            return self.__dataType
        def get_rowcount(self):
            return self.__rows
        def get_colcount(self):
            return self.__cols
        def get_loaded(self):
            if self.__loadedFirst < 0:
                return (-1, -1)
            else:
                return (self.__loadedFirst * self.cols, (self.__loadedLast + 1) * self.cols - 1)
        def get_loadedRows(self):
            return (self.__loadedFirst, self.__loadedLast)
        def get_data_item(self, i):
            if self.__dataType == QTR_TYPE_STRING:
                return self.__data[i]
            else:
                return self.__data.item(i/self.__cols, i%self.__cols)
        def set_data_item(self, i, x):
            if self.__dataType == QTR_TYPE_STRING:
                self.__data[i] = x
            else:
                self.__data.itemset(i/self.__cols, i%self.__cols, x)
        def setup_data(self, type, rows, cols):
            if (rows <= 0) or (cols <= 0):
                type = QTR_TYPE_EMPTY
                rows, cols = 0, 0
            self.__dataType = type
            self.__dataCount = rows * cols
            self.__rows = rows
            self.__cols = cols
            self.__dataSize = SIZE_OF_TYPE[type]
            if type == QTR_TYPE_STRING:
                self.__data = mmap.mmap(-1, rows*cols, access=mmap.ACCESS_WRITE)
            else:
                self.__data = numpy.array([TYPE_DEFAULTS[type]], dtype=TYPES_NUMPY[type])
                self.__data.resize((rows, cols))
            if self.__preload:
                self.__loadedFirst = min(0, self.__dataCount-1)
                self.__loadedLast = self.__rows - 1
            else:
                self.__loadedFirst = self.__loadedLast = -1
        def resize_data(self, nr, type_if_none):
            if self.__dataCount == 0:
                self.setup_data(type_if_none, nr, 1)
            else:
                if self.__dataType == QTR_TYPE_STRING:
                    self.__data.resize(nr*self.cols)
                else:
                    self.__data.resize((nr, self.cols), refcheck=0)
                self.__rows = nr
                self.__dataCount = nr * self.__cols
                self.__adjust_load()
        def __adjust_load(self):
            # in case data was resized shorter than the bound(s)
            nr = self.__rows
            if self.__dataCount:
                self.__loadedFirst = min(nr-1, self.__loadedFirst)
                self.__loadedLast = min(nr-1, self.__loadedLast)
            else:
                self.__loadedFirst = self.__loadedLast = -1
        def clear_data(self):
            self.setup_data(QTR_TYPE_EMPTY, 0, 0)
        def set_string_data(self, s):
            l = len(s)
            self.setup_data(QTR_TYPE_STRING, l, 1)
            self.data[:] = s
        def load_rows(self, first, last, do_read):
            self.__loadedFirst = first
            self.__loadedLast = last
        def unload_rows(self, do_write):
            self.__loadedFirst = self.__loadedLast = -1
        def read_rows(self, first, last):
            return self.__data[first:last+1].copy()
        def write_rows(self, first, last, arr): # flat array
            if (self.cols > 1 and isinstance(arr, list)) or (arr.shape[1] != self.cols):
                raise ValueError()
            
            if len(arr) != (last - first + 1)*self.cols:
                raise ValueError()
            self.__data[first:last+1] = arr
        def get_rows(self, first, last):
            return self.__data[first:last+1]

    class NodeFile(object):
        """Manages a memory-mapped qubtree file, either on disk or an anonymous buffer."""
        def __init__(self, path=None, buf=None, capacity=len(QTR_MAGIC), write=True):
            """Opens a file for qubtree use.

if path:
    opens the file, creating if write and not exists, and memory-maps it
elif buf:
    copies buf into a new anonymous memory map
else:
    sets up a new anonymous memory map with capacity in bytes
"""
            self.path = path
            self.buf = buf
            self.write = write
            self.mmap = None
            self.fileno = 0
            self.size = 0
            self.cap = 0
            if path:
                try:
                    exists = os.path.exists(path)
                    size = 0
                    if exists:
                        #st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime, st_mtime, st_ctime = os.stat(path)
                        size = os.stat(path)[6]
                    if write:
                        if not exists:
                            open(path, 'w').write(QTR_MAGIC)
                            size = len(QTR_MAGIC)
                            self.fileno = os.open(path, os.O_RDWR | os.O_CREAT)
                        else:
                            self.fileno = os.open(path, os.O_RDWR)
                        self.mmap = mmap.mmap(self.fileno, size, access=mmap.ACCESS_WRITE)
                    elif size:
                        self.fileno = os.open(path, os.O_RDONLY)
                        self.mmap = mmap.mmap(self.fileno, size, access=mmap.ACCESS_READ)
                    self.size = size
                    self.cap = size
                except:
                    traceback.print_exc()
            elif buf and (len(buf) >= len(QTR_MAGIC)):
                self.size = max(capacity, len(buf))
                self.cap = max(capacity, self.size)
                self.mmap = mmap.mmap(-1, self.cap, access=mmap.ACCESS_WRITE)
                self.mmap[:len(buf)] = buf
            else:
                self.size = len(QTR_MAGIC)
                self.cap = max(capacity, self.size)
                self.mmap = mmap.mmap(-1, self.cap, access=mmap.ACCESS_WRITE)
                self.mmap[:len(QTR_MAGIC)] = QTR_MAGIC
        def close(self):
            if self.mmap:
                self.mmap.close()
                self.mmap = None
            if self.fileno:
                os.close(self.fileno)
        def flush(self):
            """Flushes the memory map."""
            if self.mmap:
                self.mmap.flush()
        def alloc(self, block_size):
            # doesn't track empty spaces, just expands the file.
            offset = self.size
            if (offset + block_size) > self.cap:
                self.cap = offset + block_size
                self.mmap.resize(self.cap)
            self.size += block_size
            #print ', now %i / %i' % (self.size, self.cap)
            return offset

    def can_data_in_node(type, count):
        "Returns True if the data will fit inside the 4 byte dataPos field."""
        return 4 >= (SIZE_OF_TYPE[type] * (count + ((type == QTR_TYPE_STRING) and 1 or 0)))
    
    class NodeInFile(NodeStorage):
        """Manages the attributes and data of a Node inside a NodeFile."""
        def __init__(self, file, offset=0, template=None, name=""):
            """Instantiated by Node to either manage an existing node or create a new one within file.

if offset:
    reads in the node structure at offset bytes and set up access to its data
elif template:
    creates a new node and copy its attributes and data from NodeStorage template
else:
    creates a new empty node with name
"""
            NodeStorage.__init__(self)
            self.__file = file
            if offset:
                self.__offset = offset
                self.read()
                return
            elif template != None:
                name = name or template.name
                self.__dataType = template.dataType
                self.__dataSize = template.dataSize
                self.__dataCount = template.dataCount
                self.__rows = template.rows
                self.__cols = template.cols
                self.__preload = template.preload
                bounds = template.loadedRows
                self.__loadedFirst = bounds[0]
                self.__loadedLast = bounds[1]
            else:
                name = name
                self.__dataType = QTR_TYPE_EMPTY
                self.__dataSize = 0
                self.__dataCount = 0
                self.__rows = 0
                self.__cols = 0
                self.__preload = True
                self.__loadedFirst = -1
                self.__loadedLast = -1
            self.__dataPos = 0
            self.__childPos = 0
            self.__siblingPos = 0
            self.__data_in_node = False
            self.__name = '_@_no_@_name'
            self.set_name(name) # also sets __offset
            self.__map_data()
            if template!=None and template.dataCount:
                if self.__dataType == QTR_TYPE_STRING:
                    sdata = template.data[:]
                    self.__file.mmap[self.__dataPos:self.__dataPos+len(sdata)] = sdata
                    self.__data = sdata
                else:
                    self.__data[:] = template.data
        def get_file(self):
            return self.__file
        def get_offset(self):
            return self.__offset
        def get_name(self):
            return self.__name
        def set_name(self, x):
            """Changes the node's name.  If it's longer than the old one, the node will have to be reallocated to make room."""
            if self.__name == x: return
            self.__name = x
            namelen = len(x)
            self.__offset = self.__file.alloc(framelen+namelen)
            if namelen:
                self.__file.mmap[self.__offset+framelen:self.__offset+framelen+namelen] = self.__name
            self.clean = False
        def get_preload(self):
            return self.__preload
        def set_preload(self, x):
            if self.__preload == x: return
            self.__preload = x
            self.clean = False
        def get_data(self):
            return self.__data
        def get_dataCount(self):
            return self.__dataCount
        def get_dataSize(self):
            return self.__dataSize
        def get_dataType(self):
            return self.__dataType
        def get_dataPos(self):
            return self.__dataPos
        def get_childPos(self):
            return self.__childPos
        def set_childPos(self, x):
            if self.__childPos == x: return
            self.__childPos = x
            self.clean = False
        def get_siblingPos(self):
            return self.__siblingPos
        def set_siblingPos(self, x):
            if self.__siblingPos == x: return
            self.__siblingPos = x
            self.clean = False
        def get_rowcount(self):
            return self.__rows
        def get_colcount(self):
            return self.__cols
        def get_loaded(self):
            if self.__loadedFirst < 0:
                return (-1, -1)
            else:
                return (self.__loadedFirst * self.cols, (self.__loadedLast + 1) * self.cols - 1)
        def get_loadedRows(self):
            return (self.__loadedFirst, self.__loadedLast)
        def get_data_item(self, i):
            if self.__dataType == QTR_TYPE_STRING:
                return self.__data[i]
            else:
                return self.__data.item(i/self.__cols, i%self.__cols)
        def set_data_item(self, i, x):
            if self.__dataType == QTR_TYPE_STRING:
                self.__file.mmap[self.__dataPos+i] = x
            else:
                self.__data.itemset(i/self.__cols, i%self.__cols, x)
        def __map_data(self):
            # sets up self.__data according to type, shape, position.
            # if dataPos is 0 but count > 0, allocates space in self.file
            if self.__dataCount:
                if not self.__dataPos:
                    self.__data_in_node = can_data_in_node(self.dataType, self.dataCount)
                    if self.__data_in_node:
                        self.__dataPos = self.__offset + frame_dataoff
                    else:
                        self.__dataPos = self.__file.alloc(self.__rows*self.__cols*SIZE_OF_TYPE[self.__dataType] +
                                                         ((self.__dataType == QTR_TYPE_STRING) and 1 or 0))
                if self.__dataType == QTR_TYPE_STRING:
                    self.__data = self.__file.mmap[self.__dataPos:self.__dataPos+self.__dataCount]
                else:
                    self.__data = numpy.frombuffer(self.__file.mmap, TYPES_NUMPY[self.__dataType], self.__rows * self.__cols, self.__dataPos)
                    self.__data = self.__data.reshape((self.__rows, self.__cols))
            else:
                self.__dataPos = 0
        def setup_data(self, type, rows, cols):
            oldbytes = SIZE_OF_TYPE[self.dataType] * self.dataCount + ((self.dataType == QTR_TYPE_STRING) and 1 or 0)
            newbytes = SIZE_OF_TYPE[type] * rows * cols + ((type == QTR_TYPE_STRING) and 1 or 0)
            if (rows <= 0) or (cols <= 0):
                type = QTR_TYPE_EMPTY
                rows, cols = 0, 0
            self.__dataType = type
            self.__dataCount = rows * cols
            self.__rows = rows
            self.__cols = cols
            self.__dataSize = SIZE_OF_TYPE[type]
            if newbytes > oldbytes: # can't reuse
                self.__dataPos = 0
            self.__map_data()
            if type == QTR_TYPE_STRING:
                self.__file.mmap[self.dataPos+self.dataCount] = '\0'
            self.clean = False
            if self.__preload:
                self.__loadedFirst = min(0, self.__dataCount-1)
                self.__loadedLast = self.__rows - 1
            else:
                self.__loadedFirst = self.__loadedLast = -1
        def resize_data(self, nr, type_if_none):
            old_nr = self.rows
            if self.__dataCount == 0:
                self.setup_data(type_if_none, nr, 1)
            elif nr <= old_nr:
                self.__rows = nr
                self.__dataCount = nr * self.__cols
                self.clean = False
                self.__map_data()
            else:
                if self.__dataType == QTR_TYPE_STRING:
                    old_data = self.data[:]
                    self.setup_data(self.dataType, nr, self.cols)
                    self.__file.mmap[self.dataPos:self.dataPos+old_nr] = old_data
                else:
                    old_data = self.data.copy()
                    self.setup_data(self.dataType, nr, self.cols)
                    self.data[:old_nr] = old_data
            self.__adjust_load()
        def __adjust_load(self):
            # in case data was resized shorter than the bound(s)
            nr = self.__rows
            if self.__dataCount:
                self.__loadedFirst = min(nr-1, self.__loadedFirst)
                self.__loadedLast = min(nr-1, self.__loadedLast)
            else:
                self.__loadedFirst = self.__loadedLast = -1
        def clear_data(self):
            self.setup_data(QTR_TYPE_EMPTY, 0, 0)
        def set_string_data(self, s):
            l = len(s)
            self.setup_data(QTR_TYPE_STRING, l, 1)
            self.__file.mmap[self.dataPos:self.dataPos+l] = s
            self.__map_data() # recopy data since it's just a copy since mmap slices aren't live
        def load_rows(self, first, last, do_read):
            self.__loadedFirst = first
            self.__loadedLast = last
        def unload_rows(self, do_write):
            self.__loadedFirst = self.__loadedLast = -1
        def read_rows(self, first, last):
            if self.__dataType == QTR_TYPE_STRING:
                return self.__data[first:last+1]
            else:
                return self.__data[first:last+1].copy()
        def write_rows(self, first, last, arr):
            if (self.cols > 1 and isinstance(arr, list)) or (arr.shape[1] != self.cols):
                raise ValueError()
            
            if len(arr) != (last - first + 1)*self.cols:
                raise ValueError()
            if self.__dataType == QTR_TYPE_STRING:
                self.__file.mmap[self.dataPos+first:self.dataPos+last+1] = arr
            else:
                self.__data[first:last+1] = arr
        def get_rows(self, first, last):
            return self.__data[first:last+1]
        def read(self):
            assert(self.__offset+framelen <= len(self.__file.mmap))
            flags, a, b, type, size, count, dPos, cPos, sPos, c, d, e, f, nameLen \
                = struct.unpack(framedef, self.__file.mmap[self.__offset:self.__offset+framelen])
            self.__name = self.__file.mmap[self.__offset+framelen:self.__offset+framelen+nameLen]
            self.__dataType = type
            # on disk, "size" is the number of bytes in a row, and
            #         "count" is the number of rows
            self.__dataSize = SIZE_OF_TYPE[type]
            self.__rows = count
            self.__cols = self.__dataSize and (size / self.__dataSize) or 0
            self.__dataCount = self.__rows * self.__cols
            self.__childPos = cPos
            self.__siblingPos = sPos
            self.__preload = bool(flags and (flags & 2))
            if self.__preload and self.__dataCount:
                self.__loadedFirst = 0
                self.__loadedLast = self.__rows - 1
            else:
                self.__loadedFirst = self.__loadedLast = -1
            self.__data_in_node = bool(flags and (flags & 4))
            if self.__data_in_node:
                self.__dataPos = 0
            else:
                self.__dataPos = dPos
            # clean up QFS string data that has extra null embedded
                # as far as I know none of the offenders are 'in-node'
            if (type == QTR_TYPE_STRING) and self.__dataPos and (count > 1) and (self.__file.mmap[self.__dataPos+count-1] == '\0'):
                self.__rows -= 1
                self.__dataCount -= 1
            self.__map_data()
            self.clean = True
        def write(self):
            dataPos = self.dataPos
            if self.__data_in_node:
                dataPos = struct.unpack("i", self.__file.mmap[self.__offset+frame_dataoff:self.__offset+frame_dataoff+4])[0]
            a = b = c = d = e = f = 0
            flags = (self.preload and 2 or 0) + (self.__data_in_node and 4 or 0)
            #print framedef, flags, a, b, self.dataType, self.dataSize*self.cols, self.rows, \
            #    dataPos, self.childPos, self.siblingPos, c, d, e, f, len(self.__name)
            self.__file.mmap[self.__offset:self.__offset+framelen] = \
                struct.pack(framedef, flags, a, b, self.dataType, self.dataSize*self.cols, self.rows,
                            dataPos, self.childPos, self.siblingPos, c, d, e, f, len(self.__name))
            self.clean = True


    # top-level Node object
    class Node_numpy(object):
        """Represents the (sub-)tree under a particular node in a qub tree, in an open file or in memory.

This is the 'numpy' flavor of qub.tree.Node
"""
        def __init__(self, name="", storage=None):
            """Creates a new node, in memory, with name.

Internal use only: Manages existing NodeStorage."""
            self.__data = Data(self)
            self.storage = storage or NodeInMem(name)
            self.__parent = weakref.ref(NullNode_numpy_instance)
            self.__child = NullNode_numpy_instance
            self.__sibling = NullNode_numpy_instance
            self.__modified = storage and (not storage.clean)
            self.__ref_onDirtied = self.__onDirtied
            self.storage.OnDirtied = self.__ref_onDirtied
        def __del__(self):
            # set child nodes free as independent in-memory trees, and close any file without saving
            self.storage.OnDirtied = None
            while not self.child.isNull:
                self.remove(self.child)
            if self.storage.file:
                self.storage.file.close()
        def __repr__(self):
            return '<QUBTree Node : %s>' % self.name
        def __str__(self):
            buf = StringIO.StringIO()
            self.__re_str(buf)
            return buf.getvalue()
        def __re_str(self, fi, indent=''):
            fi.write(indent)
            # skip text flags since we give up on reading
            fi.write(self.name)
            plusindent = len(self.name)
            if self.data.count:
                fi.write(' =')
                plusindent += 2
                self.__str_data(fi, indent + plusindent*' ')
            elif self.find('QTR_LINE_COMMENT'):
                fi.write('#\\')
                fi.write(self.lineComment)
            fi.write('\n')
            
            c = self.child
            if c.name == 'QTR_LINE_COMMENT':
                c = c.sibling
            if c.isNull:
                return

            chindent = indent + '\t'
            fi.write(indent)
            fi.write('{\n')
            while c:
                if c.name != 'QTR_LINE_COMMENT':
                    nskip = 1
                    while nskip:
                        nskip = self.__str_datas(c, fi, chindent)
                        for i in xrange(nskip):
                            c = c.sibling
                    if c and (c.name != 'QTR_LINE_COMMENT'):
                        c.__re_str(fi, chindent)
                c = c.sibling
            fi.write(indent)
            fi.write('}\n')
        def __str_data(self, fi, indent):
            if self.data.type == QTR_TYPE_STRING:
                if self.storage.data.find('\r'):
                    sep = '\r\n'
                elif self.storage.data.find('\n'):
                    sep = '\n'
                else:
                    sep = None
                if sep:
                    lines = self.storage.data[:].split(sep)
                    fi.write(lines[0])
                    for i in xrange(1, len(lines)):
                        fi.write(' \\\n')
                        fi.write(lines[i])
                else:
                    fi.write(self.storage.data)
                    if self.lineComment:
                        fi.write('#\\')
                        fi.write(self.lineComment)
            elif self.data.count:
                if self.data.cols > 1:
                    fi.write('(')
                    for r in xrange(self.data.rows):
                        row = self.data.row(r)
                        for c in xrange(self.data.cols):
                            fi.write('\t')
                            fi.write(str(row[c]))
                        if r < (self.data.rows - 1):
                            fi.write('\n')
                            fi.write(indent)
                    fi.write(' )')
                else:
                    multi = self.data.count > 10
                    if multi:
                        fi.write('(')
                    for r in xrange(0, self.data.count, 10):
                        for i in xrange(min(10, self.data.count-r)):
                            if i:
                                fi.write('\t')
                            fi.write(str(self.data[r+i]))
                        if (r+10) < self.data.count:
                            fi.write('\n')
                            fi.write(indent)
                    if multi:
                        fi.write(')')
        def __str_datas(self, cf, fi, indent):
            arrlen = 0
            narr = 0
            cl = cf
            while cl:
                if (not cl.name) or (cl.name.find(' ') >= 0) or cl.child or (cl.data.cols > 1) or (cl.data.type in [QTR_TYPE_STRING, QTR_TYPE_POINTER, QTR_TYPE_UNKNOWN]):
                    break
                if not arrlen:
                    if cl.data.count < 8:
                        break
                    arrlen = cl.data.count
                elif arrlen != cl.data.count:
                    break
                narr += 1
                cl = cl.sibling
            if narr > 1:
                names = []
                seqs = []
                c = cf
                while c != cl:
                    names.append(c.name)
                    seqs.append(c.storage.data[:,0])
                    c = c.sibling
                fi.write(indent)
                fi.write('(')
                for n in names:
                    fi.write('\t')
                    fi.write(n)
                for j in xrange(arrlen):
                    fi.write('\n')
                    fi.write(indent)
                    for s in seqs:
                        fi.write('\t')
                        fi.write(str(s[j]))
                fi.write(' )\n')
                return narr
            return 0
        def print_summary(self, indent=''):
            print indent, self.name,
            if self.data.type == QTR_TYPE_STRING:
                print '=',str(self.data)
            elif 0 < self.data.count <= 10:
                print '=',str(self.data)
            elif self.data.count:
                print ':', '%d x %d' % (self.data.rows, self.data.cols)
            else:
                print
            if self.child:
                print indent,'{'
                for c in children(self):
                    c.print_summary(indent+'\t')
                print indent,'}'
        def __nonzero__(self):
            return True # a non-null node is Truthy
        def __len__(self):
            """Returns the number of child nodes."""
            count = 0
            child = self.child
            while not child.isNull:
                child = child.sibling
                count += 1
            return count
        def __getitem__(self, key):
            """Returns a child node, by number or by name.  If by name, and None, returns a newly-appended Node."""
            if isinstance(key, int):
                child = self.child
                while key:
                    child = child.sibling
                    key -= 1
                return child
            else:
                name = str(key)
                child = self.child
                prior = NullNode_numpy_instance
                while not child.isNull:
                    if child.name == name:
                        return child
                    prior = child
                    child = child.sibling
                return self.insert(name, prior)
        def __onDirtied(self, sender):
            """When storage flips from clean to unclean, sets __modified on self and all parents up to the root."""
            self.name,'dirtied'
            p = self
            while p:
                p.__modified = True
                p = p.parent
        name = property(lambda self: self.storage.name, lambda self, x: self.set_name(x), doc="the label")
        data = property(lambda self: self.__data, lambda self, x: self.set_data(x),
                        doc="""read: the Data object managing this node's data
write: replaces this node's data with the contents of a string, 
                                                        numpy.array (len(shape) <= 2),
                                                        [list of] list of numbers,
                                                        another Node.data, or
                                                        None.""")
        lineComment = property(lambda self: self.get_lineComment(), lambda self, x: self.set_lineComment(x),
                               doc="read/write the string data of the child named QTR_LINE_COMMENT")
        isNull = property(lambda self: False)
        child = property(lambda self: self.__child, doc="first child Node, or the NullNode")
        sibling = property(lambda self: self.__sibling, doc="sibling Node, or the NullNode")
        parent = property(lambda self: self.get_parent(), doc="parent Node, or the NullNode")
        path = property(lambda self: self.get_path(), doc="path of the open file containing this node, or ''")
        modified = property(lambda self: self.get_modified(), doc="whether this or any child... has changed since Open or save(as)")
        root = property(lambda self: self.get_root(), doc="the head of the tree, with no parent; possibly self.")
        def set_name(self, x):
            if (not self.storage.file) or (self != self.root):
                self.storage.name = x
        def set_data(self, x):
            if operator.isNumberType(x) and not operator.isSequenceType(x):
                x = numpy.array([x])
            if x == None:
                self.data.setup(QTR_TYPE_EMPTY, 0, 0)
            elif isinstance(x, Data):
                # copy from other Node
                if x.type == QTR_TYPE_STRING:
                    self.storage.set_string_data(str(x))
                else:
                    self.data.setup(x.type, x.rows, x.cols)
                    self.data[:] = x[:]
            elif isinstance(x, numpy.ndarray):
                if any(sh == 0 for sh in x.shape):
                    self.data.clear()
                elif len(x.shape) == 1:
                    self.data.setup(TYPE_OF_NUMPY[str(x.dtype)], x.shape[0], 1)
                    self.storage.data[:,0] = x
                else:
                    self.data.setup(TYPE_OF_NUMPY[str(x.dtype)], x.shape[0], x.shape[1])
                    if self.storage.data.shape != x.shape:
                        print self.storage.data.shape, x.shape, x
                        traceback.print_stack()
                self.storage.data[:] = x
            elif (not isinstance(x, str)) \
                 and  operator.isSequenceType(x) and all(operator.isNumberType(z) for z in x):
                # 1d list/tuple
                type = all(isinstance(k, int) for k in x) and QTR_TYPE_INT or QTR_TYPE_DOUBLE
                self.data.setup(type, len(x), 1)
                self.storage.data[:,0] = x
            elif (not isinstance(x, str)) \
                    and operator.isSequenceType(x) \
                    and all(operator.isSequenceType(z) and all(operator.isNumberType(zz) for zz in z) for z in x):
                # 2d list/tuple
                type = all(all(isinstance(k, int) for k in z) for z in x) and QTR_TYPE_INT or QTR_TYPE_DOUBLE
                self.data.setup(type, len(x), max(len(z) for z in x))
                for i,z in enumerate(x):
                    self.data.row(i)[:len(z)] = z
            else: # string or unknown written as string
                self.storage.set_string_data(str(x))
        def get_lineComment(self):
            return str(self.find("QTR_LINE_COMMENT").data)
        def set_lineComment(self, x):
            if x:
                self['QTR_LINE_COMMENT'].data = x
            else:
                self.remove(self['QTR_LINE_COMMENT'])
        def get_parent(self):
            try:
                return self.__parent() or NullNode_numpy_instance
            except:
                return NullNode_numpy_instance
        def get_path(self):
            return self.storage.file and self.storage.file.path or ''
        def get_root(self):
            root = self
            while not root.parent.isNull:
                root = root.parent
            return root
        def clone(self, deep=True, file=None):
            """Returns a copy of the (sub-)tree rooted here, copied into file, or else in-memory.  If not deep, skips the children."""
            if file:
                new_stor = NodeInFile(file, template=self.storage)
            else:
                new_stor = NodeInMem(self.storage)
            mini_me = Node_numpy(storage=new_stor)
            if deep:
                last = NullNode_numpy_instance
                for c in children(self):
                    last = mini_me.insert(c.clone(file=file), last)
            return mini_me
        def __re_store(self, file=None):
            # recursively switches the storage between file(s) and/or memory
            if file:
                self.storage = NodeInFile(file, template=self.storage)
            else:
                self.storage = NodeInMem(self.storage)
            self.__ref_onDirtied = self.__onDirtied
            self.storage.OnDirtied = self.__ref_onDirtied
            for c in children(self):
                c.__re_store(file)
        def get_modified(self):
            """Returns whether anything's different from on disk.  Adjusts child and sibling positions first in case anything's moved."""
            self.__pre_save()
            return self.__modified
        def saveAs(self, path):
            """Creates a new file at path, moves the (sub-)tree rooted here into the file, and save()s.  This node becomes the root."""
            if (not self.storage.file) or (path != self.storage.file.path):
                if os.path.exists(path):
                    os.remove(path)
                file = NodeFile(path, True)
                self.__re_store(file)
            return self.save()
        def saveAsText(self, path):
            """Writes a text representation of the (sub-)tree rooted here, to path on disk."""
            open(path, 'w').write(str(self))
        def save(self):
            """When this node is the root of an open file (self.storage.file != None), writes any modifications to disk."""
            if self != self.root:
                raise Exception("Attempt to save non-root node.")
            if not self.storage.file:
                raise Exception('save() without saveAs()')
            
            self.__pre_save()            # harmonize child, sibling pos
            self.__re_save()             # write dirty nodes
            self.storage.file.flush()
            return True
        def __pre_save(self):
            """Recursively refreshes the child and sibling positions, in case they have moved,
so nodes are accurately marked unclean and modified."""
            self.storage.childPos = self.child and self.child.storage.offset or 0
            self.storage.siblingPos = self.sibling and self.sibling.storage.offset or 0
            for c in children(self):
                c.__pre_save()
        def __re_save(self):
            """Recursively writes any modifications to disk."""
            if not self.storage.clean:
                self.storage.write()
            for c in children(self):
                c.__re_save()
            self.__modified = False
        def re_read(self):
            """Recursively reads storage and instantiates child and sibling nodes, during Open."""
            st = self.storage
            if not st.file: return
            st.read()
            if st.childPos and not self.child:
                self.__child = Node_numpy(storage=NodeInFile(st.file, offset=st.childPos))
                self.__child.__parent = weakref.ref(self)
            if st.siblingPos and not self.sibling:
                self.__sibling = Node_numpy(storage=NodeInFile(st.file, offset=st.siblingPos))
                self.__sibling.__parent = weakref.ref(self.parent)
            for c in children(self):
                c.re_read()
            self.__modified = False
        def close(self):
            """If this node is the root of an open file, moves it into memory and close the file."""
            if not self.storage.file: return True
            if self != self.root:
                raise Exception("Attempt to close non-root node.")
            file = self.storage.file
            self.__re_store()
            file.close()
            return True
        def getBytes(self):
            """Returns the binary representation of the (sub-)tree rooted here, just as it would be on disk, as a string."""
            file = NodeFile(capacity=len(QTR_MAGIC)+self.__re_measure()) # no path: anonymous mem map
            dupe = self.clone(file=file)
            dupe.save()
            return str(file.mmap[:])
        def __re_measure(self):
            """Returns the number of bytes needed to store the (sub-)tree rooted here."""
            return self.storage.measure() + sum(c.__re_measure() for c in children(self))
        def list(self):
            """Returns the names of all child Nodes."""
            return [c.name for c in children(self)]
        def find(self, name):
            """Returns the first child with name, or the NullNode."""
            child = self.child
            while not child.isNull:
                if child.name == name:
                    return child
                child = child.sibling
            return child
        def next(self, name):
            """Returns the next sibling with name, or the NullNode."""
            child = self.sibling
            while not child.isNull:
                if child.name == name:
                    return child
                child = child.sibling
            return child
        def nextSameName(self):
            """Returns the next sibling with name == self.name, or the NullNode."""
            return self.next(self.name)
        def append(self, childOrName):
            """inserts after the last child node and returns it."""
            last = self.child
            while not last.sibling.isNull:
                last = last.sibling
            return self.insert(childOrName, last)
        def insert(self, childOrName, after=None):
            """insert(str, after): inserts a new child node after the given child or first, with the given name, or '', and returns it
append(Node, after): inserts an existing node after the given child or first, first removing it from its former parent if any, and returns it"""
            if childOrName == self: return self
            if isinstance(childOrName, NullNode_numpy): return childOrName
            if isinstance(childOrName, Node_numpy):
                c = childOrName
                if not c.parent.isNull:
                    c.parent.remove(c, False)
                elif c.storage.file and (c.storage.file != self.storage.file): # root of its own open file
                    c.close()
                if c.storage.file != self.storage.file:
                    c.__re_store(self.storage.file)
                c.__parent = weakref.ref(self)
                if after:
                    c.__sibling = after.sibling
                    after.__sibling = c
                else:
                    c.__sibling = self.child
                    self.__child = c
                return c
            else:
                return self.insert(Node_numpy(name=str(childOrName)), after)
        def appendClone(self, other, deep=True):
            """Inserts a copy of other, after the last child, and returns it.  If not deep, skip other's children."""
            return self.append( other.clone(deep, self.storage.file) )
        def insertClone(self, other, deep=True, after=None):
            """Inserts a copy of other, after the given child or first, and returns it.  If not deep, skip other's children."""
            return self.insert( other.clone(deep, self.storage.file), after )
        def remove(self, child, re_store=True):
            """Removes the given child, making it its own root.

Internal use only: re_store=False: when we remove it from a file-bound tree only to re-insert it somewhere else in the same tree."""
            if not child: return
            
            if child == self.child:
                self.__child = child.sibling
                child.__parent = weakref.ref(NullNode_numpy_instance)
                if self.storage.file and re_store:
                    child.__re_store()
                return
            
            prior = self.child
            while not prior.isNull:
                if prior.sibling == child:
                    prior.__sibling = child.sibling
                    child.__parent = weakref.ref(NullNode_numpy_instance)
                    if self.storage.file and re_store:
                        child.__re_store()
                    return
                prior = prior.sibling
        def setChanged(self):
            """artifact from when data edits didn't go memmap straight to disk; had to be marked changed; does nothing."""
            pass
        def lock(self, timeoutMS=None):
            """the native flavor has a mutex, mainly to protect the ref-count.  This flavor does nothing."""
            pass
        def unlock(self):
            """the native flavor has a mutex, mainly to protect the ref-count.  This flavor does nothing."""
            pass


    def Open_numpy(path, readOnly=False):
        """Returns the root Node of the qubtree file at path, or the NullNode.  The file remains open until close()d or unref'd.

This is the 'numpy' flavor of qub.tree.Open
"""
        file = NodeFile(path=path, write=not readOnly)
        if file.size == len(QTR_MAGIC):
            root = Node_numpy(storage=NodeInFile(file))
        else:
            root = Node_numpy(storage=NodeInFile(file, offset=len(QTR_MAGIC)))
        root.re_read()
        return root

    def ReadBytes_numpy(buf):
        """Returns the root node of the qubtree copied from the binary in buf.

This is the 'numpy' flavor of qub.tree.ReadBytes
"""
        file = NodeFile(buf=buf, write=False)
        if file.size == len(QTR_MAGIC):
            root = Node_numpy(storage=NodeInFile(file))
        else:
            root = Node_numpy(storage=NodeInFile(file, offset=len(QTR_MAGIC)))
        root.re_read()
        root.close() # copy into own mem
        return root

# omitted since the originals never worked completely; were never really used.
#    def ReadText(path):
#        ##
#        pass
#    
#    def ReadString(s):
#        ##
#        pass
    
    if not have_lib:
        # this is the default flavor only if the 'native' flavor is missing
        Node = Node_numpy
        NullNode = NullNode_numpy
        Open = Open_numpy
        ReadBytes = ReadBytes_numpy
        FLAVOR = 'numpy'



    if __name__ == '__main__':        # tests for the 'numpy' flavor (node.storage is not a field in 'native')
        CHOOSE_FLAVOR('numpy')

        def presume(condition, comment=''):
            if not condition:
                raise Exception(comment)
            
        def bin_compare(lbl, apath, bpath):
            adat = open(apath, 'rb').read()
            bdat = open(bpath, 'rb').read()
            if len(adat) != len(bdat):
                print lbl,"Length diff: A:%d   B:%d" % (len(adat), len(bdat))
            for i in xrange(min(len(adat), len(bdat))):
                if adat[i] != bdat[i]:
                    print lbl,'%i:\t%s\t%s' % (i, adat[i], bdat[i])
        
        def build_links(root):
            presume(root == root)
            child = root['child']
            presume(child)
            presume(child.parent)
            presume(child.parent == root)
            left = root.insert('left')
            presume(root.child == left)
            presume(left.sibling == child)
            right = root.insert(Node('right'), child)
            presume(child.sibling == right)
            presume(right.parent == root)
            root.insertClone(root, True, left).name = 'clone'
            presume(root['clone'].sibling == child)
            root.appendClone(root, True).name = 'clown'
            presume(root.find('clown'))
            presume(root.clone().getBytes() == root.getBytes())
            root.append('child') # another
            child = root.child
            presume(child.name == 'left')
            child = child.sibling
            presume(child.name == 'clone')
            child = child.sibling
            presume(child.name == 'child')
            presume(child.next('clown').name == 'clown')
            child = child.nextSameName()
            presume(child.name == 'child')
            presume(not child.sibling)
            presume(child.sibling.isNull)
            root.remove(child)
            presume(not child.parent)
            presume(tuple(root.list()) == ('left', 'clone', 'child', 'right', 'clown'))
            presume(len(root.list()) == len(root))

        def build_data(root):
            if not root: return
            root['str'].data = 'foobar'
            presume(root['str'].data.type == QTR_TYPE_STRING)
            presume(str(root['str'].data) == 'foobar')
            root['str2'].data = root['str'].data
            presume(str(root['str2'].data) == 'foobar')
            root['str2'].data = None
            presume(root['str2'].data.count == 0)
            root.remove(root['str2'])
            root['float'].data.setup(QTR_TYPE_FLOAT, 1, 1)
            root['float'].data[0] = 4.5
            presume(root['float'].data[0] == 4.5)
            presume(str(root['float'].storage.data.dtype) == 'float32')
            root['double'].data = [3, 4.0, 5]
            presume(root['double'].data.type == QTR_TYPE_DOUBLE)
            presume(root['double'].data[2] == 5.0)
            root['1d'].data = (1, 2, 3)
            presume(root['1d'].data[0] == 1)
            root['1d'].data = [1.0, 2.0]
            presume(root['1d'].data.type == QTR_TYPE_DOUBLE)
            presume(root['1d'].data.rows == 2)
            presume(root['1d'].data.rows == root['1d'].data.count)
            presume(root['1d'].data[1] == 2.0)
            root['2d'].data = ((1,2,3),(4,5,6))
            presume(1 < root['2d'].data.rows < root['2d'].data.cols < 4)
            presume(root['2d'].data.row(1)[0] == 4)
            presume(root['2d'].data.type == QTR_TYPE_INT)
            presume(root['2d'].storage.data[0,1] == 2)
            root['2d2'].data = root['2d'].data
            open('/tmp/qtr_2d', 'wb').write(root['2d'].getBytes())
            _2dc = root['2d2'].clone()
            _2dc.name = '2d'
            open('/tmp/qtr_2d2', 'wb').write(_2dc.getBytes())
            bin_compare('2d vs 2d2', '/tmp/qtr_2d', '/tmp/qtr_2d2')
            root['2d'].data.resize(5)
            root['2d2'].data.resize(1)
            presume(root['2d'].data.count == 15)
            presume(root['2d2'].data.count == 3)
            presume(tuple(root['2d'].data[:3]) == (1,2,3))
            root['2d'].data[12:15] = (6, 6, 6)
            presume(tuple(root['2d'].storage.data[4,:]) == (6, 6, 6))
            root['2d'].data.col(2)[2:4] = (7, 8)
            presume(tuple(root['2d'].storage.data[2:4,2]) == (7, 8))
            root['clr'].data = 'foobar'
            presume(str(root['clr'].data) == 'foobar')
            root['clr'].data.clear()
            presume(root['clr'].data.type == QTR_TYPE_EMPTY)
            presume(root.data.node == root)
        

        presume(NullNode() == NullNode())
            
        root = Node('root')
        build_links(root)
        build_data(root)
        open('/tmp/qtr_gb_root', 'wb').write(root.getBytes())
        assert(not root.path)
        root.saveAs('/tmp/qtr_sa_root')
        assert(root.path == '/tmp/qtr_sa_root')
        root.close()
        assert(root.path == '')
        froot = Node('root')
        froot.saveAs('/tmp/qtr_inf_root')
        build_links(froot)
        build_data(froot)
        froot.save()
        froot.close()
        assert(root.getBytes() == froot.getBytes())

        del root
        del froot

        bin_compare('getbytes vs saveas', '/tmp/qtr_gb_root', '/tmp/qtr_sa_root')
        # these next two should differ because of the removal and the renaming:
        #bin_compare('getbytes vs buildin', '/tmp/qtr_gb_root', '/tmp/qtr_inf_root')
        #bin_compare('saveas vs buildin', '/tmp/qtr_sa_root', '/tmp/qtr_inf_root')
        # but these rebuild it compact, then test Open/ro and ReadBytes:
        f = Open('/tmp/qtr_inf_root', True)
        presume(f)
        presume(not f.modified)
        open('/tmp/qtr_infb_root', 'wb').write(f.getBytes())
        del f
        bin_compare('getbytes vs buildin/Open/getbytes', '/tmp/qtr_gb_root', '/tmp/qtr_infb_root')
        f = ReadBytes(open('/tmp/qtr_inf_root', 'rb').read())
        presume(f)
        open('/tmp/qtr_infbb_root', 'wb').write(f.getBytes())
        del f
        bin_compare('saveas vs buildin/ReadBytes/getbytes', '/tmp/qtr_gb_root', '/tmp/qtr_infb_root')

        # open, edit, save, close, reopen, check
        f = Open('/tmp/qtr_gb_root')
        f.remove(f['left'])
        f['child'].data = 4.2
        f['int'].data = 21
        presume(f.modified)
        f.save()
        f.close()
        f = Open('/tmp/qtr_gb_root', True)
        presume(not ('left' in f.list()))
        presume(f['child'].data[0] == 4.2)
        presume(f['int'].data[0] == 21)
        del f
        

# not done on purpose:
#  ReadString, ReadText, .lock(), .unlock(), mutable characters in string data in open file.




# test inaccessible files and locations; error handling
# re-group methods and classes
# what isn't explained/well?
# how much hassle for pure-fitness?

