"""pygtk support for qubtree: clipboard and visual editing"""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

from qub.tree import *
from qub.types import *
from qub.GTK import *
import qub.pyenv
import qub.pyenvGTK
import sys
import os
import re
from pprint import pprint

######### clipboard support ###############

def Copy(node, clipboard=None):
    """Copies node to the specified (def main) clipboard."""
    targets = [ ("STRING", 0, 0),
                ("TEXT", 0, 1),
                ("COMPOUND_TEXT", 0, 2),
                ("UTF8_STRING", 0, 3),
                ("QUBTREE", 0, 4) ]
    bin = [ node.getBytes() ]
    def get(clipboard, selectiondata, info, data):
        if info == 4:
            selectiondata.set("QUBTREE", 8, bin[0])
        else:
            selectiondata.set_text(str(ReadBytes(bin[0])))
    def clear(clipboard, data):
        bin[0] = ''

    clip = clipboard or gtk.clipboard_get("CLIPBOARD")
    clip.set_with_data(targets, get, clear, None)

def CanPaste(clipboard=None):
    """Returns True if there is a qub.tree.Node in the specified (def main) clipboard."""
    clip = clipboard or gtk.clipboard_get("CLIPBOARD")
    targets = clip.wait_for_targets()
    return bool(targets and ("QUBTREE" in targets))

def Paste(clipboard=None):
    """Returns a copy of the qub.tree.Node from the specified (def main) clipboard.
If there is none, or any exception occurs, node.isNull == True."""
    try:
        clip = clipboard or gtk.clipboard_get("CLIPBOARD")
        seldata = clip.wait_for_contents("QUBTREE")
        return ReadBytes(seldata.data)
    except:
        traceback.print_exc()
        return NullNode()

######### node.data <--> string (one-line entry) ###############

def acceptNumList(s):
    l = eval(s, qub.pyenv.env.globals)
    try:
        l = [0+l] # list-ify a number
    except:
        try:
            sum(l) # test they're all numbers
        except:
            sum(sum(l, [])) # or lists of numbers
    return l

def acceptNodeData(node):
    """acceptNodeData(node) -> accept
accept(string) -> action_fn
action_fn() -> modifies node.data
e.g.
try:
    acceptNodeData(node)(string)()
except:
    invalid string
"""
    def assign(x):
        def doit():
            node.data = x
        doit.val = x
        return doit
    def slice_ass(l):
        def doit():
            typ = node.data and node.data.type or QTR_TYPE_DOUBLE
            try:
                sum(l) # if it's 1-D:
                if (node.data.rows > 1) and (node.data.cols > 1):
                    node.data.setup(typ, len(l), 1)
                elif (not node.data) and (len(l) >= 1):
                    node.data = l[0]
                    node.data.resize(len(l))
                else:
                    node.data.resize(len(l))
                node.data[0:len(l)] = l
            except: # 2-D:
                nc = l and len(l[0]) or 0
                nr = len(l)
                node.data.setup(typ, nr, nc)
                for i,r in enumerate(l):
                    node.data[i*nc:(i+1)*nc] = r[0:nc]
        doit.val = l
        return doit
    def accept(s):
        if not s:
            return assign(None) # blank: clear
        elif node.data and (node.data.loadedRows != (0, node.data.rows-1)):
            raise Exception("can't handle unloaded data")
        elif node.data.type == QTR_TYPE_STRING:
            return assign(s)
        else:
            try:
                return slice_ass( acceptNumList(s) )
            except:
                if node.data:
                    raise Exception('expected number(s)')
                return assign(s)
    return accept

def captionNodeData(node):
    if node.data.type == QTR_TYPE_STRING:
        return 'Edit data for "%s" as a string:' % node.name
    elif node.data.type in [QTR_TYPE_UNKNOWN, QTR_TYPE_POINTER]:
        return 'Unknown data for "%s" -- blank to remove it.' % node.name
    elif not node.data:
        return 'Enter data for "%s" as a number, list of numbers, or a string (no quotes).' % node.name
    elif node.data.loadedRows != (0, node.data.rows-1):
        return 'Data for "%s" is not fully loaded -- blank to remove it all.' % node.name
    else:
        return 'Edit numeric data for "%s"' % node.name

ex_numeric = """Examples:
  (numeric)
    1
    1.0
    pi
    [2, 3]
    [2.0, 3.0]
    [[2.0, 3.0], [4.0, 5.0]]
    [[2.0], [3.0]]
    [float(x) for x in xrange(100)]
"""

def helpNodeData(node):
    if node.data.type == QTR_TYPE_STRING:
        return 'To change to numeric data, first clear the data and close this window.  Then enter an integer or floating point number.'
    elif node.data.type in [QTR_TYPE_UNKNOWN, QTR_TYPE_POINTER]:
        return 'details: %s\n%d x %d' % TYPENAMES[node.data.type], node.data.rows, node.data.cols
    elif not node.data:
        return ex_numeric+"""
  (string)
    whatever
"""
    elif node.data.loadedRows != (0, node.data.rows-1):
        "To edit this data, double-click the node in tree-view, or right-click it -> Properties."
    else:
        return ex_numeric+"""
data type (%s) is preserved.  If you'd like to change it, e.g. from float to int or even string, first clear the data and close this window.  Then enter a number, with or without decimal point as desired.
""" % TYPENAMES[node.data.type]

def formatNodeData(node):
    d = node.data
    t = d.type
    ld = d.loadedRows
    if t == QTR_TYPE_STRING:
        return str(node.data)
    elif t == QTR_TYPE_EMPTY:
        return ""
    elif t in [QTR_TYPE_POINTER, QTR_TYPE_UNKNOWN]:
        return "..."
    elif ld == (-1, -1):
        return "..."
    elif ld == (0, d.rows-1):
        if d.cols == 1:
            if d.rows == 1:
                return str(d[0])
            elif d.rows <= 10:
                return str([d[i] for i in xrange(d.rows)])
            else:
                return str([d[i] for i in xrange(10)]+['...'])
        else:
            r = d.row(0)
            if d.cols <= 10:
                return str([ r[:], '...']) 
            else:
                return str([ r[:10]+'...', '...'])
    else: # partial load
        return "..."

def nodeDataIsFullyShown(data):
    return not bool(re.search(r"\.\.\.", data))

######### gtk.TreeView-based browser ###############

class TreeView(gtk.TreeView):
    """gtk.Widget for browsing/editing a qub.tree.Node and its children.

properties:
    root: a qub.tree.Node
          no-one but this TreeView should edit root while it's assigned here
    changed:
          whether this TreeView has made any changes since root was assigned
    OnChange(TreeView)
"""
    def __init__(self):
        gtk.TreeView.__init__(self)
        self._build_ui()

        self.connect('button_press_event', self._onButtonPress)
        self.connect('row_activated', self._onRowActivated)
        self.itemCut.connect('activate', self._onItemCut)
        self.itemCopy.connect('activate', self._onItemCopy)
        self.itemPasteAfter.connect('activate', self._onItemPasteAfter)
        self.itemPasteChild.connect('activate', self._onItemPasteChild)
        self.itemInsertAfter.connect('activate', self._onItemInsertAfter)
        self.itemInsertChild.connect('activate', self._onItemInsertChild)
        self.itemDelete.connect('activate', self._onItemDelete)
        self.itemProperties.connect('activate', self._onItemProperties)
        self.cell_name.connect('edited', self._onEditName)
        self.cell_data.connect('edited', self._onEditData)

        self.OnChange = WeakEvent() # (TreeView)
        self.root = NullNode()
        
    def _build_ui(self):
        self.set_size_request(600, 400)
        self.popup = gtk.Menu()
        self.itemCut = gtk.MenuItem('Cut')
        self.itemCut.show()
        self.popup.append(self.itemCut)
        self.itemCopy = gtk.MenuItem('Copy')
        self.itemCopy.show()
        self.popup.append(self.itemCopy)
        self.itemPasteAfter = gtk.MenuItem('Paste After')
        self.itemPasteAfter.show()
        self.popup.append(self.itemPasteAfter)
        self.itemPasteChild = gtk.MenuItem('Paste Child')
        self.itemPasteChild.show()
        self.popup.append(self.itemPasteChild)
        self.itemInsertAfter = gtk.MenuItem('Insert After')
        self.itemInsertAfter.show()
        self.popup.append(self.itemInsertAfter)
        self.itemInsertChild = gtk.MenuItem('Insert Child')
        self.itemInsertChild.show()
        self.popup.append(self.itemInsertChild)
        self.itemDelete = gtk.MenuItem('Delete')
        self.itemDelete.show()
        self.popup.append(self.itemDelete)
        self.itemProperties = gtk.MenuItem('Properties')
        self.itemProperties.show()
        self.popup.append(self.itemProperties)
        
        self.store = gtk.TreeStore(str, str, gobject.TYPE_PYOBJECT, bool)
        self.set_model(self.store)
        self.col_name = gtk.TreeViewColumn('Name')
        self.append_column(self.col_name)
        self.cell_name = gtk.CellRendererText()
        self.cell_name.set_property('editable', True)
        self.col_name.pack_start(self.cell_name, True)
        self.col_name.add_attribute(self.cell_name, 'text', 0)
        self.col_data = gtk.TreeViewColumn('Data')
        self.append_column(self.col_data)
        self.cell_data = gtk.CellRendererText()
        #self.cell_data.set_property('editable', True)
        self.col_data.pack_start(self.cell_data, True)
        self.col_data.add_attribute(self.cell_data, 'text', 1)
        self.col_data.add_attribute(self.cell_data, 'editable', 3)

    def set_root(self, root):
        self._root = root
        self.changed = False
        self.store.clear()
        data = formatNodeData(root)
        rootI = self.store.append(None, [root.name, data, root, nodeDataIsFullyShown(data)])
        count = self.add_children(root, rootI)
        if count < 20:
            self.expand_all()
        else:
            self.expand_row((0,), False)
    root = property(lambda s: s._root, set_root)
    
    def set_changed(self, x):
        if x:
            self.OnChange(self)
        self._changed = x
    changed = property(lambda s: s._changed, set_changed)
    
    def add_children(self, node, iter):
        count = 0
        for ch in children(node):
            data = formatNodeData(ch)
            chI = self.store.append(iter, [ch.name, data, ch, nodeDataIsFullyShown(data)])
            count += 1
            count += self.add_children(ch, chI)
        return count

    def _onEditName(self, cell, path_string, new_text):
        path = tuple(re.split(':',path_string))
        if len(path) == 1: # root
            return
        node = self.store.get(self.store.get_iter(path_string), 2)[0]
        node.name = new_text
        self.store.set(self.store.get_iter(path_string), 0, new_text)
        self.changed = True
    def _onEditData(self, cell, path_string, new_text):
        node = self.store.get(self.store.get_iter(path_string), 2)[0]
        accept = acceptNodeData(node)
        def use(txt):
            accept(txt)()
            data = formatNodeData(node)
            self.store.set(self.store.get_iter(path_string), 1, data)
            self.store.set(self.store.get_iter(path_string), 3, nodeDataIsFullyShown(data))
            self.changed = True
        try:
            use(new_text)
        except:
            dlg = qub.pyenvGTK.HelpTestDialog(accept, captionNodeData(node), "data = ", helpNodeData(node),
                                              lambda x: qub.pyenv.env.globals.__setitem__('data', x.val), lambda x:'data')
            if gtk.RESPONSE_ACCEPT == dlg.run(new_text):
                try:
                    use(dlg.expr)
                except:
                    pass
            dlg.destroy()
    def _onButtonPress(self, view, event):
        if event.button == 3:
            try:
                path, col, cellx, celly = view.get_path_at_pos(int(event.x), int(event.y))
            except:
                return False
            # just in case the view doesn't alreay have focus
            view.grab_focus()
            selection = view.get_selection()
            # if this row isn't already selected, then select it before popup
            if not selection.path_is_selected(path):
                view.set_cursor( path, col, False)
            self.itemPath = path
            self.itemCopy.set_sensitive(True)
            self.itemPasteChild.set_sensitive(CanPaste())
            self.itemInsertChild.set_sensitive(True)
            self.itemCut.set_sensitive(len(path)>1)
            self.itemPasteAfter.set_sensitive(len(path)>1 and CanPaste())
            self.itemInsertAfter.set_sensitive(len(path)>1)
            self.itemDelete.set_sensitive(len(path)>1)
            self.itemProperties.set_sensitive(True)
            self.popup.popup(None, None, None, event.button, event.time)
            return True
        return False
    def _onRowActivated(self, tree, path, col):
        self.itemPath = path
        self._onItemProperties(self.itemProperties)
    def _onItemCut(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        if node == self.root:
            return
        Copy(node)
        node.parent.remove(node)
        self.store.remove(hereI)
        self.changed = True
    def _onItemCopy(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        Copy(node)
    def _onItemPasteAfter(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        if node == self.root:
            return
        parentI = self.store.get_iter(self.itemPath[:-1])
        try:
            sib = Paste()
            if sib.isNull:
                return
            node.parent.insert(sib, node)
            data = formatNodeData(sib)
            chI = self.store.insert_after(parentI, hereI, [sib.name, data, sib, nodeDataIsFullyShown(data)])
            self.add_children(sib, chI)
            self.set_cursor(self.store.get_path(chI), self.col_name, False)
            self.changed = True
        except:
            traceback.print_exc()
    def _onItemPasteChild(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        try:
            ch = Paste()
            if ch.isNull:
                return
            node.insert(ch)
            data = formatNodeData(ch)
            chI = self.store.insert_after(hereI, None, [ch.name, data, ch, nodeDataIsFullyShown(data)])
            self.add_children(ch, chI)
            self.expand_row(self.itemPath, False)
            self.set_cursor(self.store.get_path(chI), self.col_name, 0)
            self.changed = True
        except:
            traceback.print_exc()
    def _onItemInsertAfter(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        if node == self.root:
            return
        parentI = self.store.get_iter(self.itemPath[:-1])
        sib = node.parent.insert('untitled', node)
        data = formatNodeData(sib)
        chI = self.store.insert_after(parentI, hereI, [sib.name, data, sib, nodeDataIsFullyShown(data)])
        self.set_cursor(self.store.get_path(chI), self.col_name, True)
        self.changed = True
    def _onItemInsertChild(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        ch = node.insert('untitled')
        data = formatNodeData(ch)
        chI = self.store.insert_after(hereI, None, [ch.name, data, ch, nodeDataIsFullyShown(data)])
        self.expand_row(self.itemPath, False)
        self.set_cursor(self.store.get_path(chI), self.col_name, True)
        self.changed = True
    def _onItemDelete(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        if node == self.root:
            return
        node.parent.remove(node)
        self.store.remove(hereI)
        self.changed = True
    def _onItemProperties(self, item):
        hereI = self.store.get_iter(self.itemPath)
        node = self.store.get(hereI, 2)[0]
        dlg = NodeDialog()
        dlg.run(node)
        dlg.destroy()
        self.store.set(hereI, 0, node.name)
        data = formatNodeData(node)
        self.store.set(hereI, 1, data)
        self.store.set(hereI, 3, nodeDataIsFullyShown(data))
        self.changed = True


######### node properties widget and dialog ###############

class NodeView(gtk.VBox):
    """Properties panel for a single qub.tree.Node.

properties:
    node: a qub.tree.Node
          do not edit node's name or data while it is assigned here
"""
    def __init__(self):
        gtk.VBox.__init__(self)
        self._build_ui()

        self.ref = Reffer()
        self.txtName.OnChange += self.ref(self._onChangeName)
        self.mnuType.OnPopulate += self.ref(self._onPopulateType)
        self.mnuType.OnChanged += self.ref(self._onChangeType)
        self.chkPreload.connect('toggled', self._onTogglePreload)
        self.txtRows.OnChange += self.ref(self._onChangeRows)
        self.txtCols.OnChange += self.ref(self._onChangeCols)
        self.btnExport.connect('clicked', self._onPressExport)
        self.btnImport.connect('clicked', self._onPressImport)
        self.txtLoadFrom.OnChange += self.ref(self._onChangeLoadFrom)
        self.txtLoadTo.OnChange += self.ref(self._onChangeLoadTo)
        self.txtData.connect('focus_out_event', self._onFocusOutData)
        self.btnApply.connect('clicked', self._onPressApply)
        
        self.node = None

    def _build_ui(self):
        self.set_size_request(500, 200)
        self.header = gtk.HBox(True)
        h = gtk.HBox()
        lbl = gtk.Label('Name')
        h.pack_start(lbl, False, True)
        lbl.show()
        self.txtName = NumEntry("")
        h.pack_start(self.txtName, True, True)
        self.txtName.show()
        self.header.pack_start(h)
        h.show()
        h = gtk.HBox()
        lbl = gtk.Label('Data type:')
        h.pack_start(lbl, False, True)
        lbl.show()
        self.mnuType = DynamicComboBox()
        h.pack_start(self.mnuType, True, True)
        self.mnuType.show()
        self.header.pack_start(h)
        h.show()
        self.pack_start(self.header, False, True)
        self.header.show()
        self.header2 = gtk.HBox()
        h = self.header2
        self.chkPreload = gtk.CheckButton('Preload')
        h.pack_start(self.chkPreload, False, True)
        self.chkPreload.show()
        self.panRC = gtk.HBox()
        self.txtRows = NumEntry(1, acceptIntGreaterThan(-1))
        self.txtRows.set_width_chars(4)
        self.panRC.pack_start(self.txtRows, False, True)
        self.txtRows.show()
        lbl = gtk.Label(' x ')
        self.panRC.pack_start(lbl, False, True)
        lbl.show()
        self.txtCols = NumEntry(1, acceptIntGreaterThan(-1))
        self.txtCols.set_width_chars(4)
        self.panRC.pack_start(self.txtCols, False, True)
        self.txtCols.show()
        self.btnExport = gtk.Button('Export...')
        self.btnExport.connect('clicked', self._onPressExport)
        self.panRC.pack_start(self.btnExport, False, True)
        self.btnExport.show()
        self.btnImport = gtk.Button('Import...')
        self.panRC.pack_start(self.btnImport, False, True)
        self.btnImport.show()
        h.pack_start(self.panRC, True, True)
        self.panLoad = gtk.HBox()
        lbl = gtk.Label('Load:')
        self.panLoad.pack_start(lbl, False, True)
        lbl.show()
        self.txtLoadFrom = NumEntry(1, acceptIntGreaterThan(-2))
        self.txtLoadFrom.set_width_chars(4)
        self.panLoad.pack_start(self.txtLoadFrom, False, True)
        self.txtLoadFrom.show()
        lbl = gtk.Label('-')
        self.panLoad.pack_start(lbl)
        lbl.show()
        self.txtLoadTo = NumEntry(1, acceptIntGreaterThan(-2))
        self.txtLoadTo.set_width_chars(4)
        self.panLoad.pack_start(self.txtLoadTo, False, True)
        self.txtLoadTo.show()
        h.pack_start(self.panLoad, False, True)
        self.pack_start(h, False, True)
        h.show()
        
        scr = gtk.ScrolledWindow()
        scr.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.txtData = gtk.TextView()
        self.txtDataOut = TextViewAppender(self.txtData)
        scr.add(self.txtData)
        self.txtData.show()
        self.pack_start(scr, True, True)
        scr.show()
        
        self.buttons = gtk.HBox()
        h = self.buttons
        self.btnApply = gtk.Button('Apply')
        self.btnApply.set_sensitive(False)
        h.pack_end(self.btnApply, False, True)
        self.btnApply.show()
        self.pack_start(h, False, True)
        h.show()

    def set_node(self, node):
        self._node = node
        self.refresh()
    node = property(lambda s: s._node, set_node)
    def refresh(self):
        node = self.node
        if (node == None) or node.isNull:
            self.header.set_sensitive(False)
            self.txtData.set_sensitive(False)
            self.header2.hide_all()
            self.panLoad.hide_all()
        else:
            self.header.set_sensitive(True)
            self.header2.show_all()
            self.txtName.set_editable((not node.path) or (not node.parent.isNull))
            self.txtName.value = node.name
            self.mnuType.choose(TYPENAMES[node.data.type])
            self.chkPreload.set_active(node.data.preload)
            self.txtData.set_sensitive(bool(node.data))
            if node.data.type == QTR_TYPE_EMPTY:
                self.panRC.hide_all()
                self.panLoad.hide_all()
                self.txtData.get_buffer().set_text("")
            elif node.data.type in [QTR_TYPE_UNKNOWN, QTR_TYPE_POINTER]:
                self.panRC.hide_all()
                self.panLoad.hide_all()
                self.txtData.get_buffer().set_text("Can't edit this data type; choose Data-type:empty to clear.")
            elif node.data.type == QTR_TYPE_STRING:
                self.panRC.hide_all()
                self.panLoad.hide_all()
                self.txtData.get_buffer().set_text(str(node.data))
            else: # numeric
                self.panRC.show_all()
                self.txtRows.value = node.data.rows
                self.txtCols.value = node.data.cols
                if node.data.preload:
                    self.panLoad.hide_all()
                    node.data.loadRows()
                    if node.data.count == 1:
                        lst = node.data[0]
                    elif node.data.cols == 1:
                        lst = node.data[:]
                    else:
                        lst = [node.data.row(i)[0:node.data.cols] for i in xrange(node.data.rows)]
                else:
                    la, lb = node.data.loadedRows
                    self.panLoad.show_all()
                    self.txtLoadFrom.value = la
                    self.txtLoadTo.value = lb
                    if la == -1:
                        lst = []
                    else:
                        lst = [node.data.row(i)[0:node.data.cols] for i in xrange(la, lb+1)]
                self.txtData.get_buffer().set_text("")
                pprint(lst, self.txtDataOut)
        self.txtData.modify_text(gtk.STATE_NORMAL, gdk.color_parse("#000000"))
    def reshape(self, type, rows, cols):
        node = self.node
        if type == QTR_TYPE_EMPTY:
            node.data = None
            return True
        elif type in [QTR_TYPE_UNKNOWN, QTR_TYPE_POINTER]:
            gobject.idle_add(self.mnuType.choose, TYPENAMES[node.data.type])
            return False
        elif type == QTR_TYPE_STRING:
            buf = self.txtData.get_buffer()
            node.data = buf.get_text(buf.get_start_iter(), buf.get_end_iter()) or "edit_me"
            return True
        else:
            if node.data.type in [QTR_TYPE_EMPTY, QTR_TYPE_UNKNOWN, QTR_TYPE_POINTER]:
                node.data.setup(type, 1, 1)
                node.data[0] = 0
            elif node.data.type == QTR_TYPE_STRING:
                node.data.setup(type, 1, 1)
                self.reparse()
            elif (node.data.type == type) and (node.data.cols == cols):
                node.data.resize(rows)
            else:
                clone = node.clone()
                pre = node.data.preload
                ldb = node.data.loadedRows
                node.data.preload = True
                node.data.setup(type, rows, cols)
                for i in xrange(min(rows, node.data.rows)):
                    node_row = node.data.row(i)
                    clone_row = clone.data.row(i)
                    for j in xrange(min(len(node_row), len(clone_row))):
                        node_row[j] = clone_row[j]
                node.data.preload = pre
                if pre:
                    node.data.loadRows()
                else:
                    node.data.loadRows(*ldb)
            return True
    def reparse(self):
        node = self.node
        buf = self.txtData.get_buffer()
        txt = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
        if node.data.type == QTR_TYPE_STRING:
            node.data = txt
            self.btnApply.set_sensitive(False)
            self.txtData.modify_text(gtk.STATE_NORMAL, gdk.color_parse("#000000"))
        else:
            try:
                lst = acceptNumList(txt or '[]')
                try: # make uniform 2d:
                    sum(lst)
                    lst = [[x] for x in lst]
                except:
                    pass
                la, lb = node.data.loadedRows
                if la == -1:
                    return
                if node.data.preload:
                    self.reshape(node.data.type, len(lst), len(lst) and len(lst[0]))
                    la, lb = node.data.loadedRows
                for i in xrange(min(len(lst), lb-la+1)):
                    lst_row = lst[i]
                    node_row = node.data.row(la+i)
                    for j in xrange(min(node.data.cols, len(lst_row))):
                        node_row[j] = lst_row[j]
                self.txtData.modify_text(gtk.STATE_NORMAL, gdk.color_parse("#000000"))
                self.btnApply.set_sensitive(False)
            except:
                traceback.print_exc()
    def _onChangeName(self, txt, val):
        self.node.name = val
    def _onPopulateType(self, add):
        for t in TYPES:
            add(TYPENAMES[t])
    def _onChangeType(self, mnu, val):
        self.reshape(self.mnuType.active_i, max(1, self.node.data.rows), max(1, self.node.data.cols))
        self.refresh()
    def _onTogglePreload(self, chk):
        self.node.data.preload = chk.get_active()
        self.refresh()
    def _onChangeRows(self, txt, val):
        self.reshape(self.node.data.type, val, self.node.data.cols)
        self.refresh()
    def _onChangeCols(self, txt, val):
        self.reshape(self.node.data.type, self.node.data.rows, val)
        self.refresh()
    def _onPressExport(self, btn):
        dlg = gtk.FileChooserDialog('Export data as...', None, gtk.FILE_CHOOSER_ACTION_SAVE,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("Comma-separated files")
        filter.add_pattern("*.csv")
        dlg.add_filter(filter)
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            try:
                fi = open(dlg.get_filename(), "w")
                self._export(fi)
            except Exception, e:
                print "Saving %s: %s" % (dlg.get_filename(), e)
        dlg.destroy()
    def _onPressImport(self, btn):
        dlg = gtk.FileChooserDialog('Import data...', None, gtk.FILE_CHOOSER_ACTION_OPEN,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("Comma-separated files")
        filter.add_pattern("*.csv")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dlg.add_filter(filter)
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            try:
                self._import(open(dlg.get_filename(), "r"))
            except:
                traceback.print_exc()
        dlg.destroy()
    def _export(self, fi):
        node = self.node
        data = node.data
        la, lb = node.data.loadedRows
        if la == 0 and lb == data.rows-1:
            for i in xrange(data.rows):
                fi.write(','.join(map(str, data.row(i)[0:data.cols])))
                fi.write('\n')
        else:
            atatime = 1000000 / node.data.size
            for ii in xrange(0, data.rows, atatime):
                node.data.loadRows(ii,ii+atatime-1)
                for i in xrange(ii, min(data.rows,ii+atatime)):
                    fi.write(','.join(map(str, data.row(i)[0:data.cols])))
                    fi.write('\n')
            if la >= 0:
                data.loadRows(la, lb) 
            else:
                data.unloadRows()
    def _import(self, fi):
        self.txtData.get_buffer().set_text("")
        out = self.txtDataOut
        out.write('[')
        first = True
        for line in fi:
            if first:
                first = False
            else:
                out.write(',\n')
            if line:
                out.write('[%s]' % line.strip())
        out.write(']')
        self.reparse()
        self.refresh()
    def _onChangeLoadFrom(self, txt, val):
        node = self.node
        ld_to = node.data.loadedRows[1]
        if val >= 0 and ld_to < val:
            ld_to = val
        if val >= 0:
            node.data.loadRows(val, ld_to)
        else:
            node.data.unloadRows()
        self.refresh()
    def _onChangeLoadTo(self, txt, val):
        node = self.node
        ld_from = node.data.loadedRows[0]
        if val >= 0 and ld_from > val:
            ld_from = val
        if val >= 0:
            node.data.loadRows(ld_from, val)
        else:
            node.data.unloadRows()
        self.refresh()
    def _onChangeData(self, buf):
        self.txtData.modify_text(gtk.STATE_NORMAL, gdk.color_parse("#FF0000"))
        self.btnApply.set_sensitive(True)
    def _onFocusOutData(self, txt, evt):
        self.reparse()
        self.refresh()
    def _onPressApply(self, btn):
        self.reparse()
        self.refresh()

class NodeDialog(gtk.Dialog):
    """A dialog box to hold the NodeView for editing a single qub.tree.Node.

methods:
    NodeDialog(parent=None, appName="")
        parent: a gtk.Window to center the dialog upon
        appName: host application name, for the title bar
    run(node)
        show node's properties, modally
"""
    def __init__(self, parent=None, appName=""):
        gtk.Dialog.__init__(self, "", parent, gtk.DIALOG_MODAL)
        appPrefix = appName and (appName+" - ") or ""
        self.title_prefix = appPrefix+"Edit data"
        self.set_title(self.title_prefix)
        
        self.view = NodeView()
        self.vbox.pack_start(self.view, True, True)
        self.view.show()

        self.add_buttons(gtk.STOCK_CLOSE, gtk.RESPONSE_REJECT)
    def run(self, node):
        self.view.node = node
        self.set_title(self.title_prefix + " for "+node.name)
        return gtk.Dialog.run(self)

                                        
######### top-level window ###############

class TreeWindow(gtk.Window):
    """A top-level gtk.Window with a TreeView for editing a qub.tree.Node and its children.

methods:
    TreeWindow(stand_alone=False, appName="")
        stand_alone: has menus, prompts to save, quits on close
        appName: host application name, for the title bar

properties:
    view: the TreeView
    root: view.root
"""
    def __init__(self, stand_alone=False, appName=""):
        gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)
        self.stand_alone = stand_alone
        appPrefix = appName and (appName+" - ") or ""
        self.title_prefix = appPrefix+"Tree Editor"
        self.set_title(self.title_prefix)
        self._build_ui()

        self.connect('delete_event', self._onDelete)
        self.connect('destroy', self._onDestroy)

        self.root = Node('Untitled')

    def _build_ui(self):
        v = gtk.VBox()
        if self.stand_alone:
            self._buildMenus()
            v.pack_start(self.menubar, False, True)
            self.menubar.show()
        scr = gtk.ScrolledWindow()
        scr.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.view = TreeView()
        scr.add(self.view)
        self.view.show()
        v.pack_start(scr, True, True)
        scr.show()
        self.add(v)
        v.show()
        
    def set_root(self, root):
        self._root = root
        self.view.root = root
    root = property(lambda s: s._root, set_root)
    def open(self, path):
        try:
            root = Open(path)
            if not root.isNull:
                self.root = root
            self.set_title(self.title_prefix+" - "+self.get_filename())
        except:
            traceback.print_exc()
    def promptSave(self):
        if self.stand_alone and self.view.changed:
            filename = self.get_filename()
            dlg = gtk.MessageDialog(self, buttons=gtk.BUTTONS_NONE,
                                    message_format='Save changes to %s (%s)?' % (filename, self.root.name))
            dlg.add_buttons('Yes',gtk.RESPONSE_YES,
                            'No',gtk.RESPONSE_NO,
                            'Cancel',gtk.RESPONSE_CANCEL)
            response = dlg.run()
            dlg.destroy()
            if response == gtk.RESPONSE_YES:
                if not self.save(): return self.promptSave() # ask again on failure
            elif response == gtk.RESPONSE_CANCEL:
                return False
            elif response == gtk.RESPONSE_NO:
                pass
        return True
    def save(self, ipath=None):
        try:
            path = ipath or self.root.path or ""
            if not path:
                return self.saveAs()
            if path == self.root.path:
                if self.root.save():
                    self.view.changed = False
                    return True
            if self.root.saveAs(path):
                self.view.changed = False
                return True
            return False
        except Exception, e:
            dlg = gtk.MessageDialog(self, buttons=gtk.BUTTONS_OK,
                                    message_format=str(e))
            dlg.run()
            dlg.destroy()
            return False
    def saveAs(self, ipath=None):
        path = ipath or self.root.path or ""
        dlg = gtk.FileChooserDialog('Save as...', self, gtk.FILE_CHOOSER_ACTION_SAVE,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        dlg.set_filename(path)
        filter = gtk.FileFilter()
        filter.set_name("all files")
        filter.add_pattern("*.*")
        dlg.add_filter(filter)
        dlg.set_current_folder(self.get_path())
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            try:
                self.root.saveAs(dlg.get_filename())
                self.view.changed = False
                self.set_title(self.title_prefix+" - "+self.get_filename())
            except Exception, e:
                mdlg = gtk.MessageDialog(self, buttons=gtk.BUTTONS_OK, message_format=str(e))
                mdlg.run()
                mdlg.destroy()
                response = gtk.RESPONSE_CANCEL
        dlg.destroy()
        return response == gtk.RESPONSE_OK
    def saveAsText(self, ipath=None):
        path = ipath or self.root.path or ""
        print path, os.path.splitext(path)
        path = os.path.splitext(path)[0] + '.txt'
        print path
        dlg = gtk.FileChooserDialog('Save as...', self, gtk.FILE_CHOOSER_ACTION_SAVE,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        dlg.set_filename(path)
        filter = gtk.FileFilter()
        filter.set_name("text files")
        filter.add_pattern("*.txt")
        dlg.add_filter(filter)
        dlg.set_current_folder(self.get_path())
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            try:
                path = os.path.splitext(dlg.get_filename())[0] + '.txt'
                self.root.saveAsText(path)
            except Exception, e:
                mdlg = gtk.MessageDialog(self, buttons=gtk.BUTTONS_OK, message_format=str(e))
                mdlg.run()
                mdlg.destroy()
                response = gtk.RESPONSE_CANCEL
        dlg.destroy()
        return response == gtk.RESPONSE_OK
    def get_path(self):
        if self.root.path:
            return os.path.split(self.root.path)[0]
        else:
            return ""
    def get_filename(self):
        if self.root.path:
            return os.path.split(self.root.path)[1]
        else:
            return ""
    def _onDelete(self, w, evt):
        if self.promptSave():
            if self.stand_alone:
                return False # they destroy
            else:
                self.hide()
        return True # we handle
    def _onDestroy(self, window):
        if self.stand_alone:
            gtk.main_quit()
    def _buildMenus(self):
        self.mnuFile = gtk.Menu()
        self.itemNew = gtk.MenuItem("New")
        self.itemNew.connect('activate', self._onItemNew)
        self.itemNew.show()
        self.mnuFile.append(self.itemNew)
        self.itemPasteNew = gtk.MenuItem("Paste as new document")
        self.itemPasteNew.connect('activate', self._onItemPasteNew)
        self.itemPasteNew.show()
        self.mnuFile.append(self.itemPasteNew)
        self.itemOpen = gtk.MenuItem("Open...")
        self.itemOpen.connect('activate', self._onItemOpen)
        self.itemOpen.show()
        self.mnuFile.append(self.itemOpen)
        self.itemSave = gtk.MenuItem('Save')
        self.itemSave.connect('activate', self._onItemSave)
        self.itemSave.show()
        self.mnuFile.append(self.itemSave)
        self.itemSaveAs = gtk.MenuItem('Save As...')
        self.itemSaveAs.connect('activate', self._onItemSaveAs)
        self.itemSaveAs.show()
        self.mnuFile.append(self.itemSaveAs)
        self.itemSaveAsText = gtk.MenuItem('Save a copy as text...')
        self.itemSaveAsText.connect('activate', self._onItemSaveAsText)
        self.itemSaveAsText.show()
        self.mnuFile.append(self.itemSaveAsText)
        self.itemQuit = gtk.MenuItem(self.stand_alone and 'Quit' or 'Close')
        self.itemQuit.connect('activate', self._onItemQuit)
        self.itemQuit.show()
        self.mnuFile.append(self.itemQuit)

        self.menubar = gtk.MenuBar()
        self.mnuItemFile = gtk.MenuItem('File')
        self.mnuItemFile.show()
        self.mnuItemFile.set_submenu(self.mnuFile)
        self.mnuItemFile.connect('activate', self._onShowMnuFile)
        self.menubar.append(self.mnuItemFile)
    def _onShowMnuFile(self, item):
        self.itemPasteNew.set_sensitive(CanPaste())
        self.itemSave.set_sensitive((not self.root.path) or bool(self.root.modified))
    def _onItemNew(self, item):
        self.root = Node('Untitled')
        self.set_title(self.title_prefix)
    def _onItemPasteNew(self, item):
        node = Paste()
        if node.isNull: return
        self.root = node
        self.set_title(self.title_prefix)
    def _onItemOpen(self, item):
        dlg = gtk.FileChooserDialog('Open...', self, gtk.FILE_CHOOSER_ACTION_OPEN,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dlg.add_filter(filter)
        dlg.set_current_folder(self.get_path())
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            self.open(dlg.get_filename())
        dlg.destroy()
    def _onItemSave(self, item):
        self.save()
    def _onItemSaveAs(self, item):
        self.saveAs()
    def _onItemSaveAsText(self, item):
        self.saveAsText()
    def _onItemQuit(self, item):
        if self.promptSave():
            if self.stand_alone:
                self.destroy()
            else:
                self.hide()

######### dialog and function ###############

class TreeDialog(gtk.Dialog):
    """Dialog for modal editing of a qub.tree.Node and its children.

methods:
    TreeDialog(parent=None, appName="", caption="")
        parent: a gtk.Window to center the dialog upon
        appName: host application name, for the title bar
        caption: optional text for the top of the dialog

    run(node)
        Shows the dialog modally, editing node

properties:
    view: the TreeView
    root: view.root
"""
    def __init__(self, parent=None, appName="", caption=""):
        gtk.Dialog.__init__(self, "", parent, gtk.DIALOG_MODAL)
        appPrefix = appName and (appName+" - ") or ""
        self.title_prefix = appPrefix+"Tree Editor"
        self.set_title(self.title_prefix)
        
        v = self.vbox
        if caption:
            lbl = gtk.Label(caption)
            v.pack_start(lbl, False, True)
            lbl.show()
        scr = gtk.ScrolledWindow()
        scr.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.view = TreeView()
        scr.add(self.view)
        self.view.show()
        v.pack_start(scr, True, True)
        scr.show()
        
        self.root = Node('Untitled')
    def set_root(self, root):
        self._root = root
        self.view.root = root
    root = property(lambda s: s._root, set_root)
    def run(self, node):
        self.root = node
        return gtk.Dialog.run(self)

def UserEdit(node, parent=None, appName="", caption=""):
    """Creates a TreeDialog, runs it, and destroys it; returns the edited node."""
    dlg = TreeDialog(parent, appName, caption)
    dlg.run(node)
    dlg.destroy()
    return node



if __name__ == '__main__':
    path = os.path.join(os.path.expanduser('~'), '.qubtree')
    qub.pyenv.Init(path)
    
    w = TreeWindow(stand_alone=True)
    w.open(sys.argv[1])
    w.show()
    
    gtk.main()
    
    # UserEdit(w.root, caption="This is demonstrating simple UserEdit() dialog style")

