"""General-purpose types and utilities"""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import weakref
import traceback
from math import *
from string import join
import re

VERSION = "1.3.6"

ShowDropped = False

UNSET_VALUE = 86738.82583


class Event(object):
    """
    Usage:
       myEvent = Event() # a, b
       callback1 = lambda a, b: dostuff(a, b)
       callback2 = lambda a, b: otherstuff(a, b)
       myEvent += callback1
       myEvent += callback2
       myEvent(a, b) # calls both callbacks
       myEvent -= callback1
       myEvent -= callback2
       myEvent(a, b) # no callbacks; does nothing
       
    Note: this is a great source of circular references; see gc.collect()
    """
       
    def __init__(self):
        self._listeners = []
        self.enabled = True
    def __call__(self, *args, **kw):
        if self.enabled:
            for l in self._listeners:
                l(*args, **kw)
    def __iadd__(self, l):
        if not l in self._listeners:
            self._listeners.append(l)
        return self
    def __isub__(self, l):
        if l in self._listeners:
            self._listeners.remove(l)
        return self

class Stop(Exception):
    def __init__(self):
        Exception.__init__(self, 'stop requested')

    
# this makes a good stand-in for any callback
def printer(*xx):
    if len(xx) == 1:
        print xx[0]
        return xx[0]
    else:
        for x in xx:
            print x,
        print
    return xx

class WeakEvent(object):
    """
    Usage:
       myEvent = Event() # a, b
       # note: keep a ref somewhere, even to self.whatever, or it will be unlinked; see Reffer
       callback1 = lambda a, b: dostuff(a, b)
       callback2 = lambda a, b: otherstuff(a, b)
       callback3 = self.callback3
       myEvent += callback1
       myEvent += callback2
       myEvent(a, b) # calls both callbacks
       myEvent -= callback1
       myEvent -= callback2
       myEvent(a, b) # no callbacks; does nothing
    """
       
    def __init__(self, name=""):
        self._listeners = []
        self._names = []
        self.name = name
        self.enabled = True
    def __call__(self, *args, **kw):
        dead = []
        if self.enabled:
            for l in self._listeners:
                listener = l()
                if listener:
                    try:
                        listener(*args, **kw)
                    except:
                        traceback.print_exc()
                else:
                    dead.append(l)
        for l in dead:
            i = self._listeners.index(l)
            n = self._names[i]
            if self.name or n or ShowDropped:
                print "%s%sdropped callback%s%s" % (self.name, self.name and ' ', n and ': ', n)
            self._listeners.pop(i)
            self._names.pop(i)
    def __iadd__(self, ln):
        try:
            l, n = ln
        except:
            l, n = ln, ""
        if not l in self._listeners:
            self._listeners.append(weakref.ref(l)) # , lambda l:printer("released "+n)))
            self._names.append(n)
        return self
    def __isub__(self, lsub):
        dead = []
        for l in self._listeners:
            listener = l()
            if (not listener) or (listener == lsub):
                dead.append(l)
        for l in dead:
            i = self._listeners.index(l)
            self._listeners.pop(i)
            self._names.pop(i)
        if not len(dead): # no matching event found
            raise Exception('-=!~')
        return self

# for WeakEvent, we need to keep bound-method references somewhere:
# self.r = Reffer()
# self.foo.OnFoo += self.r(self._onFoo)

class Reffer(object):
    """Holds references to keep objects alive.
       Intended in combo with WeakEvent for bound method references e.g. self.onEvent,
       since self.onEvent creates a temporary instance that would otherwise be destroyed.

Usage:
    self.ref = Reffer()
    self.evt = WeakEvent()
    self.evt += self.ref(self.onEvt)
    ...
    self.evt -= self.ref(self.onEvt) # hashes the same each time
"""
    def __init__(self):
        self.r = {}
    def __call__(self, ref):
        if not self.r.has_key(ref):
            self.r[ref] = ref
        return self.r[ref]

class DeathRattle(object):
    """Prints to stdout when it is finally freed, presumably at the same time as its owner.

Usage:
    self.death_rattle = DeathRattle('Free at last')
"""    
    def __init__(self, rattle):
        self.rattle = rattle
    def __del__(self):
        print self.rattle

class Anon(object):
    """An object whose fields are the constructor keyword arguments.
Requesting any other field returns None rather than raise NameError.

Example:
    x = Anon(stuff='x',n=5)
    print x.stuff, x.n, x.whatever

x, 5, None
"""
    def __init__(self, **kw):
        for k,v in kw.iteritems():
            self.__dict__[k] = v
    # access missing field returns None instead of raising NameError
    def __getattr__(self, name):
        try:
            return self.__dict__[name]
        except:
            return None


"""
Here is a technique to estimate mean and std.dev of an unbounded sequence:
    
The basic equation you need for a single pass update is:
new S2 = old S2 + (x - old_M).(x - new_M)
where S2 = sum of squares of deviations from the mean M.

It is implemented as follows:
Start: n = 0, M = 0, S2 = 0
Then for each observation x:
dev = x - M
n = n + 1
M = M + dev/n
S2 = S2 + dev.(x - M)

When you want to calculate the sample std. devn. :
std.dev. = sqrt( S2 / (n-1) )
--
Alan Miller, Retired Scientist (Statistician)
CSIRO Mathematical & Information Sciences
a...@vic.cmis.csiro.au, mille...@ozemail.com.au
http://www.ozemail.com.au/~milleraj
"""

class RunningMean(object):
    """Accumulates the approximate mean and standard deviation of a partial dataset.

Example:
    accumulator = RunningMean()
    for x in dataset:
        accumulator.add(x)
        print "after %d points: %f +/- %f" % (accumulator.n, accumulator.mean, accumulator.stdDev)
"""
    def __init__(self):
        self.n = 0
        self.mean = 0
        self.s2 = 0
    def add(self, x):
        dev = x - self.mean
        self.n += 1
        self.mean += dev/self.n
        self.s2 += dev*(x - self.mean)
    def get_std(self):
        return (self.n > 1) and sqrt(self.s2 / (self.n - 1)) or 0.0
    stdDev = property(get_std)


def breduce(func, seq):
    """like builtin reduce(), but recursing by splitting in half.  Doesn't seem to make a difference."""
    if len(seq) < 1:
        return None
    elif len(seq) == 1:
        return seq[0]
    elif len(seq) == 2:
        return func(seq[0], seq[1])
    else:
        half = len(seq)/2
        return func(breduce(func, seq[:half]), breduce(func, seq[half:]))

def memoize(func, lbl=""):
    """Returns wrapped func that caches return values for specific args.
       All args must be immutable (e.g. tuples not lists).

Example:
    def factorial(n):
        if n > 1:
           return n * fast_fac(n-1)
        else:
           return 1
     fast_fac = memoize(factorial)
"""
    results = {}
    def wrapper(*args):
        if args in results:
            return results[args]
        else:
            result = func(*args)
            results[args] = result
            # print lbl, args, ':', result
            return result
    return wrapper

# all coords in [0,1]
def HSVtoRGB(h,s,v):
    h, s, v = map(float, (h, s, v))
    if h == 1.0: h = 0.0
    h *= 360

    if s == 0:
        return v,v,v
    
    sectorPos = h/60.0
    sectorNumber = int(floor(sectorPos))
    fractionalSector = sectorPos - sectorNumber

    p = v * (1 - s)
    q = v * (1 - (s * fractionalSector))
    t = v * (1 - (s * (1 - fractionalSector)))
    
    return {0 : (v, t, p),
            1 : (q, v, p),
            2 : (p, v, t),
            3 : (p, q, v),
            4 : (t, p, v),
            5 : (v, p, q)} [sectorNumber]

def RGBtoHSV(r,g,b):
    r, g, b = map(float, (r, g, b))
    mn = min(r, g, b)
    mx = max(r, g, b)
    v = mx
    delta = mx - mn
    if (mx == 0.0) or (delta == 0.0):
        return 0.0, 0.0, v
    s = delta / mx
    if r == mx:
        h = (g - b) / delta
    elif g == mx:
        h = 2.0 + (b - r) / delta
    else:
        h = 4.0 + (r - g) / delta
    h /= 6.0
    if h < 0: h += 360
    return h/360.0, s, v

def ScaleToColor(x, saturation=1.0): # x in [-1..1]
    """Returns r,g,b corresponding to x in [-1..1].
    The scale moves smoothly around the color wheel from pink (-) to red (+),
    with brightness proportional to abs(x)."""
    y = sin(x*pi/2)
    h = 0.44 * (1.0 - y) # leave off the last 12% of the wheel so 0 is distinguishable from 1
    v = abs(x)
    return HSVtoRGB(h, saturation, v*(2.0 - v))

