"""Unlimited undo/redo"""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

from qub.types import *

class UndoStack(object):
    """Stack of undo and redo actions.

properties:
    can_undo:  True if there is an undo action
    undo_lbl:  string describing the undo action
    can_redo:  True if there is a redo action
    redo_lbl:  string describing the redo action

remarks:
    An action consists of one or more callables (no arguments).
    When the user does something, you push one or more pairs of functions
    that undo and redo it, then seal the action with a label.

example:
    def Undo():
        print "undo"
    def Redo():
        print "redo"
    undos = UndoStack()
    undos.push_undo(Undo, Redo)
    undos.seal_undo('bogus test')
    undos.undo()
    undos.redo()
"""
    def __init__(self):
        self.OnChange = WeakEvent() # (undostack)
        self._working = False
        self.clear()
    def clear(self):
        """Resets the stack so there are no undo or redo actions."""
        if self._working: return
        self.uu = []
        self.rr = []
        self.OnChange(self)
    def get_can_undo(self):
        return bool(self.uu) # and isinstance(self.uu[-1], str))
    can_undo = property(get_can_undo)
    def get_undo_lbl(self):
        if self.can_undo:
            for i in xrange(1, len(self.uu)+1):
                if isinstance(self.uu[-i], str):
                    return self.uu[-i]
        return ""
    undo_lbl = property(get_undo_lbl)
    def get_can_redo(self):
        return bool(self.rr) # and isinstance(self.rr[-1], str))
    can_redo = property(get_can_redo)
    def get_redo_lbl(self):
        if self.can_redo:
            for i in xrange(1, len(self.rr)+1):
                if isinstance(self.rr[-i], str):
                    return self.rr[-i]
        return ""
    redo_lbl = property(get_redo_lbl)
    def push_undo(self, undo, redo):
        """Adds undo() and redo() to the stack."""
        if self._working: return
        self.uu.append( (undo, redo) )
    def seal_undo(self, lbl):
        """Groups the preceding calls to push_undo(), sets undo_lbl,
    clears the redo stack."""
        if self._working: return
        if isinstance(self.uu[-1], str):
            return
        self.uu.append( lbl )
        self.rr = []
        self.OnChange(self)
    def push_redo(self, redo, undo):
        if self._working: return
        self.rr.append( (redo, undo) )
    def seal_redo(self, lbl):
        if self._working: return
        self.rr.append( lbl )
        self.OnChange(self)
    def undo(self):
        """Calls all unsealed undo functions, and all between the last two seal_undo()s,
and transfers them to the redo stack."""
        if not self.can_undo: return
        #if self._working: return
        self._working = True
        # leftover/unsealed
        while self.uu and not isinstance(self.uu[-1], str):
            undo, redo = self.uu.pop()
            undo()
            self.rr.append( (redo, undo) )
        lbl = self.uu and self.uu.pop() or ""
        while self.uu and not isinstance(self.uu[-1], str):
            undo, redo = self.uu.pop()
            undo()
            self.rr.append( (redo, undo) )
        self.rr.append( lbl )
        self._working = False
        self.OnChange(self)
    def redo(self):
        """Does whatever was undone by undo(), transfers it back to the undo stack."""
        if not self.can_redo: return
        #if self._working: return
        self._working = True
        while self.rr and not isinstance(self.rr[-1], str):
            redo, undo = self.rr.pop()
            redo()
            self.uu.append( (undo, redo) )
        lbl = self.rr.pop()
        while self.rr and not isinstance(self.rr[-1], str):
            redo, undo = self.rr.pop()
            redo()
            self.uu.append( (undo, redo) )
        self.uu.append( lbl )
        self._working = False
        self.OnChange(self)
