/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifdef _USRDLL
  #define _QTR_PY_
#endif
#include "qubtree_node.h"

#include <strstream>
#include <stdio.h>
#include <fstream>

#if defined (_WIN32)
#include <windows.h>
#endif

PyMappingMethods qubtree_node_asMapping = {
  qubtree_node_length,
  qubtree_node_subscript,
  0 /* mp_ass_subscript */
};

PyTypeObject qubtree_node_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "QUBTree Node",
  sizeof(qubtree_node),
  0,
  qubtree_node_dealloc,
  0,          /*tp_print*/
  qubtree_node_getattr,
  qubtree_node_setattr,
  qubtree_node_compare,
  qubtree_node_repr,
  0,          /*tp_as_number*/
  0,          /*tp_as_sequence*/
  &qubtree_node_asMapping,
  0,          /*tp_hash */
  0,          /*tp_call */
  qubtree_node_str
};

PyMethodDef qubtree_module_methods[] = {
  // module-scope functions
  {"Node", qubtree_node_create, METH_VARARGS,				"Create a new Node."},
  {"NullNode", qubtree_node_createNull, METH_VARARGS,		"Make a Node object referring to no node."},
  {"CommentNode", qubtree_node_createComment, METH_VARARGS, "Create a new comment Node."},
  {"Open", qubtree_node_open, METH_VARARGS,					"open a binary file."},
  {"ReadText", qubtree_node_readText, METH_VARARGS,			"read a text tree."},
  {"ReadString", qubtree_node_readString, METH_VARARGS,		"reads a text tree from a string."},
  {"ReadBytes", qubtree_node_readBytes, METH_VARARGS,		"reads a binary tree from a string."},
  {NULL, NULL, 0, NULL}
};

PyMethodDef qubtree_node_methods[] = {
  // methods of node
  {"save", qubtree_node_save, METH_VARARGS,					"save changes if the tree is in an open file"},
  {"saveAs", qubtree_node_saveAs, METH_VARARGS,				"save a copy as a binary tree"},
  {"saveAsText", qubtree_node_saveAsText, METH_VARARGS,		"save as a text tree"},
  {"close", qubtree_node_close, METH_VARARGS,				"close the file, move the tree to memory"},
  {"clone", qubtree_node_clone, METH_VARARGS,				"create a copy of node"},
  {"list", qubtree_node_list, METH_VARARGS,					"get names of children"},
  {"find", qubtree_node_find, METH_VARARGS,					"get first named child if it exists"},
  {"next", qubtree_node_next, METH_VARARGS,					"get first named sibling if it exists"},
  {"nextSameName", qubtree_node_nextSameName, METH_VARARGS,	"get first sibling with same name if it exists"},
  {"append", qubtree_node_append, METH_VARARGS,				"append a child node"},
  {"appendClone", qubtree_node_appendClone, METH_VARARGS,   "append a clone as a child node"},
  {"insert", qubtree_node_insert, METH_VARARGS,				"insert a child node before the given node"},
  {"insertClone", qubtree_node_insertClone, METH_VARARGS,   "insert a clone as a child node before the given node"},
  {"remove", qubtree_node_remove, METH_VARARGS,				"remove a child node"},
  {"lock", qubtree_node_lock, METH_VARARGS,					"lock the fields of node against other threads"},
  {"unlock", qubtree_node_unlock, METH_VARARGS,				"unlock"},
  {"getBytes", qubtree_node_getBytes, METH_VARARGS,         "get the whole binary (sub)tree as a string"},
  {NULL, NULL, 0, NULL}
};


PySequenceMethods qubtree_data_asSequence = {
	(inquiry)qubtree_data_length,			/* sq_length */
	0,		/* sq_concat */
	0,		/* sq_repeat */
	(intargfunc)qubtree_data_item,			/* sq_item */
	(intintargfunc)qubtree_data_slice,		/* sq_slice */
	(intobjargproc)qubtree_data_ass_item,
	(intintobjargproc)qubtree_data_ass_slice,
	0,		/* sq_contains */
};

PyTypeObject qubtree_data_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "QUBTree data",
  sizeof(qubtree_data),
  0,
  qubtree_data_dealloc, /*tp_dealloc*/
  0,          /*tp_print*/
  qubtree_data_getattr,
  qubtree_data_setattr,
  0,          /*tp_compare*/
  qubtree_data_repr,
  0,          /*tp_as_number*/
  &qubtree_data_asSequence,
  0, /*&QUB_TreeMaps,*/
  0,          /*tp_hash */
  0,          /*tp_call */
  qubtree_data_str
};

PyMethodDef qubtree_data_methods[] = {
  {"resize", qubtree_data_resize, METH_VARARGS,			   "change # of elements"},
  {"clear", qubtree_data_clear, METH_VARARGS,			   "get rid of data"},
  {"setup", qubtree_data_setup, METH_VARARGS,			   "allocate space for data of a certain type and dimensions"},
  {"loadRows", qubtree_data_load, METH_VARARGS,			   "change loaded portion"},
  {"unloadRows", qubtree_data_unload, METH_VARARGS,		   "page data to disk"},
//  {"readRows", qubtree_data_readRows, METH_VARARGS,		   "get copy of some data even if unloaded"},
//  {"writeRows", qubtree_data_writeRows, METH_VARARGS,	   "modify a range of data even if unloaded"},
  {"row", qubtree_data_row, METH_VARARGS,				   "get a sequence object for one row of data"},
  {"col", qubtree_data_col, METH_VARARGS,				   "get a sequence object for one column of data"},
  {NULL, NULL, 0, NULL}
};


PySequenceMethods qubtree_dataRow_asSequence = {
	(inquiry)qubtree_dataRow_length,			/* sq_length */
	0,		/* sq_concat */
	0,		/* sq_repeat */
	(intargfunc)qubtree_dataRow_item,			/* sq_item */
	(intintargfunc)qubtree_dataRow_slice,		/* sq_slice */
	(intobjargproc)qubtree_dataRow_ass_item,
	(intintobjargproc)qubtree_dataRow_ass_slice,
	0,		/* sq_contains */
};

PyTypeObject qubtree_dataRow_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "QUBTree data row",
  sizeof(qubtree_dataRow),
  0,
  qubtree_dataRow_dealloc, /*tp_dealloc*/
  0,          /*tp_print*/
  0, // qubtree_dataRow_getattr,
  0, // qubtree_dataRow_setattr,
  0,          /*tp_compare*/
  qubtree_dataRow_repr,
  0,          /*tp_as_number*/
  &qubtree_dataRow_asSequence,
  0, /*&QUB_TreeMaps,*/
  0,          /*tp_hash */
  0,          /*tp_call */
  qubtree_dataRow_str
};

PyMethodDef qubtree_dataRow_methods[] = {
  {NULL, NULL, 0, NULL}
};


PySequenceMethods qubtree_dataCol_asSequence = {
	(inquiry)qubtree_dataCol_length,			/* sq_length */
	0,		/* sq_concat */
	0,		/* sq_repeat */
	(intargfunc)qubtree_dataCol_item,			/* sq_item */
	(intintargfunc)qubtree_dataCol_slice,		/* sq_slice */
	(intobjargproc)qubtree_dataCol_ass_item,
	(intintobjargproc)qubtree_dataCol_ass_slice,
	0,		/* sq_contains */
};

PyTypeObject qubtree_dataCol_type = {
  PyObject_HEAD_INIT(NULL)
  0,
  "QUBTree data column",
  sizeof(qubtree_dataCol),
  0,
  qubtree_dataCol_dealloc, /*tp_dealloc*/
  0,          /*tp_print*/
  0, // qubtree_dataCol_getattr,
  0, // qubtree_dataCol_setattr,
  0,          /*tp_compare*/
  qubtree_dataCol_repr,
  0,          /*tp_as_number*/
  &qubtree_dataCol_asSequence,
  0, /*&QUB_TreeMaps,*/
  0,          /*tp_hash */
  0,          /*tp_call */
  qubtree_dataCol_str
};

PyMethodDef qubtree_dataCol_methods[] = {
  {NULL, NULL, 0, NULL}
};



QTR_PY_API void init_qubtree(void) 
{
  qubtree_node_type.ob_type = &PyType_Type;
  qubtree_data_type.ob_type = &PyType_Type;
  qubtree_dataRow_type.ob_type = &PyType_Type;
  qubtree_dataCol_type.ob_type = &PyType_Type;
  
  PyObject *qubtree = Py_InitModule("_qubtree", qubtree_module_methods);
  
  // add datatype constants
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_EMPTY", QTR_TYPE_EMPTY );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_UNKNOWN", QTR_TYPE_UNKNOWN );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_POINTER", QTR_TYPE_POINTER );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_STRING", QTR_TYPE_STRING );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_UCHAR", QTR_TYPE_UCHAR );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_CHAR", QTR_TYPE_CHAR );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_USHORT", QTR_TYPE_USHORT );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_SHORT", QTR_TYPE_SHORT );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_UINT", QTR_TYPE_UINT );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_INT", QTR_TYPE_INT );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_ULONG", QTR_TYPE_ULONG );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_LONG", QTR_TYPE_LONG );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_FLOAT", QTR_TYPE_FLOAT );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_DOUBLE", QTR_TYPE_DOUBLE );
  PyModule_AddIntConstant( qubtree, "QTR_TYPE_LDOUBLE", QTR_TYPE_LDOUBLE );
}

QTR_PY_API QUB_Tree qubtree_node_obj(PyObject *node)
{ // works on qubtree_*
	return QUB_Tree( ((qubtree_node *) node)->impl );
}

QTR_PY_API QTR_Impl* qubtree_node_impl(PyObject *node)
{ // works on qubtree_*
	return ((qubtree_node *) node)->impl;
}

QTR_PY_API PyObject* qubtree_node_new(const char *name)
{
	qubtree_node *node = PyObject_New(qubtree_node, &qubtree_node_type);
	node->impl = QTR_Create( name );
	return (PyObject *) node;
}

QTR_PY_API PyObject* qubtree_node_using(QTR_Impl *impl)
{
  qubtree_node *node = PyObject_New(qubtree_node, &qubtree_node_type);
  node->impl = impl;
  if ( impl )
    QTR_INCREF( impl );
  return (PyObject *) node;
}

QTR_PY_API int       qubtree_node_check(PyObject *node)
{
	return ( node->ob_type == &qubtree_node_type );
}

void qubtree_node_dealloc(PyObject* self)
{
  qubtree_node *node = (qubtree_node *) self;
  if ( node->impl )
    QTR_DECREF( node->impl );
  PyObject_Del(self);
}

QTR_DataType qubtree_node_getSeqType(PyObject *seq)
{
  int len = PySequence_Size( seq );
  if ( len == 0 )
	  return QTR_TYPE_EMPTY;

  bool isNum = true;
  bool isInt = true;
  PyObject *elt;
  
  for ( int i=0; i<len; ++i ) {
    elt = PySequence_GetItem( seq, i );
    if ( PyInt_Check( elt ) ) {
      
    }
    else if ( PyFloat_Check( elt ) ) {
      isInt = false;
    }
    else {
      isNum = false;
    }
    Py_DECREF( elt );
  }
  
  if ( isInt )
    return QTR_TYPE_INT;
  else if ( isNum )
    return QTR_TYPE_DOUBLE;
  else
    return QTR_TYPE_UNKNOWN;
}

PyObject* qubtree_node_create(PyObject* self, PyObject* args)
{
  char *name = "";
  if (!PyArg_ParseTuple(args,"|s:Node", &name))
    return NULL;
  
  return qubtree_node_new(name);
}

PyObject* qubtree_node_createNull(PyObject* self, PyObject* args)
{
  return qubtree_node_using(NULL);
}

PyObject* qubtree_node_createComment(PyObject* self, PyObject* args)
{
  char *txt = "";
  if (!PyArg_ParseTuple(args,"|s:CommentNode", &txt))
    return NULL;
  
  QUB_Tree com = QUB_Tree::CreateComment( txt );
  return qubtree_node_using( com.getImpl() );
}

PyObject* qubtree_node_open(PyObject* self, PyObject* args)
{
	char *path;
	PyObject *roObj = 0;
	bool readOnly = false;

	if (!PyArg_ParseTuple(args,"s|O:Open", &path, &roObj))
		return NULL;
	if ( roObj )
		readOnly =(0!= PyObject_IsTrue( roObj ));

	QUB_Tree node;
	try {
	  node = QUB_Tree::Open( path, readOnly );
	}
	catch (...) {
	  node = QUB_Tree(); // null
	}
	return qubtree_node_using( node.getImpl() );
}


PyObject* qubtree_node_readText(PyObject* self, PyObject* args)
{
	char *path;
	if (!PyArg_ParseTuple(args,"s:ReadText", &path))
		return NULL;
  
	QUB_Tree node;
	try {
	  node = QUB_Tree::ReadText( path );
	}
	catch (...) {
	  node = QUB_Tree(); // null
	}
	return qubtree_node_using( node.getImpl() );
}

PyObject* qubtree_node_readString(PyObject* self, PyObject* args)
{
	char *s;
	if ( ! PyArg_ParseTuple(args, "s:ReadString", &s ) )
		return NULL;

	QUB_Tree node;
	try {
	  node = QUB_Tree::CreateFromString( s );
	}
	catch (...) {
	  node = QUB_Tree(); // null
	}
	return qubtree_node_using( node.getImpl() );
}

PyObject* qubtree_node_readBytes(PyObject* self, PyObject* args)
{
	PyObject* pyString;
	char*     bytes;

	if ( ! PyArg_ParseTuple(args, "O:ReadBytes", &pyString) )
		return NULL;
	if ( ! PyString_Check(pyString) ) {
		PyErr_SetString(PyExc_TypeError, "you can only clone a qubtree.Node");
		return NULL;
	}
	bytes = PyString_AsString(pyString);

	QUB_Tree node;
	try {
	  node = QUB_Tree::CreateFromBuffer(bytes);
	}
	catch (...) {
	  node = QUB_Tree(); // null
	}
	return qubtree_node_using( node.getImpl() );
}

PyObject* qubtree_node_save(PyObject* self, PyObject* args)
{
  try {
	return PyInt_FromLong( qubtree_node_obj(self).save() ? 1 : 0 );
  }
  catch (...) {
    return PyInt_FromLong( 0 );
  }
}

PyObject* qubtree_node_saveAs(PyObject* self, PyObject* args)
{
	char *path;
	if (!PyArg_ParseTuple(args,"s:saveAs", &path))
		return NULL;
	
	try {
	  return PyInt_FromLong( qubtree_node_obj(self).saveAs( path ) ? 1 : 0 );
	}
	catch (...) {
	  return PyInt_FromLong( 0 );
	}
}

PyObject* qubtree_node_saveAsText(PyObject* self, PyObject* args)
{
	char *path;
	if (!PyArg_ParseTuple(args,"s:saveAsText", &path))
		return NULL;

	try {  
	  return PyInt_FromLong( qubtree_node_obj(self).saveTextCopy( path ) ? 1 : 0 );
	}
	catch (...) {
	  return PyInt_FromLong( 0 );
	}
}

PyObject* qubtree_node_close(PyObject* self, PyObject* args)
{
        try {
	  return PyInt_FromLong( qubtree_node_obj(self).close() ? 1 : 0 );
	}
	catch (...) {
	  return PyInt_FromLong( 0 );
	}
}

PyObject *qubtree_node_repr(PyObject *self)
{
	QUB_Tree node = qubtree_node_obj(self);
	string name;
	if ( node.isNull() )
		name = "NULL";
	else
		name = node.name();

	ostrstream ost;
	ost << "<QUBTree Node : " << name << ">" << '\0';
	PyObject *reprStr = PyString_FromString( ost.str() );
	ost.rdbuf()->freeze(0);
	return reprStr;
}

PyObject *qubtree_node_str(PyObject *self)
{
	ostrstream ost;
	qubtree_node_obj(self).toStream(ost, "");
	ost << '\0';

	PyObject *str = PyString_FromString( ost.str() );
	ost.rdbuf()->freeze(0);
	return str;
}

int qubtree_node_compare(PyObject* v, PyObject* w)
{
  if ( qubtree_node_check( w ) ) {
    long l1 = (long) qubtree_node_impl( v );
    long l2 = (long) qubtree_node_impl( w );
    if ( l1 < l2 )
      return -1;
    else if ( l1 == l2 )
      return 0;
    else
      return 1;
  }
  else
    return -1;
}

PyObject* qubtree_node_clone(PyObject* self, PyObject* args)
{
  PyObject *deepArg = 0;
  if (!PyArg_ParseTuple(args,"|O:clone", &deepArg))
    return NULL;
  bool deep = deepArg ? (0!=PyObject_IsTrue( deepArg )) : true;
  
  return qubtree_node_using( qubtree_node_obj( self ).clone( deep ).getImpl() );
}

PyObject* qubtree_node_list(PyObject* self, PyObject* args)
{
	QTR_Impl *impl = qubtree_node_impl(self);
	PyObject *ls = PyList_New(0);

	if ( impl ) {
		for ( QTR_Impl *child = impl->child; child; child = child->sibling )
			PyList_Append( ls, PyString_FromString( QTR_Name( child ) ) );
	}

	return ls;
}

PyObject* qubtree_node_find(PyObject* self, PyObject* args)
{
	char *name;
	if (!PyArg_ParseTuple(args, "s:find", &name))
		return NULL;

	return qubtree_node_using( qubtree_node_obj(self).find( name )->getImpl() );
}

PyObject* qubtree_node_next(PyObject* self, PyObject* args)
{
	char *name;
	if (!PyArg_ParseTuple(args, "s:next", &name))
		return NULL;

	return qubtree_node_using( qubtree_node_obj(self).sibling( name ).getImpl() );
}

PyObject* qubtree_node_nextSameName(PyObject* self, PyObject* args)
{
	return qubtree_node_using( qubtree_node_obj(self).siblingSameName().getImpl() );
}


PyObject* qubtree_node_append(PyObject* self, PyObject* args)
{
	PyObject *ch;
	if (!PyArg_ParseTuple(args, "O:append", &ch))
		return NULL;
	
	if ( qubtree_node_check( ch ) ) {
		return qubtree_node_using( qubtree_node_obj(self).appendChild( qubtree_node_obj(ch) ).getImpl() );
	}
	else if ( PyString_Check( ch ) ) {
		return qubtree_node_using( qubtree_node_obj(self).appendChild( PyString_AsString(ch) ).getImpl());
	}
	else {
		PyErr_SetString(PyExc_TypeError, "you can only append a qubtree.Node or a string naming the new child");
		return NULL;
	}
}

PyObject* qubtree_node_appendClone(PyObject* self, PyObject* args)
{
	PyObject *orig, *deepArg = 0;
	if ( ! PyArg_ParseTuple(args, "O|O:appendClone", &orig, &deepArg) )
		return NULL;
	bool deep = deepArg ? (0!=PyObject_IsTrue( deepArg )) : true;

	if ( ! qubtree_node_check( orig ) ) {
		PyErr_SetString(PyExc_TypeError, "you can only clone a qubtree.Node");
		return NULL;
	}

	QUB_Tree clone = qubtree_node_obj(self).appendClone( qubtree_node_obj(orig), deep );
	return qubtree_node_using( clone.getImpl() );
}
		
PyObject* qubtree_node_insert(PyObject* self, PyObject* args)
{
	PyObject *ch;
	PyObject *after = 0;
	if (!PyArg_ParseTuple(args, "O|O:insert", &ch, &after))
		return NULL;
	if ( after && ! qubtree_node_check(after) ) {
		PyErr_SetString(PyExc_TypeError, "after (arg2) must be a qubtree.Node");
		return NULL;
	}
	QUB_Tree afterNode;
	if ( after )
		afterNode = qubtree_node_obj(after);

	if ( qubtree_node_check( ch ) ) {
		return qubtree_node_using( qubtree_node_obj(self).insertChild( afterNode, qubtree_node_obj(ch) ).getImpl() );
	}
	else if ( PyString_Check( ch ) ) {
		return qubtree_node_using( qubtree_node_obj(self).insertChild( afterNode, PyString_AsString(ch) ).getImpl() );
	}
	else {
		PyErr_SetString(PyExc_TypeError, "you can only insert a qubtree.Node or a string naming the new child");
		return NULL;
	}
}


PyObject* qubtree_node_insertClone(PyObject* self, PyObject* args)
{
	PyObject *orig, *deepArg = 0, *after = 0;
	if ( ! PyArg_ParseTuple(args, "O|OO:insertClone", &orig, &deepArg, &after) )
		return NULL;
	if ( after && ! qubtree_node_check(after) ) {
		PyErr_SetString(PyExc_TypeError, "after (arg3) must be a qubtree.Node");
		return NULL;
	}
	bool deep = deepArg ? (0!=PyObject_IsTrue( deepArg )) : true;
	QUB_Tree afterNode;
	if ( after )
		afterNode = qubtree_node_obj(after);

	if ( ! qubtree_node_check( orig ) ) {
		PyErr_SetString(PyExc_TypeError, "you can only clone a qubtree.Node");
		return NULL;
	}

	return qubtree_node_using( qubtree_node_obj(self).insertClone( afterNode, qubtree_node_obj(orig), deep ).getImpl() );
}

PyObject* qubtree_node_remove(PyObject* self, PyObject* args)
{
	PyObject *ch;
	if (!PyArg_ParseTuple(args, "O:remove", &ch))
		return NULL;
	if ( ! qubtree_node_check( ch ) ) {
		PyErr_SetString(PyExc_TypeError, "you can only remove a (child) qubtree.Node");
		return NULL;
	}

	qubtree_node_obj(self).removeChild( qubtree_node_obj(ch) );
	return Py_BuildValue("");
}
/*
PyObject* qubtree_node_setChanged(PyObject* self, PyObject* args)
{
  PyObject *isChanged = Py_None;
  if (!PyArg_ParseTuple(args,"|O:setChanged", &isChanged))
    return NULL;
  
  qubtree_node_obj( self ).setChanged( PyObject_IsTrue( isChanged ) );
  return Py_BuildValue("");
}
*/
PyObject* qubtree_node_lock(PyObject* self, PyObject* args)
{
  int timo = -1;
  if (!PyArg_ParseTuple(args,"|i:lock", &timo))
    return NULL;
  
  if ( ! qubtree_node_obj(self).lock(timo) ) {
    PyErr_SetString( PyExc_Exception, "couldn't get lock" );
    return NULL;
  }
  return Py_BuildValue("");
}

PyObject* qubtree_node_unlock(PyObject* self, PyObject* args)
{
  qubtree_node_obj(self).unlock();
  return Py_BuildValue("");
}

PyObject* qubtree_node_setData(PyObject* self, PyObject* args)
{
	QUB_Tree nobj = qubtree_node_obj( self );
	PyObject *val;
	if ( ! PyArg_ParseTuple( args, "O:setData", &val ) )
		return NULL;
  
	if ( nobj.isNull() )
		return Py_BuildValue("");
  
	if ( val == Py_None )
		nobj.clearData();
	else if ( PyString_Check( val ) )
		nobj.setData( PyString_AsString( val ) );
	else if ( PyInt_Check( val ) )
		nobj.setData( QTR_TYPE_INT, (int) PyInt_AsLong( val ) );
	else if ( PyFloat_Check( val ) )
		nobj.setData( QTR_TYPE_DOUBLE, PyFloat_AsDouble( val ) );
	else if ( PySequence_Check( val ) ) {
		// if seq of seqs, set as matrix (future)

		QTR_DataType dtype = qubtree_node_getSeqType( val );
		if ( dtype == QTR_TYPE_EMPTY ) {
			nobj.clearData();
		}
		else if ( dtype != QTR_TYPE_UNKNOWN ) {
			int len = PySequence_Length( val );
			if ( nobj.setNumData( dtype, len, 1, (void *) 0 ) ) {
				nobj.loadData();
				PyObject *data = qubtree_data_new( nobj.getImpl() );
				qubtree_data_ass_slice( (qubtree_data *) data, 0, len, val );
				Py_DECREF( data );
				return Py_BuildValue("");
			}
			else {
				PyErr_SetString( PyExc_TypeError, "couldn't setup data" );
				return NULL;
			}
		}
		else {
			PyErr_SetString( PyExc_TypeError, "unknown data type" );
			return NULL;
		}
	}
	else {
		PyErr_SetString( PyExc_TypeError, "unknown data type" );
		return NULL;
	}

	return Py_BuildValue("");
}


int qubtree_node_length(PyObject *self)
{
  qubtree_node *node = (qubtree_node *) self;
  int count = 0;
  if ( node->impl ) {
    QTR_Impl *ch = node->impl->child;
    while ( ch ) {
      ++count;
      ch = ch->sibling;
    }
  }
  return count;
}

PyObject *qubtree_node_subscript(PyObject *self, PyObject *key)
{
	QUB_Tree node = qubtree_node_obj(self);
	if ( PyString_Check( key ) ) {
		char *name = PyString_AsString( key );
		return qubtree_node_using( node[ name ].getImpl() );
	}
	else if ( PyInt_Check( key ) ) {
		int ix = (int) PyInt_AsLong( key );
		int at = 0;
		QUB_TreeMonoIter ci;
		for ( ci = node.children(); ! (*ci).isNull(); ++ci, ++at ) {
			if ( at == ix )
				break;
		}
		if ( (*ci).isNull() ) {
			PyErr_SetString( PyExc_IndexError, "index out of range" );
			return NULL;
		}
		return qubtree_node_using( (*ci).getImpl() );
	}
	else {
		PyErr_SetString( PyExc_TypeError, "choose a child by name or index only" );
		return NULL;
	}
}


PyObject *qubtree_node_getattr(PyObject *self, char *name)
{
	QUB_Tree node = qubtree_node_obj(self);
	if ( strcmp( name, "name" ) == 0 )
		return PyString_FromString( node.name().c_str() );
	else if ( strcmp( name, "lineComment" ) == 0 )
		return PyString_FromString( node.getLineComment().c_str() );
	else if ( strcmp( name, "parent" ) == 0 )
		return qubtree_node_using( node.parent().getImpl() );
	else if ( strcmp( name, "child" ) == 0 )
		return qubtree_node_using( node.child().getImpl() );
	else if ( strcmp( name, "sibling" ) == 0 )
		return qubtree_node_using( node.sibling().getImpl() );
	else if ( strcmp( name, "isNull" ) == 0 )
		return PyInt_FromLong( node.isNull() ? 1 : 0 );
	else if ( strcmp( name, "data" ) == 0 )
		return qubtree_data_new( node.getImpl() );
	else if ( strcmp( name, "refs" ) == 0 )
		return PyInt_FromLong( node.isNull() ? -1 : (node.getImpl()->refs - 1) );  // -1 for the ref held by "node"
	else if ( strcmp( name, "path" ) == 0 ) {
		if ( node.inFile() )
			return PyString_FromString( node.path().c_str() );
		else
			return Py_BuildValue("");
	}
	else if ( strcmp( name, "modified" ) == 0 )
		return PyInt_FromLong( ((node.flags() & QTR_FLAG_CHANGED) | (node.flags() & QTR_FLAG_CHILDCHANGED)) ? 1 : 0 );
		
  
	return Py_FindMethod(qubtree_node_methods, self, name);
}

int qubtree_node_setattr(PyObject *self, char *name, PyObject *val)
{
	QUB_Tree node = qubtree_node_obj(self);

	if ( strcmp( name, "name" ) == 0 ) {
		if ( PyString_Check( val ) ) {
			node.setName( PyString_AsString( val ) );
			return 0;
		}
		else {
			PyErr_SetString(PyExc_TypeError, "name must be a string");
			return -1;
		}
	}
    else if ( strcmp( name, "lineComment" ) == 0 ) {
		if ( PyString_Check( val ) ) {
			node.setLineComment( PyString_AsString( val ) );
			return 0;
		}
		else {
			PyErr_SetString(PyExc_TypeError, "line comment must be a string");
			return -1;
		}
	}
    else if ( strcmp( name, "data" ) == 0 ) {
		PyObject *args = PyTuple_New(1);
		Py_INCREF( val );
		PyTuple_SetItem( args, 0, val );
		PyObject *rslt = qubtree_node_setData( self, args );
		Py_DECREF( args );
		return ( rslt ? 0 : -1 );
    }

	PyErr_SetString(PyExc_TypeError, "can't set that");
	return -1;
}



PyObject* qubtree_node_getBytes(PyObject* self, PyObject* args)
{
	QUB_Tree node = qubtree_node_obj( self );
	unsigned int buffer_size = node.toBufferSize();
	PyObject *result = PyString_FromStringAndSize( NULL, buffer_size );
	node.toBuffer( PyString_AsString(result) );
	return result;
}


// Chris 6-16-2004 since node.data.row(0)[:] wasn't working
// However this is still not integrated, and I can't recall what I was thinking in data_{ass_}slice or column for that matter.
// (so it's only used in row_slice)

void adjust_slice(int& from, int& to, int len)
{
	if ( from < 0 )     from = len - from;   if ( to < 0 )     to = len - to;
	if ( from < 0 )     from = 0;            if ( to < 0 )     to = 0;
	if ( from > len )   from = len;          if ( to > len )   to = len;
}






PyObject *qubtree_data_new( QTR_Impl *impl )
{
  qubtree_data *data = PyObject_New(qubtree_data, &qubtree_data_type);
  data->impl = impl;
  if ( impl )
    QTR_INCREF( impl );
  return (PyObject *) data;
}

void qubtree_data_dealloc(PyObject* self)
{
  qubtree_data *data = (qubtree_data *) self;
  if ( data->impl )
    QTR_DECREF( data->impl );
  PyObject_Del(self);
}

PyObject *qubtree_data_repr(PyObject *self)
{
	QUB_Tree node = qubtree_node_obj(self);
	string name;
	if ( node.isNull() )
		name = "NULL";
	else
		name = node.name();
	int count  = node.isNull() ? 0      : node.dataCount();

	ostrstream ost;
	ost << "<QUBTree data for " << name << " (count " << count << ")>" << '\0';
	PyObject *reprStr = PyString_FromString( ost.str() );
	ost.rdbuf()->freeze(0);
	return reprStr;
}

PyObject *qubtree_data_str(PyObject *self)
{
	QUB_Tree node = qubtree_node_obj(self);
	return PyString_FromString( node.dataAsString().c_str() );
}

int qubtree_data_length(PyObject *self)
{
	return qubtree_node_obj(self).dataCount();
}

PyObject *qubtree_data_getattr(PyObject *self, char *name)
{
	QUB_Tree node = qubtree_node_obj(self);

	if ( strcmp( name, "type" ) == 0 )
		return PyInt_FromLong( node.dataType() );
	else if ( strcmp( name, "size" ) == 0 )
		return PyInt_FromLong( node.dataSize() );
	else if ( strcmp( name, "count" ) == 0 )
		return PyInt_FromLong( node.dataCount() );
	else if ( strcmp( name, "rows" ) == 0 )
		return PyInt_FromLong( node.dataRows() );
	else if ( strcmp( name, "cols" ) == 0 )
		return PyInt_FromLong( node.dataCols() );
	else if ( strcmp( name, "preload" ) == 0 )
		return PyInt_FromLong( node.isPreload() );
	else if ( strcmp( name, "node" ) == 0 )
		return qubtree_node_using( node.getImpl() );
	else if ( strcmp( name, "loaded" ) == 0 ) {
		PyObject *ld = PyTuple_New( 2 );
		PyTuple_SetItem( ld, 0, PyInt_FromLong( node.loadedFirstItem() ) );
		PyTuple_SetItem( ld, 1, PyInt_FromLong( node.loadedLastItem() ) );
		return ld;
	}
	else if ( strcmp( name, "loadedRows" ) == 0 ) {
		PyObject *ld = PyTuple_New( 2 );
		PyTuple_SetItem( ld, 0, PyInt_FromLong( node.loadedFirst() ) );
		PyTuple_SetItem( ld, 1, PyInt_FromLong( node.loadedLast() ) );
		return ld;
	}
  
	return Py_FindMethod(qubtree_data_methods, self, name);
}

int qubtree_data_setattr(PyObject *self, char *name, PyObject *val)
{
	QUB_Tree node = qubtree_node_obj(self);

	if ( node.isNull() ) {
		PyErr_SetString( PyExc_TypeError, "node is null" );
		return -1;
	}
	if ( strcmp( name, "preload" ) == 0 ) {
		node.setPreload( (0!=PyObject_IsTrue( val )) );
		return 0;
	}

	PyErr_SetString(PyExc_TypeError, "can't set that");
	return -1;
}

#ifdef _WIN32

// visual studio 6 can't handle the templates, so I have hand-inlined them

PyObject* qubtree_data_item( qubtree_data *ndata, int ix )
{
  QUB_Tree node( ndata->impl );
  QTR_DataType dtype = node.dataType();
  if ( dtype == QTR_TYPE_STRING ) {
    char *data = (char *) node.data( false );
    if ( data && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyString_FromStringAndSize( data + ix - node.loadedFirstItem(), 1 );
	else {
      PyErr_SetString(PyExc_IndexError, "index out of loaded range");
      return NULL;
	}
  }
  else if ( dtype == QTR_TYPE_UCHAR ) {
    unsigned char *ucdata = (unsigned char *) node.data( false );
    if ( ucdata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) ucdata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_CHAR ) {
    char *cdata = (char *) node.data( false );
    if ( cdata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) cdata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_USHORT ) {
    unsigned short *usdata = (unsigned short *) node.data( false );
    if ( usdata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) usdata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_SHORT ) {
    short *sdata = (short *) node.data( false );
    if ( sdata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) sdata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_UINT ) {
    unsigned int *uidata = (unsigned int *) node.data( false );
    if ( uidata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) uidata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_INT ) {
    int *idata = (int *) node.data( false );
    if ( idata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) idata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_ULONG ) {
    unsigned long *uldata = (unsigned long *) node.data( false );
    if ( uldata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) uldata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_LONG ) {
    long *ldata = (long *) node.data( false );
    if ( ldata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyInt_FromLong( (long) ldata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_FLOAT ) {
    float *fdata = (float *) node.data( false );
    if ( fdata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyFloat_FromDouble( (double) fdata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_DOUBLE ) {
    double *ddata = (double *) node.data( false );
    if ( ddata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyFloat_FromDouble( (double) ddata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_LDOUBLE ) {
    long double *lddata = (long double *) node.data( false );
    if ( lddata && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) )
      return PyFloat_FromDouble( (double) lddata[ ix - node.loadedFirstItem() ] );
    // else:
    PyErr_SetString(PyExc_IndexError, "index of out loaded range");
    return NULL;
  }
  else if ( dtype == QTR_TYPE_EMPTY ) {
    PyErr_SetString( PyExc_TypeError, "data is empty" );
    return NULL;
  }
  else {
    PyErr_SetString( PyExc_TypeError, "data type is not compatible with python" );
    return NULL;
  }
}

#else

template<class T>
PyObject *qtr_data_IntItemFrom( QUB_Tree node, int ix )
{
  T *data = (T *) node.data( false );
  if ( data && node.loadedFirstItem() <= ix && ix <= node.loadedLastItem() )
    return PyInt_FromLong( (long) data[ ix - node.loadedFirstItem() ] );
  
  PyErr_SetString(PyExc_IndexError, "index out of loaded range");
  return NULL;
}

template<class T>
PyObject *qtr_data_FloatItemFrom( QUB_Tree node, int ix )
{
  T *data = (T *) node.data( false );
  if ( data && node.loadedFirstItem() <= ix && ix <= node.loadedLastItem() )
    return PyFloat_FromDouble( (double) data[ ix - node.loadedFirstItem() ] );
  
  PyErr_SetString(PyExc_IndexError, "index out of loaded range");
  return NULL;
}

PyObject* qubtree_data_item( qubtree_data *ndata, int ix )
{
  QUB_Tree node( ndata->impl );
  QTR_DataType dtype = node.dataType();
  if ( dtype == QTR_TYPE_STRING ) {
    char *data = (char *) node.data( false );
    if ( data && node.loadedFirstItem() <= ix && ix <= node.loadedLastItem() )
      return PyString_FromStringAndSize( data + ix - node.loadedFirstItem(), 1 );
	else {
      PyErr_SetString(PyExc_IndexError, "index out of loaded range");
      return NULL;
	}
  }
  else if ( dtype == QTR_TYPE_UCHAR )
    return qtr_data_IntItemFrom<unsigned char>( node, ix );
  else if ( dtype == QTR_TYPE_CHAR )
    return qtr_data_IntItemFrom<char>( node, ix );
  else if ( dtype == QTR_TYPE_USHORT )
    return qtr_data_IntItemFrom<unsigned short>( node, ix );
  else if ( dtype == QTR_TYPE_SHORT )
    return qtr_data_IntItemFrom<short>( node, ix );
  else if ( dtype == QTR_TYPE_UINT )
    return qtr_data_IntItemFrom<unsigned int>( node, ix );
  else if ( dtype == QTR_TYPE_INT )
    return qtr_data_IntItemFrom<int>( node, ix );
  else if ( dtype == QTR_TYPE_ULONG )
    return qtr_data_IntItemFrom<unsigned long>( node, ix );
  else if ( dtype == QTR_TYPE_LONG )
    return qtr_data_IntItemFrom<long>( node, ix );
  else if ( dtype == QTR_TYPE_FLOAT )
    return qtr_data_FloatItemFrom<float>( node, ix );
  else if ( dtype == QTR_TYPE_DOUBLE )
    return qtr_data_FloatItemFrom<double>( node, ix );
  else if ( dtype == QTR_TYPE_LDOUBLE )
    return qtr_data_FloatItemFrom<long double>( node, ix );
  else if ( dtype == QTR_TYPE_EMPTY ) {
    PyErr_SetString( PyExc_TypeError, "data is empty" );
    return NULL;
  }
  else {
    PyErr_SetString( PyExc_TypeError, "data type is not compatible with python" );
    return NULL;
  }
}

#endif


PyObject* qubtree_data_slice( qubtree_data *ndata, int from, int until )
{
  QUB_Tree node( ndata->impl );

  if (from < int(node.loadedFirstItem()) )
    from = 0;
  else if (from > int(node.loadedLastItem()) + 1)
    from = node.loadedLastItem() + 1;
  if (until < from)
    until = from;
  else if (until > int(node.loadedLastItem()) + 1)
    until = node.loadedLastItem() + 1;
  
  if ( node.data(false) && int(node.loadedFirstItem()) <= from && (until-1) <= int(node.loadedLastItem()) ) {
    PyObject *lst = PyList_New( until - from );
    int ilst, idat;
    for ( ilst = 0, idat = from; idat < until; ++ilst, ++idat )
      PyList_SetItem( lst, ilst, qubtree_data_item( ndata, idat ) );
    return lst;
  }
  else {
    PyErr_SetString( PyExc_IndexError, "index out of loaded range" );
    return NULL;
  }
}


template<class T>
void QTR_SetDataItem( QUB_Tree node, int ix, T val ) {
  QTR_Impl *impl = node.getImpl();
  
  if ( impl && impl->data && int(node.loadedFirstItem()) <= ix 
		&& ix <= int(node.loadedLastItem()) ) {
    ix -= node.loadedFirstItem();
    switch ( node.dataType() ) {
      // case QTR_TYPE_UNKNOWN:
      //case QTR_TYPE_POINTER:
    case QTR_TYPE_STRING:
    case QTR_TYPE_CHAR:
      ((char *) impl->data)[ ix ] = (char) val;
      break;
    case QTR_TYPE_UCHAR:
      ((unsigned char *) impl->data)[ ix ] = (unsigned char) val;
      break;
    case QTR_TYPE_USHORT:
      ((unsigned short *) impl->data)[ ix ] = (unsigned short) val;
      break;
    case QTR_TYPE_SHORT:
      ((short *) impl->data)[ ix ] = (short) val;
      break;
    case QTR_TYPE_UINT:
      ((unsigned int *) impl->data)[ ix ] = (unsigned int) val;
      break;
    case QTR_TYPE_INT:
      ((int *) impl->data)[ ix ] = (int) val;
      break;
    case QTR_TYPE_ULONG:
      ((unsigned long *) impl->data)[ ix ] = (unsigned long) val;
      break;
    case QTR_TYPE_LONG:
      ((long *) impl->data)[ ix ] = (long) val;
      break;
    case QTR_TYPE_FLOAT:
      ((float *) impl->data)[ ix ] = (float) val;
      break;
    case QTR_TYPE_DOUBLE:
      ((double *) impl->data)[ ix ] = (double) val;
      break;
    case QTR_TYPE_LDOUBLE:
      ((long double *) impl->data)[ ix ] = (long double) val;
      break;

	  QTR_Changed(impl);
    }
  }
}


int qubtree_data_ass_item(qubtree_data *ndata, int ix, PyObject *val)
{
  QUB_Tree node( ndata->impl );

  if ( node.data(false) && int(node.loadedFirstItem()) <= ix && ix <= int(node.loadedLastItem()) ) {
    if ( PyInt_Check( val ) )
      QTR_SetDataItem( node, ix, PyInt_AsLong( val ) );
    else if ( PyFloat_Check( val ) )
      QTR_SetDataItem( node, ix, PyFloat_AsDouble( val ) );
    else if ( PyString_Check( val ) )
      QTR_SetDataItem( node, ix, PyString_AsString( val )[0] );
    else {
      PyErr_SetString( PyExc_TypeError, "not a number" );
      return -1;
    }
    return 0;
  }
  else {
    PyErr_SetString( PyExc_IndexError, "index out of loaded range" );
    return -1;
  }
}

int qubtree_data_ass_slice(qubtree_data *ndata, int from, int until, PyObject *val)
{
	if ( PySequence_Check( val ) ) {
		int len = PySequence_Length( val );
		if ( len == (until - from) ) {
			QUB_Tree node( ndata->impl );
			if ( node.data(false) && int(node.loadedFirstItem()) <= from && (until-1) <= int(node.loadedLastItem()) ) {
				bool success = true;
				for ( int i=from; i<until; ++i ) {
					PyObject *item = PySequence_GetItem( val, i - from );
					bool success = qubtree_data_ass_item( ndata, i, item ) >= 0;
					Py_DECREF( item );
					if ( ! success )
						break;
				}
				return ( success ? 0 : -1 );
			}
			else {
				PyErr_SetString( PyExc_IndexError, "index out of loaded range" );
				return -1;
			}
		}
		else {
			PyErr_SetString( PyExc_TypeError, "sequence must be same length as slice" );
			return -1;
		}
	}
	else {
		PyErr_SetString( PyExc_TypeError, "must set slice with a sequence object" );
		return -1;
	}
}

PyObject* qubtree_data_resize(PyObject *self, PyObject *args)
{
  int newcount;
  if ( ! PyArg_ParseTuple( args, "i:resize", &newcount ) )
    return NULL;
  
  if ( ! qubtree_node_obj( self ).resizeData( newcount ) ) {
    PyErr_SetString( PyExc_Exception, "couldn't resize data" );
    return NULL;
  }
  return Py_BuildValue("");
}

PyObject* qubtree_data_clear(PyObject *self, PyObject *args)
{
	qubtree_node_obj(self).clearData();
	return Py_BuildValue("");
}

PyObject* qubtree_data_load(PyObject *self, PyObject *args)
{
  QUB_Tree node = qubtree_node_obj( self );
  int first = 0;
  int last = node.dataRows() - 1;
  PyObject *doReadArg = 0;
  
  if ( ! PyArg_ParseTuple( args, "|iiO:load", &first, &last, &doReadArg ) )
    return NULL;
  
  bool doRead = doReadArg ? (0!=PyObject_IsTrue( doReadArg )) : true;
  
  if ( ! node.loadData( first, last, doRead ) ) {
    PyErr_SetString( PyExc_Exception, "couldn't load data" );
    return NULL;
  }
  return Py_BuildValue("");
}

PyObject* qubtree_data_unload(PyObject *self, PyObject *args)
{
  QUB_Tree node = qubtree_node_obj( self );
  PyObject *doWriteArg = 0;
  
  if ( ! PyArg_ParseTuple( args, "|O:unload", &doWriteArg ) )
    return NULL;

  bool doWrite = doWriteArg ? (0!=PyObject_IsTrue( doWriteArg )) : true;
  
  if ( ! node.unloadData( doWrite ) ) {
    PyErr_SetString( PyExc_Exception, "couldn't unload data" );
    return NULL;
  }
  return Py_BuildValue("");
}
/*
PyObject* qubtree_data_readRows(PyObject *self, PyObject *args)
{
}

PyObject* qubtree_data_WriteRows(PyObject *self, PyObject *args)
{
}
*/
PyObject* qubtree_data_setup(PyObject *self, PyObject *args)
{
  int type, nr=1, nc=1;
  if ( ! PyArg_ParseTuple( args, "i|ii:setup", &type, &nr, &nc ) )
    return NULL;

  QUB_Tree nobj = qubtree_node_obj( self );
  if ( ! nobj.setNumData(QTR_DataType(type), nr, nc, (void *) 0) ) {
    PyErr_SetString( PyExc_Exception, "couldn't setup data" );
    return NULL;
  }
  else
	return Py_BuildValue("");
}

PyObject* qubtree_data_row(PyObject *self, PyObject *args)
{
	int ix;
	if ( ! PyArg_ParseTuple( args, "i:row", &ix ) )
		return NULL;

	QUB_Tree node = qubtree_node_obj(self);

	if ( (ix < 0) || (ix >= int(node.dataRows())) ) {
		PyErr_SetString( PyExc_IndexError, "no such row" );
		return NULL;
	}

	return qubtree_dataRow_new( qubtree_node_impl(self), ix );
}

PyObject* qubtree_data_col(PyObject *self, PyObject *args)
{
	int ix;
	if ( ! PyArg_ParseTuple( args, "i:col", &ix ) )
		return NULL;

	QUB_Tree node = qubtree_node_obj(self);

	if ( (ix < 0) || (ix >= int(node.dataCols())) ) {
		PyErr_SetString( PyExc_IndexError, "no such column" );
		return NULL;
	}

	return qubtree_dataCol_new( qubtree_node_impl(self), ix );
}



PyObject *qubtree_dataRow_new( QTR_Impl *impl, int ix )
{
  qubtree_dataRow *row = PyObject_New(qubtree_dataRow, &qubtree_dataRow_type);
  row->impl = impl;
  row->row = ix;
  if ( impl )
    QTR_INCREF( impl );
  return (PyObject *) row;
}

void qubtree_dataRow_dealloc(PyObject* self)
{
  qubtree_dataRow *row = (qubtree_dataRow *) self;
  if ( row->impl )
    QTR_DECREF( row->impl );
  PyObject_Del(self);
}

PyObject *qubtree_dataRow_repr(PyObject *self)
{
	qubtree_dataRow *row = (qubtree_dataRow *) self;
	QUB_Tree node = qubtree_node_obj(self);
	string name;
	if ( node.isNull() )
		name = "NULL";
	else
		name = node.name();

	ostrstream ost;
	ost << "<QUBTree data for " << name << ", row " << row->row << ">" << '\0';
	PyObject *reprStr = PyString_FromString( ost.str() );
	ost.rdbuf()->freeze(0);
	return reprStr;
}

PyObject *qubtree_dataRow_str(PyObject *self)
{
	QUB_Tree node = qubtree_node_obj(self);
	return PyString_FromString( node.dataAsString().c_str() ); /// fix so it prints just this row
}

int qubtree_dataRow_length(PyObject *self)
{
	return qubtree_node_obj(self).dataCols();
}

PyObject* qubtree_dataRow_item( PyObject *row, int ix )
{
	int rowoff = ((qubtree_dataRow *) row)->row * qubtree_node_obj(row).dataCols();
	return qubtree_data_item( (qubtree_data *) row, rowoff + ix );
}

PyObject* qubtree_dataRow_slice( PyObject *row, int from, int until )
{
	int rowoff = ((qubtree_dataRow *) row)->row * qubtree_node_obj(row).dataCols();
	adjust_slice( from, until, qubtree_node_obj(row).dataCols() );
	return qubtree_data_slice( (qubtree_data *) row, rowoff + from, rowoff + until );
}

int qubtree_dataRow_ass_item(PyObject *row, int ix, PyObject *val)
{
	int rowoff = ((qubtree_dataRow *) row)->row * qubtree_node_obj(row).dataCols();
	return qubtree_data_ass_item( (qubtree_data *) row, rowoff + ix, val );
}
	
int qubtree_dataRow_ass_slice(PyObject *row, int from, int until, PyObject *val)
{
	PyErr_SetString( PyExc_IndexError, "slice assignment to rows not implemented" );
	return -1;
	// int rowoff = ((qubtree_dataRow *) row)->row * qubtree_node_obj(row).dataCols();
	// return qubtree_data_ass_slice( (qubtree_data *) row, rowoff + from, rowoff + until, val );
}


PyObject *qubtree_dataCol_new( QTR_Impl *impl, int ix )
{
  qubtree_dataCol *col = PyObject_New(qubtree_dataCol, &qubtree_dataCol_type);
  col->impl = impl;
  col->col = ix;
  if ( impl )
    QTR_INCREF( impl );
  return (PyObject *) col;
}

void qubtree_dataCol_dealloc(PyObject* self)
{
  qubtree_dataCol *col = (qubtree_dataCol *) self;
  if ( col->impl )
    QTR_DECREF( col->impl );
  PyObject_Del(self);
}

PyObject *qubtree_dataCol_repr(PyObject *self)
{
	qubtree_dataCol *col = (qubtree_dataCol *) self;
	QUB_Tree node = qubtree_node_obj(self);
	string name;
	if ( node.isNull() )
		name = "NULL";
	else
		name = node.name();

	ostrstream ost;
	ost << "<QUBTree data for " << name << ", column " << col->col << ">" << '\0';
	PyObject *reprStr = PyString_FromString( ost.str() );
	ost.rdbuf()->freeze(0);
	return reprStr;
}

PyObject *qubtree_dataCol_str(PyObject *self)
{
	QUB_Tree node = qubtree_node_obj(self);
	return PyString_FromString( node.dataAsString().c_str() ); /// fix so it prints just this column
}

int qubtree_dataCol_length(PyObject *self)
{
	return qubtree_node_obj(self).dataRows();
}

PyObject* qubtree_dataCol_item( PyObject *self, int ix )
{
	qubtree_dataCol *col = (qubtree_dataCol *) self;
	return qubtree_data_item( (qubtree_data *) col, ix * qubtree_node_obj(self).dataCols() + col->col );
}

PyObject* qubtree_dataCol_slice( PyObject *self, int from, int until )
{
	qubtree_dataCol *col = (qubtree_dataCol *) self;
	QUB_Tree node = qubtree_node_obj( self );
	int ncol = node.dataCols();

	if (from < int(node.loadedFirst()))
		from = 0;
	else if (from > int(node.loadedLast()) + 1)
		from = node.loadedLast() + 1;
	if (until < from)
		until = from;
	else if (until > int(node.loadedLast()) + 1)
		until = node.loadedLast() + 1;
  
	if ( node.data(false) && int(node.loadedFirst()) <= from && (until-1) <= int(node.loadedLast()) ) {
		PyObject *tup = PyTuple_New( until - from );
		int itup, idat;
		for ( itup = 0, idat = from; idat < until; ++itup, ++idat )
			PyTuple_SetItem( tup, itup, qubtree_data_item( (qubtree_data *) self, idat * ncol + col->col ) );
		return tup;
	}
	else {
		PyErr_SetString( PyExc_IndexError, "index out of loaded range" );
		return NULL;
	}
}

int qubtree_dataCol_ass_item(PyObject *self, int ix, PyObject *val)
{
	qubtree_dataCol *col = (qubtree_dataCol *) self;
	return qubtree_data_ass_item( (qubtree_data *) col, ix * qubtree_node_obj(self).dataCols() + col->col, val );
}

int qubtree_dataCol_ass_slice(PyObject *self, int from, int until, PyObject *val)
{
	qubtree_dataCol *col = (qubtree_dataCol *) self;
	QUB_Tree node = qubtree_node_obj( self );
	int ncol = node.dataCols();

	if ( PySequence_Check( val ) ) {
		int len = PySequence_Length( val );
		if ( len == (until - from) ) {
			if ( node.data(false) && int(node.loadedFirst()) <= from && (until-1) <= int(node.loadedLast()) ) {
				bool success = true;
				for ( int i=from; i<until; ++i ) {
					PyObject *item = PySequence_GetItem( val, i - from );
					bool success = qubtree_data_ass_item( (qubtree_data *) self, i * ncol + col->col, item ) >= 0;
					Py_DECREF( item );
					if ( ! success )
						break;
				}
				return ( success ? 0 : -1 );
			}
			else {
				PyErr_SetString( PyExc_IndexError, "index out of loaded range" );
				return -1;
			}
		}
		else {
			PyErr_SetString( PyExc_TypeError, "sequence must be same length as slice" );
			return -1;
		}
	}
	else {
		PyErr_SetString( PyExc_TypeError, "must set slice with a sequence object" );
		return -1;
	}
}

