/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifndef MAX_INTER_LL_H
#define MAX_INTER_LL_H

#include "maxill.h"

/*
  begin_html
  <h3>Maximum Interval Likelihood  (Qin et al, 1996)</h3>
  <p>
  This function computes the forward log-likelihood of data given a model.
  </p><p>
 See also:
 <ul>
 <li><a href="max_inter_ll.cpp.html">max_inter_ll.cpp</a></li>
 <li><a href="max_ll_util.h.html">max_ll_util.h</a></li>
 <li><a href="max_subi_ll.h.html">max_subi_ll.h</a></li>
 </ul>
 <b>Up:</b> <a href="index.html">Index</a>
  </p>
  
  <h4>The Model</h4>
  <p>
  Our Markov model is a graph with colored vertices.  A vertex is called a "state,"
  and its color is a nonnegative integer called its "class."  States in the same
  class are indistinguishable (same amp).  To describe the vertices, provide the array
  </p><pre>
  int clazz[Ns] = [class of each state]
  int Ns        = number of states
  </pre><p>
  Each edge is labeled with its transition rate (probability per second).  These form the matrix
  </p><p>
  Q, a Ns x Ns matrix with<br>
  \(Q_{a,b}\) = rate from state a to state b<br>
  \(Q_{a,a} = - \sum_i Q_{a,i}\) where \(i \neq a\)
  </p><p>
  Each \(Q_{a,b} = K0_{a,b} * Ligand_{a,b} * e^{K1_{a,b} * Voltage_{a,b}}\).  You provide the Ns x Ns matrices
  </p><pre>
  double **K0, **K1            of kinetic parameters
  int    **Ligand, **Voltage   index of the ligand or voltage signal influencing each rate, or 0
  </pre><p>
  with the diagonals undefined.
  </p><p>
  A pair of states is either connected in both directions or neither.  To indicate un-connectedness, set
  </p><p>
  \(K0_{a,b} = K0_{b,a} = 0\)
  </pre>
  <p>
  For constant state entry probabilities, as described in (Qin...1996), provide the array
  </p><pre>
  int P0[Ns]
  </pre><p>
  For equilibrium entry probabilities, as described in (Milescu...2005), provide
  </p><pre>
  P0 = NULL
  </pre>
  
  <h4>The Data</h4>
  <p>
  For data, you provide one or more idealized segments.  A segment is described by
  </p><pre>
  int   DwellCount            number of events
  int   Classes[DwellCount]   class index of each event
  float Durations[DwellCount] length of each dwell in seconds, less tdead
  </pre><p>
  and you give the segments together as
  </p><pre>
  int   Nseg                  number of segments
  int   **dwellCounts
  int   **classeses
  float **durationses
  </pre><p>
  tdead is the longest duration of events that can't be reliably detected.
  MIL assumes you have deleted any such events, by merging them with their prior.
  Also, for computational reasons, tdead should be subtracted from each event.
  We provide a utility which merges and shortens events, in-place:
  </p>
  end_html
*/

extern "C" MAXILL_API void __stdcall mil_prep_events( int *dwellCount, int *classes, float *durations, double tdead_ms );
extern "C" MAXILL_API void __stdcall mil_prep_segments( int Nseg, int *dwellCounts, int **classeses, float **durationses, double tdead_ms );

/*
  begin_html
  <p>
  If any \(Ligand_{a,b}\ \neq 0\) or \(Voltage_{a,b}\ \neq 0\), you must provide an array
  </p><pre>
  double *Constants
  </pre><p>
  where e.g. if \(Ligand_{a,b} = 2\) then Constants[2] holds the ligand concentration.
  We assume all segments were recorded at the same constant conditions.
  For global fitting, call max_inter_ll separately for each dataset, with different constants, and sum the LL.
  </p>
  <p><b>double *LL:</b> upon return, contains log(prob. that model generated this data)</p>
  <p><b>Returns:</b> 0 on success, error codes to be defined.</p>
  
  <h4>Limitations</h4>
  <ul>
  <li>no gradients</li>
  <li>no maximization</li>
  <li>no support for multiple identical models (channel count > 1)</li>
  </ul>
  
  end_html
*/

extern "C" MAXILL_API int __stdcall inter_ll_arr(
						 int Ns, int *clazz, double *P0_in,  double *K0_in, double *K1_in,
						 int *Ligand_in, int *Voltage_in,
						 double *Constants, double tdead,
						 int Nseg, int *dwellCounts, int **classeses, float **durationses,
						 double *ll);

extern "C" MAXILL_API int __stdcall inter_ll(
					     int Ns, int *clazz, double *P0,  double **K0, double **K1,
					     int **Ligand, int **Voltage,
					     double *Constants, double tdead,
					     int Nseg, int *dwellCounts, int **classeses, float **durationses,
					     double *LL);


#endif

