/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifndef QUB_CURVE_DECEXPO_H
#define QUB_CURVE_DECEXPO_H

#include "QUB_BasicCurve.h"

typedef struct {
	double a, b, Tau, Slope;
} DecExpoComponent;
static DecExpoComponent DECEXPO_DEFAULTS = {1.0, 0.0, 1.0, 0.0};
#define DECEXPO_NPAR 4
static char DECEXPO_PARAM_NAMES[][MAX_NAMELEN] = {"a", "b", "Tau", "Slope"};

class QUB_Curve_DecExpo : public QUB_BasicCurve<DecExpoComponent, DECEXPO_NPAR>
{
public:
	QUB_Curve_DecExpo(const char *name, IQUB_Curve *baselineCurve)
		: QUB_BasicCurve<DecExpoComponent, DECEXPO_NPAR>(
				name,
				"a + (b-a)*(1 - e^(-(x-x0) / Tau)) + Slope*(x-x0)",
				&(DECEXPO_PARAM_NAMES[0][0]),
				DECEXPO_DEFAULTS,
				baselineCurve )
	{}
	virtual void evalOne(QF_Data *data, double *yy, int c)
	{
		double a = components[c].a;
		double bMinusA = components[c].b - a;
		double ti = 1.0 / components[c].Tau;
		double Slope = components[c].Slope;
		double mx0 = - getX0();
		int n = data->n;
		double *xx = data->xx;
		for (int i=0; i<n; ++i)
			yy[i] += a + bMinusA*(1.0 - exp(- (xx[i]+mx0) * ti)) + Slope*(xx[i]+mx0);
	}
	virtual void __cdecl setComponentCount(int x) {/*disable*/}
};

typedef struct {
	double a, b, c, Tau1, Tau2, Slope;
} DblDecExpoComponent;
static DblDecExpoComponent DBLDECEXPO_DEFAULTS = {1.0, 0.5, 0.0, 1.0, 10.0, 0.0};
#define DBLDECEXPO_NPAR 6
static char DBLDECEXPO_PARAM_NAMES[][MAX_NAMELEN] = {"a", "b", "c", "Tau1", "Tau2", "Slope"};

class QUB_Curve_DblDecExpo : public QUB_BasicCurve<DblDecExpoComponent, DBLDECEXPO_NPAR>
{
public:
	QUB_Curve_DblDecExpo(const char *name, IQUB_Curve *baselineCurve)
		: QUB_BasicCurve<DblDecExpoComponent, DBLDECEXPO_NPAR>(
				name,
				"a + (b-c)*(1 - e^(-(x-x0) / Tau1)) + c*(1-e^(-(x-x0) / Tau2)) + Slope*(x-x0)",
				&(DBLDECEXPO_PARAM_NAMES[0][0]),
				DBLDECEXPO_DEFAULTS,
				baselineCurve )
	{}
	virtual void evalOne(QF_Data *data, double *yy, int ci)
	{
		double a = components[ci].a;
		double c = components[ci].c;
		double bMinusC = components[ci].b - c;
		double ti1 = 1.0 / components[ci].Tau1;
		double ti2 = 1.0 / components[ci].Tau2;
		double mSlope = - components[ci].Slope;
		double mx0 = - getX0();
		int n = data->n;
		double *xx = data->xx;
		for (int i=0; i<n; ++i) {
			double mx = - (xx[i]+mx0);
			yy[i] += a + bMinusC*(1.0 - exp(mx*ti1)) + c*(1.0 - exp(mx*ti2)) + mSlope*mx;
		}
	}
	virtual void __cdecl setComponentCount(int x) {/*disable*/}
};

// make your own factory function with this signature,
// returning a new instance of your class:

// vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
IQUB_Curve* __stdcall CreateDecExpo();
IQUB_Curve* __stdcall CreateDblDecExpo();
// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

#endif
